// ===========================================================================
//	LDynamicArray.h					1994 Metrowerks Inc. All rights reserved.
// ===========================================================================
//
//	An ordered collection of fixed-size items. Positions in the list are
//	one-based--the first item is at index 1.

#pragma once

#include "UException.h"
typedef short int	Int16 ;
typedef long			Int32 ;
typedef unsigned long	Uint32 ;
typedef short int		ResIDT ;

class	LDynamicArray {

public:
					LDynamicArray(Int32 inItemSize);
					LDynamicArray(Int32 inItemSize, Handle inItemsHandle);
	virtual			~LDynamicArray();
	
	Int32			GetItemSize() const;
	
	virtual Int32	GetCount() const;

	virtual void	InsertItemsAt(Uint32 inCount, Int32 inAtIndex,
							const void *inItem);
	virtual void	RemoveItemsAt(Uint32 inCount, Int32 inAtIndex);
	
	virtual Boolean	FetchItemAt(Int32 inAtIndex, void *outItem) const;
	virtual void	SetItemAt(Int32 inAtIndex, const void *inItem);
	
	virtual void	SwapItems(Int32 inIndexA, Int32 inIndexB);
	virtual void	MoveItem(Int32 inFromIndex, Int32 inToIndex);
	
	virtual Int32	FetchIndexOf(const void *inItem) const;
	virtual void	Remove(const void *inItem);
							
protected:
	Int32			mItemSize;
	Int32			mItemCount;
	Handle			mItemsH;
	
	virtual void*	GetItemPtr(Int32 inAtIndex) const;
	virtual void	PeekItem(Int32 inAtIndex, void *outItem) const;
	virtual void	PokeItem(Int32 inAtIndex, const void *inItem);

	virtual void	ShiftItems(Int32 inStartPos, Int32 inEndPos,
							Int32 inDestPos);
};


inline Int32
LDynamicArray::GetItemSize() const
{
	return mItemSize;
}


const	Int32	arrayIndex_Bad		= 0;
const	Int32	arrayIndex_First	= 1;
const	Int32	arrayIndex_Last		= 0x7FFFFFFF;
	
