// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include "globals.h"
#include "XDrivers.h"
#include "TDefineContentDialog.h"

TContentAtom::TContentAtom() {
	strcpy(mSymbol,"Si") ;
	mDW = 0.25 ; //LDM
	mOccupancy = 1.0 ;
	mNumInCell = 1 ;
}


TAtomEntry::TAtomEntry() {
	mNextAtomEntry = nil ;
	mAtom = nil ;
}

TAtomEntry::~TAtomEntry() {
	mNextAtomEntry = nil ;
	if (mAtom != nil)
		delete mAtom ;
}


TAtomList::TAtomList() {
	mNumAtoms = 0 ;
	mAtomEntry = nil ;
}

TAtomList::~TAtomList() {
	TAtomEntry *entry ;
	TAtomEntry *nextEntry ;
	
	// For every object entry
	entry = mAtomEntry ;
	while(entry != nil) {
		nextEntry = entry->mNextAtomEntry ;
		delete entry ;
		entry = nextEntry ;
		mNumAtoms-- ;
	} ;
}

void TAtomList::AddAtom(TContentAtom *theAtom) {
	
	TAtomEntry *aNewEntry = new TAtomEntry ;
	aNewEntry->mAtom = theAtom ;
	
	// We save time by always adding to the beginning of the list
	aNewEntry->mNextAtomEntry = mAtomEntry ;
	mAtomEntry = aNewEntry ;
	mNumAtoms++ ;
}


void TAtomList::RemoveAtom(short num) {
	TAtomEntry* entry;
	TAtomEntry* last;
	int i ;
	
	last = nil;
	short numToRemove = mNumAtoms - num - 1 ;
	for (i = 0,entry = mAtomEntry ;entry != nil; entry = entry->mNextAtomEntry,i++) {
		if ( i == numToRemove ) {			
			// is the current atom the one to be removed?
			// yes, remove it
			// if first item in list, just set first
				if ( last == nil )				
					mAtomEntry = entry->mNextAtomEntry;
				else
					last->mNextAtomEntry = entry->mNextAtomEntry ;
			
			// free the TAtomEntry
				delete entry ;
				mNumAtoms-- ;
				
			// we're done; no need to continue onto the next object
				return;
		}
		else						// no, go onto the next object in the list
			last = entry;
	}
}

TContentAtom *TAtomList::GetAtom(int number) {
	TAtomEntry* entry;
	int i,numToReturn ;
	
	if (number >= mNumAtoms || number < 0)
		return nil ;
	
	numToReturn = 	mNumAtoms - number - 1 ;	// Because of the sequence of entries
	for (i = 0,entry = mAtomEntry ;entry != nil; entry = entry->mNextAtomEntry,i++) {
		if (i == numToReturn) {
			return entry->mAtom ;
		}
	}
	return nil ;
}

TContentList::TContentList(Rect	*theListRect) 
	: TList(theListRect,sizeof(TContentAtom)) {
		mHasColLabels = true ;
}

void 	TContentList::DrawCellValue(ListCellT &inCell,Rect &cellFrame) {
	char str[256] ;
	TContentAtom	theAtom ;
	short	width = (mCanvasRect.right-mCanvasRect.left -8)/4 ;
	
	TextSize(0);
	FetchCellData(inCell,&theAtom) ;
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	sprintf(str,"%s",theAtom.mSymbol) ;
	DrawString(str) ;
	
	MoveTo(cellFrame.left + width + 4, cellFrame.bottom - 4);
	sprintf(str,"%.3f",theAtom.mDW) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 2*width + 4, cellFrame.bottom - 4);
	sprintf(str,"%.2f",theAtom.mOccupancy) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 3*width + 4, cellFrame.bottom - 4);
	sprintf(str,"%d",theAtom.mNumInCell) ;
	DrawString(str) ;
}

void 	TContentList::DrawColLabel(short colNumber) {
	short	width = (mCanvasRect.right-mCanvasRect.left -8)/4 ;
	
	Rect	cellFrame;
	
	if (mHasRowLabels)
		SetRect(&cellFrame,colNumber*mColWidth,0,(colNumber+1)*mColWidth,mRowHeight) ;
	else
		SetRect(&cellFrame,(colNumber-1)*mColWidth,0,colNumber*mColWidth,mRowHeight) ;
		
	if (colNumber != 0) 
		::OffsetRect(&cellFrame,near(mListOrigin.x),0) ;
	InsetRect(&cellFrame,1,1) ;
	SetFgColor(gGrayPixel) ;
	PaintRect(&cellFrame) ;
	SetFgColor(gDarkerGray) ;
	MoveTo(cellFrame.left,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.top) ;
	SetFgColor(gBlackPixel) ;
	MoveTo(cellFrame.right,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.bottom) ;
	InsetRect(&cellFrame,-1,-1) ;
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	DrawString("Symbol") ;
	
	MoveTo(cellFrame.left + width + 4, cellFrame.bottom - 4);
	DrawString("DW") ;

	MoveTo(cellFrame.left + 2*width + 4, cellFrame.bottom - 4);
	DrawString("Occ") ;

	MoveTo(cellFrame.left + 3*width + 4, cellFrame.bottom - 4);
	DrawString("#atoms") ;
}

TDefineContentDialog::TDefineContentDialog(TAtomList *atomList)
	: TDialog(kDialogPosX,kDialogPosY,380,230,"Specify Unit Cell Content") {

	Rect		theRect ;
	short		width,i ;
	TContentAtom		*theAtom ;
	
	width = mPortRect.right - mPortRect.left ;
		
	Rect		theListRect ;
	SetRect(&theListRect,16,16,234,128) ;
	mContentList = new TContentList(&theListRect) ;
	// Construct the content of the list
	for(i=0; i < atomList->GetNumberOfAtoms(); i++) {
		theAtom = atomList->GetAtom(i) ;
		mContentList->AddRow(theAtom) ;
	}
	AddObject(mContentList) ;	
	
	theRect = theListRect ;

	mDeleteAtom = new TButton(theRect.right+20,theRect.top,width-16,theRect.top+20,"Remove") ;	
	mDeleteAtom->Disable() ;
	AddObject(mDeleteAtom) ;

	mModifyAtom = new TButton(theRect.right+20,theRect.bottom-20,width-16,theRect.bottom,"Modify") ;	
	mModifyAtom->Disable() ;
	AddObject(mModifyAtom) ;
	
	mAddAtom = new TButton(theRect.right+20,theRect.bottom+30,width-16,theRect.bottom+50,"Add To List") ;	
	AddObject(mAddAtom) ;
	
	width = (theRect.right - theRect.left)/4 - 15 ;
	
	mChemicalSymbol = new TEditText(theRect.left,theRect.bottom+30,theRect.left+width,theRect.bottom+50,
			"Symbol","Si") ;			
	mChemicalSymbol->SetLabelPosition(kLabelOutsideTop) ;
	
	mDebyeWallerfactor = new TEditText(theRect.left+width+5,theRect.bottom+30,theRect.left+width+5+width,theRect.bottom+50,
			"DW","0.25") ; //LDM
	mDebyeWallerfactor->SetLabelPosition(kLabelOutsideTop) ;
	
	mOccupancyFactor = new TEditText(theRect.left+width+10+width,theRect.bottom+30,theRect.left+width+10+width+width,theRect.bottom+50,
			"Occ","1.0") ;
	mOccupancyFactor->SetLabelPosition(kLabelOutsideTop) ;
	
	mNumberOfAtoms = new TEditText(theRect.left+width+15+width+width,theRect.bottom+30,theRect.left+width+15+width+width+width,theRect.bottom+50,
			"#atoms","1") ;
	mNumberOfAtoms->SetLabelPosition(kLabelOutsideTop) ;
			

	theRect = mAddAtom->GetRect() ;
	mOKButton = new TButton(theRect.left,mPortRect.bottom-40,theRect.right,mPortRect.bottom-20,"OK");
	AddObject(mOKButton) ;
	mCancelButton = new TButton(theRect.left-110,mPortRect.bottom-40,theRect.left-20,mPortRect.bottom-20,"Cancel");
	AddObject(mCancelButton) ;

	// Add text edit objects in reverse order to get the tab order correct
	AddTextEditObject(mNumberOfAtoms) ;
	AddTextEditObject(mOccupancyFactor) ;
	AddTextEditObject(mDebyeWallerfactor) ;
	AddTextEditObject(mChemicalSymbol) ;
	mDefaultButton = mOKButton ;
}

void	TDefineContentDialog::DoMouseDownInObject(TObject *obj,Point *pt) {
	if (pt != nil)
		if (!obj->DoMouseDown(pt)) return ;
	char	text[256] ;
	mIsVisible = true ;
	if (obj == mOKButton) {
		mDone = true ;
		mDialogReturnCode = kDialogOK ;
	}
	else if (obj == mCancelButton) {
		mDone = true ;
		mDialogReturnCode = kDialogCancel ;
	}
	else if (obj == mAddAtom) {
		AddAtom() ;
	}
	else if (obj == mDeleteAtom) {
		DeleteAtom() ;
	}
	else if (obj == mModifyAtom) {
		ModifyAtom() ;
	}
	else if (obj == mContentList) {
		if (mContentList->HasSelection()) {
			// Put the data into the edit fields
			// Get all the defined atom entries
			TContentAtom		theAtom ;
			ListCellT	theCell = mContentList->GetSelectedCell() ;
			mContentList->FetchCellData(theCell,&theAtom) ;
			sprintf(text,"%s",theAtom.mSymbol) ;
			mChemicalSymbol->SetEditText(text) ;
			
			sprintf(text,"%.2f",theAtom.mDW) ;
			mDebyeWallerfactor->SetEditText(text) ;
			
			sprintf(text,"%.2f",theAtom.mOccupancy) ;
			mOccupancyFactor->SetEditText(text) ;
			
			sprintf(text,"%d",theAtom.mNumInCell) ;
			mNumberOfAtoms->SetEditText(text) ;
			
			mModifyAtom->Enable() ; 
			mDeleteAtom->Enable() ;
		}
		else {
			mModifyAtom->Disable() ; 
			mDeleteAtom->Disable() ;
		}
	}
}

void	TDefineContentDialog::AddAtom() {
	// Add the content to the list from the defined fields
	TContentAtom 	theAtom ;
	char	text[256] ;
	
	mChemicalSymbol->GetEditText(theAtom.mSymbol) ;
	mDebyeWallerfactor->GetEditText(text) ;
	theAtom.mDW = atof(text) ;
	mOccupancyFactor->GetEditText(text) ;
	theAtom.mOccupancy = atof(text) ;
	mNumberOfAtoms->GetEditText(text) ;
	theAtom.mNumInCell = atoi(text) ;
	
	mContentList->AddRow(&theAtom) ;
	mAddAtom->Disable() ;

	
}

void	TDefineContentDialog::DeleteAtom() {
	mContentList->RemoveRow(mContentList->GetSelectedRow()) ;
}

void	TDefineContentDialog::ModifyAtom() {
	// Add the content to the list from the defined fields
	TContentAtom 	theAtom ;
	char	text[256] ;
	
	mChemicalSymbol->GetEditText(theAtom.mSymbol) ;
	mDebyeWallerfactor->GetEditText(text) ;
	theAtom.mDW = atof(text) ;
	mOccupancyFactor->GetEditText(text) ;
	theAtom.mOccupancy = atof(text) ;
	mNumberOfAtoms->GetEditText(text) ;
	theAtom.mNumInCell = atoi(text) ;
		
	mContentList->SetCellData(mContentList->GetSelectedCell(),&theAtom) ;
	mContentList->Refresh() ;
}

void	TDefineContentDialog::DoKeyDownInObject(TObject *obj) {
	if (obj == mChemicalSymbol || obj == mDebyeWallerfactor || obj == mOccupancyFactor || 
			obj == mNumberOfAtoms) {
		mAddAtom->Enable() ;
	}
}

TAtomList	*TDefineContentDialog::GetAtomData() {
	// Get all the defined atom entries
	TContentAtom		*theAtom ;
	TAtomList	*atomList = nil ;
	ListCellT	theCell = {1,1} ;
	
	short numDefinedEntries = mContentList->GetNumberOfListItems() ;
	if (numDefinedEntries > 0) {
		atomList = new TAtomList ;
		for(int i = 0 ; i < numDefinedEntries; i++) {
			theAtom = new TContentAtom ;
			mContentList->FetchCellData(theCell,theAtom) ;
			atomList->AddAtom(theAtom) ;
			theCell.row++ ;	
		}	
	}
	return atomList ;
}
