// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------

#include "TImage.h"
#include "TImageWindow.h"
#include "tiff.h"
#include "TFile.h"
#include "XDrivers.h"

#include "TWindow.h"
#include "TImportRawImageDialog.h"

TImage *CreateImageFromTiffFile(char *filename) ;
extern	short gNumGrayLevels ;
extern	class TDMApplicationWindow *gTheControlWindow ;

int IsPowerOfTwo(long value) ;
#define log2(x)	log(x)/log(2.0)

typedef enum {kFileIsText,kFileIsBinary}
ImportFileType ;

static ImageDataType	gImageDataType = k4ByteReal;
static Boolean			gSwapBytes = false ;
static ImportFileType	gImportFileType = kFileIsBinary ;
static long				gByteOffset =  0 ;
long				gImportImageFileSize ;
short				gImportImageWidth=128 ;
short				gImportImageHeight=128 ;
short				gImportImageOffset=0 ;
extern  TiffRecord		gTiff ;
long	near(float) ;
Point	gImageDisplayPosition = {240,0} ;

extern	void	AddWindow(TWindow *window) ;
extern	void	PutMessage(char *) ;

TImage::TImage(TArray &array,Rect *selectionRect) {


    if (selectionRect != nil) {
	    mWidth = selectionRect->right - selectionRect->left ;
	    mHeight = selectionRect->bottom - selectionRect->top ;
	}
	else {
	    mWidth = array.GetWidth() ;
	    mHeight = array.GetHeight() ;
	}



	mSignature = 'imag' ;
	mGamma = 90.0 ;
	mContrast 	= 1.0 ;
	mBrightness	= 0.5 ;
	mDisplayData = nil ;
	
	mOffscreen = nil ;
	mIsVisible = false ;
	mIsDirty = false ;
	mIsRealSpace = !array.IsComplex() ;
	mDataIsComplex = array.IsComplex() ;
	mWindow = nil ;
	mRGBData = nil ;

	mDisplayOnLogScale = mIsRealSpace ? false : true ;

	mSamplingX = 1. ;
	mSamplingY = 1. ;
	mCalibration = 1. ;
	mCalibrationUnit = kPixel ;

	mData = new TArray(array,kInheritType,selectionRect) ;
	mIsDirty = false ;

	mMinValue = kUnassigned ;
	mMaxValue = kUnassigned ;
	mMeanValue = kUnassigned ;
	mStdDev = kUnassigned ;

}

TImage::TImage(const TImage &u,Rect *rect) {

	mSignature = 'imag' ;
	if (rect == nil) {
		mWidth = u.mWidth ;
		mHeight = u.mHeight ;
		mData = new TArray(u.mData) ;
		if (u.mRGBData != nil) {
			mRGBData = new unsigned char[mWidth*mHeight*4]  ;
			if (mRGBData != nil)
				memcpy(mRGBData,u.mRGBData,mWidth*mHeight*4) ;
		}
		else 
			mRGBData = nil ;

		mMinValue = u.mMinValue ;
		mMaxValue = u.mMaxValue ;
                //LDM: I don't understand this
		//mMeanValue = u.mMinValue ;
		//mStdDev = u.mMaxValue ;
                mMeanValue = u.mMeanValue ;
                mStdDev = u.mStdDev ;
	}
	else {
		mWidth = rect->right - rect->left ;
		mHeight = rect->bottom - rect->top ;
		mData = new TArray(u.mData,kInheritType,rect) ;
		mMinValue = kUnassigned ;
		mMaxValue = kUnassigned ;
		mMeanValue = kUnassigned ;
		mStdDev = kUnassigned ;
		if (u.mRGBData != nil) {
			mRGBData = new unsigned char[mWidth*mHeight*4]  ;
			if (mRGBData != nil) {
				long Top,Left,i ;
				Top = rect->top ;
				Left = rect->left ;
				for(i=0; i < mHeight; i++) {
					memcpy(mRGBData+i*mWidth*4,
						u.mRGBData + 4*((Top+i)*u.mWidth + Left),mWidth*4) ;
				}
			}
		}
		else 
			mRGBData = nil ;
	}
	mGamma = u.mGamma ;
	mName = u.mName ;
	mSamplingX = u.mSamplingX ;
	mSamplingY = u.mSamplingY ;
	mOffscreen = nil ;
	mWindow = nil ;	
	mContrast 	= 1.0 ;
	mBrightness	= 0.5 ;
	mDisplayData = nil ;
	
	mCalibration = u.mCalibration ;
	mCalibrationUnit = u.mCalibrationUnit ;

	mIsVisible = false ;
	mIsDirty = false ;
	mIsRealSpace = u.mIsRealSpace ;	
	mDataIsComplex = u.mDataIsComplex ;

	mDisplayOnLogScale = mIsRealSpace ? false : true ;
}


TImage::TImage(short w,short h,float *data ,float angle,ImageDataType dataType) 
	: mWidth(w), mHeight(h), mGamma(angle) {

	long		numElements ;
	register	long i ;
	float		*p1 ;
	
	mSignature = 'imag' ;	
	mOffscreen = nil ;
	mIsVisible = false ;
	mIsDirty = false ;
	mIsRealSpace = true ;
	mDataIsComplex = false ;
	mWindow = nil ;
	mRGBData = nil ;
	mContrast 	= 1.0 ;
	mBrightness	= 0.5 ;
	mDisplayData = nil ;

	mSamplingX = 1. ;
	mSamplingY = 1. ;
	mCalibration = 1. ;
	mCalibrationUnit = kPixel ;

	// We will convert the image to be real (or complex) 
	if (dataType == k4ByteComplex || dataType == k8ByteComplex) {
		mDataIsComplex = true ;
		mData = new TArray(mWidth,mHeight,kComplexArray) ;
	}
	else {
		mData = new TArray(mWidth,mHeight) ;
	}
	mIsDirty = false ;
	mIsRealSpace = true ;

	mDisplayOnLogScale = mIsRealSpace ? false : true ;

	mMinValue = kUnassigned ;
	mMaxValue = kUnassigned ;
	mMeanValue = kUnassigned ;
	mStdDev = kUnassigned ;

	if (data == nil) return ;

	numElements = mWidth*mHeight ;

	switch (dataType) {
		case k1ByteSigned:
			char *cp ;
			for(i=0,cp=(char *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *cp++ ;
			}
			break ;
		case k1ByteUnsigned:
			unsigned char *ucp ;
			for(i=0,ucp=(unsigned char *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *ucp++ ;
			}
			break ;
		case k2ByteSigned:
			short *sp ;
			for(i=0,sp=(short *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *sp++ ;
			}
			break ;
		case k2ByteUnsigned:
			unsigned short *usp ;
			for(i=0,usp=(unsigned short *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *usp++ ;
			}
			break ;
		case k4ByteSigned:
			long *lp ;
			for(i=0,lp=( long *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *lp++ ;
			}
			break ;
		case k4ByteUnsigned:
			unsigned long *ulp ;
			for(i=0,ulp=(unsigned long *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *ulp++ ;
			}
			break ;
		case k4ByteReal:
			memcpy(mData->GetData(),data,numElements*sizeof(float)) ;
			break ;
		case k8ByteReal:
			double *dp ;
			for(i=0,dp=(double *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *dp++ ;
			}
			break ;
		case k4ByteComplex:
			memcpy(mData->GetData(),data,2*numElements*sizeof(float)) ;
			break ;
		case k8ByteComplex:
			double *ddp ;
			for(i=0,ddp=(double *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *ddp++ ;
				*p1++ = *ddp++ ;
			}
			break ;
	}
}

TImage::TImage(TArray *array,float angle) {

	mSignature = 'imag' ;
	mWidth = array->GetWidth() ;
	mHeight = array->GetHeight() ;
	mGamma = angle ;
	mContrast 	= 1.0 ;
	mBrightness	= 0.5 ;
	mDisplayData = nil ;
	
	mOffscreen = nil ;
	mIsVisible = false ;
	mIsDirty = false ;
	mIsRealSpace = !array->IsComplex() ;
	mDataIsComplex = array->IsComplex() ;
	mWindow = nil ;
	mRGBData = nil ;

	mSamplingX = 1. ;
	mSamplingY = 1. ;
	mCalibration = 1. ;
	mCalibrationUnit = kPixel ;

	mData = new TArray(array) ;
	mIsDirty = false ;

	mMinValue = kUnassigned ;
	mMaxValue = kUnassigned ;
	mMeanValue = kUnassigned ;
	mStdDev = kUnassigned ;

	mDisplayOnLogScale = mIsRealSpace ? false : true ;
}

TImage::TImage(TFile &theFile) {
	mSignature = 'imag' ;
	mContrast 	= 1.0 ;
	mBrightness	= 0.5 ;
	mDisplayData = nil ;
	mCalibration = 1. ;
	mCalibrationUnit = kPixel ;
}


TImage::~TImage() {
	if(mOffscreen != nil) delete mOffscreen ;
	if(mData != nil) delete mData;
	if (mRGBData != nil) delete[] mRGBData ;
	if (mDisplayData != nil) delete[] mDisplayData ;

}

void	TImage::Blur() {
	TArray	*tempArray = mData ;
	register	long row,col ;
	short		numPixels,left,top,right,bottom,k,l ;

	ArrayDataType dataType = tempArray->IsComplex() ? kComplexArray : kRealArray ;
	mData = new TArray(mWidth,mHeight,dataType) ;
	
	float *p1,sum ;
	Cmplx *c1,csum ;
	
	if (dataType == kRealArray) {
		p1 = mData->GetData() ;
		for(row=0;row<mHeight;row++) {
			if (row == 0) 
				top = 0 ;
			else 
				top = row-1 ;
			if (row == mHeight-1)
				bottom = mHeight-1 ;
			else
				bottom = row+1 ;
			for(col=0;col<mWidth;col++) {
				if (col == 0) 
					left = 0 ;
				else 
					left = col-1 ;
				if (col == mWidth-1)
					right = mWidth-1 ;
				else
					right = col+1 ;
				numPixels = 0 ;
				sum = 2*(*tempArray->GetData(col,row)) ;
				for(k=top;k<=bottom;k++)
					for(l=left;l<=right;l++) {
						if (k != row && l != col) {
							sum += *tempArray->GetData(l,k) ;
							numPixels++ ;
						}
					}
				*p1++ = sum/(2+numPixels) ;
			}
		}
	}
	else {
		c1 = (Cmplx *)mData->GetData() ;
		for(row=0;row<mHeight;row++) {
			if (row == 0) 
				top = 0 ;
			else 
				top = row-1 ;
			if (row == mHeight-1)
				bottom = mHeight-1 ;
			else
				bottom = row+1 ;
			for(col=0;col<mWidth;col++) {
				if (col == 0) 
					left = 0 ;
				else 
					left = col-1 ;
				if (col == mWidth-1)
					right = mWidth-1 ;
				else
					right = col+1 ;
				numPixels = 0 ;
				csum = 2*(*(Cmplx *)tempArray->GetData(col,row)) ;
				for(k=top;k<=bottom;k++)
					for(l=left;l<=right;l++) {
						if (k != row && l != col) {
							csum += *(Cmplx *)tempArray->GetData(l,k) ;
							numPixels++ ;
						}
					}
				*c1++ = csum/(2+numPixels) ;
			}
		}
	}
	delete tempArray ;
	ZeroStatistics() ; 
	Update() ;
}

void	TImage::Sharpen() {
	TArray	*tempArray = mData ;
	register	long row,col ;
	short		numPixels,left,top,right,bottom,k,l ;

	ArrayDataType dataType = tempArray->IsComplex() ? kComplexArray : kRealArray ;
	mData = new TArray(mWidth,mHeight,dataType) ;
	
	float *p1,sum ;
	Cmplx *c1,csum ;
	
	if (dataType == kRealArray) {
		p1 = mData->GetData() ;
		for(row=0;row<mHeight;row++) {
			if (row == 0) 
				top = 0 ;
			else 
				top = row-1 ;
			if (row == mHeight-1)
				bottom = mHeight-1 ;
			else
				bottom = row+1 ;
			for(col=0;col<mWidth;col++) {
				if (col == 0) 
					left = 0 ;
				else 
					left = col-1 ;
				if (col == mWidth-1)
					right = mWidth-1 ;
				else
					right = col+1 ;
				sum = 0. ;
				numPixels = 0 ;
				for(k=top;k<=bottom;k++)
					for(l=left;l<=right;l++) {
						if (k != row && l != col) {
							numPixels++ ;
							sum += *tempArray->GetData(l,k) ;
						}
					}
				*p1++ = (numPixels+1)*(*tempArray->GetData(col,row)) - sum ;
			}
		}
	}
	else {
		c1 = (Cmplx *)mData->GetData() ;
		for(row=0;row<mHeight;row++) {
			if (row == 0) 
				top = 0 ;
			else 
				top = row-1 ;
			if (row == mHeight-1)
				bottom = mHeight-1 ;
			else
				bottom = row+1 ;
			for(col=0;col<mWidth;col++) {
				if (col == 0) 
					left = 0 ;
				else 
					left = col-1 ;
				if (col == mWidth-1)
					right = mWidth-1 ;
				else
					right = col+1 ;
				csum = 0. ;
				numPixels = 0 ;
				for(k=top;k<=bottom;k++)
					for(l=left;l<=right;l++) {
						if (k != row && l != col) {
							numPixels++ ;
							csum += *(Cmplx *)tempArray->GetData(l,k) ;
						}
					}
				*c1++ = (numPixels+1)*(*(Cmplx *)tempArray->GetData(col,row)) - csum ;
			}
		}
	}
	delete tempArray ;	
	ZeroStatistics() ; 
	Update() ;
}

void TImage::SetData(float *data,ImageDataType dataType) {

	long		numElements ;
	register	long i ;
	float		*p1 ;

	mMinValue = kUnassigned ;
	mMaxValue = kUnassigned ;
	mMeanValue = kUnassigned ;
	mStdDev = kUnassigned ;

	if (mData != nil) delete mData ;
	// We will convert the image to be real (or complex) 
	if (dataType == k4ByteComplex || dataType == k8ByteComplex) {
		mDataIsComplex = true ;
		mData = new TArray(mWidth,mHeight,kComplexArray) ;
	}
	else {
		mData = new TArray(mWidth,mHeight) ;
	}
	numElements = mWidth*mHeight ;
	switch (dataType) {
		case k1ByteSigned:
			char *cp ;
			for(i=0,cp=(char *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *cp++ ;
			}
			break ;
		case k1ByteUnsigned:
			unsigned char *ucp ;
			for(i=0,ucp=(unsigned char *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *ucp++ ;
			}
			break ;
		case k2ByteSigned:
			short *sp ;
			for(i=0,sp=(short *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *sp++ ;
			}
			break ;
		case k2ByteUnsigned:
			unsigned short *usp ;
			for(i=0,usp=(unsigned short *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *usp++ ;
			}
			break ;
		case k4ByteSigned:
			long *lp ;
			for(i=0,lp=( long *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *lp++ ;
			}
			break ;
		case k4ByteUnsigned:
			unsigned long *ulp ;
			for(i=0,ulp=(unsigned long *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *ulp++ ;
			}
			break ;
		case k4ByteReal:
			memcpy(mData->GetData(),data,numElements*sizeof(float)) ;
			break ;
		case k8ByteReal:
			double *dp ;
			for(i=0,dp=(double *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *dp++ ;
			}
			break ;
		case k4ByteComplex:
			memcpy(mData->GetData(),data,2*numElements*sizeof(float)) ;
			break ;
		case k8ByteComplex:
			double *ddp ;
			for(i=0,ddp=(double *)data,p1 = mData->GetData(); i < numElements;i++) {
				*p1++ = *ddp++ ;
				*p1++ = *ddp++ ;
			}
			break ;
	}

}

TImage&	TImage::operator=(const float v) {
	mMinValue = kUnassigned ;
	mMaxValue = kUnassigned ;
	mMeanValue = kUnassigned ;
	mStdDev = kUnassigned ;
	*this->mData = v ;
        return *this ;
}

TImage&	TImage::operator=(const Cmplx& v) {
	mMinValue = kUnassigned ;
	mMaxValue = kUnassigned ;
	mMeanValue = kUnassigned ;
	mStdDev = kUnassigned ;
	*this->mData = v ;
        return *this ;
}

TImage&	TImage::operator=(const TImage& u) {
	this->mWidth = u.mWidth ;
	this->mHeight = u.mHeight ;
	this->mGamma = u.mGamma ;
	*this->mData = u.mData ;
	this->mOffscreen = nil ;
	this->mWindow = nil ;	

	this->mMinValue = u.mMinValue ;
	this->mMaxValue = u.mMaxValue ;
        //LDM: This looks wrong
	//this->mMeanValue = u.mMinValue ;
        this->mMeanValue = u.mMeanValue ;
	//this->mStdDev = u.mMaxValue ;
        this->mStdDev = u.mStdDev ;
	this->mIsVisible ;
	this->mIsDirty = false ;
	this->mIsRealSpace = u.mIsRealSpace ;	
	this->mDataIsComplex = u.mDataIsComplex ;

	return *this ;
}

void TImage::CreateOffscreen() {
	
	float	minVal,maxVal,*p1,value,range ;
	int i,j ;
	unsigned char *ucp ;
	extern unsigned long gLut[] ;
	GrafPtr	tPort ;
	TArray	*theDisplayArray ;

	GetPort(&tPort) ;

	Rect	theRect ;
	SetRect(&theRect,0,0,mWidth,mHeight) ;
	

	if (mWindow != nil) {
		mOffscreen = new TOffscreen(mWindow,&theRect) ;
	}
	else {
		mOffscreen = new TOffscreen(tPort,&theRect) ;
	}

	if (mOffscreen == nil)  {
		fprintf(stderr,"\nmOffscreen == nil Memory Allocation Error") ; 
		fflush(stderr) ;
		return ;
	}

	if (mRGBData != nil && mIsRealSpace) {
		SetPort(mOffscreen) ;
		DrawRGBImage(mRGBData,0,0,
							mWidth,mHeight) ;
		SetPort(tPort) ;	
		return ;				
	}

	long	width = mData->GetWidth() ;
	long	height = mData->GetHeight() ;
	
	if (!mIsRealSpace) {
		mDisplayOnLogScale = true ;
	}
	else {
		// Check min/max along the cross wire and set to log scale if the dynamic range is very large
		minVal = 1.0e25 ;
		maxVal = -1.0e25 ;
		p1 = mData->GetData() ;
		double	sum = 0. ;
		long	count = 0 ;
		p1 = mData->GetData() + height/2*width ;
		for(i=0;i<width;i++) {
			value = *p1++ ;
			if (value < minVal)  minVal = value ;
			if (value > maxVal)  maxVal = value ;
		}
		p1 = mData->GetData() + width/2 ;
		for(i=0;i<height;i++) {
			value = *p1 ; 
			p1 += width ;
			if (value < minVal)  minVal = value ;
			if (value > maxVal)  maxVal = value ;
		}
		range = maxVal - minVal ;
		if (range > 10000.) mDisplayOnLogScale = true ;
	}
	
	if (mDisplayOnLogScale) {
		// We need to convert the data to displayable
		// "diffraction" type data
		theDisplayArray = new TArray(mData) ;
		if (!mIsRealSpace) {
			theDisplayArray->Intensity() ;
		}
		*theDisplayArray += 1.0 - theDisplayArray->Min() ;	// Zero + 1
		theDisplayArray->log() ;
	}
	else {		
		theDisplayArray = mData ;
	}	
	minVal = 1.0e25 ;
	maxVal = -1.0e25 ;
	p1 = theDisplayArray->GetData() ;
	double	sum = 0. ;
	long	count = 0 ;
	for(i=0;i<width*height;i++) {
		value = *p1++ ;
		if (value > 0.1) {
			sum += value ;
			count++ ;
		}
		if (value < minVal)  minVal = value ;
		if (value > maxVal)  maxVal = value ;
	}
	float meanValue = 0. ;
	if (count > 0) meanValue = sum/count ;
	mMinValue = minVal ;
	mMaxValue = maxVal ;
	
			
	p1 = theDisplayArray->GetData() ;

	if (mDisplayData != nil) delete[] mDisplayData ;
	mDisplayData = new unsigned char[width*height] ;
	unsigned char *mappedDisplayData = new unsigned char[mWidth*mHeight] ;
	ucp = mDisplayData ;
	unsigned char *ucp1 = mappedDisplayData ;
	int index ;

	if (mIsRealSpace) {
		if (!mDisplayOnLogScale) {
			mMinDisplayLimit = mMinValue ;
			mMaxDisplayLimit = mMaxValue ;
			range = mMaxDisplayLimit - mMinDisplayLimit ;
			if (range == 0.) range = 1.0 ;	// Trap divide by zero
			for(i=0;i<height;i++) {
				for(j=0;j<width;j++) {
					value = (gNumGrayLevels-1)*((*p1)- mMinDisplayLimit)/range ;
					p1++ ;
					index = (int)value ;
					*ucp++ = (unsigned char)index ;				
					*ucp1++ = (unsigned char)gLut[index] ;
				}
			}
		}
		else {
			mMinDisplayLimit = meanValue ;
			mMaxDisplayLimit = mMaxValue ;
			range = mMaxDisplayLimit - mMinDisplayLimit ;
			if (range == 0.) range = 1.0 ;	// Trap divide by zero
			for(i=0;i<height;i++) {
				for(j=0;j<width;j++) {
					value = (gNumGrayLevels-1)*((*p1)- mMinDisplayLimit)/range ;
					if (value < 0.) value = 0. ;
					p1++ ;
					index = (int)value ;
					*ucp++ = (unsigned char)index ;				
					*ucp1++ = (unsigned char)gLut[index] ;
				}
			}
		}
	}
	else {
		long h,k,halfWidth = width/2,halfHeight=height/2 ;
		mMinDisplayLimit = meanValue ;
		mMaxDisplayLimit = mMaxValue ;
		range = mMaxDisplayLimit - mMinDisplayLimit ;
		if (range == 0.) range = 1.0 ;	// Trap divide by zero
		for(i=0;i<height;i++) {
			k = -halfHeight + i ;
			for(j=0;j<width;j++) {
				h = -halfWidth + j ;
				value = *theDisplayArray->GetDataHK(h,k) ;
				value = (gNumGrayLevels-1)*(value- mMinDisplayLimit)/range ;
				if (value < 0.) value = 0. ;
				index = (int)value ;
				*ucp++ = (unsigned char)index ;				
				*ucp1++ = (unsigned char)gLut[index] ;
			}
		}
	}

	if (theDisplayArray != mData) delete theDisplayArray ;
	SetPort(mOffscreen) ;
	DrawImage(mappedDisplayData,0,0,mWidth,mHeight) ;
	delete[] mappedDisplayData ;
		
	SetPort(tPort) ;
}

void	TImage::SetContrastBrightness(float contrast,float brightness) {
	GrafPtr	tPort ;
	extern unsigned long gLut[] ;

	mContrast 	= contrast ;
	mBrightness	= brightness ;

	GetPort(&tPort) ;
	SetPort(mOffscreen) ;
	
	if (fabs(mContrast-1.0) < 0.01) mContrast = 1.0 ;
	if (fabs(mBrightness-0.5) < 0.001) mBrightness = 0.5 ;
	
	// Map the display data to new contrast and brightness levels
	unsigned char *mappedDisplayData = new unsigned char[mWidth*mHeight] ;

    // Calculate the white and black point according to the contrast and the brightness
    // When the contrast is > 0 we have the line equation
    // y = mContrast*(x-1) + (1+mContrast)*mBrightness     for 0 < x < 1  and 0 < y < 1
    // For mContrast < 0 we have
    // y = mContrast*x + (1-mContrast)*mBrightness
    float	minVal,maxVal,x1,y0 ;
    if (mContrast >= 0.) {
        y0 = -mContrast + (1+mContrast)*mBrightness ;
        x1 = 1.0 + (1.0 - (1.0+mContrast)*mBrightness)/mContrast ;
        if (y0 >= 0.) 
            minVal = 0. ;
        else 
            minVal = 1.0 - mBrightness*(1+mContrast)/mContrast ;
        if (x1 <= 1.) 
            maxVal = x1 ;
        else 
            maxVal = 1.0 ;
    }
    else {
        y0 = (1.0-mContrast)*mBrightness ;
        x1 = (mContrast-1.0)*mBrightness/mContrast ;
        if (y0 <= 1.) 
            minVal = 0. ;
        else 
            minVal = (1.0+mBrightness*(mContrast-1.0))/mContrast ;
        if (x1 <= 1.) 
            maxVal = x1 ;
        else 
            maxVal = 1.0 ;
    }
	
	float	value,range = (gNumGrayLevels-1) ;
	unsigned	char	*inData = mDisplayData ;
	unsigned	char	*outData = mappedDisplayData ;
	
	int index ;
	for(long i = 0; i < mWidth*mHeight; i++) {
	    value = (*inData++)/range ;
	    if (value < minVal) value = minVal ;
	    if (value > maxVal) value = maxVal ;
	    if (mContrast >= 0.) {
	        value = mContrast*(value-1) + (1+mContrast)*mBrightness ;
	    }
	    else {
	        value = mContrast*value + (1-mContrast)*mBrightness ;
	    }
	    if (value < minVal) value = 0. ;
	    if (value > maxVal) value = 1. ;
		value *= range ;
		index = (int)value ;
		*outData++ = (unsigned char)gLut[index] ;
	}
	DrawImage(mappedDisplayData,0,0,mWidth,mHeight) ;
	SetPort(tPort) ;
	delete[] mappedDisplayData ;
}

void	TImage::Mask(ImageMaskType maskType) {
	register long col,row,ix,iy ;
	float	*p1 ;	
	float	rad,radsq,mask,arg,cnst,piHalf ;
	long	width = mData->GetWidth() ;
	long	height = mData->GetHeight() ;
	long	h2 = height/2 ;
	long	w2 = width/2 ;

	piHalf = ::asin(1.0) ;
	float	apertureSq = w2*w2 ;
	float	hConst1,hConst2 ;
	
	hConst1 = 4*piHalf/width ;
	hConst2 = 4*piHalf/height ;
	
	switch(maskType) {
		case kCosineMask:
			p1 = mData->GetData() ;
			cnst = piHalf/::sqrt(apertureSq) ;
			for(row=0;row<height;row++) {
				iy = row-h2 ;
				for(col=0;col<width;col++) {
					ix = col-w2 ;
					radsq = ix*ix+iy*iy ;
					if (radsq >= apertureSq) {
						*p1++ = 0. ;
					}
					else {
						rad = ::sqrt(radsq) ;
						arg = cnst*rad ;
						mask = ::cos(arg) ;
						*p1++ *= mask*mask ;
					}
				}
			}
			break ;
		case kHanningMask:
			p1 = mData->GetData() ;
			float	fact1,fact2 ;
			for(row=0;row<height;row++) {
				fact2 = 1.0 - ::cos(hConst2*row) ;
				for(col=0;col<width;col++) {
					fact1 = 1.0 - ::cos(hConst1*col) ;
					mask =fact1*fact2 ;
					*p1++ *= mask ;
				}
			}
			break ;
	}
	ZeroStatistics() ; 
}

void	TImage::Invert() {
	register long col,row ;
	float	*p1 ;
	
	float	meanVal = Mean() ;
	
	long	width = mData->GetWidth() ;
	long	height = mData->GetHeight() ;
	p1 = mData->GetData() ;
	for(row = 0; row< height; row++) {
		for(col=0;col < width;col++) {
			*p1 = meanVal - *p1 ;
			p1++ ;
		}
	}
	ZeroStatistics() ; 
	Update(false) ;
}

void	TImage::FlipHorizontal() {
	register long col,row ;
	unsigned long	numBytesToMove ;
	float	*theRow,*tempRow ;
	Cmplx	*c1,*c2 ;
	float	*p1,*p2 ;
	
	long	width = mData->GetWidth() ;
	long	height = mData->GetHeight() ;
	if (mData->IsComplex()) {
		tempRow = new float[2*width] ;
		if (tempRow != nil) {
			theRow = mData->GetData() ;
			numBytesToMove = 2*width*sizeof(float) ;
			for(row = 0; row< height; row++) {
				memcpy(tempRow,theRow,numBytesToMove) ;
				c1 = (Cmplx *)mData->GetData(0,row) ;
				c2 = (Cmplx *)(tempRow + 2*(width-1)) ;
				for(col=0;col < width;col++) {
					*c1++ = *c2-- ;
				}
				theRow += 2*width ;
			}
			delete[] tempRow ;
		}
	}
	else {
		tempRow = new float[width] ;
		if (tempRow != nil) {
			theRow = mData->GetData() ;
			numBytesToMove = width*sizeof(float) ;
			for(row = 0; row< height; row++) {
				memcpy(tempRow,theRow,numBytesToMove) ;
				p1 = mData->GetData(0,row) ;
				p2 = tempRow + width-1 ;
				for(col=0;col < width;col++) {
					*p1++ = *p2-- ;
				}
				theRow += width ;
			}
			delete[] tempRow ;
		}
	}
	ZeroStatistics() ; 
	Update() ;
}

void	TImage::FlipVertical() {
	register long row ;
	float	*tempRow ;
	float	*lowRow,*highRow ;
	unsigned long	numBytesToMove ;
	
	long	width = mData->GetWidth() ;
	long	height = mData->GetHeight() ;
	if (mData->IsComplex()) {
		tempRow = new float[width*2] ;
		numBytesToMove = 2*width*sizeof(float) ;
		if (tempRow != nil) {
			lowRow = mData->GetData() ;
			highRow = lowRow + 2*(height-1)*width ;
			for(row=0;row < height/2;row++) {
				memcpy(tempRow,lowRow,numBytesToMove) ;
				memcpy(lowRow,highRow,numBytesToMove) ;
				memcpy(highRow,tempRow,numBytesToMove) ;
				highRow -= 2*width ;
				lowRow += 2*width ;
			}
			delete[] tempRow ;
		}
	}
	else {
		tempRow = new float[width] ;
		numBytesToMove = width*sizeof(float) ;
		if (tempRow != nil) {
			lowRow = mData->GetData() ;
			highRow = lowRow + (height-1)*width ;
			for(row=0;row < height/2;row++) {
				memcpy(tempRow,lowRow,numBytesToMove) ;
				memcpy(lowRow,highRow,numBytesToMove) ;
				memcpy(highRow,tempRow,numBytesToMove) ;
				highRow -= width ;
				lowRow += width ;
			}
			delete[] tempRow ;
		}
	}
	ZeroStatistics() ; 
	Update() ;
}

void TImage::RotateClockWise(float numDegrees) {

	if (numDegrees == 90.) {
		RotateRight() ;
		return ;
	}
	
	if (numDegrees == -90.) {
		RotateLeft() ;
		return ;
	}
	
	// Calculate the new width and height of the image
	float	angle = numDegrees*::asin(1.0)/90. ;
	float	csga = ::cos(angle) ;
	float	siga = ::sin(angle) ;
		
	register	long	i,j ;
	register	float	*p1,*pp ;
	register	Cmplx	*cc1,*cc2,*cc3,*cc4,*cc ;
	Cmplx		*cdata ;
	float		dx,dy,x,y,ycsga,ysiga,*data ;
	long		ix,iy,y1,x1,w02,h02,hh2,ww2,nextRowOffset ;
	long		ih,ik,h1,k1 ;
	float		h,k,dh,dk ;
	
	TArray *tempArray = mData ;

	long	width = near(fabs(mWidth*csga)+fabs(mHeight*siga)) ;
	long	height = near(fabs(mWidth*siga)+fabs(mHeight*csga)) ;

	ArrayDataType type = mData->IsComplex() ? kComplexArray : kRealArray ;	
	mData = new TArray(width,height,type) ;

	h02 = mHeight/2 ;
	w02 = mWidth/2 ;
	hh2 = height/2 ;
	ww2 = width/2 ;

	if (mData->IsComplex()) {	
		cc = (Cmplx *)mData->GetData() ;		
		cdata 	= (Cmplx *)tempArray->GetData() ;
		if (!mIsRealSpace)	 {	// This is a Fourier transform such that rotations work differently
			for(j=0;j<height;j++) {
				ik = j - hh2 ;
				ysiga = ik*siga ;
				ycsga = ik*csga ;
				for(i=0;i<width;i++) {
					ih = i - ww2 ;
					h = ih*csga + ysiga ;
					k = -ih*siga + ycsga ;
					h1 = (int)h ;
					k1 = (int)k ;
					cc	= (Cmplx *)mData->GetRealHK(ih,ik) ;
					if (k1 >= -h02 && k1 < h02-1 && h1 >= -w02 && h1 < w02-1) {
						dk = k-k1 ;
						dh = h-h1 ;
						cc1 = (Cmplx *)tempArray->GetRealHK(h1,k1) ;
						cc2 = (Cmplx *)tempArray->GetRealHK(h1+1,k1) ;
						cc3 = (Cmplx *)tempArray->GetRealHK(h1,k1+1) ;
						cc4 = (Cmplx *)tempArray->GetRealHK(h1+1,k1+1) ;
						*cc  = (1.-dk)*((1-dh)*(*cc1) + dh*(*cc2)) +
									dk*((1-dh)*(*cc3)+dh*(*cc4)) ;
					}
					else {
						*cc = 0. ;
					}
				}
			}
		}
		else {
			for(j=0;j<height;j++) {
				iy = j - hh2 ;
				ysiga = iy*siga ;
				ycsga = iy*csga ;
				for(i=0;i<width;i++) {
					ix = i - ww2 ;
					x = ix*csga + ysiga ;
					y = -ix*siga + ycsga ;
					x += w02 ;
					y += h02 ; 
					y1 = (int)y ;
					x1 = (int)x ;
					if (y1 >= 0 && y1 < mHeight-1 && x1 >= 0 && x1 < mWidth-1) {
						dy = y-y1 ;
						dx = x-x1 ;
						cc1 = cdata + y1*mWidth + x1 ;
						*cc++  = (1.-dy)*((1-dx)*(*cc1) + dx*(*(cc1+1))) +
									dy*((1-dx)*(*(cc1+mWidth))+dx*(*(cc1+mWidth+1))) ;
					}
					else {
						*cc++ = 0. ;
					}
				}
			}
		}
	}
	else {	
		pp = mData->GetData() ;	// New data	
		data 	= tempArray->GetData() ;	// old data
		for(j=0;j<height;j++) {
			iy = j - hh2 ;
			ysiga = iy*siga ;
			ycsga = iy*csga ;
			for(i=0;i<width;i++) {
				ix = i - ww2 ;
				x = ix*csga + ysiga ;
				y = -ix*siga + ycsga ;
				x += w02 ;
				y += h02 ; 
				y1 = (int)y ;
				x1 = (int)x ;
				if (y1 >= 0 && y1 < mHeight-1 && x1 >= 0 && x1 < mWidth-1) {
					nextRowOffset = mWidth ;
					dy = y-y1 ;
					dx = x-x1 ;
					p1 = data + y1*mWidth + x1 ;
					*pp++  = (1.-dy)*((1-dx)*(*p1) + dx*(*(p1+1))) +
								dy*((1-dx)*(*(p1+mWidth))+dx*(*(p1+mWidth+1))) ;
				}
				else {
					*pp++ = 0. ;
				}
			}
		}
	}
	
	delete	tempArray ;
	
	ZeroStatistics() ; 
	mWidth = width ;
	mHeight = height ;
	
}

void TImage::RotateRight() {
	short	oldWidth,oldHeight ;
	register	long row,col ;
	
	TArray	*tempArray = mData ;

	oldWidth = mWidth ;
	oldHeight = mHeight ;
	
	mWidth = oldHeight ;
	mHeight = oldWidth ;

	ArrayDataType dataType = tempArray->IsComplex() ? kComplexArray : kRealArray ;
	mData = new TArray(mWidth,mHeight,dataType) ;
	
	float *p1,*p2 ;
	Cmplx *c1,*c2 ;
	
	if (dataType == kRealArray) {
		p1 = mData->GetData() ;
		for(row=0;row<mHeight;row++) {
			p2 = tempArray->GetData(row,oldHeight-1) ;
			for(col=0;col<mWidth;col++) {
				*p1++ = *p2 ;
				p2 -= oldWidth ;
			}
		}
	}
	else {
		c1 = (Cmplx *)mData->GetData() ;
		for(row=0;row<mHeight;row++) {
			c2 = (Cmplx *)tempArray->GetData(row,oldHeight-1) ;
			for(col=0;col<mWidth;col++) {
				*c1++ = *c2 ;
				c2 -= oldWidth ;
			}
		}
	}
	ZeroStatistics() ; 
	delete tempArray ;
}

void TImage::RotateLeft() {
	short	oldWidth,oldHeight ;
	register	long row,col ;

	TArray	*tempArray = mData ;
	
	oldWidth = mWidth ;
	oldHeight = mHeight ;
	
	mWidth = oldHeight ;
	mHeight = oldWidth ;

	ArrayDataType dataType = tempArray->IsComplex() ? kComplexArray : kRealArray ;
	mData = new TArray(mWidth,mHeight,dataType) ;
	
	float *p1,*p2 ;
	Cmplx *c1,*c2 ;
	
	if (dataType == kRealArray) {
		p1 = mData->GetData() ;
		for(row=0;row<mHeight;row++) {
			p2 = tempArray->GetData(oldWidth-1-row,0) ;
			for(col=0;col<mWidth;col++) {
				*p1++ = *p2 ;
				p2 += oldWidth ;
			}
		}
	}
	else {
		c1 = (Cmplx *)mData->GetData() ;
		for(row=0;row<mHeight;row++) {
			c2 = (Cmplx *)tempArray->GetData(oldWidth-1-row,0) ;
			for(col=0;col<mWidth;col++) {
				*c1++ = *c2 ;
				c2 += oldWidth ;
			}
		}
	}
	ZeroStatistics() ; 
	delete tempArray ;
}

void 	TImage::ReSize(short width,short height) {
	register	long	i,j,ix,iy,nextRowOffset,nextColOffset,oldWidth,oldHeight ;
	register	float	x,y,dx,dy,ddx,ddy,*pp,*p1 ;
	register	Cmplx	*c1,*cc ;

	if (mWidth == width && mHeight == height) return ;
	
			
	oldWidth = mWidth ;
	oldHeight = mHeight ;
		
	ddx = (float)mWidth ;
	ddx /= width ;
	ddy = (float)mHeight ;
	ddy /= height ;
		
	pp = mData->GetData() ;

	mWidth = width ;
	mHeight = height ;	
	ArrayDataType type = mData->IsComplex()? kComplexArray : kRealArray ;
	TArray *tData = new TArray(mWidth,mHeight,type) ;
	if (mRGBData != nil) {
		// Resample the RGBData 
		unsigned char *tRGBData = new unsigned char[mWidth*mHeight*4] ;
		if (tRGBData != nil) {
			unsigned char *uc1,*ucc ;
			uc1 = mRGBData ;
			ucc = tRGBData ;
			for(j=0;j<mHeight;j++) {
				y = j*ddy ;
				iy = (int)y ;
				dy = y-iy ;
				nextRowOffset = oldWidth ;
				if (iy == oldHeight-1) {
					nextRowOffset = -iy*oldWidth;
				}
				nextRowOffset *= 4 ;
				for (i=0;i<mWidth;i++) {
					x = i*ddx ;
					ix = (int)x ;
					dx = x-ix ;
					nextColOffset = 1 ;
					if (ix == oldWidth-1) {
						nextColOffset = -ix ;
					}
					nextColOffset *= 4 ;
					uc1 = mRGBData+4*(iy*oldWidth+ix) ;
					*ucc++  = 0 ;
					uc1++ ;
					*ucc++  = (1.-dy)*((1-dx)*(*uc1) + dx*(*(uc1+nextColOffset))) +
							dy*((1-dx)*(*(uc1+nextRowOffset))+dx*(*(uc1+nextRowOffset+nextColOffset))) ;
					uc1++ ;
					*ucc++  = (1.-dy)*((1-dx)*(*uc1) + dx*(*(uc1+nextColOffset))) +
							dy*((1-dx)*(*(uc1+nextRowOffset))+dx*(*(uc1+nextRowOffset+nextColOffset))) ;
					uc1++ ;
					*ucc++  = (1.-dy)*((1-dx)*(*uc1) + dx*(*(uc1+nextColOffset))) +
							dy*((1-dx)*(*(uc1+nextRowOffset))+dx*(*(uc1+nextRowOffset+nextColOffset))) ;
				}
			}
			delete[] mRGBData ;
			mRGBData = tRGBData ;
		}
	}		
	if (!mData->IsComplex()) {	
		pp = tData->GetData() ;
		for(j=0;j<mHeight;j++) {
			y = j*ddy ;
			iy = (int)y ;
			dy = y-iy ;
			nextRowOffset = oldWidth ;
			if (iy == oldHeight-1) {
				nextRowOffset = -iy*oldWidth;
			}
			for (i=0;i<mWidth;i++) {
				x = i*ddx ;
				ix = (int)x ;
				dx = x-ix ;
				nextColOffset = 1 ;
				if (ix == oldWidth-1) {
					nextColOffset = -ix ;
				}
				p1 = mData->GetData(ix,iy) ;
				*pp  = (1.-dy)*((1-dx)*(*p1) + dx*(*(p1+nextColOffset))) +
						dy*((1-dx)*(*(p1+nextRowOffset))+dx*(*(p1+nextRowOffset+nextColOffset))) ;
				pp++ ;
			}
		}
	}
	else {
		cc = (Cmplx *)tData->GetData() ;
		for(j=0;j<mHeight;j++) {
			y = j*ddy ;
			iy = (int)y ;
			dy = y-iy ;
			nextRowOffset = oldWidth ;
			if (iy == oldHeight-1) {
				nextRowOffset = -iy*oldWidth;
			}
			for (i=0;i<mWidth;i++) {
				x = i*ddx ;
				ix = (int)x ;
				dx = x-ix ;
				nextColOffset = 1 ;
				if (ix == oldWidth-1) {
					nextColOffset = -ix ;
				}
				c1 = (Cmplx *)mData->GetData(ix,iy) ;
				*cc  = (1.-dy)*((1-dx)*(*c1) + dx*(*(c1+nextColOffset))) +
						dy*((1-dx)*(*(c1+nextRowOffset))+dx*(*(c1+nextRowOffset+nextColOffset))) ;
				cc++ ;
			}
		}
	}
	delete mData ;
	ZeroStatistics() ; 
	mData = tData ;
}

void	TImage::Scale(float percX,float percY) {
	short	w,h ;
	w = near(percX*mWidth/100.) ;
	h = near(percY*mHeight/100.) ;
	
	// In principle we should have two calibration units (x and y)
	// but I think that is unncessarily complication for this program
	mCalibration /= (percX/100.) ;
	ReSize(w,h) ;
}

void	TImage::Display(TWindow *window,Rect *rect) {
	GrafPtr	theOldPort ;
	extern 	short	gImageDepth ;

	if(mIsVisible)
		return ;


	GetPort(&theOldPort) ;
	AdjustGamma() ;
	AdjustSampling() ;

	if (window == nil && mWindow == nil) {
		// We need to create a window to display ourselves
		mWindow = new TImageWindow(gImageDisplayPosition.x,
			gImageDisplayPosition.y,mWidth,mHeight,
			gImageDepth,mName.GetString(),true,this) ;
	
		if (mOffscreen == nil) CreateOffscreen() ;
		mWindow->SetCommander((TWindow *)gTheControlWindow) ;		
		AddWindow(mWindow) ;

		gImageDisplayPosition.x += 10 ;

		// Reduce these from 1024 to 600, since not all displays are this big!
		if (gImageDisplayPosition.x > 600) {
			gImageDisplayPosition.x = 240 ;
			gImageDisplayPosition.y = 0 ;
		}
		gImageDisplayPosition.y += 10 ;
		if (gImageDisplayPosition.y > 600) {
			gImageDisplayPosition.x = 240 ;
			gImageDisplayPosition.y = 0 ;
		}
		mIsVisible = true ;
		mWindow->Update(nil) ;
		SetTargetWindow(mWindow) ;
	}
	else {
		mWindow = window ;
		if (mOffscreen == nil)  {
			CreateOffscreen() ;
		}
		if (mOffscreen == nil) return ;

		Rect	theSourceRect,theDestRect ;
		SetPort(window) ;
	
		SetRect(&theSourceRect,0,0,mWidth,mHeight) ;
		if (rect != nil) 
			theDestRect = *rect ;
		else 
			theDestRect = theSourceRect ;

		CopyBits(mOffscreen,window,&theSourceRect,&theDestRect,srcCopy) ;
	}
	SetPort(theOldPort) ;
}

void	TImage::Update(Boolean redraw) {
	GrafPtr theOldPort ;
	
	GetPort(&theOldPort) ;
	if (mWindow != nil) {
		SetPort(mWindow) ;
		if (mOffscreen != nil) {
			delete mOffscreen ;
			mOffscreen = nil ;
			CreateOffscreen() ;
		}
		Rect theRect = mWindow->GetPortRect() ;
		if (mOffscreen != nil && redraw) {
			CopyBits(mOffscreen,
					 mWindow,
					&theRect,&theRect,srcCopy) ;
		}	
		SetPort(theOldPort) ;
	}
	mContrast = 1.0 ;
	mBrightness = 0.5 ;
}

void	TImage::Draw() {
	GrafPtr	tPort ;
	GetPort(&tPort) ;
	if (mOffscreen == nil) {
		CreateOffscreen() ;
	}
	if (mOffscreen != nil) {
		Rect theRect = {0,0,mWidth,mHeight} ;		
		CopyBits(mOffscreen,
				 tPort,
				&theRect,&theRect,srcCopy) ;
	}
}

void	TImage::GetAbbreviatedName(char *name) {
	char text[256] ;
	mName.GetString(text) ;
	char	*cp ;
	if ((cp = strrchr(text,'/')) != NULL) 
		strcpy(name,cp+1) ;
	else
		strcpy(name,text) ;	
}

TImage*	TImage::FourierTransform(short direction) {
	TImage *image = new TImage(*this) ;
	switch(direction) {
		case kFFTForward:
			image->fft() ;
			break ;
		case kFFTInverse:
			image->ifft() ;
			break ;
	}
		
	return image ;
}

void	TImage::UpdateCalibration() {
	switch(mCalibrationUnit) {
		case kPixel:
			mCalibrationUnit = kInversePixel ;
			mCalibration = 1./(mCalibration*mWidth) ;
			break ;
		case kInversePixel:
			mCalibrationUnit = kPixel ;
			mCalibration = 1./(mCalibration*mWidth) ;
			break ;
		case kAngstrom:
			mCalibrationUnit = kInverseAngstrom ;
			mCalibration = 1./(mCalibration*mWidth) ;
			break ;
		case kInverseAngstrom:
			mCalibrationUnit = kAngstrom ;
			mCalibration = 1./(mCalibration*mWidth) ;
			break ;
		case kNanometer:
			mCalibrationUnit = kInverseNanometer ;
			mCalibration = 1./(mCalibration*mWidth) ;
			break ;
		case kInverseNanometer:
			mCalibrationUnit = kNanometer ;
			mCalibration = 1./(mCalibration*mWidth) ;
			break ;
	}
}

TImage 	*TImage::Extract(float x0,float y0,short width,short height) {

	// Extract the data from an image 'centered' on the spot (x,y)
	// and resampled if needed
	// The center pixel goes in the width/2,height/2 pixel of the image
	register int	i,j ;
	short	ix,iy ;
	float	w2,h2,x,y,dx,dy ;
	float	left,right,bottom,top ;
	TImage	*extractedImage = nil ;
	
	w2 = width/2. ;
	h2 = height/2. ;
	left = x0 - w2 ;
	right = x0 + w2 ;
	top = y0 - h2 ;
	bottom = y0 + h2 ;
	if (left < 0. || right > mWidth-1 || top < 0. || bottom > mHeight-1)
		return extractedImage ;
		
	extractedImage = new TImage(width,height) ;
	
	if (!mData->IsComplex()) {	
		register float	*p1 = extractedImage->GetData() ;
		for(j=0;j<height;j++) {
			y = top + j ;
			iy = (int)y ;
			dy = y - iy ;
			for(i=0;i<width;i++) {
				x = left + i ;
				ix = (int)x ;
				dx = x - ix ;
				*p1++ = (1.-dy)*((1-dx)*(*GetData(ix,iy)) + dx*(*GetData(ix+1,iy))) +
							dy*((1-dx)*(*GetData(ix,iy+1))+dx*(*GetData(ix+1,iy+1))) ;			
			}
		}
	}
	else {
		register Cmplx	*cc = (Cmplx *)extractedImage->GetData() ;
		Cmplx	*c1 ;
		for(j=0;j<height;j++) {
			y = top + j ;
			iy = (int)y ;
			dy = y - iy ;
			for(i=0;i<width;i++) {
				x = left + i ;
				ix = (int)x ;
				dx = x - ix ;
				c1 = (Cmplx *)mData->GetData(ix,iy) ;
				*cc++ = (1.-dy)*((1-dx)*(*c1) + dx*(*(c1+1))) +
							dy*((1-dx)*(*(c1+mWidth))+dx*(*(c1+mWidth+1))) ;			
			}
		}
	}
	
	return extractedImage ;
}

TImage *ImportRawFile(char	*fileName) {
	TImage	*image ;

	TImportRawImageDialog *theDialog ;
	TImportRawImageDialogRec theDialogData ;


	long	numElements ;
	unsigned char	*dataPtr,*filePtr ;
	long	i ;


	TFile *theFile = new TFile(fileName,kRead) ;

	theDialog = new TImportRawImageDialog(fileName,theFile->GetSize()) ;
	if (!theDialog->DoModalDialog()) {
		delete theFile ;
		delete theDialog ;
		return nil ;
	}
	theDialog->GetDialogData(&theDialogData) ;
	delete theDialog ;
	gSwapBytes = theDialogData.swapBytes ;
	gImportImageWidth = theDialogData.imageWidth ;
	gImportImageHeight = theDialogData.imageHeight ;
	gImageDataType = theDialogData.dataType ;
	gByteOffset = theDialogData.byteOffset ;

	filePtr = new unsigned char[theDialogData.numBytes] ;
	theFile->Read(filePtr,theDialogData.numBytes) ;
	
	dataPtr = filePtr + theDialogData.byteOffset ;
	
	unsigned long *uulp = (unsigned long *)dataPtr ;
	unsigned short *uusp = (unsigned short *)dataPtr ;
	short *sp ;
	unsigned short *usp ;
	long *lp ;
	unsigned long *ulp ;
	float	*p1,*p2,*p3,*p4 ;

	numElements = gImportImageWidth*gImportImageHeight ;
	if (gSwapBytes) {
		switch(gImageDataType) {
			case k1ByteSigned:
			case k1ByteUnsigned:
				break ;
			case k2ByteSigned:
				sp = (short *)dataPtr ;
				for(i=0; i < numElements;i++) {
					*sp++ = ByteSwapShort(*uusp++) ;
				}
				break ;
			case k2ByteUnsigned:
				usp = (unsigned short *)dataPtr ;
				for(i=0; i < numElements;i++) {
					*usp++ = ByteSwapShort(*uusp++) ;
				}
				break ;
			case k4ByteSigned:
				lp = (long *)dataPtr ;
				for(i=0; i < numElements;i++) {
					*lp++ = ByteSwapLong(*uulp++) ;
				}
				break ;
			case k4ByteUnsigned:
				ulp = (unsigned long *)dataPtr ;
				for(i=0; i < numElements;i++) {
					*ulp++ = ByteSwapLong(*uulp++) ;
				}
				break ;
			case k4ByteReal:
				p1 = (float *)dataPtr ;
				for(i=0; i < numElements;i++) {
					*p1++ = ByteSwapFloat(*uulp++) ;
				}
				break ;
			case k8ByteReal:
				p2 = (float *)dataPtr ;	// Probably wrong
				for(i=0; i < 2*numElements;i++) {
					*p2++ = ByteSwapFloat(*uulp++) ;
				}
				break ;
			case k4ByteComplex:
				p3 = (float *)dataPtr ;	
				for(i=0; i < 2*numElements;i++) {
					*p3++ = ByteSwapFloat(*uulp++) ;
				}
				break ;
			case k8ByteComplex:
				p4 = (float *)dataPtr ;	// Probably wrong
				for(i=0; i < 4*numElements;i++) {
					*p4++ = ByteSwapFloat(*uulp++) ;
				}
				break ;
		}
	}
	image = new TImage(gImportImageWidth,gImportImageHeight,
		(float *)dataPtr,90.,gImageDataType) ;
	delete theFile ;
	delete[] filePtr ;

	image->ZeroStatistics() ; 

	//image->SetName(fileName) ;
	return image ;
}


TImage *ImportTiffFile(char	*fileName) {
	TImage *image = nil ;

	image = CreateImageFromTiffFile(fileName) ;
	if (image != nil)
		image->ZeroStatistics() ; 

	return image ;
}

void	OpenImageFile(FSInfo &theFSInfo) {
	TImage	*theImage = nil  ;
	char	name[256] ;

	// If the filetype is TIFF then pass the file to ImportTiffFile
	// get the file name at this point as well
	// Check to see if it is a tiff file
	strcpy(name,theFSInfo.mFileName) ;
	Boolean isTiffFile = false ;
	// See if the file has the extension ".tif" or ".TIF" as from a PC
	char *cp = strrchr(name,'.') ;  // Does the name have a . in it?
	if (cp != NULL) {
		// If YES ! Check to se if it is of type .tif or .TIF
		if (strncmp(cp,".tif",4)==0) {
			isTiffFile = true ;
		}
		else if (strncmp(cp,".TIF",4)==0) {
				isTiffFile = true ;
		}
	}
	if (isTiffFile) {
		// Set the vol-ref num to the appropriate location
		theImage = ImportTiffFile(theFSInfo.mFileName) ;
	}
	else {	// Try to import the file as a raw data-file
			theImage = ImportRawFile(theFSInfo.mFileName) ;
	}
	if (theImage != nil) {
		theImage->SetName(theFSInfo.mFileName) ;
		theImage->Display() ;
	}
}

void TImage::Repeat(int nX,int nY) {
	int newWidth,newHeight,jold,iold  ;
	register int i,j ;
	float *p1,*p2 ;

	if (nX == 1 && nY == 1) return ;    // No work to be done

	newWidth = mWidth*nX ;
	newHeight = mHeight*nY ;

	TArray *tempArray = new TArray(newWidth,newHeight) ;

	p1 = tempArray->GetData() ;
	p2 = mData->GetData() ;

	for (j=0;j<newHeight;j++) {
		jold = j%mHeight ;
		for (i=0;i<newWidth;i++) {
			iold = i%mWidth ;
			*p1++ = *(p2+jold*mWidth+iold) ;
		}
	}
	delete mData ;
	mData = tempArray ;
	mWidth = newWidth ;
	mHeight = newHeight ;
	ZeroStatistics() ; 
}

void TImage::AdjustSampling() {
	float *p1,*p2 ,dx,dy,dxnew,dynew,aaxis,baxis,rlo,xpos,ypos,fact ;
	register 	long j,n,m ;
	long 		mlo,nlo,jlo,jhi,maxRowIndex,maxColIndex,maxArrayIndex ;

	if (mSamplingX == mSamplingY) return ;
	
	aaxis = mWidth*mSamplingX ;
	baxis = mHeight*mSamplingY ;

	dx = mSamplingX ;
	dy = mSamplingY ;

	long newWidth = mWidth ;
	long newHeight = mHeight ;
	if (mSamplingY < mSamplingX) {
		mSamplingY = mSamplingX ;
		dynew = mSamplingY ;
		newHeight = near(baxis/mSamplingY) ;
	}
	else {
		mSamplingX = mSamplingY ;
		dxnew = mSamplingX ;
		newWidth = near(aaxis/mSamplingX) ;
	}
	TArray	*tempArray = new TArray(newWidth,newHeight);
	
	p2 = mData->GetData() ;
	p1 = tempArray->GetData() ;
	
	maxRowIndex = mHeight - 1 ;
	maxColIndex	= mWidth - 1 ;
	maxArrayIndex = mWidth*mHeight - 1 ;
	if (dy < dx) {
		for (m=0;m<newHeight;m++) {
			ypos = m*mSamplingY ;
			rlo  = ypos/dy ;
			mlo  = (int) rlo ;
			fact = rlo - mlo ;
			if (mlo > maxRowIndex) { mlo = maxRowIndex ; fact = 0. ; } 
			if (mlo == maxRowIndex) {
				for (n=0;n< newWidth;n++) {
					jlo = mlo*mWidth + n ;
					*p1++ = (*(p2+jlo))  ;
				}
			}
			else {
				for (n=0;n< newWidth;n++) {
					jlo = mlo*mWidth + n ;
					jhi = jlo + mWidth ;
					*p1++ = (1.-fact)*(*(p2+jlo)) + 
							fact*(*(p2+jhi)) ;
				}
			}
		}
	}
	else {
		for (n=0;n<newWidth;n++) {
			xpos = n*mSamplingX ;
			rlo = xpos/dx ;
			nlo = (int)rlo ;
			fact = rlo - nlo ;
			if (nlo > maxColIndex) { nlo = maxColIndex ; fact = 0. ; } 
			if (nlo == maxColIndex) {
				for (m=0;m< newHeight;m++) {
					jlo = m*mWidth + nlo ;
					j = m*newWidth + n ;
					*(p1+j) = *(p2+jlo) ;
				}
			}
			else {
				for (m=0;m< newHeight;m++) {
					jlo = m*mWidth + nlo ;
					jhi = jlo + 1 ;
					j = m*newWidth + n ;
					*(p1+j) = (1.-fact)*(*(p2+jlo)) + 
								fact*(*(p2+jhi)) ;
				}
			}
		}
	}
	
	delete mData ;
	mData = tempArray ;
	mWidth = newWidth ;
	mHeight = newHeight ;
	ZeroStatistics() ; 
}

void TImage::AdjustGamma() {
	float 	dx,dy,dConstant,di,csga,siga,ri ;
	int 	ihi,ilo,numBytes ;
	register float *p1,*p2 ;
	register int j,i ;

	if (fabs(mGamma - 90.) < 0.005) return ;  // Nothing to do

	csga = ::cos(mGamma*::asin(1.0)/90.) ;
	if (fabs(csga) < 0.001)csga = 0. ;

	dx = mSamplingX ;		// Delta x
	dy = mSamplingY ;		// Delta y

	dConstant = -csga* dy/dx  ;
	           				 		// Distance from upright y-axis to
	                          		// the old skewed y-axis when multiplied
	                          		// by (j-1). The distance has been scaled
	                          		// by dividing by DEL_X to give offset in
	                          		// index.	

	 
	float *line = new float[mWidth] ;		// Place for temporary row
	p2 = line ;
	numBytes = mWidth*sizeof(float) ;
	for (j=0;j<mHeight;j++)	{			// LOOP OVER ALL THE ROWS
		p1 = mData->GetData() + j*mWidth ;
		memcpy((char *)line,(char *)p1,numBytes) ;	// Temporary store the row
		di = j*dConstant ;
		for(i=0;i<mWidth;i++)	{		// LOOP OVER ALL THE COLUMNS
			ri = i + di ;			// Old index I of this point. It is real
									// because it may lay between two indices.
			if (ri < 0.) ri += mWidth ;
			ilo = (int)ri ;			// Idex of point to the left
			dx = ri - ilo ;
			ihi = ilo + 1 ;			// I index of nearest point to the right
			ilo = ilo%mWidth ;		// Wrap around the index if necessary
			ihi = ihi%mWidth ;
			*p1++ = (1.-dx)*(*(p2+ilo)) + dx*(*(p2+ihi)) ;
		}
	}
	delete[] line ;
	siga = ::sin(mGamma*::asin(1.0)/90.) ;
	if (fabs(siga-1.0) < 0.001)siga = 1. ;
	mSamplingY = mSamplingY*siga ;
	mGamma = 90. ;
	ZeroStatistics() ; 
}

void	CopyImageToImage(TImage *sourceImage,Rect *sourceRect,TImage *destImage,Rect *destRect,float multiplicationConstant) {
	register long j,i ;


	short	widthSourceImage 	= sourceImage->Width() ;
	short	widthDestImage 		= destImage->Width() ;
	short	heightSourceImage 	= sourceImage->Height() ;
	short	heightDestImage 	= destImage->Height() ;
	Rect	imageSourceRect 	= {0,0,widthSourceImage,heightSourceImage} ;
	Rect	imageDestRect 		= {0,0,widthDestImage,heightDestImage} ;
	
	if (!RectInRect(sourceRect,&imageSourceRect))
		return ;
	if (!RectInRect(destRect,&imageDestRect))
		return ;
		
	short	sourceWidth 	= sourceRect->right - sourceRect->left ;
	short	sourceHeight 	= sourceRect->bottom - sourceRect->top ;
	short	destWidth 		= destRect->right - destRect->left ;
	short	destHeight 		= destRect->bottom - destRect->top ;
		
	if (sourceWidth != destWidth)
		return ;
	if (sourceHeight != destHeight)
		return ;
		
	TArray *sourceArray = sourceImage->GetDataArray() ;
	TArray *destArray = destImage->GetDataArray() ;
	
	if (sourceArray->IsComplex() && destArray->IsComplex()) {
		for(j=0;j<sourceHeight;j++) {
			memcpy(destArray->GetData(destRect->left,destRect->top+j),sourceArray->GetData(sourceRect->left,sourceRect->top+j),2*sourceWidth*sizeof(float)) ;
				for(i=0;i<sourceWidth; i++) {
					*((Cmplx *)destArray->GetData(destRect->left+i,destRect->top+j)) *= multiplicationConstant ;
				}
		}
	}
	else if (!sourceArray->IsComplex() && !destArray->IsComplex()) {
		for(j=0;j<sourceHeight;j++) {
			memcpy(destArray->GetData(destRect->left,destRect->top+j),sourceArray->GetData(sourceRect->left,sourceRect->top+j),sourceWidth*sizeof(float)) ;
			if (multiplicationConstant != 1.) {
				for(i=0;i<sourceWidth; i++) {
					*destArray->GetData(destRect->left+i,destRect->top+j) *= multiplicationConstant ;
				}
			}
		}
	}
	destImage->ZeroStatistics() ; 
		
}

void DisplayImageInRect(TImage *image, Rect *rect) {
	
	GrafPtr	theCurrentPort ;
	Rect	theRect ;
	
	GetPort(&theCurrentPort) ;
	
	long	w = rect->right-rect->left ;
	long	h = rect->bottom-rect->top ;
	float	fa = (float)image->mWidth/w ;
	float	fb = (float)image->mHeight/h ;
	float	f = fa > fb ? fa : fb ; 
	if (f > 1.0) {
		w = image->mWidth/f ;
		h = image->mHeight/f ;
	}
	else {
		w = image->mWidth ;
		h = image->mHeight ;
	}
	
	rect->left = (rect->right+rect->left - w)/2 ;
	rect->right = rect->left + w ;
	rect->top = (rect->top+rect->bottom -h)/2 ;
	rect->bottom = rect->top + h ;

	if (image->mOffscreen == nil) {	
		image->CreateOffscreen() ;
	}		

	SetRect(&theRect,0,0,image->mWidth,image->mHeight) ;
	CopyBits(image->mOffscreen,theCurrentPort,&theRect,rect,srcCopy) ;
}

/*
void DisplayImageInWindow(TImage *image, TWindow *window, Rect *rect,Boolean scaleToExpImageRatio) {
	
	float	minVal,maxVal,*p1,value,range ;
	int i,j ;
	GrafPtr	theOldPort ;
	unsigned char *ucp ;
	Rect	theRect ;
	
	GetPort(&theOldPort) ;
	float	maxDisplayValue = gNumGrays-1 ;

	// Create the offscreen for the image in case it doesn't exist
	if (image->mOffscreen == nil) {			
		minVal = 100000. ;
		maxVal = -100000. ;
		p1 = image->mData ;
		for(i=0;i<image->mWidth*image->mHeight;i++) {
			value = *p1++ ;
			if (value < minVal)  minVal = value ;
			if (value > maxVal)  maxVal = value ;
		}
		image->mMinValue = minVal ;
		image->mMaxValue = maxVal ;
		range = (maxVal-minVal) ;
		p1 = image->mData ;
	
		// Create the image values into our workspace
		unsigned char *baseAddress = (unsigned char *)gWorkSpace ;
		ucp = baseAddress ;
		unsigned long *ulp = (unsigned long *)baseAddress ;
		int index ;
		for(i=0;i<image->mHeight;i++) {
			for(j=0;j<image->mWidth;j++) {
				value = maxDisplayValue*((*p1)- minVal)/range ;
				p1++ ;
				index = (int)value ;
				*ulp++ = gLut[index] ;
				*ucp++ = (unsigned char)gLut[index] ;
			}
		}

		image->mOffscreen = new TOffscreen(window) ;

		SetOffscreenPort(image->mOffscreen) ;
		DrawImage(baseAddress,0,0,image->mWidth,image->mHeight) ;
		UnSetOffscreenPort(image->mOffscreen)	;
	}

	SetPort(window) ;


	Rect itemRect = *rect ;
	float fraction,ratio1,ratio2 ;
	short width = image->mWidth ;
	short height = image->mHeight ;

	ratio1 = itemRect.right-itemRect.left ;
	ratio1 /= width ;
	ratio2 = itemRect.bottom-itemRect.top ;
	ratio2 /= height ;
	fraction = ratio2 < ratio1 ? ratio2 : ratio1 ;
	if (fraction > 1.) fraction = 1. ;
		
	if (fraction != 1.0) {
		width *= fraction ;
		height *= fraction ;
	}

	if (scaleToExpImageRatio && gAspectRatio != 1.0) {
		if (width == gExperimentalImage->mWidth) {
			height = near(width*gAspectRatio) ;
		}
		else if (height == gExperimentalImage->mHeight) {
			width = near(height/gAspectRatio) ;
		}
		else {
			height = near(width*gAspectRatio) ;
		}
	}
	
	Rect theRect2 ;

	SetRect(&theRect,0,0,image->mWidth,image->mHeight) ;
	SetRect(&theRect2,(itemRect.right+itemRect.left - width)/2,
		(itemRect.bottom+itemRect.top - height)/2,
		(itemRect.right+itemRect.left + width)/2,
		(itemRect.bottom+itemRect.top + height)/2) ;
	CopyBits(image->mOffscreen,window,&theRect,&theRect2,srcCopy) ;	
	image->mWindow = window ;
	SetPort(theOldPort) ;
}
*/


TImage 	*CrossCorrelate(TImage *u,TImage *v) {
	TImage *result ;
	
	// We need to compile the FFTW library to do FFTs of images
	// that are not powers of two 
	
	TImage a = FFT(u) ;
	TImage b = FFT(v) ;
	
	result = new TImage(a.mWidth,a.mHeight,nil,90.,k4ByteComplex) ;
	Cmplx *c1,*c2,*c3 ;
	c1 = (Cmplx *)a.GetData() ;
	c2 = (Cmplx *)b.GetData() ;
	c3 = (Cmplx *)result->GetData() ;

	long size = a.mWidth*a.mHeight ;
	for(long i = 0; i < size; i++) {
		c3->x = c1->x*c2->x + c1->y*c2->y ;
		c3->y = c1->x*c2->y - c1->y*c2->y ;
		c1++ ;
		c2++ ;
		c3++ ;
	}
	result->ifft() ;
	return result ;	
}

void SaveBinaryImage(TImage *image) {
	FSInfo 	theFInfo ;
	char	imageName[256] ;
	char	text[256] ;
	
	image->GetAbbreviatedName(imageName) ; 
	
	// Strip the extension
	char	*cp ;
	if ((cp = strrchr(imageName,'.')) != NULL) 
		*cp = '\0' ;

	strcpy(theFInfo.mFileName,imageName) ;	
	sprintf(text,".bin%d_%d",image->Width(),image->Height()) ;
	strcat(theFInfo.mFileName,text) ;

	ChooseSaveFile("Save as binary floating data:",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;
	
	FILE *fp ;
	unsigned long nBytes ;
				
	fp =fopen(theFInfo.mFileName,"w") ;
	if (fp == nil) {
		sprintf(text,"Unable to open %s",theFInfo.mFileName) ;
		PutMessage(text) ;
		return ;
	}
	nBytes = image->Width()*image->Height()*sizeof(float) ;
	fwrite(image->GetData(),nBytes,1,fp) ;
	fclose(fp) ;
}


void SaveGrayscaleTiffImage(TImage *image) {
	FSInfo 	theFInfo ;
	char	imageName[256] ;
	
	image->GetAbbreviatedName(imageName) ;

	strcpy(theFInfo.mFileName,imageName) ;	
	ChooseSaveFile("Save as Tiff file:",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;
	
	register float *srcpt ;
	char	text[256] ;
	FILE *fp ;
	long nBytes ;
	register long i,j ;
	unsigned char *address ;
	unsigned char *dstpt ;
			
	long	width = image->Width() ;
	long	height = image->Height() ;
	float	minVal = image->Min() ;
	float	maxVal = image->Max() ;
	float	*data = image->GetData() ;
	fp = CreateTiffFile(theFInfo.mFileName) ;
	if (fp != nil) {
		WriteTiffDirectory(fp,kGrayScale,width,height) ;
	}
	else {
		sprintf(text,"Unable to open %s",theFInfo.mFileName) ;
		PutMessage(text) ;
		return ;
	}
	nBytes = width ;
	unsigned char *line ;
	float range = maxVal - minVal ;
	if (range == 0.) range = 1.0 ;	// Trap divide by zero
	line = (unsigned char *)malloc(width*sizeof(unsigned char)) ;
	if (line != nil) {
		for (i=0;i<height;i++) {
			srcpt = data + i*width  ;
			dstpt = line ;
			address = dstpt ;
			for(j=0;j<width;j++) {
					*dstpt++ = (unsigned char)(255.*(*srcpt - minVal)/range) ;
					srcpt++ ;
			}
			WriteTiffStrip(fp,address,nBytes) ;
		}
		free(line) ;
	}
	CloseTiffFile(fp) ;
}

void SaveRGBImage(TImage *image) {
	FSInfo 	theFInfo ;
	strcpy(theFInfo.mFileName,"") ;
	ChooseFile("Save Image to Tiff file as:","*",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;
	
	register float *srcpt ;
	char	text[256] ;
	FILE *fp ;
	long nBytes ;
	register long i,j ;
	unsigned char *address ;
	unsigned char *dstpt ;
	short	width = image->Width() ;
	short	height = image->Height() ;
	float	*data	= image->GetData() ;
				
	fp = CreateTiffFile(theFInfo.mFileName) ;
	if (fp != nil) {
		WriteTiffDirectory(fp,kRGB,image->Width(),image->Height()) ;
	}
	else {
		sprintf(text,"Unable to open %s",theFInfo.mFileName) ;
		PutMessage(text) ;
		return ;
	}
	nBytes = width ;
	unsigned char *line ;
	srcpt = data ;
	float	minValue = 1.0e+20,maxValue = -1.0e+20,value ;
	for (i=0;i<height;i++) {
		for(j=0;j<width;j++) {
			value = *srcpt++ ;
			if (value < minValue) minValue = value ;
			if (value > maxValue) maxValue = value ;
		}
	}
	float range = maxValue - minValue ;
	if (range == 0.) range = 1.0 ;	// Trap divide by zero
	line = new unsigned char[width] ;
	if (line != nil) {
		for (i=0;i<height;i++) {
			srcpt = data + i*width  ;
			dstpt = line ;
			address = dstpt ;
			for(j=0;j<width;j++) {
					*dstpt++ = (unsigned char)(255.*(*srcpt - minValue)/range) ;
					srcpt++ ;
			}
			WriteTiffStrip(fp,address,nBytes) ;
		}
		delete[] line ;
	}
	CloseTiffFile(fp) ;
}

int IsPowerOfTwo(long value) {
	float fval,rp2,dp2 ;
	int np2 ;
	
	fval = (float)value ;
	rp2 = log2(fval) ;
	np2 = (int)(rp2+0.5) ;
	dp2 = fabs(rp2 - np2) ;
	
	if (dp2 > 0.001) {
		return 0 ;
	}
		
	return np2 ;
}

/* ------------------------
void WindowToFile(TGrafPort *theWindow) {
	// ------------------------------------------------------------------ 
	// Put up the standard file dialog Box to determine the file name 
	// ------------------------------------------------------------------ 
	FSInfo 	theFInfo ;
	
	strcpy(theFInfo.mFileName,"") ;
	ChooseFile("Save Window to Tiff File as:","*",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;
	GrafPtr tPort ;
	GetPort(&tPort) ;
	SetPort(theWindow) ;
	Rect	theRect = theWindow->GetPortRect() ;
	short	width = theRect.right - theRect.left ;
	short	height = theRect.bottom - theRect.top ;
	if (ReadRGBImage((char *)gWorkSpace,0,0,width,height)) {
		WriteDisplayToTiff(theFInfo.mFileName,(unsigned char *)gWorkSpace,
					width,height) ;
	}
	SetPort(tPort) ;
}
----------------------- */

