// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include <stdlib.h>
#include <string.h>
#include "TImportRawImageDialog.h"

static char *dataTypes[] =
{ 
	"Integer 1 Byte Signed",
	"Integer 1 Byte Unsigned",
	"Integer 2 Byte Signed",
	"Integer 2 Byte Unsigned",
	"Integer 4 Byte Signed",
	"Integer 4 Byte Unsigned",
	"Real 4 Byte",
	"Real 8 Byte",
	nil
};

void	TImportRawImageDialog::DefineDialog(char *filename) {

	char text[64] ;
	TGroupBox	*theBox ;
	TStaticText	*theText ;

	theBox = new TGroupBox(10,10,250,160,"Input File Information");
	AddObject(theBox) ;

	theText = new TStaticText(20,20,120,35,"File Name");
	AddObject(theText) ;

	char	*cp ;
	if ((cp = strrchr(filename,'/')) == NULL) 
		cp = filename ;
	else 
		cp++ ;

	theText = new TStaticText(120,20,240,35,cp);
	AddObject(theText) ;

	theText = new TStaticText(20,38,120,53,"File length in bytes");
	AddObject(theText) ;

	if (mFileSize != 0)
		sprintf(text,"%d",mFileSize) ;
	else
		sprintf(text,"Unknown") ;
		
	mByteOffset = 512 ;

	mFileLengthF = new TStaticText(120,38,240,53,text);
	AddObject(mFileLengthF) ;
        // LDM Change here
	mDataTypeF = new TPopupMenu(this,20,74,180,94,
		dataTypes,2,"Data Type",kLabelOutsideTop) ;
	AddObject(mDataTypeF) ;

	mOKButton = new TButton(170,170,250,190,"OK");
	AddObject(mOKButton) ;
	mCancelButton = new TButton(70,170,160,190,"Cancel");
	AddObject(mCancelButton) ;


	mByteOffsetF = new TEditText(170,136,240,156,
		"Byte Offset","512");		
	AddTextEditObject(mByteOffsetF) ;

	mImageHeightF = new TEditText(170,115,240,135,
		"Image Height");		
	mImageHeightF->SetLabelPosition(kLabelOutsideTop) ;
	AddTextEditObject(mImageHeightF) ;

	mImageWidthF = new TEditText(20,115,90,135,
		"Image Width");
	mImageWidthF->SetLabelPosition(kLabelOutsideTop) ;
	AddTextEditObject(mImageWidthF) ;
		
	mSwapBytesF = new TCheckBox(20,136,120,156,
		"Swap Bytes");
	AddObject(mSwapBytesF) ;

	mDefaultButton = mOKButton ;
}

TImportRawImageDialog::TImportRawImageDialog(char *filename,long size) 
	: TDialog(kDialogPosX,kDialogPosY,272,202,"Import Image") {
	mFileSize = size ;

	DefineDialog(filename) ;

	char	text[256] ;

	mSwapBytes = false ;
	//mImageDataType = k4ByteReal ;
        mImageDataType = k1ByteUnsigned ;
	switch (mImageDataType) {
		case k1ByteSigned:
		case k1ByteUnsigned:
			mBytesPerElement = 1 ;
			break ;
		case k2ByteSigned:
		case k2ByteUnsigned:
			mBytesPerElement = 2 ;
			break ;
		case k4ByteSigned:
		case k4ByteUnsigned:
		case k4ByteReal:
			mBytesPerElement = 4 ;
			break ;
		case k8ByteReal:
			mBytesPerElement = 8 ;
			break ;
	}

	mDataSize = mFileSize - mByteOffset ;

	if (mFileSize != 0) {
		mNumElements = mDataSize/mBytesPerElement ;
		mImageWidth = sqrt((float)mNumElements) ;
		mImageHeight = mNumElements/mImageWidth ;
		mImageHeight += (mImageWidth*mImageHeight == mNumElements ? 0 : (long)((mNumElements - mImageHeight*mImageWidth)/mImageWidth)) ; 
	}
	else {
		mImageWidth = 128 ;
		mImageHeight = 128 ;
	}

	sprintf(text,"%d",mByteOffset) ;
	mByteOffsetF->SetEditText(text) ;

	sprintf(text,"%d",mImageWidth) ;
	mImageWidthF->SetEditText(text) ;
	SelectAll(mImageWidthF) ;
	sprintf(text,"%d",mImageHeight) ;
	mImageHeightF->SetEditText(text) ;
}

void	TImportRawImageDialog::DoKeyDownInObject(TObject *obj) {
	char	text[256] ;
	if (obj == mImageWidthF) {
		mImageWidthF->GetEditText(text) ;
		if (strlen(text) > 0) {
			mImageWidth = atoi(text) ;
			if (mImageWidth > 0) {
				if (mFileSize != 0) {

					mImageHeight = mNumElements/mImageWidth ;
					mImageHeight += (mImageWidth*mImageHeight == mNumElements ? 0 : (long)((mNumElements - mImageHeight*mImageWidth)/mImageWidth)) ; 
					sprintf(text,"%d",mImageHeight) ;
					mImageHeightF->SetEditText(text) ;
				}
			}
		}
	}
	else if (obj == mImageHeightF) {
		mImageHeightF->GetEditText(text) ;
		mImageHeight = atoi(text) ;
		if (mImageHeight > 0) {
			mImageWidth = mNumElements/mImageHeight ;
			mImageWidth += (mImageWidth*mImageHeight == mNumElements ? 0 : (long)((mNumElements - mImageHeight*mImageWidth)/mImageHeight)) ; 
			sprintf(text,"%d",mImageWidth) ;
			mImageWidthF->SetEditText(text) ;
		}
	}
	else if (obj == mByteOffsetF) {
		mByteOffsetF->GetEditText(text) ;
		mByteOffset = atoi(text) ;
		switch (mImageDataType) {
			case k1ByteSigned:
			case k1ByteUnsigned:
				mBytesPerElement = 1 ;
				break ;
			case k2ByteSigned:
			case k2ByteUnsigned:
				mBytesPerElement = 2 ;
				break ;
			case k4ByteSigned:
			case k4ByteUnsigned:
			case k4ByteReal:
				mBytesPerElement = 4 ;
				break ;
			case k8ByteReal:
			case k4ByteComplex:
				mBytesPerElement = 8 ;
				break ;
			case k8ByteComplex:
				mBytesPerElement = 16 ;
				break ;
		}
		
		mDataSize = mFileSize - mByteOffset ;
							
		mNumElements = mDataSize/mBytesPerElement ;
		mImageWidth = sqrt((float)mNumElements) ;
		mImageHeight = mNumElements/mImageWidth ;
		mImageHeight += (mImageWidth*mImageHeight == mNumElements ? 0 : (long)((mNumElements - mImageHeight*mImageWidth)/mImageWidth)) ; 

		sprintf(text,"%d",mImageWidth) ;
		mImageWidthF->SetEditText(text) ;

		sprintf(text,"%d",mImageHeight) ;
		mImageHeightF->SetEditText(text) ;
	}
		
}

void	TImportRawImageDialog::DoMouseDownInObject(TObject *obj,Point *pt) {
	if (pt != nil)
		if (!obj->DoMouseDown(pt)) return ;
	char	text[256] ;
	if (obj == mOKButton) {
		mDone = true ;
		mDialogReturnCode = kDialogOK ;
	}
	else if (obj == mCancelButton) {
		mDone = true ;
		mDialogReturnCode = kDialogCancel ;
	}
	else if (obj == mDataTypeF) {
		mImageDataType = (ImageDataType)mDataTypeF->GetSelection() ;
		long	numElements = mNumElements ;
		switch (mImageDataType) {
				case 1:
				case 2:
					mBytesPerElement = 1 ;
					mNumElements = mDataSize/mBytesPerElement ;
					break ;
				case 3:
				case 4:
					mBytesPerElement = 2 ;
					mNumElements = mDataSize/mBytesPerElement ;
					break ;
				case 5:
				case 6:
				case 7:
					mBytesPerElement = 4 ;
					mNumElements = mDataSize/mBytesPerElement ;
					break ;
				case 8:
				case 9:
					mBytesPerElement = 8 ;
					mNumElements = mDataSize/mBytesPerElement ;
					break ;
				case 10:
					mBytesPerElement = 16 ;
					mNumElements = mDataSize/mBytesPerElement ;
					break ;
			}
		if (mFileSize != 0 && numElements != mNumElements) {
			mImageWidth = sqrt((float)mNumElements) ;
			mImageHeight = mNumElements/mImageWidth ;
			mImageHeight += (mImageWidth*mImageHeight == mNumElements ? 0 : (long)((mNumElements - mImageHeight*mImageWidth)/mImageWidth)) ; 

			sprintf(text,"%d",mImageWidth) ;
			mImageWidthF->SetEditText(text) ;
			sprintf(text,"%d",mImageHeight) ;
			mImageHeightF->SetEditText(text) ;
			SelectAll(mImageWidthF) ;
		}
	}
}

void TImportRawImageDialog::GetDialogData(void *data) {
	TImportRawImageDialogRec *theData = (TImportRawImageDialogRec *)data ;

	// Get the data from the dialog
	char text[256] ;
	mImageWidthF->GetEditText(text) ;
	mImageWidth = atoi(text) ;
	mImageHeightF->GetEditText(text) ;
	mImageHeight = atoi(text) ;
	mByteOffsetF->GetEditText(text) ;
	mByteOffset = atoi(text) ;
	mImageDataType = (ImageDataType)mDataTypeF->GetSelection() ;
	mSwapBytes = mSwapBytesF->GetControlValue() ;

	theData->imageWidth = mImageWidth ;
	theData->imageHeight = mImageHeight ;
	theData->dataType = mImageDataType ;
	theData->byteOffset = mByteOffset ;
	theData->swapBytes = mSwapBytes ;
	theData->numBytes = mImageWidth*mImageHeight*mBytesPerElement+mByteOffset ;

}

void TImportRawImageDialog::SetDialogData(void *data) {
	TImportRawImageDialogRec *theData = (TImportRawImageDialogRec *)data ;
	mImageWidth = theData->imageWidth ;
	mImageHeight = theData->imageHeight ;
	mImageDataType = theData->dataType ;
	mSwapBytes = theData->swapBytes ;
	mByteOffset = theData->byteOffset ;
	mBytesPerElement = (theData->numBytes-theData->byteOffset)/(mImageWidth*mImageHeight) ;
}


