#ifndef __TLattice_H
#define __TLattice_H
#include "TButton.h"
#include "TBaseObject.h"
#include "TVector.h"
#include "TPeakList.h"
#include <math.h>

class	TImage ;
class	TImageWindow ;
typedef enum {kXaxis,kYaxis} AxisType ;

class	TLattice : public TBaseObject {
	public:
		TLattice(FPoint &o, FPoint &u, FPoint &v) ;
		TLattice(TVector &o, TVector &u, TVector &v) ;
		TLattice(TLattice *u) : mOrigin(u->mOrigin),mU(u->mU),mV(u->mV) {mSignature = 'latt' ; mWindow = nil ;} ;
		TLattice() {mOrigin.x=mOrigin.y=mU.x=mU.y=mV.x=mV.y=0;mSignature = 'latt' ; mWindow = nil ;} ;
		void	Draw2DGrid(TImage *image) ;
		void	DrawUnitVectors(TImage *image) ;
		void	Draw2DVectors(TImage *image) ;
		void	Print() ;
		float	Refine(TPeakArray *list,float maxPeakDeviation) ;
        float   RefineBiweight(TPeakArray *list, float maxPeakDeviation) ;
		float	RefineQuad(TPeakArray *list,float maxPeakDeviation, float *BX, float *BY) ;
		float	RefineCubic(TPeakArray *list,float maxPeakDeviation, float *BX, float *BY, int failed) ;
        //float   RefineBiweightQuad(TPeakArray *list, float maxPeakDeviation, float *BX, float *BY) ;
        //float   RefineBiweightCube(TPeakArray *list, float maxPeakDeviation, float *BX, float *BY) ;
		float	RefineUV(TPeakArray *list,float maxPeakDeviation) ;
		FPoint	&GetOrigin() { return mOrigin ; } ;
		FPoint	&GetU() { return mU ; } ;
		FPoint	&GetV() { return mV ; } ;
		float	GetLatticeParameterU() { return sqrt(mU.x*mU.x+mU.y*mU.y) ; } ;
		float	GetLatticeParameterV() { return sqrt(mV.x*mV.x+mV.y*mV.y) ; } ;
		float	GetLatticeAngle() { return 90.0*acos((mU.x*mV.x+mU.y*mV.y)/(GetLatticeParameterU()*GetLatticeParameterV()))/asin(1.0) ; } ;
		void	Select(bool state) { mIsSelected = state ; } ;
		bool	IsSelected() { return mIsSelected ; } ;
		bool 	InLattice(Point &pt) ;
		void	SetWindow(TImageWindow *window) { mWindow = window ; } ;
                //int     RealHValue ( float x , float y ) ;
                //int     RealKValue ( float x , float y ) ;
                //float   RealXValue ( int n, int m ) ;		
                //float   RealYValue ( int n, int m ) ;
        // TLattice	*Copy() { TLattice *lattice = new TLattice(this) ; return lattice ; }  ; 
        TBaseObject	*Copy() { TLattice *lattice = new TLattice(this) ; return (TBaseObject *)lattice ; }  ; 
	protected:
		bool	mIsSelected ;
		
	private:
		TImageWindow	*mWindow ;
		FPoint	mOrigin ;
		FPoint	mU ;
		FPoint	mV ;
		Boolean InterceptX0(long Width,long Height,int n,Point *iSect,AxisType whichAxis) ;
		Boolean InterceptXH(long Width,long Height,int n,Point *iSect,AxisType whichAxis) ;
		Boolean InterceptYH(long Width,long Height,int n,Point *iSect,AxisType whichAxis) ;
		Boolean InterceptY0(long Width,long Height,int n,Point *iSect,AxisType whichAxis) ;
} ;

#endif // __TLattice_H
