// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TList.h"
#include <stdio.h>
#include "globals.h"
#include "XDrivers.h"

extern	XFontStruct	*gMenuFont ;
extern	XFontStruct	*gDialogFont ;
extern	XFontStruct	*gWindowFont ;
extern	short	gNumGrayLevels ;

long	near(float) ;

extern	uLong	gWhitePixel,gBlackPixel,gDarkerGray,gGrayPixel ;
#define EOS '\0'

TList::TList(
	Rect	*theListRect,short	dataSize,bool hasColumnLabels) 
	: TObject(theListRect)
	
{
	
	mHasColLabels = hasColumnLabels ;
	mHasRowLabels = false ;
	mAdornRows = false ;
	mAdornColumns = false ;
	mRows = 0 ;
	mCols = 1 ;
	mCellDataSize = dataSize ;
	mColWidth = mRect.right - mRect.left - 16 ;
	mRowHeight = 15 ;
	mCellData = nil ;
	
	short	width = mRect.right - mRect.left ;
	short	height = mRect.bottom - mRect.top ;
	
	SetRect(&mCanvasRect,0,0,width-16,height-16) ;

	
	mSelectedCell.row = 0 ;
	mSelectedCell.col = 0 ;

	mCellData = new LArray((UInt32)mCellDataSize);

	mListOrigin.x = 0. ;
	mListOrigin.y = 0. ;
	
	mVScrollPos = 0. ;
	mHScrollPos = 0. ;
	mVertScrollBar 	= nil ;
	mHorizScrollBar = nil ;
	mIsInitialized = false ;
				
}

TList::~TList() {
	delete mCellData ;
}

ArrayIndexT	TList::AddRow(void *item,short afterRow)  {
	// if afterRow == -1 , we append at the end
	ArrayIndexT index ;
	if (afterRow == -1) {
		index = mCellData->AddItem(item,mCellDataSize) ;
	}
	else {
		index = mCellData->InsertItemsAt(1,afterRow,item,mCellDataSize) ;
	}
	mRows++ ;
	Refresh() ;
	return index ;
}

void	TList::RemoveRow(short theRow) {
	mCellData->RemoveItemsAt(1,theRow) ;
	mRows-- ;
	Refresh() ;
}

void	
TList::GetCellText(ListCellT &inCell,char *text) {
	sprintf(text,"%d,%d",inCell.col,inCell.row) ;
}

void	
TList::SetCellText(ListCellT &inCell,char *text) {
}

void TList::Draw(RgnHandle updateRegion) 
{
	short row,col ;
	Point	topLeftUpdate,botRightUpdate,topLeft,botRight;
	Rect	totalUpdateRect ;
	
	if (!mParentWindow->IsVisible()) {
		return ;
	}
	
	if (updateRegion == nil) {
		totalUpdateRect = mCanvasRect ;
	}
	else {
		// Get the region associated with the update
		RegionRect(updateRegion,&totalUpdateRect) ;
		
		if (totalUpdateRect.top == 0 && totalUpdateRect.left == 0 &&
			totalUpdateRect.bottom == 0 && totalUpdateRect.right == 0) {
				return ;
		}
	}
	Rect	listContentRect = mCanvasRect ;
	OffsetRect(&listContentRect,mRect.left,mRect.top) ;
	TStOffscreen stOffscreen(mParentWindow,&listContentRect) ; 

	short rightEdge,bottomEdge,savedLeft ;
	rightEdge = mCanvasRect.right ;
	bottomEdge = mCanvasRect.bottom ;
	
	if (totalUpdateRect.right > rightEdge) {
		totalUpdateRect.right = rightEdge ;
	}
	if (totalUpdateRect.bottom > bottomEdge) {
		totalUpdateRect.bottom = bottomEdge ;
	}
	savedLeft = totalUpdateRect.left ;
	if (mHasRowLabels && mRows > 0) {
		DrawColLabel(0) ;
		if (totalUpdateRect.left < mColWidth) {
			totalUpdateRect.left = mColWidth ;
		}
	}

	SetClip(totalUpdateRect) ;
	
	if (mHasColLabels && mCols > 0) {
		// Find colum label at top left of update rect
		short	leftColumn,rightColumn;	
		// Calculate the column intersected by the top left point
		// of the update region
		topLeft.x = totalUpdateRect.left ;
		topLeft.y = totalUpdateRect.top ;
		if (topLeft.y <= mRowHeight) {
			//need to draw  column labels
			leftColumn = (topLeft.x-mListOrigin.x-1)/mColWidth ;
			if (leftColumn < 0) {
				leftColumn = 0 ;
			}
			
			botRight.x = totalUpdateRect.right ;
			rightColumn = (botRight.x-mListOrigin.x+1)/mColWidth ;
			if (rightColumn > mCols) {
				rightColumn = mCols ;
			}
			if (!mHasRowLabels) {
				leftColumn++ ;
				rightColumn++ ;
			}
			
			for(col=leftColumn;col<=rightColumn;col++) {
				DrawColLabel(col) ;
			}
		}
	}
	
	if (mHasColLabels && mCols > 0) {
		totalUpdateRect.left = savedLeft ;
		if (totalUpdateRect.top < mRowHeight)
			totalUpdateRect.top = mRowHeight ;
	}
	
	SetClip(totalUpdateRect) ;
	
	if (mHasRowLabels && mRows > 0) {
		// Find colum label at top left of update rect
		short	topRow,bottomRow;	
		// Calculate the column intersected by the top left point
		// of the update region
		topLeft.x = totalUpdateRect.left ;
		topLeft.y = totalUpdateRect.top ;
		if (topLeft.x <= mColWidth) {
			// need to draw  row labels
			topRow = (topLeft.y-mListOrigin.y-1)/mRowHeight ;
			if (topRow < 0) {
				topRow = 0 ;
			}
			
			botRight.y = totalUpdateRect.bottom ;
			bottomRow = (botRight.y-mListOrigin.y+1)/mRowHeight ;
			if (bottomRow > mRows) {
				bottomRow = mRows ;
			}

			if (!mHasColLabels) {
				topRow++ ;
				bottomRow++ ;
			}
			for(row=topRow;row<=bottomRow;row++) {
				DrawRowLabel(row) ;
			}
		}
	}
	
	if (updateRegion == nil) {
		totalUpdateRect = mCanvasRect ;
	}
	else {
		// Get the region associated with the update
		RegionRect(updateRegion,&totalUpdateRect) ;
	}

	if (totalUpdateRect.right > rightEdge) {
		totalUpdateRect.right = rightEdge ;
	}
	if (totalUpdateRect.bottom > bottomEdge) {
		totalUpdateRect.bottom = bottomEdge ;
	}
	if (totalUpdateRect.top < mRowHeight && mHasColLabels) {
		totalUpdateRect.top = mRowHeight ;
	}
	if (totalUpdateRect.left < mColWidth  && mHasRowLabels) {
		totalUpdateRect.left = mColWidth ;
	}

	SetClip(totalUpdateRect) ;
	// ---------------------------------------
	// Draw the cells that need to be updated
	// ---------------------------------------
	
	// Find cell at top left of update rect
	ListCellT	topLeftCell;	
	// Calculate the cell intersected by the top left point
	// of the update region
	topLeft.x = totalUpdateRect.left - 1 ;
	topLeft.y = totalUpdateRect.top - 1 ;
	LocalToImagePoint(topLeft,topLeftUpdate) ;			
	FetchCellHitBy(topLeftUpdate, topLeftCell);
	topLeftCell.row-- ;
	topLeftCell.col-- ;
	if (topLeftCell.row < 1) {	// Lower bound is cell (1,1)
		topLeftCell.row = 1;
	}
	if (topLeftCell.col < 1) {
		topLeftCell.col = 1;
	}

								// Find cell at bottom right of update rect
	ListCellT	botRightCell;
	botRight.x = totalUpdateRect.right + 1 ;
	botRight.y = totalUpdateRect.bottom + 1 ;
	LocalToImagePoint(botRight, botRightUpdate);
	FetchCellHitBy(botRightUpdate, botRightCell);
	botRightCell.row++ ;
	botRightCell.col++ ;
								// Upper bound is cell (mRows,mCols)
	if (botRightCell.row > mRows) {
		botRightCell.row = mRows;
	}
	if (botRightCell.col > mCols) {
		botRightCell.col = mCols;
	}
	
								// Draw each cell within the update rect
	ListCellT	cell;
	for (cell.row = topLeftCell.row; cell.row <= botRightCell.row; cell.row++) {
		for (cell.col = topLeftCell.col; cell.col <= botRightCell.col; cell.col++) {
			DrawCell(cell);
		}
	}

}


void TList::DrawCell(
	ListCellT &inCell) 
{
	Rect theRect ;
	Point CellImagePt,CellLocalPt ;
	
		
	CellImagePt.x = (mHasRowLabels ? inCell.col*mColWidth : (inCell.col-1)*mColWidth) ;
	CellImagePt.y = (mHasColLabels ? inCell.row*mRowHeight : (inCell.row-1)*mRowHeight) ;
	ImageToLocalPoint(CellImagePt,CellLocalPt) ;
	SetRect(&theRect,CellLocalPt.x,CellLocalPt.y,CellLocalPt.x+mColWidth,
		CellLocalPt.y+mRowHeight) ;
	if (!(mSelectedCell.row == inCell.row && mSelectedCell.col == inCell.col)) {			
		if (mAdornColumns) {
			SetFgColor(gGrayPixel) ;
			MoveTo(theRect.left,theRect.bottom) ;
			LineTo(theRect.left,theRect.top) ;
			if (inCell.col == mCols) {
				MoveTo(theRect.right,theRect.top) ;
				LineTo(theRect.right,theRect.bottom) ;
			}
		}
		if (mAdornRows) {
			MoveTo(theRect.left,theRect.top) ;
			LineTo(theRect.right,theRect.top) ;
			if (inCell.row == mRows) {
				MoveTo(theRect.left,theRect.bottom) ;
				LineTo(theRect.right,theRect.bottom) ;
			}
		}
		if (!mIsActive) SetFgColor(gGrayPixel) ;
		else SetFgColor(gBlackPixel) ;
		DrawCellValue(inCell,theRect) ;
	}
	else {
		if (mAdornColumns) {
			SetFgColor(gGrayPixel) ;
			MoveTo(theRect.left,theRect.bottom) ;
			LineTo(theRect.left,theRect.top) ;
			if (inCell.col == mCols) {
				MoveTo(theRect.right,theRect.top) ;
				LineTo(theRect.right,theRect.bottom) ;
			}
		}
		if (mAdornRows) {
			MoveTo(theRect.left,theRect.top) ;
			LineTo(theRect.right,theRect.top) ;
			if (inCell.row == mRows) {
				MoveTo(theRect.left,theRect.bottom) ;
				LineTo(theRect.right,theRect.bottom) ;
			}
		}
		if (!mIsActive) SetFgColor(gGrayPixel) ;
		else SetFgColor(gBlackPixel) ;
		FillRect(&theRect) ;
		SetFgColor(gWhitePixel) ;
		DrawCellValue(inCell,theRect) ;
		SetFgColor(gBlackPixel) ;
	}
}

void TList::DrawCellValue(
	ListCellT &inCell,
	Rect &cellFrame)
{
	char str[256] ;
	
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	FetchCellData(inCell,str) ;
	DrawString(str) ;
}

bool
TList::FetchLocalCellFrame(
	const ListCellT	&inCell,
	Rect				&outCellFrame)
{
								// Get Top-Left in Image coordinates
	Point	cellImage;
	
	if (mHasRowLabels)
		cellImage.x = inCell.col * mColWidth;
	else
		cellImage.x = (inCell.col - 1) * mColWidth;

	if (mHasColLabels)
		cellImage.y = inCell.row * mRowHeight;
	else
		cellImage.y = (inCell.row - 1) * mRowHeight;
	
	// Check if Cell intersects the Frame
	bool	intersectsFrame = ImageRectIntersectsFrame(
									cellImage.x, cellImage.y,
									cellImage.x + mColWidth,
									cellImage.y + mRowHeight);
	
	if (intersectsFrame) {		// Convert to Local coordinates
		ImageToLocalPoint(cellImage, topLeft(outCellFrame));
		outCellFrame.right = outCellFrame.left + mColWidth;
		outCellFrame.bottom = outCellFrame.top + mRowHeight;
	}
	
	return intersectsFrame;
}


bool
TList::IsValidCell(
	const ListCellT	&inCell) const
{
	return ( (inCell.row > 0) && (inCell.row <= mRows) &&
			 (inCell.col > 0) && (inCell.col <= mCols) );
}


void
TList::FetchCellHitBy(
	const Point		&inImagePt,
	ListCellT		&outCell)
{
	// Calculate the cell hit by the point (in image coordinates)
	Point CellPoint = inImagePt ;
	mCellClickPos.x = 0 ;
	mCellClickPos.y = 0 ;

	if (mHasRowLabels) {
		CellPoint.x -= mColWidth ;
	}
	if (mHasColLabels) {
		CellPoint.y -= mRowHeight ;
	}
	
	if (CellPoint.y < 0) 
		outCell.row = 0 ;
	else {
		outCell.row = (CellPoint.y-1)/ mRowHeight + 1 ;
		mCellClickPos.y = CellPoint.y - (outCell.row-1)*mRowHeight ;
	}
	if (CellPoint.x < 0) 
		outCell.col = 0 ;
	else {
		outCell.col = (CellPoint.x-1)/ mColWidth + 1 ;
		mCellClickPos.x = CellPoint.x - (outCell.col-1)*mColWidth ;
	}
	
}

void TList::SetCellData(
	ListCellT &inCell, 
	void *data) 
{
	Int32 cellIndex ;
	cellIndex = (inCell.row-1)*mCols+inCell.col ;
	mCellData->AssignItemsAt(1,cellIndex,data) ;
}

bool TList::FetchCellData(
	ListCellT &inCell,
	void *data) 
{
	Int32 cellIndex ;
	cellIndex = (inCell.row-1)*mCols+inCell.col ;
	return(mCellData->FetchItemAt(cellIndex,data)) ;
}

void TList::SetDataCells(
	ListCellT &inFromCell, 
	ListCellT &inToCell,
	void *data) 
{
	ListCellT cell ;
	char *dataLoc ;
	
	dataLoc = (char *)data ;
	for(cell.row=inFromCell.row;cell.row < inToCell.row;cell.row++) {
		for(cell.col=inFromCell.col;cell.col < inToCell.col;cell.col++) {
			SetCellData(cell,(void *)dataLoc) ;
			dataLoc += mCellDataSize ;
		}
	}
} 
void TList::FetchDataCells(
	ListCellT &inFromCell, 
	ListCellT &inToCell,
	void *data) 
{
	ListCellT cell ;
	char *dataLoc ;
	
	dataLoc = (char *)data ;
	for(cell.row=inFromCell.row;cell.row < inToCell.row;cell.row++) {
		for(cell.col=inFromCell.col;cell.col < inToCell.col;cell.col++) {
			FetchCellData(cell,(void *)dataLoc) ;
			dataLoc += mCellDataSize ;
		}
	}
} 

void TList::ClickCell(
	ListCellT &inCell) 
{
	if (inCell.row > 0 && inCell.row <= mRows) {
		mSelectedCell = inCell ;
	}
	else {
		mSelectedCell.row = mSelectedCell.col = 0 ;
	}
	Refresh() ;
}

void	TList::ClickColLabel(short inCol) {
}


bool TList::DoClickInLabels(
	Point &inPoint) 
{

	ListCellT theCell ;
	
	// Se if we are in any of the column or row labels
	FetchCellHitBy(inPoint,theCell) ;
	if (mHasColLabels && theCell.row == 0) {
		if (theCell.col > 0 && theCell.col <= mCols) 
			ClickColLabel(theCell.col) ;
		return true ;
	}
	
	return false ;
}

bool 
TList::ImageRectIntersectsFrame(
	short left,
	short top,
	short right,
	short bottom) 
{
	Point topLeft, bottomRight ;
	
	
	topLeft.x = near(left + mListOrigin.x) ;
	topLeft.y = near(top + mListOrigin.y) ;
	
	bottomRight.x =  near(right + mListOrigin.x) ;
	bottomRight.y =  near(bottom + mListOrigin.y) ;
	
	Rect frame = mCanvasRect ;
	if (mHasRowLabels) {
		frame.left += mColWidth ;
	}
	if (mHasColLabels) {
		frame.top += mRowHeight ;
	}
	frame.right -= 16 ;	// Because of border possibly containing scrollbar
	frame.bottom -= 16 ; // likewise
	
	return (PtInRect(&topLeft,&frame) || PtInRect(&bottomRight,&frame)) ;
}

void 	
TList::ImageToLocalPoint(
	Point &inImagePt,
	Point &outLocalPt) 
{
	outLocalPt.x =  near(inImagePt.x + mListOrigin.x) ;
	outLocalPt.y =  near(inImagePt.y + mListOrigin.y) ;
}

void 	
TList::LocalToImagePoint(
	Point &inLocalPt,
	Point &outImagePt) 
{
	outImagePt.x =  near(inLocalPt.x - mListOrigin.x) ;
	outImagePt.y =  near(inLocalPt.y - mListOrigin.y) ;
}

void TList::DoClick(
	Point &inPoint) 
{

	ListCellT theCell ;
	FetchCellHitBy(inPoint,theCell) ;
	ClickCell(theCell) ;
		
}


void TList::HighlightCell(ListCellT &inCell) 
{
}


void TList::DrawColLabel(
	short colNumber) 
{
	Rect	cellFrame;
	
	if (mHasRowLabels)
		SetRect(&cellFrame,colNumber*mColWidth,0,(colNumber+1)*mColWidth,mRowHeight) ;
	else
		SetRect(&cellFrame,(colNumber-1)*mColWidth,0,colNumber*mColWidth,mRowHeight) ;
		
	if (colNumber != 0) 
		::OffsetRect(&cellFrame,near(mListOrigin.x),0) ;
	InsetRect(&cellFrame,1,1) ;
	SetFgColor(gGrayPixel) ;
	PaintRect(&cellFrame) ;
	SetFgColor(gDarkerGray) ;
	MoveTo(cellFrame.left,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.top) ;
	if (!mIsActive) SetFgColor(gGrayPixel) ;
	else SetFgColor(gBlackPixel) ;
	MoveTo(cellFrame.right,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.bottom) ;
	InsetRect(&cellFrame,-1,-1) ;
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	if (colNumber > 0) {
		DrawColLabelText(colNumber) ;
	}
}
void TList::DrawRowLabel(
	short rowNumber) 
{
	Rect	cellFrame;
	
	if (mHasColLabels)
		SetRect(&cellFrame,0,rowNumber*mRowHeight,mColWidth,(rowNumber+1)*mRowHeight) ;
	else
		SetRect(&cellFrame,0,(rowNumber-1)*mRowHeight,mColWidth,rowNumber*mRowHeight) ;
	if (rowNumber != 0)
		::OffsetRect(&cellFrame,0,near(mListOrigin.y)) ;
	InsetRect(&cellFrame,1,1) ;
	SetFgColor(gGrayPixel) ;
	PaintRect(&cellFrame) ;
	SetFgColor(gDarkerGray) ;
	MoveTo(cellFrame.left,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.top) ;
	if (!mIsActive) SetFgColor(gGrayPixel) ;
	else SetFgColor(gBlackPixel) ;
	MoveTo(cellFrame.right,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.bottom) ;
	InsetRect(&cellFrame,-1,-1) ;
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	if (rowNumber > 0) {
		DrawRowLabelText(rowNumber) ;
	}
}

void 
TList::DrawRowLabelText(
	short rowNumber) 
{
	char str[64] ;
	sprintf(str,"R%d",rowNumber) ;
	DrawString(str) ;
}

void 
TList::DrawColLabelText(
	short colNumber) 
{
	char str[64] ;
	sprintf(str,"C%d",colNumber) ;
	DrawString(str) ;
}


void TList::Draw() 
{


	if (!mIsVisible) return ;
	if (mParentWindow == nil) return ;
	if (!mParentWindow->IsVisible()) return ;

	SetPort(mParentWindow) ;	
	SetClip(mRect) ;

	if (!mIsInitialized	) {
		AdjustScrollBars() ;
		mIsInitialized = true ;
	}
	
	// Draw the inner content of the list
	Draw(nil) ;

	SetClip(mRect) ;
	
	SetFgColor(gDarkerGray) ;
	FrameRect(&mRect) ;
	
	Rect	theRect  ;
	SetFgColor(gGrayPixel) ;
	SetRect(&theRect,mRect.right-16,0,mRect.right,mRect.bottom) ;
	FillRect(&theRect) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(mRect.right-16,0,mRect.right-16,mRect.bottom) ;
	SetFgColor(gGrayPixel) ;
	SetRect(&theRect,0,mRect.bottom-16,mRect.right,mRect.bottom) ;
	FillRect(&theRect) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(0,mRect.bottom-16,mRect.right,mRect.bottom-16) ;
	DrawLine(mRect.right-1,0,mRect.right-1,mRect.bottom) ;
	DrawLine(0,mRect.bottom-1,mRect.right,mRect.bottom-1) ;

	SetFgColor(gWhitePixel) ;
	DrawLine(mRect.right-16,mRect.bottom-16,mRect.right-16,mRect.bottom) ;
	DrawLine(mRect.right-16,mRect.bottom-16,mRect.right,mRect.bottom-16) ;
	SetFgColor(gBlackPixel) ;
	
	if (mHorizScrollBar != nil) mHorizScrollBar->Draw() ;
	if (mVertScrollBar != nil) mVertScrollBar->Draw() ;
			
	SetFgColor(gBlackPixel) ;
	SetClip(mParentWindow->GetPortRect()) ;
}

void TList::AdjustScrollBars() {
	
	// Check to see if we need the scrollbar(s) 
	// If not, delete them, otherwise adjust their size and position
	long	listWidth = mHasRowLabels ? mColWidth*(mCols+1) : mColWidth*mCols ;
	long	listHeight = mHasColLabels ? mRowHeight*(mRows+1) : mRowHeight*mRows ;
	
	long	windowWidth = mRect.right - mRect.left ;
	long	windowHeight = mRect.bottom - mRect.top ;
		
	bool	hasVertScrollBar = listHeight > windowHeight-16 ? true : false ;
	bool	hasHorizScrollBar = listWidth > windowWidth-16 ? true : false ;
	
	short	horizWidth = windowWidth - 16 ;
	short	vertHeight = windowHeight - 16 ;

	short 	vertScrollMax = listHeight - vertHeight ;
	short 	horizScrollMax = listWidth - horizWidth ;
	
	float	vertScrollPos = 0. ;
	float	horizScrollPos = 0. ;
		
	if (hasHorizScrollBar) {
		// We need a horizontal scrollbar
		// Do we have one already
		if (mHorizScrollBar != nil)
			mHorizScrollBar->Resize(mRect.left,mRect.top+windowHeight-16,mRect.left+horizWidth,mRect.top+windowHeight,horizScrollMax) ;
		else {
			// We need to create one
			mHorizScrollBar =  new TScrollBar(mRect.left,mRect.top+windowHeight-16,
					mRect.left+horizWidth,mRect.top+windowHeight,kHorizontal,0,horizScrollMax,0);
			mHorizScrollBar->SetTrackingRoutine(TList::DoHScrollBar) ;
			mHorizScrollBar->SetRefCon((unsigned long)this) ;
			mHorizScrollBar->SetParentWindow(mParentWindow) ;
		}
		horizScrollPos	= mHorizScrollBar->GetControlValue() ;
	}
	else {
		// Do we have one already
		if (mHorizScrollBar != nil) {
			delete mHorizScrollBar ;
			mHorizScrollBar = nil ;
		}
	}

	if (hasVertScrollBar) {
		// We need a vertical scrollbar
		// Do we have one already
		if (mVertScrollBar != nil)
			mVertScrollBar->Resize(mRect.left+windowWidth-16,mRect.top,mRect.left+windowWidth,mRect.top+vertHeight,vertScrollMax) ;
		else {
			// We need to create one
			mVertScrollBar =  new TScrollBar(mRect.left+windowWidth-16,mRect.top,
					mRect.left+windowWidth,mRect.top+vertHeight,kVertical,0,vertScrollMax,0);
			mVertScrollBar->SetTrackingRoutine(TList::DoVScrollBar) ;		
			mVertScrollBar->SetRefCon((unsigned long)this) ;
			mVertScrollBar->SetParentWindow(mParentWindow) ;
		}
		vertScrollPos	= mVertScrollBar->GetControlValue() ;
	}
	else {
		// Do we have one already
		if (mVertScrollBar != nil) {
			delete mVertScrollBar ;
			mVertScrollBar = nil ;
		}
	}

	SetRect(&mCanvasRect,0,0,horizWidth,vertHeight) ;
	SetCanvasRect(mCanvasRect) ;
	mListOrigin.x = -horizScrollPos ;
	if (mListOrigin.x > 0.) mListOrigin.x = 0. ;
	mListOrigin.y = -vertScrollPos ;
	if (mListOrigin.y > 0.) mListOrigin.y = 0. ;

}



void TList::ScrollListBy(
	float deltaX,
	float deltaY) 
{
	mListOrigin.x -= deltaX ;
	if (mListOrigin.x > 0.)
		mListOrigin.x = 0. ;
	mListOrigin.y -= deltaY ;
	if (mListOrigin.y > 0.)
		mListOrigin.y = 0. ;
	Refresh() ;
}

void TList::ScrollListTo(
	float x,
	float y) 
{
	mListOrigin.x = -x ;
	if (mListOrigin.x > 0.)
		mListOrigin.x = 0. ;
	mListOrigin.y = -y ;
	if (mListOrigin.y > 0.)
		mListOrigin.y = 0. ;
	Refresh() ;
}

void
TList::Refresh() 
{
	GrafPtr	tPort ;
	GetPort(&tPort) ;
	SetPort(mParentWindow) ;
	AdjustScrollBars() ;
	Draw() ;
	SetPort(tPort) ;
}


void 
TList::DoHScrollBar(TObject	*object) {
	float	newVal ;
	float	scrollDistance ;
	TList	*theList ;
	
	theList = (TList *)object->GetRefCon() ;
	newVal	= theList->mHorizScrollBar->GetControlValue() ;

	scrollDistance = newVal - theList->mHScrollPos ;
	if (scrollDistance != 0.) {
		theList->mHScrollPos = newVal ;
		theList->ScrollListTo(theList->mHScrollPos,theList->mVScrollPos) ;
	}
} 

void 
TList::DoVScrollBar(TObject	*object) {
	float	newVal ;
	float	scrollDistance ;
	TList	*theList ;
	
	theList = (TList *)object->GetRefCon() ;
	newVal	= theList->mVertScrollBar->GetControlValue() ;

	scrollDistance = newVal - theList->mVScrollPos ;
	if (scrollDistance != 0) {
		theList->mVScrollPos = newVal ;
		theList->ScrollListTo(theList->mHScrollPos,theList->mVScrollPos) ;
	}
}	

Boolean TList::DoMouseDown(
	Point *inPoint) 
{
	if (mHorizScrollBar != nil) {
		if (mHorizScrollBar->InObject(inPoint)) {
			mHorizScrollBar->DoMouseDown(inPoint) ;
			return true ;
		}
	}
	if (mVertScrollBar != nil) {
		if (mVertScrollBar->InObject(inPoint)) {
			mVertScrollBar->DoMouseDown(inPoint) ;
			return true ;
		}
	}
	
	inPoint->x -= mRect.left ;		
	inPoint->y -= mRect.top ;		
	Point imagePoint ;
	LocalToImagePoint(*inPoint,imagePoint) ;
	if (!DoClickInLabels(imagePoint)) {
		DoClick(imagePoint) ;
	}	
	return true ;	
}
