// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------

#include <config.h>
#include "hand_cursor.bmp"
#include "hand_cursor_mask.bmp"
#include "ruler.bmp"
#include "pointer.bmp"
#include "zoom_cursor.bmp"
#include "zoom_cursor_mask.bmp"
#include "spot_cursor.bmp"
#include "spot_cursor_mask.bmp"
#include "selection.bmp"
#include "CircleMask.bmp"
#include "SpotMask.bmp"
#include "LineMask.bmp"
#include "WedgeMask.bmp"
#include "LatticeMask.bmp"
#include "lattice.bmp"
#include "FLattice.bmp"
#include "Aperture.bmp"
#include "Motif.bmp"
#include "XDrivers.h"

#include "TMainApplication.h"
#include "TDMApplicationWindow.h"
#include "TImage.h"
extern	short	gImageDepth ;
extern	short	gNumGrayLevels ;

char	*FS98_PATH = PREFIX ;
char	*BrowserName = BROWSER ;
char	*EDMPath ;
char	*EDMBinDir ;
char	*EDMDocDir ;
char	*EDMBrowser ;
char	edm_install_path[256] ;
char	edm_browser_name[256] ;

class TTestImage : public TImage {
	public :
		TTestImage() ;
		void CreateOffscreen() ;
} ;		

TDMApplicationWindow *gTheControlWindow = nil ;
void	SendApplicationCommand(TCommand command) ;

void	SendApplicationCommand(TCommand command) {
	Boolean	hasBeenHandled ;
	gTheControlWindow->DoCommand(command,&hasBeenHandled) ;
}

TTestImage *gTestImage = nil ;


TMainApplication::TMainApplication(int argc,char *argv[])  
	: TApplication(argc,argv) {
	
	
    sprintf(edm_install_path,"%s",PREFIX) ;
    // If PREFIX environnment variable is set, use it
    if ((EDMPath = getenv("PREFIX")) != NULL) {
	    sprintf(edm_install_path,"%s",EDMPath) ;
		FS98_PATH = edm_install_path ;
    }

    if ((EDMDocDir = getenv("EDM_DOCDIR")) == NULL) {
    	EDMDocDir = "/documentation/";
    }

    if ((EDMBinDir = getenv("EDM_BINDIR")) == NULL) {
    	EDMBinDir = "/bin/";
    }

    sprintf(edm_browser_name,"%s",BROWSER) ;
    EDMBrowser = edm_browser_name ;
    // If BROWSER environnment variable is set, use it
    if ((BrowserName = getenv("BROWSER")) != NULL) {
	    sprintf(edm_browser_name,"%s",BrowserName) ;
		EDMBrowser = edm_browser_name ;
    }
	DefineWindows() ; 

    if (argc > 1) {
	    for (int i=1;i<argc;i++) {
			if (strncmp(argv[i],"-t",2) == 0) {
				gTestImage = new TTestImage ;
				
			}
	    }
    }

} 
 
TMainApplication::~TMainApplication() {}

void TMainApplication::DefineWindows() {

	gTheControlWindow = new TDMApplicationWindow(0,0,240,742,gImageDepth,PACKAGE_STRING); //"edm 2.0") ;
	mApplicationWindow = gTheControlWindow ;
	AddWindow(gTheControlWindow) ;
}

#define __MAIN__

#include "types.h"
#include "globals.h"
void	ProcessModalEvent(XEvent *event) ;
TMainApplication	*gTheApp ; 

void	AddWindowToWindowsMenu(TWindow *theWindow) ;
void	RemoveWindowFromWindowsMenu(TWindow *theWindow) ;
void	DefineCursors() ;

// Create our cursors
Cursor	gZoomCursor ;
Cursor	gHandCursor ;
Cursor	gSelectionCursor ;
Cursor	gSpotCursor ;
Cursor	gSpotMaskCursor ;
Cursor	gLatticeMaskCursor ;
Cursor	gWedgeMaskCursor ;
Cursor	gCircleMaskCursor ;
Cursor	gLineMaskCursor ;
Cursor	gRulerCursor ;
Cursor	gApertureCursor ;

void	DefineCursors() {
	
	Pixmap shape_zoom,mask_zoom ;
	Pixmap shape_hand,mask_hand ;
	Pixmap shape_spot,mask_spot ;
		
	shape_zoom = XCreatePixmapFromBitmapData(gDisplay,RootWindow(gDisplay,gScreen),(char *)zoom_cursor_bits,
		zoom_cursor_width,zoom_cursor_height,1,0,1) ;
	mask_zoom = XCreatePixmapFromBitmapData(gDisplay,RootWindow(gDisplay,gScreen),(char *)zoom_cursor_mask_bits,
		zoom_cursor_mask_width,zoom_cursor_mask_height,1,0,1) ;
	gZoomCursor = XCreatePixmapCursor(gDisplay,shape_zoom,mask_zoom,&gColors[0],&gColors[gNumGrayLevels-1],zoom_cursor_x_hot,zoom_cursor_y_hot) ;

	
	shape_hand = XCreatePixmapFromBitmapData(gDisplay,RootWindow(gDisplay,gScreen),(char *)hand_cursor_bits,
		hand_cursor_width,hand_cursor_height,1,0,1) ;
	mask_hand = XCreatePixmapFromBitmapData(gDisplay,RootWindow(gDisplay,gScreen),(char *)hand_cursor_mask_bits,
		hand_cursor_mask_width,hand_cursor_mask_height,1,0,1) ;
	gHandCursor = XCreatePixmapCursor(gDisplay,shape_hand,mask_hand,&gColors[0],&gColors[gNumGrayLevels-1],hand_cursor_x_hot,hand_cursor_y_hot) ;


	shape_spot = XCreatePixmapFromBitmapData(gDisplay,RootWindow(gDisplay,gScreen),(char *)spot_cursor_bits,
		spot_cursor_width,spot_cursor_height,1,0,1) ;
	mask_spot = XCreatePixmapFromBitmapData(gDisplay,RootWindow(gDisplay,gScreen),(char *)spot_cursor_mask_bits,
		spot_cursor_mask_width,spot_cursor_mask_height,1,0,1) ;
	gSpotCursor = XCreatePixmapCursor(gDisplay,shape_spot,mask_spot,&gColors[0],&gColors[gNumGrayLevels-1],spot_cursor_x_hot,spot_cursor_y_hot) ;
		
	gSelectionCursor = XCreateFontCursor(gDisplay,XC_crosshair) ;

	gSpotMaskCursor = XCreateFontCursor(gDisplay,XC_target) ;
	gLatticeMaskCursor = XCreateFontCursor(gDisplay,XC_ll_angle) ;
	gWedgeMaskCursor = XCreateFontCursor(gDisplay,XC_X_cursor) ;
	gCircleMaskCursor = XCreateFontCursor(gDisplay,XC_circle) ;
	gLineMaskCursor = XCreateFontCursor(gDisplay,XC_X_cursor) ;
	gRulerCursor = XCreateFontCursor(gDisplay,XC_tcross) ;
	gApertureCursor = XCreateFontCursor(gDisplay,XC_cross) ;
	
	XFreePixmap(gDisplay,shape_zoom) ;
	XFreePixmap(gDisplay,mask_zoom) ;
	XFreePixmap(gDisplay,shape_hand) ;
	XFreePixmap(gDisplay,mask_hand) ;
	XFreePixmap(gDisplay,shape_spot) ;
	XFreePixmap(gDisplay,mask_spot) ;
}


void	AddWindow(TWindow *window) ;
void	AddWindow(TWindow *window) {
	gTheApp->AddWindow(window) ;
	AddWindowToWindowsMenu(window) ;
	
}
void	RemoveWindow(TWindow *window) {
	gTheApp->RemoveWindow(window) ;
}
int main(int argc,char *argv[]) {

	gTheApp = new TMainApplication(argc,argv) ;	
	AddWindowToWindowsMenu(gTheControlWindow) ;
	
	DefineCursors() ;
	
	if (gTestImage != nil) gTestImage->Display() ;

	gTheApp->MainLoop() ;
	delete	gTheApp ;
	return 0 ;
}

void	ProcessModalEvent(XEvent *event) {
	gTheApp->ProcessModalEvent(event) ;
}

Boolean DrawTestImage(unsigned short *image,int x,int y,int width,int height) ;

TTestImage::TTestImage()
	: TImage(256,256) {
	SetName("Color test") ;
	
}

void TTestImage::CreateOffscreen() {	
	long i,j ;
	
	
	unsigned short *ucp ;

	SetPort(mWindow) ;
	
	GrafPtr tPort ;
	GetPort(&tPort) ;
	Rect	theRect ;
	SetRect(&theRect,0,0,mWidth,mHeight) ;

	unsigned short numGrays = 	gNumGrayLevels ;
	unsigned short *mappedDisplayData = new unsigned short[mWidth*mHeight] ;
	
	ucp = mappedDisplayData ;
	
	unsigned long index ;
	
	
	ucp = mappedDisplayData ;				
	for(i=0;i<mHeight;i++) {
		for(j=0;j<mWidth;j++) {
			if (i < mHeight/2 && j < mWidth/2)
				index = gNumGrayLevels ;
			else if (i < mHeight/2 && j < mWidth)
				index = gNumGrayLevels+1 ;
			else if (i < mHeight && j < mWidth/2)
				index = gNumGrayLevels+2 ;
			else
				index = j % numGrays ;			
			*ucp++ = index ;
		}
	}


	mOffscreen = new TOffscreen(tPort,&theRect) ;

	if (mOffscreen == nil)  {
		fprintf(stderr,"\nmOffscreen == nil Memory Allocation Error") ; 
		fflush(stderr) ;
		return ;
	}

	SetPort(mOffscreen) ;
	DrawTestImage(mappedDisplayData,0,0,mWidth,mHeight) ;
	delete[] mappedDisplayData ;
	
	SetPort(tPort) ;
		
}
#define kMM		0x4D4D
#define kINTEL	0x4949
typedef	enum {kMMByteOrder = kMM, kIntelByteOrder = kINTEL} ByteOrder ;
extern	ByteOrder	gCPUByteOrder ;
typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
extern MonitorDisplayType  gMonitorDisplayType ;
extern Visual	*gVisual ;
Boolean DrawTestImage(unsigned short *image,int x,int y,int width,int height) {
	long 		i,j,ndx ;
	XImage 		*myImage ;

	unsigned char *imageRGB = nil ;
 	if (gMonitorDisplayType == PSEUDOCOLOR) {
	
		// Create a XImage for writing the image into 
		myImage = XCreateImage(gDisplay,
			gVisual,gImageDepth,ZPixmap,
			0,(char *)image,width,height,8,0) ;
	}
	else {
		imageRGB = (unsigned char *) new unsigned long[width*height] ;
		if (imageRGB == NULL)
			return(0) ;
		switch(gCPUByteOrder) {
			case kMMByteOrder:
				switch(gImageDepth) {
					case 15:
					case 16:
						myImage = XCreateImage(gDisplay,
							gVisual,gImageDepth,ZPixmap,
							0,(char *)imageRGB,width,height,16,0) ;
						ndx = 0 ;	
						for(j=0; j < height; j++) 
							for(i=0; i < width; i++) {
								XPutPixel(myImage,i,j,gColors[image[ndx]].pixel) ;
								ndx++ ;
							}	
						break ;
					case 24:
					case 32:
						myImage = XCreateImage(gDisplay,
							gVisual,gImageDepth,ZPixmap,
							0,(char *)imageRGB,width,height,32,0) ;
						ndx = 0 ;	
						for(j=0; j < height; j++) 
							for(i=0; i < width; i++) {
								XPutPixel(myImage,i,j,gColors[image[ndx]].pixel) ;
								ndx++ ;
							}	
						break ;
				}
				break ;
			case kIntelByteOrder:
				switch(gImageDepth) {
					case 15:
					case 16:
						myImage = XCreateImage(gDisplay,
							gVisual,gImageDepth,ZPixmap,
							0,(char *)imageRGB,width,height,16,0) ;
						ndx = 0 ;	
						for(j=0; j < height; j++) 
							for(i=0; i < width; i++) {
								XPutPixel(myImage,i,j,gColors[image[ndx]].pixel) ;
								ndx++ ;
							}	
							break ;
					case 24:
					case 32:
						myImage = XCreateImage(gDisplay,
							gVisual,gImageDepth,ZPixmap,
							0,(char *)imageRGB,width,height,32,0) ;
						ndx = 0 ;	
						for(j=0; j < height; j++) 
							for(i=0; i < width; i++) {
								XPutPixel(myImage,i,j,gColors[image[ndx]].pixel) ;
								ndx++ ;
							}	
						break ;
				}
				break ;
		}
	}
	
	if (myImage == (XImage *)0) {
		if (imageRGB != nil) delete[] imageRGB ;
		return(false) ;
	}
	XPutImage(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),myImage,0,0,x,y,width,height) ;
	XFlush(gDisplay) ;
	myImage->data = NULL ;
	XDestroyImage(myImage) ;
	if (gMonitorDisplayType == TRUECOLOR) {
		if (imageRGB != nil) delete[] imageRGB ;
	}
	return(true) ;
}
