// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TMenuBar.h"
#include "globals.h"
#include "XDrivers.h"
#include <stdio.h>
#include "EventNames.h"


extern	short	gDebugMode ;

TMenuBar::TMenuBar(TPane *pane,short width,short height) {
	mNumMenus = 0 ;
	mMenus = nil ;
	mWidth = width ;
	mHeight = height ;
	mWindowPane = pane ;
	mFgColor = mWindowPane->GetFgColor() ;
	mBkColor = mWindowPane->GetBkColor() ;
	XSetWindowAttributes attr ;
	attr.backing_store = WhenMapped ;
	XChangeWindowAttributes(gDisplay,mWindowPane->GetXWindow(),CWBackingStore,&attr) ;

}

void	TMenuBar::AddMenu(TMenu *menu) {
	short	x,y ;
	TMenu *theMenu ;

	x = 0 ;
	y = mHeight-1 ;
	menu->mY = mHeight-1 ;
	if (mNumMenus == 0) {
		menu->SetPosition(x,y) ;
		mMenus = menu ;
	}
	else {
		x = mMenus->mX ;
		// We must link this menu to the ones we have
		for(theMenu = mMenus; theMenu->mNextMenu != nil; 
				theMenu = theMenu->mNextMenu) {
			x += theMenu->mTitleWidth ;
		}
		x += theMenu->mTitleWidth ;
		menu->SetPosition(x,y) ;
		theMenu->mNextMenu = menu ;
	}
	menu->SetParent(mWindowPane) ;
	mNumMenus++ ;			
}

TMenu	*TMenuBar::GetMenu(Point *loc) {
	int i,x,y ;
	Rect tRect ;
	TMenu *theMenu ;
	theMenu = mMenus ;
	for (i=0;i<mNumMenus;i++) {
		x = theMenu->mX ;
		y = 0 ;
		SetRect(&tRect,x,y,x+theMenu->mTitleWidth,mHeight) ;
		if (PtInRect(loc,&tRect)) {
			if (theMenu->mEnabled)
				return theMenu ;
			else
				return nil ;
		}
	}
	return nil ;
}

void	TMenuBar::Draw() {
	int i ;
	TMenu *theMenu ;
	GrafPtr tPort ;

	GetPort(&tPort) ;
	SetPort(mWindowPane) ;

	theMenu = mMenus ;
	for (i=0;i<mNumMenus;i++) {
		DrawMenuTitle(theMenu,false) ;
		theMenu = theMenu->mNextMenu ;
	}
	SetPort(tPort) ;
}

TCommand	TMenuBar::DoMouseDown(Point *where) {
	// Search through the list of sub-menus to find out in
	// which we had a mousedown	
	int 	i,x,y,rx,ry,wx,wy ;
	TMenu 	*theMenu,*lastMenu,*theSelectedMenu,*currentMenuBarMenu,*lastMenuBarMenu ;
	TMenuItem	*menuItem,*currentMenuItem=nil,*lastMenuItem = nil ;
	Rect 	tRect ;
	unsigned int	keyButt ;
	Window	child,root ;
	Boolean	menuIsDrawn,intheMenu,inMenuBar ;
	short	lastMenuNumber,lastMenuItemNumber,selectedMenuNumber,
    		selectedMenuItemNumber,j ;
	Point	currentPt ;
	GrafPtr	tPort ;
	extern	Display	*gDisplay ;


	TCommand commandNum = 0 ;
	
	// The menu items are 1 based so we have 0 for selectedItem 
	// and lastmenuitem at start
	selectedMenuItemNumber	= 0 ;
	lastMenuItemNumber	= 0 ;

	selectedMenuNumber	= -1 ;
	lastMenuNumber		= -1 ;
	menuIsDrawn			= false ;
	intheMenu			= false ;
	inMenuBar			= false ;

	GetPort(&tPort) ;
	SetPort(mWindowPane) ;
	theMenu = mMenus ;
	
	// Let us sit in an event loop only looking for mouse moved and click events
	Boolean	done = false ;
	Boolean	pseudoMouseIsDown = true ;
	extern	Time	gClickTime ;
	extern	bool	gMouseIsDown ;
	XEvent	event ;
	XButtonEvent *buttEvent = (XButtonEvent *)&event ;
	
	XGrabPointer(gDisplay,mWindowPane->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;

	// We have a mouse down event already
	XQueryPointer(gDisplay,mWindowPane->GetXWindow(),&root,&child,&rx,&ry,&wx,&wy,&keyButt) ;
	intheMenu = false ;
	currentPt.x = wx ;
	currentPt.y = wy ;
	SetRect(&tRect,0,0,mWidth,mHeight) ;
	// The only place we can be at the beginning is within the menubar
	if (PtInRect(&currentPt,&tRect)) {	// Within the menubar	
		inMenuBar = false ;
		theMenu = mMenus ;
		for (i=0;i<mNumMenus;i++) {
			x = theMenu->mX ;
			y = 0 ;
			SetRect(&tRect,x,y,x+theMenu->mTitleWidth,mHeight) ;
			if (PtInRect(&currentPt,&tRect) && theMenu->mEnabled) { 
				inMenuBar = true ;
				selectedMenuNumber = i ;
				lastMenu = theSelectedMenu = theMenu ;
				currentMenuBarMenu = theSelectedMenu ;
				lastMenuBarMenu = currentMenuBarMenu ;
				currentMenuItem = nil ;							
				lastMenuItem = nil ;							
				lastMenuItemNumber = 0 ;
				selectedMenuNumber = 0 ;
				intheMenu = false ;
				DrawMenuTitle(currentMenuBarMenu,true) ;
				currentMenuBarMenu->Draw() ;
				menuIsDrawn = true ;
				lastMenuNumber = selectedMenuItemNumber ;
				break ;
			}
			theMenu = theMenu->mNextMenu ;
		}
	}
			
	while (!done) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonPress:
				if (pseudoMouseIsDown) done = true ;				
				break ;
			case ButtonRelease:
				gMouseIsDown = false ;
				if ((buttEvent->time - gClickTime) < 400) 
					pseudoMouseIsDown = true ;
				else
					done = true ;
				break ;
			case MotionNotify:
				if (!XQueryPointer(gDisplay,mWindowPane->GetXWindow(),&root,&child,&rx,&ry,&wx,&wy,&keyButt)) break ;
				currentPt.x = wx ;
				currentPt.y = wy ;
				inMenuBar = false ;
				intheMenu = false ;
				// ---------------------- Handle inside menubar begin -------------------------
				SetRect(&tRect,0,0,mWidth,mHeight) ;
				if (PtInRect(&currentPt,&tRect)) {	// Within the menubar	
					theMenu = mMenus ;
					for (i=0;i<mNumMenus;i++) {
						x = theMenu->mX ;
						y = 0 ;
						SetRect(&tRect,x,y,x+theMenu->mTitleWidth,mHeight) ;
						if (PtInRect(&currentPt,&tRect) && theMenu->mEnabled) { 
							inMenuBar = true ;
							selectedMenuNumber = i ;
							currentMenuBarMenu = theSelectedMenu = theMenu ;
							if (currentMenuBarMenu != lastMenuBarMenu) {	// We are in a new menu
								lastMenuItemNumber = 0 ;
								selectedMenuItemNumber = 0 ;
								intheMenu = false ;
								if (menuIsDrawn) {
									lastMenuBarMenu->Hide() ;
    					    		// Unhighligh Old title 
    					    		DrawMenuTitle(lastMenuBarMenu,false) ;
								}
								currentMenuBarMenu = theMenu ;
								lastMenuBarMenu = lastMenu = theSelectedMenu = currentMenuBarMenu ;
								DrawMenuTitle(currentMenuBarMenu,true) ;
								currentMenuBarMenu->Draw() ;
								menuIsDrawn = true ;
								currentMenuItem = nil ;							
								lastMenuItem = nil ;							
								lastMenuNumber = selectedMenuNumber ;
								break ;
							}
							else if (menuIsDrawn && currentMenuItem != nil) {
								lastMenuItemNumber = 0 ;
								selectedMenuItemNumber = 0 ;
								currentMenuItem->Draw(false) ;
								currentMenuItem = nil ;
								lastMenuItem = nil ;							
							}
						}
						theMenu = theMenu->mNextMenu ;
					}
				}
				if (inMenuBar) break ;
				if (!menuIsDrawn) break ;
				// ----------------------- Handle inside menubar end -------------------------
				
				
				// ------------------------ Handle inside menus begin ------------------------
				intheMenu = false ;
				if (theSelectedMenu->mPopUpWindow != nil) {
					if (!XQueryPointer(gDisplay,theSelectedMenu->mPopUpWindow->GetDrawable(),
						&root,&child,&rx,&ry,&wx,&wy,&keyButt))
						break ;
					currentPt.x = wx ;
					currentPt.y = wy ;
					menuItem = theSelectedMenu->mMenuItems ;
					// ------------- Check to see if we are inside the current menu --------------------
					for (j=1;j<=theSelectedMenu->mNumMenuItems;j++) {
						tRect = menuItem->mRect ;
						if (PtInRect(&currentPt,&tRect)) {
							intheMenu = true ;
							selectedMenuItemNumber = j ;
							if (selectedMenuItemNumber != lastMenuItemNumber && theSelectedMenu == lastMenu) {
								// Unhighlight old 
								currentMenuItem = theSelectedMenu->GetMenuItem(selectedMenuItemNumber) ;
								if (lastMenuItem != nil && lastMenuItem->mHierarchicalMenu !=  theSelectedMenu) {
									lastMenuItem->Draw(false) ;
								}
								// Highlight new
								lastMenuItem = currentMenuItem ;
								currentMenuItem->Draw(true) ;
								commandNum = currentMenuItem->mEnabled ? currentMenuItem->mCommand : 0 ;
								// Check to see if we just clicked in a hierarchical menu
								// If we did, then we switch to a new selected menu
								if (currentMenuItem->mIsHierarchicalMenu) {
									theSelectedMenu = currentMenuItem->mHierarchicalMenu ;
									lastMenuItemNumber = 0 ;
									break ;
								}
								lastMenuItemNumber = j ;
								lastMenu = theSelectedMenu ;
								break ;
							}
							lastMenu = theSelectedMenu ;
						}
						menuItem = menuItem->mNextItem ;
					}
					if (intheMenu) break ;
					// ------------- we were inside the current menu --------------------


					// -------------We are not inside the currently seleced menu --------------------
					// ------------- Check to see if we are inside a hierarchical menu --------------------
					if (currentMenuItem != nil) {
						// Are we inside the item which is the "title" of the hierarchical menu
						if (currentMenuItem->mIsHierarchicalMenu) {
							theMenu = currentMenuItem->mParentMenu ;
							// Check to see if we are within a new item in the old menu
							XQueryPointer(gDisplay,theMenu->mPopUpWindow->GetDrawable(),
								&root,&child,&rx,&ry,&wx,&wy,&keyButt) ;
							currentPt.x = wx ;
							currentPt.y = wy ;
							menuItem = theMenu->mMenuItems ;
							for (j = 1; j <= theMenu->mNumMenuItems; j++) {
								tRect = menuItem->mRect ;
								if (PtInRect(&currentPt,&tRect)) {
									intheMenu = true ;
									selectedMenuItemNumber = j ;
									currentMenuItem = theMenu->GetMenuItem(selectedMenuItemNumber) ;
									if (currentMenuItem != lastMenuItem) {
										// Unhighlight old 
										if (lastMenuItem != nil) {
											lastMenuItem->Draw(false) ;
										}
										// Highlight new
										theSelectedMenu = theMenu ;
										lastMenu = theSelectedMenu ;
										currentMenuItem = theSelectedMenu->GetMenuItem(selectedMenuItemNumber) ;
										currentMenuItem->Draw(true) ;
										lastMenuItem = currentMenuItem ;
										commandNum = currentMenuItem->mEnabled ? currentMenuItem->mCommand : 0 ;
										// Are we moving to a new hierarchical menu 
										if (currentMenuItem->mIsHierarchicalMenu) {
											theSelectedMenu = currentMenuItem->mHierarchicalMenu ;
											lastMenuItemNumber = 0 ;
											break ;
										}
										lastMenuItemNumber = j ;
									}
									break ;
								}
								menuItem = menuItem->mNextItem ;
							}
							if (intheMenu) break ;
						}
						else {
							// Let us check the parent menu of the current menu item
							theMenu = theSelectedMenu->mParentMenu ;
							if (theMenu != nil) {
								// Check to see if we are in the parent menu
								XQueryPointer(gDisplay,theMenu->mPopUpWindow->GetDrawable(),
									&root,&child,&rx,&ry,&wx,&wy,&keyButt) ;
								currentPt.x = wx ;
								currentPt.y = wy ;
								menuItem = theMenu->mMenuItems ;
								for (j=1;j<= theMenu->mNumMenuItems;j++) {
									tRect = menuItem->mRect ;
									if (PtInRect(&currentPt,&tRect)) {
										// We are inisde a parent menu
										// Are we inside the the menuitem that we belong to
										intheMenu = true ;
										if (menuItem->mIsHierarchicalMenu && menuItem->mParentMenu == theMenu) {
											// Unhighlight the one we came from
											currentMenuItem->Draw(false) ;
											currentMenuItem = menuItem ;
											lastMenuItem = currentMenuItem ;
											break ; // Same one we came 
										}
										selectedMenuItemNumber = j ;
										// Unhighlight old 
										currentMenuItem->Draw(false)  ;
										// Also turn off the item that is the hierarchical menu from where we came
										if (currentMenuItem->mParentMenu != nil)
											if (currentMenuItem->mParentMenu->mParentMenuItem != nil)
												currentMenuItem->mParentMenu->mParentMenuItem->Draw(false) ;
										// Highlight new
										theSelectedMenu = theMenu ;
										lastMenu = theSelectedMenu ;
										currentMenuItem = theSelectedMenu->GetMenuItem(selectedMenuItemNumber) ;
										lastMenuItem = currentMenuItem ;
										currentMenuItem->Draw(true) ;
										commandNum = currentMenuItem->mEnabled ? currentMenuItem->mCommand : 0 ;
										lastMenuItemNumber = 0 ;
										break ;
									}
									menuItem = menuItem->mNextItem ;
									
								}
								if (intheMenu) break ;
							}
						}
						if (!intheMenu) {
							theSelectedMenu->HighlightMenuItem(lastMenuItemNumber,false) ; 
							selectedMenuItemNumber = 0 ;
							lastMenuItemNumber = 0 ;
						}
					}
				}
				// ----------------------------- menu end ------------------------------------
				break ;
			default:
				break ;
		}
	
	}
	
	where->x = selectedMenuNumber ;
	where->y = selectedMenuItemNumber ;
		
	if (strcmp(theSelectedMenu->mTitle,"?") == 0) {
		// If this is the help menu we just respond to the click and report which the command kHelp
		where->x = 0 ;
		where->y = 1 ;
		commandNum = kHelp ;
	}
	if (theSelectedMenu != nil) {
		// Hide the menu 
		if (theSelectedMenu->mParentMenu != nil)
			theSelectedMenu->mParentMenu->Hide() ;
		theSelectedMenu->Hide() ;
	}
	if (lastMenuBarMenu != nil) {
		// Hide the menu 
		lastMenuBarMenu->Hide() ;
		DrawMenuTitle(lastMenuBarMenu,false) ;
	}
	gMouseIsDown = false ;
	
	XUngrabPointer(gDisplay,CurrentTime) ;
	SetPort(tPort) ;
	return commandNum ;
}

void TMenuBar::DrawMenuTitle(TMenu *menu,Boolean state) {

	if (menu->mParentMenu != nil) return ; // This is a hierarchical menu

	Rect	tRect  ;
	SetRect(&tRect,menu->mX-4,0,
		menu->mX+menu->mTitleWidth,mHeight) ;	

	if (!menu->mEnabled) {
		mWindowPane->SetFgColor(gDarkerGray) ;
		XDrawString(gDisplay,mWindowPane->GetDrawable(),
			mWindowPane->GetGC(),menu->mTitlePosition.x,menu->mTitlePosition.y,
    			menu->mTitle,strlen(menu->mTitle)) ;
		mWindowPane->SetFgColor(gBlackPixel) ;
		XFlush(gDisplay) ;
		return ;
	}

	if (state == ON) {
		if (!menu->mHighlighted)  menu->mHighlighted = true ;	
		mWindowPane->SetFgColor(gLighterGray) ;
		InsetRect(&tRect,1,1) ;
		XFillRectangle(gDisplay,mWindowPane->GetDrawable(),mWindowPane->GetGC(),
			tRect.left,tRect.top,tRect.right-tRect.left,
			tRect.bottom-tRect.top) ;
		mWindowPane->SetFgColor(gWhitePixel) ;
		XDrawLine(gDisplay,mWindowPane->GetDrawable(),
			mWindowPane->GetGC(),tRect.left,tRect.bottom,tRect.left,tRect.top) ;
		XDrawLine(gDisplay,mWindowPane->GetDrawable(),
			mWindowPane->GetGC(),tRect.left,tRect.top,tRect.right,tRect.top) ;
		InsetRect(&tRect,1,2) ;
		mWindowPane->SetFgColor(gDarkerGray) ;
		XDrawLine(gDisplay,mWindowPane->GetDrawable(),
			mWindowPane->GetGC(),tRect.right,tRect.top,tRect.right,tRect.bottom) ;
		XDrawLine(gDisplay,mWindowPane->GetDrawable(),
			mWindowPane->GetGC(),tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
		InsetRect(&tRect,-1,-1) ;
		mWindowPane->SetFgColor(gBlackPixel) ;
		XDrawLine(gDisplay,mWindowPane->GetDrawable(),
			mWindowPane->GetGC(),tRect.right,tRect.top,tRect.right,tRect.bottom) ;
		XDrawLine(gDisplay,mWindowPane->GetDrawable(),
			mWindowPane->GetGC(),tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
		if (strcmp(menu->mTitle,"?") == 0) {
			mWindowPane->SetFgColor(gRedPixel) ;
			XDrawString(gDisplay,mWindowPane->GetDrawable(),
				mWindowPane->GetGC(),menu->mTitlePosition.x,menu->mTitlePosition.y,
	    			"H",strlen("H")) ;			
			mWindowPane->SetFgColor(gBlackPixel) ;
		}
		else {
			XDrawString(gDisplay,mWindowPane->GetDrawable(),
				mWindowPane->GetGC(),menu->mTitlePosition.x,menu->mTitlePosition.y,
	    			menu->mTitle,strlen(menu->mTitle)) ;
    	}
	}
	else {
		if (menu->mHighlighted) {
			InsetRect(&tRect,0,1) ;
			XClearArea(gDisplay,mWindowPane->GetDrawable(),tRect.left,tRect.top,
				tRect.right-tRect.left,tRect.bottom-tRect.top,0) ;
			menu->mHighlighted = false ;
		}
		if (strcmp(menu->mTitle,"?") == 0) {
			mWindowPane->SetFgColor(gRedPixel) ;
			XDrawString(gDisplay,mWindowPane->GetDrawable(),
				mWindowPane->GetGC(),menu->mTitlePosition.x,menu->mTitlePosition.y,
	    			"H",strlen("H")) ;			
			mWindowPane->SetFgColor(gBlackPixel) ;
		}
		else {
			XDrawString(gDisplay,mWindowPane->GetDrawable(),
				mWindowPane->GetGC(),menu->mTitlePosition.x,menu->mTitlePosition.y,
	    			menu->mTitle,strlen(menu->mTitle)) ;
    	}
	}
	XFlush(gDisplay) ;

}
