#ifndef __TPeakList__H
#define __TPeakList__H
#include "types.h"

typedef enum {kRawPeak,kGaussianPeak,kSincPeak,kCMPeak,kHanningPeak } PeakForm ;
typedef enum {kSortAscending,kSortDescending} SortDirection ;

typedef enum {kThresholdOnMeanImageValue=1,kThresholdOnMeanAroundSpot,kNoThreshold} PeakCriteria ;
//typedef enum {kThresholdOnMeanImageValue=1,kThresholdOnMeanAroundSpot,kThresholdOnStdv,kNoThreshold} PeakCriteria ;
typedef enum {kUseXCF=1, kUseXCFMask, kUseXCFAll, kUseIntegration,kUseIntegrationBZM,kUseIntegrationBZ } IntensityMeasurementMethod ;

typedef	enum {kFoundPeak,kAddedPeak,kDeletedPeak} PeakType ;
class	TImage 
;
class TPeak {
	public:
		TPeak() ;
		TPeak(TPeak &peak) ;
		FPoint 	pos ;
		float	value ;
		short	n ;
		short	m ;
		float   Mean ;
		float	Rfactor ;
		PeakType	peakType ;
		TPeak& operator=(TPeak &peak) ;
} ;

class TPeakEntry {
	public:
		TPeakEntry() ;
		~TPeakEntry() ;
		TPeakEntry	*GetNextPeakEntry() { return mNextPeakEntry ; } ;
		TPeak		*GetPeak() { return mPeak ; } ;
		friend class TPeakList ;
	private:
		TPeak		*mPeak ;
		TPeakEntry	*mNextPeakEntry ; 
} ;

class TPeakList {
	public:
		TPeakList() ;
		~TPeakList() ;
		void 	AddPeak(TPeak &peak) ;
		void 	AddPeak(TPeak *peak) ;
		void	RemovePeak(TPeak *peak) ;
		Boolean	RemovePeak(Point &pos) ;
		void	RemovePeak(int numberToRemove) ;
		TPeakEntry	*GetFirstPeakEntry() { return mPeakEntry ; } ;
		short	GetNumberOfPeaks() { return mNumPeaks ; } ;
		TPeak	*GetPeak(int number) ;
		TPeak	*GetPeak(Point &pos) ;
	private:
		short	mNumPeaks ;
		TPeakEntry	*mPeakEntry ;
	
}  ;

typedef struct PeakCriteraRecord {
	float	minValue ;
	float	maxValue ;
	short	cmRadius ;
	short	distanceFromEdgde ;
	short	distanceBetweenPeaks ;
	Boolean	findMax ;

} PeakCriteraRecord ;

class	TPeakArray {
	public:
		TPeakArray(TPeakList *peakList) ;
		TPeakArray(long num) 
			: mNumPeaks(num) { mX = new float[num]; 
				mY = new float[num]; mN = new short[num]; mM = new short[num]; 
				mValue = new float[num]; mMean = new float[num] ; mRfactor = new float[num] ; } ;
		~TPeakArray() { delete[] mX ; delete[] mY ; delete[] mValue ; delete[] mN ; delete[] mM ;
				delete[] mMean ; delete[] mRfactor ; } ;
		
		void	Sort(SortDirection direction) ;
		long	mNumPeaks ;
		float	*mX ;
		float	*mY ;
		short	*mN ;
		short	*mM ;
		float	*mMean ;
		float	*mRfactor ;
		float	*mValue ;
} ;

TPeakArray *FindPeaks(TImage *image,PeakCriteraRecord *peakCriteria,Rect *selection = nil ) ;
PeakCriteria GetPeakCriteria() ;
IntensityMeasurementMethod GetIntensityMeasurementMethod() ;
bool	UseFriedelSymmetry() ;
bool	UseZero() ;
#endif // __TPeakList__
