// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------

#include <stdio.h>
#include "TWindowList.h"
#include "globals.h"
#include "XDrivers.h"
#include "TPotentialMapWindow.h"
#include "DMMenuCommands.h"

void	RemoveWindow(TWindow *) ;
void	AddWindow(TWindow *) ;
extern	short	gImageDepth ;
extern	long	near(float) ;
extern	void	PutMessage(char *) ;

float	gPotentialMapCalibration = 0. ;

#define kUnAssignedPhase	-9999.

CTFRecord	gCTFRecord = {300.,0.65,-400.,20.,0.15} ;

TCreatePotentialMapImageDialog::TCreatePotentialMapImageDialog(UnitCellDisplayRecord *data)
	: TDialog(kDialogPosX,kDialogPosY,300,100,"Create Potential Map Image") {
	mData = *data ;
	
	mTextNumA = new TEditText(100,20,140,40,"# Unit Cells Wide","3") ;
	mTextNumB = new TEditText(240,20,280,40,"# Unit Cells High","3") ;

	AddTextEditObject(mTextNumB) ;
	AddTextEditObject(mTextNumA) ;
	
	mDefaultButton = mOKButton = new TButton(200,60,280,80,"OK") ;
	AddObject(mOKButton) ;

	mCancelButton = new TButton(90,60,190,80,"Cancel") ;
	AddObject(mCancelButton) ;
		
}

void TCreatePotentialMapImageDialog::RetrieveDialogItems() {
	char text[256] ;
	mTextNumA->GetEditText(text) ;
	mData.numCellsA = atoi(text) ;
	
	mTextNumB->GetEditText(text) ;
	mData.numCellsB = atoi(text) ;
}

void	TPotentialDataTable::DoCommand(TCommand command,Boolean *hasBeenHandled) {
	switch(command) {
		case kHelp:
			OnCommandHelp() ;		
			*hasBeenHandled = true ;
			break ;
		case kCopy:
			DoCopy() ;
			*hasBeenHandled = true ;
			break ;	
		case kPaste:
			DoPaste() ;
			*hasBeenHandled = true ;
			break ;	
		default:
			*hasBeenHandled = false ;
			break ;
	}
}

void	TPotentialDataTable::GetCommandStatus(TCommand command,Boolean *isEnabled,
			Boolean *isChecked) {
	switch(command) {
		case kCopy:
			*isEnabled = HasSelection() ;
			break ;	
		case kPaste:
			*isEnabled = HasSelection() && gTableClipBoardList != nil ;
			break ;	
		default:
			*isEnabled = false ;
			break ;
	}
}

void	TPotentialDataTable::DoCopy() {
	register	long i ;
	
	if (gTableClipBoardList != nil) delete gTableClipBoardList ;
	gTableClipBoardList = new TTableClipBoardList(mNumReflections) ;
	switch(mTableSelection.mTopLeftCell.col) {
		case 3:	// Amplitudes 
			gTableClipBoardList->mClipboardObjectType = 'ampl' ;
			for(i=0; i < mNumReflections; i++) {
				gTableClipBoardList->mIndex1[i] = mH[i] ;
				gTableClipBoardList->mIndex2[i] = mK[i] ;
				gTableClipBoardList->mValue[i] = mAmplitude[i] ;
			}
			break ;
		case 4: // Phases 
			gTableClipBoardList->mClipboardObjectType = 'phas' ;
			for(i=0; i < mNumReflections; i++) {
				gTableClipBoardList->mIndex1[i] = mH[i] ;
				gTableClipBoardList->mIndex2[i] = mK[i] ;
				gTableClipBoardList->mValue[i] = mPhase[i] ;
			}
			break ;
		case 5: // Adjusted Phase 
			gTableClipBoardList->mClipboardObjectType = 'phas' ;
			for(i=0; i < mNumReflections; i++) {
				gTableClipBoardList->mIndex1[i] = mH[i] ;
				gTableClipBoardList->mIndex2[i] = mK[i] ;
				gTableClipBoardList->mValue[i] = mCTFCorrectedPhase[i] ;
			}
			break ;
	}
	gTableClipBoardList->mCalibration = mIndexCalibrationRecord ;
	ZeroSelection() ;

}

void	TPotentialDataTable::DoPaste() {
	// Fill in the phases from the table
	register long i,j ;
	float	scaleFactor,maxAmplitude = 0. ;	
	short	iH,iK ;
	long	II,JJ ;
	short	h,k ;	
	
	switch(mTableSelection.mTopLeftCell.col) {
		case 3:	// Amplitudes 
			
			for(j=0; j < mNumReflections; j++) {
				h = mH[j] ;
				k = mK[j] ;
				if (h != 0 && k != 0 ) {
					if (mAmplitude[j] > maxAmplitude) {
						for(i=0; i < gTableClipBoardList->mNumValues ; i++) {
							if (h == gTableClipBoardList->mIndex1[i] && k == gTableClipBoardList->mIndex2[i]) {
								maxAmplitude = mAmplitude[j] ;
								iH = h ;
								iK = k ;
								II = i ;
								JJ = j ;
								break ;
							}
						}
					}
				}
			}
				
			scaleFactor = mAmplitude[JJ]/gTableClipBoardList->mValue[II] ;
	
			for(i=0; i < mNumReflections; i++) {
				h = mH[i] ;
				k = mK[i] ;
				for(j=0; j < gTableClipBoardList->mNumValues ; j++) {
					if (gTableClipBoardList->mIndex1[j] == h && gTableClipBoardList->mIndex2[j] == k) {
						mAmplitude[i] = gTableClipBoardList->mValue[j]*scaleFactor ;
						break ;
					}
				}
			}
			break ;
		case 4: // Phases 
			for(i=0; i < mNumReflections; i++) {
				h = mH[i] ;
				k = mK[i] ;
				for(j=0; j < gTableClipBoardList->mNumValues ; j++) {
					if (gTableClipBoardList->mIndex1[j] == h && gTableClipBoardList->mIndex2[j] == k) {
						mPhase[i] = gTableClipBoardList->mValue[j] ;
						break ;
					}
				}
			}
			break ;
		case 5: // Adjusted Phases 
			for(i=0; i < mNumReflections; i++) {
				h = mH[i] ;
				k = mK[i] ;
				for(j=0; j < gTableClipBoardList->mNumValues ; j++) {
					if (gTableClipBoardList->mIndex1[j] == h && gTableClipBoardList->mIndex2[j] == k) {
						mCTFCorrectedPhase[i] = gTableClipBoardList->mValue[j] ;
						break ;
					}
				}
			}
			break ;
	}
		
	mTableSelection.ZeroSelection() ;
	Refresh() ;
}



TPotentialDataTable::TPotentialDataTable(TableAttributes *attr,TDiffractionPeaks *diffractionPeaks,short planeGroup) 
	: TTable(attr) {
	short 	h,k,hDataMax,kDataMax,hStart,groupIndex ;
	float	amplitude,astar,bstar,gammastar,twocsgastar,astarSq,bstarSq,ksqbstsq,
			astbstcsgast,kbst2csgast,csgastar ;
	long	i ;

	mDiffractionPeaks = new TDiffractionPeaks(diffractionPeaks) ;
	TArray	*dataArray = diffractionPeaks->mFourierData ;
	
	mArrayWidth		= dataArray->GetWidth() ;
	mArrayHeight	= dataArray->GetHeight() ;
	hDataMax = mArrayWidth/2-1 ;
	kDataMax = mArrayHeight/2-1 ;
	
	mPlaneGroup = planeGroup ;
	mGamma = 180-mDiffractionPeaks->mGammaStar ;
	if (mPlaneGroup > 2 && mPlaneGroup < 13) mGamma = 90. ;
	if (mPlaneGroup > 12) mGamma = 120. ;
	
	mNumReflections = attr->nRows ;
	mGVector 	= new float[mNumReflections] ;
	mAmplitude 	= new float[mNumReflections] ;
	mPhase 		= new float[mNumReflections] ;
	mH			= new short[mNumReflections] ;
	mK			= new short[mNumReflections] ;
	mIndex		= new short[mNumReflections] ;
	mCTFCorrectedPhase 		= new float[mNumReflections] ;
	
	
	for (i=0; i < mNumReflections;  mIndex[i++] = 0.)  ;				

	switch(planeGroup) {
		case 1:	// p1
		case 2: // p2
			astar = diffractionPeaks->mAStar ;
			bstar = diffractionPeaks->mBStar ;
			gammastar = diffractionPeaks->mGammaStar ;
			break ;
		case 3: // pm -  
		case 4: // pg 
		case 5: // cm 
		case 6: // pmm  -
		case 7: // pmg  
		case 8: // pgg  
		case 9: // cmm  
			astar = diffractionPeaks->mAStar ;
			bstar = diffractionPeaks->mBStar ;
			gammastar = 90. ;
			break ;
		case 10: // p4
		case 11: // p4m
		case 12: // p4g
			astar = (diffractionPeaks->mAStar+diffractionPeaks->mBStar)/2. ;
			bstar = astar ;
			gammastar = 90. ;
			break ;
		case 13: // p3
		case 14: // p3m1
		case 15: // p31m
		case 16: // p6
		case 17: // p6m
			// Gamma = 120
			astar = (diffractionPeaks->mAStar+diffractionPeaks->mBStar)/2. ;
			bstar = astar ;
			gammastar = 60. ;
			break ;
	}
	
	astarSq = astar*astar ;
	bstarSq = bstar*bstar ;
	csgastar = cos(gammastar*asin(1.0)/90.) ;
	twocsgastar = 2.*cos(gammastar*asin(1.0)/90.) ;

	mIndexCalibrationRecord.astar = astar ;
	mIndexCalibrationRecord.bstar = bstar ;
	mIndexCalibrationRecord.gammastar = gammastar ;
	mIndexCalibrationRecord.isCalibrated = mDiffractionPeaks->mCalibrationUnit == kInverseAngstrom ? true : false ;
	
	i = 0 ;
	
	TArray	*IsSet = new TArray(2*dataArray->GetWidth(),2*dataArray->GetHeight()) ;
	*IsSet = 0. ;
	
	groupIndex = 1 ;
	astbstcsgast = astar*bstar*twocsgastar ;
	float	radToDegree = 90.0/asin(1.0) ;
	// Do only upper half plane
	for(k = 0; k < kDataMax; k++) {
		ksqbstsq = k*k*bstarSq ;
		kbst2csgast = k*astbstcsgast ;
		hStart = k == 0 ? 1 : -hDataMax ;
		for(h = hStart; h < hDataMax; h++) {
			if (!(h == 0 && k == 0)) {
				amplitude = *dataArray->GetRealHK(h,k) ;
				if (amplitude > 0.001) {
					mGVector[i] = ::sqrt(h*h*astarSq + ksqbstsq + h*kbst2csgast) ;
					mAmplitude[i] = amplitude ;
					mPhase[i] = radToDegree*(*dataArray->GetImagHK(h,k)) ;
					if (fabs(mPhase[i]) < 0.01) mPhase[i] = 0. ;
					if (fabs(180.0-fabs(mPhase[i])) < 0.01) mPhase[i] = 180. ;
					mCTFCorrectedPhase[i] = mPhase[i] - 180.0  ;
					if (fabs(mCTFCorrectedPhase[i]) > 360.) {
						mCTFCorrectedPhase[i] -= 360.*(int)(mCTFCorrectedPhase[i]/360.) ;
					}
					if (fabs(mCTFCorrectedPhase[i]) < 0.01) mCTFCorrectedPhase[i] = 0. ;
					if (fabs(180.0-fabs(mCTFCorrectedPhase[i])) < 0.01) mCTFCorrectedPhase[i] = 180. ;
					mH[i] = h ;
					mK[i] = k ;
					if (*IsSet->GetDataHK(h,k) == 0.) {
						mIndex[i] = groupIndex ;
						*IsSet->GetDataHK(h,k) = (float)groupIndex ;
						// Set symmetry related reflection
						switch(mPlaneGroup) {
							case 1:	// p1
							case 2: // p2
								break ;
							case 3: // pm	|F(h,k)| = |F(-h,k)|
							case 4: // pg	|F(h,k)| = |F(-h,k)|
							case 5: // cm	|F(h,k)| = |F(-h,k)|
							case 6: // pmm	|F(h,k)| = |F(-h,k)|
							case 7: // pmg  |F(h,k)| = |F(-h,k)|
							case 8: // pgg 	|F(h,k)| = |F(-h,k)|
							case 9: // cmm	|F(h,k)| = |F(-h,k)|
								*IsSet->GetDataHK(-h,k) = groupIndex ;
								break ;
							case 10: // p4	|F(h,k)| = |F(-k,h)|
								*IsSet->GetDataHK(-k,h) = groupIndex ;
								*IsSet->GetDataHK(k,-h) = groupIndex ;
								break ;
							case 11: // p4m |F(h,k)| = |F(-k,h)| = |F(-h,k)|
							case 12: // p4g	|F(h,k)| = |F(-k,h)| = |F(-h,k)|
								*IsSet->GetDataHK(-k,h) = groupIndex ;
								*IsSet->GetDataHK(-h,k) = groupIndex ;
								*IsSet->GetDataHK(-h,-k) = groupIndex ;
								*IsSet->GetDataHK(k,-h) = groupIndex ;
								*IsSet->GetDataHK(h,-k) = groupIndex ;
								break ;
							case 13: // p3	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
								*IsSet->GetDataHK(k,-h-k) = groupIndex ;
								*IsSet->GetDataHK(-h-k,h) = groupIndex ;
								*IsSet->GetDataHK(-h,-k) = groupIndex ;
								*IsSet->GetDataHK(-k,h+k) = groupIndex ;
								*IsSet->GetDataHK(h+k,-h) = groupIndex ;
								break ;
							case 14: // p3m1	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
							case 15: // p31m	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
							case 17: // p6m		|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								*IsSet->GetDataHK(k,-h-k) = groupIndex ;
								*IsSet->GetDataHK(-h-k,h) = groupIndex ;
								*IsSet->GetDataHK(k,h) = groupIndex ;
								*IsSet->GetDataHK(-h,-k) = groupIndex ;
								*IsSet->GetDataHK(-k,h+k) = groupIndex ;
								*IsSet->GetDataHK(h+k,-h) = groupIndex ;
								*IsSet->GetDataHK(-k,-h) = groupIndex ;
								break ;
							case 16: // p6	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
								*IsSet->GetDataHK(k,-h-k) = groupIndex ;
								*IsSet->GetDataHK(-h-k,h) = groupIndex ;
								*IsSet->GetDataHK(-h,-k) = groupIndex ;
								*IsSet->GetDataHK(-k,h+k) = groupIndex ;
								*IsSet->GetDataHK(h+k,-h) = groupIndex ;
								break ;
						}
						groupIndex++ ;
					}
					else {
						mIndex[i] = (short) *IsSet->GetDataHK(h,k) ;
					}
					i++ ;
				}
			}
		}
	}
	mNumReflections = i ;

	
	delete IsSet ;
	
	mSortKey = kSortOnG ;
	SortOnG() ;		
}

TPotentialDataTable::~TPotentialDataTable() {
	if (mGVector != nil) delete[] mGVector ;
	if (mAmplitude != nil) delete[] mAmplitude ;
	if (mPhase != nil) delete[] mPhase ;
	if (mCTFCorrectedPhase != nil) delete[] mCTFCorrectedPhase ;
	if (mH != nil) delete[] mH ;
	if (mK != nil) delete[] mK ;
	if (mIndex != nil) delete[] mIndex ;
	delete	mDiffractionPeaks ;
}

void	TPotentialDataTable::Calibrate(IndexCalibrationRecord &cbr) {

	mIndexCalibrationRecord = cbr ;
	long h,k ;
	float astar = cbr.astar ;
	float bstar = cbr.bstar ;
	float csgammastar = cos(cbr.gammastar*asin(1.0)/90.) ;
	for(long i=0; i < mNumReflections; i++) {
		h = mH[i] ;
		k = mK[i] ;
		mGVector[i] = sqrt(h*h*astar*astar+k*k*bstar*bstar+2*h*k*astar*bstar*csgammastar) ;		
	}
	
}
void	TPotentialDataTable::ApplyCTFCorrections(CTFRecord &microscope,bool useScherzerLens,float gmax) {

	float	u,usq,arg,arg1,pi,x,phase,sinChi,dampBeamSpread,dampEnergySpread,envelopeFunction ;
	long	I ; //LDM: changed to I to eliminate redefinition (was i)
	bool	adjustPhase ;
	if (useScherzerLens) {
		for(I=0; I < mNumReflections; I++) {
			adjustPhase = gmax == 0. ? true : (mGVector[I] > gmax ? false : true ) ;
			if (adjustPhase) {
				mCTFCorrectedPhase[I] = mPhase[I] - 180 ;
				if (fabs(mCTFCorrectedPhase[I]) > 360.) {
					mCTFCorrectedPhase[I] -= 360.*(int)(mCTFCorrectedPhase[I]/360.) ;
				}
				if (fabs(mCTFCorrectedPhase[I]) < 0.01) mCTFCorrectedPhase[I] = 0. ;
				if (fabs(180.0-fabs(mCTFCorrectedPhase[I])) < 0.01) mCTFCorrectedPhase[I] = 180. ;
			}
			else {
				mCTFCorrectedPhase[I] = kUnAssignedPhase ;
			}
			
		}
		return ;
	}
			
    float	cs = microscope.cs*1.e+7 ;
    float	th = microscope.div*1.e-3/sqrt(log(2.0)) ;
    float	df = microscope.defocus ;
    float	del = microscope.delta/sqrt(2*log(2.0)) ;
    float	wl = 12.263652/sqrt((microscope.voltage+9.785025e-4*
            microscope.voltage*microscope.voltage)*1000.) ;
	
	pi = 2*asin(1.0) ;	
	for(int i=0; i < mNumReflections; i++) {
		u = mGVector[i] ;
		usq = u*u ;
		arg = pi*wl*usq*(df+0.5*cs*wl*wl*usq) ;
		sinChi = sin(arg) ;
		
		arg1 = pi*wl*del*usq ;
		arg1 = 0.5*arg1*arg1 ; 
		dampEnergySpread =  exp(-arg1) ;
		x = pi*th*u*(df + cs*wl*wl*usq) ;
		x = x*x  ;
		dampBeamSpread = exp(-x) ;
		
		// Only correct phases so that is sin(Chi) is negative, we subtract 180
		// and if sin(chi) is positive, we keep the phase 
		phase = mPhase[i] ;
		
		envelopeFunction = dampBeamSpread*dampEnergySpread ;
		if (envelopeFunction > 0.01) {
			mCTFCorrectedPhase[i] = phase ;
			if (sinChi < 0.) {
				mCTFCorrectedPhase[i] = phase - 180 ;
			}
			if (fabs(mCTFCorrectedPhase[i]) < 0.01) mCTFCorrectedPhase[i] = 0. ;
			if (fabs(180.0-fabs(mCTFCorrectedPhase[i])) < 0.01) mCTFCorrectedPhase[i] = 180. ;
		}
		else {
			mCTFCorrectedPhase[i] = kUnAssignedPhase ;
		}
	}
}

void	TPotentialDataTable::ClickColLabel(Int16 inCol) {
	TableCellT	firstSelectedCell,lastSelectedCell ;
	switch (inCol) {
		case 2:		// Sort on g-vector
			if (mSortKey != kSortOnG) {
				SortOnG() ;
				mSortKey = kSortOnG ;
				Refresh() ;
			}
			break ;
		case 3:		// Sort on amplitude
			if (!ControlKeyDown()) {
				if (mSortKey != kSortOnAmplitude) {
					SortOnAmpitude() ;
					mSortKey = kSortOnAmplitude ;
					Refresh() ;
				}
			}
			else {
				// Select the whole column fo copying/pasting
				if (!HasSelection()) {
					firstSelectedCell.col = inCol ;
					lastSelectedCell.col = inCol ;
					firstSelectedCell.row = 1 ;
					lastSelectedCell.row = mRows ;
					Select(firstSelectedCell,lastSelectedCell) ;
				}
				else {
					ZeroSelection() ;
				}
			}
			break ;
		case 4:
		case 5:
			// Select the whole column fo copying/pasting
			if (!HasSelection()) {
				firstSelectedCell.col = inCol ;
				lastSelectedCell.col = inCol ;
				firstSelectedCell.row = 1 ;
				lastSelectedCell.row = mRows ;
				Select(firstSelectedCell,lastSelectedCell) ;
			}
			else {
				ZeroSelection() ;
			}
			break ;
		default:
			break ;
	}
}

void TPotentialDataTable::DrawColLabel(
	Int16 colNumber) 
{
	Rect	cellFrame;
	Rect	theSortRect  = {0,0,6,6} ;
	
	FPoint	imageOrigin = mParentPane->GetPortOrigin() ;
	
	if (mHasRowLabels)
		SetRect(&cellFrame,colNumber*mColWidth,0,(colNumber+1)*mColWidth,mRowHeight) ;
	else
		SetRect(&cellFrame,(colNumber-1)*mColWidth,0,colNumber*mColWidth,mRowHeight) ;
	if (colNumber != 0) 
		::OffsetRect(&cellFrame,near(imageOrigin.x),0) ;
	InsetRect(&cellFrame,1,1) ;
	SetFgColor(gGrayPixel) ;
	PaintRect(&cellFrame) ;
	SetFgColor(gDarkerGray) ;
	MoveTo(cellFrame.left,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.top) ;
	SetFgColor(gBlackPixel) ;
	MoveTo(cellFrame.right,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.bottom) ;
	InsetRect(&cellFrame,-1,-1) ;
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	if (colNumber > 0) {
		DrawColLabelText(colNumber) ;
		switch (colNumber) {
			case 2:
				if (mSortKey == kSortOnG) {
					OffsetRect(&theSortRect,cellFrame.right - 10,cellFrame.bottom - 10) ;
					PaintRect(&theSortRect) ;					
				}
				break ;
			case 3:
				if (mSortKey == kSortOnAmplitude) {
					OffsetRect(&theSortRect,cellFrame.right - 10,cellFrame.bottom - 10) ;
					PaintRect(&theSortRect) ;					
				}
				break ;
			default :
				break ;
		}
	}
}

void TPotentialDataTable::DrawColLabelText(Int16 colNumber) {
	switch (colNumber) {
		case 1:
			DrawString("[h,k]") ;
			break ;
		case 2:
			DrawString("d-star") ;
			break ;
		case 3:
			DrawString("Amp.") ;
			break ;
		case 4:
			DrawString("Phase") ;
			break ;
		case 5:
			DrawString("Adjusted") ;
			break ;
	}
}

void TPotentialDataTable::DrawCellValue(
	TableCellT &inCell,
	Rect &cellFrame)
{
	char str[64] ;
	
	Rect	theRect = cellFrame ;
	InsetRect(&theRect,1,1) ;
	if (mIndex[inCell.row-1]%2 == 0) {
		SetFgColor(gGrayPixel) ;
		PaintRect(&theRect) ;
		SetFgColor(gBlackPixel) ;
	}
	else {	// We do this so that we can properly show up selected cells
		SetFgColor(gWhitePixel) ;
		PaintRect(&theRect) ;
		SetFgColor(gBlackPixel) ;
	}
	
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	switch (inCell.col) {
		case 1:
			sprintf(str,"[%3d,%3d]",mH[inCell.row-1],mK[inCell.row-1]) ;
			break ;
		case 2:
			sprintf(str,"%.3f",mGVector[inCell.row-1]) ;
			break ;
		case 3:
			sprintf(str,"%.2f",mAmplitude[inCell.row-1]) ;
			break ;
		case 4:
			sprintf(str,"%.2f",mPhase[inCell.row-1]) ;
			break ;
		case 5:
			if (mCTFCorrectedPhase[inCell.row-1] == kUnAssignedPhase)
				sprintf(str,"-") ;
			else 
				sprintf(str,"%.2f",mCTFCorrectedPhase[inCell.row-1]) ;
			break ;
	}
	DrawString(str) ;
}

void	TPotentialDataTable::GetCellText(TableCellT &inCell,char *text) {
	switch (inCell.col) {
		case 1:
			sprintf(text,"[%3d,%3d]",mH[inCell.row-1],mK[inCell.row-1]) ;
			break ;
		case 2:
			sprintf(text,"%.3f",mGVector[inCell.row-1]) ;
			break ;
		case 3:
			sprintf(text,"%.2f",mAmplitude[inCell.row-1]) ;
			break ;
		case 4:
			sprintf(text,"%.2f",mPhase[inCell.row-1]) ;
			break ;
		case 5:
			if (mCTFCorrectedPhase[inCell.row-1] == kUnAssignedPhase)
				sprintf(text,"-") ;
			else 
				sprintf(text,"%.2f",mCTFCorrectedPhase[inCell.row-1]) ;
			break ;
	}
}

void	TPotentialDataTable::SortOnAmpitude() {
	long gap ;
	register long i,j ;
	register short itemp ;
	register float temp ;
	for (gap=mNumReflections/2;gap>0;gap/=2)
		for(i=gap;i<mNumReflections;i++)
			for(j=i-gap;j>=0;j-=gap)
				{
					if (mAmplitude[j] >= mAmplitude[j+gap])
						break ;
						
					temp = mGVector[j] ;
					mGVector[j] = mGVector[j+gap] ;
					mGVector[j+gap] = temp ;
					
					temp = mAmplitude[j] ;
					mAmplitude[j] = mAmplitude[j+gap] ;
					mAmplitude[j+gap] = temp ;
					
					temp = mPhase[j] ;
					mPhase[j] = mPhase[j+gap] ;
					mPhase[j+gap] = temp ;

					temp = mCTFCorrectedPhase[j] ;
					mCTFCorrectedPhase[j] = mCTFCorrectedPhase[j+gap] ;
					mCTFCorrectedPhase[j+gap] = temp ;
					
					itemp = mH[j] ;
					mH[j] = mH[j+gap] ;
					mH[j+gap] = itemp ;
					
					itemp = mK[j] ;
					mK[j] = mK[j+gap] ;
					mK[j+gap] = itemp ;

					itemp = mIndex[j] ;
					mIndex[j] = mIndex[j+gap] ;
					mIndex[j+gap] = itemp ;
			  	}
			  	
	GroupSymmetryRelatedReflections() ;
}

void	TPotentialDataTable::SortOnG() {
	long gap ;
	register long i,j ;
	register short itemp ;
	register float temp ;
	for (gap=mNumReflections/2;gap>0;gap/=2)
		for(i=gap;i<mNumReflections;i++)
			for(j=i-gap;j>=0;j-=gap)
				{
					if (mGVector[j] <= mGVector[j+gap])
						break ;
						
					temp = mGVector[j] ;
					mGVector[j] = mGVector[j+gap] ;
					mGVector[j+gap] = temp ;
					
					temp = mAmplitude[j] ;
					mAmplitude[j] = mAmplitude[j+gap] ;
					mAmplitude[j+gap] = temp ;
					
					temp = mPhase[j] ;
					mPhase[j] = mPhase[j+gap] ;
					mPhase[j+gap] = temp ;

					temp = mCTFCorrectedPhase[j] ;
					mCTFCorrectedPhase[j] = mCTFCorrectedPhase[j+gap] ;
					mCTFCorrectedPhase[j+gap] = temp ;
					
					itemp = mH[j] ;
					mH[j] = mH[j+gap] ;
					mH[j+gap] = itemp ;
					
					itemp = mK[j] ;
					mK[j] = mK[j+gap] ;
					mK[j+gap] = itemp ;

					itemp = mIndex[j] ;
					mIndex[j] = mIndex[j+gap] ;
					mIndex[j+gap] = itemp ;
			  	}

	GroupSymmetryRelatedReflections() ;
}

void	TPotentialDataTable::GroupSymmetryRelatedReflections() {
	// Set the index mIndex so that symmetry related reflections appear together
	// The index is already set but needs to be reset now that they have been sorted on 
	// either g-vector or amplitude
	
	short	numInGroup,currentGroupIndex,groupIndex = 1 ;
	
	long gap ;
	register long i,j,k ;
	register short itemp ;
	register float temp ;
	float	delta = 0.0001 ;
	for (gap=mNumReflections/2;gap>0;gap/=2)
		for(i=gap;i<mNumReflections;i++)
			for(j=i-gap;j>=0;j-=gap)
				{
					if (mSortKey == kSortOnG) {
						if ((mGVector[j+gap] - mGVector[j]) > delta)
							break ;
						if (mAmplitude[j] >= mAmplitude[j+gap])
							break ;
					} 
					else if (mSortKey == kSortOnAmplitude) {
						if (mAmplitude[j] > mAmplitude[j+gap])
							break ;
						if ((mGVector[j+gap] - mGVector[j]) > delta)
							break ;
					} 
						
					if (mIndex[j] == mIndex[j+gap])
						break ;
						
					temp = mGVector[j] ;
					mGVector[j] = mGVector[j+gap] ;
					mGVector[j+gap] = temp ;
					
					temp = mAmplitude[j] ;
					mAmplitude[j] = mAmplitude[j+gap] ;
					mAmplitude[j+gap] = temp ;
					
					temp = mPhase[j] ;
					mPhase[j] = mPhase[j+gap] ;
					mPhase[j+gap] = temp ;

					temp = mCTFCorrectedPhase[j] ;
					mCTFCorrectedPhase[j] = mCTFCorrectedPhase[j+gap] ;
					mCTFCorrectedPhase[j+gap] = temp ;
					
					itemp = mH[j] ;
					mH[j] = mH[j+gap] ;
					mH[j+gap] = itemp ;
					
					itemp = mK[j] ;
					mK[j] = mK[j+gap] ;
					mK[j+gap] = itemp ;

					itemp = mIndex[j] ;
					mIndex[j] = mIndex[j+gap] ;
					mIndex[j+gap] = itemp ;
			  	}
	
	// Assign new group indicies
	for(i = 0; i < mNumReflections;) {
		currentGroupIndex = mIndex[i] ;
		for(j=i+1; j < mNumReflections && mIndex[j] == currentGroupIndex;j++) ;
		numInGroup = j - i ;
		for(k=i;k < i+numInGroup; k++) {
			mIndex[k] = groupIndex ;
		}
		i += numInGroup ;
		groupIndex++ ;		
	}
}


TPotentialMapWindow::TPotentialMapWindow(TPotentialDataTable *table)
	 : TWindow(30,30,640,340,gImageDepth,"Potential",false,false) {

	char	text[256] ;
	mX = nil ;
	mY = nil ;
	Rect	theImageRect,theRect ;
	Rect	thePotentialRect ;
	
	mSymmetrizedImage = nil ;
	mPotentialMapImage = nil ;
	mDataTable = table ;
	mDataIsCalibrated = false ;
	
	if (mDataTable->mDiffractionPeaks->mCalibrationUnit == kInverseAngstrom)
		mDataIsCalibrated = true ;
		
	mImageRect = new TFrameRect(336+10,40,336+128+10,40+128) ;
	AddObject(mImageRect) ;
	
	mPotentialRect = new TFrameRect(474+20,40,474+128+20,40+128) ;
	AddObject(mPotentialRect) ;
	
	SetBkColor(gGrayPixel) ;
	
	theRect = theImageRect = mImageRect->GetRect() ;
	thePotentialRect = mPotentialRect->GetRect() ;
	
	mIdealScherzerLens = new TRadioButton(theImageRect.left,theImageRect.bottom+10,theImageRect.right,theImageRect.bottom+30,
		"Use Ideal Scherzer Lens",true) ;
	AddObject(mIdealScherzerLens) ;
	theRect = mIdealScherzerLens->GetRect() ;
		
	mRealLens = new TRadioButton(theImageRect.left+10,theRect.bottom+5,thePotentialRect.right-68,theRect.bottom+25,
		"Use Specified Lens Below...",false) ;
	if (!mDataIsCalibrated) mRealLens->Disable() ;
	AddObject(mRealLens) ;
	theRect = mRealLens->GetRect() ;
	
	TGroupBox *grpBox = new TGroupBox(theImageRect.left,(theRect.bottom+theRect.top)/2,thePotentialRect.right,296) ;
	AddObject(grpBox) ;

	mCTFRecord = gCTFRecord ;

	
	theRect = grpBox->GetRect() ;
	sprintf(text,"%d",near(mCTFRecord.voltage)) ;
	mVoltageText = new TEditText(theRect.left+45,theRect.top+20,theRect.left+80,theRect.top+40,
		"V [kV]",text) ;
	theRect = mVoltageText->GetRect() ;
	sprintf(text,"%.2f",mCTFRecord.cs) ;
	mCsText = new TEditText(theRect.left+100,theRect.top,theRect.left+140,theRect.bottom,
		"Cs [mm]",text) ;
	theRect = mCsText->GetRect() ;
	sprintf(text,"%d",near(mCTFRecord.defocus)) ;
	mDefocusText = new TEditText(theRect.left+80,theRect.top,theRect.left+124,theRect.bottom,
		"Df [A]",text) ;
	theRect = mVoltageText->GetRect() ;
	theRect.bottom += 4 ;
	sprintf(text,"%d",near(mCTFRecord.delta)) ;
	mDeltaText = new TEditText(theRect.left,theRect.bottom+8,theRect.right,theRect.bottom+28,
		"Del [A]",text) ;
	theRect = mCsText->GetRect() ;
	theRect.bottom += 4 ;
	sprintf(text,"%.2f",mCTFRecord.div) ;
	mDivText = new TEditText(theRect.left,theRect.bottom+8,theRect.right,theRect.bottom+28,
		"Div [mRad]",text) ;

	theRect = mDefocusText->GetRect() ;
	theRect.bottom += 4 ;
	mApply = new TButton(theRect.right-60,theRect.bottom+8,theRect.right,theRect.bottom+30,"Update") ;
	if (!mDataIsCalibrated) mApply->Disable() ;
	AddObject(mApply) ;

	if (!mDataIsCalibrated) mDivText->Disable() ;
	if (!mDataIsCalibrated) mDeltaText->Disable() ;
	if (!mDataIsCalibrated) mDefocusText->Disable() ;
	if (!mDataIsCalibrated) mCsText->Disable() ;
	if (!mDataIsCalibrated) mVoltageText->Disable() ;
			
	
	mCreatePotentialMapImage = new TButton(thePotentialRect.right-100,340-34,thePotentialRect.right,340-10,"Create Image...") ;
	AddObject(mCreatePotentialMapImage) ;
		
	mTablePane = new TTablePane(this,10,40,316,256,mDataTable) ;
	AddPane(mTablePane) ;

	// Create the images
	CreateSymmetrizedImage() ;
	CreatePotentialMapImage() ;
	
	mNeedToRecalculateScherzerMap = false ;
		
	TStaticText *theText = new TStaticText(10,20,326,40,"Fourier Space Data") ;
	theText->SetJustification(kCenterJustified) ;
	AddObject(theText) ;
	
	theText = new TStaticText(theImageRect.left,theImageRect.top-20,theImageRect.right,theImageRect.top,
		"Symmetrized Image") ;
	theText->SetJustification(kCenterJustified) ;
	AddObject(theText) ;
	
	theText = new TStaticText(thePotentialRect.left,thePotentialRect.top-20,thePotentialRect.right,thePotentialRect.top,
		"'Potential Map'") ;
	theText->SetJustification(kCenterJustified) ;
	AddObject(theText) ;

	theRect = mTablePane->GetRect() ;
	mCalibrateButton = 	new TButton(theRect.right-60,340-34,theRect.right,340-14,"Calibrate") ;
	AddObject(mCalibrateButton) ;

	theRect = mCalibrateButton->GetRect() ;

	IndexCalibrationRecord cbr ;
	mDataTable->GetCalibrationRecord(cbr) ;
				
	sprintf(text,"%.2f",1./cbr.astar) ;		
	mLatticeA =  new TEditText(theRect.left-110,340-34,theRect.left-70,340-14,
		"2D Lattice Parameters[Ang.]:  A",text) ;

	sprintf(text,"%.2f",1./cbr.bstar) ;		
	mLatticeB =  new TEditText(theRect.left-50,340-34,theRect.left-10,340-14,
		"B",text) ;

	TableCellT theCell ;
	theCell.col = 2 ; 
	theCell.row = mDataTable->GetNumRows() ; 
	mDataTable->GetCellText(theCell,text) ;
	float	dmax = 1.0/atof(text) ;
	sprintf(text,"%.2f",dmax) ;		
	mScherzerGmaxText = new TEditText(thePotentialRect.right-90,thePotentialRect.bottom+12,thePotentialRect.right-60,thePotentialRect.bottom+32,
		"Res [A]",text) ;	
		
	mUpdateScherzerGmaxButton = new TButton(thePotentialRect.right-50,thePotentialRect.bottom+12,thePotentialRect.right,thePotentialRect.bottom+32,"Update") ;
	AddObject(mUpdateScherzerGmaxButton) ;
	
	if (!mDataIsCalibrated) {
		mScherzerGmaxText->Disable() ;
		mUpdateScherzerGmaxButton->Disable() ;
	}

	AddTextEditObject(mLatticeB) ;
	AddTextEditObject(mLatticeA) ;

	AddTextEditObject(mDivText) ;
	AddTextEditObject(mDeltaText) ;
	AddTextEditObject(mDefocusText) ;
	AddTextEditObject(mCsText) ;
	AddTextEditObject(mVoltageText) ;
	
	AddTextEditObject(mScherzerGmaxText) ;

	mUseRealLens = false ;
	mSymmetrizedImageInitialized = false ;

	mTablePane->Show() ;
		
	mIsVisible = true ;
}

void	TPotentialMapWindow::CreateSymmetrizedImage() {
	short	Width = mDataTable->mArrayWidth ;
	short	Height = mDataTable->mArrayHeight ;
	
	short	minimumDimension = 64 ;

	// We will use a minimum of minimumDimension (64) points per unit cell
	// This pretty much corresponds to minimum of a 
	short ucWidth = Width > minimumDimension ? Width : minimumDimension ;		
	short ucHeight = Height > minimumDimension ? Height : minimumDimension ;
	
	float	Aaxis = mDataTable->mDiffractionPeaks->mA ;
	float	Baxis = mDataTable->mDiffractionPeaks->mB ;

	if (mDataTable->mPlaneGroup > 9) {
		Aaxis = (mDataTable->mDiffractionPeaks->mA + mDataTable->mDiffractionPeaks->mB)/2 ;
		Baxis = Aaxis ;
	}
	
	int	axisRatio = Aaxis > Baxis ? near(Aaxis/Baxis)/2 : near(Baxis/Aaxis)/2 ;
	
	if (axisRatio > 1) {
		int scale = near(pow(2.0,axisRatio)) ;
		if (Aaxis > Baxis) ucWidth *= scale ;
		if (Baxis > Aaxis) ucHeight *= scale ;
	}

	TArray	*theArray = new TArray(ucWidth,ucHeight,kComplexArray) ;
	long	i,h,k ;
	float	amp,phase ;
	float	degreeToRad = asin(1.0)/90.0 ;
	
	
	for(i=0;i<mDataTable->mNumReflections;i++) {
		h = mDataTable->mH[i] ;
		k = mDataTable->mK[i] ;
		amp = mDataTable->mAmplitude[i] ;
		phase = degreeToRad*mDataTable->mPhase[i] ;
		*theArray->GetRealHK(h,k) = amp*::cos(phase) ;
		*theArray->GetImagHK(h,k) = amp*::sin(phase) ;
		*theArray->GetRealHK(-h,-k) = amp*::cos(phase) ;
		*theArray->GetImagHK(-h,-k) = -amp*::sin(phase) ;
	}

	theArray->ifft(false) ;
	
	mSymmetrizedImage = new TImage(theArray,mDataTable->mGamma) ;
	mSymmetrizedImage->SetSampling(Aaxis/theArray->GetWidth(),Baxis/theArray->GetHeight()) ;
	mSymmetrizedImage->Repeat(2,2) ;
	mSymmetrizedImage->AdjustGamma() ;
	mSymmetrizedImage->AdjustSampling() ;
	delete theArray ;

}

void	TPotentialMapWindow::CreatePotentialMapImage(float gmax) {

	short	Width = mDataTable->mArrayWidth ;
	short	Height = mDataTable->mArrayHeight ;
	
	short	minimumDimension = 64 ;

	// We will use a minimum of minimumDimension (64) points per unit cell
	// This pretty much corresponds to minimum of a 
	short ucWidth = Width > minimumDimension ? Width : minimumDimension ;		
	short ucHeight = Height > minimumDimension ? Height : minimumDimension ;
	
	float	Aaxis = mDataTable->mDiffractionPeaks->mA ;
	float	Baxis = mDataTable->mDiffractionPeaks->mB ;

	if (mDataTable->mPlaneGroup > 9) {
		Aaxis = (mDataTable->mDiffractionPeaks->mA + mDataTable->mDiffractionPeaks->mB)/2 ;
		Baxis = Aaxis ;
	}
	
	int	axisRatio = Aaxis > Baxis ? near(Aaxis/Baxis)/2 : near(Baxis/Aaxis)/2 ;
	
	if (axisRatio > 1) {
		int scale = near(pow(2.0,axisRatio)) ;
		if (Aaxis > Baxis) ucWidth *= scale ;
		if (Baxis > Aaxis) ucHeight *= scale ;
	}

	TArray	*theArray = new TArray(ucWidth,ucHeight,kComplexArray) ;
	long	i,h,k ;
	float	amp,phase,rl,im ;
	float	degreeToRad = asin(1.0)/90.0 ;

	if (gmax == 0.)  {
		for(i=0;i<mDataTable->mNumReflections;i++) {
			h = mDataTable->mH[i] ;
			k = mDataTable->mK[i] ;
			amp = mDataTable->mAmplitude[i] ;
			phase = degreeToRad*mDataTable->mCTFCorrectedPhase[i] ;
			rl = amp*::cos(phase) ;
			im = amp*::sin(phase) ;
			*theArray->GetRealHK(h,k) = rl ;
			*theArray->GetImagHK(h,k) = im ;
			*theArray->GetRealHK(-h,-k) = rl ;
			*theArray->GetImagHK(-h,-k) = -im ;
		}
	}
	else {
		for(i=0;i<mDataTable->mNumReflections;i++) {
			h = mDataTable->mH[i] ;
			k = mDataTable->mK[i] ;
			amp = mDataTable->mGVector[i] > gmax ? 0. :  mDataTable->mAmplitude[i] ;
			phase = degreeToRad*mDataTable->mCTFCorrectedPhase[i] ;
			rl = amp*::cos(phase) ;
			im = amp*::sin(phase) ;
			*theArray->GetRealHK(h,k) = rl ;
			*theArray->GetImagHK(h,k) = im ;
			*theArray->GetRealHK(-h,-k) = rl ;
			*theArray->GetImagHK(-h,-k) = -im ;
		}
	}

	theArray->ifft(false) ;

	mPotentialMapImage = new TImage(theArray,mDataTable->mGamma) ;
	mPotentialMapImage->SetSampling(Aaxis/ucWidth,Baxis/ucHeight) ;
	mPotentialMapImage->Repeat(2,2) ;
	mPotentialMapImage->AdjustGamma() ;
	mPotentialMapImage->AdjustSampling() ;
	delete theArray ;
}

void 	TPotentialMapWindow::ExportPotentialMapImage() {
	UnitCellDisplayRecord displayRecord ;
	
	TCreatePotentialMapImageDialog *theDialog = new TCreatePotentialMapImageDialog(&displayRecord) ;
	if (theDialog->DoModalDialog()) {
		theDialog->GetDialogData(&displayRecord) ;
		delete theDialog ;

		short	Width = mDataTable->mArrayWidth ;
		short	Height = mDataTable->mArrayHeight ;
		
		short	minimumDimension = 64 ;

		// We will use a minimum of minimumDimension (64) points per unit cell
		// This pretty much corresponds to minimum of a 
		short ucWidth = Width > minimumDimension ? Width : minimumDimension ;		
		short ucHeight = Height > minimumDimension ? Height : minimumDimension ;
		
		float	Aaxis = mDataTable->mDiffractionPeaks->mA ;
		float	Baxis = mDataTable->mDiffractionPeaks->mB ;

		if (mDataTable->mPlaneGroup > 9) {
			Aaxis = (mDataTable->mDiffractionPeaks->mA + mDataTable->mDiffractionPeaks->mB)/2 ;
			Baxis = Aaxis ;
		}
		
		int	axisRatio = Aaxis > Baxis ? near(Aaxis/Baxis)/2 : near(Baxis/Aaxis)/2 ;
		
		if (axisRatio > 1) {
			int scale = near(pow(2.0,axisRatio)) ;
			if (Aaxis > Baxis) ucWidth *= scale ;
			if (Baxis > Aaxis) ucHeight *= scale ;
		}

		TArray	*theArray = new TArray(ucWidth,ucHeight,kComplexArray) ;
		long	i,h,k ;
		float	amp,phase ;
		float	degreeToRad = asin(1.0)/90.0 ;

		char	text[256] ;
		mScherzerGmaxText->GetEditText(text) ;
		float	gmax = 1.0/atof(text) ; 
		
		for(i=0;i<mDataTable->mNumReflections;i++) {
			h = mDataTable->mH[i] ;
			k = mDataTable->mK[i] ;
			amp = mDataTable->mGVector[i] > gmax ? 0. :  mDataTable->mAmplitude[i] ;
			phase = degreeToRad*mDataTable->mCTFCorrectedPhase[i] ;
			*theArray->GetRealHK(h,k) = amp*::cos(phase) ;
			*theArray->GetImagHK(h,k) = amp*::sin(phase) ;
			*theArray->GetRealHK(-h,-k) = amp*::cos(phase) ;
			*theArray->GetImagHK(-h,-k) = -amp*::sin(phase) ;
		}

		theArray->ifft(false) ;

		TImage *theImage = new TImage(theArray,mDataTable->mGamma) ;
		theImage->SetSampling(Aaxis/ucWidth,Baxis/ucHeight) ;
		theImage->Repeat(displayRecord.numCellsA,displayRecord.numCellsB) ;
		theImage->AdjustGamma() ;
		theImage->AdjustSampling() ;
		delete theArray ;
		
		theImage->SetName("Potential Map") ;
		theImage->Display() ;
	}
	else 
		delete theDialog ;
}

TPotentialMapWindow::~TPotentialMapWindow() {
	if (mX != nil) delete[] mX ;
	if (mY != nil) delete[] mY ;
	if (mSymmetrizedImage != nil) delete mSymmetrizedImage ;
	if (mPotentialMapImage != nil) delete mPotentialMapImage ;
}


void	TPotentialMapWindow::DoCommand(TCommand command,Boolean *hasBeenHandled) {	
	switch(command) {
		case kHelp:
			OnCommandHelp() ;		
			*hasBeenHandled = true ;
			break ;
		default:
			mDataTable->DoCommand(command,hasBeenHandled) ;
			break ;
	}
	if (command == kPaste) {
		// Update the images
		if (mSymmetrizedImage != nil) delete mSymmetrizedImage ;
		if (mPotentialMapImage != nil) delete mPotentialMapImage ;
		CreateSymmetrizedImage() ;
		CreatePotentialMapImage() ;
		UpdateContent() ;
	}
}

void	TPotentialMapWindow::GetCommandStatus(TCommand command,Boolean *isEnabled,
			Boolean *isChecked) {
	switch(command) {
		case kHelp:
			*isEnabled = true ;
			break ;
		case kClose:
			*isEnabled = true ;
			break ;	
		default:
			mDataTable->GetCommandStatus(command,isEnabled,isChecked) ;
			break ;
	}
}

void TPotentialMapWindow::Draw() {
	GrafPtr	tPort ;
	GetPort(&tPort) ;
	SetPort(this) ;
	
	EraseRect(&mPortRect) ;	// so we can get the right color	
	TPane::Draw() ;

	mTablePane->Refresh() ;

	DrawSymmetrizedImage() ;	
	DrawPotentialMapImage() ;
	
	mPotentialRect->Draw() ;
	mImageRect->Draw() ;
		
	SetPort(tPort) ;
	
}

void TPotentialMapWindow::UpdateContent() {
	GrafPtr	tPort ;
	GetPort(&tPort) ;
	SetPort(this) ;
	
	SetClip(mPortRect) ;
	mTablePane->Refresh() ;

	DrawPotentialMapImage() ;
		
	mPotentialRect->Draw() ;
	mImageRect->Draw() ;
	
	XFlush(gDisplay) ;
		
	SetPort(tPort) ;
}

void TPotentialMapWindow::DrawSymmetrizedImage() {
	Rect	imageRect = mImageRect->GetRect() ;
	InsetRect(&imageRect,1,1) ;
	if (mSymmetrizedImage != nil) {			
		if (!mSymmetrizedImageInitialized) {
			mSymmetrizedImageInitialized = true ;
			
			float	fraction1,fraction2,scaleFactor ;
			long	cx,cy,left,top ;	
			cx = (imageRect.left+imageRect.right)/2 ;		
			cy = (imageRect.top+imageRect.bottom)/2 ;
			// Calculate the actual image in which we are displaying			
			long rectWidth = imageRect.right - imageRect.left ;
			long rectHeight = imageRect.bottom - imageRect.top ;
			
			fraction1 = ((float)rectWidth)/mSymmetrizedImage->Width() ;
			fraction2 = ((float)rectHeight)/mSymmetrizedImage->Height() ;
			
			scaleFactor = fraction1 < fraction2 ? fraction1 : fraction2 ;
			
			long imageWidth = near(scaleFactor*mSymmetrizedImage->Width()) ;
			long imageHeight = near(scaleFactor*mSymmetrizedImage->Height()) ;
			
			
			left = cx - imageWidth/2 ;
			top = cy - imageHeight/2 ;
			SetRect(&mActualImageRect,left,top,left+imageWidth,top+imageHeight) ;
			
			mActualPotentialRect = mPotentialRect->GetRect() ;
			InsetRect(&mActualPotentialRect,1,1) ;
			cx = (mActualPotentialRect.left+mActualPotentialRect.right)/2 ;		
			cy = (mActualPotentialRect.top+mActualPotentialRect.bottom)/2 ;
			left = cx - imageWidth/2 ;
			top = cy - imageHeight/2 ;
			SetRect(&mActualPotentialRect,left,top,left+imageWidth,top+imageHeight) ;
		}
		mSymmetrizedImage->Display(this,&mActualImageRect) ;
			
	}

}

void TPotentialMapWindow::DrawPotentialMapImage() {
	mPotentialMapImage->Display(this,&mActualPotentialRect) ;

}

void	TPotentialMapWindow::DoKeyDownInObject(TObject *obj) {
	if (mDataTable->mPlaneGroup < 10) return ;
	char	text[256] ;
	if (obj == mLatticeA) {
		mLatticeA->GetEditText(text) ;
		mLatticeB->SetEditText(text) ;
	}
	else if (obj == mLatticeB) {
		mLatticeB->GetEditText(text) ;
		mLatticeA->SetEditText(text) ;
	}
}

void	TPotentialMapWindow::DoMouseDownInObject(TObject *obj,Point *pt) {
	if (pt != nil)
		if (!obj->DoMouseDown(pt)) return ;
	char	text[256] ;
	float	maxG,astar,bstar,a,b,sigamma,csgammastar ;
	float	spacing ;
	long	h,k ;
	if (obj == mApply && mUseRealLens) {
		// Get the lens parameters
		mVoltageText->GetEditText(text) ;
		mCTFRecord.voltage 	= atof(text) ;	
		mCsText->GetEditText(text) ;
		mCTFRecord.cs 		= atof(text) ;	
		mDefocusText->GetEditText(text) ;
		mCTFRecord.defocus 	= atof(text) ;	
		mDeltaText->GetEditText(text) ;
		mCTFRecord.delta 	= atof(text) ;	
		mDivText->GetEditText(text) ;
		mCTFRecord.div 		= atof(text) ;	
		gCTFRecord = mCTFRecord ;
		
		// Calculate the CTF for the conditions
		mDataTable->ApplyCTFCorrections(mCTFRecord) ;
		delete	mPotentialMapImage ;
		CreatePotentialMapImage() ;
		UpdateContent() ;
	}
	else if (obj == mCreatePotentialMapImage) {
		ExportPotentialMapImage() ;
		UpdateContent() ;
	}	
	else if (obj == mCalibrateButton) {
		// Calculate the calibration 
		IndexCalibrationRecord calibrationRec ;
		mDataTable ->GetCalibrationRecord(calibrationRec) ;

		sigamma = sin(calibrationRec.gammastar*asin(1.0)/90.) ;	// The same as sin(gammastar)
		mLatticeA->GetEditText(text) ;
		a = atof(text) ;
		if (a == 0.) {
			PutMessage("A cannot be 0") ;
			return ;
		}
		astar = 1./(a*sigamma) ;

		mLatticeB->GetEditText(text) ;
		b = atof(text) ;
		if (b == 0.) {
			PutMessage("B cannot be 0") ;
			return ;
		}
		bstar = 1./(b*sigamma) ;
		
		calibrationRec.astar = astar ;
		calibrationRec.bstar = bstar ;

		calibrationRec.isCalibrated = true ;
		
		mDataTable->Calibrate(calibrationRec) ;	
		mDataTable->Refresh() ;
		mScherzerGmaxText->GetEditText(text) ;
		
		csgammastar = cos(calibrationRec.gammastar*asin(1.0)/90.) ;
		h = mDataTable->mH[mDataTable->mNumReflections-1] ;
		k = mDataTable->mK[mDataTable->mNumReflections-1] ;
		spacing = 1./sqrt(h*h*astar*astar+k*k*bstar*bstar+2.0*h*k*astar*bstar*csgammastar) ;
		sprintf(text,"%.2f",spacing) ;		
		mScherzerGmaxText->SetEditText(text) ;
		if (!mDataIsCalibrated) {
			mDivText->Enable() ;
			mDeltaText->Enable() ;
			mDefocusText->Enable() ;
			mCsText->Enable() ;
			mVoltageText->Enable() ;
			mRealLens->Enable() ;
		}
		mDataIsCalibrated = true ;
		mScherzerGmaxText->Enable() ;
		mUpdateScherzerGmaxButton->Enable() ;

	}	
	else if (obj == mUpdateScherzerGmaxButton && !mUseRealLens) {
		// Calculate the CTF for the conditions
		mScherzerGmaxText->GetEditText(text) ;
		maxG = 1.0/atof(text) ; 
		mDataTable->ApplyCTFCorrections(mCTFRecord,true,maxG) ;
		delete	mPotentialMapImage ;
		CreatePotentialMapImage(maxG) ;
		UpdateContent() ;
	}	
	else if (obj == mIdealScherzerLens && mUseRealLens) {
		mUseRealLens = false ;
		mRealLens->SetState(false) ;
		mIdealScherzerLens->SetState(true) ;
		if (mApply->IsEnabled()) mApply->Disable() ;
		// Apply Scherzer Lens conditions
		mScherzerGmaxText->GetEditText(text) ;
		maxG = 1.0/atof(text) ; 
		mDataTable->ApplyCTFCorrections(mCTFRecord,true,maxG) ;
		delete	mPotentialMapImage ;
		CreatePotentialMapImage(maxG) ;
		UpdateContent() ;
		mUpdateScherzerGmaxButton->Enable() ;
	}
	else if (obj == mRealLens && !mUseRealLens) {
		mUseRealLens = true ;
		if (!mDivText->IsEnabled()) mDivText->Enable() ;
		if (!mDeltaText->IsEnabled()) mDeltaText->Enable() ;
		if (!mDefocusText->IsEnabled()) mDefocusText->Enable() ;
		if (!mCsText->IsEnabled()) mCsText->Enable() ;
		if (!mVoltageText->IsEnabled()) mVoltageText->Enable() ;
		mIdealScherzerLens->SetState(false) ;
		mRealLens->SetState(true) ;
		mApply->Enable() ;
		mUpdateScherzerGmaxButton->Disable() ;

		// Get the lens parameters
		mVoltageText->GetEditText(text) ;
		mCTFRecord.voltage 	= atof(text) ;	
		mCsText->GetEditText(text) ;
		mCTFRecord.cs 		= atof(text) ;	
		mDefocusText->GetEditText(text) ;
		mCTFRecord.defocus 	= atof(text) ;	
		mDeltaText->GetEditText(text) ;
		mCTFRecord.delta 	= atof(text) ;	
		mDivText->GetEditText(text) ;
		mCTFRecord.div 		= atof(text) ;	
		gCTFRecord = mCTFRecord ;
		
		// Calculate the CTF for the conditions
		mDataTable->ApplyCTFCorrections(mCTFRecord) ;
		delete	mPotentialMapImage ;
		CreatePotentialMapImage() ;
		UpdateContent() ;

	}
	else if (obj == mVoltageText || obj == mCsText || obj == mDefocusText ||
			obj == mDeltaText || obj == mDivText) {
		if (!mUseRealLens) {
			mUseRealLens = true ;
			mIdealScherzerLens->SetState(false) ;
			mRealLens->SetState(true) ;
			if (!mApply->IsEnabled()) mApply->Enable() ;
		}
	}
}
