// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TWindowList.h"
void PutMessage(char *) ;
TWindowEntry::TWindowEntry() {
	mNextWindowEntry = nil ;
	mWindow = nil ;
}

TWindowEntry::~TWindowEntry() {
	mNextWindowEntry = nil ;
	if (mWindow != nil)	// The window is removed separately
		delete mWindow ;
}

TWindowList::TWindowList() {
	mNumWindows = 0 ;
	mWindowEntry = nil ;
}

TWindowList::~TWindowList() {
	TWindowEntry *entry ;
	TWindowEntry *nextEntry ;
	
	// For every window entry
	entry = mWindowEntry ;
	while(entry != nil) {
		nextEntry = entry->mNextWindowEntry ;
		delete entry ;
		entry = nextEntry ;
		mNumWindows-- ;
	} ;
}

void TWindowList::AddWindow(TWindow *theWindow) {
	
	TWindowEntry *aNewEntry = new TWindowEntry ;
	aNewEntry->mWindow = theWindow ;
	
	// We save time by always adding to the beginning of the list
	aNewEntry->mNextWindowEntry = mWindowEntry ;
	mWindowEntry = aNewEntry ;
	mNumWindows++ ;
}


void TWindowList::RemoveWindow(TWindow *theWindow) {
	TWindowEntry* entry;
	TWindowEntry* last;
	TWindow	*currentWindow ;
	
	last = nil;
	for (entry = mWindowEntry ;entry != nil; entry = entry->mNextWindowEntry) {
		currentWindow = entry->mWindow ;
		if ( currentWindow == theWindow ) {			// is the current window the one to be removed?
													// yes, remove it
			// if first item in list, just set first
				if ( last == nil )				
					mWindowEntry = entry->mNextWindowEntry;
				else
					last->mNextWindowEntry = entry->mNextWindowEntry ;
			
			// free the TWindowEntry
				delete entry ;
				mNumWindows-- ;
				
			// we're done; no need to continue onto the next document
				return;
		}
		else						// no, go onto the next document in the list
			last = entry;
	}
}

TWindow *TWindowList::GetWindow(int number) {
	TWindowEntry* entry;
	int i ;
	
	if (number >= mNumWindows || number < 0)
		return nil ;
		
	for (i = 0,entry = mWindowEntry ;entry != nil; entry = entry->mNextWindowEntry,i++) {
		if (i == number) {
			return entry->mWindow ;
		}
	}
	return nil ;
}

TWindow *TWindowList::GetWindow(char *name) {
	TWindowEntry* entry;
	char	windowName[256] ;
			
	for (entry = mWindowEntry ;entry != nil; entry = entry->mNextWindowEntry) {
		entry->mWindow->GetWindowTitle(windowName) ;
		if (strcmp(name,windowName) == 0) {
			return entry->mWindow ;
		}
	}
	return nil ;
}

TWindow *TWindowList::GetWindow(Window theXWindow) {
	TWindowEntry* entry;
	int i ;
			
	for (i = 0,entry = mWindowEntry ;entry != nil; entry = entry->mNextWindowEntry,i++) {
		if (theXWindow == entry->mWindow->GetXWindow()) {
			return entry->mWindow ;
		}
	}
	return nil ;
}

TPaneEntry::TPaneEntry() {
	mNextPaneEntry = nil ;
	mPane = nil ;
}

TPaneEntry::~TPaneEntry() {
	mNextPaneEntry = nil ;
	if (mPane != nil)	// The window is removed separately
		delete mPane ;
}

TPaneList::TPaneList() {
	mNumPanes = 0 ;
	mPaneEntry = nil ;
}

TPaneList::~TPaneList() {
	TPaneEntry *entry ;
	TPaneEntry *nextEntry ;
	
	// For every pane entry
	entry = mPaneEntry ;
	while(entry != nil) {
		nextEntry = entry->mNextPaneEntry ;
		delete entry ;
		entry = nextEntry ;
		mNumPanes-- ;
	} ;
}

void TPaneList::AddPane(TPane *thePane) {
	
	TPaneEntry *aNewEntry = new TPaneEntry ;
	aNewEntry->mPane = thePane ;
	
	// We save time by always adding to the beginning of the list
	aNewEntry->mNextPaneEntry = mPaneEntry ;
	mPaneEntry = aNewEntry ;
	mNumPanes++ ;
}


void TPaneList::RemovePane(TPane *thePane) {
	TPaneEntry* entry;
	TPaneEntry* last;
	TPane	*currentPane ;
	
	last = nil;
	for (entry = mPaneEntry ;entry != nil; entry = entry->mNextPaneEntry) {
		currentPane = entry->mPane ;
		if ( currentPane == thePane ) {		
			// if first item in list, just set first
				if ( last == nil )				
					mPaneEntry = entry->mNextPaneEntry;
				else
					last->mNextPaneEntry = entry->mNextPaneEntry ;
			
			// free the TPaneEntry
				delete entry ;
				mNumPanes-- ;
				
			// we're done; no need to continue onto the next document
				return;
		}
		else						// no, go onto the next document in the list
			last = entry;
	}
}

TPane *TPaneList::GetPane(int number) {
	TPaneEntry* entry;
	int i ;
	
	if (number >= mNumPanes || number < 0)
		return nil ;
		
	for (i = 0,entry = mPaneEntry ;entry != nil; entry = entry->mNextPaneEntry,i++) {
		if (i == number) {
			return entry->mPane ;
		}
	}
	return nil ;
}

TPane *TPaneList::GetPane(Window theXWindow) {
	TPaneEntry* entry;
	int i ;
			
	for (i = 0,entry = mPaneEntry ;entry != nil; entry = entry->mNextPaneEntry,i++) {
		if (theXWindow == entry->mPane->GetXWindow()) {
			return entry->mPane ;
		}
	}
	return nil ;
}

