#include "UMemoryMgr.h"


// ---------------------------------------------------------------------------
//	 BlocksAreEqual
// ---------------------------------------------------------------------------
//	Blocks are equal if the first n bytes pointed to by s1 have the same
//	values as the first n bytes pointed to by s2. Note that this always
//	returns true if n is zero.

Boolean
BlocksAreEqual(
	const void	*s1,
	const void	*s2,
	UInt32		n)
{
	const unsigned char	*ucp1 = (const unsigned char *) s1;
	const unsigned char	*ucp2 = (const unsigned char *) s2;
	
	while (n > 0) {
		if (*ucp1++ != *ucp2++) {
			return false;
		}
		n--;
	}
	
	return true;
}


// ---------------------------------------------------------------------------
//	 BlockCompare(const void*, const void*, UInt32)
// ---------------------------------------------------------------------------
//	Compare two equal-length blocks of memory 
//
//	Return values:
//
//		< 0		block 1 is less than block 2
//		0		block 1 is equal to block 2
//		> 0		block 1 is greater than block 2
//
// Note that this always returns 0 (equal) if n <= 0.

SInt32
BlockCompare(
	const void	*s1,
	const void	*s2,
	UInt32		n)
{
	const unsigned char	*ucp1 = (const unsigned char *) s1;
	const unsigned char	*ucp2 = (const unsigned char *) s2;
	
	while (n > 0) {
		if (*ucp1 != *ucp2) {
			return ((SInt32)(*ucp1 - *ucp2));
		}

		ucp1++;
		ucp2++;
		n--;
	}
	
	return 0;
}


// ---------------------------------------------------------------------------
//	 BlockCompare(const void*, const void*, UInt32, Unit32)
// ---------------------------------------------------------------------------
//	Compare two blocks of memory. Blocks may have different lengths.
//
//	Return values:
//
//		< 0		block 1 is less than block 2
//		0		block 1 is equal to block 2
//		> 0		block 1 is greater than block 2
//
// Note that this always returns 0 (equal) if n <= 0.

SInt32
BlockCompare(
	const void	*inPtr1,
	const void	*inPtr2,
	UInt32		inLen1,
	UInt32		inLen2)
{
	const unsigned char	*ucp1 = (const unsigned char *) inPtr1;
	const unsigned char	*ucp2 = (const unsigned char *) inPtr2;

	UInt32	len = inLen1;				// len is the shorter block length
	if (inLen2 < inLen1) {
		len = inLen2;
	}
	
	while (len > 0) {					// Compare byte by byte
		if (*ucp1 != *ucp2) {
			return ((SInt32)(*ucp1 - *ucp2));
		}

		ucp1++;
		ucp2++;
		len--;
	}
										// All bytes the same so far
	return (SInt32)(inLen1 - inLen2);	// Longer block is "bigger"
}
