// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include <stdio.h>
#include "header.h"
#include "XDrivers.h"
#include "drivers.h"
#include "graphics.h"
#include "types.h"
#include "devices.h"
#include "globals.h"
extern FILE *gPSFile ;
extern Visual *gVisual ;

// ----------------------------------------------------------------
// Write a point on the display. Presently only 2 outputs 
// are defined   WorkStation=0  LaserPrinter=1 
// All the routines in here expect the coordinate systems (0,0) 
// to be at the upper left corner. Thus any transformation to the 
// local coordinate system takes place here.	
// ----------------------------------------------------------------
void gDrawPixel(int x,int y) {
    switch(gOutDevice->type) {
		case WS:
			DrawPixel(x,y) ;
			break;
		case LP :
			y = gOutDevice->yres -y - 1 ;
			/* Generate Postscript Code */
			break ;
    }
} 

// ------------------------------------------------------- 
// Routine to draw a line in physical Device coordinates 
// ------------------------------------------------------- 
void gDrawLine(int x1,int y1,int x2,int y2) {
    switch(gOutDevice->type) {
		case WS:
			DrawLine(x1,y1,x2,y2) ;
			break;
		case LP :
			y1 = gOutDevice->yres -y1 - 1 ;
			y2 = gOutDevice->yres -y2 - 1 ;
			/* Generate Postscript Code */
			fprintf(gPSFile,
				"\nnewpath %d %d moveto %d %d lineto stroke",
				x1,y1,x2,y2) ;
			break ;
    }
}

// ----------------------------------------------------------
// Move the "pen" to a point in physical device coordinates 
// ----------------------------------------------------------
void gMoveTo(int ix,int iy) {
    switch(gOutDevice->type) {
		case WS:
    			gCurrentTextPos.x = ix ;
    			gCurrentTextPos.y = iy ;
			break;
		case LP :
			iy = gOutDevice->yres - iy - 1 ;
			/* Generate Postscript Code */
			fprintf(gPSFile,"\n%d %d moveto",ix,iy) ;
			break ;
    }
}

// ----------------------------------------------------
// Draw a character on the display with attributes     
// ---------------------------------------------------- 
void gDrawChar(int x,int y,char c) {
    switch(gOutDevice->type) {
		case WS:
			DrawChar(x,y,c) ;
			break;
		case LP :
			y = gOutDevice->yres -y - 1 ;
			/* Generate Postscript Code */
			fprintf(gPSFile,"\n%d %d moveto",x,y) ;
			fprintf(gPSFile,"\n{%c} show",c) ;
			break ;
    }
}

// -----------------------------------------------
// Routine to draw a circle in PDC units   
// -----------------------------------------------
void gDrawCircle(int x,int y,int rad) {
		
    int linewidth ;
    int ilw,y1 ;

    linewidth = GetLineWidth() ;
    ilw = linewidth ;
    switch(gOutDevice->type) {
		case WS:
			SetLineWidth(ilw) ;
			DrawCircle(x,y,rad) ;
			SetLineWidth(linewidth) ;	
			break ;
		case LP :
			y1 = gOutDevice->yres - y - 1 ;
			/* Generate Postscript Code */
			fprintf(gPSFile,"\nnewpath") ;
			fprintf(gPSFile,"\n%d %d %d 0 360 arc",x,y1,rad) ;
			fprintf(gPSFile,"\n%d setlinewidth",ilw) ;
                        fprintf(gPSFile,"\nstroke") ;
			fprintf(gPSFile,"\n%d setlinewidth",linewidth) ;
			break ;
    	}
}

void gSetFgColor(int value) {
    float gray ;
    switch(gOutDevice->type) {
		case WS:
			SetFgColor(value) ;
			break ;
		case LP:
			gray = value/127. ;
			fprintf(gPSFile,"\n%f setgray",gray) ;
			break ;
	    }
}

// -----------------------------------------------
// Routine to fill a circle in PDC units
// -----------------------------------------------
void gFillCircle(int x,int y,int rad) {
		
    int linewidth ;
    int ilw,y1 ;

    linewidth = GetLineWidth() ;
    ilw = linewidth ;
    switch(gOutDevice->type) {
		case WS:
			SetLineWidth(ilw) ;
			FillCircle(x,y,rad) ;
			SetLineWidth(linewidth) ;	
			break ;
		case LP :
			y1 = gOutDevice->yres - y - 1 ;
			/* Generate Postscript Code */
			fprintf(gPSFile,"\nnewpath") ;
			fprintf(gPSFile,"\n%d %d %d 0 360 arc",x,y1,rad) ;
			fprintf(gPSFile,"\n%d setlinewidth",ilw) ;
                        fprintf(gPSFile,"\nfill") ;
			fprintf(gPSFile,"\n%d setlinewidth",linewidth) ;
			break ;
    	}
}

// ----------------------------------------------- 
// Routine to draw a rectangle in PDC units  
// ----------------------------------------------- 
void gDrawRect(int xlow,int ylow,int xhi,int yhi) {
    int linewidth ;
    Rect tRect ;
    int ilw ;

    linewidth = GetLineWidth() ;
    ilw = linewidth ;
    switch(gOutDevice->type) {
		case WS:
			SetLineWidth(ilw) ;
			SetRect(&tRect,xlow,ylow,xhi,yhi) ;
			FrameRect(&tRect) ;
			SetLineWidth(linewidth) ;	
			break ;
		case LP :
			ylow = gOutDevice->yres -ylow - 1 ;
			yhi = gOutDevice->yres -yhi - 1 ;
			/* Generate Postscript Code */
			fprintf(gPSFile,"\nnewpath") ;
			fprintf(gPSFile,"\n%d %d moveto",xlow,ylow) ;
			fprintf(gPSFile,"\n%d %d lineto",xhi,ylow) ;
			fprintf(gPSFile,"\n%d %d lineto",xhi,yhi) ;
			fprintf(gPSFile,"\n%d %d lineto",xlow,yhi) ;
			fprintf(gPSFile,"\nclosepath") ;
			fprintf(gPSFile,"\n%d setlinewidth",ilw) ;
                        fprintf(gPSFile,"\nstroke") ;
			break ;
    	}
}

// -----------------------------------------------
// Routine to write text at x,y in PDC units
// -----------------------------------------------
void gDrawText(int x,int y,char *text) {
    switch(gOutDevice->type) {
		case WS:
    		    DrawText(x,y,text) ;
			break ;
		case LP :
			y = gOutDevice->yres -y - 1 ;
			/* Generate Postscript Code */
			fprintf(gPSFile,"\n%d %d moveto",x,y) ;
			fprintf(gPSFile,"\n(%s) show",text) ;
			break ;
    }
}

void gDrawImageLine(int x,int y,int n,unsigned char *imageline) {
    int i,ix,iy ;
   
    switch(gOutDevice->type) {
		case WS:
			DrawImageLine(x,y,n,imageline) ;
    			break ;
		case LP :
			iy = gOutDevice->yres-y-1 ;
			ix = x ;
			/* Generate Postscript Code */
    			fprintf(gPSFile,"\n%d 1 %d %d DrawImageLine",n,ix,iy) ;
    			for (i=0;i<n;i++) {
    				if (i%40==0)
    					fprintf(gPSFile,"\n") ;
    				fprintf(gPSFile,"%02x",imageline[i]) ;
			}
			break ;
    }
}

int gReadImageLine(int x,int y,int n,unsigned char *imageline) {   
    switch(gOutDevice->type) {
		case WS:
			ReadImageLine(x,y,n,imageline) ;
    		break ;
		case LP :
			y = gOutDevice->yres -y - 1 ;
			break ;
    }
    return(1) ;
}

int gDrawImage(unsigned char *image,int x,int y,int width,int height) {
    switch(gOutDevice->type) {
		case WS:
			DrawImage(image,x,y,width,height) ;
			break ;
		case LP :
			y = gOutDevice->yres -y - 1 ;
			/* Generate Postscript Code */
			break ;
    }
    return(1) ;
}


int gReadImage(unsigned char *image,int x,int y,int width,int height) {

    switch(gOutDevice->type) {
		case WS:
			// We will allow for the display to be both 
			// in PseudoColor and TrueColor			
			ReadRGBImage(image,x,y,width,height) ;
			break ;
		case LP :
			/* Generate Postscript Code */
			y = gOutDevice->yres -y - 1 ;
			break ;
    }
    return(1) ;
}

void LandScape() {
  int it ;
  switch(gOutDevice->type) 
	{
	case WS:
		break ;
	case LP:
		it = gOutDevice->xres ;
		fprintf(gPSFile,"\nLandscape") ;
		gOutDevice->xres = gOutDevice->yres ;
		gOutDevice->yres = it ;
		break ;
	}
  }
