// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
// Changed by LDM to include the possibility of having sfftw.h
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "config.h"
#ifdef HAVE_SFFTW_H
#include "sfftw.h"
#else
#include "fftw.h"
#endif
#include "TArray.h"


extern	void 	fft1d(long m,long inc,float *data,float *table) ;
extern	void 	binrev(long m,long inc,float *c)  ;
extern	void 	fft2d(float data[],long n,long m,long isign) ;
extern	long 	near(float) ;
extern	void	PutMessage(char *) ;

void vclr(float *a1,long inc, long numElements) ;
void vsmul(float *a1,long inc1,float scalar,float *a2,long inc2, long numElements) ;
void vmov(float *a1,long inc1,float *a2,long inc2,long numElements) ;
void polar(float *a1,long inc1,float *a2,long inc2,long numElements) ;
void vsq(float *a1,long inc1,float *a2,long inc2,long numElements) ;
void vsqrt(float *a1,long inc1,float *a2,long inc2,long numElements) ;
void rect(float *a1,long inc1,float *a2,long inc2,long numElements) ;
void FFTW(float *data,long width,long height) ;
void IFFTW(float *data,long width,long height) ;

#define log2(x)	log(x)/log(2.0)

void	assert(Boolean) ;
void	assert(Boolean condition) {

	if (!condition) {
		fprintf(stderr,"\n ---- Memory allocation failure. Exiting program!! ------ \n") ;
		fflush(stderr) ;
		exit(0) ;
	}
}

void FFTW(float *data,long width,long height) {
	if (width == 1) {
		fftw_plan plan ;	
		plan = fftw_create_plan(width, FFTW_FORWARD, FFTW_ESTIMATE | FFTW_IN_PLACE) ;
		fftw_one(plan, (fftw_complex *)data, NULL);
		fftw_destroy_plan(plan);
    }
	else {
		fftwnd_plan n_plan ;
		n_plan = fftw2d_create_plan(height,width,
		                               FFTW_FORWARD,FFTW_ESTIMATE | FFTW_IN_PLACE) ;
		fftwnd_one(n_plan,(fftw_complex *)data, NULL);
		fftwnd_destroy_plan(n_plan);
		
	}	
}

void IFFTW(float *data,long width,long height) {
	if (width == 1) {
		fftw_plan plan ;	
		plan = fftw_create_plan(width, FFTW_BACKWARD, FFTW_ESTIMATE | FFTW_IN_PLACE) ;
		fftw_one(plan, (fftw_complex *)data, NULL);
		fftw_destroy_plan(plan);
    }
	else {
		fftwnd_plan n_plan ;
		n_plan = fftw2d_create_plan(height,width,
		                               FFTW_BACKWARD,FFTW_ESTIMATE | FFTW_IN_PLACE) ;
		fftwnd_one(n_plan,(fftw_complex *)data, NULL);
		fftwnd_destroy_plan(n_plan);
		
	}	
}
TArray::TArray(long width,long height,ArrayDataType dataType,float *data)
	: mWidth(width),mHeight(height) {
	register float *fp ;
	register	unsigned long i ;

	mNumElements = mWidth*mHeight ;
	mIsComplex = (dataType == kComplexArray) ? true : false ;

	unsigned long numFloats = mNumElements ;
	if (mIsComplex) numFloats *= 2 ;

	mData = new float[numFloats] ;
	assert(mData != 0) ;
	if (data != nil)
		memcpy(mData,data,numFloats*sizeof(float)) ;
	else {
		for(fp = mData,i=0; i < numFloats; i++)
			*fp++ = 0. ;
	}
}


TArray::TArray(long width,ArrayDataType dataType,float *data)
	: mWidth(width),mHeight(1) {

	mNumElements = mWidth*mHeight ;
	mIsComplex = (dataType == kComplexArray) ? true : false ;

	unsigned long numFloats = mNumElements ;
	if (mIsComplex) numFloats *= 2 ;

	mData = new float[numFloats] ;
	assert(mData != 0) ;
	if (data != nil)
		memcpy(mData,data,numFloats*sizeof(float)) ;
}

TArray::TArray(const TArray& u,ArrayDataType dataType,Rect *selectionRect)
	: mIsComplex(u.mIsComplex) {

	unsigned long numFloats ;
	register	long	i ;
	register 	float	*p1,*p2 ;
	int		Top,Left,Bottom,Right ;

	if (selectionRect == nil) {
		mWidth = u.mWidth ;
		mHeight = u.mHeight ;
		Top = 0 ;
		Left = 0 ;
		Bottom = mHeight ;
		Right = mWidth ;
	}
	else {
		mWidth = selectionRect->right - selectionRect->left ;
		mHeight = selectionRect->bottom - selectionRect->top ;
		Top = selectionRect->top ;
		Left = selectionRect->left ;
		Bottom = selectionRect->bottom ;
		Right = selectionRect->right ;
	}
	mNumElements = mWidth*mHeight ;
		
	if (dataType == kInheritType || (dataType == kRealArray && !u.mIsComplex)
			|| (dataType == kComplexArray && u.mIsComplex) ) {
		// We copy the identity of the input array
		numFloats = mNumElements ;
		if (mIsComplex) numFloats *= 2 ;

		if (numFloats > 0) {
			mData = new float[numFloats] ;
			assert(mData != 0) ;
			if (selectionRect == nil)
				memcpy(mData,u.mData,numFloats*sizeof(float)) ;
			else {
				numFloats = mWidth ;
				if (mIsComplex) numFloats *= 2 ;
				p1 = mData ;
				if (mIsComplex)
					p2 = u.mData + 2*(Top*u.mWidth + Left) ;
				else
					p2 = u.mData + (Top*u.mWidth + Left) ;
				for(i=Top; i < Bottom; i++) {
					memcpy(p1,p2,numFloats*sizeof(float)) ;
					if (mIsComplex) {
						p1 += 2*mWidth ;
						p2 += 2*u.mWidth ;
					}
					else {
						p1 += mWidth ;
						p2 += u.mWidth ;
					}
				}
			}
		}
	}
	else if (dataType == kRealArray) {
		// Input array is complex. Take real part
		mData = new float[mNumElements] ;
		assert(mData != 0) ;
		p1 = mData ;
		p2 = u.mData + 2*(Top*u.mWidth + Left) ;		
		for(i=0;i<mNumElements;i++) {
			*p1++ = *p2++ ;
			p2++ ;
		}
	}
	else {	// dataType == kComplexArray  Input is real
		mData = new float[2*mNumElements] ;
		assert(mData != 0) ;
		p1 = mData ;
		p2 = u.mData + Top*u.mWidth + Left ;		
		for(i=0;i<mNumElements;i++) {
			*p1++ = *p2++ ;
			*p1++ = 0. ;
		}
	}
}

TArray::TArray(const TArray *u,ArrayDataType dataType,Rect *selectionRect)
	: mIsComplex(u->mIsComplex) {


	unsigned long numFloats ;
	register	long	i ;
	register 	float	*p1,*p2 ;
	int		Top,Left,Bottom,Right ;

	if (selectionRect == nil) {
		mWidth = u->mWidth ;
		mHeight = u->mHeight ;
		Top = 0 ;
		Left = 0 ;
		Bottom = mHeight ;
		Right = mWidth ;
	}
	else {
		mWidth = selectionRect->right - selectionRect->left ;
		mHeight = selectionRect->bottom - selectionRect->top ;
		Top = selectionRect->top ;
		Left = selectionRect->left ;
		Bottom = selectionRect->bottom ;
		Right = selectionRect->right ;
	}
	mNumElements = mWidth*mHeight ;
		
	if (dataType == kInheritType || (dataType == kRealArray && !u->mIsComplex)
			|| (dataType == kComplexArray && u->mIsComplex) ) {
		// We copy the identity of the input array
		numFloats = mNumElements ;
		if (mIsComplex) numFloats *= 2 ;

		if (numFloats > 0) {
			mData = new float[numFloats] ;
			assert(mData != 0) ;
			if (selectionRect == nil)
				memcpy(mData,u->mData,numFloats*sizeof(float)) ;
			else {
				numFloats = mWidth ;
				if (mIsComplex) numFloats *= 2 ;
				p1 = mData ;
				if (mIsComplex)
					p2 = u->mData + 2*(Top*u->mWidth + Left) ;
				else
					p2 = u->mData + (Top*u->mWidth + Left) ;
				for(i=Top; i < Bottom; i++) {
					memcpy(p1,p2,numFloats*sizeof(float)) ;
					if (mIsComplex) {
						p1 += 2*mWidth ;
						p2 += 2*u->mWidth ;
					}
					else {
						p1 += mWidth ;
						p2 += u->mWidth ;
					}
				}
			}
		}
	}
	else if (dataType == kRealArray) {
		// Input array is complex. Take real part
		mData = new float[mNumElements] ;
		assert(mData != 0) ;
		p1 = mData ;
		p2 = u->mData + 2*(Top*u->mWidth + Left) ;		
		for(i=0;i<mNumElements;i++) {
			*p1++ = *p2++ ;
			p2++ ;
		}
	}
	else {	// dataType == kComplexArray  Input is real
		mData = new float[2*mNumElements] ;
		assert(mData != 0) ;
		p1 = mData ;
		p2 = u->mData + Top*u->mWidth + Left ;		
		for(i=0;i<mNumElements;i++) {
			*p1++ = *p2++ ;
			*p1++ = 0. ;
		}
	}
}


TArray::~TArray() {
	if (mData != nil) delete[] mData ;
}


void	TArray::Print() {
	register	long i ;
	Cmplx	*c ;
	fprintf(stderr,"\n") ;
	if (mIsComplex) 
		for(i=0;i<Length();i++) {
			c = (Cmplx *)GetData(i) ;
			fprintf(stderr,"(%.4f,%.4f) ",c->x,c->y) ;
		}
	else
		for(i=0;i<Length();i++) {
			fprintf(stderr,"%.4f ",*GetData(i)) ; 
		}
}

void	TArray::SetData(float *data,long w,long h) {
	unsigned long numFloats ;
	if (w != 0) {
		// Are these new dimensions
		if (mWidth != w || mHeight != h) {
			if (mData != nil) delete[] mData ;
			mWidth = w ;
			mHeight = h ;
			mNumElements = mWidth*mHeight ;
			numFloats = mIsComplex ? 2*mNumElements : mNumElements ;
			mData = new float[numFloats] ;
			assert(mData != nil) ;
		}
	}
	numFloats = mIsComplex ? 2*mNumElements : mNumElements ;
	memcpy(mData,data,numFloats*sizeof(float)) ;
}

void TArray::MakeComplex() {
	if (mIsComplex) return ;
	mIsComplex = true ;
	
	register	long	i ;
	register 	float	*p1,*p2 ;

	float *tData = new float[mNumElements*2] ;
	assert(tData != nil) ;

	p1 = mData ;
	p2 = tData ;
	for(i=0;i<this->mNumElements;i++) {
		*p2++ = *p1++ ;
		*p2++ = 0. ;
	}
	delete[] mData ;
	mData = tData ;
}

void TArray::MakeReal() {
	if (!mIsComplex) return ;
	mIsComplex = false ;
	
	register	long	i ;
	register 	float	*p1,*p2 ;

	float *tData = new float[mNumElements] ;
	assert(tData != nil) ;

	p1 = mData ;
	p2 = tData ;
	for(i=0;i<this->mNumElements;i++) {
		*p2++ = *p1++ ;
		p1++  ;			// Skipping any imaginary part
	}
	delete[] mData ;
	mData = tData ;
}

float TArray::Mean() {

	register 	float	*p1 ;
	register	double	sum ;
	float		meanValue ;
	register 	long	i ;

	// If we have complex arrays, just operate on the real
	// part of the array
	sum = 0. ;
	p1 = mData ;
	if (mIsComplex) {
		for(i=0;i<mNumElements;i++) {
			sum += *p1++ ;
			p1++  ;			// Skipping any imaginary part
		}
	}
	else {
		for(i=0;i<mNumElements;i++) {
			sum += *p1++ ;
		}
	}
	meanValue = sum/mNumElements ;
	return meanValue ;
}

float TArray::Sum() {

	register 	float	*p1 ;
	register	double	sum ;
	register 	long	i ;

	// If we have complex arrays, just operate on the real
	// part of the array
	sum = 0. ;
	p1 = mData ;
	if (mIsComplex) {
		for(i=0;i<mNumElements;i++) {
			sum += *p1++ ;
			p1++  ;			// Skipping any imaginary part
		}
	}
	else {
		for(i=0;i<mNumElements;i++) {
			sum += *p1++ ;
		}
	}
	return sum ;
}

float TArray::Min() {
	register 	float	*p1,val ;
	float		minValue ;
	register 	long	i ;

	minValue = 1.0e+20 ;
	p1 = mData ;
	if (mIsComplex) {
		for(i=0;i<this->mNumElements;i++) {
			val = *p1++ ;
			if (val < minValue) minValue = val ;
			p1++  ;			// Skipping any imaginary part
		}
	}
	else {
		for(i=0;i<this->mNumElements;i++) {
			val = *p1++ ;
			if (val < minValue) minValue = val ;
		}
	}
	return minValue ;
}

float TArray::Max() {
	register 	float	*p1,val ;
	float		maxValue ;
	register 	long	i ;

	maxValue = -1.0e+20 ;
	p1 = mData ;
	if (mIsComplex) {
		for(i=0;i<this->mNumElements;i++) {
			val = *p1++ ;
			if (val > maxValue) maxValue = val ;
			p1++  ;			// Skipping any imaginary part
		}
	}
	else {
		for(i=0;i<this->mNumElements;i++) {
			val = *p1++ ;
			if (val > maxValue) maxValue = val ;
		}
	}
	return maxValue ;
}

float TArray::Min(short &x,short &y) {
	register 	float	*p1,val ;
	float		minValue ;
	register 	short	i,j ;

	minValue = 1.0e+20 ;
	p1 = mData ;
	if (mIsComplex) {
		for(j=0;j<this->mHeight;j++) 
			for(i=0;i<this->mWidth;i++) {
				val = *p1++ ;
				if (val < minValue) {
					x = i ;
					y = j ;
					minValue = val ;
				}
				p1++  ;			// Skipping any imaginary part
			}
	}
	else {
		for(j=0;j<this->mHeight;j++) 
			for(i=0;i<this->mNumElements;i++) {
				val = *p1++ ;
				if (val < minValue) {
					x = i ;
					y = j ;
					minValue = val ;
				}
			}
	}
	return minValue ;
}

float TArray::Max(short &x,short &y) {
	register 	float	*p1,val ;
	float		maxValue ;
	register 	short	i,j ;

	maxValue = -1.0e+20 ;
	p1 = mData ;
	if (mIsComplex) {
		for(j=0;j<this->mHeight;j++) 
			for(i=0;i<this->mWidth;i++) {
				val = *p1++ ;
				if (val > maxValue) {
					x = i ;
					y = j ;
					maxValue = val ;
				}
				p1++  ;			// Skipping any imaginary part
			}
	}
	else {
		for(j=0;j<this->mHeight;j++) 
			for(i=0;i<this->mWidth;i++) {
				val = *p1++ ;
				if (val > maxValue) {
					x = i ;
					y = j ;
					maxValue = val ;
				}
			}
	}
	return maxValue ;
}

float TArray::Stdv() {
	register 	float	*p1,dev ;
	register	double	sum ;
	float		meanValue ;
	register 	long	i ;

	meanValue = Mean() ;
	sum = 0. ;
	p1 = mData ;
	if (mIsComplex) {
		for(i=0;i<this->mNumElements;i++) {
			dev = *p1++ - meanValue ;
			sum += dev*dev ;
			p1++  ;			// Skipping any imaginary part
		}
	}
	else {
		for(i=0;i<this->mNumElements;i++) {
			dev = *p1++ - meanValue ;
			sum += dev*dev ;
		}
	}
	dev = ::sqrt(sum/mNumElements) ;
	return dev ;
}

float	TArray::VectorLength() {
	register float	*p1 ;
	register long	i ;
	register float 	rlu,imu ;
	
	register	float product = 0. ;

	// Operation will vary depending on the 
	// nature of the array to be equaled
	if (mIsComplex) {
		p1 = mData ;
		for(i=0;i<mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			product += rlu*rlu + imu*imu ;
		}
	}
	else {
		p1 = mData ;
		for(i=0;i<mNumElements;i++) {
			product += *p1 * (*p1) ;
			p1++ ;
		}
	}
	return ::sqrt(product) ;
}



void TArray::Modulus(bool centerImage) {
	register	float	*p1,*p2,rlu,imu ;
	register	long	i ;
	
	if (!mIsComplex) return ; // Already done
	mIsComplex = false ;

	float *tData = new float[mNumElements] ; 
	assert(tData != nil) ;

	p1 = mData ;
	p2 = tData ;	
	for(i=0;i<mNumElements;i++) {
		rlu = *p1++ ;
		imu = *p1++ ;
		*p2++ = ::sqrt(rlu*rlu+imu*imu) ;
	}
	delete[] mData ;
	mData = tData ;	
}

void TArray::Intensity(bool centerImage) {
	register	float	*p1,*p2,rlu,imu ;
	register	long	i ;
	

	if (mIsComplex) {
		float *tData = new float[mNumElements] ; 
		assert(tData != nil) ;

		p1 = mData ;
		p2 = tData ;	
		for(i=0;i<mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			*p2++ = rlu*rlu+imu*imu ;
		}
		delete[] mData ;
		mData = tData ;	
	}
	else {
		p1 = mData ;
		for(i=0;i<mNumElements;i++) {
			rlu = *p1 ;
			*p1++ = rlu*rlu ;
		}
	}
	mIsComplex = false ;
}

void TArray::Phase(bool centerImage) {
	register	float	*p1,*p2,rlu,imu ;
	register	long	i ;
	

	float PiOver2 = ::asin(1.0) ;
	float MinusPiOver2 = -PiOver2 ;
	float *tData ;
	
	if (mIsComplex) {
		tData = new float[mNumElements] ;
		p1 = mData ;
		p2 = tData ;
		for(i=0;i<mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			if (rlu == 0.)
				if (imu > 0.)
					*p2++ = PiOver2 ;
				else 
					*p2++ = MinusPiOver2 ;
			else
				*p2++ = ::atan2(imu,rlu) ;
		}
		delete[] mData ;
		mData = tData ;
	}
	else {
		p1 = mData ;
		for(i=0;i<mNumElements;i++) {
				*p1++ = 0. ;
		}
	}
	mIsComplex = false ;
}

void TArray::RealPart(bool centerImage) {
	register	float	*p1,*p2 ;
	register	long	i ;
	
	if (!mIsComplex) return ; // Already done
	mIsComplex = false ;

	float *tData = new float[mNumElements] ; 
	assert(tData != nil) ;

	p1 = mData ;
	p2 = tData ;	
	for(i=0;i<mNumElements;i++) {
		*p2++ = *p1++ ;
		p1++ ;
	}
	delete[] mData ;
	mData = tData ;	
}

void TArray::ImaginaryPart(bool centerImage) {
	register	float	*p1,*p2 ;
	register	long	i ;
	
	if (!mIsComplex) return ; // Already done
	mIsComplex = false ;

	float *tData = new float[mNumElements] ; 
	assert(tData != nil) ;

	p1 = mData ;
	p2 = tData ;	
	for(i=0;i<mNumElements;i++) {
		p1++ ;
		*p2++ = *p1++ ;
	}
	delete[] mData ;
	mData = tData ;	
}

void TArray::FFT() {
	// I should just remove the capital name routines
	fft() ;
}

void TArray::IFFT(Boolean keepComplex) {
	// I should just remove the capital name routines
	ifft(keepComplex) ;
}

void TArray::fft() {
	if (!mIsComplex)
		MakeComplex() ;

	if (mHeight == 1) {
		fftw_plan plan ;	
		plan = fftw_create_plan(mWidth, FFTW_FORWARD, FFTW_ESTIMATE | FFTW_IN_PLACE) ;
		fftw_one(plan, (fftw_complex *)mData, NULL);
		fftw_destroy_plan(plan);
    }
	else {
		fftwnd_plan n_plan ;
		n_plan = fftw2d_create_plan(mHeight,mWidth,
		                               FFTW_FORWARD,FFTW_ESTIMATE | FFTW_IN_PLACE) ;
		fftwnd_one(n_plan,(fftw_complex *)mData, NULL);
		fftwnd_destroy_plan(n_plan);
		
	}	
}

void TArray::ifft(Boolean keepComplex) {
	if (!mIsComplex)
		MakeComplex() ;
		
	if (mHeight == 1) {
		fftw_plan plan ;	
		plan = fftw_create_plan(mWidth, FFTW_BACKWARD, FFTW_ESTIMATE | FFTW_IN_PLACE) ;
		fftw_one(plan, (fftw_complex *)mData, NULL);
		fftw_destroy_plan(plan);
    }
	else {
		fftwnd_plan n_plan ;
		n_plan = fftw2d_create_plan(mHeight,mWidth,
		                               FFTW_BACKWARD,FFTW_ESTIMATE | FFTW_IN_PLACE) ;
		fftwnd_one(n_plan,(fftw_complex *)mData, NULL);
		fftwnd_destroy_plan(n_plan);
		
	}	
  	
	if (!keepComplex)
		MakeReal() ;
		
	// The inverse needs to be scaled by the size
	float	scaleFactor = 1./(mWidth*mHeight) ;
	*this *= scaleFactor ;
}

void TArray::log() {
	register	float	*p1,*p2,a,p ;
	register	Cmplx	*c ;
	register	long	i ;

	
	p1 = mData ;
	p2 = mData ;
	
	if (mIsComplex) {
		// If we have a complex number Ae(iP) = exp( lnA + iP)
		// then the Ln [ Aexp(iP)] =  (lnA,P)
		c = (Cmplx *)mData ;
		for(i=0;i<mNumElements;i++) {
			a = c->Modulus() ;
			p = c->Phase() ;
			*p2++ = ::log(a) ;
			*p2++ = p ;
			c++  ;
		}
	}
	else {
		for(i=0;i<mNumElements;i++) {
			*p2++ = ::log(*p1++) ;
		}
	}
}

void TArray::log10() {
	register	float	*p1,*p2,a,p ;
	register	Cmplx	*c ;
	register	long	i ;

	
	p1 = mData ;
	p2 = mData ;
	
	if (mIsComplex) {
		// If we have a complex number Ae(iP) = exp( lnA + iP)
		// then the Ln [ Aexp(iP)] =  (lnA,P)
		c = (Cmplx *)mData ;
		for(i=0;i<mNumElements;i++) {
			a = c->Modulus() ;
			p = c->Phase() ;
			*p2++ = ::log10(a) ;
			*p2++ = p ;
			c++  ;
		}
	}
	else {
		for(i=0;i<mNumElements;i++) {
			*p2++ = ::log10(*p1++) ;
		}
	}
}

void TArray::pow(float x) {
	register	float	*p1,*p2,a,p ;
	register	Cmplx	*c ;
	register	long	i ;

	
	p1 = mData ;
	p2 = mData ;
	
	if (mIsComplex) {
		// If we have a complex number Ae(iP) = exp( lnA + iP)
		// then the Ln [ Aexp(iP)] =  (lnA,P)
		c = (Cmplx *)mData ;
		for(i=0;i<mNumElements;i++) {
			a = c->Modulus() ;
			p = c->Phase() ;
			*p2++ = ::pow(a,x) ;
			*p2++ = p ;
			c++  ;
		}
	}
	else {
		for(i=0;i<mNumElements;i++) {
			*p2++ = ::pow(*p1++,x) ;
		}
	}
}

void TArray::sqrt() {
	register	float	*p1,*p2,a,p ;
	register	Cmplx	*c ;
	register	long	i ;

	
	p1 = mData ;
	p2 = mData ;
	
	if (mIsComplex) {
		// If we have a complex number Ae(iP), 
		// then sqrt(AeiP) = sqrt(A)e(iP/2)
		c = (Cmplx *)mData ;
		for(i=0;i<mNumElements;i++) {
			a = ::sqrt(c->Modulus()) ;
			p = c->Phase()/2 ;
			*p2++ = a*::cos(p) ;
			*p2++ = a*::sin(p) ;
			c++ ;
		}
	}
	else {
		for(i=0;i<mNumElements;i++) {
			*p2++ = ::sqrt(*p1++) ;
		}
	}
}

void TArray::sq() {
	register	float	*p1,*p2,a,p ;
	register	Cmplx	*c ;
	register	long	i ;


	p1 = mData ;
	p2 = mData ;
	
	if (mIsComplex) {
		// If we have a complex number Ae(iP), 
		// then sq(AeiP) = A*A(i2P)
		c = (Cmplx *)mData ;
		for(i=0;i<mNumElements;i++) {
			a = c->Modsq() ;
			p = 2*c->Phase() ;
			*p2++ = a*::cos(p) ;
			*p2++ = a*::sin(p) ;
			c++ ;
		}
	}
	else {
		for(i=0;i<mNumElements;i++) {
			*p2++ = (*p1)*(*p1) ; p1++ ;
		}
	}
}

void TArray::exp() {
	register	float	*p1,*p2,rlu,imu,a ;
	register	long	i ;


	p1 = mData ;
	p2 = mData ;
	
	if (mIsComplex) {
		// If we have a complex number x+iy, 
		// then exp(x+iy) = exp(x)*(cosy + isiny)
		for(i=0;i<mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			a = ::exp(rlu) ;
			*p2++ = a*::cos(imu) ;
			*p2++ = a*::sin(imu) ;
			p1 += 2 ;
		}
	}
	else {
		for(i=0;i<mNumElements;i++) {
			*p2++ = ::exp(*p1++) ;
		}
	}
}

void TArray::sin() {
	// I only know how to do this on real arrays
	register	float	*p1,*p2 ;
	register	long	i ;

	
	p1 = mData ;
	p2 = mData ;	
	for(i=0;i<mNumElements;i++) {
		*p2++ = ::sin(*p1++) ;
	}
}

void TArray::cos() {
	// I only know how to do this on real arrays
	register	float	*p1,*p2 ;
	register	long	i ;

	
	p1 = mData ;
	p2 = mData ;
	
	for(i=0;i<mNumElements;i++) {
		*p2++ = ::cos(*p1++) ;
	}
}

void TArray::tan() {
	register	float	*p1,*p2 ;
	register	long	i ;

	
	p1 = mData ;
	p2 = mData ;
	
	for(i=0;i<mNumElements;i++) {
		*p2++ = ::tan(*p1++) ;
	}
}

void TArray::asin() {
	register	float	*p1,*p2 ;
	register	long	i ;

	
	p1 = mData ;
	p2 = mData ;
	
	for(i=0;i<mNumElements;i++) {
		*p2++ = ::asin(*p1++) ;
	}
}

void TArray::acos() {
	// I only know how to do this on real arrays
	register	float	*p1,*p2 ;
	register	long	i ;

	
	p1 = mData ;
	p2 = mData ;
	
	for(i=0;i<mNumElements;i++) {
		*p2++ = ::acos(*p1++) ;
	}
}

void TArray::atan() {
	// I only know how to do this on real arrays
	register	float	*p1,*p2 ;
	register	long	i ;

	
	p1 = mData ;
	p2 = mData ;
	
	for(i=0;i<mNumElements;i++) {
		*p2++ = ::atan(*p1++) ;
	}
}

TArray& TArray::operator=(const float u) {
	register 	float	*p1 ;
	register 	long	i ;
	// Assignment will vary depending on the 
	// nature of the array to be equaled
	if (this->mIsComplex) {
		p1 = this->mData ;
		for(i=0;i<this->mNumElements;i++) {
			*p1++ = u ;
			*p1++ = 0. ;
		}
	}
	else {
		p1 = this->mData ;
		for(i=0;i<this->mNumElements;i++) {
			*p1++ = u ;
		}
	}
	return *this ;
}

TArray& TArray::operator=(const Cmplx& u) {
	register 	float	*p1 ;
	long	i ;

	if (this->mIsComplex) {
		p1 = this->mData ;
		for(i=0;i<this->mNumElements;i++) {
			*p1++ = u.x ;
			*p1++ = u.y ;
		}
	}
	else {
		// The array is too small. Reassign
		if (this->mData != nil) delete[] this->mData ;
		this->mData = new float[this->mNumElements*2] ;
		assert(this->mData != nil) ;
		p1 = this->mData ;
		this->mIsComplex = true ;
		for(i=0;i<this->mNumElements;i++) {
			*p1++ = u.x ;
			*p1++ = u.y ;
		}
	}
	return *this ;
}

TArray& TArray::operator=(const TArray& u) {
	float	*p1,*p2 ;
	long	i ;

	assert(this->mNumElements == u.mNumElements) ;

	// Assignment will vary depending on the 
	// nature of the array to be equaled
	if (this->mIsComplex == u.mIsComplex) {
		// straight memory copy
		if (this->mIsComplex) 
			memcpy(this->mData,u.mData,u.mNumElements*2*sizeof(float)) ;
		else
			memcpy(this->mData,u.mData,u.mNumElements*sizeof(float)) ;
	}
	else if (this->mIsComplex && !u.mIsComplex) {
		p1 = this->mData ;
		p2 = u.mData ;
		for(i=0;i<this->mNumElements;i++) {
			*p1++ = *p2++ ;
			*p1++ = 0. ;
		}
	}
	else {	// u is complex, but we are not
		// We must convert the array to complex
		this->mIsComplex = true ;
		delete[] this->mData ;
		this->mData = new float[u.mNumElements*2] ;
		memcpy(this->mData,u.mData,u.mNumElements*2*sizeof(float)) ;
	}
	return *this ;
}


TArray&	TArray::operator+=(const float v) {
	if (v == 0.) return *this;
	register	long i ;
	
	float	*fp = this->mData ;
	if (this->mIsComplex) {
		for(i=0;i<this->mNumElements;i++) {
			*fp++ += v ;
			fp++ ;
		}
	}
	else {
		for(i=0;i<this->mNumElements;i++) 
			*fp++ += v ;
	}
	return *this ;
}

TArray&	TArray::operator-=(const float v) {
	if (v == 0.) return *this;

	register	long i ;
	float	*fp = this->mData ;
	if (this->mIsComplex) {
		for(i=0;i<this->mNumElements;i++) {
			*fp++ -= v ;
			fp++ ;
		}
	}
	else {
		for(i=0;i<this->mNumElements;i++) 
			*fp++ -= v ;
	}
	return *this ;
}
TArray&	TArray::operator*=(const float v) {

	register	long i ;
	float	*fp = this->mData ;
	if (this->mIsComplex) {
		for(i=0;i<this->mNumElements;i++) {
			*fp++ *= v ;
			*fp++ *= v ;
		}
	}
	else {
		for(i=0;i<this->mNumElements;i++) 
			*fp++ *= v ;
	}
	return *this ;
}

TArray&	TArray::operator^=(const float v) {

	register	long i ;
	float	*fp = this->mData ;
	if (this->mIsComplex) {
		for(i=0;i<this->mNumElements;i++) {
			if (*fp != 0.) 
				*fp++ = ::pow(*fp,v) ;	// Wrong
			else
				*fp++ = 0. ;	// Wrong
			if (*fp != 0.) 
				*fp++ = ::pow(*fp,v) ;	// Wrong
			else
				*fp++ = 0. ;	// Wrong
		}
	}
	else {
		for(i=0;i<this->mNumElements;i++) 
			if (*fp != 0.) 
				*fp++ = ::pow(*fp,v) ;
			else
				*fp++ = 0. ;
	}
	return *this ;
}

TArray&	TArray::operator/=(const float v) {

	if (v == 0.) return *this;
	//fprintf(stderr,"\n %f Divided by %f = ",*this->mData,v) ;
	register	long i ;
	float	val = 1./v ;
	float	*fp = this->mData ;
	if (this->mIsComplex) {
		for(i=0;i<this->mNumElements;i++) {
			*fp++ *= val ;
			*fp++ *= val ;
		}
	}
	else {
		for(i=0;i<this->mNumElements;i++) 
			*fp++ *= val ;
	}
	//fprintf(stderr,"%f ",*this->mData) ;
	return *this ;
}

TArray&	TArray::operator+=(const Cmplx& v) {
	float	*tData ;
	register	long	i ;
	register 	float	*p1,*p2 ;	
	register	float rl,im ;

	// Don't have Altivec routines for this one yet.
	
	rl = v.x ;
	im = v.y ;
	// We use floating point pointers instead of complex logic for speed
	if (!this->mIsComplex) {
		// We must convert the array to complex
		tData = new float[this->mNumElements*2] ;
		assert(tData != 0) ;
		this->mIsComplex = true ;
		p1 = tData ;
		p2 = this->mData ;	// This is only half the length
		for(i=0;i<this->mNumElements;i++) {
			*p1++ = *p2++ + rl ;
			*p1++ = im ;
		}
		delete[] this->mData ;
		this->mData = tData ;
	}
	else {
		p1 = this->mData ;
		for(i=0;i<this->mNumElements;i++) {
			*p1++ += rl ;
			*p1++ += im ;
		}
	}		
	return *this ;
}

TArray&	TArray::operator-=(const Cmplx& v) {
	float	*tData ;
	register	long	i ;
	register 	float	*p1,*p2 ;	
	register	float rl,im ;

	// Don't have Altivec routines for this one yet.

	rl = v.x ;
	im = v.y ;
	if (!this->mIsComplex) {
		// We must convert the array to complex
		tData = new float[this->mNumElements*2] ;
		assert(tData != 0) ;
		this->mIsComplex = true ;
		p1 = tData ;
		p2 = this->mData ;
		for(i=0;i<this->mNumElements;i++) {
			*p1++ = *p2++ - rl ;
			*p1++ = -im ;
		}
		delete[] this->mData ;
		this->mData = tData ;
	}
	else {
		p1 = this->mData ;
		for(i=0;i<this->mNumElements;i++) {
			*p1++ -= rl ;
			*p1++ -= im ;
		}
	}		
	return *this ;
}

TArray&	TArray::operator*=(const Cmplx& v) {
	float	*tData ;
	register	long	i ;
	register 	float	*p1,*p2 ;	
	register	float rl,im,rl2,im2 ;

	rl = v.x ;
	im = v.y ;
	if (!this->mIsComplex) {
		// We must convert the array to complex
		tData = new float[this->mNumElements*2] ;
		assert(tData != 0) ;
		this->mIsComplex = true ;
		p1 = tData ;
		p2 = this->mData ;
		for(i=0;i<this->mNumElements;i++) {
			rl2 = *p2++;
			*p1++ = rl2*rl ;
			*p1++ = rl2*im ;
		}
		delete[] this->mData ;
		this->mData = tData ;
	}
	else {
		p1 = this->mData ;
		for(i=0;i<this->mNumElements;i++) {
			rl2 = *p1++;
			im2 = *p1-- ;
			*p1++ = rl2*rl - im2*im ;
			*p1++ = rl2*im + im2*rl ;
		}
	}		
	return *this ;
}

TArray&	TArray::operator/=(const Cmplx& v) {
	float		*tData ;
	register	long		i ;
	register 	float	*p1,*p2 ;	
	register	float rl,im,l,rl2,im2 ;

	rl = v.x ;
	im = v.y ;
	l = v.x*v.x+v.y*v.y ;
	if (!this->mIsComplex) {
		// We must convert the array to complex
		tData = new float[this->mNumElements*2] ;
		assert(tData != 0) ;
		this->mIsComplex = true ;
		p1 = tData ;
		p2 = this->mData ;
		for(i=0;i<this->mNumElements;i++) {
			rl2 = *p2++;
			*p1++ = rl2*rl/l ;
			*p1++ = -rl2*im/l ;
		}
		delete[] this->mData ;
		this->mData = tData ;
	}
	else {
		p1 = this->mData ;
		for(i=0;i<this->mNumElements;i++) {
			rl2 = *p1++;
			im2 = *p1-- ;
			*p1++ = (rl2*rl + im2*im)/l ;
			*p1++ = (im2*rl - rl2*im)/l ;
		}
	}		
	return *this ;
}

TArray& TArray::operator+=(const TArray& v) {
	register	float	*p1,*p2,*p3 ;
	register	long	i ;

	// Operation will vary depending on the 
	// nature of the array to be equaled
	
	
	if (this->mIsComplex == v.mIsComplex) {
		// straight addition
		if (this->mIsComplex) {
			p1 = this->mData ;
			p2 = v.mData ;
			for(i=0;i<this->mNumElements;i++) {
				*p1++ += *p2++ ;
				*p1++ += *p2++ ;
			}
		}
		else {
			p1 = this->mData ;
			p2 = v.mData ;
			for(i=0;i<this->mNumElements;i++) 
				*p1++ += *p2++ ;
		}
	}
	else if (this->mIsComplex && !v.mIsComplex) {
		p1 = this->mData ;
		p2 = v.mData ;
		for(i=0;i<this->mNumElements;i++) {
			*p1++ += *p2++ ;
			p1++ ;
		}
	}
	else {
		// The receiving array is real but is being operated
		// on by a complex array so we need to adjust size
		// and data type
		float	*tData = new float[this->mNumElements*2] ;
		assert(tData != 0) ;
		this->mIsComplex = true ;
		p1 = this->mData ;
		p2 = v.mData ;
		p3 = tData ;
		for(i=0;i<this->mNumElements;i++) {
			*p3++ = *p1++ + *p2++ ;
			*p3++ = *p2++ ;
		}
		delete[] this->mData ;
		this->mData = tData ;
	}
	return *this ;
}

TArray& TArray::operator-=(const TArray& v) {
	register float	*p1,*p2,*p3 ;
	register long	i ;

	assert(this->mNumElements == v.mNumElements) ;


	// Operation will vary depending on the 
	// nature of the array to be equaled
	if (this->mIsComplex == v.mIsComplex) {
		// straight subtraction
		if (this->mIsComplex) {
			p1 = this->mData ;
			p2 = v.mData ;
			for(i=0;i<this->mNumElements;i++) {
				*p1++ -= *p2++ ;
				*p1++ -= *p2++ ;
			}
		}
		else {
			p1 = this->mData ;
			p2 = v.mData ;
			for(i=0;i<this->mNumElements;i++) 
				*p1++ -= *p2++ ;
		}
	}
	else if (this->mIsComplex && !v.mIsComplex) {
		p1 = this->mData ;
		p2 = v.mData ;
		for(i=0;i<this->mNumElements;i++) {
			*p1++ -= *p2++ ;
			p1++ ;
		}
	}
	else {
		// The receiving array is real but is being operated
		// on by a complex array so we need to adjust size
		// and data type
		float	*tData = new float[this->mNumElements*2] ;
		assert(tData != 0) ;
		this->mIsComplex = true ;
		p1 = this->mData ;
		p2 = v.mData ;
		p3 = tData ;
		for(i=0;i<this->mNumElements;i++) {
			*p3++ = *p1++ - *p2++ ;
			*p3++ = - *p2++ ;
		}
		delete[] this->mData ;
		this->mData = tData ;
	}
	return *this ;
}

TArray& TArray::operator*=(const TArray& v) {
	register float	*p1,*p2,*p3 ;
	register long	i ;
	register float 	rl,im,rlv,imv ;

	assert(this->mNumElements == v.mNumElements) ;
		
	
	// Operation will vary depending on the 
	// nature of the array to be equaled
	if (this->mIsComplex == v.mIsComplex) {
		// straight multiplication
		if (this->mIsComplex) {
			p1 = this->mData ;
			p2 = v.mData ;
			for(i=0;i<this->mNumElements;i++) {
				rlv = *p2++ ;
				imv = *p2++ ;
				rl  = *p1++ ;
				im  = *p1-- ;
				*p1++ = rl*rlv - im*imv ;
				*p1++ = rl*imv + im*rlv ;
			}
		}
		else {
			p1 = this->mData ;
			p2 = v.mData ;
			for(i=0;i<this->mNumElements;i++) 
				*p1++ *= *p2++ ;
		}
	}
	else if (this->mIsComplex && !v.mIsComplex) {
		p1 = this->mData ;
		p2 = v.mData ;
		for(i=0;i<this->mNumElements;i++) {
			rlv = *p2++ ;
			*p1++ *= rlv ;
			*p1++ *= rlv ;
		}
	}
	else {
		// The receiving array is real but is being operated
		// on by a complex array so we need to adjust size
		// and data type
		float	*tData = new float[this->mNumElements*2] ;
		assert(tData != 0) ;
		this->mIsComplex = true ;
		p1 = this->mData ;
		p2 = v.mData ;
		p3 = tData ;
		for(i=0;i<this->mNumElements;i++) {
			rlv = *p2++ ;
			imv = *p2++ ;
			rl  = *p1++ ;
			*p3++ = rl*rlv ;
			*p3++ = rl*imv ;
		}
		delete[] this->mData ;
		this->mData = tData ;
	}
	return *this ;
}

TArray& TArray::operator/=(const TArray& v) {
	register float	*p1,*p2,*p3 ;
	register long	i ;
	register float 	rl,im,rlv,imv,l ;

	assert(this->mNumElements == v.mNumElements) ;
	
	// Operation will vary depending on the 
	// nature of the array to be equaled
	if (this->mIsComplex == v.mIsComplex) {
		// straight division
		if (this->mIsComplex) {
			p1 = this->mData ;
			p2 = v.mData ;
			for(i=0;i<this->mNumElements;i++) {
				rlv = *p2++ ;
				imv = *p2++ ;
				l = rlv*rlv+imv*imv ;
				rl  = *p1++ ;
				im  = *p1-- ;
				*p1++ = (rl*rlv + im*imv)/l ;
				*p1++ = (rlv*im - imv*rl)/l ;
			}
		}
		else {
			p1 = this->mData ;
			p2 = v.mData ;
			for(i=0;i<this->mNumElements;i++) 
				*p1++ /= *p2++ ;
		}
	}
	else if (this->mIsComplex && !v.mIsComplex) {
		p1 = this->mData ;
		p2 = v.mData ;
		for(i=0;i<this->mNumElements;i++) {
			rlv = *p2++ ;
			*p1++ /= rlv ;
			*p1++ /= rlv ;
		}
	}
	else {
		// The receiving array is real but is being operated
		// on by a complex array so we need to adjust size
		// and data type
		float	*tData = new float[this->mNumElements*2] ;
		assert(tData != 0) ;
		this->mIsComplex = true ;
		p1 = this->mData ;
		p2 = v.mData ;
		p3 = tData ;
		for(i=0;i<this->mNumElements;i++) {
			rlv = *p2++ ;
			imv = *p2++ ;
			rl  = *p1++ ;
			l = rlv*rlv+imv*imv ;
			*p3++ = rl*rlv/l ;
			*p3++ = - imv*rl/l ;
		}
		delete[] this->mData ;
		this->mData = tData ;
	}
	return *this ;
}

float InnerProduct(const TArray& u,const TArray& v) {
	register float	*p1,*p2 ;
	register long	i ;
	register float 	rlu,imu,rlv,imv ;
	
	register	float product = 0. ;

	assert(u.mNumElements == v.mNumElements) ;

	// Operation will vary depending on the 
	// nature of the array to be equaled
	if (u.mIsComplex == v.mIsComplex) {
		// U  times V*
		if (u.mIsComplex) {
			p1 = u.mData ;
			p2 = v.mData ;
			for(i=0;i<u.mNumElements;i++) {
				rlu = *p1++ ;
				imu = *p1++ ;
				rlv = *p2++ ;
				imv = *p2++ ;
				product += rlu*rlv + imu*imv ;
			}
		}
		else {
			p1 = u.mData ;
			p2 = v.mData ;
			for(i=0;i<u.mNumElements;i++) {
				product += *p1++ * (*p2++) ;
			}
		}
	} 
	return product ;
}


TArray operator*(const TArray& u,const TArray& v) {
	register float	*p1,*p2,*p3 ;
	register long	i ;
	register float 	rlu,imu,rlv,imv ;

	assert(u.mNumElements == v.mNumElements) ;

	ArrayDataType type = (u.mIsComplex || v.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	// Operation will vary depending on the 
	// nature of the array to be equaled
	if (u.mIsComplex == v.mIsComplex) {
		// straight multiplication
		if (u.mIsComplex) {
			p1 = u.mData ;
			p2 = v.mData ;
			p3 = theArray.mData ;
			for(i=0;i<u.mNumElements;i++) {
				rlu = *p1++ ;
				imu = *p1++ ;
				rlv = *p2++ ;
				imv = *p2++ ;
				*p3++ = rlu*rlv - imu*imv ;
				*p3++ = rlu*imv + imu*rlv ;
			}
		}
		else {
			p1 = u.mData ;
			p2 = v.mData ;
			p3 = theArray.mData ;
			for(i=0;i<u.mNumElements;i++) {
				*p3++ = *p1++ * (*p2++) ;
			}
		}
	} 
	else if (u.mIsComplex && !v.mIsComplex) {
		p1 = u.mData ;
		p2 = v.mData ;
		p3 = theArray.mData ;
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			rlv = *p2++ ;
			*p3++ = rlu*rlv ;
			*p3++ = imu*rlv ;
		}
	}
	else {
		p1 = u.mData ;
		p2 = v.mData ;
		p3 = theArray.mData ;
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			rlv = *p2++ ;
			imv = *p2++ ;
			*p3++ = rlu*rlv ;
			*p3++ = rlu*imv ;
		}
	}
	return theArray ;
}


TArray operator/(const TArray& u,const TArray& v) {
	register float	*p1,*p2,*p3 ;
	register long	i ;
	register float 	rlu,imu,rlv,imv,l ;

	assert(u.mNumElements == v.mNumElements) ;

	ArrayDataType type = (u.mIsComplex || v.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;
	// Operation will vary depending on the 
	// nature of the array to be equaled
	if (u.mIsComplex == v.mIsComplex) {
		// straight division
		if (u.mIsComplex) {
			p1 = u.mData ;
			p2 = v.mData ;
			p3 = theArray.mData ;
			for(i=0;i<u.mNumElements;i++) {
				rlu = *p1++ ;
				imu = *p1++ ;
				rlv = *p2++ ;
				imv = *p2++ ;
				l = rlv*rlv+imv*imv ;
				*p3++ = (rlu*rlv + imu*imv)/l ;
				*p3++ = (rlv*imu - imv*rlu)/l ;
			}
		}
		else {
			p1 = u.mData ;
			p2 = v.mData ;
			p3 = theArray.mData ;
			for(i=0;i<u.mNumElements;i++) {
				*p3++ = *p1++ / (*p2++) ;
			}
		}
	} 
	else if (u.mIsComplex && !v.mIsComplex) {
		p1 = u.mData ;
		p2 = v.mData ;
		p3 = theArray.mData ;
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			rlv = *p2++ ;
			*p3++ = rlu/rlv ;
			*p3++ = imu/rlv ;
		}
	}
	else {
		p1 = u.mData ;
		p2 = v.mData ;
		p3 = theArray.mData ;
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			rlv = *p2++ ;
			imv = *p2++ ;
			l = rlv*rlv+imv*imv ;
			*p3++ = rlu*rlv/l ;
			*p3++ = -imv*rlu/l ;
		}
	}
	return theArray ;

}

TArray operator+(const TArray& u,const TArray& v) {
	register float	*p1,*p2,*p3 ;
	register long	i ;

	assert(u.mNumElements == v.mNumElements) ;

	ArrayDataType type = (u.mIsComplex || v.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;
	// Operation will vary depending on the 
	// nature of the array to be equaled
	
	if (u.mIsComplex == v.mIsComplex) {
		if (u.mIsComplex) {
			p1 = u.mData ;
			p2 = v.mData ;
			p3 = theArray.mData ;
			for(i=0;i<u.mNumElements;i++) {
				*p3++ = *p1++ + *p2++ ;
				*p3++ = *p1++ + *p2++ ;
			}
		}
		else {
			p1 = u.mData ;
			p2 = v.mData ;
			p3 = theArray.mData ;
			for(i=0;i<u.mNumElements;i++) 
				*p3++ = *p1++ + *p2++ ;
		}
	}
	else if (u.mIsComplex && !v.mIsComplex) {
		p1 = u.mData ;
		p2 = v.mData ;
		p3 = theArray.mData ;
		for(i=0;i<u.mNumElements;i++) {
			*p3++ = *p1++ + *p2++ ;
			*p3++ = *p1++ ;
		}
	}
	else {
		p1 = u.mData ;
		p2 = v.mData ;
		p3 = theArray.mData ;
		for(i=0;i<u.mNumElements;i++) {
			*p3++ = *p1++ + *p2++ ;
			*p3++ = *p2++ ;
		}
	}
	return theArray ;
}

TArray operator-(const TArray& u,const TArray& v) {
	register float	*p1,*p2,*p3 ;
	register long	i ;

	assert(u.mNumElements == v.mNumElements) ;

	ArrayDataType type = (u.mIsComplex || v.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;
	// Operation will vary depending on the 
	// nature of the array to be equaled
	if (u.mIsComplex == v.mIsComplex) {
		if (u.mIsComplex) {
			p1 = u.mData ;
			p2 = v.mData ;
			p3 = theArray.mData ;
			for(i=0;i<u.mNumElements;i++) {
				*p3++ = *p1++ - *p2++ ;
				*p3++ = *p1++ - *p2++ ;
			}
		}
		else {
			p1 = u.mData ;
			p2 = v.mData ;
			p3 = theArray.mData ;
			for(i=0;i<u.mNumElements;i++) 
				*p3++ = *p1++ - *p2++ ;
		}
	}
	else if (u.mIsComplex && !v.mIsComplex) {
		p1 = u.mData ;
		p2 = v.mData ;
		p3 = theArray.mData ;
		for(i=0;i<u.mNumElements;i++) {
			*p3++ = *p1++ - *p2++ ;
			*p3++ = *p1++ ;
		}
	}
	else {
		p1 = u.mData ;
		p2 = v.mData ;
		p3 = theArray.mData ;
		for(i=0;i<u.mNumElements;i++) {
			*p3++ = *p1++ - *p2++ ;
			*p3++ = *p2++ ;
		}
	}
	return theArray ;
}

TArray operator^(const TArray& u,const float v) {
	register	float	*p1,*p2 ;
	register	long	i ;
	
	ArrayDataType type = (u.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;

	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = ::pow(*p1++,v) ;	// Wrong
			*p2++ = ::pow(*p1++,v) ;	// Wrong
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			if (*p1 != 0.) 
				*p2++ = ::pow(*p1++,v) ;
			else {
				*p2++ = 0. ;
				p1++ ;
			}
		}
	}
	return theArray ;
}

TArray operator*(const TArray& u,const float v) {
	register	float	*p1,*p2 ;
	register	long	i ;
	
	ArrayDataType type = (u.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;

	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ * v ;
			*p2++ = *p1++ * v ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ * v ;
		}
	}
	return theArray ;
}

TArray operator/(const TArray& u,const float v) {
	register	float	*p1,*p2 ;
	register	long	i ;
	
	ArrayDataType type = (u.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;

	float	val = 1./v ;

	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ * val ;
			*p2++ = *p1++ * val ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ * val ;
		}
	}
	return theArray ;
}

TArray operator+(const TArray& u,const float v) {
	register	float	*p1,*p2 ;
	register	long	i ;
	
	ArrayDataType type = (u.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;

	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ + v ;
			*p2++ = *p1++ ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ + v ;
		}
	}
	return theArray ;
}

TArray operator-(const TArray& u,const float v) {
	register	float	*p1,*p2 ;
	register	long	i ;
	
	ArrayDataType type = (u.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;

	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ - v ;
			*p2++ = *p1++ ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ - v ;
		}
	}
	return theArray ;
}

TArray operator*(const float v,const TArray& u) {
	register	float	*p1,*p2 ;
	register	long	i ;
	
	ArrayDataType type = (u.mIsComplex) ? kComplexArray : kRealArray ;
		
	TArray	theArray(u.mWidth,u.mHeight,type) ;
	
	p1 = u.mData ;
	p2 = theArray.mData ;

	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = v * *p1++ ;
			*p2++ = v * *p1++;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = v * *p1++ ;
		}
	}
	return theArray ;
}

TArray operator+(const float v,const TArray& u) {
	register	float	*p1,*p2 ;
	register	long	i ;
	
	ArrayDataType type = (u.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;

	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = v + *p1++ ;
			*p2++ = *p1++;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = v + *p1++ ;
		}
	}
	return theArray ;
}

TArray operator-(const float v,const TArray& u) {
	register	float	*p1,*p2 ;
	register	long	i ;
	
	ArrayDataType type = (u.mIsComplex) ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;
	
	p1 = u.mData ;
	p2 = theArray.mData ;

	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = v - *p1++ ;
			*p2++ = - *p1++;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = v - *p1++ ;
		}
	}
	return theArray ;
}

TArray operator*(const TArray& u,const Cmplx& v) {
	register	float	*p1,*p2,rlu,rlv,imu,imv ;
	register	long	i ;
	
	TArray	theArray(u.mWidth,u.mHeight,kComplexArray) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	rlv = v.x ;
	imv = v.y ;
	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			*p2++ = rlu*rlv - imu*imv ;
			*p2++ = rlu*imv + imu*rlv ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			*p2++ = rlu*rlv ;
			*p2++ = rlu*imv ;
		}
	}
	return theArray ;
}

TArray operator/(const TArray& u,const Cmplx& v) {
	register	float	*p1,*p2,rlu,rlv,imu,imv,l ;
	register	long	i ;
	
	TArray	theArray(u.mWidth,u.mHeight,kComplexArray) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	rlv = v.x ;
	imv = v.y ;
	l = rlv*rlv+imv*imv ;
	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			*p2++ = (rlu*rlv + imu*imv)/l ;
			*p2++ = (imu*rlv - rlu*imv)/l  ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			*p2++ = rlu*rlv/l ;
			*p2++ = -rlu*imv/l  ;
		}
	}
	return theArray ;

}

TArray operator+(const TArray& u,const Cmplx& v) {
	register	float	*p1,*p2,rlv,imv ;
	register	long	i ;
	
	TArray	theArray(u.mWidth,u.mHeight,kComplexArray) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	rlv = v.x ;
	imv = v.y ;
	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ + rlv ;
			*p2++ = *p1++ + imv ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ + rlv ;
			*p2++ = imv ;
		}
	}
	return theArray ;
}

TArray operator-(const TArray& u,const Cmplx& v) {
	register	float	*p1,*p2,rlv,imv ;
	register	long	i ;
	
	TArray	theArray(u.mWidth,u.mHeight,kComplexArray) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	rlv = v.x ;
	imv = v.y ;
	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ - rlv ;
			*p2++ = *p1++ - imv ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = *p1++ - rlv ;
			*p2++ = -imv ;
		}
	}
	return theArray ;
}

TArray operator*(const Cmplx& u,const TArray& v) {
	register	float	*p1,*p2,rlu,imu,rlv,imv ;
	register	long	i ;
	
	TArray	theArray(v.mWidth,v.mHeight,kComplexArray) ;

	p1 = v.mData ;
	p2 = theArray.mData ;
	
	rlu = u.x ;
	imu = u.y ;
	if (v.mIsComplex) {
		for(i=0;i<v.mNumElements;i++) {
			rlv = *p1++ ;
			imv = *p1++ ;
			*p2++ = rlu*rlv - imu*imv ;
			*p2++ = rlu*imv + imu*rlv ;
		}
	}
	else {
		for(i=0;i<v.mNumElements;i++) {
			rlv = *p1++ ;
			*p2++ = rlu*rlv ;
			*p2++ = imu*rlv ;
		}
	}
	return theArray ;
}

TArray operator+(const Cmplx& u,const TArray& v) {
	register	float	*p1,*p2,rlu,imu ;
	register	long	i ;
	
	TArray	theArray(v.mWidth,v.mHeight,kComplexArray) ;

	p1 = v.mData ;
	p2 = theArray.mData ;
	
	rlu = u.x ;
	imu = u.y ;
	if (v.mIsComplex) {
		for(i=0;i<v.mNumElements;i++) {
			*p2++ = rlu + *p1++ ;
			*p2++ = imu + *p1++ ;
		}
	}
	else {
		for(i=0;i<v.mNumElements;i++) {
			*p2++ = rlu + *p1++ ;
			*p2++ = imu ;
		}
	}
	return theArray ;
}

TArray operator-(const Cmplx& u,const TArray& v) {
	register	float	*p1,*p2,rlu,imu ;
	register	long	i ;
	
	TArray	theArray(v.mWidth,v.mHeight,kComplexArray) ;

	p1 = v.mData ;
	p2 = theArray.mData ;
	
	rlu = u.x ;
	imu = u.y ;
	if (v.mIsComplex) {
		for(i=0;i<v.mNumElements;i++) {
			*p2++ = rlu - *p1++ ;
			*p2++ = imu - *p1++ ;
		}
	}
	else {
		for(i=0;i<v.mNumElements;i++) {
			*p2++ = rlu - *p1++ ;
			*p2++ = imu ;
		}
	}
	return theArray ;
}

TArray Modulus(TArray &u) {
	register	float	*p1,*p2,rlu,imu ;
	register	long	i ;
	
	TArray	theArray(u.mWidth,u.mHeight) ;	// Real

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			*p2++ = sqrt(rlu*rlu+imu*imu) ;
		}
	}
	else {
		// Just copy the data into the new array
		memcpy(theArray.mData,u.mData,u.mNumElements*sizeof(float)) ;
	}
	return theArray ;
	
}

TArray Intensity(TArray &u) {
	register	float	*p1,*p2,rlu,imu ;
	register	long	i ;
	
	TArray	theArray(u.mWidth,u.mHeight) ;	// Real

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			*p2++ = rlu*rlu+imu*imu ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			*p2++ = rlu*rlu ;
		}
	}
	return theArray ;
	
}

TArray Phase(TArray &u) {
	register	float	*p1,*p2,rlu,imu ;
	register	long	i ;
	
	TArray	theArray(u.mWidth,u.mHeight) ;	// Real

	p1 = u.mData ;
	p2 = theArray.mData ;
	float PiOver2 = asin(1.0) ;
	float MinusPiOver2 = -PiOver2 ;
	if (u.mIsComplex) {
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			if (rlu == 0.)
				if (imu > 0.)
					*p2++ = PiOver2 ;
				else 
					*p2++ = MinusPiOver2 ;
			else
				*p2++ = ::atan2(imu,rlu) ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
				*p2++ = 0. ;
		}
	}
	return theArray ;
}


TArray log(TArray &u) {
	register	float	*p1,*p2,a,p ;
	register	Cmplx	*c ;
	register	long	i ;

	
	ArrayDataType type = u.mIsComplex ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	if (u.mIsComplex) {
		// If we have a complex number Ae(iP) = exp( lnA + iP)
		// then the Ln [ Aexp(iP)] =  (lnA,P)
		c = (Cmplx *)u.mData ;
		for(i=0;i<u.mNumElements;i++) {
			a = c->Modulus() ;
			p = c->Phase() ;
			*p2++ = ::log(a) ;
			*p2++ = p ;
			c++  ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = ::log(*p1++) ;
		}
	}
	return theArray ;
}

TArray log10(TArray &u) {
	register	float	*p1,*p2,a,p ;
	register	Cmplx	*c ;
	register	long	i ;

	
	ArrayDataType type = u.mIsComplex ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	if (u.mIsComplex) {
		// If we have a complex number Ae(iP) = exp( lnA + iP)
		// then the Ln [ Aexp(iP)] =  (lnA,P)
		c = (Cmplx *)u.mData ;
		for(i=0;i<u.mNumElements;i++) {
			a = c->Modulus() ;
			p = c->Phase() ;
			*p2++ = ::log10(a) ;
			*p2++ = p ;
			c++  ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = ::log10(*p1++) ;
		}
	}
	return theArray ;
}

TArray pow(TArray &u,float x) {
	register	float	*p1,*p2 ;
	register	long	i ;

	
	ArrayDataType type = u.mIsComplex ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	if (u.mIsComplex) {
		// If we have a complex number , figure out what to do
		/*
		c = (Cmplx *)u.mData ;
		for(i=0;i<u.mNumElements;i++) {
			a = c->Modulus() ;
			p = c->Phase() ;
			*p2++ = ::log10(a) ;
			*p2++ = p ;
			c++  ;
		}
		*/
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = ::pow(*p1++,x) ;
		}
	}
	return theArray ;
}

TArray sqrt(TArray &u) {
	register	float	*p1,*p2,a,p ;
	register	Cmplx	*c ;
	register	long	i ;

	
	ArrayDataType type = u.mIsComplex ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	if (u.mIsComplex) {
		// If we have a complex number Ae(iP), 
		// then sqrt(AeiP) = sqrt(A)e(iP/2)
		c = (Cmplx *)u.mData ;
		for(i=0;i<u.mNumElements;i++) {
			a = ::sqrt(c->Modulus()) ;
			p = c->Phase()/2 ;
			*p2++ = a*::cos(p) ;
			*p2++ = a*::sin(p) ;
			c++ ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = ::sqrt(*p1++) ;
		}
	}
	return theArray ;
}

TArray sq(TArray &u) {
	register	float	*p1,*p2,a,p ;
	register	Cmplx	*c ;
	register	long	i ;

	
	ArrayDataType type = u.mIsComplex ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	if (u.mIsComplex) {
		// If we have a complex number Ae(iP), 
		// then sq(AeiP) = A*A(i2P)
		c = (Cmplx *)u.mData ;
		for(i=0;i<u.mNumElements;i++) {
			a = c->Modsq() ;
			p = 2*c->Phase() ;
			*p2++ = a*::cos(p) ;
			*p2++ = a*::sin(p) ;
			c++ ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = (*p1)*(*p1) ; p1++ ;
		}
	}
	return theArray ;
}

TArray exp(TArray &u) {
	register	float	*p1,*p2,rlu,imu,a ;
	register	long	i ;

	
	ArrayDataType type = u.mIsComplex ? kComplexArray : kRealArray ;
	TArray	theArray(u.mWidth,u.mHeight,type) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	if (u.mIsComplex) {
		// If we have a complex number x+iy, 
		// then exp(x+iy) = exp(x)*(cosy + isiny)
		for(i=0;i<u.mNumElements;i++) {
			rlu = *p1++ ;
			imu = *p1++ ;
			a = ::exp(rlu) ;
			*p2++ = a*cos(imu) ;
			*p2++ = a*sin(imu) ;
			p1 += 2 ;
		}
	}
	else {
		for(i=0;i<u.mNumElements;i++) {
			*p2++ = ::exp(*p1++) ;
		}
	}
	return theArray ;
}

TArray sin(TArray &u) {
	// I only know how to do this on real arrays
	register	float	*p1,*p2 ;
	register	long	i ;

	
	TArray	theArray(u.mWidth,u.mHeight) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	for(i=0;i<u.mNumElements;i++) {
		*p2++ = ::sin(*p1++) ;
	}
	return theArray ;
}

TArray cos(TArray &u) {
	// I only know how to do this on real arrays
	register	float	*p1,*p2 ;
	register	long	i ;

	
	TArray	theArray(u.mWidth,u.mHeight) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	for(i=0;i<u.mNumElements;i++) {
		*p2++ = ::cos(*p1++) ;
	}
	return theArray ;
}

TArray tan(TArray &u) {
	// I only know how to do this on real arrays
	register	float	*p1,*p2 ;
	register	long	i ;

	
	TArray	theArray(u.mWidth,u.mHeight) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	for(i=0;i<u.mNumElements;i++) {
		*p2++ = ::tan(*p1++) ;
	}
	return theArray ;
}

TArray asin(TArray &u) {
	// I only know how to do this on real arrays
	register	float	*p1,*p2 ;
	register	long	i ;

	
	TArray	theArray(u.mWidth,u.mHeight) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	for(i=0;i<u.mNumElements;i++) {
		*p2++ = ::asin(*p1++) ;
	}
	return theArray ;
}

TArray acos(TArray &u) {
	// I only know how to do this on real arrays
	register	float	*p1,*p2 ;
	register	long	i ;

	
	TArray	theArray(u.mWidth,u.mHeight) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	for(i=0;i<u.mNumElements;i++) {
		*p2++ = ::acos(*p1++) ;
	}
	return theArray ;
}

TArray atan(TArray &u) {
	// I only know how to do this on real arrays
	register	float	*p1,*p2 ;
	register	long	i ;

	
	TArray	theArray(u.mWidth,u.mHeight) ;

	p1 = u.mData ;
	p2 = theArray.mData ;
	
	for(i=0;i<u.mNumElements;i++) {
		*p2++ = ::atan(*p1++) ;
	}
	return theArray ;
}


TArray	Real(TArray &u) {
	if (!u.mIsComplex) 
		return u ;
		
	TArray theArray(u.mWidth,u.mHeight) ;
		
	float	*p1,*p2 ;
	long	i ;
	p1 = u.mData ;
	p2 = theArray.GetData() ;
	for(i=0;i < u.mNumElements;i++) {
	    *p2++  = *p1++ ;
	    p1++ ;
	}
	return theArray ;
}

TArray	Imag(TArray &u) {
	TArray theArray(u.mWidth,u.mHeight) ;
	if (!u.mIsComplex) {
		theArray = 0. ;
	}
	else {
		float	*p1,*p2 ;
		long	i ;
		p1 = u.mData ;
		p2 = theArray.mData+1 ;
		for(i=0;i < u.mNumElements;i++) {
		    *p2++  = *p1++ ;
		    p1++ ;
		}
	}
	return theArray ;
}

TArray	*TArray::Real() {
	TArray *theArray = new TArray(mWidth,mHeight) ;
	if (theArray != nil) {
		float	*p1,*p2 ;
		long	i ;
		p1 = mData ;
		p2 = theArray->GetData() ;
		for(i=0;i < mNumElements;i++) {
		    *p2++  = *p1++ ;
		    p1++ ;
		}
	}
	return theArray ;
}

TArray	*TArray::Imag() {
	TArray *theArray = new TArray(mWidth,mHeight) ;
	if (theArray != nil) {
		float	*p1,*p2 ;
		long	i ;
		p1 = mData ;
		p2 = theArray->mData+1 ;
		for(i=0;i < mNumElements;i++) {
		    *p2++  = *p1++ ;
		    p1++ ;
		}
	}
	return theArray ;
}

float *TArray::GetRealHK(long h,long k) {
	long nh,nk,ndx,maxDim = 2*(mWidth*mHeight-1) ;
	nh = h < 0 ? mWidth+h : h ;
	nk = k < 0 ? mHeight+k : k ;
	ndx = 2*(nk*mWidth + nh) ;
	if (ndx < 0 || ndx > maxDim) { fprintf(stderr," Error: Index (%d,%d) is outside array bounds [%d,%d]\n",h,k,mWidth,mHeight) ; ndx = 0 ; }
	return (mData + ndx) ;
}
 
float *TArray::GetImagHK(long h,long k) {
	long nh,nk,ndx,maxDim = 2*(mWidth*mHeight-1)+1 ;
	nh = h < 0 ? mWidth+h : h ;
	nk = k < 0 ? mHeight+k : k ;
	ndx = 2*(nk*mWidth + nh) + 1 ;
	if (ndx < 0 || ndx > maxDim) { fprintf(stderr," Error: Index (%d,%d) is outside array bounds [%d,%d]\n",h,k,mWidth,mHeight) ; ndx = 0 ; }
	return (mData + ndx) ;
}

float *TArray::GetDataHK(long h,long k) {
	long nh,nk ;
	nh = h < 0 ? mWidth+h : h ;
	nk = k < 0 ? mHeight+k : k ;
	return (mData + nk*mWidth + nh) ;
}

float TArray::GetPixelAmplitude(float x,float y) {
	// Operates only on real arrays
	if (x < 0.) return 0. ;
	if (y < 0.) return 0. ;
	if (x > mWidth-1) return 0. ;
	if (y > mHeight-1) return 0. ;
	int	ix = (int)x ;
	int iy = (int)y ;
	float dx = x-ix ;
	float dy = y-iy ;
	float value = (1.-dy)*((1-dx)*(*GetData(ix,iy)) + dx*(*GetData(ix+1,iy))) +
							dy*((1-dx)*(*GetData(ix,iy+1))+dx*(*GetData(ix+1,iy+1))) ;			

	return value ;
}

void	TArray::SetPixel(float x,float y,float val) {
	// Set the value of the pixel or distribute the intensity on 
	// the neighbors if we are somewhere in the middle
	if (x < 0.) return ;
	if (y < 0.) return ;
	if (x > mWidth-1) return  ;
	if (y > mHeight-1) return  ;
	int	ix = (int)x ;
	int iy = (int)y ;
	float dx = x-ix ;
	float dy = y-iy ;
	*(mData+iy*mWidth+ix) = (1.0-dx)*(1.0-dy)*val ;
	if (ix< mWidth-1)
		*(mData+iy*mWidth+ix+1) = dx*(1.0-dy)*val ;
	if (iy< mHeight-1)
		*(mData+(iy+1)*mWidth+ix) = (1.0-dx)*dy*val ;
	if (iy< mHeight-1 && ix< mWidth-1)
		*(mData+(iy+1)*mWidth+ix+1) = dx*dy*val ;
}	


// -------- Fourier routines for performing FFTs on the arrays
void fft2d(float data[],long width,long height,long isign) {
    long n2,m2,inc,maxSize ;
	float *table ;
	float pi=3.14159265 ,rn2,rm2;
	register float *rp,ang,dang ;
	long nh ;
	register long i,j,k,l ;
	register float *rp1,*rp2,*c,*w,wr,wi,tr,ti ;
	long le,nu,newVal,last,inc2 ;
	register long ia,le1 ;
	
	rn2 = log2((float)width) ;
	rm2 = log2((float)height) ;
	n2 = near(rn2) ;
	m2 = near(rm2) ;
	
	maxSize = (width > height ? width : height) ;
	table = (float *)malloc(maxSize*sizeof(float)) ;
	if (table == (float *)0) {
		PutMessage("Out of Memory") ;
		return ;
	}
	
	/* Set up the table */
	last = (width-1)/2 ;
	ang = 0. ;
	dang = -isign*2.*pi/width ;
	for(i=0,rp=table;i<=last;i++,ang+=dang) {
		*rp++ = cos(ang) ;
		*rp++ = sin(ang) ;
	}
	
	
	// Do the rows
	inc = 1 ;	// increment of 1
	k = (n2+1)/2 ;
	nh = width/2 ;
//LDM: added float definition - should it be double?
	last = width - 1 - ::pow((float)2.,(float)k) ;
	for(i=0,rp=data;i<height;i++,rp+=2*width) {
		// Bin-reverse
		c = rp ;
		j = nh ;
		for(l=1;l<=last;l++) {
			if (l < j) {
				rp1 = c + l + l ;
				rp2 = c + j + j ;
				tr        	= *rp1 ;
				ti    		= *(rp1+1) ;
				*rp1       	= *rp2 ;
				*(rp1+1)   	= *(rp2+1) ;
				*rp2       	= tr ;
				*(rp2+1)   	= ti ;
			}
			k = nh ;
			while (j>=k) {
				j -= k ;
				k >>= 1 ;
			}
			j += k ;
		}
		
		// FFT the rows				
		le=2 ;
		for(l=0;l<n2;l++) {
			le1 = le/2 ;
			nu = 0 ;
			newVal = width/le ;
			for(j=0;j<le1;j++) {
				w = table + nu + nu ;
				wr = *w ;
				wi = *(w+1) ;
				for(ia=j;ia<width;ia += le) {
					rp1 = rp+ia+ia ;
					rp2 = rp1 + le1 + le1 ;
					tr      = wr*(*rp2) - wi*(*(rp2+1)) ; 	// T = W*Y  
					ti  	= wr*(*(rp2+1)) + wi*(*rp2) ; 	//  "   "  
					*rp2     = *rp1 - tr ;						// Y = X-T  
					*(rp2+1) = *(rp1+1) - ti ;				//  "   "   
					*rp1     += tr ;							// X = X+T  
					*(rp1+1) += ti ;						//  "   "  
					
				}
				nu += newVal ;
			}
			le += le ;
		}
	}
	
	if (height == 1) return ; /* 1-Dimensional FFT */
	
	// Do the columns
	if (height != width) {	/* Set up a new table */
		last = (height-1)/2 ;
		ang = 0. ;
		dang = -isign*2.*pi/height ;
		for(i=0,rp=table;i<=last;i++,ang+=dang) {
			*rp++ = cos(ang) ;
			*rp++ = sin(ang) ;
		}
	}
		
	inc = width ;
	inc2 = 2*inc ;
	k = (m2+1)/2 ;
	nh = height/2 ;
	last = height - 1 - ::pow((float)2.,(float)k) ; //LDM: added float here
	for(i=0,rp=data;i<width;i++,rp+=2) {
		// Bin-reverse
		c = rp ;
		j = nh ;
		for(l=1;l<=last;l++) {
			if (l < j) {
				rp1 = c + l*inc2 ;
				rp2 = c + j*inc2 ;
				tr        	= *rp1 ;
				ti    		= *(rp1+1) ;
				*rp1       	= *rp2 ;
				*(rp1+1)   	= *(rp2+1) ;
				*rp2       	= tr ;
				*(rp2+1)   	= ti ;
			}
			k = nh ;
			while (j>=k) {
				j -= k ;
				k >>= 1 ;
			}
			j += k ;
		}
		le=2 ;
		for(l=0;l<m2;l++) {
			le1 = le/2 ;
			nu = 0 ;
			newVal = height/le ;
			for(j=0;j<le1;j++) {
				w = table + nu + nu ;
				wr = *w ;
				wi = *(w+1) ;
				for(ia=j;ia<height;ia += le) {
					rp1 = rp+ia*inc2 ;
					rp2 = rp1 + le1*inc2 ;
					tr      = wr*(*rp2) - wi*(*(rp2+1)) ; 	// T = W*Y  
					ti  	= wr*(*(rp2+1)) + wi*(*rp2) ; 	//  "   "  
					*rp2     = *rp1 - tr ;					// Y = X-T  
					*(rp2+1) = *(rp1+1) - ti ;				//  "   "   
					*rp1     += tr ;						// X = X+T  
					*(rp1+1) += ti ;						//  "   "  
					
				}
				nu += newVal ;
			}
			le += le ;
		}
	}
	free(table) ;
}
void fft1d(long m,long inc,float *data,float *table) {
  	float ct[2] ;
	long n,le,nu,newVal ;
	register float *tp,*rp1,*rp2,*w ; 
	register long j,l,ia,le1,inc2 ;
//LDM: Added float definition here - should it be double?	
	n = ::pow((float)2.,(float)m) ; 
	tp = ct ;
	
	le=2 ;
	inc2 = 2*inc ;
	for(l=0;l<m;l++) {
		le1 = le/2 ;
		nu = 0 ;
		newVal = n/le ;
		for(j=0;j<le1;j++) {
			w = table + nu + nu ;
			for(ia=j;ia<n;ia += le) {
				if(inc2==2) {
					rp1 = data+ia+ia ;
					rp2 = rp1 + le1 + le1 ;
				}
				else {
					rp1 = data + ia*inc2 ;
					rp2 = rp1 + le1*inc2 ;
				}
				*tp      = *w*(*rp2) - *(w+1)*(*(rp2+1)) ; 	// T = W*Y  
				*(tp+1)  = *w*(*(rp2+1)) + *(w+1)*(*rp2) ; 	//  "   "  
				*rp2     = *rp1 - *tp ;						// Y = X-T  
				*(rp2+1) = *(rp1+1) - *(tp+1) ;				//  "   "   
				*rp1     += *tp ;							// X = X+T  
				*(rp1+1) += *(tp+1) ;						//  "   "  
				
			}
			nu += newVal ;
		}
		le += le ;
	}
}

void binrev(long m,long inc,float *c) {
  	float cmplx[2] ;
	long last,nh,n ;
	register long i,j,k,inc2 ;
	register float *rp1,*rp2,*ct ;
//LDM: Added float here - should it be double?	
	n = ::pow((float)2.,(float)m) ;
	k = (m+1)/2 ;
	nh = n/2 ;
	last = n - 1 - ::pow((float)2.,(float)k) ; //LDM: added float (double?)
	j = nh ;
	inc2 = 2*inc ;
	ct = cmplx ;
	for(i=1;i<=last;i++) {
		if (i < j) {
			if (inc==2) {
				rp1 = c + i + i ;
				rp2 = c + j + j ;
			}
			else {
				rp1 = c + i*inc2 ;
				rp2 = c + j*inc2 ;
			}
			*ct        = *rp1 ;
			*(ct+1)    = *(rp1+1) ;
			*rp1       = *rp2 ;
			*(rp1+1)   = *(rp2+1) ;
			*rp2       = *ct ;
			*(rp2+1)   = *(ct+1) ;
		}
		k = nh ;
		while (j>=k) {
			j -= k ;
			k >>= 1 ;
		}
		j += k ;
	}
}	

