        real*4 function Fproject(Y,F,YY,MAXMIX,MEMORY,DELTA)
!       Projects the current residue onto prior gradient information
        implicit real*4 (a-h,o-z)
        dimension Y(MAXMIX,MEMORY), YY(MEMORY,MEMORY)
        dimension F(MAXMIX)
        real*4 YYINV(MEMORY,MEMORY),YPROJ(MEMORY),YTG(MEMORY),
     X           YT(MEMORY,MEMORY)
!--------------------------------------------------------------------
!       Estimate the NFL Projection
!
!       Input Parameters
!       MAXMIX:         Length of the variable vector
!       MEMORY:         Total number of memory values to use
!                          Most recent is last
!       Y:              Conventional Y arrays
!       YY:             Matrix of Y*Y values
!       F:              -Grad(MAXMIX) at the current point (residue)
!
!       Output
!       FProject        Ratio of Undefined/Defined
!
!       Delta: Term to add to Diagonal for inverse
!        Parameter (Delta=1D-6)
!
!--------------------------------------------------------------------
!        allocate (YYINV(MEMORY,MEMORY),YPROJ(MEMORY),YTG(MEMORY))
!        allocate (YT(MEMORY,MEMORY) )
        fproject=-1.D0

!       Invert YY
        ninverts=0
        DO N=1,MEMORY
         DO M=1,MEMORY
           YT(M,N)=YY(M,N)
         ENDDO
         YT(N,N)=YY(N,N)+Delta
        ENDDO
!
        ADD=1D-4
1000    continue
        DO N=1,MEMORY
         DO M=1,MEMORY
           YYINV(M,N)=YT(M,N)
         ENDDO
        ENDDO
        call minverse(YYINV,MEMORY,MEMORY,IFAIL)
!       If the matrix is singular, add a diagonal component to it
        if(ifail .lt. 0)then
                  ninverts=ninverts+1
                  do J=1,MEMORY
                     YT(J,J)=YT(J,J)+ADD
                  enddo
                  ADD=ADD*2.D0
                  write(6,*)'Adding ',add
                  if(ninverts .lt. 10)goto 1000
!                 Bomb
                  goto 100
        endif
!
        DO N=1,MEMORY
!               Dot product Y_T . F
                YTG(N)=0
                DO K=1,MAXMIX
                   YTG(N)=YTG(N)+Y(K,N)*F(K)
                ENDDO
        ENDDO
!       Multiply by YYINV
        DO N=1,MEMORY
           YPROJ(N)=0
           DO K=1,MEMORY
              YPROJ(N)=YPROJ(N)+YTG(K)*YYINV(K,N)
           ENDDO
        ENDDO
!       Create the undefined part
        Residue=0.0
        Defined=0.0
        DO J=1,MAXMIX
           SP = F(J)
           Defined = Defined+SP*SP
           DO N=1,MEMORY
              SP = SP - Y(J,N)*YPROJ(N)
           ENDDO
           Residue = Residue + SP*SP
        ENDDO
        Defined=sqrt(Defined)
        if(Residue .gt. 1D-30)then
                FProject=sqrt(Residue)/Defined
        else
                FProject=0.0D0
        endif
100     continue !deallocate(YT, YTG, YPROJ, YYINV)
        return
        end


           real*4 Function BBProject(Y,S,YY,F,MAXMIX,MEMORY,DELTA)
           implicit real*4 (a-h,o-z)
           dimension F(MAXMIX),YY(MEMORY,MEMORY)
           dimension Y(MAXMIX,MEMORY),S(MAXMIX,MEMORY)
           real*4 WTMP(MAXMIX),ZTMP(MAXMIX)
!--------------------------------------------------------------------
!          Find the BB residue to known ratio
!
!          Input Parameters
!          MAXMIX:         Length of the variable vector
!          MEMORY:         Total number of memory values to use
!                          Most recent is last
!          Y,S:            Conventional Y and S arrays
!          YY:             Matrix of Y*Y values
!          F:              -Grad(MAXMIX) at the current point (residue)
!
!          Output
!          Value           |Unknown|/|Input|
!--------------------------------------------------------------------
!
!           allocate(WTMP(MAXMIX),ZTMP(MAXMIX))
           TEMPZW = 0
           DO J=1,MAXMIX
              TEMPZW=TEMPZW-Y(J,MEMORY)*F(J)
           ENDDO
           TEMPZW=  TEMPZW/YY(MEMORY,MEMORY)

           DO J=1,MAXMIX
              Wtmp(J)= -F(J)
              Ztmp(J)=  S(J,MEMORY)*TEMPZW
           ENDDO
           DO J=0,MEMORY-2
              TEMPW=TEMPZW
              TEMPZW=0.0D0
              DO K=1,MAXMIX
                 Wtmp(K) = Wtmp(K)- Y(K,MEMORY-J)*TEMPW
                 TEMPZW =  TEMPZW+Y(K,MEMORY-J-1)*Wtmp(K)
              ENDDO
              TEMPZW = TEMPZW/YY(MEMORY-J-1,MEMORY-J-1)
              DO K=1,MAXMIX
                 Ztmp(K)=Ztmp(K)+S(K,MEMORY-J-1)*TEMPZW
              ENDDO
           ENDDO
!
!           Finish up
           Defined=0
           Residue=0
           DO J=1,MAXMIX
              Wtmp(J)=Wtmp(J) - Y(J,1)*TEMPZW
              Defined=Defined+F(J)*F(J)
!             Residue
              Residue=Residue+Wtmp(J)*Wtmp(J)
           enddo
           Defined=sqrt(Defined)
           if(Residue .gt. 1D-30)then
                BBProject=sqrt(Residue)/Defined
           else
                BBProject=0.0D0
           endif
!           deallocate(Wtmp, Ztmp)

           return
           end

      Real*4 Function GBMProject(Y,S,SY,F,MAXMIX,MEMORY,DELTA)
           implicit real*4 (a-h,o-z)
           dimension F(MAXMIX)
           dimension Y(MAXMIX,MEMORY),S(MAXMIX,MEMORY)
           dimension SY(MEMORY,MEMORY)
           real*4 SYINV(MEMORY,MEMORY),YPROJ(MEMORY),
     X           YTG(MEMORY), YT(MEMORY,MEMORY)
C           real*4,allocatable :: Undef(:), Def(:)
!--------------------------------------------------------------------
!          Find the Good-Broyden (Closest Point) Residue
!
!          Input Parameters
!          MAXMIX:         Length of the variable vector
!          MEMORY:         Total number of memory values to use
!                          Most recent is last
!          Y,S:            Conventional Y and S arrays
!          SY:             Matrix of S*Y values
!          F:              -Grad(MAXMIX) at the current point (residue)
!
!          Output
!          Ration of Unknown/Known
!
!       Delta: Term to add to Diagonal for inverse
!        Parameter (Delta=1D-6)
!
!--------------------------------------------------------------------
!        allocate (SYINV(MEMORY,MEMORY),YPROJ(MEMORY),YTG(MEMORY))
!        allocate (YT(MEMORY,MEMORY) )
!       Invert YY
        ninverts=0
!       Note: SY is negative here
        DO N=1,MEMORY
         DO M=1,MEMORY
           YT(M,N)=SY(M,N)
         ENDDO
         YT(N,N)=SY(N,N)-Delta
        ENDDO
!
        ADD=1D-4
1000    continue
        DO N=1,MEMORY
         DO M=1,MEMORY
           SYINV(M,N)=YT(M,N)
         ENDDO
        ENDDO
!
        call minverse(sYINV,MEMORY,MEMORY,IFAIL)
!       If the matrix is singular, add a diagonal component to it
        if(ifail .lt. 0)then
                  ninverts=ninverts+1
                  do J=1,MEMORY
                     YT(J,J)=YT(J,J)+ADD
                  enddo
                  ADD=ADD*2.D0
                  if(ninverts .lt. 10)goto 1000
!                 Bomb to Pratt
                  IFAIL=-1
                  GBMProject=-999
                  goto 100
        endif
!
        ntries=0
2000    continue
        DO N=1,MEMORY
!       Dot product Y_T . F
                YTG(N)=0
                DO K=1,MAXMIX
                   YTG(N)=YTG(N)+S(K,N)*F(K)
                ENDDO
        ENDDO
!       Multiply by SYINV
        DO N=1,MEMORY
           YPROJ(N)=0
           DO K=1,MEMORY
              YPROJ(N)=YPROJ(N)+YTG(K)*SYINV(N,K)
           ENDDO
        ENDDO
!       Create the undefined part
        Residue=0.0
        Defined=0.0
        DO J=1,MAXMIX
           SP = F(J)
           Defined = Defined+SP*SP
           DO N=1,MEMORY
              SP = SP - Y(J,N)*YPROJ(N)
           ENDDO
           Residue = Residue + SP*SP
        ENDDO
        Defined=sqrt(Defined)
        if(Residue .gt. 1D-30)then
                GBMProject=sqrt(Residue)/Defined
        else
                GBMProject=0.0D0
        endif
100     continue !deallocate (SYINV, YTG, YPROJ, YT )
        RETURN
        END

      real*4 Function GBProject(Y,S,SY,SS,F,MAXMIX,MEMORY,DMIX,IFAIL
     X ,DELTA)
           implicit real*4 (a-h,o-z)
           dimension F(MAXMIX)
           dimension Y(MAXMIX,MEMORY),S(MAXMIX,MEMORY)
           dimension SS(MEMORY,MEMORY),SY(MEMORY,MEMORY)
           real*4 YYINV(MEMORY,MEMORY),YPROJ(MEMORY),YTG(MEMORY),
     X     YT(MEMORY,MEMORY)
!           real*4 Undef(:), Def(:)
!--------------------------------------------------------------------
!          Good Broyden according to LDM
!
!          Input Parameters
!          MAXMIX:         Length of the variable vector
!          MEMORY:         Total number of memory values to use
!                          Most recent is last
!          Y,S:            Conventional Y and S arrays
!          YY:             Matrix of Y*Y values
!          SY:             Matrix of S*Y values
!          F:              -Grad(MAXMIX) at the current point (residue)
!          HZERO:          Diagonal initial matrix
!          DMIX:           Scaler for initial matrix
!
!          Output
!          STEP            Multi-Secant Step
!
!
!--------------------------------------------------------------------
!        allocate (YYINV(MEMORY,MEMORY),YPROJ(MEMORY),YTG(MEMORY))
!        allocate (YT(MEMORY,MEMORY) )
!       Invert YY
        ninverts=0
!       
        DO N=1,MEMORY
         DO M=1,MEMORY
           YT(M,N)=SY(M,N)
         ENDDO
        ENDDO
!       Difference between BB and Multi-Secant, from Byrd Nocedal Schnabel 6.2
        DO I=2,MEMORY
                DO J=1,I-1
                        YT(I,J)=dmix*YT(I,J)-SS(I,J)
                ENDDO
        ENDDO

!
!       Note, Add is negative since matrix is negative
        ADD=-DELTA
1000    continue
        DO N=1,MEMORY
         DO M=1,MEMORY
           YYINV(M,N)=YT(M,N)
         ENDDO
        ENDDO
        call minverse(YYINV,MEMORY,MEMORY,IFAIL)
!
!       If the matrix is singular, add a diagonal component to it
        if(ifail .lt. 0)then
                  Write(21,*)':DIRE Singular, adding ',add
                  ninverts=ninverts+1
                  do J=1,MEMORY
                     YT(J,J)=YT(J,J)+ADD
                  enddo
                  ADD=ADD*2.D0
                  if(ninverts .lt. 10)goto 1000
!                 Bomb to Pratt
                  IFAIL=-1
                  goto 100
        endif
!
        ntries=0
2000    continue
        DO N=1,MEMORY
!       Dot product Y_T . F
                YTG(N)=0
                DO K=1,MAXMIX
                   YTG(N)=YTG(N)+S(K,N)*F(K)
                ENDDO
        ENDDO
!       Multiply by YYINV
        DO N=1,MEMORY
           YPROJ(N)=0
           DO K=1,MEMORY
              YPROJ(N)=YPROJ(N)+YTG(K)*YYINV(N,K)
           ENDDO
        ENDDO
!       Create the step
        Residue=0
        Defined=0
        DO J=1,MAXMIX
!          New Position
           SP = F(J)
           Defined = Defined+SP*SP
           DO N=1,MEMORY
              SP = SP - Y(J,N)*YPROJ(N)
           ENDDO
           Residue = Residue + SP*SP
        ENDDO
        Defined=sqrt(Defined)
        if(Residue .gt. 1D-30)then
                GBProject=sqrt(Residue)/Defined
        else
                GBProject=0.0D0
        endif
        IFAIL=0
100     continue !deallocate (YYINV, YTG, YPROJ )
        RETURN
        END

