      subroutine feval(M,X,RESID)
C
C     Calculate the robust residual
C     Input
C            F          Array of measurements
C            NSET       Number of Sets
C            NRNG       Range to scan
C            SCL        Scaling
C
C     Output
C            WT         Array of relative weights for each pt
C
C
      implicit real*8 (a-h,o-z)
      include 'msizes.h'
      dimension PTS(500)
      real*4 X(M),RESID
      parameter (fit2=-1.01056534E-02, fit1=9.14666467E-02,
     X          fit0=9.18710524E-01)
      logical init
C      data init /.true./
C      save BERR,INIT
      common /ceval/berr,power,init
C
C     Power rescaling for normal data
      if(abs(power-1.).lt.1e-6)then
            call fevalone(M,X,RESID)
            return
      else if(abs(power-2.).lt.1e-6)then
            call fevaltwo(M,X,RESID)
            return
      endif
      prescale = 1./(fit2*power*power+fit1*power+fit0)
      NSET=M+1
      scl(1)=1.D0
C      write(6,*)M,NSET,(X(J),J=1,M)
      do j=2,NSET
         scl(j)=abs(X(j-1))
      enddo
      ERR=0.D0
      BASE=0.D0
      do j=-IXBIG,IXBIG
           do k=-IYBIG,IYBIG
              N=0
              do i=1,NSET
                if( F(i,j,k).gt.-1)then
                        N=N+1
                        PTS(N)=F(i,j,k)*scl(i)
                endif
              enddo
              IF(N .GT. 3)then
                   DO I1=1,N-1
                      DO I2=I1+1,N
                         T=abs(PTS(I1)-PTS(I2))
                         err=err+T**power !abs(PTS(I1)-PTS(I2))
                         base=base+(PTS(I1)+PTS(I2))**power
                      ENDDO
                   ENDDO
              ENDIF
          enddo
      enddo
      err=100*((err/base)**(1./power))*prescale
      IF(init)then
        BERR=ERR
        init=.false.
      ENDIF
      ERR=ERR-BERR
      RESID=ERR
C      write(6,*)ERR,M
      return
      end
      subroutine feval2(M,X,RESID)
C
C     Calculate the robust residual
C     Input
C            F          Array of measurements
C            NSET       Number of Sets
C            NRNG       Range to scan
C            SCL        Scaling
C
C     Output
C            WT         Array of relative weights for each pt
C
C
      implicit real*8 (a-h,o-z)
      include 'msizes.h'
      dimension PTS(500),W(500)
      real*4 X(M),RESID
      logical init
C      data init /.true./
C      save BERR,INIT
      common /ceval/berr,power,init
C
      NSET=M+1
      scl(1)=1.D0
C      write(6,*)M,NSET,(X(J),J=1,M)
      do j=2,NSET
         scl(j)=abs(X(j-1))
      enddo
      ERR=0.D0
      do j=-IXBIG,IXBIG
           do k=-IYBIG,IYBIG
              N=0
              do i=1,NSET
                if( F(i,j,k).gt.-1)then
                        N=N+1
                        PTS(N)=F(i,j,k)*scl(i)
                        W(N)=wt(i,j,k)
                endif
              enddo
              IF(N .GT. 3)then
                   DO I1=1,N-1
                      DO I2=I1+1,N
                         T=abs(PTS(I1)-PTS(I2))
                         err=err+(T**power)*(W(I1)+W(I2)) !abs(PTS(I1)-PTS(I2))
                      ENDDO
                   ENDDO
              ENDIF
          enddo
      enddo
      IF(init)then
        BERR=ERR
        init=.false.
      ENDIF
      ERR=ERR-BERR
      RESID=ERR
C      write(6,*)ERR,M
      return
      end
      subroutine fevalone(M,X,RESID)
C
C     Calculate the robust residual
C     Input
C            F          Array of measurements
C            NSET       Number of Sets
C            NRNG       Range to scan
C            SCL        Scaling
C
C     Output
C            WT         Array of relative weights for each pt
C
C
      implicit real*8 (a-h,o-z)
      include 'msizes.h'
      dimension PTS(500)
      real*4 X(M),RESID
      parameter (fit2=-1.01056534E-02, fit1=9.14666467E-02,
     X          fit0=9.18710524E-01)
      logical init
C      data init /.true./
C      save BERR,INIT
      common /ceval/berr,power,init
C
C     Power rescaling for normal data

      prescale = 1./(fit2+fit1+fit0)
      NSET=M+1
      scl(1)=1.D0
      do j=2,NSET
         scl(j)=abs(X(j-1))
      enddo
      ERR=0.D0
      BASE=0.D0
      do j=-IXBIG,IXBIG
           do k=-IYBIG,IYBIG
              N=0
              do i=1,NSET
                if( F(i,j,k).gt.-1)then
                        N=N+1
                        PTS(N)=F(i,j,k)*scl(i)
                endif
              enddo
              IF(N .GT. 3)then
                   DO I1=1,N-1
                      DO I2=I1+1,N
                         T=abs(PTS(I1)-PTS(I2))
                         err=err+T
                         base=base+(PTS(I1)+PTS(I2))
                      ENDDO
                   ENDDO
              ENDIF
          enddo
      enddo
      err=100*((err/base))*prescale
      IF(init)then
        BERR=ERR
        init=.false.
      ENDIF
      ERR=ERR-BERR
      RESID=ERR
C      write(6,*)ERR,M
      return
      end
      subroutine fevaltwo(M,X,RESID)
C
C     Calculate the rms residual
C     Input
C            F          Array of measurements
C            NSET       Number of Sets
C            NRNG       Range to scan
C            SCL        Scaling
C
C     Output
C            WT         Array of relative weights for each pt
C
C
      implicit real*8 (a-h,o-z)
      include 'msizes.h'
      dimension PTS(500)
      real*4 X(M),RESID
      parameter (fit2=-1.01056534E-02, fit1=9.14666467E-02,
     X          fit0=9.18710524E-01)
      logical init
C      data init /.true./
C      save BERR,INIT
      common /ceval/berr,power,init
C

      prescale = 1./(fit2*4+fit1*2+fit0)
      NSET=M+1
      scl(1)=1.D0
      do j=2,NSET
         scl(j)=abs(X(j-1))
      enddo
      ERR=0.D0
      BASE=0.D0
      do j=-IXBIG,IXBIG
           do k=-IYBIG,IYBIG
              N=0
              do i=1,NSET
                if( F(i,j,k).gt.-1)then
                        N=N+1
                        PTS(N)=F(i,j,k)*scl(i)
                endif
              enddo
              IF(N .GT. 3)then
                   DO I1=1,N-1
                      DO I2=I1+1,N
                         T=abs(PTS(I1)-PTS(I2))
                         err=err+T*T
                         T=(PTS(I1)+PTS(I2))
                         base=base+T*T
                      ENDDO
                   ENDDO
              ENDIF
          enddo
      enddo
      err=100*(sqrt(err/base))*prescale
      IF(init)then
        BERR=ERR
        init=.false.
      ENDIF
      ERR=ERR-BERR
      RESID=ERR
C      write(6,*)ERR,M
      return
      end

