/*
 * atom_bk.c
 * converted from ATOM.F by Bird & King
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "utils.h"
#include "complex.h"
#include "atom_bk.h"

#define ATOM_TABLE "mbfitbk.dat"
#define MS2_THRESH 2.00
#define BUFLEN 256

double quad_int(double x[], double xx, double f[])
{
  /* Use Lagrange interpolating quadratic */
  double x1, x2, x3, x12, x13, x23;
  x1 = xx - x[0];
  x2 = xx - x[1];
  x3 = xx - x[2];
  x12 = x[0] - x[1];
  x13 = x[0] - x[2];
  x23 = x[1] - x[2];
  return f[0] * x2 * x3 / (x12 * x13)
    - f[1] * x1 * x3 / (x12 * x23)
      + f[2] * x1 * x2 / (x13 * x23);
}


double fabs_bi_quad_int(
  double m, double ms2, int i_ms, int i_ms2,
  double gm[], double gms2[], double cf[NUM_GM][NUM_GMS2],
  double r_mlt, double r_add
  )
{
  double x[3], f[3], xx[3], ff[3], expms2;
  int i, j;
  for(j=0; j<3; j++){
    expms2 = exp(gms2[j + i_ms2 - 2]);
    for(i=0; i<3; i++){
      x[i] = gm[i + i_ms - 2];
      f[i] = expms2 * ( cf[i + i_ms - 2][j + i_ms2 - 2] * r_mlt - r_add );
    }
    ff[j] = quad_int(x, m, f);
    xx[j] = gms2[j + i_ms2 - 2];
  }
  return quad_int(xx, ms2, ff);
}


double cub_int(double x[], double xx, double f[])
{
  /* Use Lagrange interpolating cubic */
  double x1, x2, x3, x4, x12, x13, x14, x23, x24, x34;
  x1 = xx - x[0];
  x2 = xx - x[1];
  x3 = xx - x[2];
  x4 = xx - x[3];
  x12 = x[0] - x[1];
  x13 = x[0] - x[2];
  x14 = x[0] - x[3];
  x23 = x[1] - x[2];
  x24 = x[1] - x[3];
  x34 = x[2] - x[3];
  return
    f[0] * x2 * x3 * x4 / (x12 * x13 * x14) -
      f[1] * x1 * x3 * x4 / (x12 * x23 * x24) +
        f[2] * x1 * x2 * x4 / (x13 * x23 * x34) -
          f[3] * x1 * x2 * x3 / (x14 * x24 * x34);
}


double fabs_bi_cub_int(
  double m, double ms2,
  double gm[], double gms2[], double cf[NUM_GM][NUM_GMS2],
  double r_mlt, double r_add
  )
{
  double rf[4], ss[4], fatm[4], satm[4], expms2;
  int i, j, i_ms, i_ms2;
  
  i_ms  = 1; do{ i_ms++; } while(m >= gm[i_ms] && i_ms < NUM_GM - 2);
  i_ms2 = 1; do{ i_ms2++; } while(ms2 >= gms2[i_ms2] && i_ms2 < NUM_GMS2 - 2);
  for(j=0; j<4; j++){
    expms2 = exp(gms2[i_ms2 + j - 2]);
    for(i=0; i<4; i++){
      rf[i] = expms2 * (cf[i + i_ms - 2][j + i_ms2 - 2] * r_mlt - r_add);
      ss[i] = gm[i + i_ms - 2];
    }
    fatm[j] = cub_int(ss, m, rf);
    satm[j] = gms2[j + i_ms2 - 2];
  }
  return cub_int(satm, ms2, fatm);
}


#if 0
double bi_cub_int(
  double m, double ms2, int n_gm, int n_gms2,
  double gm[], double gms2[], double cf[NUM_GM][NUM_GMS2]
  )
{
  double x[4], f[4], xatm[4], fatm[4];
  int i, j, i_ms, i_ms2;
  i_ms = 1;  do{ i_ms++; }  while(m >= gm[i_ms] && i_ms < NUM_GM - 2);
  i_ms2 = 1; do{ i_ms2++; } while(ms2 >= gms2[i_ms2] && i_ms2 < NUM_GMS2 - 2);
  for(j=0; j<4; j++){
    for(i=0; i<4; i++){
      x[i] = gm[i + i_ms - 2]; f[i] = cf[i + i_ms - 2][j + i_ms2 - 2];
    }
    fatm[j] = cub_int(x, m, f);
    xatm[j] = gms2[j + i_ms2 - 2];
  }
  return cub_int(xatm, ms2, fatm);
}
#endif


double cub_int_2(double xx, int n, double x[], double f[])
{
  /*
   * Use Lagrange interpolating cubic
   *  x: x[0] to x[n-1], f: f[0] to f[n-1]
   */
  double x1, x2, x3, x4, x12, x13, x14, x23, x24, x34;
  int i0;
  
  i0 = 1; do{i0++;} while(xx >= x[i0] && i0 < n - 2);
  /* printf("$$ i0 = %d\n",i0);*/
  x1 = xx - x[i0 - 2];
  x2 = xx - x[i0 - 1];
  x3 = xx - x[i0];
  x4 = xx - x[i0 + 1];
  x12 = x[i0 - 2] - x[i0 - 1];
  x13 = x[i0 - 2] - x[i0];
  x14 = x[i0 - 2] - x[i0 + 1];
  x23 = x[i0 - 1] - x[i0];
  x24 = x[i0 - 1] - x[i0 + 1];
  x34 = x[i0]     - x[i0 + 1];
  return
    f[i0 - 2] * x2 * x3 * x4 / (x12 * x13 * x14) -
      f[i0 - 1] * x1 * x3 * x4 / (x12 * x23 * x24) +
        f[i0    ] * x1 * x2 * x4 / (x13 * x23 * x34) -
          f[i0 + 1] * x1 * x2 * x3 / (x14 * x24 * x34);
}


int atom_bk(
  ATOMF_BK ftbl, double s, double m, complex *atomf, double kvolts
  )
     /*
      * Elastic and inelastic (absorptive) atomic scattering factor
      *  converted from ATOM (Bird & King)
      * 				!!!!! (without exp(-B s^2)) !!!!!
      */
{
  double	beta, sq_s, ms2, r_mlt, r_add;
  /*	double	ms, expms2;*/
  /*	double	ss[4], rf[4], fatm[4], satm[4]; */
  int i_ms, i_ms2;
  /*	int i, j, status, nt;*/
  
  atomf->re = 0.0;	atomf->im = 0.0;
  sq_s = s * s;	ms2 = m * sq_s;
  
  if(m < 0.0 || m > ftbl.gm[NUM_GM-1]){return 2;}
  if(s < 0.0){return 3;}
  if(kvolts <= 0.0 || kvolts > 1e+4){
    printf("$$ invalid kvolts (=%lf).\n",kvolts); return 4;
  }
  if( s > ftbl.dts[NUM_DT - 1]){
    printf("$$ s (=%lf) must be less than %lf\n", s, ftbl.dts[NUM_DT - 1]); return 1;
  }
  
  beta = sqrt( 1.0 - 261127.0 / ( 261127.0 + kvolts * ( kvolts + 1022.0) ) );
  /* printf("beta = %lf \n",beta); */
  
  /* ----- Elastic part ----- */
  /* atomf->re = exp(- m * sq_s) * cub_int_2(s, NUM_DT, ftbl.dts, ftbl.dt) / 1e+3; */
  atomf->re = cub_int_2(s, NUM_DT, ftbl.dts, ftbl.dt) / 1e+3;
  
  /* ----- Inelastic (absorption) part ----- */
  if(m < ftbl.gm[0]){ return 0; }					/* Success */
  if(ms2 > ftbl.gms2[NUM_GMS2 - 1]){ return 0; }	/* Success (modified by KT) */
  
  r_mlt = 10.0 / ftbl.mlt_fct;	r_add = 1e-6 * ftbl.add_fct;
  if( ms2 >= MS2_THRESH){			/* Interpolate absorptive part using bi-cubic */
    atomf->im = atomf->re / beta
      * fabs_bi_cub_int( m, ms2, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add );
  }
  else{							/* Interpolate absorptive part using bi-quadratic */
    i_ms = 0;  do{ i_ms++; }  while(m   >= ftbl.gm[i_ms]    && i_ms  < NUM_GM - 1);
    i_ms2 = 0; do{ i_ms2++; } while(ms2 >= ftbl.gms2[i_ms2] && i_ms2 < NUM_GMS2 - 1);
    
    if (i_ms == 1  && i_ms2 == 1 ){
      i_ms = 2; i_ms2 = 2;
      atomf->im = fabs_bi_quad_int(m, ms2, i_ms, i_ms2, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add);
    }
    if (i_ms == 1  && i_ms2 != 1 ){
      i_ms = 2;
      atomf->im = (
        fabs_bi_quad_int(m, ms2, i_ms, i_ms2,   ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        + fabs_bi_quad_int(m, ms2, i_ms, i_ms2+1, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        ) / 2.0;			
    }
    if (i_ms != 1  && i_ms2 == 1 ){
      i_ms2 = 2;
      atomf->im = (
        fabs_bi_quad_int(m, ms2, i_ms,   i_ms2, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        + fabs_bi_quad_int(m, ms2, i_ms+1, i_ms2, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        ) / 2.0;			
    }
    else if (i_ms == NUM_GM-1  && i_ms2 == NUM_GMS2-1 ){
      atomf->im = fabs_bi_quad_int(m, ms2, i_ms, i_ms2, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add);
    }
    else if(i_ms == NUM_GM-1 && i_ms2 != NUM_GMS2-1){
      atomf->im = (
        fabs_bi_quad_int(m, ms2, i_ms, i_ms2,   ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        + fabs_bi_quad_int(m, ms2, i_ms, i_ms2+1, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        ) / 2.0;			
    }
    else if(i_ms != NUM_GM-1 && i_ms2 == NUM_GMS2-1){
      atomf->im = (
        fabs_bi_quad_int(m, ms2, i_ms,   i_ms2, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        + fabs_bi_quad_int(m, ms2, i_ms+1, i_ms2, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        ) / 2.0;			
    }
    else{
      atomf->im = (
        fabs_bi_quad_int(m, ms2, i_ms,   i_ms2,   ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        + fabs_bi_quad_int(m, ms2, i_ms,   i_ms2+1, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        + fabs_bi_quad_int(m, ms2, i_ms+1, i_ms2,   ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        + fabs_bi_quad_int(m, ms2, i_ms+1, i_ms2+1, ftbl.gm, ftbl.gms2, ftbl.f, r_mlt, r_add)
        ) / 4.0;			
    }
    atomf->im *= atomf->re / beta;
  }
  return 0;	/* Success */
}


void dump_ftbl(ATOMF_BK ftbl)
{
  int i,j;
  printf("ftbl:\n");
  printf("mlt_fct = %lf, add_fct = %lf\n", ftbl.mlt_fct, ftbl.add_fct);
  for(i=0; i<NUM_DT; i++){
    printf("dts[%d] = %lf, dt[%d] = %lf\n", i, ftbl.dts[i], i, ftbl.dt[i]);
  }
  for(j=0; j<NUM_GMS2; j++){
    for(i=0; i<NUM_GM; i++){
      printf("gm[%d] = %lf, gms2[%d] = %lf, f[%d][%d] = %lf\n",
             i, ftbl.gm[i], j, ftbl.gms2[j], i,j, ftbl.f[i][j]);
    }
  }
}


ATOMF_BK read_atom_dat_table(char *element)
{
  /*
   * Atomic scattering factor
   */
  char buf[BUFLEN];
  double t[14], tt[60];
  int i;
  ATOMF_BK atmf;
  FILE *fp;

  
  char mbfit_atom_table[256];
  
  sprintf(mbfit_atom_table,"%s/mbfit/mbfitbk.dat",PREFIX);
  fp = fopen(mbfit_atom_table,"r");
  
/*  fp = fopen_chk_rmode(ATOM_TABLE); */

  comment_skip_fgets(sizeof buf, buf, fp);
  if( strncmp(buf,"DTS",3) == 0 ){
    comment_skip_fgets(sizeof buf, buf, fp);  str_to_fs(9, buf, t);
    for(i=0; i<9; i++){ atmf.dts[i] = t[i]; }
    comment_skip_fgets(sizeof buf, buf, fp);  str_to_fs(9, buf, t);
    for(i=0; i<9; i++){ atmf.dts[i+9] = t[i]; }
    comment_skip_fgets(sizeof buf, buf, fp);  str_to_fs(9, buf, t);
    for(i=0; i<9; i++){ atmf.dts[i+18] = t[i]; }
  }
  comment_skip_fgets(sizeof buf, buf, fp);
  if( strncmp(buf,"GM",2) == 0 ){
    comment_skip_fgets(sizeof buf, buf, fp);  str_to_fs(6, buf, t);
    for(i=0; i<6; i++){ atmf.gm[i] = t[i]; }
  }
  comment_skip_fgets(sizeof buf, buf, fp);
  if( strncmp(buf,"GMS2",4) == 0 ){
    comment_skip_fgets(sizeof buf, buf, fp);  str_to_fs(5, buf, t);
    for(i=0; i<5; i++){ atmf.gms2[i] = t[i]; }
    comment_skip_fgets(sizeof buf, buf, fp);  str_to_fs(5, buf, t);
    for(i=0; i<5; i++){ atmf.gms2[i+5] = t[i]; }
  }

  while( comment_skip_fgets(sizeof buf, buf, fp) != NULL ){/* element */
    if( strcmp(rm_cr(buf), rm_cr(element)) == 0 ){/* Found */
      comment_skip_fgets(sizeof buf, buf, fp);
      str_to_fs(2, buf, t);	atmf.mlt_fct = t[0];	atmf.add_fct = t[1];
      comment_skip_fgets(sizeof buf, buf, fp);	str_to_fs(9, buf, t);
      for(i=0; i<9; i++){ atmf.dt[i] = t[i]; }
      comment_skip_fgets(sizeof buf, buf, fp);	str_to_fs(9, buf, t);
      for(i=0; i<9; i++){ atmf.dt[i+9] = t[i]; }
      comment_skip_fgets(sizeof buf, buf, fp);	str_to_fs(9, buf, t);
      for(i=0; i<9; i++){ atmf.dt[i+18] = t[i]; }

      comment_skip_fgets(sizeof buf, buf, fp); str_to_fs(13, buf, t);
      for(i=0;  i<13; i++){tt[i] = t[i];}
      comment_skip_fgets(sizeof buf, buf, fp); str_to_fs(13, buf, t);
      for(i=0;  i<13; i++){tt[i+13] = t[i];}
      comment_skip_fgets(sizeof buf, buf, fp); str_to_fs(13, buf, t);
      for(i=0;  i<13; i++){tt[i+26] = t[i];}
      comment_skip_fgets(sizeof buf, buf, fp); str_to_fs(13, buf, t);
      for(i=0;  i<13; i++){tt[i+39] = t[i];}
      comment_skip_fgets(sizeof buf, buf, fp); str_to_fs(13, buf, t);
      for(i=0;  i<8; i++){tt[i+52] = t[i];}
      for(i=0; i<60; i++){ atmf.f[i % NUM_GM][i / NUM_GM] = tt[i];}
      /* for(i=0; i<60; i++){ atmf.f[i / NUM_GMS2][i % NUM_GMS2] = tt[i];} */
      fclose(fp);
      return  atmf;
    }
    else{  /* skip */
      for(i=0; i<9; i++){
        if ( comment_skip_fgets(sizeof buf, buf, fp) == NULL ){
          break;
        }
      }
    }
  }
  printf("$$ ERROR: Element %s was not found in %s\n", element, ATOM_TABLE);
  fclose(fp);
  exit(EXIT_FAILURE);
  return  atmf;  /* dummy */
}
