/*
 *	cmatutil.c -- complex matrix
 */

#include <stdio.h>
#include <stdlib.h>
#include "complex.h"
#include "vecmat.h"
#include "cmatutil.h"


complex c_innerproduct(int n, complex *u, complex *v)
{
  int i;
  complex s;

  s = c_conv(0.0,0.0); 
  for (i = 0; i < n; i++) {
    s = c_add( s, c_mul( u[i], v[i]) ) ;
  }
  return s;
}


void c_vecprint(int n, int perline, char *format, complex *v)
{
  int j, k;

  k = 0;
  for (j = 0; j < n; j++) {
    printf(format, v[j].re, v[j].im);
    if (++k >= perline) {
      k = 0;  printf("\n");
    }
  }
  if (k != 0) printf("\n");
}



void c_matcopy2(int n, complex **a, complex **b)
/*
 * copy matrix a to b (square matrix ) 
 * result ==> b
 */
{
	int i,j;

	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			b[i][j] = a[i][j];
		}
	 }
}


void c_matmult(int n, int m, int l, complex **c, complex **a, complex **b)
/*
 * nxl, lxm,  --> nxm C_MATRIX
 * result ==> c
 */
{
	int i, j, k;
	double re, im;
	complex x, y;
/*	complex s;
	for (i = 0; i < n; i++){
		for (j = 0; j < m; j++) {
			s = c_conv(0.0, 0.0);
			for (k = 0; k < l; k++){
				s = c_add( s, c_mul(a[i][k] , b[k][j]) );
			}
			c[i][j] = s;
		}
	}
*/
	for (i = 0; i < n; i++){
		for (j = 0; j < m; j++) {
			re = 0.0;	im = 0.0;
			for (k = 0; k < l; k++){
				x = a[i][k];	y = b[k][j];
				re += x.re * y.re - x.im * y.im;
				im += x.im * y.re + x.re * y.im;
			}
			c[i][j].re = re;	c[i][j].im = im;
		}
	}

}


void c_matmult2(int n, complex **a, complex **b)
/*
 * multiply
 * for square matrix
 * result  ==> a
 */
{
	int i, j, k;
	double re, im;
	complex x, y;
	/* complex s; */
	complex *tmp;  /* a temporary argument */

	tmp = new_vector(complex, n);	chk_vector(tmp);
/*
	for (i = 0; i < n; i++) {
		for (k = 0; k < n; k++) { tmp[k] = a[i][k]; }
		for (j = 0; j < n; j++) {
			s = c_conv(0.0,0.0);
			for (k = 0; k < n; k++) {
				s = c_add(s, c_mul( tmp[k], b[k][j] ));
			}
			a[i][j] = s;
		}
	}
*/
	for (i = 0; i < n; i++) {
		for (k = 0; k < n; k++) { tmp[k] = a[i][k]; }
		for (j = 0; j < n; j++) {
			re = 0.0;	im = 0.0;
			for (k = 0; k < n; k++) {
				x = tmp[k];	y = b[k][j];
				re += x.re * y.re - x.im * y.im;
				im += x.im * y.re + x.re * y.im;
			}
			a[i][j].re = re;	a[i][j].im = im;
		}
	}

	free_vector(tmp);
}


void c_matadd2(int n, complex **a, complex **b)
/*
 * add
 * for square matrix
 * result ==> a
 */
{
	int i, j;

	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			a[i][j] = c_add(a[i][j], b[i][j] );
		}
	}
}


void c_matsub2(int n, complex **a, complex **b)
/*
 * subtract
 * for square matrix
 * result ==> a
 */
{
	int i, j;

	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			a[i][j] = c_sub(a[i][j], b[i][j] );
		}
	}
}


void c_matunit(int n, complex **a)
/*
 * unit matrix
 * for square matrix
 * result ==> a
 */
{
	int i, j;

	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			a[i][j] = c_conv(1.0, 0.0);
		}
	}
}
