/*
 * fitting.h	(for fitting.c)
 */
#include "complex.h"
#include "cmatutil.h"
#include "lattice.h"
#include "strct_f.h"
#include "rfsearch.h"
#include "utils.h"
#include "vecmat.h"

#define  NAME_LEN 32			/* length of _name */
#define  CSTR_LEN 256			/* length of _cstr */
#define  STRSIZE 256

typedef struct{
  int		flg;	/* flag: 0:fix, 1:restraint, 2:fitting, -1:constraint */
  double	dlt;	/* delta: step width to evaluate differential of the parameter */
  double	esd;	/* estimated standard deviation */
  int		ntp;	/* nth parameter */
  char	name[NAME_LEN];		/* name of the parameter */
  char	cstr[CSTR_LEN];		/* constraint */
} FIT_CTRL;

typedef struct{
  FIT_CTRL	a, b, c, alpha, beta, gamma;
} LATPAR_FIT;

typedef struct{
  FIT_CTRL	x,y,z,b,b11,b22,b33,b12,b23,b13,occ,fabs,db;
} ATOM_FIT;

typedef struct{
  FIT_CTRL	re, im, ela_re, ela_im,	*fe, *fi;
  /*
   * lof_type=1 ==> re, im: Fg
   * lof_type=2 ==> ela_re, ela_im: elastic Fg
   * lof_type=3 ==> elastic f and inela. f: *fe, *fi (i=0 to n_atom-1)
   * lof_type=4 ==> elastic f: *fe (i=0 to n_atom-1)
   */
} LO_FG_FIT;


typedef struct{	/* Crystal structure data */
  /* ----- space group and lattice parameter ----- */
  int		space_group;
  int		num_sym_op;
  SYM_OP	*sym_op;
  char	lattice_type[2];
  LAT_PAR lp;			LATPAR_FIT lp_f;
  /* ----- atoms ----- */
  int		n_atom;
  ATOM	*atm;		ATOM_FIT *atm_f;
  XDT		*xdyl;
  ATOMF_BK *ftbl;
  /* ----- Low order Fg ----- */
  int		num_lofg;
  int		lof_type;
  HKL		*g_lofg;
  LO_FG	*fg_lofg;	LO_FG_FIT *fg_lofg_f;
} CRYSTAL_STRCT_DATA;	/* <<<<< */


typedef struct{	/* a set of experimental data for FIT */
  UVW		incidence, surf_norm;
  R_HKL	zolzbase1, zolzbase2;
  double H_tilt; FIT_CTRL H_tilt_f;
  double G_tilt; FIT_CTRL G_tilt_f;
  double  accel_kvol;      FIT_CTRL accel_kvol_f;

  int  num_exact, num_bethe;
  HKL  *reflex;
  BEAM_SEARCH_PAR  beam_search;

  double  *scale_factor;   FIT_CTRL *scale_factor_f; /* (*) ==> it */

  int     num_thickness; 
  double  *thickness;   FIT_CTRL *thickness_f;  /* (*) ==> it */
  double  *thickwidth;  FIT_CTRL *thickwidth_f; /* (*) ==> it */

  int		*num_avrpix;
  double	***avrpix_wt, *gaussian_par;

  int  calc_type, calc_range_flag;
  int  calc_numh, calc_numg;
  double  calc_hbgn, calc_hend, calc_gbgn, calc_gend;

  char	intensity_file[STRSIZE];
  int		num_rg;
  /* HKL  *obs_rg;*/ /*  (*)  ==>  ig */
  double  *g, *art_weight;  /*  (*)  ==>  ig */
  /* int  *num_data_points; */  /*  (*)  ==>  ig */
  double  **H_shift;  FIT_CTRL **H_shift_f;  /* (**) ==> ig, it */
  double  **G_shift;  FIT_CTRL **G_shift_f;  /* (**) ==> ig, it */
  
  /* Caution:
   It is assumed that H_shift[ig][it], G_shift[ig][it],
     obs_H[ig][it] and obs_G[ig][it] are common for all 'it',
     if 'ig' is the same.
   */
} FIT_EXP_DATASET;


typedef struct{	/* a set of experimental data for FIT */
  char intensity_file[STRSIZE];
  int  num_rg;
  HKL  *obs_rg;           /*  (*)  ==>  ig */
  int  num_thickness; 
  int  *num_data_points;  /*  (*)  ==>  ig */
  double  ***obs_H, ***obs_G;
  double  ***obs_intensity, ***obs_sig, ***cal_intensity;
          /* (***) ==> ig, it, ik */
} FIT_EXP_INTSET;

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

#include "mbfunc.h"
#include "mb_debug.h"
#include "fitfunc.h"

void mb_fitting( FILE *fi, FILE *fpo, char *infile );
/*
void alloc_csd(int num_sym_op, int n_atom, int num_low_order_g, CRYSTAL_STRCT_DATA *csd_new);
void alloc_csd_sub(
	int n_atom, int num_low_order_g, int lof_type, CRYSTAL_STRCT_DATA *csd_new
);
void free_csd(CRYSTAL_STRCT_DATA *csd);
 */
void simple_beam_search_wg(
  int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  double H_tilt, double G_tilt, double accel_kvol, CRYSTAL_STRCT_DATA *csd
  );
void simple_beam_search_sg(
  int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  double H_tilt, double G_tilt, double accel_kvol, CRYSTAL_STRCT_DATA *csd
  );
void beam_search_sg_multi(
  int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  double H_tilt, double G_tilt, double accel_kvol, CRYSTAL_STRCT_DATA *csd
  );

void beam_search_sphere(
  int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  double H_tilt, double G_tilt, double accel_kvol, CRYSTAL_STRCT_DATA *csd
  );


void auto_ed_beams(FIT_EXP_DATASET *ed, CRYSTAL_STRCT_DATA *csd);
void nksize_evaluation(
  FIT_EXP_DATASET *ed, double *G_bgn, double *G_end,
  double *H_bgn, double *H_end,
  int *num_points_h
  );
/*void free_fed_vector(int num_exp_dataset, FIT_EXP_DATASET *fed);*/
void input_fit_exp_dataset(FIT_EXP_DATASET *ed, FIT_EXP_INTSET *ei, FILE *fi);

#include "cnstfunc.h"
