/*
 * lattice.h  ( for lattice.c )
 */

typedef struct {
	double a,b,c,alpha,beta,gamma;
} LAT_PAR;
typedef struct {
	double s11, s22, s33, s12, s13, s23, vc;
	double aas, bbs, ccs, abs, bcs, acs;		/* metric tensor (reciprocal space) */
	double aa, bb, cc, ab, bc, ac;				/* metric tensor (real space) */
} CELL_PAR;
typedef struct {int h,k,l;} HKL;		/* reciprocal lattice vector */
typedef struct {double h,k,l;} R_HKL;	/* reciprocal lattice vector */
typedef struct {int u,v,w;} UVW;		/* real lattice vector */
typedef struct {double u,v,w;} R_UVW;	/* real lattice vector */

CELL_PAR bravai(LAT_PAR lp);

double inner_product(R_HKL g1, R_HKL g2, CELL_PAR bv);
double sq_abs(R_HKL g, CELL_PAR bv);

R_HKL hkl2rhkl(HKL g);
R_HKL zolz_ort_base(UVW u, R_HKL g1, CELL_PAR bv);
void zolz_ort_base_two_uvw(UVW u, R_HKL g1, CELL_PAR bv, R_UVW *u1, R_UVW *u2);
R_HKL proj_to_zolz(
	HKL g, UVW u, R_HKL zolzbase1, CELL_PAR bv,	double *coef_x, double *coef_y
);
R_HKL proj_to_zolz2(
	HKL g, UVW u, R_HKL zolzbase1, R_HKL zolzbase2, CELL_PAR bv,
	double *coef_x, double *coef_y
);
R_HKL normalize_vec(R_HKL g, R_HKL h, CELL_PAR bv);

int extinction_rule(HKL g, char *lattice_type);

double inner_product_uvw_hkl(UVW u, R_HKL g);
double inner_product_uvw(UVW u1, UVW u2, CELL_PAR bv);
double inner_product_r_uvw(R_UVW u1, R_UVW u2, CELL_PAR bv);
double abs_uvw(UVW u, CELL_PAR bv);
double abs_r_uvw(R_UVW u, CELL_PAR bv);

UVW incident_direction(HKL h1, HKL h2);
/* double chi(double accel_kvol); */
double wavelength(double accel_kvol);
double relativ_corr(double accel_kvol);
double evdash(double accel_kvol);
