/*
 * mbu_output.c
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "mbconfig.h"
#include "mbu_conf.h"
#include "fitting.h"
#include "mb_input.h"
#include "mbu_output.h"
#include "vecmat.h"
extern EPS_CTRL	epsilon;

/* #define EPS 1e-7 */
#define SKIP_FACTOR	0.0	/* Distance between two adjacent disks */
#define BUFSIZE 256


void chk_array( int i, int max )
{
  if( i < 0 || i >= max ){
    printf("$$ Error: array range over. Abort.\n");
    exit(EXIT_FAILURE);
  }
}


void setmat_obs_cal_dif(
  double **mat, double Hbgn, double Hend, double Gbgn, double Gend,
  int np, int nx, int ny, int nskip,
  double H, double G, double obs, double cal
  )
{
  int  x_obs, y_obs, x_cal, y_cal, x_dif, y_dif;
  
  x_obs = (int) ( (double)(np-1) * (H-Hbgn) / (Hend-Hbgn) + 0.5 );
  y_obs = (ny - 1) - (int) ( (double)(np-1) * (G-Gbgn) / (Gend-Gbgn) + 0.5 );
  x_cal = x_obs + nskip + np;
  y_cal = y_obs;
  x_dif = x_obs + (nskip + np) * 2;
  y_dif = y_obs;
/*  printf("$$ x_obs, y_obs = %d %d\n", x_obs, y_obs);
  printf("$$ x_cal, y_cal = %d %d\n", x_cal, y_cal);
  printf("$$ x_dif, y_dif = %d %d\n", x_dif, y_dif);*/
  
  chk_array(x_obs, nx);  chk_array(y_obs, ny);
  chk_array(x_cal, nx);  chk_array(y_cal, ny);
  chk_array(x_dif, nx);  chk_array(y_dif, ny);
  
  mat[x_obs][y_obs] = obs;
  mat[x_cal][y_cal] = cal;
  mat[x_dif][y_dif] = fabs(obs - cal);
}


void output_pgm_image(
  int image_width, int image_height, double **double_image,
  double maxval, double minval, double gamma_coef, int mag_coef,
  int contrast_reverse_flag, char *fn_out
  )
{
  /* Output an 8-bit gray-scale pgm image */
  long int  ct;
  int  px, py, px_end, py_end;
  /*  int  cut_image_width, cut_image_height, offset_x, offset_y; */
  unsigned char *proc_image8;
  unsigned long  t_size;
  double  tmp, dist, intensity;
  FILE   *fp_;

  if(gamma_coef > 1.0 || gamma_coef <= 0.0){
    printf("$$ gamma_coef(=%lf) should be within [0,1].  Abort.\n", gamma_coef);
    exit(EXIT_FAILURE);
  }
  if(mag_coef < 1){
    printf("$$ mag_coef(=%d) should be an integer number\n", mag_coef);
    printf("$$  and greater than 0.  Abort.\n", mag_coef);
    exit(EXIT_FAILURE);
  }
  dist = maxval - minval;
  if(dist <= 0.0){
    printf("$$ max - min (=%lf) should be greater than 0.0.  Abort.\n", dist);
    exit(EXIT_FAILURE);
  }
  proc_image8
    = new_vector(
      unsigned char, image_width * image_height * sq(mag_coef)
      );
  chk_vector(proc_image8);

  px_end = image_width * mag_coef;
  py_end = image_height * mag_coef;
  ct = 0;
/*  for(py=0; py<py_end; py++){ */
  for(py=py_end-1; py>=0; py--){
    for(px=0; px<px_end; px++){
      /* caution: fabs */
      intensity = fabs( double_image[px/mag_coef][py/mag_coef] - minval );
      /* gamma curve */
      if(gamma_coef == 1.0){
        tmp = 255.0 * intensity / dist;
      }
      else{
        tmp = 255.0 * pow( intensity / dist, gamma_coef );
      }
      if(contrast_reverse_flag == 0){
        if(tmp < 0){
          proc_image8[ct++] = 0;
        }
        else if (tmp > 255){
          proc_image8[ct++] = 255;
        }
        else{
          proc_image8[ct++] = (unsigned char) tmp;
        }
      }
      else if(contrast_reverse_flag == 1){
        if(tmp < 0){
          proc_image8[ct++] = 255;
        }
        else if (tmp > 255){
          proc_image8[ct++] = 0;
        }
        else{
          proc_image8[ct++] = (unsigned char) 255 - tmp;
        }
      }
    }
  }

  /* Check */
  if (ct != image_width * image_height * sq(mag_coef)){
    printf(
      "$$ ct = %ld is not equal to %ld. Abort.\n",
      ct, image_width * image_height
      );
    exit(EXIT_FAILURE);
  }

  if( (fp_ = fopen(fn_out, "wb")) == NULL ){
    printf("$$ Cannot open %s\n", fn_out);
    exit(EXIT_FAILURE);
  }
  fprintf(fp_, "P5\n");
  fprintf(fp_, "#%s\n", fn_out);
  fprintf(fp_, "%d %d\n", image_width * mag_coef, image_height * mag_coef);
  fprintf(fp_, "255\n"); /* full-scale */
  t_size = fwrite(
    proc_image8, sizeof(unsigned char),
    image_width * image_height * sq(mag_coef), fp_
    );
  fclose(fp_);
  free_vector(proc_image8);
}


void mb_fit_result_visualize(
  char *resfil, char *infofil, int mag_coef, int cont_rev_flag,
  double preset_max_val, double gamma_max
  )
{
  int  i, j, n_points, num_data, np, nx, ny, nskip, nth_file;
  double  ycal, yobs, x[6], Hbgn, Hend, Gbgn, Gend, G, H;
  double  **mat, min, max, min_val, max_val, gamma_coef=1.0;

  char  buf[BUFSIZE], hklbuf[BUFSIZE];

  char  infil[BUFSIZE], pgm_file[BUFSIZE], htmlfile[BUFSIZE];
  /* char  batfile[BUFSIZE], shfile[BUFSIZE]; */
  FILE  *fi_res, *fi_info, *fo_html; /* *fo_bat, *fo_sh; */
  
  printf("\n***** Visualize FIT_RESULT *****\n\n");
  fi_res = fopen_chk_rmode(resfil);
  fi_info = fopen_chk_rmode(infofil);

  strcpy(infil, remove_extention(resfil));
  /* html output */
  sprintf(htmlfile, "%s.html", infil);
  printf("Result: html ==> %s\n", htmlfile);
  fo_html = fopen_chk_wmode(htmlfile);

  /* 
  sprintf(batfile, "%s.cmd", infil);
  printf("Script: cmd (win) ==> %s\n", batfile);
  fo_bat = fopen_chk_wmode(batfile);
  sprintf(shfile, "%s.sh", infil);
  printf("Script: sh (win) ==> %s\n", shfile);
  fo_sh = fopen_chk_wmode(shfile);
  fprintf(fo_sh, "#!/bin/sh\n");
    */

  /* info file */
  np = fget_i(fi_info);
  if( np <= 0 ){
    printf("$$ np must be greater than 0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  fget_4d(&Hbgn, &Hend, &Gbgn, &Gend, fi_info);
  chk_data("INFO_END", fi_info);
  fclose(fi_info);
  
  /* html */
  fprintf(fo_html, "<html>\n");
  fprintf(fo_html, "<title>%s</title>\n", infil);
  fprintf(fo_html, "<body>\n");
  fprintf(fo_html, "<p>Data file: %s<br>\n", infil);
  fprintf(fo_html, "Number of horizontal pixels = %d<br>\n", np);
  fprintf(
    fo_html,
    "H_begin, H_end, G_begin, G_end = %lf, %lf, %lf, %lf<br>\n",
    Hbgn, Hend, Gbgn, Gend
    );
  fprintf(fo_html, "Magnification = %d, ", mag_coef);
  fprintf(fo_html, "gamma_max = %lf", gamma_max );
  if(preset_max_val > 0.0){
    fprintf(fo_html, ", common_scale_max = %lf", preset_max_val );
  }
  fprintf(fo_html, "</p>\n");
  fprintf(fo_html, "<HR>\n");
  fprintf(fo_html, "obs, <spacer size=\"30\" type=\"horizontal\">\n");
  fprintf(fo_html, "cal, <spacer size=\"30\" type=\"horizontal\">\n");
  fprintf(fo_html, "dif, <spacer size=\"30\" type=\"horizontal\">\n");
  fprintf(fo_html, " hkl, [min_intensity, max_intensity]<p>\n");
  
  nskip = (int) (np * SKIP_FACTOR);
  nx = np * 3 + nskip * 2;
  ny = np;
  /*printf("$$nx, ny = %d %d.\n", nx, ny);*/
  mat = new_matrix(double, nx, ny);  chk_matrix(mat);
  num_data = 0;
  nth_file = 0;
  while( comment_skip_fgets(sizeof buf, buf, fi_res) != NULL ){
    nth_file++;
    strcpy(hklbuf, buf);
    num_data = fget_i(fi_res);
    /*printf("$$num_data = %d\n", num_data);*/
    for(i=0; i<nx; i++){
      for(j=0; j<ny; j++){
        mat[i][j] = 0.0;
      }
    }
    for(i=0; i<num_data; i++){
      fget_4d( &H, &G, &yobs, &ycal, fi_res );
      /*printf("$$ %lf %lf %lf %lf\n", H, G, yobs, ycal);*/
      setmat_obs_cal_dif(
        mat, Hbgn, Hend, Gbgn, Gend, np, nx, ny, nskip, H, G, yobs, ycal
        );
    }
    findmaxmin2(nx, ny, mat, &max, &min);
    /*printf("$$ max, min = %lf %lf\n",  max, min);*/
    /*
    fprintf(fo_bat, "pnmtopng %d.pgm > %d.png\n", nth_file, nth_file);
    fprintf(fo_sh,  "pnmtopng %d.pgm > %d.png\n", nth_file, nth_file);
      */
    fprintf(fo_html, "<img src=\"%d.png\" align=\"middle\">\n", nth_file);
    fprintf(fo_html, "%s", rm_cr(hklbuf));	/* hkl */
    fprintf(fo_html, "\t[%.0lf, %.0lf]\n", min, max);

    if(preset_max_val > 0.0){
      max_val = preset_max_val * pow( max / preset_max_val, 1.0-gamma_max );
      printf("$$ max_val = %lf\n", max_val);
      /* <<<<< 030814 */
      min_val = 0.0;  /*min_val = min;*/
    }
    else{
      max_val = max;
      min_val = 0.0;  /*min_val = min;*/
    }
    sprintf(pgm_file, "%d.pgm", nth_file);
    printf("$$ Output pgmfile ==> %s\n", pgm_file);
    output_pgm_image(
      nx, ny, mat, max_val, min_val,
      gamma_coef, mag_coef, cont_rev_flag, pgm_file
      );
  }
  fprintf(fo_html, "</body>\n</html>\n");
  fclose(fi_res);
  fclose(fo_html);  /* fclose(fo_bat);  fclose(fo_sh); */
  free_matrix(mat);
}


void sim_visual(
  char *infil, double gamma_coef, int mag_coef, int cont_rev_flag,
  double preset_max_val, FILE *fo_html
  )
{
  /*
    Visualize SIM_RESULT
   */
  int  ik, ig, nk, ng, ix, iy, nx, ny;
  int  center_x, center_y;
  double  **mat;
  double  t[7], thickness, *hpos, *gpos, max_H, min_H, max_G, min_G;
  double  *Hval, *Gval, val;
  double  max_val, min_val;
  int  min_hpix, max_hpix, min_gpix, max_gpix, *hpix, *gpix;
  int  posx_g, posy_g, max_x, max_y, min_x, min_y;
  UVW  incidence;
  R_HKL  zolzbase1;
  HKL  g, *cal_g;
  char  buf[BUFSIZE], filenm[BUFSIZE];
  char  pgm_file[BUFSIZE], png_file[BUFSIZE];
  CELL_PAR  bv;
  FILE *fp;

  fp = fopen_chk_rmode(infil);
  printf("$$ reading %s\n", infil);

  chk_data("SIM_RESULT", fp);
  chk_data("VISUAL{", fp);
  fget_2d(&max_val, &min_val, fp);
  if(preset_max_val > 0.0){
    max_val = preset_max_val;
  }
  fget_d(fp);  /* gamma_coef dummy */
  fget_i(fp);  /* mag_coef dummy */
  fget_i(fp);  /* cont_rev_flag  dummy */
  chk_data("}VISUAL", fp);
  
  chk_data("CR_PARAMETERS{", fp);
  comment_skip_fgets(sizeof buf, buf, fp);
  str_to_fs(7, buf, t);
  bv.s11 = t[0]; bv.s22 = t[1]; bv.s33 = t[2];
  bv.s12 = t[3]; bv.s13 = t[4]; bv.s23 = t[5]; bv.vc = t[6];
  comment_skip_fgets(sizeof buf, buf, fp); 	str_to_fs(6, buf, t);
  bv.aas = t[0]; bv.bbs = t[1]; bv.ccs = t[2];
  bv.abs = t[3]; bv.bcs = t[4]; bv.acs = t[5];
  comment_skip_fgets(sizeof buf, buf, fp); 	str_to_fs(6, buf, t);
  bv.aa = t[0]; bv.bb = t[1]; bv.cc = t[2];
  bv.ab = t[3]; bv.bc = t[4]; bv.ac = t[5];
  
  fget_3i( &(incidence.u), &(incidence.v), &(incidence.w), fp );
  fget_3d( &(zolzbase1.h), &(zolzbase1.k), &(zolzbase1.l), fp );
  thickness = fget_d(fp);
  chk_data("}CR_PARAMETERS", fp);
  
  chk_data("RF_INDEX{", fp);
  ng = fget_i(fp);
  cal_g = new_vector(HKL, ng);  chk_vector(cal_g);
  hpos = new_vector(double, ng);  chk_vector(hpos);
  gpos = new_vector(double, ng);  chk_vector(gpos);
  for(ig = 0; ig < ng; ig++){
    comment_skip_fgets(sizeof buf, buf, fp);
    str_to_fs(5, buf, t);
    cal_g[ig].h = (int) t[0];
    cal_g[ig].k = (int) t[1];
    cal_g[ig].l = (int) t[2];
    hpos[ig] = t[3];  gpos[ig] = t[4];
  }
  chk_data("}RF_INDEX", fp);
  
  chk_data("COORD{", fp);
  nk = fget_i(fp);  /* number of pixels */
  Hval = new_vector(double, nk);	chk_vector(Hval);
  Gval = new_vector(double, nk);	chk_vector(Gval);
  hpix = new_vector(int, nk);		chk_vector(hpix);
  gpix = new_vector(int, nk);		chk_vector(gpix);
  fget_2d(&max_H, &min_H, fp);
  fget_2d(&max_G, &min_G, fp);
  fget_2i(&max_hpix, &min_hpix, fp);
  fget_2i(&max_gpix, &min_gpix, fp);
  for(ik=0; ik<nk; ik++){
    comment_skip_fgets(sizeof buf, buf, fp);
    str_to_fs(4, buf, t);
    Hval[ik] = t[0];  Gval[ik] = t[1];
    hpix[ik] = (int)t[2];  gpix[ik] = (int)t[3];
  }
  chk_data("}COORD", fp);

    /* html output*/
  fprintf(fo_html, "<hr>\n");
  fprintf(
    fo_html, "<p>[%d %d %d], t = %lg [A], gamma_coef = %lg</p>\n",
    incidence.u, incidence.v, incidence.w, thickness, gamma_coef
    );

  /* Evaluation of the image size */
  max_x = min_x = max_y = min_y = 0;
  for(ig = 0; ig < ng; ig++){
    if(max_x < (int) hpos[ig] + max_hpix){ max_x = (int) hpos[ig] + max_hpix; }
    if(min_x > (int) hpos[ig] + min_hpix){ min_x = (int) hpos[ig] + min_hpix; }
    if(max_y < (int) gpos[ig] + max_gpix){ max_y = (int) gpos[ig] + max_gpix; }
    if(min_y > (int) gpos[ig] + min_gpix){ min_y = (int) gpos[ig] + min_gpix; }
  }
  nx = max_x - min_x;
  ny = max_y - min_y;
  if( nx <= 0 ){
    printf("$$ (mb_visualization) Error: nx <= 0; Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( ny <= 0 ){
    printf("$$ (mb_visualization) Error: ny <= 0; Abort.\n");
    exit(EXIT_FAILURE);
  }
  nx = ( (int) (nx / 4) + 2) * 4;	/* <<<<< */
  ny = ( (int) (ny / 4) + 2) * 4;	/* <<<<< */
  center_x = - min_x + 4;
  center_y = - min_y + 4;
  /*
  printf(
    "$$ nx, ny = %d %d, center_x, center_y = %d %d\n",
    nx, ny, center_x, center_y
    );
   */
  mat = new_matrix(double, nx, ny);  chk_matrix(mat);
  /* initialize */
  for(ix = 0; ix < nx; ix++){
    for(iy = 0; iy < ny; iy++){
      mat[ix][iy] = 0.0;
    }
  }
  
  for(ig=0; ig<ng; ig++){
    chk_data("RF{", fp);
    fget_3i(&(g.h), &(g.k), &(g.l), fp);
    if( g.h != cal_g[ig].h || g.k != cal_g[ig].k || g.l != cal_g[ig].l ){
      printf("$$ (mb_visualization) mismatch in hkl. Abort.\n");
      exit(EXIT_FAILURE);
    }
    posx_g = (int) hpos[ig];
    posy_g = (int) gpos[ig];
    for(ik=0; ik<nk; ik++){
      val = fget_d(fp);
      ix = center_x + posx_g + hpix[ik];
      iy = center_y + posy_g + gpix[ik];
      if( ix >= 0 && ix < nx && iy >= 0 && iy < ny ){
        mat[ix][iy] += val;
      }
      else{
        printf("$$ (mb_visualization) (%d,%d) WARNING: out of area.\n", ix, iy);
      }
    }
    chk_data("}RF", fp);
  }
  strcpy(filenm, remove_extention(infil));

  sprintf(pgm_file, "%s.pgm", filenm);
  sprintf(png_file, "%s.png", filenm);
  min_val = 0.0; /* <<<<< 030806 */
  /* findmaxmin2(....) */
  output_pgm_image(
    nx, ny, mat, max_val, min_val,
    gamma_coef, mag_coef, cont_rev_flag, pgm_file
    );
#if defined(LINUX_MODE) /* for Linux */
  sprintf(buf, "./pnmtopng %s > %s", pgm_file, png_file);
  system(buf);
  sprintf(buf, "rm %s", pgm_file);
  system(buf);
#else   /* win mode */
  sprintf(buf, "pnmtopng %s > %s", pgm_file, png_file);
  system(buf);
  sprintf(buf, "del %s", pgm_file);
  system(buf);
#endif

  /* html output*/
  fprintf(fo_html, "<p><img src=\"%s\" align=\"middle\"></p>\n", png_file);

  free_vector(Hval);  free_vector(Gval);
  free_vector(hpix);  free_vector(gpix);
  free_vector(cal_g);
  free_vector(hpos);  free_vector(gpos);
  free_matrix(mat);
  fclose(fp);
}


void mb_sim_result_visualize(
  char *infil, double gamma_coef, int mag_coef, int cont_rev_flag,
  double preset_max_val, FILE *fo_html
  )
{
  if( gamma_coef < 0.0 ){
    printf("$$ (mb_visualization) Error: gamma < 0.0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( mag_coef <= 0 ){
    printf("$$ (mb_visualization) Error: mag_coef <= 0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  sim_visual(
    infil, gamma_coef, mag_coef, cont_rev_flag, preset_max_val, fo_html
    );
}


void sim_reflex_visual(
  char *infil, double gamma_coef, int mag_coef, int cont_rev_flag,
  double preset_max_val, FILE *fo_html
  )
{
  /*
    Visualize SIM_RESULT
   */
  int  ik, ig, nk, ng, ix, iy, nx, ny;
  int  center_x, center_y;
  double  **mat;
  double  t[7], thickness, *hpos, *gpos, max_H, min_H, max_G, min_G;
  double  *Hval, *Gval, val;
  double  max_val, min_val;
  int  min_hpix, max_hpix, min_gpix, max_gpix, *hpix, *gpix;
  int  posx_g, posy_g, max_x, max_y, min_x, min_y;
  UVW  incidence;
  R_HKL  zolzbase1;
  HKL  g, *cal_g;
  char  buf[BUFSIZE], fname[BUFSIZE];
  char  pgm_file[BUFSIZE], png_file[BUFSIZE];
  CELL_PAR  bv;
  FILE *fp;

  fp = fopen_chk_rmode(infil);  /* SIM_RESULT file */
  printf("$$ reading %s\n", infil);

  /* Data input */
  chk_data("SIM_RESULT", fp);
  chk_data("VISUAL{", fp);
  fget_2d(&max_val, &min_val, fp);
  if(preset_max_val > 0.0){
    max_val = preset_max_val;
  }
  fget_d(fp);  /* gamma_coef dummy */
  fget_i(fp);  /* mag_coef dummy */
  fget_i(fp);  /* cont_rev_flag  dummy */
  chk_data("}VISUAL", fp);
  
  chk_data("CR_PARAMETERS{", fp);
  comment_skip_fgets(sizeof buf, buf, fp);
  str_to_fs(7, buf, t);
  bv.s11 = t[0]; bv.s22 = t[1]; bv.s33 = t[2];
  bv.s12 = t[3]; bv.s13 = t[4]; bv.s23 = t[5]; bv.vc = t[6];
  comment_skip_fgets(sizeof buf, buf, fp); 	str_to_fs(6, buf, t);
  bv.aas = t[0]; bv.bbs = t[1]; bv.ccs = t[2];
  bv.abs = t[3]; bv.bcs = t[4]; bv.acs = t[5];
  comment_skip_fgets(sizeof buf, buf, fp); 	str_to_fs(6, buf, t);
  bv.aa = t[0]; bv.bb = t[1]; bv.cc = t[2];
  bv.ab = t[3]; bv.bc = t[4]; bv.ac = t[5];
  
  fget_3i( &(incidence.u), &(incidence.v), &(incidence.w), fp );
  fget_3d( &(zolzbase1.h), &(zolzbase1.k), &(zolzbase1.l), fp );
  thickness = fget_d(fp);
  chk_data("}CR_PARAMETERS", fp);
  
  chk_data("RF_INDEX{", fp);
  ng = fget_i(fp);
  cal_g = new_vector(HKL, ng);  chk_vector(cal_g);
  hpos = new_vector(double, ng);  chk_vector(hpos);
  gpos = new_vector(double, ng);  chk_vector(gpos);
  for(ig = 0; ig < ng; ig++){
    comment_skip_fgets(sizeof buf, buf, fp);
    str_to_fs(5, buf, t);
    cal_g[ig].h = (int) t[0];
    cal_g[ig].k = (int) t[1];
    cal_g[ig].l = (int) t[2];
    hpos[ig] = t[3];
    gpos[ig] = t[4];
  }
  chk_data("}RF_INDEX", fp);
  
  chk_data("COORD{", fp);
  nk = fget_i(fp);  /* number of pixels */
  Hval = new_vector(double, nk);  chk_vector(Hval);
  Gval = new_vector(double, nk);  chk_vector(Gval);
  hpix = new_vector(int, nk);  chk_vector(hpix);
  gpix = new_vector(int, nk);  chk_vector(gpix);
  fget_2d(&max_H, &min_H, fp);
  fget_2d(&max_G, &min_G, fp);
  fget_2i(&max_hpix, &min_hpix, fp);
  fget_2i(&max_gpix, &min_gpix, fp);
  for(ik=0; ik<nk; ik++){
    comment_skip_fgets(sizeof buf, buf, fp);
    str_to_fs(4, buf, t);
    Hval[ik] = t[0];  Gval[ik] = t[1];
    hpix[ik] = (int)t[2];  gpix[ik] = (int)t[3];
  }
  chk_data("}COORD", fp);

    /* html output*/
  fprintf(fo_html, "<hr>\n");
  fprintf(
    fo_html, "<p>[%d %d %d], t = %lg [A], gamma_coef = %lg</p>\n",
    incidence.u, incidence.v, incidence.w, thickness, gamma_coef
    );

  nx = max_hpix - min_hpix + 1;
  ny = max_gpix - min_gpix + 1;
  center_x = - min_hpix;
  center_y = - min_gpix;
  if( nx <= 0 ){
    printf("$$ (mb_visualization) Error: nx <= 0; Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( ny <= 0 ){
    printf("$$ (mb_visualization) Error: ny <= 0; Abort.\n");
    exit(EXIT_FAILURE);
  }

  strcpy(fname, remove_extention(infil));
  mat = new_matrix(double, nx, ny);  chk_matrix(mat);
  for(ig=0; ig<ng; ig++){
    for(ix = 0; ix < nx; ix++){
      for(iy = 0; iy < ny; iy++){
        mat[ix][iy] = 0.0;  /* initialize */
      }
    }
    chk_data("RF{", fp);
    fget_3i(&(g.h), &(g.k), &(g.l), fp);
    if( g.h != cal_g[ig].h || g.k != cal_g[ig].k || g.l != cal_g[ig].l ){
      printf("$$ (mb_visualization) mismatch in hkl. Abort.\n");
      exit(EXIT_FAILURE);
    }
    /*
    posx_g = (int) hpos[ig];
    posy_g = (int) gpos[ig];
      */
    for(ik=0; ik<nk; ik++){
      val = fget_d(fp);
      ix = center_x + hpix[ik];
      iy = center_y + gpix[ik];
      if( ix >= 0 && ix < nx && iy >= 0 && iy < ny ){
        mat[ix][iy] += val;
      }
      else{
        printf("$$ (mb_visualization) (%d,%d) WARNING: out of area.\n", ix, iy);
      }
    }

    min_val = 0.0;  /* <<<<< 030806 */
    /* findmaxmin2(....) */

    sprintf(pgm_file, "%s_%d.pgm", fname, ig);
    sprintf(png_file, "%s_%d.png", fname, ig);
    output_pgm_image(
      nx, ny, mat, max_val, min_val,
      gamma_coef, mag_coef, cont_rev_flag, pgm_file
      );
#if defined(LINUX_MODE) /* for Linux */
    sprintf(buf, "./pnmtopng %s > %s", pgm_file, png_file);
    system(buf);
    sprintf(buf, "rm %s", pgm_file);
    system(buf);
#else   /* win mode */
    sprintf(buf, "pnmtopng %s > %s", pgm_file, png_file);
    system(buf);
    sprintf(buf, "del %s", pgm_file);
    system(buf);
#endif
    chk_data("}RF", fp);

    /* html output*/
    fprintf(fo_html, "<p><img src=\"%s\" align=\"middle\">\n", png_file);
    fprintf(
      fo_html, "hkl= %d %d %d</p>\n", cal_g[ig].h, cal_g[ig].k, cal_g[ig].l
      );
  }
  free_vector(Hval);  free_vector(Gval);
  free_vector(hpix);  free_vector(gpix);
  free_vector(cal_g);
  free_vector(hpos);  free_vector(gpos);
  free_matrix(mat);
  fclose(fp);
}


void mb_sim_result_reflex_visualize(
  char *infil, double gamma_coef, int mag_coef, int cont_rev_flag,
  double preset_max_val, FILE *fo_html
  )
{
  if( gamma_coef < 0.0 ){
    printf("$$ (mb_visualization) Error: gamma < 0.0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( mag_coef <= 0 ){
    printf("$$ (mb_visualization) Error: mag_coef <= 0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  sim_reflex_visual(
    infil, gamma_coef, mag_coef, cont_rev_flag, preset_max_val, fo_html
    );
}


#if 0
void mb_visualization(
  int command_line_visual, char *infil,
  double gamma_coef, int mag_coef, int cont_rev_flag,
  double preset_max_val
  )
{
  /*
   * If preset_max_val > 0.0 then the value of preset_max_val will be used
   *  instead of that of max_val.
   */
  double  max_val, min_val, t[3];
  char  rsl_fil[BUFSIZE], identifier[BUFSIZE], buf[BUFSIZE];
  FILE	*fp;
  
  if (command_line_visual == 1){		/* command line mode */
    sprintf(rsl_fil, "%s.rsl", infil);
    if( (fp = fopen(rsl_fil,"r")) == NULL ){				/* rsl file (input) */
      fprintf(stderr,"Cannot open %s\n", rsl_fil);
      exit(EXIT_FAILURE);
    }
    printf("reading %s\n", rsl_fil);
    comment_skip_fgets(80, buf, fp);
    strcpy(identifier, rm_cr(buf));
    fget_2d(&max_val, &min_val, fp);
    if(preset_max_val > 0.0){
      max_val = preset_max_val;
    }
    fget_d(fp);		/* gamma_coef dummy */
    fget_i(fp);		/* mag_coef dummy */
    fget_i(fp); 	/* cont_rev_flag  dummy */
    
  }
  else if (command_line_visual == 0){				/* console mode */
    printf("Data filename without extension (rsl) (SIM: disp_type=7) \n>>");
    fgets(infil,sizeof(infil),stdin);
    sprintf(rsl_fil, "%s.rsl", infil);
    if( (fp = fopen(rsl_fil,"r")) == NULL ){				/* rsl file (input) */
      fprintf(stderr,"Cannot open %s\n", rsl_fil);
      exit(EXIT_FAILURE);
    }
    
    printf("reading %s\n", rsl_fil);
    comment_skip_fgets(80, buf, fp);
    strcpy(identifier, rm_cr(buf));
    fget_2d(&max_val, &min_val, fp);
    if(preset_max_val > 0.0){
      max_val = preset_max_val;
    }
    gamma_coef = fget_d(fp);
    mag_coef = fget_i(fp);
    cont_rev_flag = fget_i(fp);
    
    printf("max, min ? (std.: %lf, %lf)>>", max_val, min_val);
    fgets(buf,sizeof(buf),stdin);	
    str_to_fs(2, buf, t);	max_val = t[0], min_val = t[1];
    printf("gamma coef. ? (>0) >>");
    fgets(buf,sizeof(buf),stdin);	
    str_to_fs(1, buf, t);	gamma_coef = t[0];
    printf("magnification ? (int) >>");
    fgets(buf,sizeof(buf),stdin);	mag_coef = atoi(buf);
    printf("contrast reverse ? [1:on, 0:off] >>");
    fgets(buf,sizeof(buf),stdin);	cont_rev_flag = atoi(buf);
  }
  
  /* Check */
  printf("max, min = %lf, %lf, ", max_val, min_val);
  printf("gamma_coef = %lf, ", gamma_coef);
  printf("mag_coef = %d, ", mag_coef);
  printf("cont_rev_flag = %d\n", cont_rev_flag);
  if( gamma_coef < 0.0 ){
    printf("$$ (mb_visualization) Error: gamma < 0.0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( mag_coef <= 0 ){
    printf("$$ (mb_visualization) Error: mag_coef <= 0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( cont_rev_flag != 0 && cont_rev_flag != 1){
    printf("$$ (mb_visualization) Error: cont_rev_flag must be 0 or 1. Abort.\n");
    exit(EXIT_FAILURE);
  }
  
  if (strcmp(identifier, "VISUAL") == 0){
    sim_visualization(
      fp, infil, gamma_coef, max_val, min_val, mag_coef, cont_rev_flag
      );
  }
  else if (strcmp(identifier, "VISUAL_COHERENT") == 0){
    sim_visualization_coherent(
      fp, infil, gamma_coef, max_val, min_val, mag_coef, cont_rev_flag
      );
  }
  else {
    printf("$$ Identifier %s is not valid. Abort.\n", identifier);
  }
  fclose(fp);
}


void mb_visualization_diff(
  int command_line_visual, char infil1[], char infil2[],
  double gamma_coef, int mag_coef, int cont_rev_flag,
  double preset_max_val
  )
{
  /*
   * If preset_max_val > 0.0 then the value of preset_max_val will be used
   *  instead of that of max_val.
   */
  double  max_val, min_val, t[3];
  char  rsl_fil1[BUFSIZE], rsl_fil2[BUFSIZE], identifier[BUFSIZE], buf[BUFSIZE];
  FILE  *fp1, *fp2;
  
  if (command_line_visual == 1){		/* command line mode */
    sprintf(rsl_fil1, "%s.rsl", infil1);
    if( (fp1 = fopen(rsl_fil1,"r")) == NULL ){				/* rsl file (input) */
      fprintf(stderr,"Cannot open %s\n", rsl_fil1);
      exit(EXIT_FAILURE);
    }
    sprintf(rsl_fil2, "%s.rsl", infil2);
    if( (fp2 = fopen(rsl_fil2,"r")) == NULL ){  /* rsl file (input) */
      fprintf(stderr,"Cannot open %s\n", rsl_fil2);
      exit(EXIT_FAILURE);
    }
    printf("Reading data from %s and %s\n", rsl_fil1, rsl_fil2);
    comment_skip_fgets(80, buf, fp2); /* dummy */
    comment_skip_fgets(80, buf, fp1);
    strcpy(identifier, rm_cr(buf));
    fget_2d(&max_val, &min_val, fp2); /* dummy */
    fget_2d(&max_val, &min_val, fp1);
    if(preset_max_val > 0.0){
      max_val = preset_max_val;
    }
    fget_d(fp2);		/* gamma_coef dummy */
    fget_i(fp2);		/* mag_coef dummy */
    fget_i(fp2);	 	/* cont_rev_flag  dummy */
    fget_d(fp1);		/* gamma_coef dummy */
    fget_i(fp1);		/* mag_coef dummy */
    fget_i(fp1);	 	/* cont_rev_flag  dummy */
    
  }
  else if (command_line_visual == 0){  /* console mode */
    printf("Data filename 1 without extension (rsl) (SIM: disp_type=7) \n>>");
    fgets(infil1,sizeof(infill),stdin);
    sprintf(rsl_fil1, "%s.rsl", infil1);
    if( (fp1 = fopen(rsl_fil1,"r")) == NULL ){  /* rsl file (input) */
      fprintf(stderr,"Cannot open %s\n", rsl_fil1);
      exit(EXIT_FAILURE);
    }
    printf("Data filename 2 without extension (rsl) (SIM: disp_type=7) \n>>");
    fgets(infil2,sizeof(infil2),stdin);
    sprintf(rsl_fil2, "%s.rsl", infil2);
    if( (fp2 = fopen(rsl_fil2,"r")) == NULL ){  /* rsl file (input) */
      fprintf(stderr,"Cannot open %s\n", rsl_fil2);
      exit(EXIT_FAILURE);
    }
    printf("Reading data from %s and %s\n", rsl_fil1, rsl_fil2);
    comment_skip_fgets(80, buf, fp2);
    comment_skip_fgets(80, buf, fp1);
    strcpy(identifier, rm_cr(buf));
    fget_2d(&max_val, &min_val, fp2);
    fget_2d(&max_val, &min_val, fp1);
    if(preset_max_val > 0.0){
      max_val = preset_max_val;
    }
    gamma_coef = fget_d(fp2);
    mag_coef = fget_i(fp2);
    cont_rev_flag = fget_i(fp2);
    gamma_coef = fget_d(fp1);
    mag_coef = fget_i(fp1);
    cont_rev_flag = fget_i(fp1);
    
    printf("max, min ? (std.: %lf, %lf)>>", max_val, min_val);
    fgets(buf,sizeof(buf),stdin);	
    str_to_fs(2, buf, t);	max_val = t[0], min_val = t[1];
    printf("gamma coef. ? (>0) >>");
    fgets(buf,sizeof(buf),stdin);	
    str_to_fs(1, buf, t);	gamma_coef = t[0];
    printf("magnification ? (int) >>");
    fgets(buf,sizeof(buf),stdin);	
    mag_coef = atoi(buf);
    printf("contrast reverse ? [1:on, 0:off] >>");
    fgets(buf,sizeof(buf),stdin);	
    cont_rev_flag = atoi(buf);
  }
  
  /* Check */
  printf("max, min = %lf, %lf, ", max_val, min_val);
  printf("gamma_coef = %lf, ", gamma_coef);
  printf("mag_coef = %d, ", mag_coef);
  printf("cont_rev_flag = %d\n", cont_rev_flag);
  if( gamma_coef < 0.0 ){
    printf("$$ (mb_visualization) Error: gamma < 0.0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( mag_coef <= 0 ){
    printf("$$ (mb_visualization) Error: mag_coef <= 0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( cont_rev_flag != 0 && cont_rev_flag != 1){
    printf("$$ (mb_visualization) Error: cont_rev_flag must be 0 or 1. Abort.\n");
    exit(EXIT_FAILURE);
  }
  
  if (strcmp(identifier, "VISUAL") == 0){
    sim_visualization_diff(
      fp1, fp2, infil1, infil2, gamma_coef, max_val, min_val,
      mag_coef, cont_rev_flag
      );
  }
  else if (strcmp(identifier, "VISUAL_COHERENT") == 0){
    printf(
      "$$ Sorry, coherent mode has not been supported in VISDIF mode. Bye.\n"
      );
    exit(EXIT_FAILURE);
  }
  else {
    printf(
      "$$ (mb_visualization) Identifier %s is not valid. Abort.\n", identifier
      );
  }
  fclose(fp1);
  fclose(fp2);
}
#endif


#if 0
void ass_out(char *outfil, double **mat, int nx, int ny)
     /* ASCII-special-format output */
{
  int i, j;
  double max, min;
  FILE *fo;
  
  if ( (fo = fopen(outfil,"w")) == NULL ){
    printf("Can't open file: %s\n", outfil);	exit(EXIT_FAILURE);
  }
  
  fprintf(fo, "%d\t%d\n", ny, nx);
  findmaxmin2(nx, ny, mat, &max, &min);
  fprintf(fo, "%lf\t%lf\n", max, min);
  for(i=0; i<ny-1; i++){ fprintf(fo, "%d\t", i); } fprintf(fo, "%d\n", ny-1); 
  for(i=0; i<nx-1; i++){ fprintf(fo, "%d\t", i); } fprintf(fo, "%d\n", nx-1); 
  
  
  for(j=0; j<ny; j++){
    for(i=0; i<nx-1; i++){
      fprintf(fo, "%lf\t", mat[i][j]);
    }
    fprintf(fo, "%lf\n", mat[nx-1][j]);
  }
  
  fclose(fo);
}
#endif


#if 0
void bmp_out(
  char *output_filename, double **val, int nx, int ny, int fmag
  )
{
  /*
   * BMP-format output (Gray 8-bit, inverted)
   *  fmag: Magnification of image (integer)
   *  nx and ny should be multiples of 4.
   */
  long int i, j, ct, num_pix;
  double y, pix_val, max_val, min_val, max_val_rsl, min_val_rsl;
  unsigned short  b2[2]; /* 2 bytes */
  unsigned long	b4[2], bfSize, bfOffBits;	/* 4 bytes */
  unsigned char	ctable[256*4], *pix;		/* 1 byte */
  FILE  *fo;
  
  if ( (fo = fopen(output_filename,"wb")) == NULL ){
    printf("Can't open file: %s\n", output_filename);
    exit(EXIT_FAILURE);
  }
  
  findmaxmin2(nx, ny, val, &max_val_rsl, &min_val_rsl);
  
  /* tech. */
  max_val = max_val_rsl;
  min_val = 0.0;
  if (max_val <= 0.0){
    max_val = EPS;
  }
  
  num_pix = nx * ny;
  bfOffBits = (unsigned long) (14 + 40 + 256 * 4);
  bfSize = bfOffBits + (unsigned long) ( num_pix * sq(fmag) );	/* byte */
  /*	printf("$$ bfOffBits, bfSize = %d %d\n", bfOffBits, bfSize); */
  
  /* file header (14 bytes) */
  b2[0] = 0x4D42U; /* bfType 4D42 */				fwrite(b2, 2, 1, fo);
  b4[0] = bfSize;	/* bfSize: file size (byte) */	fwrite(b4, 4, 1, fo);
  b2[0] = 0x0000U;	/* bfReserved1 */			fwrite(b2, 2, 1, fo);
  b2[0] = 0x0000U;	/* bfReserved2 */			fwrite(b2, 2, 1, fo);
  b4[0] = bfOffBits;		/* bfOffBits */			fwrite(b4, 4, 1, fo);
  
  /* info header (40 bytes) */
  b4[0] = 40UL;		/* bcSize (Win format) */	fwrite(b4, 4, 1, fo);
  b4[0] = (unsigned long) (nx * fmag);	/* bcWidth */	fwrite(b4, 4, 1, fo);
  b4[0] = (unsigned long) (ny * fmag);	/* bcHeight */	fwrite(b4, 4, 1, fo);
  b2[0] = 1U;		/* bcPlanes */					fwrite(b2, 2, 1, fo);
  b2[0] = 8U;		/* bcBitCount */				fwrite(b2, 2, 1, fo);
  b4[0] = 0x0L;		fwrite(b4, 4, 1, fo);	/* biCompression */
  b4[0] = 0x0L;		fwrite(b4, 4, 1, fo);	/* biSizeImage */
  b4[0] = 0L;			fwrite(b4, 4, 1, fo);	/* biXPelsPerMeter */
  b4[0] = 0L;			fwrite(b4, 4, 1, fo);	/* biYPelsPerMeter */
  b4[0] = 0x0L;		fwrite(b4, 4, 1, fo);	/* biClrUsed */
  b4[0] = 0x0L;		fwrite(b4, 4, 1, fo);	/* biClrImportant */
  
  
  /* Colour table (8-bit, gray) (4*256 bytes) */
  for(i=0; i<256; i++){
    ctable[i * 4 + 0] = (unsigned char) (255-i);	/* B */
    ctable[i * 4 + 1] = (unsigned char) (255-i);	/* G */
    ctable[i * 4 + 2] = (unsigned char) (255-i);	/* R */
    ctable[i * 4 + 3] = 0x00U;				/* N */
  }
  fwrite(ctable, 1, 4*256, fo);
  
  /* Pixel data (1byte data) (nx * ny bytes) */
  pix  = new_vector(unsigned char, num_pix * sq(fmag));	chk_vector(pix);
  ct = 0;
  for(j = ny * fmag -1; j >= 0; j--){
    for(i = 0; i < nx * fmag; i++){
      y = val[i / fmag][j / fmag];
      pix_val = (
        255.0 * (y - min_val) / (max_val - min_val)
        );
      if(pix_val < 0.0){
        pix[ct] = 0U;
      }
      else if(pix_val > 255.0){
        pix[ct] = 255U;
      }
      else{
        pix[ct] = (unsigned char) pix_val;
      }
      ct++;
    }
  }
  fwrite(pix, 1, (unsigned long) (num_pix * sq(fmag)), fo);
  
  free_vector(pix);
  fclose(fo);
}
#endif


#if 0
void bmp_out_cont(
  char *output_filename, double **val, int nx, int ny,
  int fmag, double gamma, double max_val, double min_val, int cont_rv_flag
  )
{
  /*
   * BMP-format output (Gray 8-bit, inverted)
   *  fmag: Magnification of image (integer)
   *  nx and ny should be multiples of 4.
   *  gamma: gamma coefficient
   *  max_val, min_val: max and min for val[]
   *  cont_rv_flag: 1:contrast reversed, 2:not reversed
   */
  long int i, j, j_end, i_end, ct, num_pix;
  double y, pix_val, span;
  unsigned short  b2[2];						/* 2 bytes */
  unsigned long	b4[2], bfSize, bfOffBits;	/* 4 bytes */
  unsigned char	ctable[256*4], *pix;		/* 1 byte */
  FILE  *fo;
  
  if ( (fo = fopen(output_filename,"wb")) == NULL ){
    printf("Can't open file: %s\n", output_filename);
    exit(EXIT_FAILURE);
  }
  
  /* Check */
  if (max_val < min_val){
    printf("$$ (bmp_out_cont) Error: max <= min. Abort.\n");
    exit(EXIT_FAILURE);
  }
  if (max_val == min_val){
    max_val = min_val + epsilon.eps;
  }
  if (gamma <= 0.0){
    printf("$$ (bmp_out_cont) Error: gamma coef. < 0.0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  if (nx % 4 != 0 || ny % 4 != 0){
    printf("$$ nx and ny should be a multiple of 4. Abort.\n");
    exit(EXIT_FAILURE);
  }
  
  num_pix = (long) nx * (long) ny;
  bfOffBits = (unsigned long) (14 + 40 + 256 * 4);
  bfSize = bfOffBits + (unsigned long) ( num_pix * sq(fmag) );	/* byte */
  /*	printf("$$ bfOffBits, bfSize = %d %d\n", bfOffBits, bfSize); */
  
  /* file header (14 bytes) */
  b2[0] = 0x4D42U; /* bfType 4D42 */				fwrite(b2, 2, 1, fo);
  b4[0] = bfSize;	/* bfSize: file size (byte) */	fwrite(b4, 4, 1, fo);
  b2[0] = 0x0000U;	/* bfReserved1 */			fwrite(b2, 2, 1, fo);
  b2[0] = 0x0000U;	/* bfReserved2 */			fwrite(b2, 2, 1, fo);
  b4[0] = bfOffBits;		/* bfOffBits */			fwrite(b4, 4, 1, fo);
  
  /* info header (40 bytes) */
  b4[0] = 40UL;		/* bcSize (Win format) */	fwrite(b4, 4, 1, fo);
  b4[0] = (unsigned long) (nx * fmag);	/* bcWidth */	fwrite(b4, 4, 1, fo);
  b4[0] = (unsigned long) (ny * fmag);	/* bcHeight */	fwrite(b4, 4, 1, fo);
  b2[0] = 1U;		/* bcPlanes */					fwrite(b2, 2, 1, fo);
  b2[0] = 8U;		/* bcBitCount */				fwrite(b2, 2, 1, fo);
  b4[0] = 0x0L;		fwrite(b4, 4, 1, fo);	/* biCompression */
  b4[0] = 0x0L;		fwrite(b4, 4, 1, fo);	/* biSizeImage */
  b4[0] = 0L;			fwrite(b4, 4, 1, fo);	/* biXPelsPerMeter */
  b4[0] = 0L;			fwrite(b4, 4, 1, fo);	/* biYPelsPerMeter */
  b4[0] = 0x0L;		fwrite(b4, 4, 1, fo);	/* biClrUsed */
  b4[0] = 0x0L;		fwrite(b4, 4, 1, fo);	/* biClrImportant */
  
  /* Colour table (8-bit, gray) (4*256 bytes) */
  if(cont_rv_flag == 1){	/* contrast reversed */
    for(i=0; i<256; i++){
      ctable[i * 4 + 0] = (unsigned char) (255-i);	/* B */
      ctable[i * 4 + 1] = (unsigned char) (255-i);	/* G */
      ctable[i * 4 + 2] = (unsigned char) (255-i);	/* R */
      ctable[i * 4 + 3] = 0x00U;				/* N */
    }
  }
  else{
    for(i=0; i<256; i++){
      ctable[i * 4 + 0] = (unsigned char) i;	/* B */
      ctable[i * 4 + 1] = (unsigned char) i;	/* G */
      ctable[i * 4 + 2] = (unsigned char) i;	/* R */
      ctable[i * 4 + 3] = 0x00U;				/* N */
    }
  }
  fwrite(ctable, 1, 4*256, fo);
  
  /* Pixel data (1byte data) (nx * ny bytes) */
  pix  = new_vector(unsigned char, num_pix * sq(fmag));	chk_vector(pix);
  span = max_val - min_val;
  
  ct = 0;
  /* j_bgn = ny * fmag - 1; */
  j_end = ny * fmag;
  i_end = nx * fmag;
  /*	for(j = j_bgn; j >= 0; j--){ */
  for(j = 0; j < j_end; j++){
    for(i = 0; i < i_end; i++){
      y = val[i / fmag][j / fmag];
      if(y > max_val){ y = max_val; }
      if(y < min_val){ y = min_val; }
      pix_val = pow( (y - min_val) / span, gamma );	/* Gamma curve */
      pix[ct] = (unsigned char) (pix_val * 255.0);
      ct++;
    }
  }
  fwrite(pix, 1, (unsigned long) (num_pix * sq(fmag)), fo);
  free_vector(pix);
  fclose(fo);
}
#endif


#if 0
void sim_visualization_diff(
  FILE *fp1, FILE *fp2, char infil1[], char infil2[],
  double gamma_coef, double max_val, double min_val, int mag_coef, int cont_rev_flag
  )
     /* Visualization of SIM intensities: difference plot */
{
  int		ik, ig, nk, ng, ix, iy, nx, ny;
  int		center_x, center_y;
  /*	double	scale; */
  double	**mat;
  double	t[7], thickness, *hpos, *gpos, max_H, min_H, max_G, min_G;
  double	*Hval, *Gval, val1, val2;
  int		min_hpix, max_hpix, min_gpix, max_gpix, *hpix, *gpix;
  int		posx_g, posy_g, max_x, max_y, min_x, min_y;
  int		nz, np_zolz, np_holz, h1, k1, l1, h2, k2, l2;
  double	ruz, rdz, ruh, rdh;	/* R-factor */
  double	dscr_zolz, dscr_holz;	/* Descrepancy index */
  UVW		incidence;
  R_HKL	zolzbase1;
  /*	R_HKL	zolzbase2; */
  HKL		g, *cal_g;
  /*	char	rsl_fil[BUFSIZE]; */
  char	buf1[BUFSIZE], buf2[BUFSIZE], bmp_file[BUFSIZE];
  CELL_PAR	bv;
  
  printf("***** VISUALIZATION : Difference plot mode\n");
  printf("Difference between %s.rsl and %s.rsl\n", infil1, infil2);
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  str_to_fs(7, buf1, t);
  bv.s11 = t[0]; bv.s22 = t[1]; bv.s33 = t[2];
  bv.s12 = t[3]; bv.s13 = t[4]; bv.s23 = t[5]; bv.vc = t[6];
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  str_to_fs(6, buf1, t);
  bv.aas = t[0]; bv.bbs = t[1]; bv.ccs = t[2];
  bv.abs = t[3]; bv.bcs = t[4]; bv.acs = t[5];
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  str_to_fs(6, buf1, t);
  bv.aa = t[0]; bv.bb = t[1]; bv.cc = t[2];
  bv.ab = t[3]; bv.bc = t[4]; bv.ac = t[5];
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  str_to_fs(3, buf1, t);
  incidence.u = (int) t[0];	incidence.v = (int) t[1];	incidence.w = (int) t[2];
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  str_to_fs(3, buf1, t);
  zolzbase1.h = t[0];	zolzbase1.k = t[1];	zolzbase1.l = t[2];
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  thickness = atof(buf1);
  
  chk_data("PARAMETERS_END", fp1);
  chk_data("PARAMETERS_END", fp2);
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  ng = atoi(buf1);
  
  cal_g = new_vector(HKL, ng);		chk_vector(cal_g);
  hpos = new_vector(double, ng);		chk_vector(hpos);
  gpos = new_vector(double, ng);		chk_vector(gpos);
  
  for(ig = 0; ig < ng; ig++){
    comment_skip_fgets(100, buf1, fp1);
    comment_skip_fgets(100, buf2, fp2);
    str_to_fs(5, buf1, t);
    cal_g[ig].h = (int) t[0];	cal_g[ig].k = (int) t[1];	cal_g[ig].l = (int) t[2];
    hpos[ig] = t[3];	gpos[ig] = t[4];
  }
  chk_data("RF_INDEX_END", fp1);
  chk_data("RF_INDEX_END", fp2);
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  nk = atoi(buf1);
  
  Hval = new_vector(double, nk);	chk_vector(Hval);
  Gval = new_vector(double, nk);	chk_vector(Gval);
  hpix = new_vector(int, nk);		chk_vector(hpix);
  gpix = new_vector(int, nk);		chk_vector(gpix);
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  str_to_fs(2, buf1, t);
  max_H = t[0];	min_H = t[1];
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  str_to_fs(2, buf1, t);
  max_G = t[0];	min_G = t[1];
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  str_to_fs(2, buf1, t);
  max_hpix = (int)t[0]; min_hpix = (int)t[1];
  
  comment_skip_fgets(100, buf1, fp1);
  comment_skip_fgets(100, buf2, fp2);
  str_to_fs(2, buf1, t);
  max_gpix = (int)t[0]; min_gpix = (int)t[1];
  
  for(ik=0; ik<nk; ik++){
    comment_skip_fgets(100, buf1, fp1);
    comment_skip_fgets(100, buf2, fp2);
    str_to_fs(4, buf1, t);
    Hval[ik] = t[0];		Gval[ik] = t[1];
    hpix[ik] = (int)t[2];	gpix[ik] = (int)t[3];
  }
  chk_data("COORD_END", fp1);
  chk_data("COORD_END", fp2);
  
  /* Evaluation of the image size */
  max_x = min_x = max_y = min_y = 0;
  for(ig = 0; ig < ng; ig++){
    if(max_x < (int) hpos[ig] + max_hpix){ max_x = (int) hpos[ig] + max_hpix; }
    if(min_x > (int) hpos[ig] + min_hpix){ min_x = (int) hpos[ig] + min_hpix; }
    if(max_y < (int) gpos[ig] + max_gpix){ max_y = (int) gpos[ig] + max_gpix; }
    if(min_y > (int) gpos[ig] + min_gpix){ min_y = (int) gpos[ig] + min_gpix; }
  }
  nx = max_x - min_x;
  ny = max_y - min_y;
  if( nx <= 0 ){
    printf("$$ (mb_visualization) Error: nx <= 0; Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( ny <= 0 ){
    printf("$$ (mb_visualization) Error: ny <= 0; Abort.\n");
    exit(EXIT_FAILURE);
  }
  nx = ( (int) (nx / 4) + 2) * 4;	/* <<<<< */
  ny = ( (int) (ny / 4) + 2) * 4;	/* <<<<< */
  center_x = - min_x + 4;
  center_y = - min_y + 4;
  printf(
    "$$ nx, ny = %d %d, center_x, center_y = %d %d\n",
    nx, ny, center_x, center_y
    );
  mat = new_matrix(double, nx, ny);	chk_matrix(mat);
  /* initialize */
  for(ix = 0; ix < nx; ix++){
    for(iy = 0; iy < ny; iy++){
      mat[ix][iy] = 0.0;
    }
  }
  ruz = rdz = ruh = rdh = 0.0;	/* R-factor */
  dscr_zolz = dscr_holz = 0.0;	/* Discrepancy index */
  np_zolz = np_holz = 0;
  
  for(ig=0; ig<ng; ig++){
    comment_skip_fgets(sizeof buf1, buf1, fp1);
    str_to_fs(3, buf1, t);
    h1 = g.h = (int)t[0];	k1 = g.k = (int)t[1]; l1 = g.l = (int)t[2];
    if( g.h != cal_g[ig].h || g.k != cal_g[ig].k || g.l != cal_g[ig].l ){
      printf("$$ (mb_visualization) mismatch in hkl. Abort.\n");
      exit(EXIT_FAILURE);
    }
    comment_skip_fgets(sizeof buf2, buf2, fp2);
    str_to_fs(3, buf2, t);
    h2 = (int)t[0];	k2 = (int)t[1]; l2 = (int)t[2];
    if( h2 != h1 || k2 != k1 || l2 != l1 ){
      printf("$$ (mb_visualization) mismatch in hkl. Abort.\n");
      exit(EXIT_FAILURE);
    }
    nz = h1 * incidence.u + k1 * incidence.v + l1 * incidence.w;
    
    posx_g = (int) hpos[ig];	posy_g = (int) gpos[ig];
    for(ik=0; ik<nk; ik++){
      comment_skip_fgets(100, buf1, fp1);
      comment_skip_fgets(100, buf2, fp2);
      val1 = atof(buf1);
      val2 = atof(buf2);
      ix = center_x + posx_g + hpix[ik];
      iy = center_y + posy_g + gpix[ik];
      if( ix >= 0 && ix < nx && iy >= 0 && iy < ny ){
        mat[ix][iy] += fabs(val1 - val2);
      }
      else{
        printf("$$ (mb_visualization) (%d,%d) WARNING: out of area.\n", ix, iy);
      }
      
      if( nz == 0 ){	/* ZOLZ */
        ruz += fabs(val1 - val2);
        rdz += fabs(val1);
        if(val1 > 0.0){
          dscr_zolz += sq(val1 - val2) / val1;	/* <<<<< */
          np_zolz++;
        }
      }
      else if ( nz > 0 ){	/* HOLZ */
        ruh += fabs(val1 - val2);
        rdh += fabs(val1);
        if(val1 > 0.0){
          dscr_holz += sq(val1 - val2) / val1;	/* <<<<< */
          np_holz++;
        }
      }
      else{
        printf("$$ (sim_visualization_diff) Invalid laue zone. Abort.");
      }
    }
    chk_data("RF_END", fp1);
    chk_data("RF_END", fp2);
  }
  
  /* R-factors */
  if(rdz + rdh > 0.0){
    printf("R-factor(TOTAL)= %lf\n", (ruz + ruh) / (rdz + rdh));
  }
  if(rdz > 0.0){
    printf("R-factor(ZOLZ) = %lf\n", ruz / rdz);
  }
  if(rdh > 0.0){
    printf("R-factor(HOLZ) = %lf\n", ruh / rdh);
  }
  
  /* Discrepancy indices */
  if(np_zolz + np_holz > 0){
    printf("Discrepancy index (TOTAL)= %lf\n",
           (dscr_zolz + dscr_holz) / (double)(np_zolz + np_holz) );
  }
  if(np_zolz > 0){
    printf("Number of data points (ZOLZ) = %d\n", np_zolz);
    printf("Discrepancy index (ZOLZ) = %lf\n", dscr_zolz / np_zolz);
  }
  if(np_holz > 0){
    printf("Number of data points (HOLZ) = %d\n", np_holz);
    printf("Discrepancy index (HOLZ) = %lf\n", dscr_holz / np_holz);
  }
  /* Bitmap file */
  sprintf(bmp_file, "%s-d.bmp", infil2);
  printf("BMP file output --> %s\n", bmp_file);
  bmp_out_cont(
    bmp_file, mat, nx, ny, mag_coef, gamma_coef, max_val, min_val, cont_rev_flag
    );
  free_vector(Hval);	free_vector(Gval);
  free_vector(hpix);	free_vector(gpix);
  free_vector(cal_g);
  free_vector(hpos);	free_vector(gpos);
  free_matrix(mat);
}


void sim_visualization(
  FILE *fp, char infil[],
  double gamma_coef, double max_val, double min_val,
  int mag_coef, int cont_rev_flag
  )
     /* Visualization of SIM intensities */
{
  int		ik, ig, nk, ng, ix, iy, nx, ny;
  int		center_x, center_y;
  double	**mat;
  double	t[7], thickness, *hpos, *gpos, max_H, min_H, max_G, min_G;
  double	*Hval, *Gval, val;
  int		min_hpix, max_hpix, min_gpix, max_gpix, *hpix, *gpix;
  int		posx_g, posy_g, max_x, max_y, min_x, min_y;
  UVW		incidence;
  R_HKL	zolzbase1;
  HKL		g, *cal_g;
  char  buf[BUFSIZE], bmp_file[BUFSIZE], pgm_file[BUFSIZE];
  CELL_PAR	bv;
  
  printf("***** VISUALIZATION mode\n");
#if 0
  if (command_line_visual == 1){		/* command line mode */
    
    printf("reading %s\n", rsl_fil);
    chk_data("VISUAL", fp);
    fget_2d(&max_val, &min_val, fp);
    fget_d(fp);		/* gamma_coef dummy */
    fget_i(fp);		/* mag_coef dummy */
    fget_i(fp); 	/* cont_rev_flag  dummy */
    
  }
  else if (command_line_visual == 0){				/* console mode */
    printf("Data filename without extension (rsl) (SIM: disp_type=7) \n>>");
    fgets(infil,sizeof(infil),stdin);
    sprintf(rsl_fil, "%s.rsl", infil);
    if( (fp = fopen(rsl_fil,"r")) == NULL ){				/* rsl file (input) */
      fprintf(stderr,"Cannot open %s\n", rsl_fil);
      exit(EXIT_FAILURE);
    }
    
    printf("reading %s\n", rsl_fil);
    chk_data("VISUAL", fp);
    fget_2d(&max_val, &min_val, fp);
    gamma_coef = fget_d(fp);
    mag_coef = fget_i(fp);
    cont_rev_flag = fget_i(fp);
    
    printf("max, min ? (std.: %lf, %lf)>>", max_val, min_val);
    fgets(buf,sizeof(buf),stdin);	
    str_to_fs(2, buf, t);	max_val = t[0], min_val = t[1];
    printf("gamma coef. ? (>0) >>");
    fgets(buf,sizeof(buf),stdin);	
    str_to_fs(1, buf, t);	gamma_coef = t[0];
    printf("magnification ? (int) >>");
    fgets(buf,sizeof(buf),stdin);	
    mag_coef = atoi(buf);
    printf("contrast reverse ? [1:on, 0:off] >>");
    fgets(buf,sizeof(buf),stdin);	
    cont_rev_flag = atoi(buf);
  }
  
  /* Check */
  printf("max, min = %lf, %lf\n", max_val, min_val);
  printf("gamma_coef = %lf\n", gamma_coef);
  printf("mag_coef = %d\n", mag_coef);
  printf("cont_rev_flag = %d\n", cont_rev_flag);
  if( gamma_coef < 0.0 ){
    printf("$$ (mb_visualization) Error: gamma < 0.0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( mag_coef <= 0 ){
    printf("$$ (mb_visualization) Error: mag_coef <= 0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( cont_rev_flag != 0 && cont_rev_flag != 1){
    printf("$$ (mb_visualization) Error: cont_rev_flag must be 0 or 1. Abort.\n");
    exit(EXIT_FAILURE);
  }
#endif
  
  comment_skip_fgets(sizeof buf, buf, fp); 	str_to_fs(7, buf, t);
  bv.s11 = t[0]; bv.s22 = t[1]; bv.s33 = t[2];
  bv.s12 = t[3]; bv.s13 = t[4]; bv.s23 = t[5]; bv.vc = t[6];
  comment_skip_fgets(sizeof buf, buf, fp); 	str_to_fs(6, buf, t);
  bv.aas = t[0]; bv.bbs = t[1]; bv.ccs = t[2];
  bv.abs = t[3]; bv.bcs = t[4]; bv.acs = t[5];
  comment_skip_fgets(sizeof buf, buf, fp); 	str_to_fs(6, buf, t);
  bv.aa = t[0]; bv.bb = t[1]; bv.cc = t[2];
  bv.ab = t[3]; bv.bc = t[4]; bv.ac = t[5];
  
  fget_3i( &(incidence.u), &(incidence.v), &(incidence.w), fp );
  fget_3d( &(zolzbase1.h), &(zolzbase1.k), &(zolzbase1.l), fp );
  thickness = fget_d(fp);
  chk_data("PARAMETERS_END", fp);
  
  ng = fget_i(fp);
  cal_g = new_vector(HKL, ng);	chk_vector(cal_g);
  hpos = new_vector(double, ng);		chk_vector(hpos);
  gpos = new_vector(double, ng);		chk_vector(gpos);
  for(ig = 0; ig < ng; ig++){
    comment_skip_fgets(sizeof buf, buf, fp);
    str_to_fs(5, buf, t);
    cal_g[ig].h = (int) t[0];
    cal_g[ig].k = (int) t[1];
    cal_g[ig].l = (int) t[2];
    hpos[ig] = t[3];	gpos[ig] = t[4];
  }
  chk_data("RF_INDEX_END", fp);
  
  nk = fget_i(fp);
  Hval = new_vector(double, nk);	chk_vector(Hval);
  Gval = new_vector(double, nk);	chk_vector(Gval);
  hpix = new_vector(int, nk);		chk_vector(hpix);
  gpix = new_vector(int, nk);		chk_vector(gpix);
  fget_2d(&max_H, &min_H, fp);
  fget_2d(&max_G, &min_G, fp);
  fget_2i(&max_hpix, &min_hpix, fp);
  fget_2i(&max_gpix, &min_gpix, fp);
  for(ik=0; ik<nk; ik++){
    comment_skip_fgets(sizeof buf, buf, fp); 	str_to_fs(4, buf, t);
    Hval[ik] = t[0];		Gval[ik] = t[1];
    hpix[ik] = (int)t[2];	gpix[ik] = (int)t[3];
  }
  chk_data("COORD_END", fp);
  
  /* Evaluation of the image size */
  max_x = min_x = max_y = min_y = 0;
  for(ig = 0; ig < ng; ig++){
    if(max_x < (int) hpos[ig] + max_hpix){ max_x = (int) hpos[ig] + max_hpix; }
    if(min_x > (int) hpos[ig] + min_hpix){ min_x = (int) hpos[ig] + min_hpix; }
    if(max_y < (int) gpos[ig] + max_gpix){ max_y = (int) gpos[ig] + max_gpix; }
    if(min_y > (int) gpos[ig] + min_gpix){ min_y = (int) gpos[ig] + min_gpix; }
  }
  nx = max_x - min_x;
  ny = max_y - min_y;
  if( nx <= 0 ){
    printf("$$ (mb_visualization) Error: nx <= 0; Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( ny <= 0 ){
    printf("$$ (mb_visualization) Error: ny <= 0; Abort.\n");
    exit(EXIT_FAILURE);
  }
  nx = ( (int) (nx / 4) + 2) * 4;	/* <<<<< */
  ny = ( (int) (ny / 4) + 2) * 4;	/* <<<<< */
  center_x = - min_x + 4;
  center_y = - min_y + 4;
  
  printf(
    "$$ nx, ny = %d %d, center_x, center_y = %d %d\n",
    nx, ny, center_x, center_y
    );
  
  mat = new_matrix(double, nx, ny);  chk_matrix(mat);
  /* initialize */
  for(ix = 0; ix < nx; ix++){
    for(iy = 0; iy < ny; iy++){
      mat[ix][iy] = 0.0;
    }
  }
  
  for(ig=0; ig<ng; ig++){
    fget_3i(&(g.h), &(g.k), &(g.l), fp);
    if( g.h != cal_g[ig].h || g.k != cal_g[ig].k || g.l != cal_g[ig].l ){
      printf("$$ (mb_visualization) mismatch in hkl. Abort.\n");
      exit(EXIT_FAILURE);
    }
    posx_g = (int) hpos[ig];	posy_g = (int) gpos[ig];
    for(ik=0; ik<nk; ik++){
      val = fget_d(fp);
      ix = center_x + posx_g + hpix[ik];
      iy = center_y + posy_g + gpix[ik];
      if( ix >= 0 && ix < nx && iy >= 0 && iy < ny ){
        mat[ix][iy] += val;
      }
      else{
        printf("$$ (mb_visualization) (%d,%d) WARNING: out of area.\n", ix, iy);
      }
    }
    chk_data("RF_END", fp);
  }
  sprintf(pgm_file, "%s.pgm", infil);
  output_pgm_image(
    nx, ny, mat, max_val, min_val,
    gamma_coef, mag_coef, cont_rev_flag, pgm_file
    );
  
  free_vector(Hval);	free_vector(Gval);
  free_vector(hpix);	free_vector(gpix);
  free_vector(cal_g);
  free_vector(hpos);	free_vector(gpos);
  free_matrix(mat);
}


void sim_visualization_coherent(
  FILE *fp, char infil[],
  double gamma_coef, double max_val, double min_val, int mag_coef, int cont_rev_flag
  )
     /* Visualization of SIM intensities */
{
  int		ik, ig, nk, ng, ix, iy, nx, ny;
  int		center_x, center_y;
  double	val_re, val_im, **mat;
  complex	**cmat, ta;
  double	t[7], thickness, *hpos, *gpos, max_H, min_H, max_G, min_G;
  double	*Hval, *Gval;
  int		min_hpix, max_hpix, min_gpix, max_gpix, *hpix, *gpix;
  int		posx_g, posy_g, max_x, max_y, min_x, min_y;
  UVW		incidence;
  R_HKL	zolzbase1;
  HKL		g, *cal_g;
  char	buf[BUFSIZE], bmp_file[BUFSIZE];
  CELL_PAR	bv;
  
  printf("***** VISUALIZATION (coherent) mode\n");
  
  comment_skip_fgets(100, buf, fp); 	str_to_fs(7, buf, t);
  bv.s11 = t[0]; bv.s22 = t[1]; bv.s33 = t[2];
  bv.s12 = t[3]; bv.s13 = t[4]; bv.s23 = t[5]; bv.vc = t[6];
  comment_skip_fgets(100, buf, fp); 	str_to_fs(6, buf, t);
  bv.aas = t[0]; bv.bbs = t[1]; bv.ccs = t[2];
  bv.abs = t[3]; bv.bcs = t[4]; bv.acs = t[5];
  comment_skip_fgets(100, buf, fp); 	str_to_fs(6, buf, t);
  bv.aa = t[0]; bv.bb = t[1]; bv.cc = t[2];
  bv.ab = t[3]; bv.bc = t[4]; bv.ac = t[5];
  
  fget_3i( &(incidence.u), &(incidence.v), &(incidence.w), fp );
  fget_3d( &(zolzbase1.h), &(zolzbase1.k), &(zolzbase1.l), fp );
  thickness = fget_d(fp);
  chk_data("PARAMETERS_END", fp);
  
  ng = fget_i(fp);
  cal_g = new_vector(HKL, ng);	chk_vector(cal_g);
  hpos = new_vector(double, ng);		chk_vector(hpos);
  gpos = new_vector(double, ng);		chk_vector(gpos);
  for(ig = 0; ig < ng; ig++){
    comment_skip_fgets(100, buf, fp); 	str_to_fs(5, buf, t);
    cal_g[ig].h = (int) t[0];	cal_g[ig].k = (int) t[1];	cal_g[ig].l = (int) t[2];
    hpos[ig] = t[3];	gpos[ig] = t[4];
  }
  chk_data("RF_INDEX_END", fp);
  
  nk = fget_i(fp);
  Hval = new_vector(double, nk);	chk_vector(Hval);
  Gval = new_vector(double, nk);	chk_vector(Gval);
  hpix = new_vector(int, nk);		chk_vector(hpix);
  gpix = new_vector(int, nk);		chk_vector(gpix);
  fget_2d(&max_H, &min_H, fp);
  fget_2d(&max_G, &min_G, fp);
  fget_2i(&max_hpix, &min_hpix, fp);
  fget_2i(&max_gpix, &min_gpix, fp);
  for(ik=0; ik<nk; ik++){
    comment_skip_fgets(100, buf, fp); 	str_to_fs(4, buf, t);
    Hval[ik] = t[0];		Gval[ik] = t[1];
    hpix[ik] = (int)t[2];	gpix[ik] = (int)t[3];
  }
  chk_data("COORD_END", fp);
  
  /* Evaluation of the image size */
  max_x = min_x = max_y = min_y = 0;
  for(ig = 0; ig < ng; ig++){
    if(max_x < (int) hpos[ig] + max_hpix){ max_x = (int) hpos[ig] + max_hpix; }
    if(min_x > (int) hpos[ig] + min_hpix){ min_x = (int) hpos[ig] + min_hpix; }
    if(max_y < (int) gpos[ig] + max_gpix){ max_y = (int) gpos[ig] + max_gpix; }
    if(min_y > (int) gpos[ig] + min_gpix){ min_y = (int) gpos[ig] + min_gpix; }
  }
  nx = max_x - min_x;
  ny = max_y - min_y;
  if( nx <= 0 ){
    printf("$$ (mb_visualization) Error: nx <= 0; Abort.\n");
    exit(EXIT_FAILURE);
  }
  if( ny <= 0 ){
    printf("$$ (mb_visualization) Error: ny <= 0; Abort.\n");
    exit(EXIT_FAILURE);
  }
  nx = ( (int) (nx / 4) + 2) * 4;	/* <<<<< */
  ny = ( (int) (ny / 4) + 2) * 4;	/* <<<<< */
  center_x = - min_x + 4;
  center_y = - min_y + 4;
  
  mat = new_matrix(double, nx, ny);	chk_matrix(mat);
  cmat = new_matrix(complex, nx, ny);	chk_matrix(cmat);
  
  /* initialize */
  for(ix = 0; ix < nx; ix++){
    for(iy = 0; iy < ny; iy++){
      mat[ix][iy] = 0.0;
      cmat[ix][iy] = c_conv(0.0, 0.0);
    }
  }
  
  for(ig=0; ig<ng; ig++){
    fget_3i(&(g.h), &(g.k), &(g.l), fp);
    if( g.h != cal_g[ig].h || g.k != cal_g[ig].k || g.l != cal_g[ig].l ){
      printf("$$ (mb_visualization) mismatch in hkl. Abort.\n");
      exit(EXIT_FAILURE);
    }
    posx_g = (int) hpos[ig];	posy_g = (int) gpos[ig];
    for(ik=0; ik<nk; ik++){
      fget_2d(&val_re, &val_im, fp);
      ix = center_x + posx_g + hpix[ik];
      iy = center_y + posy_g + gpix[ik];
      if( ix >= 0 && ix < nx && iy >= 0 && iy < ny ){
        cmat[ix][iy] = c_add( cmat[ix][iy], c_conv(val_re, val_im) );
      }
      else{
        printf("$$ (mb_visualization) WARNING: out of area.\n");
      }
    }
    chk_data("RF_END", fp);
  }
  
  /* amplitude to intensity */
  for(ix = 0; ix < nx; ix++){
    for(iy = 0; iy < ny; iy++){
      ta = cmat[ix][iy];
      mat[ix][iy] = ( sq(ta.re) + sq(ta.im) );
    }
  }
  
  /* find max min */
  findmaxmin2(nx, ny, mat, &max_val, &min_val);
  printf("$$ (sim_visualization_coherent) max, min = %lf %lf.\n", max_val, min_val);
  
  sprintf(bmp_file, "%s.bmp", infil);
  bmp_out_cont(
    bmp_file, mat, nx, ny, mag_coef, gamma_coef, max_val, min_val, cont_rev_flag
    );
  
  /* >>>>> ass part here */
  
  free_vector(Hval);	free_vector(Gval);
  free_vector(hpix);	free_vector(gpix);
  free_vector(cal_g);
  free_vector(hpos);	free_vector(gpos);
  free_matrix(mat);	free_matrix(cmat);
}
#endif


/* ----- */
#if 0
void set_mat_lnp(
  double **obs_mat, double **cal_mat, double **sig_mat,
  double Hbgn, double Hend, double Gbgn, double Gend,
  int np, double H, double G,
  double obs, double cal, double sig
  )
{
  int	x_obs, y_obs;
  
  /*	printf("H, G = %lf %lf\n", H, G);
	printf("$$ Hbgn, Gbgn, Hend, Gend= %lf %lf %lf %lf\n",  Hbgn, Gbgn, Hend, Gend);
   */
  /* Pixel coordinate: shi-sha go-nyu */
  x_obs = (int) ( (double)(np - 1) * (H - Hbgn) / (Hend - Hbgn) + 0.5 );		/* Shi-sha go-nyu */
  y_obs = (np - 1) - (int) ( (double)(np - 1) * (G - Gbgn) / (Gend - Gbgn) + 0.5 );
  /*	printf("$$ x, y = %d %d\n", x_obs, y_obs); */
  
  chk_array(x_obs, np);	chk_array(y_obs, np);
  
  obs_mat[x_obs][y_obs] = obs;
  cal_mat[x_obs][y_obs] = cal;
  sig_mat[x_obs][y_obs] = sig;
}


void 	dbg_matout(double **mat, int np)
{
  int i, j;
  FILE *fp;
  
  if ( (fp = fopen("dbgimg.out","w")) == NULL ){
    printf("Can't open file: dbgimg.out\n");	exit(EXIT_FAILURE);
  }
  
  printf("$$ Writing the image to dbgimg.out (matrix %d x %d)\n", np, np);
  
  for(j=0; j<np; j++){
    for(i=0; i<np; i++){
      fprintf(fp, "%lf\t", mat[i][j]);
    }
    fprintf(fp, "\n");
  }
  fclose(fp);
}


double cub_intpl(double x[], double xx, double f[])
     /* Use Lagrange interpolating cubic */
{
  double x1, x2, x3, x4, x12, x13, x14, x23, x24, x34, res;
  x1 = xx - x[0];
  x2 = xx - x[1];
  x3 = xx - x[2];
  x4 = xx - x[3];
  x12 = x[0] - x[1];
  x13 = x[0] - x[2];
  x14 = x[0] - x[3];
  x23 = x[1] - x[2];
  x24 = x[1] - x[3];
  x34 = x[2] - x[3];
  res =
    f[0] * x2 * x3 * x4 / (x12 * x13 * x14) -
      f[1] * x1 * x3 * x4 / (x12 * x23 * x24) +
        f[2] * x1 * x2 * x4 / (x13 * x23 * x34) -
          f[3] * x1 * x2 * x3 / (x14 * x24 * x34);
  return res;
}


double bicub_intpl(
  double xx, double yy, int nx, int ny, double *x, double *y, double **val
  )
     /* 2 dimensional */
{
  double h[4], f[4], hh[4], ff[4], res;
  int i, j, ix, iy;
  
  ix = 1; do{ ix++; } while(xx >= x[ix] && ix < nx - 2);
  iy = 1; do{ iy++; } while(yy >= y[iy] && iy < ny - 2);
  /*printf("$$ ix, iy = %d %d \n", ix, iy);*/
  
  for(j=0; j<4; j++){
    for(i=0; i<4; i++){
      h[i] = x[i + ix - 2];
      f[i] = val[i + ix - 2][j + iy - 2];
      /*	printf("$$ i f h = %d %lf %lf\n", i, f[i], h[i]);	*/
    }
    ff[j] = cub_intpl(h, xx, f);
    hh[j] = y[j + iy - 2];
    /* printf("$$ j ff hh = %d %lf %lf\n", j, ff[j], hh[j]);	*/
  }
  res = cub_intpl(hh, yy, ff);
  /*	printf("$$ res=%lf\n", res);	*/
  
  return res;
}


void lnp_out(
  char *datafil, char *outfil,
  double **obs_mat, double **cal_mat, double **sig_mat,
  double *xpix, double *ypix, int np,
  int n_avr, int lx_bgn, int ly_bgn, int lx_end, int ly_end
)
/* lnprf output */
{
  int		i, j, np_lnp;
  double	x, y, length, sx, sy, shiftx, shifty, x1, y1, x2, y2;
  double	obs, cal, sig, av_obs, av_cal, av_sig;
  double	sum_obs, sum_cal, sum_sig;
  FILE	*fo;
  
  if ( (fo = fopen(outfil,"w")) == NULL ){
    printf("Can't open file: %s\n", outfil);	exit(EXIT_FAILURE);
  }
  
  np_lnp = (int) sqrt( (double) sq(lx_end - lx_bgn) + (double) sq(ly_end - ly_bgn) );
  printf("$$np_lnp=%d\n", np_lnp);
  
  fprintf(fo, "%%line profile of %s\n", datafil);
  fprintf(fo, "%%(%d,%d)-(%d,%d)\n", lx_bgn, ly_bgn, lx_end, ly_end);
  fprintf(fo, "%%number of points\n");
  fprintf(fo, "%d\n", np_lnp);
  
  length = sqrt( sq( (double)(lx_end - lx_bgn) ) + sq((double)(ly_end - ly_bgn))  );
  sx = (double)(lx_end - lx_bgn) / length;
  sy = (double)(ly_end - ly_bgn) / length;
  shiftx = sy;
  shifty = -sx;
  
  fprintf(fo, "%%obs, cal, sig, dif\n");
  
  for(j=0; j<np_lnp; j++){
    x = (double) lx_bgn
      + (double)(lx_end - lx_bgn) * (double)j / (double)(np_lnp - 1);
    y = (double) ly_bgn
      + (double)(ly_end - ly_bgn) * (double)j / (double)(np_lnp - 1);
    
    /*	printf("%lf, %lf\n", x, y);	*/
    obs = bicub_intpl( x, y, np, np, xpix, ypix, obs_mat );
    cal = bicub_intpl( x, y, np, np, xpix, ypix, cal_mat );
    sig = bicub_intpl( x, y, np, np, xpix, ypix, sig_mat );
    
    /* printf("%lf\t%lf\n", obs_mat[(int)x][(int)y]); */
    /* averaging along the direction perpendicular to the line */
    sum_obs = obs;	sum_cal = cal; sum_sig = sig;
    for(i=1; i<=n_avr; i++){
      x1 = x + (double) i * shiftx;
      y1 = y + (double) i * shifty;
      x2 = x - (double) i * shiftx;
      y2 = y - (double) i * shifty;
      
      /* printf("%lf\t%lf\t%lf\t%lf\n", x, y, x1, y1); */
      
      sum_obs += bicub_intpl( x1, y1, np, np, xpix, ypix, obs_mat )
        + bicub_intpl( x2, y2, np, np, xpix, ypix, obs_mat );
      sum_cal += bicub_intpl( x1, y1, np, np, xpix, ypix, cal_mat )
        + bicub_intpl( x2, y2, np, np, xpix, ypix, cal_mat );
      sum_sig += bicub_intpl( x1, y1, np, np, xpix, ypix, sig_mat )
        + bicub_intpl( x2, y2, np, np, xpix, ypix, sig_mat );
    }
    av_obs = sum_obs / (double) (1 + 2 * n_avr);
    av_cal = sum_cal / (double) (1 + 2 * n_avr);
    av_sig = sum_sig / (double) (1 + 2 * n_avr);
    
    fprintf(fo, "%lf\t%lf\t%lf\t%lf\n", av_obs, av_cal, av_sig, av_obs-av_cal);
  }
  
  fclose(fo);
}


void mb_lnprf(void)
{
  int		i, j, n_points, num_data, np, nth_file, n_avr;
  int		lx_bgn, ly_bgn, lx_end, ly_end;
  double	ycal, yobs, ysig, x[6], Hbgn, Hend, Gbgn, Gend, G, H;
  double	**obs_mat, **cal_mat, **sig_mat, *xpix, *ypix;
  char	infil[BUFSIZE], outfil[BUFSIZE], resfil[BUFSIZE], infofil[BUFSIZE];
  char	buf[BUFSIZE], hklbuf[BUFSIZE];
  FILE	*fp, *fi;
  
  printf("*****  lnprf  ******************************\n");
  printf(" Extract lineprofile data from a res-file.\n");
  printf("********************************************\n\n");
  
  printf("Files ***.res and ***.inf are required.\n\n");
  
  printf("Data filename without extension (mbfit-output-format)\n>>");
  fgets(infil,sizeof(infil),stdin);
  
  strcpy(resfil, infil);	strcat(resfil, ".res");
  strcpy(infofil, infil);	strcat(infofil, ".inf");
  
  if( (fp = fopen(resfil,"r")) == NULL ){				/* res file (input) */
    fprintf(stderr,"Cannot open %s\n",resfil);
    exit(EXIT_FAILURE);
  }
  
  if( (fi = fopen(infofil,"r")) == NULL ){			/* info file (input) */
    fprintf(stderr,"Cannot open %s\n",infofil);
    printf("Input info file (without extension inf)\n>>");
    fgets(infofil,sizeof(infofil),stdin);	
    strcat(infofil,".inf");
    
    if( (fi = fopen(infofil,"r")) == NULL ){
      fprintf(stderr,"Cannot open %s\n",infofil);
      exit(EXIT_FAILURE);
    }
  }
  
  /* Read from info file */	
  np = fget_i(fi);
  if( np <= 0 ){
    printf("$$ np must be greater than 0. Abort.\n");	exit(EXIT_FAILURE);
  }
  fget_4d(&Hbgn, &Hend, &Gbgn, &Gend, fi);
  chk_data("INFO_END", fi);
  fclose(fi);
  
  printf("Line profile: px_bgn, py_bgn, px_end, py_end (pixel coord.)>>");
  fgets(buf,sizeof(buf),stdin);	str_to_fs(4,buf,x);
  lx_bgn = (int)x[0];	ly_bgn = (int)x[1];
  lx_end = (int)x[2]; ly_end = (int)x[3];
  printf("Number of lines : for averaging [0: no average]>>");
  fgets(buf,sizeof(buf),stdin);	n_avr = atoi(buf);
  if( n_avr < 0 ){
    printf("$$ n_avr >= 0. Abort.\n");	exit(EXIT_FAILURE);
  }
  
  /* Intensity data */
  obs_mat = new_matrix(double, np, np);		chk_matrix(obs_mat);
  cal_mat = new_matrix(double, np, np);		chk_matrix(cal_mat);
  sig_mat = new_matrix(double, np, np);		chk_matrix(sig_mat);
  xpix	= new_vector(double, np);			chk_vector(xpix);
  ypix	= new_vector(double, np);			chk_vector(ypix);
  
  num_data = 0;	nth_file = 0;
  
  while( comment_skip_fgets(80, buf, fp) != NULL ){
    
    nth_file++;
    strcpy(hklbuf, buf);
    printf("$$%s\n", rm_cr(hklbuf));
/*    exit_fgets(buf, sizeof buf, fp);
    str_to_fs(1,buf,x);	n_points = (int) x[0]; */
    n_points = fget_i(fp);
    num_data += n_points;
    /* initialize */
    for(i=0; i<np; i++){
      xpix[i] = ypix[i] = (double) i;
      for(j=0; j<np; j++){
        obs_mat[i][j] = cal_mat[i][j] = sig_mat[i][j] = 0.0;
      }
    }
    
    /* Set data to matrices */
    for(i=0; i<n_points; i++){
/*      exit_fgets(buf, sizeof buf, fp);
      str_to_fs(5, buf, x);
      H = x[0];	G = x[1];	yobs = x[2];	ycal = x[3];	ysig = x[4]; */
      fget_5d( &H, &G, &yobs, &ycal, &ysig, fp );
      set_mat_lnp(
        obs_mat, cal_mat, sig_mat, Hbgn, Hend, Gbgn, Gend,
        np, H, G, yobs, ycal, ysig
        );
    }
    dbg_matout(obs_mat, np);
    
    sprintf(outfil, "%d.lnp", nth_file);
    printf("* created file: %s\n", outfil);
    lnp_out(
      infil, outfil, obs_mat, cal_mat, sig_mat, xpix, ypix, np,
      n_avr, lx_bgn, ly_bgn, lx_end, ly_end
      );
    
  }
  
  fclose(fp);
  
  printf("data file: %s\n\n", infil);
  printf("Number of data points = %d\n\n", num_data);
  
  free_matrix(obs_mat);	free_matrix(cal_mat);	free_matrix(sig_mat);
  free_vector(xpix);		free_vector(ypix);
}
#endif
