/*
 * strct_f.h
 */
#include "atomf.h"
#include "atom_bk.h"

typedef struct {double x,y,z;} XYZ;	/* coordinate */
typedef struct {double r11,r21,r31,r12,r22,r32,r13,r23,r33,t1,t2,t3;} SYM_OP;
									/* The values of r_ij are 1, 0 or -1. */
typedef struct {
	char element[5];	/* element name (capital) */
	int bf,af;
	/*
	 * bf: B-factor model.
	 *     bf=0 ==> isotropic B-factor
	 *     bf=1 ==> anisotropic
	 * af: Atomic scattering factor and absorption model.
	 *     af=0 ==> Mott formula + inela BK
	 *     af=1 ==> Bird & King
	 */
	double x, y, z, mul, b, b11, b22, b33, b12, b13, b23, occ;
	/*
	 * x,y,z: fractional coordinate
	 * mul: multiplicity
	 * b:	isotropic B-factor
	 * bij: anisotropic B-factor (beta_ij)
	 * occ:	occupancy
	 */
} ATOM;

#if 0
typedef struct {	/* not in use at the moment */
	double **r;	/* constraint: rotation part */
	double *t;	/* constraint: translation part */
} CST_ATOM;
#endif

typedef struct {
	double	re, im,	ela_re, ela_im, 
			*fe, *fi;	/* (*) ==> i=0 to n_atom-1 */
	/*
	 * lof_type=1 ==> re, im: Fg
	 * lof_type=2 ==> ela_re, ela_im: elastic Fg
	 * lof_type=3 ==> elastic f and inela. f: *fe, *fi (i=0 to n_atom-1)
	 * lof_type=4 ==> elastic f: *fe (i=0 to n_atom-1)
	 */
} LO_FG;

/* -------------------------------------------------- */

complex fg_to_vg(complex fg, CELL_PAR bv);
complex vg_to_fg(complex vg, CELL_PAR bv);
complex fg_to_ug(complex fg, double acc_voltage, CELL_PAR bv);
complex ug_to_fg(complex ug, double acc_voltage, CELL_PAR bv);
double fg_to_xig(complex fg, double acc_voltage, CELL_PAR bv);
double fg_to_xig2(
	complex fg, double relativistic_correction, double wave_length, CELL_PAR bv
);
double vg_to_xig(
	complex vg, double relativistic_correction, double wave_length
);
complex vg_to_ug( complex vg, double relativistic_correction );
complex vg_to_ug_2( complex vg, double accel_kvol );
complex ug_to_vg( complex vg, double relativistic_correction );

/* ----- */
SYM_OP *read_sp_group_tbl( int num_sp_group,
			int *num_sym_op, int *inversion, char *lattice_type );
SYM_OP decode_sym_op(char *str);
void read_latt_type( int num_sp_group, char *lattice_type);
void newpos_sym_op(
	int n_atom, ATOM *atm, int *num_nthatm, XYZ **pos,
	int num_sym_op, SYM_OP *sym_op
);
int search_g_table(int n_gtable, HKL g, HKL *gtable );
void atom_multiplicity(
		int n_atom, ATOM *atm, int num_sym_op, SYM_OP *sym_op
);

int ij_to_i(int ij);
double mul_aniso_bij(int ij, ATOM *atm);
/* void cst_aniso_bij(ATOM *atm); */

/* ----- */
void stock_fg_tbl_sym(
	HKL g, complex fg, int *ct_g_table,
	int g_table_size, HKL *g_table, complex *fg_table,
	int num_sym_op, SYM_OP *sym_op
);
complex structure_factor_ext(
	HKL g, int num_sym_op, SYM_OP sym_op[],
	int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[],
	double kvolts, CELL_PAR bv, char *lattice_type
);
complex structure_factor_ext_tbl(
	HKL g, int num_sym_op, SYM_OP sym_op[],
	int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[],
	double kvolts, CELL_PAR bv, char *lattice_type,
	int *ct_fg_table, int g_table_size, HKL *g_table, complex *fg_table
);
complex calc_z_factor(
	HKL g, int num_sym_op, SYM_OP sym_op[],
	int n_atom, ATOM atm[],	XDT xdyl[], double kvolts, CELL_PAR bv, char *lattice_type
);
complex calc_structure_factor(
	HKL g, int num_sym_op, SYM_OP sym_op[],
	int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[], double kvolts, CELL_PAR bv
);
double calc_F_0_re(
	int num_sym_op, int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[], CELL_PAR bv
);
complex calc_structure_factor_f_out(
	FILE *fo, HKL g, int num_sym_op, SYM_OP sym_op[],
	int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[], double kvolts, CELL_PAR bv
);
complex calc_Fg_ela(
	HKL g, int num_sym_op, SYM_OP sym_op[],
	int n_atom, ATOM atm[], XDT xdyl[], ATOMF_BK ftbl[], double kvolts, CELL_PAR bv
);
complex calc_Fg_inela(
	HKL g, int num_sym_op, SYM_OP sym_op[],
	int n_atom, ATOM atm[], ATOMF_BK ftbl[], double kvolts, CELL_PAR bv
);
/* ----- */
complex lofg_to_fg(
	HKL g, int lof_type, LO_FG *Fg_lo, int num_sym_op, SYM_OP sym_op[],
	int n_atom, ATOM atm[], ATOMF_BK ftbl[], XDT *xdyl, double kvolts, CELL_PAR bv
);
/*
void calc_cst_lofg(
	int n_atom, int num_low_order_g, int lof_type,
	HKL *g_lo, LO_FG fg_lo[], CST_LO_FG *cl
);
*/
/* ----  */
complex calc_Xray_Fg_ela(
	HKL g, int num_sym_op, SYM_OP sym_op[],
	int n_atom, ATOM atm[], XDT xdyl[], double kvolts, CELL_PAR bv
);
