#ifndef TCrystal_h
#define TCrystal_h

#include <stdio.h>

#include "TWindow.h"
#include "TAtom.h"
#include "TMatrix.h"
#include "TVector.h"
#include "CSelection.h"


#include "PhasegratingRecord.h"
#include "ProjCellRecord.h"
#include "symops.h"
#include "TImage.h"

#ifndef __RefinementGroup
#define __RefinementGroup
typedef struct RefinementGroup {
	long numInGroup ;
	long *atomIndex ;
	unsigned char *refine ;
} RefinementGroup ;

typedef struct RefinementList {
	long numGroups ;
	RefinementGroup *groupPtr ;
} RefinementList ;

#endif


typedef struct RMatrix { 
	float m[3][3] ;
} RMatrix ;
typedef RMatrix *RMatrixPtr ;
typedef struct SMatrix { 
	short m[3][3] ;
} SMatrix ;
typedef SMatrix *SMatrixPtr ;
typedef struct LMatrix { 
	long m[3][3] ;
} LMatrix ;
typedef LMatrix *LMatrixPtr ;


class TCrystal {
	public:
				
		TCrystal(TWindow *window,Rect &theDisplayRect, NucellPtr nucell) ;
		~TCrystal(void) ;
		void	Initialize( NucellPtr nucell) ;
		void	DrawCrystal() ;
		void	DrawCrystal(TVector &zone) ;
		void 	Print() ;
		void 	Rotate(void) ;
		void 	Invalidate(void)  ;
		void	ChangeMagnification(float factor)  ;
		void	SetMagnification(float magnification) ;
		void	PanView(Point &loc) ;
		void	ResetView() ;
		void	SizeAtoms(Boolean reduceSize) ;
		void	CreateRefinementList(RefinementList *list) ;
		void	UseRefinementList(RefinementList *list) ;
		void	RedisplayShiftedAtoms() ;
		void	SetSelection(CSelection *selection) { mSelection = selection ; } ;
		void	SetSymmops(short numOrigSymops,GENERATOR **symop,RMatrixPtr transfMatrix) ;
		void	SetSymmops(short spacegroupNumber,Boolean secondSetting,RMatrixPtr transfMatrix) ;
		Boolean		mDrawOutline ;
		Boolean		mDrawOrigin;
		
		SAtom		mFirstAtom ;
		SAtom		mSecondAtom ;
		void 		MeasureDistance(const Point& where) ;
		void 		MeasureAngle(const Point& where) ;
		void		Inquire(const Point& where) ;
		void		AddAtomToBeRefined(const Point& where) ;
		void		RemoveAtomToBeRefined(const Point& where) ;	
		void		DeselectAllAtomsToBeRefined() ;	
		void		AddAtomsToBeRefined(FRect &rect) ;
		void		AddAtomsToBeRefined(Rect &rect) ;
		void		AddGroupedAtomsToBeRefined(Rect &rect) ;		
		void		ShowSelectedAtom(TAtom *atom) ;
		TAtom		*GetAtom(short ndx) { return (mAtomPtr + ndx) ; } ;
		void 		SwitchSymmetryView(Boolean useSymmetry) ;
		void		UseSymmetries(Boolean state) { mUseSymmetries = state ;SwitchSymmetryView(state) ; } ;
		void		SetRefinementState(short index,unsigned char state) { (mAtomPtr+index)->refine = state ; }  ; 
		unsigned 	char	GetRefinementState(short index) { return ((mAtomPtr+index)->refine) ; }  ; 
	private:
	
		void	WriteInfo(char *) ;
		void	Draw() ;
		void	ApplySymmops() ;
		void	Assert(void *ptr) ;
		void	SortScreenAtoms(void) ;
		void 	AddViewAtom(float x,float y, float z,float TZ0,float scale,short i,float rad,RGBColor *color) ;
		void 	SetTMatrix() ;
		void 	SetSMatrix() ;
		void 	SetDisRotMatrix(TypeRotAxis whichAxis,float angle) ;
		void 	Draw2DScreenBalls(void) ;
		void 	SetUCellFaces(void) ;
		void	DrawInvisibleUCellFaces(void) ;
		void	DrawVisibleUCellFaces(void) ;
		void 	DrawAtom(SAtom *atom) ;
		Boolean IsFaceVisible(UCellFace face) ;
		void 	DrawFace(UCellFace face) ;
		void 	SortFaces(void) ;
		Boolean FindAtom(Point& mouseLoc,SAtom *whichAtom) ;	
		SAtom	*GetAtom(Point& mouseLoc) ;
		long 	CalculateNumFaceAtoms(void) ;
		void 	CreateView(short na,short nb,short nc) ;
		void	UpdateViewAtoms() ;
		void	SetTransformationMatrix(TVector &view) ;
		void	CalculateRealZoneFromReciprocal(TVector uvw) ; 
		void 	RecipToRlVector(TVector &rcpVec,TVector &rlVec) ;
		void 	RlToRecipVector(TVector &rcpVec,TVector &rlVec) ;
		float 	LengthOfRecipVector(TVector &vec) ;
		float 	LengthOfRealVector(TVector &vec) ;
		void 	RecipToCartVector(TVector &rcVec,TVector &cartVec) ;		// new
		void	SetScale() ;
		void	SetCenter(void) ;
		void	CreateUnitCellViewAtoms() ;
		void	DrawUnitCell() ;
		void	DrawOrigin();  // 9/96
		void	LoadSpacegroup(char *spcgrp,short number) ;
		Point 	VectorToPoint(TVector inV);
		void	TransformGenerator(GENERATOR *symop) ;
		Boolean	NewSymop(GENERATOR *p_symop,long num_symop) ;
		Boolean Equivalent(TAtom *atomList,float x,float y,float z,int icount,int ndx) ;
		void	HiliteAtom(TAtom *theAtom, Point AtomPt);  // bjf 12/96
		void	SelectThisAtom(float x,float y,float z,float TZ0,float scale,TAtom *theAtom) ;		
		short	mAtomsInUCell ;	/* Number of atoms in the cell	*/
		float 	mA ;				/* Unit cell axis in Angstrom	*/
		float 	mB ;				/* Unit cell axis in Angstrom	*/
		float 	mC ;				/* Unit cell axis in Angstrom	*/
		float 	mAlpha ;			/* Unit cell angle in degrees	*/
		float 	mBeta ;			/* Unit cell angle in degrees	*/
		float 	mGamma ;			/* Unit cell angle in degrees	*/
		TVector mZone ;			/* Current Viewing Direction	*/
		
		Boolean		mUseSymmetries ;
		long		mAllocatedAtoms ;
		
		float		mLightX ;
		float		mLightY ;
		float		mLightZ ;
		short		mXViewCells ;
		short		mYViewCells ;
		short		mZViewCells ;
		short		mNumFaceAtoms ;
		int			mShiftX ;
		int			mShiftY ;
		int			mCurrentGroup ;
		
		long		mSnumAtoms ;
		long		mNumViewAtoms ;

		AtomDef		mWAtom ;
		AnglePt		mAnglePt ;
		PlanePt		mPlanePt ;
		Point		mFirstPt ;
		Point		mSecondPt ;
		
		float		mMagnification ;
		float		mAtomScaleFactor ;
				
		UCellFace	mUCellFace[6] ;
		XSection	mXSection ;
		
		TWindow		*mDocWindow ;
		Rect		mDisplayRect ;
		short		mWidth ;
		short		mHeight ;
		
		CSelection	*mSelection ;
		short		mNumBasisAtoms ;
		short		mNumUnitCellAtoms ;
		short		mNumSymOps ;
		
		TAtom  	*mAtomPtr ;				/* Pointer to the record of 	*/
										/* atoms in the unit cell		*/
		TAtom	*mBasisPtr ;
		TMatrix	*mSymOps ;				/* Holds the symetry operators for this "unit cell"	*/
		SAtom	*mScreenAtom ;			/* Pointer to a list of screen atoms	*/
		TMatrix mXyzToScreenMatrix ;	/* Transformation matrix that 	*/
										/* transforms coordinates in	*/
										/* unit cell coordinates to the	*/
										/* coordinate system that is 	*/
										/* current at a particular time	*/
										/* This would usually be that 	*/
										/* of the interface				*/
		TMatrix mRotMatrix ;			/* Current Rotation Matrix		*/
		TMatrix mTmatrix ;				/* Matrix for transforming between the unit cell 	*/
										/* and the local orthonormal coordinates system  	*/
										/* for the crystal (a,b,c) = T(xl,yl,zl)			*/
		TMatrix	mSmatrix ;				/* (a,b,c) = S (ast,bst,cst) 	for the crystal 	*/
		TMatrix mTst ;					/* The equivalent of T, but for the reciprocal	 	*/
										/* lattice.					 						*/
		TMatrix	mT ;	// The matrix that takes the old axes into the new axes of the zone
		TMatrix	mU ;	// U is just the transpose of the inverse of T
		TVector mO ;	// mO is the origin of the new unit cell if shifted from the original		
} ;

#endif // TCrystal_h

