#ifndef CTable_h
#define CTable_h
#include <Controls.h>
#include <Events.h>
//#include <Windows.h>
typedef short int	Int16 ;
typedef long		Int32 ;
typedef unsigned long	Uint32 ;
typedef Int16	ResIDT ;
typedef	long	TableReference ;
#include <LVariableArray.h>

typedef struct {
	short row ;
	short col ;
} TableCellT ;


//typedef short int Int16 ;
//typedef int	Int32 ;


typedef struct {
	Int16	nRows ;
	Int16	nCols ;
	Int16	DataSize ;
	Int16	ColWidth ;
	Int16	RowHeight ;
	Int16	WindowWidth ;
	Int16	WindowHeight ;
	Boolean	AdornColumns ;
	Boolean	AdornRows ;
	Boolean	HasColumnLabels ;
	Boolean	HasRowLabels ;
	char	*WindowTitle ;
}	TableAttributes ;

class CTable {

public:
	CTable(Int16 rows,Int16 cols,Int16 cellSize,Int16 colWidth,Int16 rowHeight,char *tableTitle) ;
	CTable(TableAttributes *attributes) ;
	CTable(void) ;
	~CTable() ;
	virtual void 	SetRowHeight(Int16 height) ;
	virtual void 	SetColWidth(Int16 width) ;
	virtual void 	Draw() ;
	virtual void 	Show() ;
	virtual void	Print() ;

	virtual void	SetTableTitle(char *title) ;
	virtual void	GetTableTitle(char *title) ;
	virtual void 	SetCellData(TableCellT &inCell,void *data) ;
	virtual Boolean FetchCellData(TableCellT &inCell,void *data) ;
	virtual void 	SetDataCells(TableCellT &inFromCell, TableCellT &inToCell,void *data) ; 
	virtual void 	FetchDataCells(TableCellT &inFromCell, TableCellT &inToCell,void *data) ; 
	virtual void 	ClickCell(TableCellT &inCell) ;
	virtual void 	EditCell(TableCellT &inCell) ;
	virtual void 	DoClick(EventRecord *inEvent) ;
	virtual	void 	Update() ;
	virtual	void 	DoGrow(EventRecord *inEvent) ;
	virtual void	SetColLabel(Int16 inCol,char *title) ;
	virtual void	SetRowLabel(Int16 inRow,char *title) ;
	inline friend  	WindowPtr	GetWindowPtr(CTable	*table) { return table->mWindow ; }  ;
	
protected:

	virtual	void 	DrawCellValue(TableCellT &inCell,Rect &cellFrame) ;
	virtual void 	DrawColLabel(Int16 colNumber) ;
	virtual void 	DrawRowLabel(Int16 rowNumber) ;
	virtual void 	DrawCell(TableCellT &inCell) ;
	virtual void 	ScrollTableBy(Int16 deltaX,Int16 deltaY) ;
	virtual void 	HighlightCell(TableCellT &inCell) ;
	virtual	void 	Refresh() ;
	virtual void	InitTable() ;	
	virtual void	DrawRowLabelText(Int16 rowNumber) ;
	virtual void	DrawColLabelText(Int16 colNumber) ;
	
	Boolean ImageRectIntersectsFrame(Int16 left,Int16 top,Int16 right,Int16 bottom) ;
	void 	ImageToLocalPoint(Point &inImagePt,Point &outLocalPt) ;
	void 	LocalToImagePoint(Point &inImagePt,Point &outLocalPt) ;
	void	FetchCellHitBy(const Point &inImagePt,TableCellT &outCell) ;
	void	InvalPortRgn(RgnHandle inUpdateRgn) ;
	Boolean IsValidCell(const TableCellT	&inCell) const ;
	inline 	RgnHandle	GetLocalUpdateRgn() { return mUpdateRgnH ; } ;
	Boolean FetchLocalCellFrame(const TableCellT &inCell,Rect &outCellFrame) ;
	static pascal 	void DoHScrollBar(ControlHandle control,short partCode) ;
	static pascal 	void DoVScrollBar(ControlHandle control,short partCode) ;
	
	ControlActionUPP mHScrollBarActionProc ;
	ControlActionUPP mVScrollBarActionProc ;
	WindowPtr		mWindow ;
	ControlHandle 	mVertScrollH ;
	ControlHandle 	mHorizScrollH ;
	Boolean			mHasColLabels ;
	Boolean			mHasRowLabels ;
	Boolean			mAdornRows ;
	Boolean			mAdornColumns ;
	Pattern			mAdornPattern;
	RgnHandle		mUpdateRgnH;
	Point			mImageOrigin ;			// The origin of the table with respect to 
											// the window.  Local Point is Port point.
											// Image Point is with respect to the table origin
	Rect			mRevealedRect ;
	void			ScrollBits(Int16 inLeftDelta,Int16 inTopDelta) ;
	
	// these are arrays of 16 characters (max of labels)
	LArray			mRowLabels;
	LArray			mColLabels;
	
	LArray			*mCellData;
	
	Int16			mRows ;
	Int16			mCols ;
	Int16			mColWidth ;
	Int16			mRowHeight ;
	Int16			mCellDataSize ;
	TableCellT		mSelectedCell ;
	long			mClickTime ;
	long			mLastClick ;
} ;

#endif // CTable_h

