// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "globals.h"
#include "funcdefs.h"
#include "files.h"
#include "header.h"
#include "PhasegratingRecord.h"
#include "ProjCellRecord.h"
#include "MslicRecord.h"
#include "displ.h"
#include "filenames.h"
#include "dfpdspl.h"
#include "errors.h"
#include "storage.h"
#include "devices.h"
extern 	struct 	{ float wavel ; long nslice ; float sigma ; } gMSRec ;
extern 	struct { 	
	long abs_scale ;
	float scale_min,scale_max ; 
} mapping ;
extern 	char 	stext[] ;
#define fabs(x) 	((x) < 0. ? -(x) : (x))
#define min(x,y)  	((x) < (y) ? (x) : (y)) 
#define max(x,y)  	((x) > (y) ? (x) : (y))

static  long	lWhichImage,lNumImages ;
// --------------------------------------------------------------------------
// Main routine that gets invoked when the image(s) gets computed.
// Prior to this routine, the exit wavefunction has been calculated 
// and this routine calculates the images by taking into account 
// the effect of the lens on the scattered electrons
// --------------------------------------------------------------------------
extern Boolean gNoGUI ;
void CalculateImages(long *nobeam,float *sum,float *range,long *ierr) {

    float *out,*zphi,thickness,cs,div,time,*wrkspc,fact ;
    long hist,jmax,ndefocus,nthick,num_image,j,i,nb,maxReflections,
	NeededToAllocate ;
	
    // ----------------------------------------------- 
    // Set the parameters for the lens and the display 
    // ----------------------------------------------- 
    if (!gNoGUI) SetWatchCursor() ;
    SetLensParameters(&gDisplayRec.cs,&gDisplayRec.div,&cs,&div,
		&gPhsgrtRec->gmax,&gPhsgrtRec->nbeam,&gPhsgrtRec->mbeam,
           	&gPhsgrtRec->nz) ;

    if (gIOError) return ;
    jmax = gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;
    zphi = gWorkSpace ;
    out = zphi + 2*jmax ;
    fact = 0.25*gDisplayRec.aprtrd*gDisplayRec.aprtrd/(gPhsgrtRec->gmax*gPhsgrtRec->gmax) ;
    maxReflections = jmax*fact*1.1 ; /* Leave some room for round-off error */
	if (maxReflections > (gMaxArraySize-4*jmax)/3) {
		NeededToAllocate = 1 ;
		wrkspc = (float*)malloc(3*maxReflections*sizeof(float)) ;
		if (wrkspc == (float *)NULL) {
			erase_message() ;
			write_message("Failed to Allocate memory !") ;
			gAbortStatus = 1 ;
			return ;
		}
    }
    else {
		NeededToAllocate = 0 ;
		wrkspc = out+2*jmax ;
    }
	
    if (!gNoGUI) SetWatchCursor() ;

    /* sort out the number of thicknesses to be calculated	*/
    nthick = 1 ;
    if (gMsliceRec.tts)		/* If thru-thickness series	*/
		nthick = (gMsliceRec.t2-gMsliceRec.t1)/gMsliceRec.dt + 1 ;
    else
		gMsliceRec.t2 = gMsliceRec.thick ;	/* Single thickness	*/
	
    ndefocus = 1 ;
    if (gDisplayRec.tfs)		/* If thru-defocus series	*/
		ndefocus = fabs(gDisplayRec.f2-gDisplayRec.f1)/fabs(gDisplayRec.df) + 1 ;
    else
		gDisplayRec.f1 = gDisplayRec.dfocus ;	/* Single defocus	*/
		
    jmax = gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;
	
	if(!open_input_file(flname.mout)) {
		gIOError = ERR_OP ;
		if (NeededToAllocate) {
			free(wrkspc) ;
		}
		if (!gNoGUI) InitCursor() ;
		return ;
	}

    if (!open_output_file(flname.iout)) {
		gIOError = ERR_OP ;
		if (NeededToAllocate) {
			free(wrkspc) ;
		}
		if (!gNoGUI) InitCursor() ;
		return ;
    }
		
    num_image = 0 ;
    thickness = gMsliceRec.thick ;

    /* Now calculate the images */
    for(i=0;i<nthick;i++) {
		if (gMsliceRec.tts) 
			thickness = gMsliceRec.t1 + i*gMsliceRec.dt ;
		erase_message() ;
		write_message("Reading Input File...") ;
		if(!read_input_file(i+1,(char *)zphi,2*jmax)) {
			gIOError = ERR_RD ;
			if (NeededToAllocate) {
				free(wrkspc) ;
			}
			erase_message() ;
			if (!gNoGUI) InitCursor() ;
			return ;
		}
		erase_message() ;
		mmfft(zphi,gPhsgrtRec->nbeam,gPhsgrtRec->mbeam,&time) ;  
		// Get the diffraction amplitudes and store
	    // The exit wavefunction amplitudes and phases are stored since
		// they don't have to be computed again for another image at the
		// same thickness, but at a different defocus	

		erase_message() ; // for printing the status line
		lNumImages = ndefocus*nthick ;
		for(j=0;j<ndefocus;j++) {
			lWhichImage = i*ndefocus + j + 1 ;
			sprintf(stext,
				"STATUS:  ** IMAGE ** Calculating Image # %d of %d",lWhichImage,lNumImages) ;
			write_message(stext) ;

			if (gDisplayRec.tfs) gDisplayRec.dfocus = gDisplayRec.f1+j*gDisplayRec.df ;
			// ------------------------------------------------
			// Take into account the effect of the lens  
			// ------------------------------------------------ 
			ApplyObjectiveLens(zphi,out,cs,div,gPhsgrtRec->nbeam,gPhsgrtRec->mbeam,jmax,nobeam,
				&nb,sum,wrkspc,maxReflections) ;
			// Set the image intensity 
			SetImageIntensities(out,jmax) ;

			// ------------------------------------------------
			// Write the image out to the file "IOUT"    
			// ------------------------------------------------ 
			if(!write_output_file((char *)out,jmax)) {
				gIOError = ERR_WR ;
				if (NeededToAllocate) {
					free(wrkspc) ;
				}
				if (!gNoGUI) InitCursor() ;
				return ;
			}
		}
    }
    close_output_file() ;
    close_input_file() ;
    if (!gNoGUI) InitCursor() ;
    if (NeededToAllocate) {
		free(wrkspc) ;
    }
    erase_message() ;
}

// --------------------------------------------------------------------------
// This routine doesn't really do much. It just sets up some things
// --------------------------------------------------------------------------
void SetLensParameters(float *cs1,float *div1,float *cs,float *div,float *gmax,
		long *nbeam,long *mbeam,long *nz) {
    long ierr ;
    float ewavel,evolt ;
  
    *cs = *cs1*1.0e+7 ;
    *div = *div1*1.0e-3 ;
    if(*nbeam == 0) {
		SetProjectedCell(&ierr) ;
		SetCalculationDimensions(&gProjCell->a,&gProjCell->b,gmax,nbeam,mbeam) ;
    }
    evolt = gMsliceRec.volt ;
    ewavel = 1.2263652e-7/sqrt((evolt+9.785025e-4*evolt*evolt)*1000.) ;
    gMSRec.sigma = 2.08886e+6*ewavel*sqrt(1.0+5.886579e-20/(ewavel*ewavel)) ;
    gMSRec.wavel = ewavel*1.0e+8 ;
}

// --------------------------------------------------------------------------
// Because of the way the calculation has been changed, the image
// intensities are already computed as a result of taking the
// inverse FFT of the image frequency spectrum. Thus we only need 
// to arrange the array so to get rid of the complex part (0.)
// --------------------------------------------------------------------------
void SetImageIntensities(float *out,long jmax) {

    register long i ;
    register float *rp1,*rp2 ;
	
    /* Just condense the complex array (x,0) long a real array (x) */
    rp1 = out ;
    rp2 = out ;
    for(i=0;i<jmax;i++,rp2++) 
    	*rp1++ = *rp2++ ;
}

// --------------------------------------------------------------------------
// scale the image in the range (0 - *prange) if automatic scaling
// If black/white is set (absolute scaling), scale the image in 
// the range (min - max) set by the values in 'mapping'
// --------------------------------------------------------------------------
void ScaleImageIntensities(long *pnp,long *pmp,float *prange,long *ierr) {
    long 	np,mp ;
    float 	range ;
    register 	long j ;
    register 	float a,b,*rp ;
    register 	float vmax,vmin ;
    long 	jmax ;
    float 	span ;
			
			
    np = *pnp ;
    mp = *pmp ;
    range = *prange ;
    vmin = 1.0e+20 ;
    vmax = -1.0e+20 ;
    jmax = np*mp ;
    if (mapping.abs_scale) {
		vmin = mapping.scale_min ;
		vmax = mapping.scale_max ;
    }
    else {
		for(j=0,rp=gWorkSpace;j<jmax;j++,rp++) {
			vmax = max(*rp,vmax) ;
			vmin = min(*rp,vmin) ;
		}
    }
    mapping.scale_min = vmin ;
    mapping.scale_max = vmax ;
    span = vmax-vmin ;
    if (span == 0.) {
		*ierr = -1 ;
		return ;
    }
	
    a = range/span ;
    b = -vmin*a ;
    for(j=0,rp=gWorkSpace;j<jmax;j++,rp++) {
		*rp = *rp*a + b ;
		if (*rp < 0.) *rp = 0. ;
		if (*rp > range) *rp = range ;
    }
}

// --------------------------------------------------------------------------
// Adjust the image according to the contrast and brightness levels	
// that have been set by the user.		
// --------------------------------------------------------------------------
void contrst(long *pnp,long *pmp,float *pcntrst,float *prange,float *pbrite) {

    long 	np,mp,index ;
    float 	cntrst,range,brite ;
    register 	long j,jmax ;
    register 	float *rp,b,a,frange ;
	
    np = *pnp ;
    mp = *pmp ;
    cntrst = *pcntrst ;
    range = *prange ;
    brite = *pbrite ;
    a = cntrst ;
    jmax = np*mp ;
    if (a >= 0.)
		b = (1.+a)*brite*range - a*range ;
    else
		b = (1.-a)*brite*range ;
    for(j=0,rp=gWorkSpace;j<jmax;j++,rp++) {
		*rp = a*(*rp) + b ;
		if (*rp > range) *rp = range ;
		if (*rp < 0. ) *rp = 0. ;
		if (OutDevice->type == WS) {
			index = *rp ;
			*rp = gLut[index] ;
		}
    }
}

// --------------------------------------------------------------------------
// Perform a histogram equalization on the image
// --------------------------------------------------------------------------
void histeq(long *pnp,long *pmp) {

    long 	np,mp ;
    register 	long jmax,i,k,j ;
    register 	float *rp ;
    long 	ip[256],ivalue,nlevls,it ;
    extern 	float RANGE ;
    float 	fact,t[256] ;
	
    np = *pnp ;
    mp = *pmp ;
    ivalue = RANGE ;      	/* Highest intensity value allowed */
    nlevls = ivalue + 1 ; 	/* Number of different grey levels */
    jmax = np*mp ;
    fact = 1./jmax ;
    for(i=0;i<nlevls;ip[i++]=0)
		;
	
    // --------- Find the probability function (unscaled) ------ 
    for(j=0,rp=gWorkSpace;j<jmax;j++,rp++) {
		i = (long)(*rp+0.5) ;  	/* Nearest intensity level */
		ip[i]++ ;		/* Increment the count for this intensity */
	}
	
    // ------ Find the cumulative distribution function ------------ 
    for(k=0;k<nlevls;k++) {
		t[k] = 0. ;
		for(i=0;i<k;t[k] += ip[i++])
			;
		t[k] *= fact ;
	}
	
	
    // ----------- Find the transformation from old to new ---------------
    for(k=0;k<nlevls;k++) {
		t[k] *= nlevls ;       		// Find the nearest integer from 
		it = (long)(t[k]+0.5) ; 		// 0 - ivalue                  
		t[k] = min(it,ivalue) ;		// Make sure to not exceed ivalue
	}
	
    // -------------- Do the actual transformation on the data ----------- *
    for(j=0,rp=gWorkSpace;j<jmax;j++,rp++) {
		i = (long)(*rp + 0.5) ;
		*rp = t[i] ;
	}
}

float cxa1,cxa2,cxd ;
// --------------------------------------------------------------------------
// this is the workhorse in the image routine. The function of this
// routine is to simulate the action of the objective lens on the
// diffracted electrons. Thus any attempt to optimize the code 
// should be performed on this routine. There is also the routine
// that is called by this one, which sits in a loop and could 
// be optimized.							
// --------------------------------------------------------------------------
void ApplyObjectiveLens(float *zphi,float *ihk,float cs,float theta,long nbeam,long mbeam,
	long jmax,long *nobeam,long *num_beam,float *sum,float *wrkspc,long maxRefl) {

    float aprtsq,abga,testint,astsq,bstsq,test,th,tk,*rp1,
          *rp2,*tp1,*tp2,zrl,zim ;
    float hcapert,kcapert,hoaxis,koaxis,one_ov_jmax,*zfi,
    	  zrl1,zrl2,zim1,zim2 ;
    float c02,c04,arg,usq,cnst,usqmax ;
    float hpr,kpr,hprh,kprk,u1sq,u2sq,u1frth,u2frth,arg1,arg2,h1,k1,h2,k2 ;
    long nh,h,nk,k,i,j,hmax,kmax,mem_size ;
    short *ip1,*ip2,*ih,*ik ;
    float pi=3.14159265,xcoeff,time,vibcnst ;
	
    aprtsq = gDisplayRec.aprtrd*gDisplayRec.aprtrd ;
    abga = -2.*gProjCell->ast*gProjCell->bst*gProjCell->csga ;
    c02 = pi*gMSRec.wavel*gDisplayRec.dfocus ;
    c04 = 0.5*pi*1.e+7*gDisplayRec.cs*gMSRec.wavel*gMSRec.wavel*gMSRec.wavel ;
    cxd = -0.5*pi*pi*gMSRec.wavel*gMSRec.wavel*gDisplayRec.del*gDisplayRec.del ;
    cxa1 = gDisplayRec.dfocus*pi*1.e-3*gDisplayRec.div*0.77 ;
    cxa2 = 1.e+7*gDisplayRec.cs*gMSRec.wavel*gMSRec.wavel*pi*1.e-3*gDisplayRec.div*0.77 ;

    testint = pow(10.,-dfpdspl.min_int) ;
	
    hmax = nbeam/2 ;
    kmax = mbeam/2 ;
    *sum = 0. ;
    *nobeam = 0 ;
    *num_beam = 0 ;
    astsq = gProjCell->ast*gProjCell->ast ;
    bstsq = gProjCell->bst*gProjCell->bst ;
    hcapert = gDisplayRec.hc ;
    kcapert = gDisplayRec.kc ;
    hoaxis = gDisplayRec.ho ;
    koaxis = gDisplayRec.ko ;
    one_ov_jmax = 1./jmax ;
    tp1 = &zrl ;
    tp2 = &zim ;

    vibcnst = -2.*pi*pi*gDisplayRec.vibration*gDisplayRec.vibration ;
    
    zfi = wrkspc ;
    mem_size = 2*maxRefl ;
    zfi += mem_size ;
    ih = (short *)zfi ;
    mem_size = maxRefl ;
    ik = ih + mem_size ;
    zfi = wrkspc ;

	
    // ------------------------------------------------------------------- 
    // Fill the array zfi with intensities > testint within the aperture 
    // -------------------------------------------------------------------
    rp2 = zfi ;
    ip1 = ih ;
    ip2 = ik ;
    sprintf(stext,"Finding Reflections with intensities greater than %e !",
	testint) ;
    erase_message() ;
    write_message(stext) ;
    for(nh=0;nh<nbeam;nh++) {
		h = nh ;
		if (h>hmax) h -= nbeam ;
		th = h - hcapert ;
		for(nk=0;nk<mbeam;nk++) {
			k = nk ;
			if (k > kmax) k -= mbeam ;
			tk = k - kcapert ;
			test = th*th*astsq+tk*tk*bstsq+th*tk*abga ;
			if (test <= aprtsq) {
				rp1 = zphi + 2*(nk*nbeam+nh) ;
				*tp1 = *(rp1++)*one_ov_jmax ;
				*tp2 = *rp1*one_ov_jmax ;
				if (((*tp1)*(*tp1)+(*tp2)*(*tp2)) > testint) {
					*rp2++ = *tp1 ;
					*rp2++ = *tp2 ;
					*ip1++ = h ;
					*ip2++ = k ;
					(*num_beam)++ ;
				}
			}
		}
	}
	
	*nobeam = *num_beam ;
	/* Set the image intensities to zero */
	for(i=0;i<2*jmax;ihk[i++]=0.)
		;
		
	if (abandn()) {
		InitCursor() ;
		return ; 	/* Allow the user to CANCEL	*/
	}
	erase_message() ;
	
	float   fOutputInterval = *num_beam/10.0 ;
	float   fOutput = fOutputInterval ;
	long	outputInterval = near(fOutputInterval) ;
	long	percentCompleted = 10 ;
	for(i=0;i<*num_beam;i++) {
    	if (abandn()) {
			InitCursor() ;
			return ; 	/* Allow the user to CANCEL	*/
		}
		hpr = *(ih+i) ;
		kpr = *(ik+i) ;
		h1 = hpr - hoaxis ;
		k1 = kpr - koaxis ;
		u1sq = h1*h1*astsq + k1*k1*bstsq + h1*k1*abga ;
		u1frth = u1sq*u1sq ;
		arg1 = c02*u1sq + c04*u1frth ;
		rp1 = zfi + 2*i ;
		*tp1 = cos(arg1) ;
		*tp2 = -sin(arg1) ;
		zrl1 = *rp1*(*tp1) - *(rp1+1)*(*tp2) ;
		zim1 = *rp1*(*tp2) + *(rp1+1)*(*tp1) ;
		for(j=0;j<*num_beam;j++) {
			hprh = *(ih+j) ;
			kprk = *(ik+j) ;
			h = hpr - hprh ;
			k = kpr - kprk ;
			if (h < hmax && h > -hmax && k < kmax && k > -kmax) {
				h2 = hprh-hoaxis ;
				k2 = kprk-koaxis ;
				u2sq = h2*h2*astsq + k2*k2*bstsq + h2*k2*abga ;
				u2frth = u2sq*u2sq ;
				arg2 = c02*u2sq + c04*u2frth ;
				*tp1 = cos(arg2) ;
				*tp2 = -sin(arg2) ;
				rp1 = zfi + 2*j ;
				zrl2 = *rp1*(*tp1) - *(rp1+1)*(*tp2) ;
				zim2 = *rp1*(*tp2) + *(rp1+1)*(*tp1) ;
				nh = h ;
				nk = k ;
				if (nh < 0) nh += nbeam ;
				if (nk < 0) nk += mbeam ;
				zrl = zrl1*zrl2+zim1*zim2 ;
				zim = zim1*zrl2-zrl1*zim2 ;
				CrossTransferFunction(h1,k1,h2,k2,u1sq,u2sq,
				       astsq,bstsq,abga,&xcoeff) ;
				rp2 = ihk + 2*(nk*nbeam+nh) ;
				*rp2 += zrl*xcoeff ;
				*(rp2+1) += zim*xcoeff ;
			}
		}		
		if (i != 0 && i%outputInterval == 0) {
			sprintf(stext," STATUS:  Image %d (%d) Completed %d percent ",lWhichImage,lNumImages,percentCompleted) ;
			write_message(stext) ;
			fOutput += fOutputInterval ;
			outputInterval = near(fOutput) ;
			percentCompleted += 10 ;
		}
		if (percentCompleted == 100 && i == *num_beam-1) {
			sprintf(stext," STATUS:  Image %d (%d) Completed %d percent ",lWhichImage,lNumImages,percentCompleted) ;
			write_message(stext) ;
		}
	}
    *sum = ihk[0] ;
    if (gDisplayRec.vibration != 0.) { /* We need to apply vibration */
		usqmax = 4.*aprtsq ;
		for(nh=0;nh<nbeam;nh++) {
			h = nh ;
			if (h > hmax) h -= nbeam ;
			for(nk=0;nk<mbeam;nk++) {
				k = nk ;
				if (h > kmax) k -= mbeam ;
				usq = h*h*astsq+k*k*bstsq+h*k*abga ;
				if (usq < usqmax) {
					arg = vibcnst*usq ;
					rp1 = ihk + 2*(nk*nbeam+nh) ;
					rp2 = rp1+1 ;
					if (arg > -9.2)  {
						cnst = exp(arg) ;
						*rp1 *= cnst ;
						*rp2 *= cnst ;
					}
					else {
						*rp1 = 0. ;
						*rp2 = 0. ;
					}
				}
			}
		}
    }

    mmffti(ihk,nbeam,mbeam,&time) ;
}

void CrossTransferFunction(float h1,float k1,float h2,float k2,float u1sq,float u2sq,
	float astsq,float bstsq,float abga,float *xcoeff) {
    float u,v,arg1,arg2 ;
	
    u = cxa1*(h1-h2) + cxa2*(u1sq*h1-u2sq*h2) ;
    v = cxa1*(k1-k2) + cxa2*(u1sq*k1-u2sq*k2) ;
    arg1 = cxd*(u1sq-u2sq)*(u1sq-u2sq) ;
    arg2 = u*u*astsq + v*v*bstsq + u*v*abga ;
    *xcoeff = exp(arg1-arg2) ;
}
