// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "funcdefs.h"
#include "globals.h"
#include "menu_index.h"
#include "TImage.h"
#include "devices.h"
#include "TCompareImagesPage.h"
#include "TStructureRefinementAnnealingPage.h"
#include "TRefineParametersPage.h"
#include "TControlWindow.h"

void DisplayQuantitativeMenu() ;
void DrawQuantitativePage() ;
void CompareWithSimulation() ;
void ShowCompareAndRefineParameterPage() ;
void ShowCompareAndRefineStructurePage() ;
void ReturnFromQuantitativeMenu() ;
void DrawExpImageDisplayArea() ;
void DrawStdDevImageDisplayArea() ;
void DisplayImageInWindow(TImage *, TWindow *, Rect *rect,Boolean );

extern	TImage *gStandardDeviationImage ;
extern	TImage *gExperimentalImage ;

TImage	*gExpDisplayImage = nil ;
TImage	*gStdDevDisplayImage = nil ;

TImage *gPreviousStandardDeviationImage = nil ;
TImage *gPreviousExperimentalImage = nil ;

TCompareImagesPage *gCompareImagesPage = nil ;
TStructureRefinementAnnealingPage *gStructureRefinementAnnealingPage = nil ;
TRefineParametersPage *gRefineParametersPage = nil ;
void	DisplayMessage(char *text) ;

void SetPage(TPage *);

void DrawQuantitativePage() {
	displ_menu() ;
	DrawExpImageDisplayArea() ;
	DrawStdDevImageDisplayArea() ;
}

void DisplayQuantitativeMenu() {
	extern COMMAND c_compareImageWithSimulation,c_refineParameters,c_refineStructure ;
	ClearDisplay() ;
	set_menu(QUANTITATIVE) ;
	DrawQuantitativePage() ;
	if (gExperimentalImage != nil) {
		if (c_compareImageWithSimulation.disabled) {
			EnableControl(&c_compareImageWithSimulation) ;
			EnableControl(&c_refineParameters) ;
			EnableControl(&c_refineStructure) ;
		}
	}
}

void DrawExpImageDisplayArea() {
	Rect	tRect ;
	short 	x,y,width,height ;
	extern	unsigned long	gWhitePixel,gBlackPixel ;
	COMMAND	theCommand = {10,200,490,680} ;

	set_video_normal() ;
	width = (theCommand.xh - theCommand.xl)/2 ;
	height = (theCommand.yh - theCommand.yl)/2 ;
	x = theCommand.xl/2 ;
	y = Terminal.yres - theCommand.yh ;
	y /= 2 ;
	SetRect(&tRect,x,y,x+width,y+height) ;
	InsetRect(&tRect,1,1) ;
	SetFgColor(gDarkerGray) ;
	FillRect(&tRect) ;
	PenSize(2,2) ;
	SetFgColor(gBlackPixel) ;
	DrawLine(tRect.left,tRect.bottom,tRect.left,tRect.top) ;
	DrawLine(tRect.left,tRect.top,tRect.right,tRect.top) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(tRect.right,tRect.top,tRect.right,tRect.bottom) ;
	DrawLine(tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
	PenSize(1,1) ;
	SetFgColor(gBlackPixel) ;

	if (gExperimentalImage == nil) {
		SetFgColor(gWhitePixel) ;
		width = StringWidth("NO IMAGE LOADED") ;
		x = (tRect.left+tRect.right-width)/2 ;
		y = (tRect.top+tRect.bottom)/2 + 10 ;
		DrawText(x,y,"NO IMAGE LOADED") ;
		SetFgColor(gBlackPixel) ;
		XFlush(gDisplay) ;
		return ;
	}

	if (gPreviousExperimentalImage != gExperimentalImage) {
		gPreviousExperimentalImage = gExperimentalImage ;
		if (gExpDisplayImage != nil) {
			DeleteImage(gExpDisplayImage) ;
			gExpDisplayImage = nil ;
		}
	}
	if (gExpDisplayImage == nil) {
		gExpDisplayImage = CopyImage(gExperimentalImage) ;
		if (gExpDisplayImage != nil) {
			Rect	theRect = tRect ;
			InsetRect(&theRect,9,9) ;
			float	ratio1,ratio2,fraction ;
			short	width = gExpDisplayImage->mWidth ;
			short	height = gExpDisplayImage->mHeight ;
			ratio1 = theRect.right-theRect.left ;
			ratio1 /= width ;
			ratio2 = theRect.bottom-theRect.top ;
			ratio2 /= height ;
			fraction = ratio2 < ratio1 ? ratio2 : ratio1 ;
			if (fraction != 1.) {
				width *= fraction ;
				height *= fraction ;
				AdjustImageSize(&gExpDisplayImage,width,height) ;
			}
		}
	}
	if (gExpDisplayImage != nil) {
		InsetRect(&tRect,4,4) ;
		DisplayImageInWindow(gExpDisplayImage,gControlWindow,
			&tRect,false );
		InsetRect(&tRect,-4,-4) ; 
		tRect.top = tRect.bottom + 1 ;
		tRect.bottom = tRect.top+12 ;
		EraseRect(&tRect) ;
		x = tRect.left ;
		y = tRect.bottom - 2 ;
		DrawText(x,y,gExperimentalImage->mName) ;
		XFlush(gDisplay) ;
	}
}

void DrawStdDevImageDisplayArea() {
	Rect	tRect ;
	short 	x,y,width,height ;
	extern	unsigned long	gWhitePixel,gBlackPixel ;
	COMMAND	theCommand = {510,200,990,680} ;

	set_video_normal() ;
	width = (theCommand.xh - theCommand.xl)/2 ;
	height = (theCommand.yh - theCommand.yl)/2 ;
	x = theCommand.xl/2 ;
	y = Terminal.yres - theCommand.yh ;
	y /= 2 ;
	SetRect(&tRect,x,y,x+width,y+height) ;
	InsetRect(&tRect,1,1) ;
	SetFgColor(gDarkerGray) ;
	FillRect(&tRect) ;
	PenSize(2,2) ;
	SetFgColor(gBlackPixel) ;
	DrawLine(tRect.left,tRect.bottom,tRect.left,tRect.top) ;
	DrawLine(tRect.left,tRect.top,tRect.right,tRect.top) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(tRect.right,tRect.top,tRect.right,tRect.bottom) ;
	DrawLine(tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
	PenSize(1,1) ;
	SetFgColor(gBlackPixel) ;
	if (gStandardDeviationImage == nil) {
		SetFgColor(gWhitePixel) ;
		width = StringWidth("NO IMAGE LOADED") ;
		x = (tRect.left+tRect.right-width)/2 ;
		y = (tRect.top+tRect.bottom)/2 + 10 ;
		DrawText(x,y,"NO IMAGE LOADED") ;
		SetFgColor(gBlackPixel) ;
		XFlush(gDisplay) ;
		return ;
	}

	if (gPreviousStandardDeviationImage != gStandardDeviationImage) {
		gPreviousStandardDeviationImage = gStandardDeviationImage ;
		if (gStdDevDisplayImage != nil) {
			DeleteImage(gStdDevDisplayImage) ;
			gStdDevDisplayImage = nil ;
		}
	}
	if (gStdDevDisplayImage == nil) {
		gStdDevDisplayImage = CopyImage(gStandardDeviationImage) ;
		if (gStdDevDisplayImage != nil) {
			Rect	theRect = tRect ;
			InsetRect(&theRect,9,9) ;
			float	ratio1,ratio2,fraction ;
			short	width = gStdDevDisplayImage->mWidth ;
			short	height = gStdDevDisplayImage->mHeight ;
			ratio1 = theRect.right-theRect.left ;
			ratio1 /= width ;
			ratio2 = theRect.bottom-theRect.top ;
			ratio2 /= height ;
			fraction = ratio2 < ratio1 ? ratio2 : ratio1 ;
			if (fraction != 1.) {
				width *= fraction ;
				height *= fraction ;
				AdjustImageSize(&gStdDevDisplayImage,width,height) ;
			}
		}
	}
	if (gStdDevDisplayImage != nil) {
		InsetRect(&tRect,4,4) ;
		DisplayImageInWindow(gStdDevDisplayImage,gControlWindow,
			&tRect,false );
		InsetRect(&tRect,-4,-4) ; 
		tRect.top = tRect.bottom + 1 ;
		tRect.bottom = tRect.top+12 ;
		EraseRect(&tRect) ;
		x = tRect.left ;
		y = tRect.bottom - 2 ;
		DrawText(x,y,gStandardDeviationImage->mName) ;
		XFlush(gDisplay) ;
	}
}
void CompareWithSimulation() {
	if (gCompareImagesPage == nil)
		gCompareImagesPage = new TCompareImagesPage(gControlWindow) ;
	SetPage(gCompareImagesPage) ;
	gCompareImagesPage->Show() ;
}

void ShowCompareAndRefineParameterPage() {
	if (gRefineParametersPage == nil)
		gRefineParametersPage = 
			new TRefineParametersPage(gControlWindow) ;
	SetPage(gRefineParametersPage) ;
	gRefineParametersPage->Initialize() ;
	gRefineParametersPage->Show() ;
}

void ShowCompareAndRefineStructurePage() {
	if (gStructureRefinementAnnealingPage == nil)
		gStructureRefinementAnnealingPage = 
			new TStructureRefinementAnnealingPage(gControlWindow) ;
	SetPage(gStructureRefinementAnnealingPage) ;
	gStructureRefinementAnnealingPage->Show() ;
}

void ReturnFromQuantitativeMenu() {
	ClearDisplay() ;
	set_menu(DSPL_SOURCE) ;
	displ_menu() ;
	set_menu(DSPL_OPER) ;
	displ_menu() ;
	set_menu(DSPL_SOURCE) ;
}
