// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include <math.h>
#include "header.h"
#include "storage.h"
#include "filenames.h"
#include "devices.h"
#include "menudef.h"
#include "globals.h"
#include "PhasegratingRecord.h"
#include "ProjCellRecord.h"
#include "funcdefs.h"
#include "graphics.h"
#define max(x,y) ((x) > (y) ? (x) : (y))
#define fabs(x) 	((x) < 0. ? -(x) : (x))
extern GDevice *OutDevice ;
extern long LAYERED ;

#include "menu_index.h"

#include "TAmplitudePage.h"
#include "TMainPage.h"
extern short gPrintValues ;
void	SetPage(TPage *) ;
extern	TMainPage *gMainPage ;
void 	run_help(char *) ;

void	ExitAmplitudePage(TAmplitudePage *page) ;

TAmplitudePage::TAmplitudePage(TWindow *window) : TPage(window) {
	TGroupBox 	*grpBox ;
	TStaticText	*theText ;
	char		text[256] ;
	long		i ;

	mHelp = new TButton(490,2,510,20,"H") ;
	AddObject(mHelp) ;

	grpBox = new TGroupBox(20,40,280,290,"Choose reflections to view");
	AddObject(grpBox) ;

	grpBox = new TGroupBox(290,40,500,290,"Options");
	AddObject(grpBox) ;

	grpBox = new TGroupBox(320,60,480,120,"Intensity / Amplitude");
	AddObject(grpBox) ;

	grpBox = new TGroupBox(320,140,480,250,"Multiple beam options");
	AddObject(grpBox) ;

	theText = new TStaticText(330,190,465,240,"If individual plots is chosen\nyou must click in the window\nto display the next reflection") ;
	AddObject(theText) ;


	mPlotIntensity = new TRadioButton(330,70,470,90,"Plot intensity");
	AddObject(mPlotIntensity) ;

	mPlotAmplitude = new TRadioButton(330,90,470,110,"Plot amplitude");
	AddObject(mPlotAmplitude) ;

	mSuperImposeBeams = new TRadioButton(330,150,470,170,"Superimpose plots");
	AddObject(mSuperImposeBeams) ;

	mIndividualBeams = new TRadioButton(330,170,470,190,"Plot individual beams");
	AddObject(mIndividualBeams) ;

	mPlotPhases = new TCheckBox(320,255,470,275,"Plot phases");
	AddObject(mPlotPhases) ;

	mOK = new TButton(405,330,495,360,"Plot");
	AddObject(mOK) ;

	mCancel = new TButton(290,330,390,360,"Return");
	AddObject(mCancel) ;

	for(i=0;i<10;i++) 
		mBeams[i] = nil ;

	FILE *fp_aout=fopen(flname.aout,"r") ;
	if(fp_aout==NULL) {
		// Handle error codes
		sprintf(text,"Unable to open the file %s",flname.aout) ;
		PutMessage(text) ;
		return ;
	}

	mZplot = gWorkSpace ;

	fread(mTitle,40*sizeof(char),1,fp_aout) ;
	fread(&mA,sizeof(float),1,fp_aout) ;
	fread(&mB,sizeof(float),1,fp_aout) ;
	fread(&mC,sizeof(float),1,fp_aout) ;
	fread(&mN,sizeof(long),1,fp_aout) ;
	fread(&mM,sizeof(long),1,fp_aout) ;
	fread(&mVolt,sizeof(float),1,fp_aout) ;
	fread(&mNumBeamsStored,sizeof(long),1,fp_aout) ;
	fread(&mNumSlices,sizeof(long),1,fp_aout) ;
	fread(&mSliceThickness,sizeof(float),1,fp_aout) ;
	fread(&mGMax,sizeof(float),1,fp_aout) ;
	fread(mHStored,sizeof(long),mNumBeamsStored,fp_aout) ;
	fread(mKStored,sizeof(long),mNumBeamsStored,fp_aout) ;
	fread(mLStored,sizeof(long),mNumBeamsStored,fp_aout) ;
	i = 2*mNumSlices*mNumBeamsStored ;   // Complex data
	fread(mZplot,sizeof(float),i,fp_aout) ;

	fclose(fp_aout) ;

	INTENSITY 	= true ;
	IMANY		= true ;
	PLOTPHS		= false ;
	mNumBeamsToPlot = 0 ;

	mPlotIntensity->SetControlValue(INTENSITY) ;
	mSuperImposeBeams->SetControlValue(IMANY) ;
	mPlotPhases->SetControlValue(PLOTPHS) ;

	for(i=0;i<mNumBeamsStored;i++) {
		sprintf(text,"%3d %3d %3d",mHStored[i],mKStored[i],mLStored[i]) ;
		mBeams[i] = new TCheckBox(30,60+i*20,200,80+i*20,text) ;
		AddObject(mBeams[i]) ;
	}

}

void	TAmplitudePage::Draw() {
	EraseRect(&mWindow->GetPortRect()) ;
	TPage::Draw() ;	
}

TAmplitudePage::~TAmplitudePage() {
}

void	TAmplitudePage::PlotBeams() {
	float 	*th,zim,*rp1,*rp2,*amp,
		*phase,ampmax[10],phsmax[10] ;
	long 	i,j,n,ndx,ndxi ;
	long 	k,ierr ;
	float 	tst,ymax,ymaxp,ymaxa ;
	float 	xmax,ymin,ymax2,size,two,ymina,yminp ;
	char 	buffer[132] ;
	char 	ixname[80],iyname[80],ident[80],
			ihkl[80],igmax[80],ivolt[80],islcth[80] ;

	th = mZplot+20000 ;
	amp = th+10000 ;
	phase = amp+10000 ;

	if (PLOTPHS) {
		/* adjust the phase as to get a continuous curve */
		two = 2. ;
		for(n=0;n<mNumBeamsStored;n++)
			for(j=1;j<mNumSlices;j++) {
				rp1 = mZplot + 2*(n*mNumBeamsStored+j)  ;	// Pointer to cmplx [n][j] 
				rp2 = rp1 - 2*mNumBeamsStored ;         	//  --- [n-1][j]        
				rp1++ ;   					// Imaginary part
				rp2++ ;   					// Imaginary part
				tst = fabs(*rp1-*rp2) ;
				if (tst >= 1.) {
					zim = *rp2 ;
					zim = (zim > 0. ? two : -two) ;
					*rp1 += zim ;
					tst = fabs(*rp1-*rp2) ;
					if (tst > 1.) *rp1 += zim ;
				}
			}
	}

	// set all the output arrays
	for(i=0;i<mNumBeamsToPlot;i++)
		for(j=0;j<mNumBeamsStored;j++)
			if (mHPlot[i]==mHStored[j] && mKPlot[i]==mKStored[j] && mLPlot[i]==mLStored[j]) {
			ampmax[i] = 0. ;
			phsmax[i] = 0. ;
			ndxi = i*1000 ;
			for(n=0;n<mNumSlices;n++) {
				ndx = ndxi+n ;
				rp1 = mZplot + 2*(n*mNumBeamsStored+j) ;
				if (PLOTPHS) {
					 *(phase+ndx) = *(rp1+1) ;
					 phsmax[i] = max(phsmax[i],fabs(*(phase+ndx))) ;
				}
				*(amp+ndx) = *rp1 ;
				if (INTENSITY) {
					*(amp+ndx) *= *(amp+ndx) ;
				}
				ampmax[i] = max(ampmax[i],*(amp+ndx)) ;
			}
			if (PLOTPHS) phsmax[i] *= 1.2 ;
			ampmax[i] *= 1.2 ;
	}
	// plot the arrays
	xmax = mNumSlices*mSliceThickness  ;

	eushft('!') ;
	sprintf(ident,"%s",mTitle) ;
	sprintf(ivolt,"V = %7.1f kV",mVolt) ;
	instrct('#') ;
	sprintf(igmax,"G#lh.7max#lxhx = %3.1f !D#eh.5-1#exhx",mGMax) ;
	sprintf(islcth,"!CZ = %4.1f !D",mSliceThickness) ;

	for(i=0;i<mNumSlices;i++)
		th[i] = (i+1)*mSliceThickness ;

	size = 1.5 ;

	switch(OutDevice->type) {
		case WS:
			SetOffscreenPort(gOffscreenPtr) ;
			CLEAR_DISPLAY ;
			break ;
		case FS:
			break ;
		case LP:
			open_ps_file("laserprinter.ps") ;
			LandScape() ;
			break ;
	}
	if (!IMANY) {
		// Individual plots
		for(k=0;k<mNumBeamsToPlot;k++)  {
			sprintf(ihkl,"%d %d %d",mHPlot[k],mKPlot[k],mLPlot[k]) ;
			if (INTENSITY) {
				strcpy(buffer,"INTENSITY") ;
			}
			else {
				strcpy(buffer,"AMPLITUDE") ;
			}
			if (PLOTPHS) {
				physor(0.15,0.55) ;
				area2d(0.65,0.4) ;
				frame() ;
				yname(buffer,size) ;
			}
			else {
				physor(0.15,0.1) ;
				area2d(0.8,0.7) ;
				frame() ;
				strcpy(ixname,"Thickness !D") ;
				messag(0.15,0.89,1.5,0.,ivolt) ;
				messag(0.15,0.86,1.5,0.,igmax) ;
				messag(0.15,0.83,1.5,0.,islcth) ;
				messag(0.18,0.75,2.0,0.,ihkl) ;
			}
			headin(ident,2.0) ;
			ymax = ampmax[k] ;
			if (gScaling.automatic) {
				ymaxa = ymax ;
				ymina = 0. ;
				gScaling.amin = ymina ;
				gScaling.amax = ymaxa ;
			}
			else {
				ymaxa = gScaling.amax ;
				ymina = gScaling.amin ;
			}
			graph2d(0.,xmax,ymina,ymaxa) ;
			curve(th,amp+k*1000,mNumSlices,0) ;
			if (PLOTPHS) {
				physor(0.15,0.10) ;
				area2d(0.65,0.4) ;
				frame() ;
				strcpy(buffer,"PHASE (!B)") ;
				yname(buffer,size)  ;
				strcpy(buffer,"Thickness !D") ;
				xname(buffer,size) ;
				ymax2 = phsmax[k] ;
				ymin = - ymax2 ;
				if (gScaling.automatic) {
					gScaling.pmin = ymin ;
					gScaling.pmax = ymax2 ;
				}
				else {
					ymax2 = gScaling.pmax ;
					ymin = gScaling.pmin ;
				}
				graph2d(0.,xmax,ymin,ymax2) ;
				line(0.,0.,xmax,0.) ;
				curve(th,phase+k*1000,mNumSlices,0) ;
				messag(0.83,0.93,1.5,0.,ivolt) ;
				messag(0.83,0.90,1.5,0.,igmax) ;
				messag(0.83,0.87,1.5,0.,islcth) ;
				messag(0.18,0.45,2.0,0.,ihkl) ;
				messag(0.18,0.90,2.0,0.,ihkl) ;
			}
			if(k != mNumBeamsToPlot-1) {
				if (OutDevice->type == WS) {
					UnSetOffscreenPort(gOffscreenPtr) ;
					SetPort(gControlWindow) ;
					UpdateImageScreen() ;
					cpage() ;
					SetOffscreenPort(gOffscreenPtr) ;
					CLEAR_DISPLAY ;
				}
				else {
					cpage() ;
				}
			}
		} 
	}
	else  {   // Several plots together 
		if (INTENSITY)
			strcpy(iyname,"INTENSITY") ;
		else
			strcpy(iyname,"AMPLITUDE") ;

		if (PLOTPHS) {
			physor(0.15,0.55) ;
			area2d(0.65,0.4) ;
			frame() ;
			yname(iyname,size) ;
		}
		else {
			physor(0.15,0.1) ;
			area2d(0.8,0.7) ;
			frame() ;
			strcpy(buffer,"Thickness !D") ;
			xname(buffer,size) ;
			yname(iyname,size) ;
			messag(0.15,0.89,1.5,0.,ivolt) ;
			messag(0.15,0.86,1.5,0.,igmax) ;
			messag(0.15,0.83,1.5,0.,islcth) ;
		}
		ymaxa = 0. ;
		ymaxp = 0. ;
		for(k=0;k<mNumBeamsToPlot;k++) {
			ymaxa = max(ymaxa,ampmax[k]) ;
			ymaxp = max(ymaxp,phsmax[k]) ;
		}
		if (gScaling.automatic) {
			yminp = -ymaxp ;
			ymina = 0. ;
			gScaling.amin = ymina ;
			gScaling.amax = ymaxa ;
			gScaling.pmin = yminp ;
			gScaling.pmax = ymaxp ;
		}
		else {
			ymina = gScaling.amin ;
			ymaxa = gScaling.amax ;
			yminp = gScaling.pmin ;
			ymaxp = gScaling.pmax ;
		}
		graph2d(0.,xmax,ymina,ymaxa) ;
		for(k=0;k<mNumBeamsToPlot;k++) {
			// marker(k) ;
			curve(th,amp+k*1000,mNumSlices,0) ;
		}
		if (PLOTPHS) {
			physor(0.15,0.10) ;
			area2d(0.65,0.4) ;
			frame() ;
			strcpy(buffer,"PHASE (!B)") ;
			yname(buffer,size) ;
			strcpy(buffer,"Thickness !D") ;
			xname(buffer,size) ;
			ymin = -ymaxp ;
			graph2d(0.,xmax,yminp,ymaxp) ;
			line(0.,0.,xmax,0.) ;
			for(k=0;k<mNumBeamsToPlot;k++) {
				// marker(k) ;
				curve(th,phase+k*1000,mNumSlices,0) ;
				sprintf(ihkl,"%d %d %d",mHPlot[k],mKPlot[k],mLPlot[k]) ;
				// lines(ihkl,ipkray,k) ;
			}
			messag(0.83,0.93,1.5,0.,ivolt) ;
			messag(0.83,0.90,1.5,0.,igmax) ;
			messag(0.83,0.87,1.5,0.,islcth) ;
			// legend(ipkray,IW,7.6,4.2) ;
		}
	}

	switch(OutDevice->type) {
		case WS:
			UnSetOffscreenPort(gOffscreenPtr) ;
			UpdateImageScreen() ;
			SetPort(gControlWindow) ;
			break ;
		case FS:
			break ;
		case LP:
			close_ps_file() ;
			break ;
	}

}

void	TAmplitudePage::PrintBeams() {
	FILE 	*fp ;
	char	text[256] ;
	float 	*th,zim,*rp1,*rp2,*amp,
		*phase ;
	long 	i,j,n,ndx,ndxi ;
	long 	ierr ;
	float 	tst ;
	float 	two ;
	char	filename[256] ;

	th = mZplot+20000 ;
	amp = th+10000 ;
	phase = amp+10000 ;

	eraser(0,75,1000,105) ;
	temdtx(10,80,"Print the values to which File -> ") ;
	ReadString(filename,80) ;
	eraser(0,75,1000,105) ;
	if ((fp = fopen(filename,"w")) == NULL) {
		sprintf(text,"Unable to open the file %s",filename) ;
		PutMessage(text) ;
		return ; 
	}
	if (gPhsgrtRec->nbeam == 0) {
		SetProjectedCell(&ierr) ;
		SetCalculationDimensions(&gProjCell->a,&gProjCell->b,&gPhsgrtRec->gmax,
		&gPhsgrtRec->nbeam,&gPhsgrtRec->mbeam) ;
	}
	gPhsgrtRec->slcth = gProjCell->c/gPhsgrtRec->nz ;
	if (PLOTPHS) {
		/* adjust the phase as to get a continuous curve */
		two = 2. ;
		for(n=0;n<mNumBeamsStored;n++)
			for(j=1;j<mNumSlices;j++) {
				rp1 = mZplot + 2*(n*mNumBeamsStored+j)  ;	// Pointer to cmplx [n][j] 
				rp2 = rp1 - 2*mNumBeamsStored ;         	//  --- [n-1][j]        
				rp1++ ;   					// Imaginary part
				rp2++ ;   					// Imaginary part
				tst = fabs(*rp1-*rp2) ;
				if (tst >= 1.) {
					zim = *rp2 ;
					zim = (zim > 0. ? two : -two) ;
					*rp1 += zim ;
					tst = fabs(*rp1-*rp2) ;
					if (tst > 1.) *rp1 += zim ;
				}
			}
	}

	// set all the output arrays
	for(i=0;i<mNumBeamsToPlot;i++)
		for(j=0;j<mNumBeamsStored;j++)
			if (mHPlot[i]==mHStored[j] && mKPlot[i]==mKStored[j] && mLPlot[i]==mLStored[j]) {
			ndxi = i*1000 ;
			fprintf(fp,"\nReflection: [%d,%d,%d]",
				mHPlot[i],mKPlot[i],mLPlot[i]) ;
			for(n=0;n<mNumSlices;n++) {
				ndx = ndxi+n ;
				rp1 = mZplot + 2*(n*mNumBeamsStored+j) ;
				if (PLOTPHS) {
					 *(phase+ndx) = *(rp1+1) ;
				}
				*(amp+ndx) = *rp1 ;
				if (INTENSITY) {
					*(amp+ndx) *= *(amp+ndx) ;
				}
				fprintf(fp,"\nT = %6.2f",(n+1)*gPhsgrtRec->slcth) ;
				if (INTENSITY) 
					fprintf(fp,"\tIntensity: ") ;
				else
					fprintf(fp,"\tAmplitude: ") ;
				fprintf(fp,"%10.6f\t",*(amp+ndx)) ;
				if (PLOTPHS) {
					fprintf(fp,"Phase: %10.6f",*(phase+ndx)) ;
				}
			}
	}
	fprintf(fp,"\n") ;
	fclose(fp) ;
}

void	TAmplitudePage::DoMouseDownInObject(TObject *obj,Point *pt) {
	if (obj->DoMouseDown(pt)) {
		if (obj == mHelp) {
			run_help("TAmplitudes") ;
			return ;
		}
		if (obj == mPlotAmplitude) {
			if (mPlotIntensity->GetControlValue()) 
				mPlotIntensity->SetControlValue(false) ;
		}
		else if (obj == mPlotIntensity) {
			if (mPlotAmplitude->GetControlValue()) 
				mPlotAmplitude->SetControlValue(false) ;
		}
		else if (obj == mSuperImposeBeams) {
			if (mIndividualBeams->GetControlValue()) 
				mIndividualBeams->SetControlValue(false) ;
		}
		else if (obj == mIndividualBeams) {
			if (mSuperImposeBeams->GetControlValue()) 
				mSuperImposeBeams->SetControlValue(false) ;
		}

		else if (obj == mCancel) {
			ExitAmplitudePage(this) ;
		}
		else if (obj == mOK) {
			// See how many beams were checked for plotting
			mNumBeamsToPlot = 0 ;
			char	text[256] ;
			for(long i=0;i<mNumBeamsStored;i++) {
				if (mBeams[i]->GetControlValue()) {
					mBeams[i]->GetItemText(text) ;
					sscanf(text,"%3d %3d %3d",
						&mHPlot[mNumBeamsToPlot],
						&mKPlot[mNumBeamsToPlot],
						&mLPlot[mNumBeamsToPlot]) ;
					mNumBeamsToPlot++ ;
				}
			}
			INTENSITY 	= mPlotIntensity->GetControlValue() ;
			IMANY 		= mSuperImposeBeams->GetControlValue() ;
			PLOTPHS		= mPlotPhases->GetControlValue() ;
			if (mNumBeamsToPlot > 0) {
				if (gPrintValues) 
					PrintBeams() ;
				else
					PlotBeams() ;
			}
			else {
				PutMessage("No beams have been selected") ;
			}
		}
	}
}

void	ExitAmplitudePage(TAmplitudePage *page) {
	delete page ;
	if (LAYERED) 
		set_menu(MAIN_LAY_STR) ;
	else
		set_menu(MAIN_SNGL_STR) ;
	ClearDisplay() ;
	SetPage(gMainPage) ;
	gMainPage->Show() ;
}
