// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdlib.h>
#include "TApplication.h"
#include "globals.h"
#include "layers.h"
#include "devices.h"

long	near(float) ;

//#define	DEBUG
#ifdef DEBUG
#include "EventNames.h"
#endif

#define gFontName "-adobe-times-medium-r-normal--*-100-*-*-p-*"
#define gAltFontName "-Adobe-Times-Medium-R-Normal--10-100-75-75-P-54-ISO8859-1"
#define gAltFontName1 "-Adobe-Times-Medium-R-Normal--8-80-75-75-P-44-ISO8859-1"


typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
MonitorDisplayType  gMonitorDisplayType = PSEUDOCOLOR ;
Visual	*gVisual ;
short	gImageDepth ;
Atom	gDeleteWindowAtom ;
extern	Boolean gOptionKeyDown,gShiftKeyDown ;
extern	short	gReportVisuals ;
Time gClickTime,gLastClickTime=0 ;

#define	kDoubleTime	400
Boolean	gIsDoubleClick = false ;
Boolean IsDoubleTime() ;

short	gRedBitShift ;
short	gGreenBitShift ;
short	gBlueBitShift ;

#define kMM		0x4D4D
#define kINTEL	0x4949
typedef	enum {kMMByteOrder = kMM, kIntelByteOrder = kINTEL} ByteOrder ;
extern	ByteOrder	gCPUByteOrder ;
extern	ByteOrder	gServerByteOrder ;
short gBitsPerPRGB ;

unsigned long gRedMask ;
unsigned long gGreenMask ;
unsigned long gBlueMask ;

extern  Boolean gNoGUI ;

unsigned char *gMapFrom16To24BitColor = nil ;
unsigned char *gMapFrom24To16BitColor = nil ;

Boolean	IsDoubleTime() { return gIsDoubleClick ; } 
TApplication::TApplication(long argc,char *argv[]) {

	gMapFrom16To24BitColor = nil ;
	gMapFrom24To16BitColor = nil ;
	mWindowList = nil ;
	
	if (!gNoGUI) {
		mWindowList = new TWindowList ;
		gClickTime = 0 ;
		mDone = false ;
	
		int argcc = (int)argc ;
		fl_initialize(&argcc,argv,"XApplication",0,0) ;
		gDisplay = fl_get_display() ;
	}
    
}

TApplication::~TApplication() {
	if (mWindowList != nil) delete mWindowList ;
    if (gMapFrom16To24BitColor != nil) delete[] gMapFrom16To24BitColor ;
    if (gMapFrom24To16BitColor != nil) delete[] gMapFrom24To16BitColor ;
}

void TApplication::Initialize() {
	char thisDisplay[256],*displayName = NULL ;
	Boolean	forcePseudoColor ;

	strcpy(thisDisplay,XDisplayName(displayName)) ;

	/* Get the screen size for the display	*/
	gScreen = DefaultScreen(gDisplay) ;
	gDisplayWidth = DisplayWidth(gDisplay,gScreen) ;
	gDisplayHeight = DisplayHeight(gDisplay,gScreen) ;
    
	forcePseudoColor = false ;
	gImageDepth = DefaultDepth(gDisplay,gScreen);

	short i ;
    gMapFrom16To24BitColor = new unsigned char[32] ;
    for(i=0; i < 32; i++) {
    	gMapFrom16To24BitColor[i] = near(i*255.0/31.0) ;
    }
    gMapFrom24To16BitColor = new unsigned char[256] ;
    for(i=0; i < 256; i++) {
    	gMapFrom24To16BitColor[i] = near(i*31.0/255.0) ;
    }

	if (gImageDepth == 8 ) {
		forcePseudoColor = true ;
	}
   
	GetVisual(forcePseudoColor,&gVisualInfo) ;
	gVisual = DefaultVisual(gDisplay,gScreen) ; /* New for LUT */
	gVisualInfo.visual = gVisual ;		/* New for LUT */



	gRedMask = gVisualInfo.red_mask ;
	gGreenMask = gVisualInfo.green_mask ;
	gBlueMask = gVisualInfo.blue_mask ;
	
	gBitsPerPRGB = gVisualInfo.bits_per_rgb ;
	
	// Find out how many bits to shift for each
	unsigned long bitmask ;
	unsigned long mask_mask ;
	unsigned long bit ;
	
	if (gCPUByteOrder == kMMByteOrder) {
		if (gVisualInfo.depth == 32) {
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			// We probably have 8 bits per color
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
		else if (gVisualInfo.depth == 24) {
			mask_mask = 0x00FFFFFF ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
		else if (gVisualInfo.depth == 16 || gVisualInfo.depth == 15) {
			mask_mask = 0x0000FFFF ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32 ; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
	}
	else {
		// Testing
		/*
		unsigned long test = 15 ;
		fprintf(stderr,"\n test %u = %0X",test,test) ;
		test = 1 ;
		fprintf(stderr,"\n test for Testing = %u = %0X",test,test) ;
		for(i = 0; i < 32; i++) {
			fprintf(stderr,"\n test shifted left %d = %u = %0X",i,test,test) ;
			test <<= 1 ;
		}
		test = 1 ;
		for(i=0; i < 32; i++) {
			fprintf(stderr,"\n test shifted right %d = %u = %0X",i,test,test) ;
			test >>= i ;
		}
		fprintf(stderr,"\n gRedMask %u = %0X",gRedMask,gRedMask) ;
		fprintf(stderr,"\n gGreenMask %u = %0X",gGreenMask,gGreenMask) ;
		fprintf(stderr,"\n gBlueMask %u = %0X",gBlueMask,gBlueMask) ;
		*/
			 
		if (gVisualInfo.depth == 32) {
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
		else if (gVisualInfo.depth == 24) {
			mask_mask = 0x00FFFFFF ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
		else if (gVisualInfo.depth == 16 || gVisualInfo.depth == 15) {
			mask_mask = 0x0000FFFF ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32 ; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
	}
	if (gReportVisuals) {
		fprintf(stderr,"\n gRedBitShift = %d",gRedBitShift) ;
		fprintf(stderr,"\n gGreenBitShift = %d",gGreenBitShift) ;
		fprintf(stderr,"\n gBlueBitShift = %d",gBlueBitShift) ;
		fprintf(stderr,"\n gRedBitMask = %0X",gRedMask) ;
		fprintf(stderr,"\n gGreenBitMask = %0X",gGreenMask) ;
		fprintf(stderr,"\n gBlueBitMask = %0X",gBlueMask) ;
	}
    
	if (strncmp(thisDisplay,":0.0",4) == 0) {
		gFontStructure = XLoadQueryFont(gDisplay,gAltFontName) ;
	}
	else {
		gFontStructure = XLoadQueryFont(gDisplay,gAltFontName) ;
	}
	if (gFontStructure == NULL) {
		fprintf(stderr,"\nFailed to load the font !\n") ;
		fflush(stderr) ;
		exit(0) ;
	}	
	gImageAttributes.background_pixel = WhitePixel(gDisplay,gScreen) ;
	gImageAttributes.border_pixel = BlackPixel(gDisplay,gScreen) ;
	gImageAttributes.backing_store = WhenMapped ;
  
	gDeleteWindowAtom = XInternAtom(gDisplay,"WM_DELETE_WINDOW",false) ;

}

void TApplication::MainLoop() {
	while (!mDone) {
		ProcessEvents() ;
	}
	CleanUp() ;
}

void TApplication::DoKeyDown(XEvent *event)  {
	XKeyEvent *keyEvent = (XKeyEvent *)event ;
	TWindow *theWindow = FindWindow(keyEvent->window) ;
	if (theWindow != nil) { 
		theWindow->DoKeyDown(event) ;
	}
}

void TApplication::DoKeyUp(XEvent *event)  {
	if (!(gOptionKeyDown || gShiftKeyDown)) return ;
	XKeyEvent *keyEvent = (XKeyEvent *)event ;
	TWindow *theWindow = FindWindow(keyEvent->window) ;
	if (theWindow != nil) { 
		theWindow->DoKeyUp(event) ;
	}
}

void TApplication::DoUpdate(XEvent *event)  {
	TWindow *theWindow = FindWindow(event->xexpose.window) ;
	if (theWindow != nil) { 
		theWindow->Update(event) ;
	}
}

void TApplication::DoMouseDown(XEvent *event) {
	XButtonEvent *buttonEvent = (XButtonEvent *)event ;
	TWindow *theWindow = FindWindow(buttonEvent->window) ;
	if (theWindow != nil) {
		theWindow->DoMouseDown(event) ;
	}
}

void TApplication::CleanUp() {
	delete mWindowList ;
}

void TApplication::SelectCursor() {
	// Find the window that is in the foreground and 
	// select the cursor accordingly
}

void TApplication::DoIdle() {
	SelectCursor() ;	// Select the correct cursor
	// Also do idle for the active window
	// for all windows do idle
} 

void TApplication::DoResizeWindow(XEvent *event) {
	TWindow *theWindow = FindWindow(event->xexpose.window) ;
	if (theWindow != nil) 
		theWindow->Resize(event) ;
}

void TApplication::DoClientMessage(XEvent *event) {
	XClientMessageEvent *theMessage = (XClientMessageEvent *)event ;
	TWindow *theWindow = FindWindow(theMessage->window) ;
	if (theWindow != nil) 
		theWindow->DoClientMessage(theMessage) ;
}

void TApplication::GetVisual(Boolean forcePseudoColor,XVisualInfo *vinfo) {
    long i ;
    Status retInfo ;
    XVisualInfo	*vInfo,*visPtr ;
    extern	short	gReportVisuals ;
    
    int	numReturned ;

    vInfo = XGetVisualInfo(gDisplay,VisualNoMask,NULL,&numReturned) ;
    visPtr = vInfo ;
    retInfo = 0 ;
    
    
	if (gReportVisuals) {
		// Report all visuals
		fprintf(stderr,"\nThese are the available Visuals");
		for(i=0;i<numReturned;i++,vInfo++) {
			switch(vInfo->c_class) {
				case StaticGray:
					fprintf(stderr,"\nClass = StaticGray , Bit Depth = %d",vInfo->depth) ;
					break ;
				case GrayScale:
					fprintf(stderr,"\nClass = GrayScale , Bit Depth = %d",vInfo->depth) ;
					break ;
				case StaticColor:
					fprintf(stderr,"\nClass = StaticColor , Bit Depth = %d",vInfo->depth) ;
					break ;
				case PseudoColor:
					fprintf(stderr,"\nClass = PseudoColor , Bit Depth = %d",vInfo->depth) ;
					break ;
				case TrueColor:
					fprintf(stderr,"\nClass = TrueColor , Bit Depth = %d",vInfo->depth) ;
					break ;
				case DirectColor:
					fprintf(stderr,"\nClass = DirectColor , Bit Depth = %d",vInfo->depth) ;
					break ;
			}					
		}
		vInfo = visPtr ;
	}
    
    if (forcePseudoColor) {
		if (gReportVisuals) 
	        fprintf(stderr,"Forcing pseudocolor\n");
	    // Try and find an 8-bit display
    	for(i=0;i<numReturned;i++,vInfo++) {
			if (vInfo->c_class == PseudoColor && vInfo->depth == 8) {
				retInfo = XMatchVisualInfo(gDisplay,
							DefaultScreen(gDisplay),
							vInfo->depth,
							vInfo->c_class,
							vinfo) ;
				gMonitorDisplayType = PSEUDOCOLOR ;
				if (gReportVisuals) 
	    			fprintf(stderr,"Got 8bit Pseudocolor\n");
				break ;
			}
		}
    }
    else {
	    // Do we have 24 bit?
        vInfo = visPtr ;
		if (gReportVisuals) 
	        fprintf(stderr,"Looing for 24bit\n");
    	for(i=0;i<numReturned;i++,vInfo++) {
			if (vInfo->c_class == TrueColor && (vInfo->depth == 24 || vInfo->depth == 32)) {
				retInfo = XMatchVisualInfo(gDisplay,
							DefaultScreen(gDisplay),
							vInfo->depth,
							vInfo->c_class,
							vinfo) ;
				if (gReportVisuals) 
	    			fprintf(stderr,"Got 24bit TrueColor\n");
				gMonitorDisplayType = TRUECOLOR ;
				break ;
			}
		}
    if (!retInfo) { //We have not found it yet, try 16 bit 
    
        vInfo = visPtr ;
		if (gReportVisuals) 
	        fprintf(stderr,"Looing for 16bit\n");
    	for(i=0;i<numReturned;i++,vInfo++) {
			if (vInfo->c_class == TrueColor && (vInfo->depth == 16 || vInfo->depth == 15)) {
				retInfo = XMatchVisualInfo(gDisplay,
							DefaultScreen(gDisplay),
							vInfo->depth,
							vInfo->c_class,
							vinfo) ;
				if (gReportVisuals) 
	    			fprintf(stderr,"Got 16bit TrueColor\n");
				gMonitorDisplayType = TRUECOLOR ;
				break ;
			}
		}
		}		
    if (!retInfo) { //Still not found it , try 8 bit pseudocolor
        vInfo = visPtr ;
		if (gReportVisuals) 
	        fprintf(stderr,"Looing for 8bit\n");
    	for(i=0;i<numReturned;i++,vInfo++) {
			if (vInfo->c_class == PseudoColor && vInfo->depth == 8) {
				retInfo = XMatchVisualInfo(gDisplay,
							DefaultScreen(gDisplay),
							vInfo->depth,
							vInfo->c_class,
							vinfo) ;
				if (gReportVisuals) 
	    			fprintf(stderr,"Got 8bit Pseudocolor\n");
				gMonitorDisplayType = TRUECOLOR ;
				break ;
			}
		}
		}		
    } 
    if (!retInfo) {
		printf("\n Unable to locate the right Visual. Exiting...\n") ;
		exit(-1) ;
    }
    XFree(visPtr) ; 
	
}

void TApplication::SetupColors() {
}

void	TApplication::ProcessModalEvent(XEvent *event) {
	// Other windows only get update and client events
	switch(event->type) {
		case Expose:    
			DoUpdate(event) ;
			break ;
		case ClientMessage:
			DoClientMessage(event) ;
			break ;
		case ConfigureNotify:
			break ;
	}
}

void TApplication::ProcessEvents() {
	XEvent event ;
	FL_OBJECT	*theFormsObject ;
	XButtonEvent *buttEvent = (XButtonEvent *)&event ;

	DoIdle() ;
	theFormsObject = fl_do_forms() ;
	if (theFormsObject == FL_EVENT) {
		fl_XNextEvent(&event) ;
		switch(event.type) {
			case ConfigureNotify:
				DoResizeWindow(&event) ;
				break ;
			case Expose:    
				DoUpdate(&event) ;
				break ;
			case ButtonPress:
				gIsDoubleClick = false ; 
				if ((buttEvent->time - gClickTime) < 400) gIsDoubleClick = true ;
				gClickTime	= buttEvent->time ;
				DoMouseDown(&event) ;
				break ;
			case KeyPress:
				DoKeyDown(&event) ;
				break ;
			case KeyRelease:
				DoKeyUp(&event) ;
				break ;
			case ClientMessage:
				DoClientMessage(&event) ;
				break ;
#ifdef DEBUG
			default:
				fprintf(stderr,"Event '%s' not handled\n",event_names[event.type]) ;
				break ;
#endif
		}
	}
}
