// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "TChemicalConstraintsDialog.h"
#include "funcdefs.h"
extern	char *chemsymb[] ;

typedef struct {
	float top ;
	float left ;
	float bottom ;
	float right ;
} FRect ;

typedef struct	{
	short number ;
	short index[20] ;
} ElementRecord,*ElementPtr ;
extern	ElementPtr MyElements ;

typedef struct {
	short 	typeAtom1 ;
	short 	typeAtom2 ;
	float 	optimumDistance ;
	float	weight ;
	Boolean	use3Ddistance ;
	Boolean	isActive ;
} ChemicalBondRecord ;

typedef struct ChemicalBondData {
	short numberOfBonds ;
	ChemicalBondRecord bond[20] ;
} ChemicalBondData ;

typedef enum {kCrossCorrelation,kChiSquared,kFTAmplCrossCorrelation } FittingMethod ;

typedef struct {
	float	maxAmplitude ;	// In Angstrom
	Boolean	useActiveRect ;
	FRect	activeRect ;
} RelaxationAtomRecord ;

typedef struct {
	float	start ;			// Starting Temperature
	float	end ;			// Ending Temperature
	float	current ;		// current temperature
	float	dTFraction ;	// Temperature decrement
	long		maxNumTriesPerTemperature ;
	long		maxNumGoodTries ;
} RelaxationTemperatureRecord ;

typedef struct {
	RelaxationAtomRecord		atom ;	// In Angstrom
	RelaxationTemperatureRecord	temperature ;
	float previousEnergy ;
	float thisEnergy ;
	float minEnergy ;
	long  numTriesThisTemperature ;
	FittingMethod fittingMethod ;
} RelaxationParamRecord ;

extern	RelaxationParamRecord gRelaxationParamRecord ;
extern	Boolean gIncludeBondDistances ;

TChemicalConstraintsDialog::TChemicalConstraintsDialog() 
	: TXForm("Chemical Constraints") {

	DefineForm() ;
	mData = (void *) new ChemicalBondData ;
}

TChemicalConstraintsDialog::~TChemicalConstraintsDialog() {
	delete mData ;
}

void TChemicalConstraintsDialog::DefineForm() {
	FL_OBJECT	*obj ;
	mXForm->mForm = fl_bgn_form(FL_NO_BOX, 391, 271);
	fl_add_box(FL_UP_BOX,0,0,391,271,"");
	mAtomList1 = fl_add_browser(FL_HOLD_BROWSER,15,33,60,110,"");
	mAtomList2 = fl_add_browser(FL_HOLD_BROWSER,115,33,60,110,"");
	mBondList = fl_add_browser(FL_HOLD_BROWSER,15,145,160,70,"");
	fl_add_labelframe(FL_ENGRAVED_FRAME,10,25,170,195,"Bond criteria");
	fl_add_box(FL_UP_BOX,75,90,40,10,"");
	mOKButton = fl_add_button(FL_NORMAL_BUTTON,300,230,70,30,"OK");
	mCancelButton = fl_add_button(FL_NORMAL_BUTTON,190,230,90,30,"Cancel");
	mAdd = fl_add_button(FL_NORMAL_BUTTON,190,195,50,25,"Add");
	mModify = fl_add_button(FL_NORMAL_BUTTON,250,195,60,25,"Modify");
	mDelete = fl_add_button(FL_NORMAL_BUTTON,320,195,50,25,"Delete");
	mR0 = fl_add_input(FL_NORMAL_INPUT,290,40,60,20,"r0 [Ang.]");
	m3D = fl_add_checkbutton(FL_PUSH_BUTTON,200,70,150,20,"3d constraint (vs 2-d)");
	fl_add_labelframe(FL_ENGRAVED_FRAME,190,25,180,80,"Optimum distance");
	mWeight = fl_add_input(FL_NORMAL_INPUT,290,125,60,20,"Weight (W)");
	obj = fl_add_text(FL_NORMAL_TEXT,210,150,150,30,
		"E = W*(4/(r-r0)**10-\n            5/(r-r0)**8)");
	fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
	fl_add_labelframe(FL_ENGRAVED_FRAME,190,115,180,70,"Energy contribution");
	fl_end_form();

	fl_set_clipping(0,0,391, 271) ;
	mXForm->mForm->fdui = mXForm;
}

void TChemicalConstraintsDialog::DoObject(FL_OBJECT *obj) {

	short 		i,ndx ;
	char 		text[256],*aName,atomName1[8],atomName2[8],s2d3d[4] ;
	float		distance ;
	float		weight ;
	long		line1,line2,line3 ;

	if (obj == mAtomList1) {
		if (!((line1 = fl_get_browser(mAtomList1)) > 0 &&
				(line2 = fl_get_browser(mAtomList2)) > 0))
			DeactivateButton(mAdd) ;
		
	}
	if (obj == mAtomList2) {
		if (!((line1 = fl_get_browser(mAtomList1)) > 0 &&
				(line2 = fl_get_browser(mAtomList2)) > 0))
			DeactivateButton(mAdd) ;
	}
	if (obj == mBondList) {
		// If something is selected after the click, activate DELETE and MODIFY atom
		if ((line3 = fl_get_browser(mBondList)) > 0) {
			strcpy(text,fl_get_browser_line(mBondList,line3)) ;
					
			sscanf(text,"%s - %s : %f : %f : %s",
				atomName1,atomName2,&distance,&weight,s2d3d) ;;
			sprintf(text,"%.2f",distance) ;
			fl_set_input(mR0,text) ;
			sprintf(text,"%.3f",weight) ;
			if (strcmp(s2d3d,"2d")==0) 
				fl_set_button(m3D,0) ;
			else
				fl_set_button(m3D,1) ;
			// Find out which items should be selected in the atom lists
			fl_freeze_form(mXForm->mForm) ;
			fl_deselect_browser(mAtomList1) ;
			fl_deselect_browser(mAtomList2) ;
			for (i = 0; i < MyElements->number;i++) {
				ndx = MyElements->index[i] ;
				aName = chemsymb[ndx] ;
				if (strcmp(atomName1,aName)==0) {
					fl_select_browser_line(mAtomList1,i+1) ;
				}
				if (strcmp(atomName2,aName)==0) {
					fl_select_browser_line(mAtomList2,i+1) ;
				}
			}
			fl_unfreeze_form(mXForm->mForm) ;
			ActivateButton(mDelete) ;
			ActivateButton(mModify) ;
		}
		else {
			DeactivateButton(mDelete) ;
			DeactivateButton(mModify) ;
		}	
	}
	if (obj == mAdd) {
		if ((line1 = fl_get_browser(mAtomList1)) > 0) {
			if ((line2 = fl_get_browser(mAtomList2)) > 0) {
				// Get the text in the two lists
				strcpy(atomName1,fl_get_browser_line(mAtomList1,line1)) ;
				strcpy(atomName2,fl_get_browser_line(mAtomList2,line2)) ;
						
				// Get the bond-length and the weight
				strcpy(text,fl_get_input(mR0)) ;
				distance = strlen(text) == 0 ? 2.0 : atof(text) ;
						
				strcpy(text,fl_get_input(mWeight)) ;
				weight = strlen(text) == 0 ? 0.01 : atof(text) ;

				if (fl_get_button(m3D))
					sprintf(text,"%-2.2s - %-2.2s : %.2f : %.3f : 3d",
						atomName1,atomName2,distance,weight) ;
				else
					sprintf(text,"%-2.2s - %-2.2s : %.2f : %.3f : 2d",
							atomName1,atomName2,distance,weight) ;
				
				fl_add_browser_line(mBondList,text) ;
				gIncludeBondDistances = true ;
				//fl_select_browser_line(mBondList,fl_get_browser_maxline(mBondList));
			}
			else 
				DeactivateButton(mAdd) ;
		}
		else 
			DeactivateButton(mAdd) ;
	}
	if (obj == mDelete) {
		// Delete the bond that is defined (highlighted) in the Bond list
		if ((line3 = fl_get_browser(mBondList)) > 0) {
			// Delete this row
			fl_delete_browser_line(mBondList,line3) ;
			//long lastLine = fl_get_browser_maxline(mBondList) ;
			//if (line3 > lastLine)
			//	fl_select_browser_line(mBondList,lastLine) ;
			//else
			//	fl_select_browser_line(mBondList,line3) ;
		}
	}
	if (obj == mModify) {
		// Change parameters for the bond that is highlighted in the Bond list
		if ((line3 = fl_get_browser(mBondList)) > 0)
		   	if ((line1 = fl_get_browser(mAtomList1)) > 0)
				if ((line2 = fl_get_browser(mAtomList2)) > 0 ) {
							
			// Get the text in the two lists
			strcpy(atomName1,fl_get_browser_line(mAtomList1,line1)) ;
			strcpy(atomName2,fl_get_browser_line(mAtomList2,line2)) ;
							
			// Get the bond-length and the weight
			strcpy(text,fl_get_input(mR0)) ;
			distance = strlen(text) == 0 ? 2.0 : atof(text) ;
			strcpy(text,fl_get_input(mWeight)) ;
			weight = strlen(text) == 0 ? 0.01 : atof(text) ;
							
			if (fl_get_button(m3D))
				sprintf(text,"%-2.2s - %-2.2s : %.2f : %.3f : 3d",
					atomName1,atomName2,distance,weight) ;
			else
				sprintf(text,"%-2.2s - %-2.2s : %.2f : %.3f : 2d",
					atomName1,atomName2,distance,weight) ;
			fl_replace_browser_line(mBondList,line3,text) ;
		}
	}
}

void TChemicalConstraintsDialog::SetFormData(void *data) {
	
	ChemicalBondData *theData = (ChemicalBondData *)mData ;
	*theData = *((ChemicalBondData *)data) ;

	short 		i,ndx ;
	char 		text[256],*aName,atomName1[8],atomName2[8] ;
	float		distance ;
	float		weight ;
	
	fl_freeze_form(mXForm->mForm);
	fl_clear_browser(mAtomList1);
	fl_clear_browser(mAtomList2);
	fl_clear_browser(mBondList);
	for (i=0;i<MyElements->number;i++) {
		ndx = MyElements->index[i] ;
		aName = chemsymb[ndx] ;
		fl_add_browser_line(mAtomList1,aName) ;
		fl_add_browser_line(mAtomList2,aName) ;
	}
	for(i=0;i<theData->numberOfBonds;i++) {
		strcpy(atomName1,chemsymb[theData->bond[i].typeAtom1-1]) ;
		strcpy(atomName2,chemsymb[theData->bond[i].typeAtom2-1]) ;
		distance = theData->bond[i].optimumDistance ;
		weight = theData->bond[i].weight ;
		if (theData->bond[i].use3Ddistance) 
			sprintf(text,"%-2.2s - %-2.2s : %.2f : %.3f : 3d",
				atomName1,atomName2,distance,weight) ;
		else
			sprintf(text,"%-2.2s - %-2.2s : %.2f : %.3f : 2d",
				atomName1,atomName2,distance,weight) ;
		fl_add_browser_line(mBondList,text) ;
	}
	fl_select_browser_line(mAtomList1,1) ;
	fl_select_browser_line(mAtomList2,1) ;
	fl_unfreeze_form(mXForm->mForm);

	DeactivateButton(mModify) ;
	DeactivateButton(mDelete) ;

	if (theData->numberOfBonds == 0) {
		distance = 2.0 ;
		switch(gRelaxationParamRecord.fittingMethod) {
			case kCrossCorrelation:
			case kFTAmplCrossCorrelation:
				weight = 0.01 ;
				break ;
			case kChiSquared:
				weight = 0.1 ;
				break ;
		}
	}
	else {
		distance = theData->bond[0].optimumDistance ;
		weight = theData->bond[0].weight ;
	}
	sprintf(text,"%.2f",distance) ;
	fl_set_input(mR0,text) ;
	sprintf(text,"%.3f",weight) ;
	fl_set_input(mWeight,text) ;

}

void TChemicalConstraintsDialog::GetFormData(void *data) {
	long 		i;
	char 		text[256],atomName1[8],atomName2[8],s2d3d[4] ;
	float		distance ;
	float		weight ;

	ChemicalBondData *theData = (ChemicalBondData *)mData ;
	// Find out how many cells we have in the list
	theData->numberOfBonds = fl_get_browser_maxline(mBondList) ;

	for(i=0;i<theData->numberOfBonds;i++) {
		strcpy(text,fl_get_browser_line(mBondList,i+1)) ;
			
		sscanf(text,"%s - %s : %f : %f : %s",
			atomName1,atomName2,&distance,&weight,s2d3d) ;
		theData->bond[i].typeAtom1 = atomno(atomName1) ;
		theData->bond[i].typeAtom2 = atomno(atomName2) ;
		theData->bond[i].optimumDistance = distance ;
		theData->bond[i].weight = weight ;
		theData->bond[i].use3Ddistance = 
			strcmp(s2d3d,"3d") == 0 ? true : false ;
		theData->bond[i].isActive = true ;
	}

	theData = (ChemicalBondData *)data ;
	*theData = *((ChemicalBondData *)mData) ;
}
