// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "menu_index.h"
#include "TImage.h"
#include "TButton.h"
#include "TCompareImagesPage.h"
#include "TMainPage.h"

#include "TRefineParametersPage.h"
#include "globals.h"

void	SetPage(TPage *) ;
extern	TMainPage *gMainPage ;
void	PutMessage(char *) ;
void 	DrawQuantitativePage() ;
void	ExitFromComparison(TCompareImagesPage *) ;

extern	TRefineParametersPage *gRefineParametersPage ;

extern	TImage *gStandardDeviationImage ;
extern	TImage *gExperimentalImage ;
Boolean AdjustImageSize(TImage **image,long width,long height) ;

void DisplayImageInWindow(TImage *, TWindow *, Rect *rect,Boolean );
void run_help(char *text) ;
extern	unsigned char tool_bits[11][60] ;

typedef enum {kXCF,kChiSquare,kRMSD,kDifferenceImage,kFMAD,kFTXCF}
ComparisonMethod ;

ComparisonMethod gComparisonMethod = kXCF ;
Boolean gEstimateImageShift = false ;
Boolean gCalculateExactFTXCF = true ;
FRect	gComparisonRect = {0.,0.,1.,1.} ;
Boolean	gCompareEntireImage = true ;

void CompareExperimentalImageWithSimulation() ;

TCompareImagesPage::TCompareImagesPage(TWindow *window) 
		: TPage(window) {

	TGroupBox 	*grpBox ;
	TFrame		*theFrame ;
	TStaticText	*theText ;
	TIcon	*icon ;
	
	mHelp = new TButton(490,2,510,20,"H") ;
	AddObject(mHelp) ;

	theText = new TStaticText(20,16,184,32,"Exp. Image Name:") ;
	AddObject(theText) ;

	theText = new TStaticText(20,52,184,68,"Std. Dev. Image Name:") ;
	AddObject(theText) ; 

	if (gExperimentalImage == nil) { 
		mExpImageName = new TStaticText(20,32,184,44,"None") ;
	}
	else {
		mExpImageName = new TStaticText(20,32,184,44,gExperimentalImage->mName) ;
	}
	AddObject(mExpImageName) ;

	if (gStandardDeviationImage == nil) {
		mStdDevName = new TStaticText(20,68,184,80,"None") ;
	}
	else {
		mStdDevName = new TStaticText(20,68,184,80,gStandardDeviationImage->mName) ;
	}
	AddObject(mStdDevName) ;

	icon = new TIcon(20,20,tool_bits[0]) ;
	mArrowButton	= new TPushButton(204,40,228,64,icon) ;
	AddObject(mArrowButton) ;

	icon = new TIcon(20,20,tool_bits[1]) ;
	mRectButton	= new TPushButton(204,68,228,92,icon) ;
	AddObject(mRectButton) ;

	icon = new TIcon(20,20,tool_bits[4]) ;
	mMagnifyButton	= new TPushButton(204,96,228,120,icon) ;
	AddObject(mMagnifyButton) ;

	icon = new TIcon(20,20,tool_bits[9]) ;
	mHandButton	= new TPushButton(204,124,228,148,icon) ;
	AddObject(mHandButton) ;

	mCurrentTool = kComparePointer ;
	mArrowButton->SetState(true) ;

	grpBox = new TGroupBox(8,8,504,176, "Experimental Image & Standard Deviation") ;
	AddObject(grpBox) ;

	grpBox = new TGroupBox(20,90,190,143, "Area to compare") ;
	AddObject(grpBox) ;

	mCompareEntireImage = new TRadioButton(25,100,180,116,"Compare Entire Image") ;
	mCompareEntireImage->SetControlValue(true) ;
	AddObject(mCompareEntireImage) ;

	mCompareSelection	= new TRadioButton(25,120,130,136,"Compare Selection") ;
	mCompareSelection->Disable() ;
	AddObject(mCompareSelection) ;

	mSetNewSelection = new TButton(135,122,185,138,"Modify") ;
	mSetNewSelection->Disable() ;
	AddObject(mSetNewSelection) ;

	mResetView = new TButton(20,148,100,168,"Reset View") ;
	mResetView->Disable() ;
	AddObject(mResetView) ;

	mExpImageFrame = new TUserRect(232,40,360,168) ;
	AddObject(mExpImageFrame) ;
	mExpImageFrame->Enable() ;

	Rect	theRect = mImageRect = mExpImageFrame->GetRect() ;
	theText = new TStaticText(theRect.left,theRect.top-20,
		theRect.right,theRect.top-2,"Experimental Image") ;
	theText->SetJustification(kCenterJustified) ;
	AddObject(theText) ;

	mStdDevImageFrame = new TUserRect(364,40,492,168) ;
	AddObject(mStdDevImageFrame) ;

	theRect = mStdDevRect = mStdDevImageFrame->GetRect() ;
	theText = new TStaticText(theRect.left,theRect.top-20,
		theRect.right,theRect.top-2,"Std. Deviation Image") ;
	theText->SetJustification(kCenterJustified) ;
	AddObject(theText) ;

	grpBox = new TGroupBox(8,184,504,340, "Comparison Method") ;
	AddObject(grpBox) ;

	mRealCCC = new TRadioButton(44,200,280,216,"Cross Correlation Real Space") ;
	mRealCCC->SetControlValue(true) ; mSelectedMethod = mRealCCC ;
	AddObject(mRealCCC) ;

	mFTCCC = new TRadioButton(44,224,238,240,"Cross Correlation Fourier Space") ;
	AddObject(mFTCCC) ;

	grpBox = new TGroupBox(288,204,496,280,"Fourier Method (Compares Entire Image)") ;
	AddObject(grpBox) ;

	theFrame = new TFrame(238,232,288,232) ;
	AddObject(theFrame) ;

	mExactFTCCC = new TRadioButton(292,218,452,234,"Exact") ;
	mExactFTCCC->SetControlValue(true) ;
	mExactFTCCC->Disable() ;
	AddObject(mExactFTCCC) ;

	mAmplitudeFTCCC = new TRadioButton(292,240,452,256,"Amplitudes Only") ;
	mAmplitudeFTCCC->Disable() ;
	AddObject(mAmplitudeFTCCC) ;

	mEstimateShift	= new TCheckBox(324,260,484,276,"Estimate Image Shift") ;
	mEstimateShift->Disable() ;
	AddObject(mEstimateShift) ;

	mChiSquare = new TRadioButton(44,248,280,264,"Chi-Square") ;
	AddObject(mChiSquare) ;

	mRMSD = new TRadioButton(44,272,280,288,"Root Mean Square Difference") ;
	AddObject(mRMSD) ;

	mDiffImage = new TRadioButton(44,296,280,312,"Difference Image") ;
	AddObject(mDiffImage) ;

	mFMAD = new TRadioButton(44,320,280,336,"Fractional Mean Absolute Difference") ;
	AddObject(mFMAD) ;

	mOK	= new TButton(432,356,500,376,"Compare") ;
	AddObject(mOK) ;

	mCancel = new TButton(352,356,416,376,"Return") ;
	AddObject(mCancel) ;

	mExpImage = nil ;
	mStdDevImage = nil ;

	mImageSelectionRect = gComparisonRect ;

	mImageSelection = nil ;
	mImageCenter.x = (mImageRect.right-mImageRect.left)/2 ;
	mImageCenter.y =  (mImageRect.bottom-mImageRect.top)/2 ;
	mStdDevCenter.x = (mStdDevRect.right-mStdDevRect.left)/2 ;
	mStdDevCenter.y =  (mStdDevRect.bottom-mStdDevRect.top)/2 ;
	mImageMagnification = 1.0 ;
	
	mIsInsideActiveRect = false ;

}
 
TCompareImagesPage::~TCompareImagesPage() {
}

void TCompareImagesPage::Draw() {
	EraseRect(&mWindow->GetPortRect()) ;
	TPage::Draw() ;
	DrawExpImageDisplayArea() ;
	DrawStdDevImageDisplayArea() ;
}


void 
TCompareImagesPage::DrawImageComparisonRect() {
	long left,top,right,bottom ;
	float width,height ;
	Rect activeRect ;
	extern	unsigned long	gBlackPixel,gRedPixel ;

	width = mActualImageRect.right - mActualImageRect.left ;
	height = mActualImageRect.bottom - mActualImageRect.top ;
	left = mActualImageRect.left + mImageSelectionRect.left*width ;
	right = mActualImageRect.left + mImageSelectionRect.right*width ;
	top = mActualImageRect.top + mImageSelectionRect.top*height ;
	bottom = mActualImageRect.top + mImageSelectionRect.bottom*height ;
	SetRect(&activeRect,left,top,right,bottom) ;
	
	SetFgColor(gRedPixel) ;
	FrameRect(&activeRect) ;
	SetFgColor(gBlackPixel) ;

}


void 
TCompareImagesPage::DrawStdDevComparisonRect() {
	long left,top,right,bottom ;
	float width,height ;
	Rect activeRect ;
	extern	unsigned long	gBlackPixel,gRedPixel ;

	width = mActualStdDevRect.right - mActualStdDevRect.left ;
	height = mActualStdDevRect.bottom - mActualStdDevRect.top ;
	left = mActualStdDevRect.left + mImageSelectionRect.left*width ;
	right = mActualStdDevRect.left + mImageSelectionRect.right*width ;
	top = mActualStdDevRect.top + mImageSelectionRect.top*height ;
	bottom = mActualStdDevRect.top + mImageSelectionRect.bottom*height ;
	SetRect(&activeRect,left,top,right,bottom) ;
	
	SetFgColor(gRedPixel) ;
	FrameRect(&activeRect) ;
	SetFgColor(gBlackPixel) ;

}

void TCompareImagesPage::DrawExpImageDisplayArea() {
	Rect	tRect ;
	short 	x,y,width ;
	extern	unsigned long	gWhitePixel,gBlackPixel ;

	tRect = mExpImageFrame->GetRect() ;
	if (gExperimentalImage == nil) {
		SetFgColor(gDarkerGray) ;
		FillRect(&tRect) ;
		SetFgColor(gWhitePixel) ;
		width = StringWidth("NO IMAGE LOADED") ;
		x = (tRect.left+tRect.right-width)/2 ;
		y = (tRect.top+tRect.bottom)/2 + 10 ;
		DrawText(x,y,"NO IMAGE LOADED") ;
		SetFgColor(gBlackPixel) ;
	}
	else {
		DrawExperimentalImage() ;
	}

	PenSize(2,2) ;
	DrawLine(tRect.left,tRect.bottom,tRect.left,tRect.top) ;
	DrawLine(tRect.left,tRect.top,tRect.right,tRect.top) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(tRect.right,tRect.top,tRect.right,tRect.bottom) ;
	DrawLine(tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
	PenSize(1,1) ;
	SetFgColor(gBlackPixel) ;
	DrawExperimentalImage() ;
}
void TCompareImagesPage::DrawStdDevImageDisplayArea() {
	Rect	tRect ;
	short 	x,y,width,height ;
	extern	unsigned long	gWhitePixel,gBlackPixel ;

	tRect = mStdDevImageFrame->GetRect() ;

	if (gStandardDeviationImage == nil) {
		SetFgColor(gDarkerGray) ;
		FillRect(&tRect) ;
		SetFgColor(gWhitePixel) ;
		width = StringWidth("NO IMAGE LOADED") ;
		x = (tRect.left+tRect.right-width)/2 ;
		y = (tRect.top+tRect.bottom)/2 + 10 ;
		DrawText(x,y,"NO IMAGE LOADED") ;
		SetFgColor(gBlackPixel) ;
	}
	else {
		DrawStdDevImage() ;
	}
	PenSize(2,2) ;
	DrawLine(tRect.left,tRect.bottom,tRect.left,tRect.top) ;
	DrawLine(tRect.left,tRect.top,tRect.right,tRect.top) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(tRect.right,tRect.top,tRect.right,tRect.bottom) ;
	DrawLine(tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
	PenSize(1,1) ;
	SetFgColor(gBlackPixel) ;
}


void	TCompareImagesPage::DrawExperimentalImage() {
	Rect	tRect ;
	extern	unsigned long	gBlackPixel ;


	if (gExperimentalImage == nil) {
		return ;
	}

	tRect = mExpImageFrame->GetRect() ;
	InsetRect(&tRect,6,6) ;

	if (mExpImage == nil) {
		InsetRect(&tRect,1,1) ;
		mExpImage = CopyImage(gExperimentalImage) ;
		if (mExpImage != nil) {
			float	ratio1,ratio2,fraction ;
			short	width = mExpImage->mWidth ;
			short	height = mExpImage->mHeight ;
			ratio1 = tRect.right-tRect.left ;
			ratio1 /= width ;
			ratio2 = tRect.bottom-tRect.top ;
			ratio2 /= height ;
			fraction = ratio2 < ratio1 ? ratio2 : ratio1 ;
			// if (fraction > 1.) fraction = 1. ;		
			width *= fraction ;
			height *= fraction ;
			SetRect(&mActualImageRect,mImageCenter.x - width/2,
				mImageCenter.y - height/2,
				mImageCenter.x + width/2,
				mImageCenter.y + height/2) ;
			if (AdjustImageSize(&mExpImage,width,height)) {
				CreateImageOffscreen(mExpImage,mWindow) ;
			}
		}
	}

	if (mExpImage == nil) return ;

	tRect = mExpImageFrame->GetRect() ;
	InsetRect(&tRect,1,1) ;
	TStOffscreen offScreen(mWindow,&tRect) ;
	SetFgColor(gDarkerGray) ;
	FillRect(&gCurrentPort->GetPortRect()) ;
	SetFgColor(gBlackPixel) ;

	DisplayImageInRect(mExpImage,&mActualImageRect);
	DrawImageComparisonRect() ;

}

void	TCompareImagesPage::DrawStdDevImage() {
	Rect	tRect ;
	extern	unsigned long	gBlackPixel ;


	if (gStandardDeviationImage == nil) {
		return ;
	}

	tRect = mStdDevImageFrame->GetRect() ;
	InsetRect(&tRect,6,6) ;

	if (mStdDevImage == nil) {
		InsetRect(&tRect,1,1) ;
		mStdDevImage = CopyImage(gStandardDeviationImage) ;
		if (mStdDevImage != nil) {
			float	ratio1,ratio2,fraction ;
			short	width = gStandardDeviationImage->mWidth ;
			short	height = gStandardDeviationImage->mHeight ;
			ratio1 = tRect.right-tRect.left ;
			ratio1 /= width ;
			ratio2 = tRect.bottom-tRect.top ;
			ratio2 /= height ;
			fraction = ratio2 < ratio1 ? ratio2 : ratio1 ;
			// if (fraction > 1.) fraction = 1. ;		
			width *= fraction ;
			height *= fraction ;
			SetRect(&mActualStdDevRect,mStdDevCenter.x - width/2,
					mStdDevCenter.y - height/2,
					mStdDevCenter.x + width/2,
					mStdDevCenter.y + height/2) ;
			if (AdjustImageSize(&mStdDevImage,width,height)) {
				CreateImageOffscreen(mStdDevImage,mWindow) ;
			}
		}
	}

	if (mStdDevImage == nil) return ;

	tRect = mStdDevImageFrame->GetRect() ;
	InsetRect(&tRect,1,1) ;
	TStOffscreen offScreen(mWindow,&tRect) ;
	SetFgColor(gDarkerGray) ;
	FillRect(&gCurrentPort->GetPortRect()) ;
	SetFgColor(gBlackPixel) ;

	DisplayImageInRect(mStdDevImage,&mActualStdDevRect);
	DrawStdDevComparisonRect() ;
}

void	TCompareImagesPage::DoMouseDownInObject(TObject *obj,Point *pt) {
	Rect	tRect ;
	if (obj->DoMouseDown(pt)) {
		if (obj == mHelp) {
			run_help("TCompareImagesPage") ;
			return ;
		}
		if (obj == mCompareEntireImage) {
			if (mCompareSelection->GetControlValue()) 
				mCompareSelection->SetControlValue(false) ;
				if (mImageSelection != nil) {
					delete mImageSelection ;
					mImageSelection = nil ;
				}
				mImageSelectionRect.left = 0. ;
				mImageSelectionRect.top = 0. ;
				mImageSelectionRect.right = 1. ;
				mImageSelectionRect.bottom = 1. ;
				DrawExperimentalImage() ;
				DrawStdDevImage() ;
				mCompareSelection->Disable() ;
				mSetNewSelection->Disable() ;
		}
		else if (obj == mCompareSelection) {
			if (mCompareEntireImage->GetControlValue()) 
				mCompareEntireImage->SetControlValue(false) ;
				if (mImageSelection != nil) {
					Rect selectionRect ;
					tRect = mExpImageFrame->GetRect() ;
					mImageSelection->GetSelectionRect(selectionRect) ;
					// Compute the actual rectange mImageSelectionRect
					Rect sizeOfImage = mActualImageRect ;
					OffsetRect(&sizeOfImage,tRect.left,tRect.top) ;
					float x1,y1,x2,y2,width,height ;
					width = sizeOfImage.right - sizeOfImage.left ;
					height = sizeOfImage.bottom - sizeOfImage.top ;
					x1 = selectionRect.left - sizeOfImage.left ;
					x1 = x1 < 0. ? 0. : x1/width ;
					y1 = selectionRect.top - sizeOfImage.top ;
					y1 = y1 < 0. ? 0. : y1/height ;
					x2 = sizeOfImage.right - selectionRect.right ;
					x2 = x2 < 0. ? 1. : 1.0 - x2/width ;
					y2 = sizeOfImage.bottom - selectionRect.bottom ;
					y2 = y2 < 0. ? 1. : 1.0 - y2/height ;
					mImageSelectionRect.left = x1 ;
					mImageSelectionRect.top = y1 ;
					mImageSelectionRect.right = x2 ;
					mImageSelectionRect.bottom = y2 ;
					
					delete mImageSelection ;
					mImageSelection = nil ;
					DrawExperimentalImage() ;
					DrawStdDevImage() ;
					mCompareSelection->Disable() ;
					mSetNewSelection->Disable() ;
				}
		}
		else if (obj == mSetNewSelection) {
				if (mImageSelection != nil) {
					Rect selectionRect ;
					tRect = mExpImageFrame->GetRect() ;
					mImageSelection->GetSelectionRect(selectionRect) ;
					// Compute the actual rectange mImageSelectionRect
					Rect sizeOfImage = mActualImageRect ;
					OffsetRect(&sizeOfImage,tRect.left,tRect.top) ;
					float x1,y1,x2,y2,width,height ;
					width = sizeOfImage.right - sizeOfImage.left ;
					height = sizeOfImage.bottom - sizeOfImage.top ;
					x1 = selectionRect.left - sizeOfImage.left ;
					x1 = x1 < 0. ? 0. : x1/width ;
					y1 = selectionRect.top - sizeOfImage.top ;
					y1 = y1 < 0. ? 0. : y1/height ;
					x2 = sizeOfImage.right - selectionRect.right ;
					x2 = x2 < 0. ? 1. : 1.0 - x2/width ;
					y2 = sizeOfImage.bottom - selectionRect.bottom ;
					y2 = y2 < 0. ? 1. : 1.0 - y2/height ;
					mImageSelectionRect.left = x1 ;
					mImageSelectionRect.top = y1 ;
					mImageSelectionRect.right = x2 ;
					mImageSelectionRect.bottom = y2 ;
					
					delete mImageSelection ;
					mImageSelection = nil ;
					DrawExperimentalImage() ;
					DrawStdDevImage() ;
					mCompareSelection->Disable() ;
					mSetNewSelection->Disable() ;
				}
		}
		else if (obj == mRealCCC && mSelectedMethod != mRealCCC) {
			mSelectedMethod->SetControlValue(false) ;
			if (mSelectedMethod == mFTCCC) {
				mExactFTCCC->Disable() ;
				mAmplitudeFTCCC->Disable() ;
				mEstimateShift->Disable() ;
			}
			mSelectedMethod = (TRadioButton *)obj ;
		}
		else if (obj == mFTCCC && mSelectedMethod != mFTCCC) {
			mSelectedMethod->SetControlValue(false) ;
			mSelectedMethod = (TRadioButton *)obj ;
			mExactFTCCC->Enable() ;
			mAmplitudeFTCCC->Enable() ;
			if (mAmplitudeFTCCC->GetControlValue())
				mEstimateShift->Enable() ;
		}
		else if (obj == mChiSquare && mSelectedMethod != mChiSquare) {
			mSelectedMethod->SetControlValue(false) ;
			if (mSelectedMethod == mFTCCC) {
				mExactFTCCC->Disable() ;
				mAmplitudeFTCCC->Disable() ;
				mEstimateShift->Disable() ;
			}
			mSelectedMethod = (TRadioButton *)obj ;
		}
		else if (obj == mRMSD && mSelectedMethod != mRMSD) {
			mSelectedMethod->SetControlValue(false) ;
			if (mSelectedMethod == mFTCCC) {
				mExactFTCCC->Disable() ;
				mAmplitudeFTCCC->Disable() ;
				mEstimateShift->Disable() ;
			}
			mSelectedMethod = (TRadioButton *)obj ;
		}
		else if (obj == mDiffImage && mSelectedMethod != mDiffImage) {
			mSelectedMethod->SetControlValue(false) ;
			if (mSelectedMethod == mFTCCC) {
				mExactFTCCC->Disable() ;
				mAmplitudeFTCCC->Disable() ;
				mEstimateShift->Disable() ;
			}
			mSelectedMethod = (TRadioButton *)obj ;
		}
		else if (obj == mFMAD && mSelectedMethod != mFMAD) {
			mSelectedMethod->SetControlValue(false) ;
			if (mSelectedMethod == mFTCCC) {
				mExactFTCCC->Disable() ;
				mAmplitudeFTCCC->Disable() ;
				mEstimateShift->Disable() ;
			}
			mSelectedMethod = (TRadioButton *)obj ;
		}
		else if (obj == mExactFTCCC) {
			if (mAmplitudeFTCCC->GetControlValue()) 
					mAmplitudeFTCCC->SetControlValue(false) ;
			mEstimateShift->Disable() ;
		}
		else if (obj == mAmplitudeFTCCC) {
			if (mExactFTCCC->GetControlValue()) 
					mExactFTCCC->SetControlValue(false) ;
			mEstimateShift->Enable() ;
		}
		else if (obj == mArrowButton) {
			switch(mCurrentTool) {
				case kCompareSelection:
					mRectButton->SetState(false) ;
					break ;
				case kCompareMagnify:
					mMagnifyButton->SetState(false) ;
					break ;
				case kCompareHand:
					mHandButton->SetState(false) ;
					break ;
			}
			mCurrentTool = kComparePointer ;
			mArrowButton->SetState(true) ;
		}
		else if (obj == mRectButton) {
			switch(mCurrentTool) {
				case kComparePointer:
					mArrowButton->SetState(false) ;
					break ;
				case kCompareMagnify:
					mMagnifyButton->SetState(false) ;
					break ;
				case kCompareHand:
					mHandButton->SetState(false) ;
					break ;
			}
			mCurrentTool = kCompareSelection ;
			mRectButton->SetState(true) ;
		}
		else if (obj == mMagnifyButton) {
			switch(mCurrentTool) {
				case kCompareSelection:
					mRectButton->SetState(false) ;
					break ;
				case kComparePointer:
					mArrowButton->SetState(false) ;
					break ;
				case kCompareHand:
					mHandButton->SetState(false) ;
					break ;
			}
			mCurrentTool = kCompareMagnify ;
			mMagnifyButton->SetState(true) ;
		}
		else if (obj == mHandButton) {
			switch(mCurrentTool) {
				case kCompareSelection:
					mRectButton->SetState(false) ;
					break ;
				case kCompareMagnify:
					mMagnifyButton->SetState(false) ;
					break ;
				case kComparePointer:
					mArrowButton->SetState(false) ;
					break ;
			}
			mCurrentTool = kCompareHand ;
			mHandButton->SetState(true) ;
		}
		else if (obj == mResetView) {
				mImageCenter.x = (mImageRect.right-mImageRect.left)/2 ;
				mImageCenter.y =  (mImageRect.bottom-mImageRect.top)/2 ;
				mStdDevCenter.x = (mStdDevRect.right-mStdDevRect.left)/2 ;
				mStdDevCenter.y =  (mStdDevRect.bottom-mStdDevRect.top)/2 ;
				mImageMagnification = 1.0 ;
				PositionImages() ;
				DrawExperimentalImage() ;
				DrawStdDevImage() ;
				mResetView->Disable() ;
		}
		else if (obj == mExpImageFrame) {
			DoMouseDownInImage(*pt) ;
		}
		else if (obj == mOK) {
			// Get and assemble all the relevant parameters and
			// compare the images
			if (mExpImage != nil)
				DeleteImage(mExpImage) ;	
			mExpImage = nil ;
			if (mStdDevImage != nil)
				DeleteImage(mStdDevImage) ;	
			mStdDevImage = nil ;
			if (mRealCCC->GetControlValue())
				gComparisonMethod = kXCF ;
			else if (mFTCCC->GetControlValue())
				gComparisonMethod = kFTXCF ;
			else if (mChiSquare->GetControlValue())
				gComparisonMethod = kChiSquare ;
			else if (mRMSD->GetControlValue())
				gComparisonMethod = kRMSD ;
			else if (mDiffImage->GetControlValue())
				gComparisonMethod = kDifferenceImage ;
			else if (mFMAD->GetControlValue())
				gComparisonMethod = kFMAD ;

			gEstimateImageShift = mEstimateShift->GetControlValue() ;
			gCalculateExactFTXCF = mExactFTCCC->GetControlValue() ;
			gCompareEntireImage = mCompareEntireImage->GetControlValue() ;
			gComparisonRect = mImageSelectionRect ;
			CompareExperimentalImageWithSimulation() ;
		}		
		else if (obj == mCancel) {
			if (mExpImage != nil)
				DeleteImage(mExpImage) ;	
			mExpImage = nil ;
			if (mStdDevImage != nil)
				DeleteImage(mStdDevImage) ;	
			mStdDevImage = nil ;
			Hide() ;
			ExitFromComparison(this) ;
		}		
	}
} 

void	
TCompareImagesPage::SizeImages() {
}

void
TCompareImagesPage::PositionImages() {
	
	short width = mExpImage->mWidth*mImageMagnification ;
	short height = mExpImage->mHeight*mImageMagnification ;
	
	SetRect(&mActualImageRect,mImageCenter.x - width/2,
		mImageCenter.y - height/2,
		mImageCenter.x + width/2,
		mImageCenter.y + height/2) ;

	SetRect(&mActualStdDevRect,mStdDevCenter.x - width/2,
		mStdDevCenter.y - height/2,
		mStdDevCenter.x + width/2,
		mStdDevCenter.y + height/2) ;
}

void		
TCompareImagesPage::DoMouseDownInImage(Point &loc) {
	short 	xc,yc,dx,dy ;
	float	zoomfactor ;
	
	switch (mCurrentTool) {
		case kComparePointer:			
			if (mImageSelection != nil) {
				mImageSelection->Modify(loc) ;
			}
			break ;
		case kCompareMagnify:
			float factor ;
			if (::OptionKeyDown()) {
				if (::ShiftKeyDown()) {
					factor = 2.0 ;
				}
				else {
					factor = 1.1 ;
				}
				
			}
			else {
				if (::ShiftKeyDown()) {
					factor = 0.5 ;
				}
				else {
					factor = 1./1.1 ;
				}
			}
			xc = (mImageRect.right-mImageRect.left)/2 ;
			yc = (mImageRect.bottom-mImageRect.top)/2 ;
			mImageMagnification /= factor ;
			zoomfactor = 1./factor ;
			dx = xc + (mImageCenter.x - xc)/factor - mImageCenter.x ;
			dy = yc + (mImageCenter.y - yc)/factor - mImageCenter.y ;
			mImageCenter.x = xc + (mImageCenter.x - xc)/factor ;
			mImageCenter.y = yc + (mImageCenter.y - yc)/factor ;
			xc = (mStdDevRect.right-mStdDevRect.left)/2 ;
			yc = (mStdDevRect.bottom-mStdDevRect.top)/2 ;
			mStdDevCenter.x = xc + (mStdDevCenter.x - xc)/factor ;
			mStdDevCenter.y = yc + (mStdDevCenter.y - yc)/factor ;
			SizeImages() ;
			PositionImages() ;
			DrawExperimentalImage() ;
			DrawStdDevImage() ;
			if (mImageSelection != nil) {
				mImageSelection->Zoom(zoomfactor,dx,dy) ;
				mImageSelection->Update(&mImageRect) ;
			}			
			if (!mResetView->IsEnabled()) mResetView->Enable() ;
			break ;
		case kCompareHand:	
			// Let us try to pan the view
			Point lastPoint,currentPoint ;
			long dx,dy ;
			lastPoint = loc ;
			currentPoint = loc ;
			while(Button()) {
				GetMouse(&currentPoint) ;
				dx = currentPoint.x - lastPoint.x ;
				dy = currentPoint.y - lastPoint.y ;
				if (dx != 0 || dy != 0) {
					mImageCenter.x += dx ;
					mImageCenter.y += dy ;
					mStdDevCenter.x += dx ;
					mStdDevCenter.y += dy ;
					PositionImages() ;
					DrawExperimentalImage() ;
					DrawStdDevImage() ;
					if (mImageSelection != nil) {
						mImageSelection->Pan(dx,dy) ;
						mImageSelection->Update(&mImageRect) ;
					}			
					lastPoint = currentPoint ;
				}
			}
			if (lastPoint.x != loc.x || lastPoint.y != loc.y) {
				if (!mResetView->IsEnabled()) 	mResetView->Enable() ;
			}
			break ;		
		case kCompareSelection:
			InsetRect(&mImageRect,1,1) ;

			if (mImageSelection != nil) {
				mImageSelection->DoMouseDown(loc) ;
			}
			else {
				mImageSelection = new CSelection(&mImageRect) ;
				mImageSelection->DoMouseDown(loc) ;
			}
			if (mImageSelection->IsNullSelection()) {
				delete mImageSelection ;
				mImageSelection = nil ;
			}
			InsetRect(&mImageRect,-1,-1) ;
			if (mImageSelection != nil) {
				if (mCompareSelection->GetControlValue())
					mSetNewSelection->Enable() ;
				else
					mCompareSelection->Enable() ;
			}
			else 
				mCompareSelection->Disable() ;
			break ;
	}
}

void	ExitFromComparison(TCompareImagesPage *thePage) {
	set_menu(QUANTITATIVE) ;
	ClearDisplay() ;
	SetPage(gMainPage) ;
	DrawQuantitativePage() ;
}
