// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "TImage.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "globals.h"
#include "funcdefs.h"	
#include "types.h"
#include "PhasegratingRecord.h"
#include "ProjCellRecord.h"
#include "MslicRecord.h"
#include "displ.h"
#include "filenames.h"
#include "files.h"
#include "errors.h"
#include "storage.h"
#include "dfpdspl.h"
#include "tiff.h"
#include "TFile.h"

#include "TImageWindow.h"
#include "TImportRawImageDialog.h"

extern struct { float wavel ; short nslice ; float sigma ; } gMSRec ;
extern char stext[] ;
extern short gComplexInputFile,gComplexOutputFile ;
void GetOutputImageDimensions(long *width,long *height) ;
extern long NUM_DEFOC,NUM_THICK ;
extern long CreateOffScreenRect(OffScreenPtr *offScrPtr,GrafPtr orgPort,Rect *rect) ;

void WindowToFile(TGrafPort *window) ;

void	DrawExpImageDisplayArea() ;
void	DrawStdDevImageDisplayArea() ;
Boolean CreateImageFromTiffFile(char *,TImage **) ;
TImage *ImportRawFile(char	*fileName) ;
TImage *ImportTiffFile(char	*fileName) ;
void	CopyImage(TImage *image1, TImage *image2) ;
float	ImageStandardDeviation(TImage *image,float mean) ;
long	AskSaveResult(char *) ;
void 	GetNM(long *width,long *height) ;
void DisplayImageInWindow(TImage *image, TWindow *window, Rect *rect,Boolean scaleToExpImageRatio) ;

extern float	gAspectRatio ;

Boolean IsPowerOfTwo(long value) ;
#define log2(x)	log(x)/log(2.0)

extern short gXRes,gYRes ;
typedef enum {kFileIsText,kFileIsBinary}
ImportFileType ;

static ImageDataType	gImageDataType = k4ByteReal;
static Boolean			gSwapBytes = false ;
static ImportFileType	gImportFileType = kFileIsBinary ;
long				gImportImageFileSize ;
short				gImportImageWidth ;
short				gImportImageHeight ;
short				gImportImageOffset ;
static long			gByteOffset =  0 ;
char				gExperimentalImageName[64] = {'\0'} ;
extern	TImage			*gExperimentalImage;
extern	TImage			*gStandardDeviationImage;

static 	Boolean			gIsPower2Sampled = false ;
extern  TiffRecord		gTiff ;

void	AddWindow(TWindow *window) ;
Point gImageDisplayPosition = {50,50} ;
float gMeanOfExperiment ;
float gSigmaOfExperiment ;

void LoadExperimentalImage() ;
void LoadStandardDeviationImage() ;
void CompareWithSimulation() ;
void CompareAndRefine() ;
Boolean AdjustImageSize(TImage **image,long width,long height) ;
Boolean HasExperimentalImage() ;
Boolean HasStandardDeviationImage() ;
Boolean HasExperimentalImage() {
	return (gExperimentalImage != nil) ;
}

Boolean HasStandardDeviationImage() {
	return (gStandardDeviationImage != nil) ;
}

void LoadExperimentalImage() {
	TImage *temp ;
	extern COMMAND c_compareImageWithSimulation,c_refineParameters,c_refineStructure ;
	
	temp = ImportImage() ;
	if (temp != nil) {
		if (gExperimentalImage != nil) {
			DeleteImage(gExperimentalImage) ;
		}
		gExperimentalImage = temp ;
		//DisplayImage(gExperimentalImage) ;
		DrawExpImageDisplayArea() ;
		gMeanOfExperiment = ImageMean(gExperimentalImage) ;
		gSigmaOfExperiment = ImageStandardDeviation(gExperimentalImage,
			gMeanOfExperiment) ;
		if (c_compareImageWithSimulation.disabled) {
			EnableControl(&c_compareImageWithSimulation) ;
			EnableControl(&c_refineParameters) ;
			EnableControl(&c_refineStructure) ;
		}
	}
}

void LoadStandardDeviationImage() {
	TImage *temp ;
	
	temp = ImportImage() ;
	if (temp != nil) {
		if (gStandardDeviationImage != nil) {
			DeleteImage(gStandardDeviationImage) ;
		}
		gStandardDeviationImage = temp ;
		//DisplayImage(gStandardDeviationImage) ;
		DrawStdDevImageDisplayArea() ;
	}
}

Boolean ReadImageFromCalculation(TImage *image,short whichImage,Boolean isSimulationSampled) {
	float	*data ; 
	long	jmax ;
	long	width,height ;
	long	na,nb ; 
	
	
	data = gWorkSpace ;
	jmax = gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;
	
	if(!read_input_file(whichImage+1,(char *)data,jmax)) {
		strcat(gErrorText,flname.iout) ;
		return false;
	}
	image->mData = data ;
		
	width = gPhsgrtRec->nbeam ;
	height = gPhsgrtRec->mbeam ;
	na = 1 ;
	nb = 1 ;
	
	if (!isSimulationSampled) {
		/* Adjust the number of plotting points in the 2 directions */
		AdjustSampling(&width,&height,na,nb) ;
		image->mSimulationSampling = false ;
	}
	else {
		image->mSimulationSampling = true ;
	}
	
	image->mWidth = width ;
	image->mHeight = height ;
	image->mMeanIsZero = false ;
	image->mName = nil ;
	image->mIsDirty = false ;
	image->mIsRealSpace = true ;
	image->mGamma = gPhsgrtRec->gamma ;
	
	// Because the calculated images are sored with the bottom row first
	// instead of the normal convention of the top row, we need to change the
	// rows so that the simulated images have the same logic of the ccordinate system (y-down)
	FlipImage(image) ;
	
	return true ;
}

TImage *CreateImageFromData(short width,short height,void *data,ImageDataType dataType) {
	TImage *image ;
	long	size,numElements ;
	register	long i ;
	float	*p1 ;
	
	image = (TImage *)malloc(sizeof(TImage)) ;
	if (image == nil) {
		return image ;
	}
	image->mWidth = width ;
	image->mHeight = height ;
	size = width*height*sizeof(float) ;
	image->mName = nil ;
	image->mOffscreen = nil ;
	image->mIsVisible = false ;

	image->mData = (float *)malloc(size) ;
	if (image->mData == nil) {
		free(image) ;
		return nil ;
	}
	numElements = width*height ;
	switch (dataType) {
		case k1ByteSigned:
			char *cp ;
			for(i=0,cp=(char *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *cp++ ;
			}
			break ;
		case k1ByteUnsigned:
			unsigned char *ucp ;
			for(i=0,ucp=(unsigned char *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *ucp++ ;
			}
			break ;
		case k2ByteSigned:
			short *sp ;
			for(i=0,sp=(short *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *sp++ ;
			}
			break ;
		case k2ByteUnsigned:
			unsigned short *usp ;
			for(i=0,usp=(unsigned short *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *usp++ ;
			}
			break ;
		case k4ByteSigned:
			long *lp ;
			for(i=0,lp=( long *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *lp++ ;
			}
			break ;
		case k4ByteUnsigned:
			unsigned long *ulp ;
			for(i=0,ulp=(unsigned long *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *ulp++ ;
			}
			break ;
		case k4ByteReal:
			memcpy(image->mData,data,size) ;
			break ;
		case k8ByteReal:
			double *dp ;
			for(i=0,dp=(double *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *dp++ ;
			}
			break ;
	}
	image->mMeanIsZero = false ;
	image->mIsDirty = false ;
	image->mIsRealSpace = true ;
	return image ;
}

void	CopyImage(TImage *image0, TImage *image) {
	long	size ;
	image->mWidth = image0->mWidth ;
	image->mHeight = image0->mHeight ;
	size = image0->mWidth*image0->mHeight*sizeof(float) ;
	image->mIsVisible = false ;
	memcpy(image->mData,image0->mData,size) ;
	image->mMeanIsZero = image0->mMeanIsZero ;
	image->mMinValue = image0->mMinValue ;
	image->mMaxValue = image0->mMaxValue ;
	image->mSimulationSampling = image0->mSimulationSampling ;
	image->mGamma = image0->mGamma ;
	image->mWindow = nil ;
	image->mOffscreen = nil ;
	image->mIsDirty = false ;	// It has to be forced to true 
	image->mIsRealSpace = image0->mIsRealSpace ;
}

TImage *CopyImage(TImage *image0) {
	TImage *image ;
	long	size ;
	
	image = (TImage *)malloc(sizeof(TImage)) ;
	if (image != nil) {
		image->mWidth = image0->mWidth ;
		image->mHeight = image0->mHeight ;
		size = image0->mWidth*image0->mHeight*sizeof(float) ;
		image->mData = (float *)malloc(size) ;
		if (image->mData != nil) {
			image->mIsVisible = false ;
			image->mName = nil ;
			memcpy(image->mData,image0->mData,size) ;
			image->mMeanIsZero = image0->mMeanIsZero ;
			image->mMinValue = image0->mMinValue ;
			image->mMaxValue = image0->mMaxValue ;
			image->mSimulationSampling = image0->mSimulationSampling ;
			image->mGamma = image0->mGamma ;
			image->mWindow = nil ;
			image->mOffscreen = nil ;
			image->mIsDirty = false ;	// It has to be forced to true 
			image->mIsRealSpace = image0->mIsRealSpace ;

		}
		else {
			free(image) ;
			image = nil ;
		}
	}
	return image ;
}

TImage *CreateImageFromImageArea(short width,short height,short dx,short dy,TImage *area) {
	TImage *image ;
	long	size,row ;
	float	*data ;
	
	image = (TImage *)malloc(sizeof(TImage)) ;
	if (image != nil) {
		image->mWidth = width ;
		image->mHeight = height ;
		size = width*height*sizeof(float) ;
		image->mData = (float *)malloc(size) ;
		if (image->mData != nil) {
			image->mIsVisible = false ;
			image->mName = nil ;
			image->mMeanIsZero = false ;
			image->mWindow = nil ;
			image->mOffscreen = nil ;
			image->mIsDirty = false ;	// It has to be forced to true 
			image->mGamma = 90. ;
			if (image->mData == nil) {
				free(image) ;
				return nil ;
			}
			size = width*sizeof(float) ;
			for(row=0;row < height;row++) {
				data = area->mData+(row+dy)*area->mWidth+dx ;
				memcpy(image->mData,data,size) ;
			}
		}
		else {
			free(image) ;
			image = nil ;
		}
	}
	return image ;
}

float ImageMean(TImage *image) {
	float mean,*rp ;
	long i, sizeofImage ;
	
	sizeofImage = image->mWidth*image->mHeight ;
	mean = 0. ;
	for(i=0,rp=image->mData;i<sizeofImage;i++) {
		mean += *rp++ ;
	}
	mean /= sizeofImage ;
	return mean ;
}

float ImageStandardDeviation(TImage *image,float mean) {
	float sigma,*rp,delta ;
	long i, sizeofImage ;
	
	sizeofImage = image->mWidth*image->mHeight ;
	sigma = 0. ;
	for(i=0,rp=image->mData;i<sizeofImage;i++,rp++) {
		delta = *rp - mean ;
		sigma += delta*delta ;
	}
	sigma /= sizeofImage ;
	sigma = sqrt(sigma) ;
	return (sigma) ;
}

float ImageMeanOfArea(TImage *image,FRect *rect) {
	float mean,*rp ;
	long i,j, sizeofImageArea,top,left,right,bottom ;
	
	top = (long)(image->mHeight*rect->top) ;
	left = (long)(image->mWidth*rect->left) ;
	bottom = (long)(image->mHeight*rect->bottom) ;
	right = (long)(image->mWidth*rect->right) ;
	sizeofImageArea = 0 ;
	mean = 0. ;
	for(j=top; j < bottom;j++) {
		rp = image->mData + j*image->mWidth + left ;
		for(i=left; i < right; i++) {
			sizeofImageArea++ ;
			mean += *rp++ ;
		}
	}
	mean /= sizeofImageArea ;
	return mean ;
}

TImage *ImportRawFile(char	*fileName) {
	TImage	*image ;

	TImportRawImageDialog *theDialog ;
	TImportRawImageDialogRec theDialogData ;


	long	numElements,simulationWidth,simulationHeight ;
	unsigned char	*dataPtr,*filePtr ;
	long	i ;

	image = (TImage *)nil ;

	TFile *theFile = new TFile(fileName,kRead) ;

	theDialog = new TImportRawImageDialog(fileName,theFile->GetSize()) ;
	if (!theDialog->DoModalDialog()) {
		delete theFile ;
		delete theDialog ;
		return image ;
	}
	theDialog->GetDialogData(&theDialogData) ;
	delete theDialog ;



	gSwapBytes = theDialogData.swapBytes ;
	gImportImageWidth = theDialogData.imageWidth ;
	gImportImageHeight = theDialogData.imageHeight ;
	gImageDataType = theDialogData.dataType ;
	strcpy(gExperimentalImageName,fileName) ;

	
	gSwapBytes = theDialogData.swapBytes ;
	gImportImageWidth = theDialogData.imageWidth ;
	gImportImageHeight = theDialogData.imageHeight ;
	gImageDataType = theDialogData.dataType ;
	gByteOffset = theDialogData.byteOffset ;

	filePtr = new unsigned char[theDialogData.numBytes] ;
	theFile->Read(filePtr,theDialogData.numBytes) ;
		
	dataPtr = filePtr + theDialogData.byteOffset ;
	
	unsigned long *uulp = (unsigned long *)dataPtr ;
	unsigned short *uusp = (unsigned short *)dataPtr ;
	short *sp ;
	unsigned short *usp ;
	long *lp ;
	unsigned long *ulp ;
	float	*p1,*p2 ;

	numElements = gImportImageWidth*gImportImageHeight ;
	if (gSwapBytes) {
		switch(gImageDataType) {
			case k1ByteSigned:
			case k1ByteUnsigned:
				break ;
			case k2ByteSigned:
				sp = (short *)dataPtr ;
				for(i=0; i < numElements;i++) {
					*sp++ = ByteSwapShort(*uusp++) ;
				}
				break ;
			case k2ByteUnsigned:
				usp = (unsigned short *)dataPtr ;
				for(i=0; i < numElements;i++) {
					*usp++ = ByteSwapShort(*uusp++) ;
				}
				break ;
			case k4ByteSigned:
				lp = (long *)dataPtr ;
				for(i=0; i < numElements;i++) {
					*lp++ = ByteSwapLong(*uulp++) ;
				}
				break ;
			case k4ByteUnsigned:
				ulp = (unsigned long *)dataPtr ;
				for(i=0; i < numElements;i++) {
					*ulp++ = ByteSwapLong(*uulp++) ;
				}
				break ;
			case k4ByteReal:
				p1 = (float *)dataPtr ;
				for(i=0; i < numElements;i++) {
					*p1++ = ByteSwapFloat(*uulp++) ;
				}
				break ;
			case k8ByteReal:
				p2 = (float *)dataPtr ;	// Probably wrong
				for(i=0; i < 2*numElements;i++) {
					*p2++ = ByteSwapFloat(*uulp++) ;
				}
				break ;
		}
	}
	image = CreateImageFromData(gImportImageWidth,gImportImageHeight,dataPtr,gImageDataType) ;
	delete theFile ;
	delete[] filePtr ;
	
	
	GetNM(&simulationWidth,&simulationHeight) ;
	image->mSimulationSampling = 
		(gImportImageWidth == simulationWidth && 
		gImportImageHeight == simulationHeight) ;

	image->mName = nil ;
	image->mWindow = nil ;
	image->mOffscreen = nil ;
	image->mIsDirty = false ;
	SetImageName(image,gExperimentalImageName) ;
	image->mIsVisible = false ;

	return image ;
}


TImage *ImportTiffFile(char	*fileName) {
	TImage *image = (TImage *)nil ;

	if (CreateImageFromTiffFile(fileName,&image)) {
		
		gIsPower2Sampled = false ;
		if (IsPowerOfTwo(gTiff.width) && IsPowerOfTwo(gTiff.height)) {
			gIsPower2Sampled = true ;
		}
		gImportImageWidth = gTiff.width ;
		gImportImageHeight = gTiff.height ;
		
		long simulationWidth,simulationHeight ;
		GetNM(&simulationWidth,&simulationHeight) ;
		image->mSimulationSampling = (gImportImageWidth == simulationWidth && gImportImageHeight == simulationHeight) ;
		
		image->mName = nil ;
		image->mWindow = nil ;
		image->mOffscreen = nil ;
		image->mIsDirty = false ;
		SetImageName(image,fileName) ;
		image->mIsVisible = false ;
	}
	else {
		PutMessage("Failed in creating the Image from the TIFF file") ;
	}
	return image ;
}

TImage *ImportImage() {
	FSInfo	fsInfo ;
	TImage	*image ;
	char	name[256] ;
	
	image = (TImage	*)nil ;
	
	ChooseFile("Choose Image File","*",fsInfo) ;
	if (fsInfo.mReply) {
		// If the filetype is TIFF then pass the file to ImportTiffFile
		// get the file name at this point as well
		// Check to see if it is a tiff file
		strcpy(name,fsInfo.mFileName) ;
		Boolean isTiffFile = false ;
		// See if the file has the extension ".tif" or ".TIF" as from a PC
		char *cp = strrchr(name,'.') ;  // Does the name have a . in it?
		if (cp != NULL) {
			// If YES ! Check to se if it is of type .tif or .TIF
			if (strncmp(cp,".tif",4)==0) {
				isTiffFile = true ;
			}
			else if (strncmp(cp,".TIF",4)==0) {
					isTiffFile = true ;
			}
		}
		if (isTiffFile) {
			// Set the vol-ref num to the appropriate location
			image = ImportTiffFile(fsInfo.mFileName) ;
		}
		else {	// Try to import the file as a raw data-file
			image = ImportRawFile(fsInfo.mFileName) ;
		}
	}
	return image ;

}

void FillImageFromImageArea(short width,short height,short dx,short dy,TImage *area,TImage *image) {
	long	size,row ;
	float	*data,*p1 ;
	
	size = width*sizeof(float) ;
	for(row=0,p1 = image->mData;row < height;row++) {
		data = area->mData+(row+dy)*area->mWidth+dx ;
		memcpy(p1,data,size) ;
		p1 += width ;
	}
	image->mMeanIsZero = false ;
}

void FlipImage(TImage *image) {
	register long row ;
	long	 numElements ; 
	float	*tempRow ;
	float	*lowRow,*highRow ;
	unsigned long	numBytesToMove ;
	
	tempRow = (float *)malloc(image->mWidth*sizeof(float)) ;
	numBytesToMove = image->mWidth*sizeof(float) ;
	if (tempRow != nil) {
		numElements = image->mWidth*image->mHeight ;
		lowRow = image->mData ;
		highRow = image->mData + (image->mHeight-1)*image->mWidth ;
		for(row=0;row < image->mHeight/2;row++) {
			memcpy(tempRow,lowRow,numBytesToMove) ;
			memcpy(lowRow,highRow,numBytesToMove) ;
			memcpy(highRow,tempRow,numBytesToMove) ;
			highRow -= image->mWidth ;
			lowRow += image->mWidth ;
		}
		free(tempRow) ;
	}
}

void SetImageData(TImage *image,float *data,ImageDataType dataType) {
	long		size,numElements ;
	register	long i ;
	float		*p1 ;
	
	numElements = image->mWidth*image->mHeight ;
	size = numElements*sizeof(float) ;
	switch (dataType) {
		case k1ByteSigned:
			char *cp ;
			for(i=0,cp=(char *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *cp++ ;
			}
			break ;
		case k1ByteUnsigned:
			unsigned char *ucp ;
			for(i=0,ucp=(unsigned char *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *ucp++ ;
			}
			break ;
		case k2ByteSigned:
			short *sp ;
			for(i=0,sp=(short *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *sp++ ;
			}
			break ;
		case k2ByteUnsigned:
			unsigned short *usp ;
			for(i=0,usp=(unsigned short *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *usp++ ;
			}
			break ;
		case k4ByteSigned:
			long *lp ;
			for(i=0,lp=( long *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *lp++ ;
			}
			break ;
		case k4ByteUnsigned:
			unsigned long *ulp ;
			for(i=0,ulp=(unsigned long *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *ulp++ ;
			}
			break ;
		case k4ByteReal:
			memcpy(image->mData,data,size) ;
			break ;
		case k8ByteReal:
			double *dp ;
			for(i=0,dp=(double *)data,p1 = image->mData; i < numElements;i++) {
				*p1++ = *dp++ ;
			}
			break ;
	}
	image->mMeanIsZero = false ;

}

TImage *CreateNewImage(short width,short height) {
	TImage *image ;
	long	size,i,length ;
	float	*fp ;
	
	image = (TImage *)malloc(sizeof(TImage)) ;
	if (image != nil) {
		image->mWidth = width ;
		image->mHeight = height ;
		size = width*height*sizeof(float) ;
		image->mData = (float *)malloc(size) ;
		if (image->mData == nil) {
			free(image) ;
			return nil ;
		}
		image->mName = nil ;
		image->mMeanIsZero = false ;
		image->mIsVisible = false ;
		image->mIsDirty = false ;	// It has to be forced to true 
		image->mWindow = nil ;
		image->mOffscreen = nil ;
		
		// Zero the image
		fp = image->mData ;
		length = width*height ;
		for( i=0; i < length ; i++) {
			*fp++ = 0. ;
		}
	}
	return image ;
}

void ZeroMeanOfImage(TImage *image) {
	float mean ;
	register long i ;
	register float *p1 ;
	long length ;
	
	if (image->mMeanIsZero)
		return ;
	
	mean = 0. ;
	length = image->mWidth*image->mHeight ;
	// Find the mean of the image
	for(i=0,p1 = image->mData;i<length;i++) {
		mean += *p1++ ;
	}
	
	mean /= length ;
	
	// Subtract the mean
	for(i=0,p1 = image->mData;i<image->mWidth*image->mHeight;i++) {
		*p1 -= mean ;
		p1++ ;
	}
	image->mMeanIsZero = true ;
}

void ZeroMeanOfImageArea(TImage *image,FRect *rect) {
	float mean,*rp ;
	long i,j,top,left,right,bottom ;
	
	mean = ImageMeanOfArea(image,rect) ;
	top = (long)(image->mHeight*rect->top) ;
	left = (long)(image->mWidth*rect->left) ;
	bottom = (long)(image->mHeight*rect->bottom) ;
	right = (long)(image->mWidth*rect->right) ;
	for(j=top; j < bottom;j++) {
		rp = image->mData + j*image->mWidth + left ;
		for(i=left; i < right; i++) {
			*rp = *rp - mean ;
			rp++ ;
		}
	}
}

Boolean FourierTransformImage(TImage *image) {
	float *workSpace,*rp1,*rp2,i ;
	long lengthOfData,numBytes ;
	float	time ;
	
	lengthOfData = image->mWidth*image->mHeight ;
	numBytes = 2*lengthOfData*sizeof(float) ;
	workSpace = (float *)malloc(numBytes) ;
	if (workSpace == NULL) 
		return false ;
		
	rp2 = workSpace ;
	rp1 = image->mData ;	
	for (i=0;i<lengthOfData;i++) {
	  	*rp2++ = *rp1++ ;
	  	*rp2++ = 0. ;
	}

	mmfft(workSpace,image->mWidth,image->mHeight,&time) ;
	memcpy(image->mData,workSpace,numBytes/2) ;
	image->mIsRealSpace = false ;
	free(workSpace) ; 
	return true ;
	
}

Boolean AdjustImageSize(TImage **image,long width,long height) {
	long i,j,x1,y1,nextRowOffset,nextColOffset,oldWidth,oldHeight ;
	float	x,y,dx,dy,ddx,ddy,*pp,*p1 ;
	char	imageName[256] ;
	Boolean	hasName = false ;
	
	// scale the image to new width and height
	if ((*image)->mWidth == width && (*image)->mHeight == height)
		return(true) ;	// nothing to do
		
	if ((*image)->mName != NULL) {
		strcpy(imageName,(*image)->mName) ;
		hasName= true ;
	}
		
	TImage *tempImage = CopyImage(*image) ;
	
	oldWidth = (*image)->mWidth ;
	oldHeight = (*image)->mHeight ;
		
	ddx = (float)(*image)->mWidth ;
	ddx /= width ;
	ddy = (float)(*image)->mHeight ;
	ddy /= height ;

	DeleteImage(*image) ;
	
	*image = CreateNewImage(width,height) ;
	if (*image == nil) {
		PutMessage("Memory allocation error in 'AdjustImageSize'") ;
		return false ;
	}
	if (hasName) {
		SetImageName(*image,imageName) ;
	}
	
	pp = (*image)->mData ;
	
	for(j=0;j<height;j++) {
		y = j*ddy ;
		y1 = (long)y ;
		dy = y-y1 ;
		nextRowOffset = oldWidth ;
		if (y1 == oldHeight-1) {
			nextRowOffset = -y1*tempImage->mWidth;
		}
		for (i=0;i<width;i++) {
			x = i*ddx ;
			x1 = (long)x ;
			dx = x-x1 ;
			nextColOffset = 1 ;
			if (x1 == oldWidth-1) {
				nextColOffset = -x1 ;
			}
			p1 = tempImage->mData + y1*tempImage->mWidth + x1 ;
			*pp  = (1.-dy)*((1-dx)*(*p1) + dx*(*(p1+nextColOffset))) +
						dy*((1-dx)*(*(p1+nextRowOffset))+dx*(*(p1+nextRowOffset+nextColOffset))) ;
			pp++ ;
		}
	}
	DeleteImage(tempImage) ;
	return true ;
}

typedef enum {kSave,kNoSave,kCancelClose} SaveCode ;
SaveCode PromptSave(char *) ;
void DeleteImage(TImage *image) {
	if (image == nil) return ;
	// First find out if we should save the data
	if (image->mIsDirty) {
		SaveCode save ;
		char	text[256] ;
		if (image->mName == nil) {
			strcpy(text,"Do you want to save the image 'untitled'?") ;
		}
		else {
			sprintf(text,"Do you want to save the image '%s'?",image->mName) ;
		}
		save = PromptSave(text) ;
		if (save == kSave) {
			SaveImage(image) ;	// Save
		}
		else if (save == kCancelClose) {	// Cancel
				return ;
		}
	}
	if (image->mData != nil) free (image->mData) ;
	image->mWidth = 0 ;
	image->mHeight = 0 ;
	image->mMeanIsZero = false ;
	if (image->mName != nil) {
		free(image->mName) ;
	}
	if (image->mOffscreen != nil) {
		delete image->mOffscreen ;
	}
	if (image == gExperimentalImage) {
		gExperimentalImage = nil;
	}
	else if (image == gStandardDeviationImage) {
		gStandardDeviationImage = nil;
	}
	free(image) ;
}

Boolean ImageIsModified(TImage *image) {
	return image->mIsDirty ;
}

void SetImageModifiedFlag(TImage *image,Boolean state) {
	image->mIsDirty = state ;
}

void SaveImage(TImage *image) {
	FSInfo 	theFInfo ;
	
	strcpy(theFInfo.mFileName,"") ;
	ChooseFile("Save Image to Tiff file as:","*",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;
	
	register float *srcpt ;
	char	text[256] ;
	FILE *fp ;
	long nBytes ;
	register long i,j ;
	unsigned char *address ;
	unsigned char *dstpt ;

				
	fp = CreateTiffFile(theFInfo.mFileName) ;
	if (fp != nil) {
		WriteTiffDirectory(fp,kGrayScale,image->mWidth,image->mHeight) ;
	}
	else {
		sprintf(text,"Unable to open %s",theFInfo.mFileName) ;
		PutMessage(text) ;
		return ;
	}
	nBytes = image->mWidth ;
	unsigned char *line ;
	float range = image->mMaxValue - image->mMinValue ;
	line = (unsigned char *)malloc(image->mWidth*sizeof(unsigned char)) ;
	if (line != nil) {
		for (i=0;i<image->mHeight;i++) {
			srcpt = image->mData + i*image->mWidth  ;
			dstpt = line ;
			address = dstpt ;
			for(j=0;j<image->mWidth;j++) {
					*dstpt++ = (unsigned char)(255.*(*srcpt - image->mMinValue)/range) ;
					srcpt++ ;
			}
			WriteTiffStrip(fp,address,nBytes) ;
		}
		free(line) ;
	}
	CloseTiffFile(fp) ;
}

void SaveRGBImage(TImage *image) {
	FSInfo 	theFInfo ;
	
	strcpy(theFInfo.mFileName,"") ;
	ChooseFile("Save Image to Tiff file as:","*",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;
	
	register float *srcpt ;
	char	text[256] ;
	FILE *fp ;
	long nBytes ;
	register long i,j,n ;
	unsigned char *address ;
	unsigned char *dstpt ;

				
	fp = CreateTiffFile(theFInfo.mFileName) ;
	if (fp != nil) {
		WriteTiffDirectory(fp,kRGB,image->mWidth,image->mHeight) ;
	}
	else {
		sprintf(text,"Unable to open %s",theFInfo.mFileName) ;
		PutMessage(text) ;
		return ;
	}
	nBytes = image->mWidth ;
	unsigned char *line ;
	float range = image->mMaxValue - image->mMinValue ;
	line = (unsigned char *)malloc(image->mWidth*sizeof(unsigned char)) ;
	if (line != nil) {
		for (i=0;i<image->mHeight;i++) {
			srcpt = image->mData + i*image->mWidth  ;
			dstpt = line ;
			address = dstpt ;
			for(j=0;j<image->mWidth;j++) {
					*dstpt++ = (unsigned char)(255.*(*srcpt - image->mMinValue)/range) ;
					srcpt++ ;
			}
			WriteTiffStrip(fp,address,nBytes) ;
		}
		free(line) ;
	}
	CloseTiffFile(fp) ;
}

void SetImageName(TImage *image,char *name) {
	if (image->mName != nil) {
		free(image->mName) ;
	}
	image->mName = (char *)malloc(strlen(name)+1) ;
	strcpy(image->mName,name) ;
}

void UpdateImage(TImage *image) {
	GrafPtr theOldPort ;
	
	GetPort(&theOldPort) ;
	if (image->mWindow != nil) {
		SetPort(image->mWindow) ;
		Rect theRect = image->mWindow->GetPortRect() ;
		if (image->mOffscreen != nil) {
			CopyBits(image->mOffscreen,
					 image->mWindow,
					&theRect,&theRect,srcCopy) ;
		}	
		SetPort(theOldPort) ;
	}
}


void DisplayImageInRect(TImage *image, Rect *rect) {
	
	GrafPtr	theCurrentPort ;
	Rect	theRect ;
	
	GetPort(&theCurrentPort) ;

	if (image->mOffscreen == nil) {	
		return ;
	}
		

	SetRect(&theRect,0,0,image->mWidth,image->mHeight) ;
	CopyBits(image->mOffscreen,theCurrentPort,&theRect,rect,srcCopy) ;
}

void CreateImageOffscreen(TImage *image, TWindow *window) {
	
	float	minVal,maxVal,*p1,value,range,maxDisplayValue ;
	long i,j ;
	unsigned char *ucp ;
	GrafPtr	tPort ;

	GetPort(&tPort) ;
	
	maxDisplayValue = gNumGrays-1 ;
	
	if (image->mOffscreen == nil) {			
		minVal = 100000. ;
		maxVal = -100000. ;
		p1 = image->mData ;
		for(i=0;i<image->mWidth*image->mHeight;i++) {
			value = *p1++ ;
			if (value < minVal)  minVal = value ;
			if (value > maxVal)  maxVal = value ;
		}
		image->mMinValue = minVal ;
		image->mMaxValue = maxVal ;
		range = (maxVal-minVal) ;
		p1 = image->mData ;
	
		// Create the image values into our workspace
		unsigned char *baseAddress = (unsigned char *)gWorkSpace ;
		ucp = baseAddress ;
		unsigned long *ulp = (unsigned long *)baseAddress ;
		long index ;
		for(i=0;i<image->mHeight;i++) {
			for(j=0;j<image->mWidth;j++) {
				value = maxDisplayValue*((*p1)- minVal)/range ;
				p1++ ;
				index = (long)value ;
				*ulp++ = gLut[index] ;
				*ucp++ = (unsigned char)gLut[index] ;
			}
		}

		image->mOffscreen = new TOffscreen(window) ;

		SetOffscreenPort(image->mOffscreen) ;
		DrawImage(baseAddress,0,0,image->mWidth,image->mHeight) ;
		UnSetOffscreenPort(image->mOffscreen)	;
	}
	SetPort(tPort) ;
}

void WindowToFile(TGrafPort *theWindow) {
	// ------------------------------------------------------------------ 
	// Put up the standard file dialog Box to determine the file name 
	// ------------------------------------------------------------------ 
	FSInfo 	theFInfo ;
	
	strcpy(theFInfo.mFileName,"") ;
	ChooseFile("Save Window to Tiff File as:","*",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;
	GrafPtr tPort ;
	GetPort(&tPort) ;
	SetPort(theWindow) ;
	Rect	theRect = theWindow->GetPortRect() ;
	short	width = theRect.right - theRect.left ;
	short	height = theRect.bottom - theRect.top ;
	if (ReadRGBImage((unsigned char *)gWorkSpace,0,0,width,height)) {
		WriteDisplayToTiff(theFInfo.mFileName,(unsigned char *)gWorkSpace,
					width,height) ;
	}
	SetPort(tPort) ;
}

void DisplayImageInWindow(TImage *image, TWindow *window, Rect *rect,Boolean scaleToExpImageRatio) {
	
	float	minVal,maxVal,*p1,value,range ;
	long i,j ;
	GrafPtr	theOldPort ;
	unsigned char *ucp ;
	Rect	theRect ;
	
	GetPort(&theOldPort) ;
	float	maxDisplayValue = gNumGrays-1 ;

	// Create the offscreen for the image in case it doesn't exist
	if (image->mOffscreen == nil) {			
		minVal = 100000. ;
		maxVal = -100000. ;
		p1 = image->mData ;
		for(i=0;i<image->mWidth*image->mHeight;i++) {
			value = *p1++ ;
			if (value < minVal)  minVal = value ;
			if (value > maxVal)  maxVal = value ;
		}
		image->mMinValue = minVal ;
		image->mMaxValue = maxVal ;
		range = (maxVal-minVal) ;
		p1 = image->mData ;
	
		// Create the image values into our workspace
		unsigned char *baseAddress = (unsigned char *)gWorkSpace ;
		ucp = baseAddress ;
		unsigned long *ulp = (unsigned long *)baseAddress ;
		long index ;
		for(i=0;i<image->mHeight;i++) {
			for(j=0;j<image->mWidth;j++) {
				value = maxDisplayValue*((*p1)- minVal)/range ;
				p1++ ;
				index = (long)value ;
				*ulp++ = gLut[index] ;
				*ucp++ = (unsigned char)gLut[index] ;
			}
		}

		image->mOffscreen = new TOffscreen(window) ;

		SetOffscreenPort(image->mOffscreen) ;
		DrawImage(baseAddress,0,0,image->mWidth,image->mHeight) ;
		UnSetOffscreenPort(image->mOffscreen)	;
	}

	SetPort(window) ;


	Rect itemRect = *rect ;
	float fraction,ratio1,ratio2 ;
	short width = image->mWidth ;
	short height = image->mHeight ;

	ratio1 = itemRect.right-itemRect.left ;
	ratio1 /= width ;
	ratio2 = itemRect.bottom-itemRect.top ;
	ratio2 /= height ;
	fraction = ratio2 < ratio1 ? ratio2 : ratio1 ;
	if (fraction > 1.) fraction = 1. ;
		
	if (fraction != 1.0) {
		width *= fraction ;
		height *= fraction ;
	}

	if (scaleToExpImageRatio && gAspectRatio != 1.0) {
		if (width == gExperimentalImage->mWidth) {
			height = near(width*gAspectRatio) ;
		}
		else if (height == gExperimentalImage->mHeight) {
			width = near(height/gAspectRatio) ;
		}
		else {
			height = near(width*gAspectRatio) ;
		}
	}
	
	Rect theRect2 ;

	SetRect(&theRect,0,0,image->mWidth,image->mHeight) ;
	SetRect(&theRect2,(itemRect.right+itemRect.left - width)/2,
		(itemRect.bottom+itemRect.top - height)/2,
		(itemRect.right+itemRect.left + width)/2,
		(itemRect.bottom+itemRect.top + height)/2) ;
	CopyBits(image->mOffscreen,window,&theRect,&theRect2,srcCopy) ;	
	image->mWindow = window ;
	SetPort(theOldPort) ;
}

void DisplayImage(TImage *image) {

	float	minVal,maxVal,*p1,value,range ;
	long i,j ;
	GrafPtr	theOldPort ;
	unsigned char *ucp ;
	extern 	short	gImageDepth ;

	if(image->mIsVisible)
		return ;

	GetPort(&theOldPort) ;

	TWindow *window = new TImageWindow(gImageDisplayPosition.x,
		gImageDisplayPosition.y,image->mWidth,image->mHeight,
		gImageDepth,image->mName,true,image) ;
	
	AddWindow(window) ;

	
	minVal = 100000. ;
	maxVal = -100000. ;
	p1 = image->mData ;
	for(i=0;i<image->mWidth*image->mHeight;i++) {
		value = *p1++ ;
		if (value < minVal)  minVal = value ;
		if (value > maxVal)  maxVal = value ;
	}
	image->mMinValue = minVal ;
	image->mMaxValue = maxVal ;
	range = (maxVal-minVal) ;
	p1 = image->mData ;
	
	float	maxDisplayValue = gNumGrays-1 ;
	// Create the image values into our workspace
	unsigned char *baseAddress = (unsigned char *)gWorkSpace ;
	ucp = baseAddress ;
	unsigned long *ulp = (unsigned long *)baseAddress ;
	long index ;
	for(i=0;i<image->mHeight;i++) {
		for(j=0;j<image->mWidth;j++) {
			value = maxDisplayValue*((*p1)- minVal)/range ;
			p1++ ;
			index = (long)value ;
			*ulp++ = gLut[index] ;
			*ucp++ = (unsigned char)gLut[index] ;
		}
	}

	image->mOffscreen = new TOffscreen(window) ;

	SetOffscreenPort(image->mOffscreen) ;
	DrawImage(baseAddress,0,0,image->mWidth,image->mHeight) ;
	UnSetOffscreenPort(image->mOffscreen)	;

	SetPort(theOldPort) ;
	gImageDisplayPosition.x += 20 ;
	if (gImageDisplayPosition.x > 1024) {
		gImageDisplayPosition.x = 50 ;
		gImageDisplayPosition.y = 70 ;
	}
	gImageDisplayPosition.y += 20 ;
	if (gImageDisplayPosition.y > 1024) {
		gImageDisplayPosition.x = 70 ;
		gImageDisplayPosition.y = 50 ;
	}
	image->mIsVisible = true ;
	image->mWindow = window ;

}
