#ifndef TImage_h
#define TImage_h
#include "TWindow.h"
typedef struct TImage {
		short mWidth ;
		short mHeight ;
		float mGamma ;
		char  *mName ;
		float mMinValue ;
		float mMaxValue ;
		Boolean mMeanIsZero ;
		Boolean mSimulationSampling ;
		Boolean mIsVisible ;
		Boolean mIsDirty ;
		Boolean mIsRealSpace ;
		float	*mData ;
		TOffscreen	*mOffscreen ;
		TWindow		*mWindow ;
} TImage ;

typedef struct {
	float xlow ;
	float ylow ;
	float zlow ;
	float xhi ;
	float yhi ;
	float zhi ;
} F3Box ;
typedef struct {
	float top ;
	float left ;
	float bottom ;
	float right ;
} FRect ;

typedef enum {k1ByteSigned=1,k1ByteUnsigned,k2ByteSigned,k2ByteUnsigned
	,k4ByteSigned,k4ByteUnsigned,k4ByteReal,k8ByteReal }
ImageDataType ;

float RealSpaceCCCompare(TImage *image1,TImage *image2) ;
float ReciprocalSpaceCCCompare(TImage *image1,TImage *image2) ;
float ComputeImageChiSquare(TImage *image1,TImage *image2,TImage *image3,FRect *rect,long numParameters) ;
float ComputeFMAD(TImage *image1,TImage *image2,FRect *rect) ;
float ComputeMeanSquareDifference(TImage *image1,TImage *image2,FRect *rect) ;
float ComputeResidualImage(TImage *image1,TImage *image2,TImage *image3,FRect *rect,TImage *image4,long nPar) ;
Boolean ReadImageFromCalculation(TImage *image,short whichImage,Boolean resize) ;
Boolean FourierTransformImage(TImage *image) ;
void UpdateImageWindow(WindowPtr window) ;
void UpdateImage(TImage *image) ;
void FlipImage(TImage *image) ;
Boolean ImageIsModified(TImage *image) ;
void SetImageModifiedFlag(TImage *image,Boolean state) ;
void SaveImage(TImage *image) ;
void SaveRGBImage(TImage *image) ;
void ZeroMeanOfImage(TImage *image) ;
void ZeroMeanOfImageArea(TImage *image,FRect *rect) ;
void DeleteImage(TImage *image) ;
void SetImageName(TImage *image,char *name) ;
void DisplayImage(TImage *image) ;
TImage *CopyImage(TImage *image) ;
void SetImageData(TImage *image,float *data,ImageDataType dataType) ;
TImage *CreateNewImage(short width,short height) ;
float ImageMean(TImage *image) ;
float ImageMeanOfArea(TImage *image,FRect *rect) ;
TImage *ImportImage() ;
TImage *CreateImageFromData(short width,short height,void *data,ImageDataType dataType) ;
TImage *CreateImageFromImageArea(short width,short height,short dx,short dy,TImage *area) ;
void FillImageFromImageArea(short width,short height,short dx,short dy,TImage *area,TImage *image) ;
Boolean AdjustImageSize(TImage **image,long width,long height) ;
void DisplayImageInWindow(TImage *, TWindow *, Rect *rect,Boolean scaleToExp);
void ZoomImageToRect(TImage *, TWindow *, Rect *rect) ;
void CreateImageOffscreen(TImage *, TWindow *) ;
void DisplayImageInRect(TImage *image, Rect *rect) ;
#endif

