// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdlib.h>
#include "devices.h"
#include "globals.h"
#include "types.h"
#include "TNCEMSSApp.h"
#include "TControlWindow.h"
#include "TDisplayWindow.h"
#include "XDrivers.h"

/* Define different menu-associated items */
#include "funcdefs.h"
#include "graphics.h"
#include "menudef.h"
#include "devices.h"
#include "menu_index.h"

#include "dfpdspl.h"
#include "errors.h"
#include "microscopes.h"
#include "xtal.h"
#include "params.h"
#include "atoms.h"
#include "ProjCellRecord.h"
#include "PhasegratingRecord.h"
#include "MslicRecord.h"
#include "displ.h"
#include "filenames.h"
#include "storage.h"
#include "layers.h"

UInt32	gRedPixel, gBluePixel, gGreenPixel ;
UInt32	gLighterGray ;

#ifdef ALPHA
#define ALIGN64BYTE_BOUNDARY 
#endif
#ifdef IBM
#define ALIGN64BYTE_BOUNDARY
#endif
#ifdef SGI 
#define ALIGN64BYTE_BOUNDARY
#endif
#ifdef VMS
#include <decw$include/cursorfont.h>
#else
#include <X11/cursorfont.h>
#endif


void SetupColorTable() ;
void get_microscopes() ;
char *Allocate(long numBytes) ;
void button1() ;
void r_unass() ;
void set_menu(long number) ;
void enableast() ;
void GetXEvents() ;
void ExitX() ;

extern  Boolean gNoGUI ;
extern  Cursor gWatchCursor ;

typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
extern MonitorDisplayType  gMonitorDisplayType ;

#include "TImage.h"
TImage				*gExperimentalImage = nil;
TImage				*gStandardDeviationImage = nil;


extern MENUS gMenuList[] ;

float 	RANGE ; 
char	*NcemssDir,*NcemssPool,*NcemssPath ;
extern 	short	gImageDepth ;
short	gReportVisuals = 0 ;
char	*NCEMSSDIR = "." ;
char	*NCEMSSPOOL = "." ;


long	gBlackPixelIndex ;
long	gLighterGrayIndex ;
long	gGrayPixelIndex ;
long	gDarkerGrayIndex ;
long	gWhitePixelIndex ;
long	gRedPixelIndex ;
long	gGreenPixelIndex ;
long	gBluePixelIndex ;

char	*NCEMSS_PATH = PREFIX ;

char	ncemss_data_path[256] ;
char	ncemss_common_path[256] ;
char	ncemss_install_path[256] ;

TNCEMSSApp::TNCEMSSApp(long argc,char *argv[]) 
	: TApplication(argc,argv) {

    long	i,idim ;
    long	iatoms ;
	char	text[256],*cp=text ;
    
    
    sprintf(ncemss_data_path,"%s/ncemssdir",PREFIX) ;
    sprintf(ncemss_common_path,"%s/ncemsscommon",PREFIX) ;
    sprintf(ncemss_install_path,"%s",PREFIX) ;

	// If PREFIX environnment variable is set, use it
    if ((NcemssPath = getenv("PREFIX")) != NULL) {
	    sprintf(ncemss_data_path,"%s/ncemssdir",NcemssPath) ;
	    sprintf(ncemss_common_path,"%s/ncemsscommon",NcemssPath) ;
	    sprintf(ncemss_install_path,"%s",NcemssPath) ;
		NcemssDir = ncemss_data_path ;	
		NcemssPool = ncemss_common_path ;  
		NcemssPath = ncemss_install_path ;  
		NCEMSS_PATH = ncemss_install_path ;
    }
    else {
	    NcemssPath = ncemss_install_path ;
	    NCEMSS_PATH = ncemss_install_path ;
    }
	// Check to see if we have the NCEMSSDIR and NCEMSSPOOL variables
	// If we do, over-ride
	if ((NcemssDir = getenv("NCEMSSDIR")) == NULL) NcemssDir = ncemss_data_path ;		
	if ((NcemssPool = getenv("NCEMSSPOOL")) == NULL) NcemssPool = ncemss_common_path ;
	
	fprintf(stderr,"DIR %s \n",NcemssDir);
	fprintf(stderr,"COM %s \n",NcemssPool);
	//exit(1);
    OutDevice = &WorkStation ;   
    // Get the optional paranmeters
    if (argc > 1) {
	    for (i=1;i<argc;i++) {
			strcpy(cp,argv[i]) ;
			if (strcmp(argv[i],"-s") == 0) {
				if (strcmp(argv[i+1],"small") == 0) {
	    				WorkStation.xres = 512 ;
	    				WorkStation.yres = 384 ;
				}
				else if (strcmp(argv[i+1],"medium") == 0) {
	    				WorkStation.xres = 768 ;
	    				WorkStation.yres = 512 ;
				}
				else if (strcmp(argv[i+1],"large") == 0) {
	    				WorkStation.xres = 768 ;
	    				WorkStation.yres = 768 ;
				}
			}
			if (strcmp(argv[i],"-v") == 0) {
				gReportVisuals = 1 ;
			}
			if (strcmp(argv[i],"-d") == 0) {
				idim = atoi(argv[i+1]) ;
				if (idim >= 32 && idim <= 2048) {
					gMaxArraySize = idim*idim*3*2 ;
				}
			}
			if (strcmp(argv[i],"-a") == 0) {
				iatoms = atoi(argv[i+1]) ;
				if (iatoms >= 1000 && iatoms <= 40000) {
					gMaxAtoms = iatoms ;
				}
			}
			if (strcmp(argv[i],"-h") == 0) {
				PrintUsage() ;
				exit(0) ;
			}
	    }
    }
    RANGE = OutDevice->range ;
	gNumGrays = 64 ;
	gNumColors = 40 ;

    AllocateMemory() ;
    dfpdspl.xcenter = OutDevice->xres/2 ;
    dfpdspl.ycenter = OutDevice->yres/2 ;
    enableast() ;      
    get_microscopes() ;  

	if (!gNoGUI) {
		// ------------------------------------------------------------
		// Display the message about origin and authorship		
		// ------------------------------------------------------------ 
		gCurrentMenu = &gMenuList[0] ;  // Initialize the menu-ponter
		set_menu(0) ;					// Set up the first menu to display
		
		Initialize() ;
    
		switch(gImageDepth) {
			case 8:
				gNumGrays = 64 ;
				gNumColors = 40 ;
				OutDevice->range = gNumGrays-1 ;
				RANGE = OutDevice->range ;
				break ;
			case 15:
			case 16:
				gNumGrays = 256 ;
				gNumColors = 40 ;
				OutDevice->range = gNumGrays-1 ;
				RANGE = gNumGrays-1 ;
				break ;
			case 24:
			case 32:
				gNumGrays = 256 ;
				gNumColors = 40 ;
				OutDevice->range = gNumGrays-1 ;
				RANGE = gNumGrays-1 ;
				break ;
		}
  
		SetupColors() ;
		DefineXForms() ;
		ShowXForms() ;
		DefineWindows() ;
		
		if (gReportVisuals) {
			fprintf(stderr,"\nImageDepth = %d\n",gImageDepth) ;
			fflush(stderr) ;
		}
	}
}

TNCEMSSApp::~TNCEMSSApp() {
    FreeMemory() ;
}

void TNCEMSSApp::DefineXForms() {
}

void TNCEMSSApp::ShowXForms() {
}

void TNCEMSSApp::FreeMemory() {
	delete[] gLut ;
	delete[] gColors ;
	delete[] mStorage ;
	delete[] gZplot ;
	delete[] gProjCell->x ;
	delete[] gProjCell->y ;
	delete[] gProjCell->z ;
	delete[] gProjCell->dw ;
	delete[] gProjCell->occ ;
	delete[] gProjCell->atno ;
	delete[] gProjCell->ntype ;
	delete gProjCell ;
	delete[] gPhsgrtRec->x ;
	delete[] gPhsgrtRec->y ;
	delete[] gPhsgrtRec->z ;
	delete[] gPhsgrtRec->dw ;
	delete[] gPhsgrtRec->occf ;
	delete[] gPhsgrtRec->atno ;
	delete[] gPhsgrtRec->ntype ;
	delete gPhsgrtRec ;
	delete[] atom ;
	delete[] basis ;
}

void TNCEMSSApp::AllocateMemory() {
	unsigned long evenAddressOffset ;
	char *address ;

	
#ifdef ALIGN64BYTE_BOUNDARY 
	address = (char *)gWorkSpace ;
	evenAddressOffset = (unsigned long)address%64 ;
	if (evenAddressOffset != 0) 
		evenAddressOffset = 64 - evenAddressOffset ;
	address += evenAddressOffset ;
	gWorkSpace = (float *)address ;
#endif

	gPhsgrtRec = new PhasegratingRecord ;
	gProjCell = new ProjCellRecord ;
	
	gProjCell->x = new float[gMaxAtoms] ;
	gProjCell->y = new float[gMaxAtoms] ;
	gProjCell->z = new float[gMaxAtoms] ;
	gProjCell->dw = new float[gMaxAtoms] ;
	gProjCell->occ = new float[gMaxAtoms] ;
	gProjCell->atno = new short[gMaxAtoms] ;
	gProjCell->ntype = new short[gMaxAtoms] ;
	gPhsgrtRec->x = new float[gMaxAtoms] ;
	gPhsgrtRec->y = new float[gMaxAtoms] ;
	gPhsgrtRec->z = new float[gMaxAtoms] ;
	gPhsgrtRec->dw = new float[gMaxAtoms] ;
	gPhsgrtRec->occf = new float[gMaxAtoms] ;
	gPhsgrtRec->atno = new short[gMaxAtoms] ;
	gPhsgrtRec->ntype = new short[gMaxAtoms] ;

	gZplot = new float[MAXPLOT] ;
	atom = new AtomRec[gMaxAtoms] ;
	basis = new AtomRec[gMaxAtoms] ;


	gColors = new XColor[512] ;
	gLut = new unsigned long[512] ;


	gWorkSpace = new float[gMaxArraySize] ;
	mStorage = gWorkSpace ;


}

void TNCEMSSApp::DefineWindows() {

	gDisplayWindow = new TDisplayWindow(210,10,OutDevice->xres,OutDevice->yres,
    		gImageDepth,"NCEMSS-IMAGES",true) ;
    
	gOffscreenPtr = gDisplayWindow->GetOffscreen() ;
	AddWindow(gDisplayWindow) ;

 	gControlWindow = new TControlWindow(10,gDisplayHeight-440,
		512,394,gImageDepth,"NCEMSS",0) ;
	AddWindow(gControlWindow) ;

	XSetBackground(gDisplay,gControlWindow->GetGC(),gGrayPixel) ; 


	SetPort(gControlWindow) ; 
	SetBkColor(gGrayPixel) ;
  
	BringToFront(gControlWindow) ;

	/* Create the watch cursor so that it can be used later */
	gWatchCursor = XCreateFontCursor(gDisplay,XC_watch) ;

}

void TNCEMSSApp::PrintUsage(){
	fprintf(stderr,"\n -------------------------------------------------------- ") ;
	fprintf(stderr,"\n Valid flags are :") ;
	fprintf(stderr,"\n -s small,medium of large;    Indicating window size") ;
	fprintf(stderr,"\n -d number;                   Indicating the maximum") ;
	fprintf(stderr,"\n                array size to use: i.e 1024 -> 1024**2") ;
	fprintf(stderr,"\n                Default: 512") ;
	fprintf(stderr,"\n -a number                    Giving the maximum number") ;
	fprintf(stderr,"\n                of atoms that can be used. Default: 5000") ;
	fprintf(stderr,"\n ---------------------------------------------------------") ;	
	fprintf(stderr,"\n") ;
}

void TNCEMSSApp::SetupColors() {

    long		i,base = 0 ;

    gColorMap = DefaultColormap(gDisplay,gScreen) ;
    
    if (gMonitorDisplayType == PSEUDOCOLOR) {
    	XAllocColorCells(gDisplay,gColorMap,0,NULL,0,gLut,
    		     	gNumGrays+gNumColors) ;
     	for(i=0;i < gNumGrays;i++) {
    		gColors[i].pixel	= gLut[i] ;
    		base				= i*65535/(gNumGrays-1) ;
    		gColors[i].red 		= base ;
    		gColors[i].green 	= base ;
    		gColors[i].blue 	= base ;
    		gColors[i].flags	= DoRed | DoGreen | DoBlue ;
    	}
   }
   else {
    	for(i=0;i < gNumGrays;i++) {
    		gColors[i].pixel	= i ;
    		base				= i*65535/(gNumGrays-1) ;
    		gColors[i].red 		= base ;
    		gColors[i].green 	= base ;
    		gColors[i].blue 	= base ;
    		gColors[i].flags	= DoRed | DoGreen | DoBlue ;
    	}
    }
    SetMyColors(gColors+gNumGrays,gLut+gNumGrays,gNumColors) ;


   if (gMonitorDisplayType == PSEUDOCOLOR) 
     	for(i=gNumGrays;i < gNumGrays+gNumColors;i++) 
    		gColors[i].pixel	= gLut[i] ;
   else 
    	for(i=gNumGrays;i < gNumGrays+gNumColors;i++) 
     		gColors[i].pixel	= i ;

    if (gMonitorDisplayType == PSEUDOCOLOR) {
   		XStoreColors(gDisplay,gColorMap,gColors,gNumGrays+gNumColors) ;
    }
    else {
	  	for(i=0;i < gNumGrays+gNumColors;i++) {
	   		gColors[i].pixel = i ;
	   		XAllocColor(gDisplay,gColorMap,gColors+i) ;
	   		gLut[i] =  i ;
	  	}
    }
    gImageAttributes.background_pixel = WhitePixel(gDisplay,gScreen) ;
    gImageAttributes.border_pixel = BlackPixel(gDisplay,gScreen) ;
    if (gMonitorDisplayType == PSEUDOCOLOR) {
		gImageAttributes.colormap = gColorMap ;
		gImageDepth = 8 ;	// Force the depth to 8
		gImageLineBuffer = new unsigned char[1024*4] ;
	    	gImageLine = XCreateImage(gDisplay,gVisualInfo.visual,gImageDepth,
	    			ZPixmap,0,(char *)gImageLineBuffer,4096,1,8,0) ;
    }
    else {
    	switch(gImageDepth) {
    		case 15:
			case 16:
				gImageLineBuffer = new unsigned char[2048*4] ;
			    	gImageLine = XCreateImage(gDisplay,gVisualInfo.visual,gImageDepth,
			    			ZPixmap,0,(char *)gImageLineBuffer,4096,1,16,0) ;
    			break ;
    		case 24:
			case 32:
				gImageLineBuffer = new unsigned char[4096*4] ;
			    	gImageLine = XCreateImage(gDisplay,gVisualInfo.visual,gImageDepth,
			    			ZPixmap,0,(char *)gImageLineBuffer,4096,1,32,0) ;
    			break ;
    	}
    }

	gBlackPixelIndex 	= 0 ;
	gLighterGrayIndex 	= near(0.797*gNumGrays) ;
	gGrayPixelIndex 	= near(0.718*gNumGrays) ;
	gDarkerGrayIndex 	= near(0.547*gNumGrays) ;
	gWhitePixelIndex 	= gNumGrays - 1 ;
	gRedPixelIndex 		= gNumGrays ;
	gGreenPixelIndex 	= gNumGrays + 6 ;
	gBluePixelIndex 	= gNumGrays + 16 ;
	

    gBlackPixel 	= gColors[gBlackPixelIndex].pixel ;
    gLighterGray 	= gColors[gLighterGrayIndex].pixel ;
    gGrayPixel 		= gColors[gGrayPixelIndex].pixel ;
    gDarkerGray 	= gColors[gDarkerGrayIndex].pixel ;
	gHighlightColor	= gColors[near(0.8125*gNumGrays)].pixel ;
	gDarkFrameColor = gColors[near(0.46875*gNumGrays)].pixel ;
	gDarkestFrameColor = gColors[near(0.3125*gNumGrays)].pixel ;
	gLightFrameColor= gHighlightColor ;
    gWhitePixel 	= gColors[gWhitePixelIndex].pixel ;
    gRedPixel 		= gColors[gRedPixelIndex].pixel ;
    gGreenPixel 	= gColors[gGreenPixelIndex].pixel ;
    gBluePixel 		= gColors[gBluePixelIndex].pixel ;

}
void TNCEMSSApp::SetMyColors(XColor *color,unsigned long *pixel,unsigned long numColors){
    XColor *thisColor ;
    unsigned long	*thisPixel ;

    thisColor = color ;
    thisPixel = pixel ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 65535 ;	/* 43961 ; */
    thisColor->green 	= 0 ; // 12037 ;	/* 997 ; */
    thisColor->blue 	= 0 ; // 64675 ;	/* 65535 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 65535 ;
    thisColor->green 	= 10155 ;	/* 1040 ; */
    thisColor->blue 	= 29346 ;	/* 46311 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 65535 ;
    thisColor->green 	= 11065 ;	/* 602 ; */
    thisColor->blue 	= 9815 ;	/* 3105 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 65535 ;
    thisColor->green 	= 32688 ;	/* 27792 ; */
    thisColor->blue 	= 10105 ;	/* 1317 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 65535 ;
    thisColor->green 	= 61965 ;	/* 51958 ; */
    thisColor->blue 	= 10279 ;	/* 599 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 45330 ;	/* 60128 ; */
    thisColor->green 	= 65535 ;
    thisColor->blue 	= 10665 ;	/* 447 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 0 ; // 12409 ;	/* 26973 ; */
    thisColor->green 	= 65535 ;
    thisColor->blue 	= 0 ; // 11557 ;	/* 2210 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 11137 ;	/* 1080 ; */
    thisColor->green 	= 65535 ;
    thisColor->blue 	= 45654 ;	/* 19407 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 11627 ;	/* 669 ; */
    thisColor->green 	= 62405 ;	/* 65535 ; */
    thisColor->blue 	= 65535 ;	/* 45801 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 12001 ;	/* 1398 ; */
    thisColor->green 	= 30261 ; 	/* 57373 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 14117 ;	/* 1619 ; */
    thisColor->green 	= 12181 ; 	/* 35557 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;


    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 28943;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 13982;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 7643;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 33450;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 0 ; // 8050 ;	/* 1348 ; */
    thisColor->green 	= 0 ; // 2965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;
    gBluePixel = thisColor->pixel ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 4050 ;	/* 1348 ; */
    thisColor->green 	= 19965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 5535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 18050 ;	/* 1348 ; */
    thisColor->green 	= 42965 ;	/* 3561 ; */
    thisColor->blue 	= 25535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 28050 ;	/* 1348 ; */
    thisColor->green 	= 52965 ;	/* 3561 ; */
    thisColor->blue 	= 15535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 15535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 675;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 55535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 28050 ;	/* 1348 ; */
    thisColor->green 	= 52965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 40050 ;	/* 1348 ; */
    thisColor->green 	= 19965 ;	/* 3561 ; */
    thisColor->blue 	= 45535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 2965 ;	/* 3561 ; */
    thisColor->blue 	= 5535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 450 ;	/* 1348 ; */
    thisColor->green 	= 2965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 38050 ;	/* 1348 ; */
    thisColor->green 	= 32965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 41050 ;	/* 1348 ; */
    thisColor->green 	= 42965 ;	/* 3561 ; */
    thisColor->blue 	= 35535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 22050 ;	/* 1348 ; */
    thisColor->green 	= 62965 ;	/* 3561 ; */
    thisColor->blue 	= 25535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 965 ;	/* 3561 ; */
    thisColor->blue 	= 55535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 11150 ;	/* 1348 ; */
    thisColor->green 	= 32965 ;	/* 3561 ; */
    thisColor->blue 	= 44535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;
}
