// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdlib.h>
#include "menu_index.h"
#include "TImage.h"
#include "TButton.h"
#include "TStructureRefinementAnnealingPage.h"
#include "TMainPage.h"

#include "globals.h"

void	SetPage(TPage *) ;
extern	TMainPage *gMainPage ;
void	PutMessage(char *) ;
void 	DrawQuantitativePage() ;
void	ExitFromStructureAnnealing() ;
void	SetActiveAreaAndAtoms() ;
void 	SetActiveElements() ;
void	CompareAndRefineStructure() ;
void 	GetChemicalBondData() ;
void	GetBondValenceParameters() ;
void 	run_help(char *) ;

typedef enum {kXCF,kChiSquare,kRMSD,kDifferenceImage,kFMAD,kFTXCF}
ComparisonMethod ;

extern	ComparisonMethod gComparisonMethod ;
extern	Boolean gEstimateImageShift ;
extern	Boolean gCalculateExactFTXCF ;
extern	FRect	gComparisonRect ;
extern	Boolean	gCompareEntireImage ;

float gImageComparisonWeight = 1.0 ;
float gBondValenceEnergy = -1. ;
float gLastBondValenceEnergy = -1. ;
float BondValenceEnergy() ;
float BondValenceSum(short ndx,short index) ;
Boolean	gIncludeImageComparison = true ;
Boolean	gIncludeBondDistances = false ;
Boolean	gIncludeBondValenceSum = false ;
Boolean	gSaveLogFile = false ;
void CompareExperimentalImageWithSimulation() ;


float 	gChemicalEnergy = -1. ; // At start of a run
float   gChangeInChemicalEnergy ;
float	gMatchingEnergy = 10000. ;

short	gNumActiveElements = 0 ;
short 	gActiveElementNumber[20] ;	// Maximum number of active elements

#ifndef __RefinementGroup
#define __RefinementGroup
typedef struct RefinementGroup {
	long numInGroup ;
	long *atomIndex ;
	unsigned char	*refine ;

} RefinementGroup ;

typedef struct RefinementList {
	long numGroups ;
	RefinementGroup *groupPtr ;
} RefinementList ;
#endif


typedef enum {kCrossCorrelation,kChiSquared,kFTAmplCrossCorrelation } FittingMethod ;

RefinementList gRefinementList = { 0, NULL} ;

typedef struct {
	float	maxAmplitude ;	// In Angstrom
	Boolean	useActiveRect ;
	FRect	activeRect ;
} RelaxationAtomRecord ;

typedef struct {
	float	start ;			// Starting Temperature
	float	end ;			// Ending Temperature
	float	current ;		// current temperature
	float	dTFraction ;	// Temperature decrement
	long		maxNumTriesPerTemperature ;
	long		maxNumGoodTries ;
} RelaxationTemperatureRecord ;

typedef struct {
	unsigned long	ID ;
	float	minValue ;
	float	maxValue ;
	float	startingValue[2] ;
	float	maxChangeInValue ;
	float	currentValue[3] ;	// We may need to store a magnitude and a direction or possible magnitude and components
} ParameterRecord ;

typedef struct {
	short	numFittingParameters ;
	ParameterRecord parameter[10] ;	// only space for 10 different parameters to fit
	RelaxationTemperatureRecord temperature ;
	float previousEnergy ;
	float thisEnergy ;
	float minEnergy ;
	long  numTriesThisTemperature ;
	FittingMethod fittingMethod ;
} ParameterFittingRecord ;

typedef struct {
	RelaxationAtomRecord		atom ;	// In Angstrom
	RelaxationTemperatureRecord	temperature ;
	float previousEnergy ;
	float thisEnergy ;
	float minEnergy ;
	long  numTriesThisTemperature ;
	FittingMethod fittingMethod ;
} RelaxationParamRecord ;

RelaxationParamRecord gRelaxationParamRecord = {0.5,false,0.0,0.0,1.0,1.0,0.01,0.0001,0.1,0.9,25,5,0.,0.,0.005,0,kCrossCorrelation} ;

ParameterFittingRecord gParameterFittingRecord = {-1} ;

TStructureRefinementAnnealingPage::TStructureRefinementAnnealingPage(TWindow *window) 
		: TPage(window) {

	TGroupBox *grpBox ;
	TFrame	*theFrame ;
	char	text[256] ;

	mHelp = new TButton(490,2,510,20,"H") ;
	AddObject(mHelp) ;

	grpBox = new TGroupBox(10,20,270,190, "Temperature") ;
	AddObject(grpBox) ;

	grpBox = new TGroupBox(290,20,500,190, "Refinement Parameters") ;
	AddObject(grpBox) ;

	grpBox = new TGroupBox(10,200,500,320, "Fitting Method") ;
	AddObject(grpBox) ;

	grpBox = new TGroupBox(30,120,260,180, "Per Degree of Freedom") ;
	AddObject(grpBox) ;

  	sprintf(text,"%.3f",gRelaxationParamRecord.atom.maxAmplitude) ;
	mMaxShift = new TEditText(410,30,490,50,"Max. Atom Shift[A]",text) ;
	AddTextEditObject(mMaxShift) ;

  	sprintf(text,"%d",gRelaxationParamRecord.temperature.maxNumGoodTries) ;
	mNumSuccessfulAttempts = new TEditText(180,155,250,175,
		"# Successful Attempts",text) ;
	AddTextEditObject(mNumSuccessfulAttempts) ;

  	sprintf(text,"%d",
		gRelaxationParamRecord.temperature.maxNumTriesPerTemperature) ;
	mAttemptsPerTemp = new TEditText(180,130,250,150,"Attempts Per. Temp.",
		text) ;
	AddTextEditObject(mAttemptsPerTemp) ; 

  	sprintf(text,"%.4f",gRelaxationParamRecord.temperature.dTFraction) ;
	mFractionalChange = new TEditText(160,90,260,110,"Fractional Change",text) ;
	AddTextEditObject(mFractionalChange) ;

  	sprintf(text,"%.5f",gRelaxationParamRecord.temperature.end) ;
	mEndingTemperature = new TEditText(160,60,260,80,"Ending",text) ;
	AddTextEditObject(mEndingTemperature) ;

  	sprintf(text,"%.5f",gRelaxationParamRecord.temperature.start) ;
	mStartingTemperature = new TEditText(160,30,260,50,"Starting",text) ;
	AddTextEditObject(mStartingTemperature) ;


	mSetActiveElements	= new TButton(300,55,490,83,"Set Active Elements...") ;
	AddObject(mSetActiveElements) ;

	mSetFitAreaAndAtoms	= new TButton(300,88,490,116,"Set Fitting Area & Atoms to Refine...") ;
	AddObject(mSetFitAreaAndAtoms) ;

	mSetChemicalConstraints	= new TButton(300,121,490,149,"Set Chemical Constraints...") ;
	AddObject(mSetChemicalConstraints) ;

	mSetBondValences	= new TButton(300,154,490,182,"Set Bond Valences...") ;
	AddObject(mSetBondValences) ;

	mCCCRealSpace = new TRadioButton(20,230,255,250,"Cross Correlation Real Space") ;
	mCCCRealSpace->SetControlValue(true) ;
	AddObject(mCCCRealSpace) ;

	mChiSquare = new TRadioButton(20,255,255,275,"Chi-Square") ;
	AddObject(mChiSquare) ;

	mFTAmplitudes = new TRadioButton(20,280,255,300,
		"Cross Correlation Fourier Space Amplitudes") ;
	AddObject(mFTAmplitudes) ;


	mCompareImageFit = new TCheckBox(40,210,210,230,"Include Image Agreement");
	mCompareImageFit->SetControlValue(true) ;
	AddObject(mCompareImageFit) ;

	grpBox = new TGroupBox(290,220,490,310, "Options") ;
	AddObject(grpBox) ;

	mUseBondDistances = new TCheckBox(300,235,460,255,"Include Chemical Constraints");
	AddObject(mUseBondDistances) ;

	mUseBondStrengths = new TCheckBox(300,260,460,280,"Include Bond Valence Sum");
	AddObject(mUseBondStrengths) ;

	mCreateLogFile = new TCheckBox(300,285,460,305,"Create a log file of E vs time");
	AddObject(mCreateLogFile) ;

	mRefine	= new TButton(380,350,500,380,"Refine") ;
	AddObject(mRefine) ;

	mCancel = new TButton(240,350,360,380,"Cancel") ;
	AddObject(mCancel) ;

	theFrame = new TFrame(15,220,260,310) ;
	AddObject(theFrame) ;

	Initialize() ;
}
 
TStructureRefinementAnnealingPage::~TStructureRefinementAnnealingPage() {
}

void	TStructureRefinementAnnealingPage::Initialize() {
	char	text[256] ;
  	sprintf(text,"%.5f",gRelaxationParamRecord.temperature.start) ;
	mStartingTemperature->SetEditText(text) ;
	mStartingTemperature->SelectAll() ;
	
  	sprintf(text,"%.5f",gRelaxationParamRecord.temperature.end) ;
	mEndingTemperature->SetEditText(text) ;
		
  	sprintf(text,"%.4f",gRelaxationParamRecord.temperature.dTFraction) ;
	mFractionalChange->SetEditText(text) ;
	
  	sprintf(text,"%d",gRelaxationParamRecord.temperature.maxNumTriesPerTemperature) ;
	mAttemptsPerTemp->SetEditText(text) ;
	
  	sprintf(text,"%d",gRelaxationParamRecord.temperature.maxNumGoodTries) ;
	mNumSuccessfulAttempts->SetEditText(text) ;
	
  	sprintf(text,"%.3f",gRelaxationParamRecord.atom.maxAmplitude) ;
	mMaxShift->SetEditText(text) ;

}

void TStructureRefinementAnnealingPage::Draw() {
	EraseRect(&mWindow->GetPortRect()) ;
	TPage::Draw() ;
}

void	TStructureRefinementAnnealingPage::DoMouseDownInObject(TObject *obj,Point *pt) {
	char	text[256] ;
	if (obj->DoMouseDown(pt)) {
		if (obj == mHelp) {
			run_help("TStructureRefinement") ;
			return ;
		}
		if (obj == mCCCRealSpace) {
			mChiSquare->SetControlValue(false) ;
			mFTAmplitudes->SetControlValue(false) ;
			sprintf(text,"%.5f",0.01) ;
			mStartingTemperature->SetEditText(text) ;
			sprintf(text,"%.4f",0.0001) ;
			mEndingTemperature->SetEditText(text) ;
		}
		else if (obj == mChiSquare) {
			mCCCRealSpace->SetControlValue(false) ;
			mFTAmplitudes->SetControlValue(false) ;
			sprintf(text,"%.5f",0.1) ;
			mStartingTemperature->SetEditText(text) ;
			sprintf(text,"%.4f",0.001) ;
			mEndingTemperature->SetEditText(text) ;
		}
		else if (obj == mFTAmplitudes) {
			mCCCRealSpace->SetControlValue(false) ;
			mChiSquare->SetControlValue(false) ;
			sprintf(text,"%.5f",0.01) ;
			mStartingTemperature->SetEditText(text) ;
			sprintf(text,"%.4f",0.0001) ;
			mEndingTemperature->SetEditText(text) ;
		}
		else if (obj == mSetActiveElements) {
			SetActiveElements() ;
		}
		else if (obj == mSetFitAreaAndAtoms) {
			SetActiveAreaAndAtoms() ;
		}
		else if (obj == mSetChemicalConstraints) {
			// Get the fitting method so we can use this
			// information to set the weights
			FittingMethod savedMethod = gRelaxationParamRecord.fittingMethod ;
			if (mCCCRealSpace->GetControlValue()) {
				gRelaxationParamRecord.fittingMethod = kCrossCorrelation ;
			}
			else if (mChiSquare->GetControlValue()) {
				gRelaxationParamRecord.fittingMethod = kChiSquared ;
			}
			else {
				gRelaxationParamRecord.fittingMethod = kFTAmplCrossCorrelation ;
			}
			GetChemicalBondData() ;
			// Reset in case we cancel
			gRelaxationParamRecord.fittingMethod = savedMethod ;
		}
		else if (obj == mSetBondValences) {
			GetBondValenceParameters() ;
		}
		else if (obj == mRefine) {
			// Get and assemble all the relevant parameters and
			// do the refinement

	  		mStartingTemperature->GetEditText(text) ;
			gRelaxationParamRecord.temperature.start = atof(text) ;
		
	  		mFractionalChange->GetEditText(text) ;
			gRelaxationParamRecord.temperature.dTFraction = atof(text) ;
		
	  		mEndingTemperature->GetEditText(text) ;
			gRelaxationParamRecord.temperature.end = atof(text) ;
			
	  		mAttemptsPerTemp->GetEditText(text) ;
			gRelaxationParamRecord.temperature.maxNumTriesPerTemperature = 
				atoi(text) ;
		
	  		mNumSuccessfulAttempts->GetEditText(text) ;
			gRelaxationParamRecord.temperature.maxNumGoodTries = atoi(text) ;
		
	  		mMaxShift->GetEditText(text) ;
			gRelaxationParamRecord.atom.maxAmplitude = atof(text) ;
				
			gSaveLogFile = mCreateLogFile->GetControlValue() ;

			gIncludeImageComparison = mCompareImageFit->GetControlValue() ;
			gImageComparisonWeight = (gIncludeImageComparison  ? 1.0 : 0.) ;

			gIncludeBondDistances = mUseBondDistances->GetControlValue() ;
			gIncludeBondValenceSum = mUseBondStrengths->GetControlValue() ;

			if (mCCCRealSpace->GetControlValue()) {
				gRelaxationParamRecord.fittingMethod = kCrossCorrelation ;
			}
			else if (mChiSquare->GetControlValue()) {
				gRelaxationParamRecord.fittingMethod = kChiSquared ;
			}
			else {
				gRelaxationParamRecord.fittingMethod = kFTAmplCrossCorrelation ;
			}
			if (gNumActiveElements == 0) {
				PutMessage("There are no active elements defined") ;
			}
			else {
				if (gRefinementList.numGroups == 0) {
					PutMessage("There are no atoms defined to be moved") ;		
				}
				else {
					CompareAndRefineStructure() ;
				}
			}
		}		
		else if (obj == mCancel) {
			ExitFromStructureAnnealing() ;
		}		
	}
} 

void	ExitFromStructureAnnealing() {
	set_menu(QUANTITATIVE) ;
	ClearDisplay() ;
	SetPage(gMainPage) ;
	DrawQuantitativePage() ;
}
