// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdlib.h>
#include <math.h>
#include "types.h"
#include "menu_index.h"
#include "TImage.h"
#include "TButton.h"
#include "TMainPage.h"

#include "TStructureRefinementProgressPage.h"
#include "globals.h"

void	SetPage(TPage *) ;
extern	TMainPage *gMainPage ;
void	PutMessage(char *) ;
void 	DrawQuantitativePage() ;
void	ExitFromStructureRefinementProgress() ;
void	SaveRefinedStructure() ;

#include "funcdefs.h"	
#include "PhasegratingRecord.h"
#include "ProjCellRecord.h"
#include "displ.h"
#include "MslicRecord.h"
#include "storage.h"
#include "types.h"
#include "files.h"
#include "errors.h"
#include "globals.h"
#include "dfpdspl.h"
#include "symops.h"
#include "TAtom.h"

extern	RefinementList gRefinementList ;
extern	char	*chemsymb[] ;
extern	Boolean	gUseSymmetryElements ;

extern	TImage *gStandardDeviationImage ;
extern	TImage *gExperimentalImage ;
Boolean AdjustImageSize(TImage **image,long width,long height) ;

void DisplayImageInWindow(TImage *, TWindow *, Rect *rect,Boolean );


typedef struct {
	short numBasisAtoms ;
	TAtom *atom ;
} TBasis ;

extern	TBasis gBasis ;

typedef enum {kCrossCorrelation,kChiSquared,kFTAmplCrossCorrelation } FittingMethod ;

typedef struct {
	float	maxAmplitude ;	// In Angstrom
	Boolean	useActiveRect ;
	FRect	activeRect ;
} RelaxationAtomRecord ;

typedef struct {
	float	start ;			// Starting Temperature
	float	end ;			// Ending Temperature
	float	current ;		// current temperature
	float	dTFraction ;	// Temperature decrement
	long		maxNumTriesPerTemperature ;
	long		maxNumGoodTries ;
} RelaxationTemperatureRecord ;

typedef struct {
	RelaxationAtomRecord		atom ;	// In Angstrom
	RelaxationTemperatureRecord	temperature ;
	float previousEnergy ;
	float thisEnergy ;
	float minEnergy ;
	long  numTriesThisTemperature ;
	FittingMethod fittingMethod ;
} RelaxationParamRecord ;

extern	RelaxationParamRecord gRelaxationParamRecord ;

extern	TStructureRefinementProgressPage *gStructureRefinementProgressPage ;
extern	Boolean	gSaveLogFile ;
extern	UInt32	gRedPixel, gBluePixel, gGreenPixel ;
extern	long	gIndex ;
extern	float	gDx,gDy,gDz,gDdw,gDocc ;
extern	long 	gShiftN,gShiftM;
extern	TMatrix	*gSymOps ;	


TStructureRefinementProgressPage::TStructureRefinementProgressPage(TWindow *window,ProjCellRecord * nucell)
	: TPage(window) {
	TStaticText	*theText ;
	Rect	theRect ;

	mEnergyFrame = new TUserRect(5,240,505,340);
	AddObject(mEnergyFrame) ;

	mOffscreen = new TOffscreen(mWindow,&mEnergyFrame->GetRect()) ;

	mModelFrame = new TUserRect(5,30,165,190);
	AddObject(mModelFrame) ;

	theRect = mModelFrame->GetRect() ;
	theText = new TStaticText(theRect.left,theRect.top-20,theRect.right,theRect.top-5,
		"Current Model");
	theText->SetJustification(kCenterJustified) ;
	AddObject(theText) ;

	mCurrentImageFrame = new TUserRect(175,30,335,190);
	AddObject(mCurrentImageFrame) ;

	theRect = mCurrentImageFrame->GetRect() ;
	theText = new TStaticText(theRect.left,theRect.top-20,theRect.right,theRect.top-5,
		"Current Computed Image");
	theText->SetJustification(kCenterJustified) ;
	AddObject(theText) ;

	mExperimentalImageFrame = new TUserRect(345,30,505,190);
	AddObject(mExperimentalImageFrame) ;

	theRect = mExperimentalImageFrame->GetRect() ;
	theText = new TStaticText(theRect.left,theRect.top-20,theRect.right,theRect.top-5,
		"Target Image");
	theText->SetJustification(kCenterJustified) ;
	AddObject(theText) ;

	mReturn = new TButton(430,350,505,380,"Return");
	AddObject(mReturn) ;

	mSaveReturn = new TButton(260,350,420,380,"Save Refined Model & Return");
	AddObject(mSaveReturn) ;

	theText = new TStaticText(220,215,280,235,"Log File: ");
	AddObject(theText) ;

	mLogFileName = new TStaticText(280,215,460,235,"None");
	AddObject(mLogFileName) ;

	mCurrentStatus = new TStaticText(120,345,210,365,"Idle");
	AddObject(mCurrentStatus) ;

	theText = new TStaticText(5,195,115,215,"Current Temperature:");
	AddObject(theText) ;

	mCurrentTemperature = new TStaticText(130,195,200,215,"");
	AddObject(mCurrentTemperature) ;

	mCurrentChangeInStructure = new TStaticText(320,195,490,215,"");
	AddObject(mCurrentChangeInStructure) ;

	theText = new TStaticText(5,215,110,235,"Configuration \"Energy\":");
	AddObject(theText) ;

	theText = new TStaticText(220,195,320,215,"Last Structural Change:");
	AddObject(theText) ;

	mCurrentEnergy = new TStaticText(120,215,190,235,"");
	AddObject(mCurrentEnergy) ;

	theText = new TStaticText(5,365,215,385,"Press 'Control C' to Cancel");
	AddObject(theText) ;

	theText = new TStaticText(5,345,115,365,"Status of Calculation:");
	AddObject(theText) ;

	mExperimentalImage = nil ;
	mComputedImage = nil ;
	mCrystal	= new TCrystal(mWindow,mModelFrame->GetRect(),nucell) ;

	Initialize() ;
}

TStructureRefinementProgressPage::~TStructureRefinementProgressPage() {
	if (mExperimentalImage != nil) DeleteImage(mExperimentalImage) ;
	if (mComputedImage != nil) DeleteImage(mComputedImage) ;
	if (mCrystal != nil) delete mCrystal ;
	if (mOffscreen != nil) delete mOffscreen ;
}

void	TStructureRefinementProgressPage::Draw() {
	EraseRect(&mWindow->GetPortRect()) ;
	TPage::Draw() ;
	DrawExperimentalImage() ;
	DrawCurrentComputedImage() ;
	DrawCurrentModel() ;
	DrawEnergyWindow() ;
}

void	TStructureRefinementProgressPage::DoMouseDownInObject(TObject *obj,Point *pt) {
	if (obj == mReturn) {
		Hide() ;
		ExitFromStructureRefinementProgress() ;
		if (gSymOps != nil) {
			delete[] gSymOps ;
			gSymOps = nil ;
		}
	}
	else if (obj == mSaveReturn) {
		Hide() ;
		ExitFromStructureRefinementProgress() ;
		SaveRefinedStructure() ;		
		if (gSymOps != nil) {
			delete[] gSymOps ;
			gSymOps = nil ;
		}
	}

}


void	TStructureRefinementProgressPage::Initialize() {
	char 	text[256] ;
	short	ix,iy ;
	Rect	evolutionRect = mEnergyFrame->GetRect() ;

	// Initialize the energy window
	if (mOffscreen != nil) {
		OffsetRect(&evolutionRect,-evolutionRect.left,-evolutionRect.top) ;
		SetPort(mOffscreen) ;
	}
	FrameRect(&evolutionRect) ;
	InsetRect(&evolutionRect,1,1) ;
	SetFgColor(gWhitePixel) ;
	PaintRect(&evolutionRect) ;
	SetFgColor(gBlackPixel) ;
				
	// Allow for 10% growth in energy
	// and minimum to be 0
	if (gRelaxationParamRecord.thisEnergy > 0.0) {	
		mMaxValueOfEnergy = gRelaxationParamRecord.thisEnergy * 1.1 ;
		mMinValueOfEnergy = 0.0 ;
	}
	else {
		mMaxValueOfEnergy = 0.9*gRelaxationParamRecord.thisEnergy ;
		mMinValueOfEnergy = 10.0*gRelaxationParamRecord.thisEnergy ;  
	}
	mScaleFactor = (evolutionRect.bottom - evolutionRect.top) / 
			(mMaxValueOfEnergy-mMinValueOfEnergy) ;
	ix = evolutionRect.left ;
	iy = evolutionRect.bottom - 
			(gRelaxationParamRecord.thisEnergy-mMinValueOfEnergy)*mScaleFactor ;
	if (iy < evolutionRect.top) iy = evolutionRect.top ;
	if (iy > evolutionRect.bottom) iy = evolutionRect.bottom ;
	mCurrentEnergyPt.x = ix ;
	mCurrentEnergyPt.y = iy ;
	mNumDisplayIndex = 1 ;
	if (mOffscreen != nil) {
		SetPort(mWindow) ;
		CopyBits(mOffscreen,mWindow,&evolutionRect,&mEnergyFrame->GetRect()) ;
	}

					
	mCurrentStatus->SetItemText("Running !") ;

	sprintf(text,"%f",gRelaxationParamRecord.temperature.current) ;
	mCurrentTemperature->SetItemText(text) ;

	mReturn->Disable() ;
	mCurrentEnergy->SetItemText("") ;
	mSaveReturn->Disable() ;

}

void	TStructureRefinementProgressPage::DrawEnergyWindow() {

	if (mOffscreen != nil) {
		Rect	evolutionRect = mEnergyFrame->GetRect() ;
		OffsetRect(&evolutionRect,-evolutionRect.left,-evolutionRect.top) ;
		CopyBits(mOffscreen,mWindow,&evolutionRect,&mEnergyFrame->GetRect()) ;
	}
}

void	TStructureRefinementProgressPage::SetLogFileName(char *text) {
	mLogFileName->SetItemText(text) ;
}


void	TStructureRefinementProgressPage::UpdateCurrentParameters() {
	char 	text[256] ;
	long	i ;
		
	// Update the parameter window
	sprintf(text,"%f",gRelaxationParamRecord.temperature.current) ;
	mCurrentTemperature->SetItemText(text) ;
															
	// Write current energy to the window
	UpdateCurrentEnergy(gRelaxationParamRecord.thisEnergy) ;

	// Show the latest change in the structure

	i = *((gRefinementList.groupPtr+gIndex)->atomIndex) ;
	if (gUseSymmetryElements) {
		sprintf(text,"%s: %.3f,%.3f,%.3f (%.3f,%.3f)",
			chemsymb[(gBasis.atom+i)->no-1],
			(gBasis.atom+i)->xc,
			(gBasis.atom+i)->yc,
			(gBasis.atom+i)->zc,
			(gBasis.atom+i)->dw,
			(gBasis.atom+i)->occ) ;
	}
	else {
		sprintf(text,"%s: %.3f,%.3f,%.3f (%.3f,%.3f)",
			chemsymb[gProjCell->atno[i]-1],
			gProjCell->x[i],
			gProjCell->y[i],
			gProjCell->z[i],
			gProjCell->dw[i],
			gProjCell->occ[i]) ;
	}
	mCurrentChangeInStructure->SetItemText(text) ;
	
}

void	TStructureRefinementProgressPage::UpdateCurrentEnergy(float energy) {
	char 	text[256] ;
	short	ix,iy ;
	Rect	evolutionRect = mEnergyFrame->GetRect() ;

	if (mOffscreen != nil) {
		OffsetRect(&evolutionRect,-evolutionRect.left,-evolutionRect.top) ;
		SetPort(mOffscreen) ;
	}

	ix = mNumDisplayIndex%(evolutionRect.right-evolutionRect.left)
		+evolutionRect.left ;
	iy = evolutionRect.bottom - 
		(gRelaxationParamRecord.thisEnergy-mMinValueOfEnergy)*mScaleFactor ;
	if (iy > evolutionRect.bottom-3) {
		if (gRelaxationParamRecord.thisEnergy > 0.0) {	
			mMaxValueOfEnergy = gRelaxationParamRecord.thisEnergy * 1.2 ;
			mMinValueOfEnergy = 0.0 ;
		}
		else {
			mMaxValueOfEnergy = 0.8*gRelaxationParamRecord.thisEnergy ;
			mMinValueOfEnergy = 10.0*gRelaxationParamRecord.thisEnergy ;  
		}
		mScaleFactor = (evolutionRect.bottom - evolutionRect.top) / 
			(mMaxValueOfEnergy-mMinValueOfEnergy) ;
		iy = evolutionRect.bottom - 
			(gRelaxationParamRecord.thisEnergy-mMinValueOfEnergy)*mScaleFactor ;
		SetFgColor(gRedPixel) ;
		MoveTo(ix,evolutionRect.bottom) ;
		LineTo(ix,evolutionRect.top) ;
		mCurrentEnergyPt.x = ix ;
		mCurrentEnergyPt.y = iy ;
	} 
	if (iy < evolutionRect.top) iy = evolutionRect.top ;
	if (iy > evolutionRect.bottom) iy = evolutionRect.bottom ;
					
	if (ix == evolutionRect.left) {	// We just did a jump to go back
		SetFgColor(gWhitePixel) ;
		PaintRect(&evolutionRect) ;
		mCurrentEnergyPt.x = ix ;
		mCurrentEnergyPt.y = iy ;
	}
	else {	
		SetFgColor(gBluePixel) ;
		DrawLine(mCurrentEnergyPt.x,mCurrentEnergyPt.y,ix,iy) ;
		mCurrentEnergyPt.x = ix ;
		mCurrentEnergyPt.y = iy ;
	}
	SetFgColor(gBlackPixel) ;
	if (mOffscreen != nil) {
		SetPort(mWindow) ;
		CopyBits(mOffscreen,mWindow,&evolutionRect,&mEnergyFrame->GetRect()) ;
	}

	sprintf(text,"%f",energy) ;

	if (gRelaxationParamRecord.fittingMethod == kFTAmplCrossCorrelation) {
		sprintf(text,"%.4f, Shift: (%d,%d)",
			energy,gShiftN,gShiftM) ;
	}
	else {
		sprintf(text,"%.4f",energy) ;
	}
	mCurrentEnergy->SetItemText(text) ;
	mNumDisplayIndex++ ;
	XFlush(gDisplay) ;
}

void	TStructureRefinementProgressPage::DisplayLatestParameters() {

	// Write current energy to the window
	UpdateCurrentEnergy(gRelaxationParamRecord.thisEnergy) ;

	if (gAbortStatus) {
		mCurrentStatus->SetItemText("Cancelled by User !") ;
	}
	else {
		mCurrentStatus->SetItemText("Finished !") ;
	}
	mReturn->Enable() ;
	mSaveReturn->Enable() ;
	XFlush(gDisplay) ;

}

void	TStructureRefinementProgressPage::DrawExperimentalImage() {
	Rect	tRect ;
	extern	unsigned long	gWhitePixel,gBlackPixel ;

	tRect = mExperimentalImageFrame->GetRect() ;
	InsetRect(&tRect,1,1) ;
	SetFgColor(gDarkerGray) ;
	FillRect(&tRect) ;
	PenSize(2,2) ;
	SetFgColor(gBlackPixel) ;
	DrawLine(tRect.left,tRect.bottom,tRect.left,tRect.top) ;
	DrawLine(tRect.left,tRect.top,tRect.right,tRect.top) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(tRect.right,tRect.top,tRect.right,tRect.bottom) ;
	DrawLine(tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
	PenSize(1,1) ;
	SetFgColor(gBlackPixel) ;
	InsetRect(&tRect,4,4) ;
	if (mExperimentalImage == nil) {
		mExperimentalImage = CopyImage(gExperimentalImage) ;
		if (mExperimentalImage != nil) {
			DisplayImageInWindow(mExperimentalImage,mWindow,&tRect,false );
		}
	}
	else {
		DisplayImageInWindow(mExperimentalImage,mWindow,&tRect,false );
	}
	XFlush(gDisplay) ;
}

void	TStructureRefinementProgressPage::DrawCurrentComputedImage() {
	Rect	tRect ;
	extern	unsigned long	gWhitePixel,gBlackPixel ;

	if (mComputedImage == nil) return ;
	tRect = mCurrentImageFrame->GetRect() ;
	InsetRect(&tRect,1,1) ;
	SetFgColor(gDarkerGray) ;
	FillRect(&tRect) ;
	PenSize(2,2) ;
	SetFgColor(gBlackPixel) ;
	DrawLine(tRect.left,tRect.bottom,tRect.left,tRect.top) ;
	DrawLine(tRect.left,tRect.top,tRect.right,tRect.top) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(tRect.right,tRect.top,tRect.right,tRect.bottom) ;
	DrawLine(tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
	PenSize(1,1) ;
	SetFgColor(gBlackPixel) ;
	InsetRect(&tRect,4,4) ;
	DisplayImageInWindow(mComputedImage,mWindow,&tRect,true );
	XFlush(gDisplay) ;

}

void	TStructureRefinementProgressPage::UpdateComputedImage(TImage *image) {
	Rect	tRect ;
	extern	unsigned long	gWhitePixel,gBlackPixel ;

	if (image == nil) return ;
	tRect = mCurrentImageFrame->GetRect() ;
	InsetRect(&tRect,1,1) ;
	if (mComputedImage == nil) {
		SetFgColor(gDarkerGray) ;
		FillRect(&tRect) ;
		PenSize(2,2) ;
		SetFgColor(gBlackPixel) ;
		DrawLine(tRect.left,tRect.bottom,tRect.left,tRect.top) ;
		DrawLine(tRect.left,tRect.top,tRect.right,tRect.top) ;
		SetFgColor(gWhitePixel) ;
		DrawLine(tRect.right,tRect.top,tRect.right,tRect.bottom) ;
		DrawLine(tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
		PenSize(1,1) ;
		SetFgColor(gBlackPixel) ;
	}
	InsetRect(&tRect,4,4) ;
	if (mComputedImage != nil) {
		DeleteImage(mComputedImage) ;
		mComputedImage = nil ;
	}
	mComputedImage = CopyImage(image) ;
	if (mComputedImage != nil) {
		DisplayImageInWindow(mComputedImage,mWindow,&tRect,true );
	}
	XFlush(gDisplay) ;
}

void	TStructureRefinementProgressPage::DrawCurrentModel() {
	mCrystal->DrawCrystal() ; 
	FrameRect(&mModelFrame->GetRect()) ;
	XFlush(gDisplay) ;
}

void	ExitFromStructureRefinementProgress() {
	delete gStructureRefinementProgressPage ;
	gStructureRefinementProgressPage = nil ;
	set_menu(QUANTITATIVE) ;
	ClearDisplay() ;
	SetPage(gMainPage) ;
	DrawQuantitativePage() ;
}

