// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include <memory.h>
#include <X11/keysym.h>
#include "XDrivers.h"
#include "globals.h" 
#include "devices.h"
#include "numbers.h"

typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
extern MonitorDisplayType  gMonitorDisplayType ;
extern Visual	*gVisual ;

#define kMM		0x4D4D
#define kINTEL	0x4949
typedef	enum {kMMByteOrder = kMM, kIntelByteOrder = kINTEL} ByteOrder ;
extern	ByteOrder	gCPUByteOrder ;

Boolean	gShiftKeyDown = false ;
Boolean	gOptionKeyDown = false ;

extern unsigned char *gMapFrom16To24BitColor ;
extern unsigned char *gMapFrom24To16BitColor ;

extern unsigned short ByteSwapShort(unsigned short value) ;
long	near(float) ;


extern	short	gRedBitShift ;
extern	short	gGreenBitShift ;
extern	short	gBlueBitShift ;
extern	short	gReportVisuals ;

extern	ByteOrder	gCPUByteOrder ;
extern	ByteOrder	gServerByteOrder ;

Boolean	gCommandKeyDown = false ;
extern short gImageDepth ;
extern short gBitsPerPRGB ;

extern unsigned long gRedMask ;
extern unsigned long gGreenMask ;
extern unsigned long gBlueMask ;

extern short gImageDepth ;
extern Pixmap	gStipple ;
void PenPat(Pattern *pattern=nil) ;
#define	EOS	'\0'
void	SetModifierKeyDown(XEvent *event) ;
void	SetModifierKeyUp(XEvent *event) ;
Boolean	ControlCDown() ;
void DrawPixel(long x,long y){
	// -------------------------------------------------
	// Write a point on the current port	
	// -------------------------------------------------
	XDrawPoint(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),x,y) ;
} 

void DrawLine(long x1,long y1,long x2,long y2){
	// ------------------------------------------------------- 
	// Routine to draw a line in physical Device coordinates 
	// -------------------------------------------------------
	XDrawLine(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),x1,y1,x2,y2) ;
}

void DrawCross(short x,short y){
	// ------------------------------------------------------- 
	// Routine to draw a line in physical Device coordinates
	// ------------------------------------------------------- 
 
	XDrawLine(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),x-2,y,x+2,y) ;
	XDrawLine(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),x,y-2,x,y+2) ;
}

void MoveTo(long ix,long iy){
	// ---------------------------------------------------------
	// Move the "pen" to a point in physical device coordinates
	// ---------------------------------------------------------
	gCurrentTextPos.x = ix ;
	gCurrentTextPos.y = iy ;
}

void LineTo(long ix,long iy) {
	// -------------------------------------------------------
	// Routine to draw a line in physical Device coordinates
	// ------------------------------------------------------- 
	XDrawLine(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),gCurrentTextPos.x,gCurrentTextPos.y,
		ix,iy) ;
	gCurrentTextPos.x = ix ;
	gCurrentTextPos.y = iy ;
}	

void DrawChar(long x,long y,char c){
	// ----------------------------------------------------
	// Draw a character on the display with attributes 
	// ----------------------------------------------------
	char text[2] ;
	text[0] = c ; text[1] = '\0' ;
	XDrawString(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),x,y,text,1) ;
}

void DrawRect(long xlow,long ylow,long xhi,long yhi) {
	// -----------------------------------------------
	// Routine to draw a rectangle in PDC units 
	// -----------------------------------------------
	long linewidth ;
	Rect tRect ;
	long ilw ;

	linewidth = GetLineWidth() ;
	ilw = linewidth ;
	SetLineWidth(ilw) ;
	SetRect(&tRect,xlow,ylow,xhi,yhi) ;
	FrameRect(&tRect) ;
	SetLineWidth(linewidth) ;	
}

void DrawFilledText(long x,long y,char *text){
	// -----------------------------------------------
	// Routine to write text at x,y in PDC units     
	// -----------------------------------------------
	long numChars ;
	numChars = strlen(text) ;
	if (numChars > 0) {
		XDrawImageString(gDisplay,gCurrentPort->GetDrawable(),
			gCurrentPort->GetGC(),x,y,text,numChars) ;
		gCurrentTextPos.y = y ;
		gCurrentTextPos.x = x + 
			XTextWidth(gFontStructure,text,numChars) ;
	}
}

void	DrawFilledString(char *text) {
	long numChars ;
	numChars = strlen(text) ;
	if (numChars > 0) {
		XDrawImageString(gDisplay,gCurrentPort->GetDrawable(),
			gCurrentPort->GetGC(),gCurrentTextPos.x,gCurrentTextPos.y,
			text,numChars) ;
		gCurrentTextPos.x += XTextWidth(gFontStructure,text,numChars) ;
	}
}

void DrawText(long x,long y,char *text){
	// -----------------------------------------------
	// Routine to write text at x,y in PDC units 
	// -----------------------------------------------
	long numChars ;
	numChars = strlen(text) ;
	if (numChars > 0) {
		XDrawString(gDisplay,gCurrentPort->GetDrawable(),
			gCurrentPort->GetGC(),x,y,text,numChars) ;
		gCurrentTextPos.y = y ;
		gCurrentTextPos.x = x + 
			XTextWidth(gFontStructure,text,numChars) ;
	}
}

void	DrawString(char *text) {
	long numChars ;
	numChars = strlen(text) ;
	if (numChars > 0) {
		XDrawString(gDisplay,gCurrentPort->GetDrawable(),
			gCurrentPort->GetGC(),gCurrentTextPos.x,
			gCurrentTextPos.y,text,numChars) ;
		gCurrentTextPos.x += XTextWidth(gFontStructure,text,numChars) ;
	}
}

void DrawImageLine(long x,long y,long n,unsigned char *imageline) {
	long i ;

	if (gMonitorDisplayType == PSEUDOCOLOR) {
		gImageLine->data = (char *)imageline ;
	}
	else {
		for(i=0;i < n;i++) {
			XPutPixel(gImageLine,i,0,gColors[imageline[i]].pixel) ;
		}
	}
  
	XPutImage(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),gImageLine,0,0,x,y,n,1) ;
}

void DrawRGBImageLine(unsigned char *imageline,long x,long y,long n) {

	gImageLine->data = (char *)imageline ;
	// In this case the data coming in has been set up 
	// as indicies into a look up table that has as values
	// a long number where if the display is PSEUDOCOLOR then
	// the value is a number in the range 0 - NUM_GRAYS+NUM_COLORS
	// If the display is TRUECOLOR, the value consists of a long number
	// where we have 00 XX XX XX as 0 - RED - GREEN - BLUE
	XPutImage(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),gImageLine,0,0,x,y,n,1) ;
}

Boolean ReadImageLine(long x,long y,long n,unsigned char *imageline){
	long i ;
	unsigned char *ucp ;
	XImage *myRImageLine ;
    
	myRImageLine = XGetImage(gDisplay,gCurrentPort->GetDrawable(),
		x,y,n,1,AllPlanes,ZPixmap) ;
	if (myRImageLine == (XImage *)0) 
		return(false) ;

	if (gMonitorDisplayType == PSEUDOCOLOR) {
		// Copy all the stuff into imageline
		memcpy(imageline,myRImageLine->data,n) ;
	}
	else {
		// We are reading back rgb information which need to be	
		// converted back to unsigned char		
		// We really are just catching the Red component
		ucp = (unsigned char *)myRImageLine->data ;
		unsigned	short	*usp  = (unsigned short *)myRImageLine->data ;
		switch(gCPUByteOrder) {
			case kMMByteOrder:
				switch(gImageDepth) {
					case 15:
					case 16:
						for(i=0;i<n;i++) {
							imageline[i] = *usp & (0x1F << 5) ;
							usp++ ;
						}
						break ;
					case 24:
					case 32:
						ucp++ ;
						for(i=0;i<n;i++) {
							imageline[i] = *ucp ;
							ucp += 4 ;
						}
						break ;
				}
				break ;
			case kIntelByteOrder:
				switch(gImageDepth) {
					case 15:
					case 16:
						for(i=0;i<n;i++) {
							imageline[i] = *usp & (0x1F << 5) ;
							usp++ ;
						}
						break ;
					case 24:
					case 32:
						for(i=0;i<n;i++) {
							imageline[i] = *ucp ;
							ucp += 4 ;
						}
						break ;
				}
				break ;
		}
	}
	XDestroyImage(myRImageLine) ;
	return(true) ;
}

Boolean ReadRGBImageLine(long x,long y,long n,unsigned char *imageline){
	XImage *myRImageLine ;
    
	myRImageLine = XGetImage(gDisplay,gCurrentPort->GetDrawable(),
		x,y,n,1,AllPlanes,ZPixmap) ;
	if (myRImageLine == (XImage *)0) 
		return(false) ;

	if (gMonitorDisplayType == PSEUDOCOLOR) {
		// Copy all the stuff into imageline
		memcpy(imageline,myRImageLine->data,n) ;
	}
	else {
		switch(gImageDepth) {
			case 15:
			case 16:
				memcpy(imageline,myRImageLine->data,n*sizeof(short)) ;
				break ;
			case 24:
			case 32:
				memcpy(imageline,myRImageLine->data,n*sizeof(long)) ;
				break ;
		}
	}
	XDestroyImage(myRImageLine) ;
	return(true) ;
}

Boolean DrawImage(unsigned char *image,long x,long y,long width,long height) {
	long 		i,j,ndx ;
	XImage 		*myImage ;
	unsigned char *imageRGB,*ucp ;
	unsigned	short *usp ;

	imageRGB = nil ;
 	if (gMonitorDisplayType == PSEUDOCOLOR) {
	
		// Create a XImage for writing the image into 
		myImage = XCreateImage(gDisplay,
			gVisual,gImageDepth,ZPixmap,
			0,(char *)image,width,height,8,0) ;
	}
	else {
		imageRGB = (unsigned char *) new unsigned long[width*height] ;
		if (imageRGB == NULL)
			return(0) ;
		ucp = imageRGB ;
		usp = (unsigned short *)imageRGB ;
		switch(gCPUByteOrder) {
			case kMMByteOrder:
				switch(gImageDepth) {
					case 15:
					case 16:
						myImage = XCreateImage(gDisplay,
							gVisual,gImageDepth,ZPixmap,
							0,(char *)imageRGB,width,height,16,0) ;
						ndx = 0 ;	
						for(j=0; j < height; j++) 
							for(i=0; i < width; i++) {
								XPutPixel(myImage,i,j,gColors[image[ndx]].pixel) ;
								ndx++ ;
							}	
						break ;
					case 24:
					case 32:
						myImage = XCreateImage(gDisplay,
							gVisual,gImageDepth,ZPixmap,
							0,(char *)imageRGB,width,height,32,0) ;
						ndx = 0 ;	
						for(j=0; j < height; j++) 
							for(i=0; i < width; i++) {
								XPutPixel(myImage,i,j,gColors[image[ndx]].pixel) ;
								ndx++ ;
							}	
						break ;
				}
				break ;
			case kIntelByteOrder:
				switch(gImageDepth) {
					case 15:
					case 16:
						myImage = XCreateImage(gDisplay,
							gVisual,gImageDepth,ZPixmap,
							0,(char *)imageRGB,width,height,16,0) ;
						ndx = 0 ;	
						for(j=0; j < height; j++) 
							for(i=0; i < width; i++) {
								XPutPixel(myImage,i,j,gColors[image[ndx]].pixel) ;
								ndx++ ;
							}	
							break ;
					case 24:
					case 32:
						myImage = XCreateImage(gDisplay,
							gVisual,gImageDepth,ZPixmap,
							0,(char *)imageRGB,width,height,32,0) ;
						ndx = 0 ;	
						for(j=0; j < height; j++) 
							for(i=0; i < width; i++) {
								XPutPixel(myImage,i,j,gColors[image[ndx]].pixel) ;
								ndx++ ;
							}	
						break ;
				}
				break ;
		}
	}
	
	if (myImage == (XImage *)0) {
		if (imageRGB != nil) delete[] imageRGB ;
		return(false) ;
	}
	XPutImage(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),myImage,0,0,x,y,width,height) ;
	XFlush(gDisplay) ;
	myImage->data = NULL ;
	XDestroyImage(myImage) ;
	if (gMonitorDisplayType == TRUECOLOR) {
		if (imageRGB != nil) delete[] imageRGB ;
	}
	return(true) ;
}

Boolean DrawRGBImage(unsigned char *image,long x,long y,long width,long height){
	XImage *myImage ;
	extern short gImageDepth ;
	
 	if (gMonitorDisplayType == PSEUDOCOLOR) {
	
		// Create a XImage for writing the image into 
		myImage = XCreateImage(gDisplay,
			gVisual,gImageDepth,ZPixmap,
			0,(char *)image,width,height,8,0) ;
	}
	else {
		/* Create a XImage for writing the image into */
		switch(gImageDepth) {
			case 15:
			case 16:
				myImage = XCreateImage(gDisplay,
					gVisual,gImageDepth,ZPixmap,
					0,(char *)image,width,height,16,0) ;
				break ;
			case 24:
			case 32:
				myImage = XCreateImage(gDisplay,
					gVisual,gImageDepth,ZPixmap,
					0,(char *)image,width,height,32,0) ;
				break ;
		}	
	}
	if (myImage == (XImage *)0) {
		return(false) ;
	}

	// Make sure we crop to the display size
	long ww = OutDevice->xres - x - 2 ;
	long hh = OutDevice->yres - y - 2 ;

	XPutImage(gDisplay,gCurrentPort->GetDrawable(),
		gCurrentPort->GetGC(),myImage,0,0,x,y,ww,hh) ;
	XFlush(gDisplay) ;
	myImage->data = NULL ;
	XDestroyImage(myImage) ;
	return(true) ;
}

Boolean ReadImage(unsigned char *image,long x,long y,long width,long height) {
	long	i ;
	XImage *myImage ;
	unsigned char *ucp ;
       
	// Create a XImage for reading the image into
	myImage = XGetImage(gDisplay,gCurrentPort->GetDrawable(),
		x,y,width,height,AllPlanes,ZPixmap) ;
	if (myImage == (XImage *)0) {
		return(false) ;
	}

	// Copy the data into 'image' */
	if (gMonitorDisplayType == PSEUDOCOLOR) {
		// Copy all the stuff into imageline
		memcpy(image,myImage->data,width*height) ;
	}
	else {
		// We are reading back rgb information which need to be	
		// converted back to unsigned char	
		ucp = (unsigned char *)myImage->data ;
		unsigned	short	*usp  = (unsigned short *)myImage->data ;
		switch(gCPUByteOrder) {
			case kMMByteOrder:
				switch(gImageDepth) {
					case 15:
					case 16:
						for(i=0;i<width*height;i++) {
							image[i] = (*usp & 0x1F) ;	// Red component
							usp++ ;
						}
						break ;
					case 24:
					case 32:
						ucp++ ;
						for(i=0;i<width*height;i++) {
							image[i] = *ucp ;
							ucp += 4 ;
						}
						break ;
				}
				break ;
			case kIntelByteOrder:
				switch(gImageDepth) {
					case 15:
					case 16:
						for(i=0;i<width*height;i++) {
							image[i] = *usp & 0x1F ;
							usp++ ;
						}
						break ;
					case 24:
					case 32:
						for(i=0;i<width*height;i++) {
							image[i] = *ucp ;
							ucp += 4 ;
						}
						break ;
				}
				break ;
		}
		

	}
	XFlush(gDisplay) ;
	XDestroyImage(myImage) ;
	return(true) ;
}

Boolean ReadRGBImage(unsigned char *image,long x,long y,long width,long height) {
	XImage *myImage ;
       
	// Create a XImage for reading the image into
	myImage = XGetImage(gDisplay,gCurrentPort->GetDrawable(),
		x,y,width,height,AllPlanes,ZPixmap) ;
	if (myImage == (XImage *)0) {
		return(false) ;
	}

	// Copy the data into 'image' */
	if (gMonitorDisplayType == PSEUDOCOLOR) {
		// Copy all the stuff into imageline
		memcpy(image,myImage->data,width*height) ;
	}
	else {
		switch(gImageDepth) {
			case 15:
			case 16:
				memcpy(image,myImage->data,width*height*sizeof(short)) ;
				break ;
			case 24:
			case 32:
				memcpy(image,myImage->data,width*height*sizeof(long)) ;
				break ;
		}
	}
	XFlush(gDisplay) ;
	XDestroyImage(myImage) ;
	return(true) ;
}

void RDrawRect(Rect *tRect){
	long width,height ;

	width = tRect->right - tRect->left ;
	height = tRect->bottom - tRect->top ;
	XDrawRectangle(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
		tRect->left,tRect->top,width,height) ;
}

void FrameRect(Rect *tRect) {
	XDrawRectangle(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
		tRect->left,tRect->top,tRect->right-tRect->left,
		tRect->bottom-tRect->top) ;
}

void FillRect(Rect *tRect){
	XFillRectangle(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
		tRect->left,tRect->top,tRect->right-tRect->left,
		tRect->bottom-tRect->top) ;
}

void InvertRect(Rect *tRect) {
	XSetFunction(gDisplay,gCurrentPort->GetGC(),GXxor) ;
	XSetFillStyle(gDisplay,gCurrentPort->GetGC(),FillOpaqueStippled) ;
	XFillRectangle(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
		tRect->left,tRect->top,tRect->right-tRect->left,
		tRect->bottom-tRect->top) ;
	XSetFillStyle(gDisplay,gCurrentPort->GetGC(),FillSolid) ;
	XSetFunction(gDisplay,gCurrentPort->GetGC(),GXcopy) ;
}

void EraseRect(Rect *tRect){
	uLong GetBkColor() ;
	unsigned long savefg ;

	savefg = gCurrentPort->GetFgColor() ;
	gCurrentPort->SetFgColor(gCurrentPort->GetBkColor()) ;
	XFillRectangle(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
        	tRect->left,tRect->top,tRect->right-tRect->left,
        	tRect->bottom-tRect->top) ;
	gCurrentPort->SetFgColor(savefg) ;
}

void EraseCircle(short x,short y,short rad){
	unsigned long oldFg ;
	long x1,y1,width,height ;
	uLong GetBkColor() ;

	x1 = x - rad ;
	y1 = y - rad ;
	width = 2*rad ;
	height = 2*rad ;
	oldFg = gCurrentPort->GetFgColor() ;
	gCurrentPort->SetFgColor(gCurrentPort->GetBkColor()) ;
	XFillArc(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
		x1,y1,width,height,0,360*64) ;
	gCurrentPort->SetFgColor(oldFg) ;
} 
   
void ClearDisplay() {
	Rect tRect ;
	//unsigned long oldFg,GetBkColor() ;

	tRect = gCurrentPort->GetPortRect() ;
	EraseRect(&tRect) ;
	//oldFg = gCurrentPort->GetFgColor() ;
	//XSetForeground(gDisplay,gCurrentPort->GetGC(),GetBkColor()) ;
	//FillRect(&tRect) ;
	//XSetForeground(gDisplay,gCurrentPort->GetGC(),oldFg) ;
}

void GetMouseDownPosition(long *x,long *y){
	XEvent event ;
	Point pt ;   
	long done = 0 ;
	Window root,child ;
	int root_x,root_y,win_x,win_y ;
	unsigned int keys_buttons ;

	// Does not return until the mouse is pressed down in the window
	// that is the current port	
	while (!done) {
		XNextEvent(gDisplay,&event) ;
		if (event.type == ButtonPress) {
			GetMouse(&pt) ;
			if (XQueryPointer(gDisplay,gCurrentPort->GetDrawable(),&root,&child,
					&root_x,&root_y,&win_x,&win_y,&keys_buttons)) {
				*x = win_x ;
				*y = win_y ;
				done = 1 ;
			}
		}
	}
}

void GetPort(GrafPtr *port) {
	*port = gCurrentPort ;
}

void SetPort(GrafPtr port) {
	gCurrentPort = port ;
}

void SetRect(Rect *tRect,short x1,short y1,short x2,short y2) {

	tRect->left	= x1 ;
	tRect->top  = y1 ;
	tRect->right = x2 ;
	tRect->bottom = y2 ;
}

Boolean PtInRect(Point *pt,Rect *tRect) {

	if (pt->x < tRect->left)
		return(false) ;
	if (pt->x > tRect->right)
		return(false) ;
	if (pt->y < tRect->top)
		return(false) ;
	if (pt->y > tRect->bottom)
		return(false) ;
	return(true) ;
}

void OffsetRect(Rect *tRect,short x,short y){

	tRect->left += x ;
	tRect->right += x ;
	tRect->top += y ;
	tRect->bottom += y ;
}

void InsetRect(Rect *tRect,short x,short y){
	tRect->left += x ;
	tRect->right -= x ;
	tRect->top += y ;
	tRect->bottom -= y ;
}

void DrawCircle(short x,short y,short rad){
	long x1,y1,width,height ;
	x1 = x - rad ;
	y1 = y - rad ;
	width = 2*rad ;
	height = 2*rad ;
	XDrawArc(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
		x1,y1,width,height,0,360*64) ;
}

void FillCircle(short x,short y,short rad){
	long x1,y1,width,height ;
	x1 = x - rad ;
	y1 = y - rad ;
	width = 2*rad ;
	height = 2*rad ;
	XFillArc(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
		x1,y1,width,height,0,360*64) ;
}

void FrameOval(Rect *rect) {
	long x1,y1,width,height,rad ;
	x1 = rect->left ;
	y1 = rect->top ;
	rad = (rect->left+rect->right)/2 ;
	width = 2*rad ;
	height = 2*rad ;
	XDrawArc(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
		x1,y1,width,height,0,360*64) ;
}

void PaintOval(Rect *rect) {
	long x1,y1,width,height,rad ;
	x1 = rect->left ;
	y1 = rect->top ;
	rad = (rect->left+rect->right)/2 ;
	width = 2*rad ;
	height = 2*rad ;
	XFillArc(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
		x1,y1,width,height,0,360*64) ;
}

unsigned long GetFgColor() {
	return(gCurrentPort->GetFgColor()) ;
}

unsigned long GetBkColor() {
	return(gCurrentPort->GetBkColor()) ;
}

void SetFgColor(unsigned long value) {
	gCurrentPort->SetFgColor(value) ;
}

void SetBkColor(unsigned long value) {
  	gCurrentPort->SetBkColor(value) ;
}

void TextSize(short size){
}

void TextFace(short face){
}

void PenSize(short width,short height) {
	XGCValues gcv ;

	gcv.line_width = width ;
	XChangeGC(gDisplay,gCurrentPort->GetGC(),GCLineWidth,&gcv) ;
}

void PenMode(DrawingMode mode) {
	switch(mode) {
		case kCopy:
			XSetFunction(gDisplay,gCurrentPort->GetGC(),GXcopy) ;
			break ;
		case kXOr:
			XSetFunction(gDisplay,gCurrentPort->GetGC(),GXinvert) ;
			//XSetFunction(gDisplay,gCurrentPort->GetGC(),GXxor) ;
			break ;
		case kOr:
			XSetFunction(gDisplay,gCurrentPort->GetGC(),GXor) ;
			break ;
	}
}

void PenPat(Pattern *pattern) {
	//XSetFillStyle(gDisplay,gCurrentPort->GetGC(),FillStippled) ;	
}

void PenNormal() {
	XGCValues gcv ;

	gcv.line_width = 1 ;
	XChangeGC(gDisplay,gCurrentPort->GetGC(),GCLineWidth,&gcv) ;
	XSetFunction(gDisplay,gCurrentPort->GetGC(),GXcopy) ;
	//XSetFillStyle(gDisplay,gCurrentPort->GetGC(),FillSolid) ;
} 
   
void SetWatchCursor() {
	extern Cursor gWatchCursor ;
	XDefineCursor(gDisplay,gCurrentPort->GetDrawable(),gWatchCursor) ;
}

void InitCursor() {
	XUndefineCursor(gDisplay,gCurrentPort->GetDrawable()) ;
}

KeySym	GetEventKey(XEvent *theEvent) {
	KeySym 	keysym ;
	long	count ;
	XComposeStatus compose ;
	char 	InputBuffer[80];

	count = XLookupString((XKeyEvent *)theEvent,InputBuffer,80,&keysym,&compose) ;
	return 	keysym ; 
}

void SetTrackingMode() {
	// ------------------------------------------------
	// Set up our mode such that we will be notified
	// about cursor movements and mouseup events
	// ------------------------------------------------
	XSelectInput(gDisplay,gCurrentPort->GetDrawable(),ExposureMask |
		ButtonPressMask | ButtonReleaseMask | 
		ButtonMotionMask | PointerMotionHintMask) ;
}

void UnSetTrackingMode() {
	// ------------------------------------------------
	// Set up our mode such that we will be notified
	// about cursor movements and mouseup events
	// ------------------------------------------------	
	XSelectInput(gDisplay,gCurrentPort->GetDrawable(),
		ExposureMask | ButtonPressMask) ;
}

Boolean	ButtonDown() {
	static Boolean isDown = false ;
	if (!isDown) {
		SetTrackingMode() ;
		isDown = true ;
	}
	if (Button())
		return true ; 
	else {
		isDown = false ;
		UnSetTrackingMode() ;
		return false ;
	}
#ifndef __VMS
   return false ;
#endif
}
	
Boolean Button() {
	// ------------------------------------------------
	// Return true if no mouse up event is in the queue
	// signifying that the mouse is still down from a
	// previous mouse down event	
	// This routine only works if SetTrackingMode has
	// been called prior to calling Button	
	// ------------------------------------------------	
	XEvent event ;
	if (XCheckMaskEvent(gDisplay,ButtonReleaseMask,&event))
		return(false) ;

	return(true) ;
}

/*
void GetGlobalMouse(Point *where) {
	// ------------------------------------------------
	// Get the location of the mouse in global coordinates
	// ------------------------------------------------	
	Window root,child ;
	long root_x,root_y,win_x,win_y ;
	unsigned long keys_buttons ;

	XQueryPointer(gDisplay,gCurrentPort->GetDrawable(),&root,&child,
    		&root_x,&root_y,&win_x,&win_y,&keys_buttons) ;
	where->x = root_x ;
	where->y = root_y ;
}
*/
void GetMouse(Point *where){
	// ------------------------------------------------
	// Get the location of the mouse in local window
	// coordinates.  Relative to Image Window	
	// ------------------------------------------------	
	Window root,child ;
	int root_x,root_y,win_x,win_y ;
	unsigned int keys_buttons ;

	if (XQueryPointer(gDisplay,gCurrentPort->GetDrawable(),&root,&child,
    		&root_x,&root_y,&win_x,&win_y,&keys_buttons)) {
		where->x = win_x ;
		where->y = win_y ;
	}
	else {
		where->x = root_x ;
		where->y = root_y ;
	}
}

void	GetGlobalMouse(Point *where) {
	// ------------------------------------------------
	// Get the location of the mouse in global coordinates
	// ------------------------------------------------	
	Window root,child ;
	int root_x,root_y,win_x,win_y ;
	unsigned int keys_buttons ;

	XQueryPointer(gDisplay,RootWindow(gDisplay,gScreen),&root,&child,
    		&root_x,&root_y,&win_x,&win_y,&keys_buttons) ;
	where->x = win_x ;
	where->y = win_y ;
}

long StringWidth(char *text) {
	return(XTextWidth(gFontStructure,text,strlen(text))) ;
}

long	TextHeight() {
	return gFontStructure->ascent ;
}

void UnionRect(Rect *srcRect1,Rect *srcRect2,Rect *destRect){
	Rect aRect,bRect ;

	aRect = *srcRect1 ;
	bRect = *srcRect2 ;
	destRect->left = (aRect.left < bRect.left ? aRect.left : bRect.left) ;
	destRect->top = (aRect.top < bRect.top ? aRect.top : bRect.top) ;
	destRect->right = (aRect.right > bRect.right ? aRect.right : bRect.right) ;
	destRect->bottom = (aRect.bottom > bRect.bottom ? 
    			aRect.bottom : bRect.bottom) ;
}

void RegionRect(Region rgn,Rect *rect) {
	XRectangle theXRect ;
	XClipBox(rgn,&theXRect) ;
	rect->left 		= theXRect.x ;
	rect->top 		= theXRect.y ;
	rect->right 	= theXRect.x+theXRect.width ;
	rect->bottom 	= theXRect.y+theXRect.height ;
}

void RectRgn(Region rgn,Rect *rect) {
	XRectangle theXRect ;
	theXRect.x		= rect->left ;
	theXRect.y		= rect->top ;
	theXRect.width	= rect->right-rect->left ;
	theXRect.height	= rect->bottom-rect->top ;
	XUnionRectWithRegion(&theXRect,rgn,rgn) ;	
}

Boolean RectInRgn(Rect *rect,Region rgn) {
	XRectangle theXRect ;
	theXRect.x		= rect->left ;
	theXRect.y		= rect->top ;
	theXRect.width	= rect->right-rect->left ;
	theXRect.height	= rect->bottom-rect->top ;
	if (XRectInRegion(rgn,theXRect.x,theXRect.y,theXRect.width,theXRect.height) == RectangleOut)
		return false ;
	return true ;
}

Boolean RectISectRect(Rect *aRect,Rect *bRect){
    
    if (aRect->right < bRect->left) 
    	return(false) ;
    else if (aRect->left > bRect->right)
    	return(false) ;
    else if (aRect->bottom < bRect->top)
    	return(false) ;
    else if (aRect->top > bRect->bottom)
    	return(false) ;
    else
    	return(true) ;
}

void	GetClip(RgnHandle region) {
	Rect theRect ;
	RegionRect(gCurrentPort->GetClipRegion(),&theRect) ;
	RectRgn(region,&theRect) ;
}
	
void	SetClip(RgnHandle region) {
 
	XSetRegion(gDisplay,gCurrentPort->GetGC(),region) ;
	gCurrentPort->SetClipRegion(region) ;
}

void	SetClip(Rect &rect) {
	XRectangle clipRect ;
	clipRect.x = rect.left ;
	clipRect.y = rect.top ;
	clipRect.width = rect.right - rect.left ;
	clipRect.height = rect.bottom - rect.top ;
	XSetClipRectangles(gDisplay,gCurrentPort->GetGC(),0,0,
				&clipRect,1,Unsorted) ;
	gCurrentPort->SetClipRegion(rect) ;
}

void	FormRGBForeColor(RGBColor *color) {
	unsigned	long	red,green,blue ;
	float		val ;
	val	= color->red/65535.*255. ;
	red	= (unsigned long)val ;
	val	= color->green/65535.*255. ;
	green	= (unsigned long)val ;
	val	= color->blue/65535.*255. ;
	blue	= (unsigned long)val ;
	fl_mapcolor(FL_FREE_COL1,red,green,blue) ;
	SetFgColor(fl_get_pixel(FL_FREE_COL1)) ;
}

void	FormRGBBackColor(RGBColor *color) {
	unsigned	long	red,green,blue ;
	float		val ;
	val	= color->red/65535.*255. ;
	red	= (unsigned long)val ;
	val	= color->green/65535.*255. ;
	green	= (unsigned long)val ;
	val	= color->blue/65535.*255. ;
	blue	= (unsigned long)val ;
	fl_mapcolor(FL_FREE_COL1,red,green,blue) ;
	SetBkColor(fl_get_pixel(FL_FREE_COL1)) ;
}

void	SetFormFgColor(FL_COLOR color) {
	SetFgColor(fl_get_pixel(color)) ;
}

void	SetFormBkColor(FL_COLOR color) {
	SetBkColor(fl_get_pixel(color)) ;
}

void	RGBForeColor(RGBColor *color) {
}

void	RGBBackColor(RGBColor *color) {
}

void	SetModifierKeyDown(XEvent *event) {
	long 	count ;
	KeySym 	keysym ;
	XComposeStatus compose ;
	char 	InputBuffer[80] ;

	count = XLookupString((XKeyEvent *)event,InputBuffer,
			80,&keysym,&compose) ;
	InputBuffer[count] = EOS ;
	if (keysym == XK_Alt_L || keysym == XK_Alt_R
			|| keysym == XK_Meta_L || keysym == XK_Meta_R
			|| keysym == XK_Control_R || keysym == XK_Control_L)
		gOptionKeyDown = true ;
	if (keysym == XK_Shift_L || keysym == XK_Shift_R)
		gShiftKeyDown = true ;
} 

Boolean	ControlCDown() {
	long 	count ;
	KeySym 	keysym ;
	XComposeStatus compose ;
	char 	InputBuffer[80] ;
	XEvent	theEvent ;
	static	Boolean	ControlKeyDown = false ;

	if (XCheckTypedWindowEvent(gDisplay,gCurrentPort->GetDrawable(),
			KeyPress,&theEvent)) {
		count = XLookupString((XKeyEvent *)&theEvent,InputBuffer,
			80,&keysym,&compose) ;
		InputBuffer[count] = EOS ;
		if (keysym == XK_Control_L || keysym == XK_Control_R) {
			ControlKeyDown = true ;
			// We will just returned true even if the second key is not c
			return true ;
		}
		else {
			if (keysym == XK_c) {
				if (ControlKeyDown) {
					ControlKeyDown = false ;
					return true ;
				}
				return true ;
			}
			ControlKeyDown = false ;
			return false ;
		}
	}
	return false ;
}

void	SetModifierKeyUp(XEvent *event) {
	long 	count ;
	KeySym 	keysym ;
	XComposeStatus compose ;
	char 	InputBuffer[80] ;

	count = XLookupString((XKeyEvent *)event,InputBuffer,
			80,&keysym,&compose) ;
	InputBuffer[count] = EOS ;
	if (keysym == XK_Alt_L || keysym == XK_Alt_R
			|| keysym == XK_Meta_L || keysym == XK_Meta_R 
			|| keysym == XK_Control_R || keysym == XK_Control_L)
		gOptionKeyDown = false ;
	if (keysym == XK_Shift_L || keysym == XK_Shift_R)
		gShiftKeyDown = false ;
} 

Boolean	OptionKeyDown() {
	return gOptionKeyDown ;
} 

Boolean	ShiftKeyDown() {
	return gShiftKeyDown ;

} 


void CopyBits(GrafPtr source,GrafPtr dest,Rect *srcRect,Rect *dstRect,CopyMode mode) {

	short	destWidth,destHeight,srcWidth,srcHeight ;
	unsigned short	w1,w2,w3,w4 ;
	switch(mode) {
		case srcCopy:
			XSetFunction(gDisplay,dest->GetGC(),GXcopy) ;
			break ;
		case srcOr:
			XSetFunction(gDisplay,dest->GetGC(),GXor) ;
			break ;
	}
	
	destWidth = dstRect->right - dstRect->left ;
	destHeight = dstRect->bottom - dstRect->top ;
	srcWidth = srcRect->right - srcRect->left ;
	srcHeight = srcRect->bottom - srcRect->top ;
	if (srcWidth == destWidth && srcHeight == destHeight) {
		XCopyArea(gDisplay,source->GetDrawable(),dest->GetDrawable(),
			dest->GetGC(),srcRect->left,
			srcRect->top,srcRect->right-srcRect->left,
			srcRect->bottom-srcRect->top,dstRect->left,dstRect->top) ;
	}
	else {
		long 	row,col ;
		float	n_npr,m_mpr ;
		int		n1,m1 ;
		unsigned char *ucp ;
		XImage *myImage ;
 		unsigned char	*destDataC,*srcDataC ;
  		unsigned short	*usp,*destDataS ;
  		   
		// Create a XImage and read the source pixels 		
		myImage = XGetImage(gDisplay,source->GetDrawable(),
			srcRect->left,srcRect->top,srcWidth,
			srcHeight,AllPlanes,ZPixmap) ;
		if (myImage == (XImage *)0) {
			return ;
		}


		XImage	*destImage = nil ;
		unsigned char *destData = nil ;
		
		n_npr = ((float)srcWidth)/destWidth ;
		m_mpr = ((float)srcHeight)/destHeight ;
		

  		int	src_bytes_per_line = myImage->bytes_per_line ;
		
		// Create an image to hold the destination area	
	    if (gMonitorDisplayType == PSEUDOCOLOR) {
	    	destData = new unsigned char[destWidth*destHeight] ;
			if (destData == nil) {
				return ;
			}
			gImageAttributes.colormap = gColorMap ;
			gImageDepth = 8 ;	// Force the depth to 8
	    	destImage = XCreateImage(gDisplay,gVisualInfo.visual,gImageDepth,
	    			ZPixmap,0,(char *)destData,destWidth,destHeight,8,0) ;
			if (destImage == (XImage *)0) 
				return ;
				
	    }
	    else {
	    	switch(gImageDepth) {
	    		case 15:
	    		case 16:
			    	destData = new unsigned char[sizeof(short)*destWidth*destHeight] ;
					if (destData == nil) {
						return ;
					}
			    	destImage = XCreateImage(gDisplay,gVisualInfo.visual,gImageDepth,
			    			ZPixmap,0,(char *)destData,destWidth,destHeight,16,0) ;
					if (destImage == (XImage *)0) {
						return ;
					}
	    			break ;
	    		case 24:
	    		case 32:
			    	destData = new unsigned char[sizeof(long)*destWidth*destHeight] ;
					if (destData == nil) {
						return ;
					}
			    	destImage = XCreateImage(gDisplay,gVisualInfo.visual,gImageDepth,
			    			ZPixmap,0,(char *)destData,destWidth,destHeight,32,0) ;
					if (destImage == (XImage *)0) {
						return ;
					}
	    			break ;
	    	}
	    }

  		int	dest_bytes_per_line = destImage->bytes_per_line ;
		unsigned	long *ulp,*destDataL ;
		if (gMonitorDisplayType == PSEUDOCOLOR) {
			if (gReportVisuals) fprintf(stderr,"Case Pseudocolor\n");
			// All the data is unsigned char
			ucp = (unsigned char *)myImage->data ;
			for(row=0;row<destHeight;row++) {
				m1 = row*m_mpr ;
				m1 *= src_bytes_per_line ;
				destDataC = (unsigned char *)(destImage->data + row*dest_bytes_per_line) ;
				for(col=0;col<destWidth;col++) {
					n1 = col*n_npr ;
					*destDataC++ = (unsigned char)*(ucp+m1+n1) ;
				}
			}
			XPutImage(gDisplay,dest->GetDrawable(),
				dest->GetGC(),destImage,0,0,dstRect->left,dstRect->top,destWidth,destHeight) ;
		}
		else {
			// It now depends on the depth of the screen
			long	ioffset ;
			// Check to see if any of steps these are integers
			long	n2=0,m2=0 ;
			if (NumberIsInteger(n_npr)) n2 = near(n_npr) ;
			if (NumberIsInteger(m_mpr)) m2 = near(n_npr) ;
			long	rowjump ;
			if (n2 != 0 && m2 != 0) {
				switch(gImageDepth) {
					case 15:
					case 16:
						srcDataC = (unsigned char *)myImage->data ;
						destDataC = (unsigned char *)destImage->data ;
						destDataS = (unsigned short *)destImage->data ;
						ucp = srcDataC ;
						rowjump = m2*src_bytes_per_line ;
						for(row=0,m1=0;row<destHeight;row++,m1+=m2) {
							// row index into source image
							destDataS = (unsigned short *)destDataC ;
							for(col=0,n1=0;col<destWidth;col++,n1 += n2) {
								usp = (unsigned short *)ucp ;
								usp += n1 ;
								*destDataS++ = *usp ;
							}
							ucp	+= rowjump ;
							destDataC += dest_bytes_per_line ;
						}
						XPutImage(gDisplay,dest->GetDrawable(),
							dest->GetGC(),destImage,0,0,dstRect->left,dstRect->top,destWidth,destHeight) ;
						break ;
					case 24:
					case 32:
						srcDataC = (unsigned char *)myImage->data ;
						destDataC = (unsigned char *)destImage->data ;
						destDataL = (unsigned long *)destImage->data ;
						ucp = srcDataC ;
						rowjump = m2*src_bytes_per_line ;
						for(row=0,m1=0;row<destHeight;row++,m1+=m2) {
							// row index into source image
							destDataL = (unsigned long *)destDataC ;
							for(col=0,n1=0;col<destWidth;col++,n1 += n2) {
								ulp = (unsigned long *)ucp ;
								ulp += n1 ;
								*destDataL++ = *ulp ;
							}
							ucp	+= rowjump ;
							destDataC += dest_bytes_per_line ;
						}
						XPutImage(gDisplay,dest->GetDrawable(),
							dest->GetGC(),destImage,0,0,dstRect->left,dstRect->top,destWidth,destHeight) ;
						break ;
				}
			}
			else if (n2 != 0) {
				switch(gImageDepth) {
					case 15:
					case 16:
						srcDataC = (unsigned char *)myImage->data ;
						destDataC = (unsigned char *)destImage->data ;
						destDataS = (unsigned short *)destImage->data ;
						for(row=0;row<destHeight;row++) {
							// row index into source image
							m1 = (int)(row*m_mpr) ;
							destDataS = (unsigned short *)destDataC ;
							ucp = srcDataC + m1*src_bytes_per_line ;
							for(col=0,n1=0;col<destWidth;col++,n1 += n2) {
								usp = (unsigned short *)ucp ;
								usp += n1 ;
								*destDataS++ = *usp ;
							}
							destDataC += dest_bytes_per_line ;
						}
						XPutImage(gDisplay,dest->GetDrawable(),
							dest->GetGC(),destImage,0,0,dstRect->left,dstRect->top,destWidth,destHeight) ;
						break ;
					case 24:
					case 32:
						srcDataC = (unsigned char *)myImage->data ;
						destDataC = (unsigned char *)destImage->data ;
						destDataL = (unsigned long *)destImage->data ;
						for(row=0;row<destHeight;row++) {
							// row index into source image
							m1 = (int)(row*m_mpr) ;
							destDataL = (unsigned long *)destDataC ;
							ucp = srcDataC + m1*src_bytes_per_line ;
							for(col=0,n1=0;col<destWidth;col++,n1 += n2) {
								ulp = (unsigned long *)ucp ;
								ulp += n1 ;
								*destDataL++ = *ulp ;
							}
							destDataC += dest_bytes_per_line ;
						}
						XPutImage(gDisplay,dest->GetDrawable(),
							dest->GetGC(),destImage,0,0,dstRect->left,dstRect->top,destWidth,destHeight) ;
						break ;
				}
			}
			else if (m2 != 0) {
				switch(gImageDepth) {
					case 15:
					case 16:
						srcDataC = (unsigned char *)myImage->data ;
						destDataC = (unsigned char *)destImage->data ;
						destDataS = (unsigned short *)destImage->data ;
						ucp = srcDataC ;
						rowjump = m2*src_bytes_per_line ;
						for(row=0,m1=0;row<destHeight;row++,m1+=m2) {
							// row index into source image
							destDataS = (unsigned short *)destDataC ;
							for(col=0;col<destWidth;col++) {
								n1 = col*n_npr ;
								usp = (unsigned short *)ucp ;
								usp += n1 ;
								*destDataS++ = *usp ;
							}
							ucp	+= rowjump ;
							destDataC += dest_bytes_per_line ;
						}
						XPutImage(gDisplay,dest->GetDrawable(),
							dest->GetGC(),destImage,0,0,dstRect->left,dstRect->top,destWidth,destHeight) ;
						break ;
					case 24:
					case 32:
						srcDataC = (unsigned char *)myImage->data ;
						destDataC = (unsigned char *)destImage->data ;
						destDataL = (unsigned long *)destImage->data ;
						rowjump = m2*src_bytes_per_line ;
						for(row=0,m1=0;row<destHeight;row++,m1+=m2) {
							// row index into source image
							destDataL = (unsigned long *)destDataC ;
							for(col=0;col<destWidth;col++) {
								n1 = col*n_npr ;
								ulp = (unsigned long *)ucp ;
								ulp += n1 ;
								*destDataL++ = *ulp ;
							}
							ucp	+= rowjump ;
							destDataC += dest_bytes_per_line ;
						}
						XPutImage(gDisplay,dest->GetDrawable(),
							dest->GetGC(),destImage,0,0,dstRect->left,dstRect->top,destWidth,destHeight) ;
						break ;
				}
			}
			else {
				switch(gImageDepth) {
					case 15:
					case 16:
						srcDataC = (unsigned char *)myImage->data ;
						destDataC = (unsigned char *)destImage->data ;
						destDataS = (unsigned short *)destImage->data ;
						for(row=0;row<destHeight;row++) {
							// row index into source image
							m1 = (int)(row*m_mpr) ;
							destDataS = (unsigned short *)destDataC ;
							ucp = srcDataC + m1*src_bytes_per_line ;
							for(col=0;col<destWidth;col++) {
								n1 = col*n_npr ;
								usp = (unsigned short *)ucp ;
								usp += n1 ;
								*destDataS++ = *usp ;
							}
							destDataC += dest_bytes_per_line ;
						}
						XPutImage(gDisplay,dest->GetDrawable(),
							dest->GetGC(),destImage,0,0,dstRect->left,dstRect->top,destWidth,destHeight) ;
						break ;
					case 24:
					case 32:
						srcDataC = (unsigned char *)myImage->data ;
						destDataC = (unsigned char *)destImage->data ;
						destDataL = (unsigned long *)destImage->data ;
						for(row=0;row<destHeight;row++) {
							// row index into source image
							m1 = (int)(row*m_mpr) ;
							destDataL = (unsigned long *)destDataC ;
							ucp = srcDataC + m1*src_bytes_per_line ;
							for(col=0;col<destWidth;col++) {
								n1 = col*n_npr ;
								ulp = (unsigned long *)ucp ;
								ulp += n1 ;
								*destDataL++ = *ulp ;
							}
							destDataC += dest_bytes_per_line ;
						}
						XPutImage(gDisplay,dest->GetDrawable(),
							dest->GetGC(),destImage,0,0,dstRect->left,dstRect->top,destWidth,destHeight) ;
						break ;
				}
			}
		}
		if (myImage != nil) XDestroyImage(myImage) ;
		if (destImage != nil) XDestroyImage(destImage) ;	// This also deletes destData
	}
	XSetFunction(gDisplay,dest->GetGC(),GXcopy) ;
}

void SetOffscreenPort(OffScreenPtr portPtr){
    GetPort(&gSavedPort) ;
    SetPort((GrafPtr)portPtr) ;
}

void UnSetOffscreenPort(OffScreenPtr portPtr) {
    SetPort(gSavedPort) ;
}
