// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include "header.h"
#include "funcdefs.h"
#include "menudef.h"
#include "menu_index.h"
#include "filenames.h"
#include "params.h"
extern COMMAND par_atoms[],parameter[],par_symop[] ;
extern COMMAND *m_ch_atom[],*m_ch_main_par[],*m_ch_bas_at[],*m_ch_symop[] ;

#include "TWindow.h"
extern 	TWindow *gControlWindow ;

extern  Boolean	load_batch_structure_file(char *structname) ;

Boolean	load_batch_structure_file(char *structname) {
	
	long i ;
	Boolean stat ;
	FILE *fp ;
	
	// Set up all the filenames to be used in the program
	
	strcpy(flname.pin,structname) ; strcat(flname.pin,".at") ;
	strcpy(flname.pout,structname) ; strcat(flname.pout,".pout") ;
	strcpy(flname.mout,structname) ; strcat(flname.mout,".mout") ;
	strcpy(flname.aout,structname) ; strcat(flname.aout,".aout") ;
	strcpy(flname.iout,structname) ; strcat(flname.iout,".iout") ;
	strcpy(flname.pprint,structname) ; strcat(flname.pprint,".p_prnt") ;
	strcpy(flname.mprint,structname) ; strcat(flname.mprint,".m_prnt") ;
	
	fp = fopen(flname.pin,"r") ;
	stat = read_file(fp) ;
	fclose(fp) ;
	
	return stat ;
}

// -----------------------------------------------------------------
// Display a single structure defined by the ATOM file             
// and reset various parameters upon start of a new structure     
// -----------------------------------------------------------------

void r_sngle_structure(char *structname,char *status) {

	long i,stat ;
	FILE *fp ;

	gNumInputParameters = NUM_INPUT_PARAMETERS ;  /* set the no. of parameters */
	                   /* This is the number of parameters in the */
	                   /* main menu                               */

	/* Set up all the filenames to be used in the program */
	/* ..but first initaialize the names                  */
	for (i=0;i<NAMELEN;i++) {
		flname.pin[i] = ' ' ;
		flname.pout[i] = ' ' ;
		flname.mout[i] = ' ' ;
		flname.aout[i] = ' ' ;
		flname.iout[i] = ' ' ;
		flname.pprint[i] = ' ' ;
		flname.mprint[i] = ' ' ;
	}
	
	strcpy(flname.pin,structname) ; strcat(flname.pin,".at") ;
	strcpy(flname.pout,structname) ; strcat(flname.pout,".pout") ;
	strcpy(flname.mout,structname) ; strcat(flname.mout,".mout") ;
	strcpy(flname.aout,structname) ; strcat(flname.aout,".aout") ;
	strcpy(flname.iout,structname) ; strcat(flname.iout,".iout") ;
	strcpy(flname.pprint,structname) ; strcat(flname.pprint,".p_prnt") ;
	strcpy(flname.mprint,structname) ; strcat(flname.mprint,".m_prnt") ;

	/* this screen makes all the input data active such that changes */
	/* can be performed */
	for (i=1;i<=gNumInputParameters;i++) m_ch_main_par[i] = &parameter[i-1] ;
	m_ch_main_par[i] = &null ;     /* terminate with the null structure */

	/* This makes the atom-parameters active for change  */
	for (i=1;i<=120;i++) m_ch_atom[i] = &par_atoms[i-1] ;
	
	m_ch_atom[i] = &null ;     /* terminate with the null structure */
	for (i=1;i<=120;i++) m_ch_bas_at[i] = &par_atoms[i-1] ;
	
	m_ch_bas_at[i] = &null ;     /* terminate with the null structure */
	for (i=1;i<=40;i++) m_ch_symop[i] = &par_symop[i-1] ;
	
	m_ch_symop[i] = &null ;


	if (strcmp(status,"new")==0) {
		CLEAR_DISPLAY ;
		create_file() ;
		CLEAR_DISPLAY ;
	}
	else  {       /* Old input file */
		fp = fopen(flname.pin,"r") ;
		write_message("Reading the Structure File.......") ;
		stat = read_file(fp) ;
		fclose(fp) ;
		if (!stat) {
			BELL ;
			return ;
		}
		CLEAR_DISPLAY ;
	}
	set_menu(MAIN_SNGL_STR) ;
	displ_var() ;             /* display the parameters */
	displ_menu() ;            /* display the controlmenu */

	char text[256] ; strcpy(text,"NCEMSS - ") ; strcat(text,structname) ;
	gControlWindow->SetWindowTitle(text) ;

	ViewNewUnitCell() ;
}

//-----------------------------------------------------------------
// Display a layered structure defined by the ATOM file            
// and reset various parameters upon start of a new structure     
//-----------------------------------------------------------------
void r_lay_structure(char *structname,char *status) {

	long i,stat ;
	FILE *fp ;

	gNumInputParameters = NUM_INPUT_PARAMETERS ; /* set the no. of parameters */
                               /* This is the number of parameters in the */
                               /* main menu                               */


	/*                ***************                             */
	/* The proper names for the input to the multislice program   */
	/* is the sequence of output files from single-structure      */
	/* calculations previously performed                          */
	/*                ***************                             */


	/* Set up all the filenames to be used in the program */
	/* ..but first initaialize the names                  */
	for (i=0;i<NAMELEN;i++) {
		flname.pin[i] = ' ' ;
		flname.pout[i] = ' ' ;
		flname.mout[i] = ' ' ;
		flname.aout[i] = ' ' ;
		flname.iout[i] = ' ' ;
		flname.pprint[i] = ' ' ;
		flname.mprint[i] = ' ' ;
	}
	strcpy(flname.pin,structname) ; strcat(flname.pin,".lay") ;
	strcpy(flname.mout,structname) ; strcat(flname.mout,".mout") ;
	strcpy(flname.aout,structname) ; strcat(flname.aout,".aout") ;
	strcpy(flname.iout,structname) ; strcat(flname.iout,".iout") ;
	strcpy(flname.mprint,structname) ; strcat(flname.mprint,".m_prnt") ;
  
	/* this screen makes all the input data active such that changes */
	/* can be performed */
	for (i=1;i<=gNumInputParameters;i++) m_ch_main_par[i] = &parameter[i-1] ;
	m_ch_main_par[i] = &null ;     /* terminate with the null structure */

  
	if (strcmp(status,"new")==0) {
		CLEAR_DISPLAY ;
		create_file() ;
		CLEAR_DISPLAY ;
	}
	else  {       /* Old input file */
		fp = fopen(flname.pin,"r") ;
		write_message("Reading the Structure File.......") ;
		stat = read_file(fp) ;
		fclose(fp) ;
		if (!stat) {
			BELL ;
			return ;
		}
		CLEAR_DISPLAY ;
	}
	set_menu(MAIN_LAY_STR) ;
	displ_var() ;         /* display the parameters */
	displ_menu() ;        /* display the controlmenu */

	char text[256] ; strcpy(text,"NCEMSS - ") ; strcat(text,structname) ;
	gControlWindow->SetWindowTitle(text) ;
}

