// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "header.h"
#include "globals.h"
#include "funcdefs.h"
#include "menudef.h"
#include "menu_index.h"
#include "MslicRecord.h"
#include "layers.h"
#include "params.h"
#include "globals.h"
extern MENUS gMenuList[] ;
extern COMMAND *m_set_pg_seq[],*m_rem_pg_lay[],pgseq[],pg[] ;
extern char stext[] ;
extern long gCalculationIsChanged,gMSisChanged ;
/* PSECT for different phasegratings to be used in LAYERED structure */
struct  LayerRec {
           short  num      ;   /* Number of different phasegratings MAX=8 */
           float     delz     ;   /* Slice-thickness for each phasegrating   */
           char      name[32] ;   /* Name of the individual strucures        */
        } layer[MAXLAYERS] ;

long NUM_LAYERS = 0 ;         /* Number of layers defined                     */
long NUM_SLICES = 0 ;         /* Total number of layers to make up the spec.  */
long PG_SEQ[MAXSLICES] ;      /* The sequence of PG's to be used in the calc. */
long PG_NUM = 0 ;             /* Total number of slices                       */
float TOT_THICKNESS = 0. ;   /* Total thickness defined with layers          */
float PG_THICK[MAXSLICES] ;
static long gCurrPosX,gCurrPosY ;
long DEPOSIT = FALSE ;
long INSERT  = FALSE ;
long gREMOVE  = FALSE ;        /* Various flags to be used in LAYERS           */
long REPEAT  = FALSE ;
long RESET   = FALSE ;
long TOP     = 550 ;
long BOTTOM  = 100 ;
long COLUMN  = 450 ;
char *null_text = "" ;
void displ_pick() ;

// --------------------------------------------------------------------------
// Program handles a list of layers  
// Title stands for view composition
// --------------------------------------------------------------------------
void r_v_comp() {
	short  i ;

	set_menu(LAY_PG) ;
	NUM_LAYERS = gLayerRec.num_pg ;
	PG_NUM = gLayerRec.num_slc ;
	for (i=0; i< NUM_LAYERS ;i++) {
		layer[i].num = i + 1 ;
		layer[i].delz = gLayerRec.slcth[i] ;
		strcpy(layer[i].name,gLayerRec.pg[i]) ;
		pg[i].xl = 220 + ((long)(i/8))*120 ;
		pg[i].yl = 560 - 30*(i%8) ;
		pg[i].xh = pg[i].xl+20;
		pg[i].yh = pg[i].yl + 20 ;
		pg[i].action = mod_pg ;
		pg[i].txt = null_text ;
		pg[i].help_msg = null_text ;
		m_set_pg_seq[5+i] = &pg[i] ;
	}      
	m_set_pg_seq[5+NUM_LAYERS] = &null ;

	/* Set initial sequence of the structure */
	for (i=0;i<PG_NUM;i++) {
		PG_SEQ[i] = gLayerRec.phgr_seq[i] ; 
	}
	CLEAR_DISPLAY ;
	temdtx(100,760,"***** Construct Phase-grating sequence *****") ;
	temdln(0,750,1000,750) ;
	/* temdtx(10,720," LAYER#   NAME    DZ(A)   LAYER#   NAME    DZ(A)") ;*/
	displ_lay() ;
	temdln(0,590,1000,590) ;
	temdln(0,50,1000,50) ;
	displ_menu() ;
	displ_pick() ;
	displ_spec() ;
	temdtx(220,280,"PHASE-GRATING SEQUENCE :") ;
}

void SetSequence() {
	long i,deltaX ;

	TOT_THICKNESS = 0. ;
	gCurrPosX = 224 ;
	gCurrPosY = 250 ;
	for (i=0; i< PG_NUM;i++) {
		TOT_THICKNESS += layer[PG_SEQ[i]-1].delz ;
		PG_THICK[i] = TOT_THICKNESS ;
		sprintf(stext,"%d ",PG_SEQ[i]) ;
		deltaX = 2*StringWidth(stext) ;
		pgseq[i].xl = gCurrPosX ;		/*	224 + 27*(i%35) ; */
		pgseq[i].xh = pgseq[i].xl + deltaX ;
		pgseq[i].yl = gCurrPosY ;
		pgseq[i].yh = pgseq[i].yl + 20 ;
		pgseq[i].action = rem_pg ;
		pgseq[i].txt = null_text ;
		pgseq[i].help_msg = null_text ;
		gCurrPosX += deltaX ;
		if (gCurrPosX > 820) {
			gCurrPosX = 224 ;
			gCurrPosY -= 30 ;
		}
		m_rem_pg_lay[i+1] = &pgseq[i] ;
	}
	m_rem_pg_lay[PG_NUM+1] = &null ;
}

// --------------------------------------------------------------------------
// Deposit layers at the bottom of the specimen      
// --------------------------------------------------------------------------
void r_deposit() {

	if (gCurrentMenu == &gMenuList[LAY_PG]) {
		gCurrentCommand = gCurrentButton ;
		on_highlight() ;
		change_cursor() ;
		set_menu(SET_PG_SEQ) ;
		gCalculationIsChanged = TRUE ;
		gMSisChanged = TRUE ;
		DEPOSIT = TRUE ;
	}
	else {
		if (!DEPOSIT) {
			off_highlight() ;
			gCurrentCommand = gCurrentButton ;
			on_highlight() ;
			change_cursor() ;
			INSERT = FALSE ;
			gREMOVE = FALSE ;
			REPEAT = FALSE ;
			RESET  = FALSE ;
			DEPOSIT = TRUE ;
			gCalculationIsChanged = TRUE ;
			gMSisChanged = TRUE ;
		}
		else {
			set_menu(LAY_PG) ;
			off_highlight() ;
			reset_cursor() ;
			DEPOSIT = FALSE ;
		}
	}
}

// --------------------------------------------------------------------------
// Insert a layer into the phasegrating sequence  
// --------------------------------------------------------------------------
void r_insert() {

	if (gCurrentMenu == &gMenuList[LAY_PG]) {
		gCurrentCommand = gCurrentButton ;
		on_highlight() ;
		change_cursor() ;
		set_menu(SET_PG_SEQ) ;
		gCalculationIsChanged = TRUE ;
		gMSisChanged = TRUE ;
		INSERT = TRUE ;
	}
	else {
		if (!INSERT) {
			off_highlight() ;
			gCurrentCommand = gCurrentButton ;
			on_highlight() ;
			change_cursor() ;
			DEPOSIT = FALSE ;
			gREMOVE = FALSE ;
			REPEAT = FALSE ;
			RESET  = FALSE ;
			INSERT = TRUE ;
			gCalculationIsChanged = TRUE ;
			gMSisChanged = TRUE ;
		}
		else {
			reset_cursor() ;
			set_menu(LAY_PG) ;
			off_highlight() ;
			INSERT = FALSE ;
		}
	}
}

// --------------------------------------------------------------------------
// Repeat current Phasegrating sequence N times   
// --------------------------------------------------------------------------
void r_repeat() {

	if (gCurrentMenu == &gMenuList[LAY_PG]) {
		gCurrentCommand = gCurrentButton ;
		on_highlight() ;
		REPEAT = TRUE ;
		do_repeat() ;
		off_highlight() ;
		gCalculationIsChanged = TRUE ;
		gMSisChanged = TRUE ;
		displ_spec() ;
		REPEAT = FALSE ;
	}
	else {
		off_highlight() ;
		gCurrentCommand = gCurrentButton ;
		on_highlight() ;
		INSERT = FALSE ;
		gREMOVE = FALSE ;
		DEPOSIT = FALSE ;
		RESET  = FALSE ;
		REPEAT = FALSE ;
		do_repeat() ;
		gCalculationIsChanged = TRUE ;
		gMSisChanged = TRUE ;
		off_highlight() ;
		set_menu(LAY_PG) ;
		displ_spec() ;
	}
}

// --------------------------------------------------------------------------
// Resets the entire specimen to zero thickness  
// --------------------------------------------------------------------------
void r_reset() {

	if (gCurrentMenu == &gMenuList[LAY_PG]) {
		gCurrentCommand = gCurrentButton ;
		on_highlight() ;
		do_reset() ;
		gCalculationIsChanged = TRUE ;
		gMSisChanged = TRUE ;
		off_highlight() ;
		RESET = FALSE ;        
		displ_spec() ;
	}
	else {
		off_highlight() ;
		gCurrentCommand = gCurrentButton ;
		on_highlight() ;
		INSERT = FALSE ;
		gREMOVE = FALSE ;
		DEPOSIT = FALSE ;
		REPEAT = FALSE ;
		RESET = FALSE ;
		do_reset() ;
		off_highlight() ;
		set_menu(LAY_PG) ;
		displ_spec() ;
		gCalculationIsChanged = TRUE ;
		gMSisChanged = TRUE ;
	}
}

// --------------------------------------------------------------------------
// Define a unique layer to be used         
// --------------------------------------------------------------------------
void r_def_layer() {
  
	gCurrentCommand = gCurrentButton ;
	on_highlight() ;
	temdtx(0,10," Enter the name for the layer -> ") ;
	ReadString(layer[NUM_LAYERS].name,32) ; 
	eraser(0,0,800,40) ;
	temdtx(0,10," Enter unit cell thickness -> ") ;
	layer[NUM_LAYERS].delz = get_float(3.) ;
	eraser(0,0,800,40) ;
	off_highlight() ;
	layer[NUM_LAYERS].num = NUM_LAYERS+1 ;
	pg[NUM_LAYERS].xl = 220 + ((long)(NUM_LAYERS/8))*120 ;
	pg[NUM_LAYERS].yl = 560 - 30*(NUM_LAYERS%8) ;
	pg[NUM_LAYERS].xh = pg[NUM_LAYERS].xl+20 ;
	pg[NUM_LAYERS].yh = pg[NUM_LAYERS].yl + 20 ;
	pg[NUM_LAYERS].action = mod_pg ;
	pg[NUM_LAYERS].txt = null_text ;
	pg[NUM_LAYERS].help_msg = null_text ;
	m_set_pg_seq[5+NUM_LAYERS] = &pg[NUM_LAYERS] ;
	m_set_pg_seq[6+NUM_LAYERS] = &null ;
	NUM_LAYERS++ ;
	displ_lay() ;
	eraser(220,310,840,580) ;
	displ_pick() ;
	gCalculationIsChanged = TRUE ;
	gMSisChanged = TRUE ;
}

// --------------------------------------------------------------------------
// Deletes an already defined layer  
// --------------------------------------------------------------------------
void r_del_layer() {
	long num,i ;

	gCurrentCommand = gCurrentButton ;
	on_highlight() ;
	eraser(0,0,800,40) ;
	temdtx(0,10," Enter the number of layer -> ") ;
	num = get_int(1) ;
	eraser(0,0,800,40) ;
	off_highlight() ;
	NUM_LAYERS-- ;
	for (i=num-1;i<NUM_LAYERS;i++) {
		layer[i].num = i+1 ;
		layer[i].delz = layer[i+1].delz ;
		strcpy(layer[i].name,layer[i+1].name) ;
	}
	m_set_pg_seq[5+NUM_LAYERS] = &null ;
	eraser(0,595,800,745) ;
	displ_lay() ;
	eraser(220,310,840,580) ;
	displ_pick() ;
	gCalculationIsChanged = TRUE ;
	gMSisChanged = TRUE ;
}

void displ_lay() {
	short  i ;

	for (i=0;i<(NUM_LAYERS > 20 ? 20 : NUM_LAYERS);i++) {
		sprintf(stext,"%d %5s %5.2f",
			layer[i].num,layer[i].name,layer[i].delz) ;
		temdtx(20+(long)(i/5)*180,720-(i%5)*30,stext) ; 
	}
}

void displ_pick() {
    short  i ;
	extern uLong gPattern[] ;

	for (i=0;i<(NUM_LAYERS > 32 ? 32 : NUM_LAYERS);i++) {
		DrawControl(&pg[i]) ;
		SetFgColor(gLut[gPattern[i]-gNumGrays]) ;
		temdtx(pg[i].xh+10,pg[i].yl,layer[i].name) ;
		SetFgColor(gBlackPixel) ;
	}
}

// --------------------------------------------------------------------------
// Displayes the content of the specimen as a vertical 
// bar on the terminal                 
// --------------------------------------------------------------------------
void displ_spec() {
	short  i,j,ix=20,iy=250,yb,yt,last_th ;
	float beg_thick,length ;
	long deltaX ;
	char stemp[80] ;


	SetSequence() ;
	eraser(10,BOTTOM-30,215,TOP+25) ;
	temdtx(ix,TOP+5,"TOP - 0.00 A") ;
	last_th = TOP+5 ;
	if (PG_NUM == 0) {
		temdln(ix,TOP,ix+30,TOP) ;
		sprintf(stext,"BOTTOM") ;
		sprintf(stemp,"- %6.2f A",TOT_THICKNESS) ;
		strcat(stext,stemp) ;
		temdtx(ix,TOP-25,stext) ; 
	}
	else {
		j=0 ;
		beg_thick = 0. ;
		yt = TOP ;
		do {
			for (i=j;i<PG_NUM;i++) {
				/* Traverse until a compositional change or until the end */
				if (i==PG_NUM-1 || PG_SEQ[i+1] != PG_SEQ[i]) {
					length = (PG_THICK[i]-beg_thick)/TOT_THICKNESS ;
					yb = yt - length*COLUMN ;
					yb = MAX(yb,BOTTOM) ;
					pattern(ix,yb,ix+30,yt,PG_SEQ[i]) ;
					if ((last_th-yb) > 25) {
						temdln(ix+30,yb,ix+36,yb) ;
						sprintf(stext,"%6.2f",PG_THICK[i]) ;
						temdtx(ix+40,yb,stext) ; 
						last_th = yb ;
					}
					j = i+1 ;
					beg_thick = PG_THICK[i] ;
					yt = yb ;
					break ;
				}
			}
		} while (j < PG_NUM) ;
		
		sprintf(stext,"BOTTOM") ;
		sprintf(stemp,"- %6.2f A",TOT_THICKNESS) ;
		strcat(stext,stemp) ;
		temdtx(ix,BOTTOM-25,stext) ; 
	}
	eraser(220,60,1000,270) ;
	if (PG_NUM != 0) {
		/* Print out the phasegrating sequence */
		i = 0 ;
		for(i=0;i<PG_NUM;i++) {	
			sprintf(stext,"%d",PG_SEQ[i]) ;
			temdtx(pgseq[i].xl+1,pgseq[i].yl,stext) ;
		}          
	}
}

// --------------------------------------------------------------------------
// Returns control to main menu and sets necassary data 
// --------------------------------------------------------------------------
void r_ret_lay() {
	long i ;

	if (DEPOSIT || gREMOVE || INSERT || RESET || REPEAT) 
		return ;  /* All selections should be turned off first */
	/* Transfer the necessary data back in common */
	gLayerRec.num_pg = NUM_LAYERS ;
	gLayerRec.num_slc = PG_NUM ;
	
	for (i=0 ; i< NUM_LAYERS;i++) {
		gLayerRec.slcth[i] = layer[i].delz ;
		strcpy(gLayerRec.pg[i],layer[i].name) ;
	}
	for (i=0; i< PG_NUM; i++) {
		gLayerRec.phgr_seq[i] = PG_SEQ[i] ;
	}
	CLEAR_DISPLAY ;
	set_menu(MAIN_LAY_STR) ;
	gInputParameter[21].fv = TOT_THICKNESS ;
	displ_var() ;
	displ_menu() ;
	if (gMsliceRec.tts && TOT_THICKNESS < gMsliceRec.t2) {
		sprintf(stext,
			"* WARNING * ; Thickness TOO large, specimen undefined at %.0f A",
		gMsliceRec.t2) ;
		REVERSE_VIDEO ;
		temdtx(0,130,stext) ;
		NORMAL_VIDEO ;
		BELL ;
	}
}

// --------------------------------------------------------------------------
// Routine to deal with adding,removing and inserting
// layers into the phasegrating sequence  
// --------------------------------------------------------------------------
void mod_pg(long num) {
	short  num_slices,i,j,new_pg_num ;
	float thick,del_thick ;
	long deltaX ;

	num -= 5 ;  
	on_off(&pg[num]) ;
	if (DEPOSIT) {
		eraser(0,0,800,40) ;
		temdtx(0,10," Deposit what thickness (A) -> ") ;
		thick = get_float(3.) ;
		eraser(0,0,800,40) ;
		num_slices = thick/layer[num].delz ;
		new_pg_num = PG_NUM + num_slices ;
		if (new_pg_num > MAXSLICES) { /* Hardcoded upper limit on 	*/
			/* the number of pg's 		*/
			eraser(0,0,800,40) ;
			REVERSE_VIDEO ;
			temdtx(0,10,
				"The number of slices exceeds maximum allowed - 1000") ;
			NORMAL_VIDEO ;
			BELL ;
			on_off(&pg[num]) ;
			return ;
		}
		for (i=0;i<num_slices;i++) {
			PG_SEQ[PG_NUM] = num+1 ;
			TOT_THICKNESS += layer[num].delz ;
			PG_THICK[PG_NUM] = TOT_THICKNESS ;
			PG_NUM++ ;
		}
	}
	else if (INSERT) {
		eraser(0,0,800,40) ;
		temdtx(0,10," Insert at what thickness -> ") ;
		thick = get_float(3.) ;
		eraser(0,0,800,40) ;
		temdtx(0,10," Insert how much (A) -> ") ;
		del_thick = get_float(3.) ;
		num_slices = del_thick/layer[num].delz ;      
		new_pg_num = PG_NUM + num_slices ;
		if (new_pg_num > MAXSLICES) { /* Hardcoded upper limit */
			eraser(0,0,800,40) ;
			REVERSE_VIDEO ;
			temdtx(0,10,
			"The number of slices exceeds maximum allowed - 1000") ;
			NORMAL_VIDEO ;
			BELL ;
			on_off(&pg[num]) ;
			return ;
		}
		del_thick = num_slices*layer[num].delz ;
		eraser(0,0,800,40) ;
		/* Find the location to insert the material */
		for (i=0; PG_THICK[i] <= thick; i++)
			;
		/* Push the list down to open room for the insertion */ 
		for (j=PG_NUM+num_slices; j >= i+num_slices; ) {
			j-- ;
			PG_SEQ[j] = PG_SEQ[j-num_slices] ;
			PG_THICK[j] = PG_THICK[j-num_slices] + del_thick ;
		}
		/* Insert the material */
		for (j=0; j < num_slices; j++) {
			PG_SEQ[i+j] = num+1 ;
			PG_THICK[i+j] = PG_THICK[i+j-1]+layer[num].delz ;
		}
		PG_NUM += num_slices ;
		TOT_THICKNESS += del_thick ;
	}      
	on_off(&pg[num]) ;
	displ_spec() ;      
}

// --------------------------------------------------------------------------
// routine to remove a single layer from the pg_seq. 
// --------------------------------------------------------------------------
void rem_pg(long i) {
	long j ;

	i-- ;
	/* We are going to remove the i'th entry from the list     */
	/* and update everytime. However, the menu is still active */
	/* until specifically turned off                           */

	/* Shift the list down */
	PG_NUM-- ;
	TOT_THICKNESS -= layer[PG_SEQ[i]-1].delz ;
	for (j=i; j < PG_NUM; j++) {
		PG_SEQ[j] = PG_SEQ[j+1] ;
		PG_THICK[j] = PG_THICK[j+1]-layer[PG_SEQ[i]-1].delz ;
	}
	m_rem_pg_lay[PG_NUM+1] = &null ;
	displ_spec() ;
}

// --------------------------------------------------------------------------
// Does the actual repeating of current sequence num_repeat times
// --------------------------------------------------------------------------
void do_repeat() {
	long i,num_repeat,j,old_pg_num,new_pg_num ;
	long deltaX ;

	eraser(0,0,800,40) ;
	temdtx(0,10," Repeat current sequence how many times -> ") ;
	num_repeat = get_int(1) ;
	eraser(0,0,800,40) ;
	old_pg_num = PG_NUM ;
	new_pg_num = old_pg_num*num_repeat ;
	if (new_pg_num > MAXSLICES) { /* Hardcoded limit on the number of pg's */
		eraser(0,0,800,40) ;
		REVERSE_VIDEO ;
		temdtx(0,10,"The number of slices exceeds maximum allowed - 1000") ;
		NORMAL_VIDEO ;
		BELL ;
		return ;
	}
	PG_NUM = new_pg_num ;
	TOT_THICKNESS = 0. ;
	for (i=0; i< PG_NUM;i++) {
		j = i%old_pg_num ;
		PG_SEQ[i] = PG_SEQ[j] ;
		TOT_THICKNESS += layer[PG_SEQ[i]-1].delz ;
		PG_THICK[i] = TOT_THICKNESS ;
	}
}

// --------------------------------------------------------------------------
// Actually sets the specimen to zero thickness 
// --------------------------------------------------------------------------
void do_reset() {
   
	PG_NUM = 0 ;
	TOT_THICKNESS = 0. ;
	m_rem_pg_lay[1] = &null ;
}

