// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include "funcdefs.h"
#include "header.h"
#include "menudef.h"
#include "params.h"
#include "layers.h"
#include "PhasegratingRecord.h"
#include "MslicRecord.h"
#include "displ.h"
#include "symops.h"
#include "atoms.h"
#include "dfpdspl.h"

#include "TAtomRefinementPage.h"
extern	TAtomRefinementPage *gAtomRefinementPage ;

extern COMMAND c_set_ull_slice,c_set_pseudo_slice ;
extern short gSliceThreeD ;
extern long gNumSymops,gNumBasisAtoms,gNumUnitCellAtoms,LAYERED ;
extern long gSpacegroupNumber,gSymopsAreChanged ;
extern long NUM_THICK,NUM_DEFOC ;
extern char title[],micname[],plots[],name[],*chemsymb[] ;
extern struct	{ long status, oldstatus ; } calculation ;
extern long gCalculationIsChanged,gPGisChanged,gBasisIsChanged ;
void	ResetActiveElements() ;
void	UpdateCalculationStatus() ;

#define	EOS		'\0'
#define NEWLINE	'\n'
#define	CARRIAGE_RETURN	'\r'
#define	BLANK	' '

typedef struct	ElementRecord {
	short number ;
	short index[20] ;
} ElementRecord,*ElementPtr ;
ElementPtr MyElements = nil ;

// ------------------------------------------------------
// Read a single line from the file
// ------------------------------------------------------
char *ReadLine(char *buffer,long numChars,FILE *fp) {
	char c ;
	int i ;
	
	for(i=0;i<numChars;) {
		c = fgetc(fp) ;
		if (c == EOF) {
			if (i == 0) {
				return NULL ;
			}
			buffer[i] = '\0' ;
			return buffer ;
		}	
		if (c == NEWLINE || c == CARRIAGE_RETURN) {
			if (i == 0) {
				// Skip this as a valid line
			}
			else {
				buffer[i] = '\0' ;
				return buffer ;
			}
		}
		else {
			if (c == ' ' && i == 0) {
				// Gobble up empty spaces at the beginning of the line
			}
			else {
				buffer[i++] = c ;
			}
		}
	}
	buffer[numChars-1] = '\0' ;
	return buffer ;
}


// ---------------------------------------------------------------------
// Function to read the input file --.at containing all input 
//                                 --.lay if LAYERED       
//  parameters for the calculation                          
// ---------------------------------------------------------------------
long read_file(FILE *fp) {

	long i,maxPossibleAtoms,iz,j,n,k,inword,numVariablesAssigned ;
	char c,buffer[256],text[256] ;
	float ftemp ;
	

	i = 0 ;
	/* read in the title */
	if (ReadLine(title,40,fp) == NULL) {
		write_message("Error Reading the title") ;
		return false ;
	}
	gInputParameter[0].cv=title ;

	/* read in the space group number */
	if (ReadLine(buffer,256,fp) == NULL) {
		write_message("Error Reading the space group number") ;
		return false ;
	}
	sscanf(buffer,"%d",&gSpacegroupNumber) ;
	gInputParameter[1].iv = gSpacegroupNumber ;

	/* read in a,b,c,alpha,beta,gamma */
	if (ReadLine(buffer,256,fp) == NULL) {
		write_message("Error Reading the lattice parameters") ;
		return false ;
	}
	sscanf(buffer,"%f %f %f %f %f %f",&gInputParameter[2].fv,&gInputParameter[3].fv,&gInputParameter[4].fv,
		&gInputParameter[5].fv,&gInputParameter[6].fv,&gInputParameter[7].fv) ;

	/* read in gmax */
	if (ReadLine(buffer,256,fp) == NULL) {
		write_message("Error Reading gmax") ;
		return false ;
	}
	sscanf(buffer,"%f",&gInputParameter[8].fv) ;
		
	/* read in the incoming beam direction */
	if (ReadLine(buffer,256,fp) == NULL) {
		return false ;
	}
	sscanf(buffer,"%d %d %d",&gPhsgrtRec->iuvw[0],&gPhsgrtRec->iuvw[1],&gPhsgrtRec->iuvw[2]) ;

	gSliceThreeD = FALSE ;
	c_set_ull_slice.status = OFF ;
	c_set_pseudo_slice.status = ON ;
	if (!LAYERED) {
		/* gNumSymops, nz */
		if (ReadLine(buffer,256,fp) == NULL) {
			write_message("Error Reading number of symmetry operators") ;
			return false ;
		}
		gSliceThreeD = false ;
		numVariablesAssigned = sscanf(buffer,"%d %d %d",&gInputParameter[10].iv,&gInputParameter[11].iv,&gSliceThreeD) ;  
		gNumSymops = gInputParameter[10].iv ;
		if (gInputParameter[11].iv != 1 && numVariablesAssigned == 3) {
			if (gSliceThreeD) {
				c_set_ull_slice.status = ON ;
				c_set_pseudo_slice.status = OFF ;
			}
			else {
				c_set_ull_slice.status = OFF ;
				c_set_pseudo_slice.status = ON ;
			}
		}
		else gSliceThreeD = false ;

		/* natom,ntype */
		if (ReadLine(buffer,256,fp) == NULL) {
			write_message("Error Reading number of atoms and types of atoms") ;
			return false ;
		}
		sscanf(buffer,"%d %d",&gInputParameter[12].iv,&gInputParameter[13].iv) ;
		gNumBasisAtoms = gInputParameter[12].iv ;
		maxPossibleAtoms = gNumBasisAtoms*gNumSymops ;
		if (maxPossibleAtoms > gMaxAtoms) {
			sprintf(text,"The Number of possible atoms in the structure (%d = %d*%d) exceeds Limit (%d) Use 'ncemss -a %d' instead",
			maxPossibleAtoms,gNumBasisAtoms,gNumSymops,gMaxAtoms,maxPossibleAtoms) ;
			write_message(text) ;
			return 0 ;
		}
		for (n=0;n<gNumBasisAtoms;n++)    {
			if (ReadLine(buffer,256,fp) == NULL) {
				write_message("Error Reading the atom information") ;
				return false ;
			}
			if ((numVariablesAssigned=sscanf(buffer,"%d %s %f %f %f %f %f",
						&(basis+n)->type.val.iv,name,
						&(basis+n)->xc.val.fv,&(basis+n)->yc.val.fv,&(basis+n)->zc.val.fv,
						&(basis+n)->dw.val.fv,&(basis+n)->occ.val.fv)) < 5) {	
				fprintf(stderr,"Error Reading the atom information for %d atoms\n",gNumBasisAtoms) ; fflush(stderr) ;
				write_message("Error Reading the atom information") ;
				return false ;
			}
			(basis+n)->no.val.iv = (iz=atomno(name)) ;         /* atomic number   */  
			(basis+n)->name.val.cv = chemsymb[iz-1] ;          /* chemical symbol */
		}
		sort(basis,gNumBasisAtoms) ;
	}
	else {
		if (ReadLine(buffer,256,fp) == NULL) {
			write_message("Error Reading the layered information") ;
			return false ;
		}
		sscanf(buffer,"%d %d",&gLayerRec.num_pg,&gLayerRec.num_slc) ;
		for (n=0; n< gLayerRec.num_pg;n++) {
			if (ReadLine(buffer,256,fp) == NULL) return false ;
			sscanf(buffer,"%f %s",&gLayerRec.slcth[n],&gLayerRec.pg[n][0]) ;
		}
		for (n=0; n<  gLayerRec.num_slc;n++ ) {
			if (fscanf(fp,"%d",&gLayerRec.phgr_seq[n]) == EOF) {
				return false ;
			}
		}
		ReadLine(buffer,80,fp) ; // just to get past the end of line after the number of slices
		i = 14 ; /* set i to point to the correct parameter if LAYERED */
	}


	if (ReadLine(buffer,256,fp) == NULL) {
		write_message("Error Reading the microscope information") ;
		return false ;
	}

	gDisplayRec.TwoFoldAstigMag = 0. ;
	gDisplayRec.TwoFoldAstigAngle = 0. ;
	gDisplayRec.ThreeFoldAstigMag = 0. ;
	gDisplayRec.ThreeFoldAstigAngle = 0. ;

	if ((numVariablesAssigned=sscanf(buffer,"%s %f %f %f %f %f %f %f",
			micname,&gInputParameter[15].fv,&gInputParameter[16].fv,&gInputParameter[17].fv,
			&gDisplayRec.TwoFoldAstigMag,&gDisplayRec.TwoFoldAstigAngle,
			&gDisplayRec.ThreeFoldAstigMag,&gDisplayRec.ThreeFoldAstigAngle)) < 4) {
		
		return false ;
	}
	gInputParameter[14].cv = micname ;                          /* microscope name */

	if (ReadLine(buffer,256,fp) == NULL) {
		write_message("Error Reading the voltage information") ;
		return false ;
	}
	sscanf(buffer,"%f",&gInputParameter[18].fv) ;  /* Voltage  Lh,Lk  */

	if (ReadLine(buffer,256,fp) == NULL) {
		write_message("Error Reading the Lh,Lk information") ;
		return false ;
	}
	sscanf(buffer,"%f %f",&gInputParameter[19].fv,&gInputParameter[20].fv) ;  /* Voltage  Lh,Lk  */

	dfpdspl.theta = gInputParameter[17].fv ;
	/* Get the thickness and possibly a thickness series */
	gMsliceRec.tts = FALSE ;
	if (ReadLine(buffer,256,fp) == NULL) {
		write_message("Error Reading the thickness information") ;
		return false ;
	}
	if (strchr(buffer,',') != NULL) {
		sscanf(buffer,"%f,%f,%f",&gMsliceRec.t1,&gMsliceRec.t2,&gMsliceRec.dt) ;
		gMsliceRec.tts = TRUE ;
		NUM_THICK = (gMsliceRec.t2-gMsliceRec.t1)/gMsliceRec.dt + 1 ;
	}
	else {
		sscanf(buffer,"%f",&gMsliceRec.t1) ;
		gMsliceRec.t2 = gMsliceRec.t1 ;
		gMsliceRec.dt = 0. ;
		NUM_THICK = 1 ;
	}
	gInputParameter[21].fv = gMsliceRec.t2 ;
	i = 22 ;
	gMsliceRec.iplot = FALSE ;
	if (ReadLine(plots,256,fp) == NULL) {
		write_message("Error Reading the pendellossung information") ;
		return false ;
	}
	gInputParameter[i++].cv = plots ;
	if (plots[0] == 'y' || plots[0] == 'Y') {
		gMsliceRec.iplot = TRUE ;
		if (ReadLine(buffer,256,fp) == NULL) {
			write_message("Error reading the Pendellosung Option Indices") ;
			fclose(fp) ;
			return false ;
		}
		if ((numVariablesAssigned = sscanf(buffer,"%d",&gMsliceRec.nw)) != 1) {
			write_message("Error reading the Pendellosung Option Indices") ;
			fclose(fp) ;
			return false ;
		}
		for (k=0;k<gMsliceRec.nw;k++) {
			if (ReadLine(buffer,256,fp) == NULL) {
				write_message("Error reading the Pendellosung Option Indices") ;
				fclose(fp) ;
				return false ;
			}
          	if ((numVariablesAssigned = sscanf(buffer,"%d %d %d",&gMsliceRec.ihw[k],&gMsliceRec.ikw[k],&gMsliceRec.ilw[k])) != 3) {
				write_message("Error reading the Pendellosung Option Indices") ;
				fclose(fp) ;
				return false ;
			}
		}
	}
	
	
	gDisplayRec.tfs = FALSE ;
	if (ReadLine(buffer,256,fp) == NULL) {
		write_message("Error reading the defocus information") ;
		fclose(fp) ;
		return false ;
	}
    if (strchr(buffer,',') != NULL) {
        sscanf(buffer,"%f,%f,%f",&gDisplayRec.f1,&gDisplayRec.f2,&gDisplayRec.df) ;
		gDisplayRec.tfs = TRUE ;
		ftemp = (gDisplayRec.f2-gDisplayRec.f1)/gDisplayRec.df ;
		ftemp = (ftemp > 0. ? ftemp : -ftemp) ;
		NUM_DEFOC = (long)ftemp + 1 ;
	}
    else {
        sscanf(buffer,"%f",&gDisplayRec.f1) ;
		gDisplayRec.f2 = gDisplayRec.f1 ;
		gDisplayRec.df = 0. ;
		NUM_DEFOC = 1 ;
	}

	gInputParameter[23].fv = gDisplayRec.f2 ;
	i = 24 ;
	if (ReadLine(buffer,256,fp) == NULL) {
		write_message("Error reading microscope objective lens aperture") ;
		fclose(fp) ;
		return false ;
	}
    if ((numVariablesAssigned = sscanf(buffer,"%f",&gInputParameter[i++].fv)) < 1) {
		write_message("Error reading microscope objective lens aperture") ;
		fclose(fp) ;
		return false ;
	}
	if (ReadLine(buffer,256,fp) == NULL) {
		write_message("Error reading microscope information") ;
		fclose(fp) ;
		return false ;
	}
    if ((numVariablesAssigned = sscanf(buffer,"%f %f %f %f",&gInputParameter[i].fv,&gInputParameter[i+1].fv,     
                            &gInputParameter[i+2].fv,&gInputParameter[i+3].fv)) != 4) {
		write_message("Error reading the center of the Laue circle and optic axis parameters") ;
		fclose(fp) ;
		return false ;
	}
	
	if (!LAYERED) {
	    for (i=0;i<gNumSymops;i++) {
			if (ReadLine(buffer,256,fp) == NULL) {
	    		sprintf(text,"Error reading symmetry operator #d",i+1) ;
				write_message(text) ;
				fclose(fp) ;
				return false ;
			}        		
	        do {
	            for (j=0,n=0;(c=buffer[n])!=EOS;n++)
	               if (c!=BLANK) symop_txt[i][j++] = c ;
			} while (j==0) ;
	        symop_txt[i][j] = EOS ;
			txtptr_symop[i] = &symop_txt[i][0] ;
		}
		/* Apply the symmetry operators to the basis */
		gSymopsAreChanged = TRUE ;
		ApplySymmetries() ;   
		gSymopsAreChanged = FALSE ;

		ResetActiveElements() ;
	}

	gCalculationIsChanged = FALSE ;
	gBasisIsChanged = FALSE ;
	gPGisChanged = FALSE ;

	if (ReadLine(buffer,256,fp) != NULL) {	
		sscanf(buffer,"%d",&calculation.status) ;
	}
	else {
		calculation.status = 1 ;	/* Nothing has been done	*/
	}
	
	calculation.oldstatus = calculation.status ;	/* For reference  */
	SetCalculationState() ;
	UpdateCalculationStatus() ;	/* based on if the files actally exists */

	if (gAtomRefinementPage != nil) {
		delete gAtomRefinementPage ;
		gAtomRefinementPage = nil ;
	}

	return 1 ;
}

void	ResetActiveElements() {
	extern	short	gNumActiveElements ;
	extern	short	gActiveElementNumber[20] ;
	
	long currentType = -1 ;
	gNumActiveElements = 0 ;
	
	if (MyElements == nil) {
		MyElements = new ElementRecord ;
	}
	MyElements->number = 0 ;
	for (long n=0;n<gNumBasisAtoms;n++)    {
		if ((basis+n)->type.val.iv != currentType) {
			gActiveElementNumber[gNumActiveElements++] = (basis+n)->no.val.iv ;
			currentType = (basis+n)->type.val.iv ;
			MyElements->index[MyElements->number++] = (basis+n)->no.val.iv - 1 ;
		}
	}
}

