// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "globals.h"
#include "config.h"
#include "tiff.h"

#include "TImage.h"

#ifdef IBM
#define WORDS_BIGENDIAN 
#endif
#ifdef WORDS_BIGENDIAN
#define BYTEORDER 0x4D4D
#else 
#define BYTEORDER 0x4949
#endif

#define kMM		0x4D4D
#define kINTEL	0x4949
#define max(x,y) (x) > (y) ? (x) : (y)
typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
extern	MonitorDisplayType  gMonitorDisplayType;
extern	short	gImageDepth ;
short gInverseColorTable[256] ;
extern short	gNumGrays ;
void	MapDataToFloat(TiffRecord *tiff,float *imagedata) ;
void	SetImageDataFromTiffData(TImage *image,TiffRecord *tiff) ;
void	PutMessage(char *) ;
void 	MapTiffDataToDisplayData(unsigned char **data) ;
Boolean WriteImageDataToByteTiff(char *filename,float *data,int ncols,int nrows,Boolean flipImage) ;
typedef	enum {kMMByteOrder = kMM, kIntelByteOrder = kINTEL} ByteOrder ;
extern	unsigned char *gMapFrom16To24BitColor ;
extern	unsigned char *gMapFrom24To16BitColor ;


ByteOrder	gCPUByteOrder = (ByteOrder)BYTEORDER ;
ByteOrder	gServerByteOrder = (ByteOrder)BYTEORDER  ;	// By default

TiffRecord gTiff = {0,0,8,1,1,1,0,0,0,0,0,0,0,0,0,0} ;
extern	TiffRecord gTiff ;

Boolean	Write8BitTiffFile(char *filename,long *n,long *m,unsigned char *data) ;

Boolean	Write8BitTiffFile(char *filename,long *n,long *m,unsigned char *data) {

	Boolean status = false ;
	unsigned char *address ;
	long 	nBytes,i,nrows,ncols ;
	FILE	*fp ;
	
	ncols = *n ;
	nrows = *m ;
	
	if ((fp = CreateTiffFile(filename)) != NULL) {
		WriteTiffDirectory(fp,kGrayScale,ncols,nrows) ;
		nBytes = ncols*sizeof(char) ;
		for (i=0,address= (unsigned char *)data;i<nrows;i++,address += ncols) {
			WriteTiffStrip(fp,address,nBytes) ;
		}
		CloseTiffFile(fp) ;
		status = true ;
	}
	return status ;
}

Boolean WriteImageDataToTiff(char *filename,float *data,int ncols,int nrows,Boolean flipImage) {

	Boolean status = false ;
	float *address ;
	int 	nBytes,i ;
	FILE	*fp ;
	
	if ((fp = CreateTiffFile(filename)) != NULL) {
		WriteTiffDirectory(fp,kFloatingPoint,ncols,nrows) ;
		nBytes = ncols*sizeof(float) ;
		if (flipImage) {
			address = data+(nrows-1)*ncols ;
			for (i=0;i<nrows;i++,address -= ncols) {
				WriteTiffStrip(fp,address,nBytes) ;
			}
		}
		else {
			for (i=0,address=data;i<nrows;i++,address += ncols) {
				WriteTiffStrip(fp,address,nBytes) ;
			}
		}
		CloseTiffFile(fp) ;
		status = true ;
	}
	return status ;
}

Boolean WriteImageDataToByteTiff(char *filename,float *data,int ncols,int nrows,Boolean flipImage) {

	Boolean status = false ;
	float 	*p1,minVal,maxVal,fval,scale ;
	unsigned	char	*address,*c1,*byteData  ;
	int 	nBytes,i ;
	FILE	*fp ;

	// Find min and max so we can scale
	p1 = data ;
	minVal = 1.0e+20 ;
	maxVal = -minVal ;
	for(i=0;i<nrows*ncols;i++) {
		if (*p1 > maxVal)
			maxVal = *p1 ;
		if (*p1 < minVal)
			minVal = *p1 ;
		p1++ ;
	}
	scale = 255./(maxVal-minVal) ; 
	byteData = new unsigned char[nrows*ncols] ;

	c1 = byteData ;
	p1 = data ;
	for(i=0;i<nrows*ncols;i++) {
		fval = (*p1++-minVal)*scale ;
		*c1++ = (unsigned char)fval ;
	}
		
	if ((fp = CreateTiffFile(filename)) != NULL) {
		WriteTiffDirectory(fp,kGrayScale,ncols,nrows) ;
		nBytes = ncols*sizeof(unsigned char) ;
		if (flipImage) {
			address = byteData+(nrows-1)*ncols ;
			for (i=0;i<nrows;i++,address -= ncols) {
				WriteTiffStrip(fp,address,nBytes) ;
			}
		}
		else {
			for (i=0,address=byteData;i<nrows;i++,address += ncols) {
				WriteTiffStrip(fp,address,nBytes) ;
			}
		}
		CloseTiffFile(fp) ;
		status = true ;
	}
	delete byteData ;
	return status ;
}

Boolean WriteDisplayToTiff(char *filename,unsigned char *data,int ncols,int nrows) {
	Boolean status = false ;
	PaletteColor *table ;
	unsigned char *address ;
	unsigned char	*strip,*ucp,*usp ;
	int 	nBytes,i,j ;
	FILE	*fp ;
 	
	if ((fp = CreateTiffFile(filename)) != NULL) {
		if (gMonitorDisplayType == PSEUDOCOLOR) {
			nBytes = 256*sizeof(PaletteColor) ;
			table = (PaletteColor *)malloc(nBytes) ;
			if (table == NULL) {
				printf("\nError allocating enough memory\n") ; fflush(stdout) ;
				return status ;
			}
			WriteTiffDirectory(fp,kPalette,ncols,nrows) ;
 			CreateAppColorMap(table) ;
 			WriteTiffPalette(fp,table,256) ;
			nBytes = ncols ;
 			for (i=0,address=data;i<nrows;i++,address += ncols) {
				WriteTiffStrip(fp,address,nBytes) ;
			}
			free(table) ;
		}
		else {
			// Write the RGB data
			WriteTiffDirectory(fp,kRGB,ncols,nrows) ;
			// The next must be fixed
			nBytes = 3*ncols ;
			strip = new unsigned char[nBytes] ;
			unsigned short	*uusp ; 
			switch(gCPUByteOrder) {
				case kMMByteOrder:
					// The data coming back from the display is RGB0
					switch(gImageDepth) {
						case 15:
			    		case 16:
							// We will convert the data to 24 bit color
				 			for (i=0;i<nrows;i++) {
								ucp = data + 2*i*ncols ;
								uusp = (unsigned short *)ucp ;
								usp = strip ;
								for(j=0;j<ncols;j++) {
									*usp = (*uusp & (0x1F << 10)) >> 10  ;
									*(usp+1) = (*uusp & (0x1F << 5)) >> 5  ;
									*(usp+2) = *uusp & 0x1F  ;
									*usp = gMapFrom16To24BitColor[*usp]  ;
									*(usp+1) = gMapFrom16To24BitColor[*(usp+1)]  ;
									*(usp+2) = gMapFrom16To24BitColor[*(usp+2)]  ;
									usp += 3 ;
									uusp++ ;
								}
								WriteTiffStrip(fp,strip,nBytes) ;
							}
							break ;
						case 24:
			    		case 32:
				 			for (i=0;i<nrows;i++) {
								ucp = data + 4*i*ncols + 1 ;
								usp = strip ;
								for(j=0;j<ncols;j++) {
									*usp++ = *ucp++ ;
									*usp++ = *ucp++ ;
									*usp++ = *ucp++ ;
									ucp++ ;
								}
								WriteTiffStrip(fp,strip,nBytes) ;
							}
							break ;
					}	
					break ;
				case kIntelByteOrder:
					// The data coming back from the display is 0BGR
					switch(gImageDepth) {
						case 15:
			    		case 16:
				 			for (i=0;i<nrows;i++) {
								ucp = data + 2*i*ncols ;
								uusp = (unsigned short *)ucp ;
								usp = strip ;
								for(j=0;j<ncols;j++) {
									*usp = (*uusp & (0x1F << 11)) >> 11  ;
									*(usp+1) = (*uusp & (0x1F << 6)) >> 6  ;
									*(usp+2) = *uusp & 0x1F  ;
									*usp = gMapFrom16To24BitColor[*usp]  ;
									*(usp+1) = gMapFrom16To24BitColor[*(usp+1)]  ;
									*(usp+2) = gMapFrom16To24BitColor[*(usp+2)]  ;
									usp += 3 ;
									uusp++ ;
								}
								WriteTiffStrip(fp,strip,nBytes) ;
							}
							break ;
						case 24:
			    		case 32:
				 			for (i=0;i<nrows;i++) {
								ucp = data + 4*i*ncols ;
								usp = strip ;
								for(j=0;j<ncols;j++) {
									*(usp+2) = *ucp++ ;
									*(usp+1) = *ucp++ ;
									*usp = *ucp++ ;
									usp += 3 ;
									ucp++ ;
								}
								WriteTiffStrip(fp,strip,nBytes) ;
							}
							break ;
					}
					break ;
			}
			delete[] strip ;
		}
		CloseTiffFile(fp) ;
		status = true ;
	}
	return status ;
}

FILE *CreateTiffFile(char * filename) {
	FILE *fp ;
	
	fp = fopen(filename,"w") ;
	return fp ;
}

void CloseTiffFile(FILE *fp) {
	fclose(fp) ;
}

PIFD CreateTiffIFD(short numDE) {
	PIFD pifd ;
	
	pifd = new IFD ; // (PIFD) malloc(sizeof(IFD)) ;
	pifd->numEntries = 0 ;
	pifd->entry = new DIRENTRY[numDE] ; // (PDIRENTRY)malloc(numDE*sizeof(DIRENTRY)) ;
	pifd->nextIFD = 0 ;
	return(pifd) ;
}

void WriteTiffIFD(FILE *fp,PIFD pifd) {
	int numBytes ;
	
	numBytes = sizeof(short) ;
	fwrite(&pifd->numEntries,numBytes,1,fp) ;
	numBytes = pifd->numEntries*sizeof(DIRENTRY) ;
	fwrite((char *)pifd->entry,numBytes,1,fp) ;
	pifd->nextIFD = 0 ;
	numBytes = sizeof(int) ; /* long on some machines */
	fwrite((char *)&pifd->nextIFD,numBytes,1,fp) ;
}
void AddTiffTag(PIFD ifd,short TagID,short TagType,int numValues,int value) {
	
	ifd->entry[ifd->numEntries].deTag = TagID ;
	ifd->entry[ifd->numEntries].deType = TagType ;
	ifd->entry[ifd->numEntries].deLength = numValues ;
#ifdef WORDS_BIGENDIAN	
	/* We need to do this for Motorola Byte order */
	if (TagType == TIFFSHORT && numValues == 1) 
		value <<= 16 ;
#endif
	ifd->entry[ifd->numEntries].deVal = value ;
	ifd->numEntries++ ;
}
void WriteTiffDirectory(FILE *fp,TiffImageType imageType,int width,int height) {
	PIFD pifd ;
	int rational[4],numBytes ;
	int	value ;
	int tgxoffset,tgyoffset,tgstripoffset,tgpaletteoffset ;
	
	switch(imageType) {
		case kBiLevel:
			break ;
		case kGrayScale:
			tgxoffset = sizeof(TIFFHDR)+sizeof(short)+NEGRAYSCALE*sizeof(DIRENTRY)+sizeof(int)  ;
			tgyoffset = tgxoffset + 2*sizeof(int) ;
			tgstripoffset = tgyoffset + 2*sizeof(int) ;
			WriteTiffHeader(fp) ;
			pifd = CreateTiffIFD(NEGRAYSCALE) ;
			AddTiffTag(pifd,TGNEWSUBFILETYPE,TIFFLONG,1,0) ;
			AddTiffTag(pifd,TGIMAGEWIDTH,TIFFLONG,1,width) ;
			AddTiffTag(pifd,TGIMAGELENGTH,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGBITSPERSAMPLE,TIFFSHORT,1,8) ;
			AddTiffTag(pifd,TGCOMPRESSION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGPHOTOMETRICINTERPRETATION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGSTRIPOFFSETS,TIFFLONG,1,tgstripoffset) ;
			AddTiffTag(pifd,TGROWSPERSTRIP,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGSTRIPBYTECOUNTS,TIFFLONG,1,width*height) ;
			AddTiffTag(pifd,TGXRESOLUTION,TIFFRATIONAL,1,tgxoffset) ;
			AddTiffTag(pifd,TGYRESOLUTION,TIFFRATIONAL,1,tgyoffset) ;
			AddTiffTag(pifd,TGRESOLUTIONUNIT,TIFFSHORT,1,2) ;
			WriteTiffIFD(fp,pifd) ;
			rational[0] = 72 ;
			rational[1] = 1 ;
			rational[2] = 72 ;
			rational[3] = 1 ;
			numBytes = 4*sizeof(int) ;
			fwrite((char *)rational,numBytes,1,fp) ;
			delete[] pifd->entry ; // free(pifd->entry) ;
			delete pifd ; //free(pifd) ;
			break ;
		case kPalette:
			tgxoffset = sizeof(TIFFHDR)+sizeof(short)+NEPALETTE*sizeof(DIRENTRY)+sizeof(int) ;
			tgyoffset = tgxoffset + 2*sizeof(int) ;
			tgpaletteoffset = tgyoffset + 2*sizeof(int) ;
			tgstripoffset = tgpaletteoffset + 3*256*sizeof(short) ;	/* The colortable	*/
			WriteTiffHeader(fp) ;
			pifd = CreateTiffIFD(NEPALETTE) ;
			AddTiffTag(pifd,TGNEWSUBFILETYPE,TIFFLONG,1,0) ;
			AddTiffTag(pifd,TGIMAGEWIDTH,TIFFLONG,1,width) ;
			AddTiffTag(pifd,TGIMAGELENGTH,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGBITSPERSAMPLE,TIFFSHORT,1,8) ;
			AddTiffTag(pifd,TGCOMPRESSION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGPHOTOMETRICINTERPRETATION,TIFFSHORT,1,3) ;
			AddTiffTag(pifd,TGSTRIPOFFSETS,TIFFLONG,1,tgstripoffset) ;
			AddTiffTag(pifd,TGROWSPERSTRIP,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGSTRIPBYTECOUNTS,TIFFLONG,1,width*height) ;
			AddTiffTag(pifd,TGXRESOLUTION,TIFFRATIONAL,1,tgxoffset) ;
			AddTiffTag(pifd,TGYRESOLUTION,TIFFRATIONAL,1,tgyoffset) ;
			AddTiffTag(pifd,TGRESOLUTIONUNIT,TIFFSHORT,1,2) ;
			AddTiffTag(pifd,TGCOLORMAP,TIFFSHORT,768,tgpaletteoffset) ;
			WriteTiffIFD(fp,pifd) ;
			rational[0] = 72 ;
			rational[1] = 1 ;
			rational[2] = 72 ;
			rational[3] = 1 ;
			numBytes = 4*sizeof(int) ;
			fwrite((char *)rational,numBytes,1,fp) ;
			delete[] pifd->entry ; // free(pifd->entry) ;
			delete pifd ; //free(pifd) ;
			break ;
		case kRGB:
			tgxoffset = sizeof(TIFFHDR)+sizeof(short)+NERGB*sizeof(DIRENTRY)+sizeof(int) ;
			tgyoffset = tgxoffset + 2*sizeof(int) ;
			tgstripoffset = tgyoffset + 2*sizeof(int) ;
			WriteTiffHeader(fp) ;
			pifd = CreateTiffIFD(NERGB) ;
			AddTiffTag(pifd,TGNEWSUBFILETYPE,TIFFLONG,1,0) ;
			AddTiffTag(pifd,TGIMAGEWIDTH,TIFFLONG,1,width) ;
			AddTiffTag(pifd,TGIMAGELENGTH,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGBITSPERSAMPLE,TIFFSHORT,1,8) ;
			AddTiffTag(pifd,TGCOMPRESSION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGPHOTOMETRICINTERPRETATION,TIFFSHORT,1,2) ;
			AddTiffTag(pifd,TGSAMPLESPERPIXEL,TIFFSHORT,1,3) ;
			AddTiffTag(pifd,TGSTRIPOFFSETS,TIFFLONG,1,tgstripoffset) ;
			AddTiffTag(pifd,TGROWSPERSTRIP,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGSTRIPBYTECOUNTS,TIFFLONG,1,3*width*height) ;
			AddTiffTag(pifd,TGXRESOLUTION,TIFFRATIONAL,1,tgxoffset) ;
			AddTiffTag(pifd,TGYRESOLUTION,TIFFRATIONAL,1,tgyoffset) ;
			AddTiffTag(pifd,TGRESOLUTIONUNIT,TIFFSHORT,1,2) ;
			WriteTiffIFD(fp,pifd) ;
			rational[0] = 72 ;
			rational[1] = 1 ;
			rational[2] = 72 ;
			rational[3] = 1 ;
			numBytes = 4*sizeof(int) ;
			fwrite((char *)rational,numBytes,1,fp) ;
			delete[] pifd->entry ; // free(pifd->entry) ;
			delete pifd ; //free(pifd) ;
			break ;
		case kFloatingPoint:
			tgxoffset = sizeof(TIFFHDR)+sizeof(short)+NEFLPOINT*sizeof(DIRENTRY)+sizeof(int)  ;
			tgyoffset = tgxoffset + 2*sizeof(int) ;
			tgstripoffset = tgyoffset + 2*sizeof(int) ;
			WriteTiffHeader(fp) ;
			pifd = CreateTiffIFD(NEFLPOINT) ;
			AddTiffTag(pifd,TGNEWSUBFILETYPE,TIFFLONG,1,0) ;
			AddTiffTag(pifd,TGIMAGEWIDTH,TIFFLONG,1,width) ;
			AddTiffTag(pifd,TGIMAGELENGTH,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGBITSPERSAMPLE,TIFFSHORT,1,8*sizeof(float)) ;
			AddTiffTag(pifd,TGCOMPRESSION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGPHOTOMETRICINTERPRETATION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGSAMPLEFORMAT,TIFFSHORT,1,3) ;
			AddTiffTag(pifd,TGSAMPLESPERPIXEL,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGSTRIPOFFSETS,TIFFLONG,1,tgstripoffset) ;
			AddTiffTag(pifd,TGROWSPERSTRIP,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGSTRIPBYTECOUNTS,TIFFLONG,1,width*height*sizeof(float)) ;
			AddTiffTag(pifd,TGXRESOLUTION,TIFFRATIONAL,1,tgxoffset) ;
			AddTiffTag(pifd,TGYRESOLUTION,TIFFRATIONAL,1,tgyoffset) ;
			AddTiffTag(pifd,TGRESOLUTIONUNIT,TIFFSHORT,1,2) ;
			WriteTiffIFD(fp,pifd) ;
			rational[0] = 72 ;
			rational[1] = 1 ;
			rational[2] = 72 ;
			rational[3] = 1 ;
			numBytes = 4*sizeof(int) ;
			fwrite((char *)rational,numBytes,1,fp) ;
			delete[] pifd->entry ; // free(pifd->entry) ;
			delete pifd ; //free(pifd) ;
			break ;
		case kComplexFloatingPoint:
			tgxoffset = sizeof(TIFFHDR)+sizeof(short)+NECFLPOINT*sizeof(DIRENTRY)+sizeof(int)  ;
			tgyoffset = tgxoffset + 2*sizeof(int) ;
			tgstripoffset = tgyoffset + 2*sizeof(int) ;
			WriteTiffHeader(fp) ;
			pifd = CreateTiffIFD(NECFLPOINT) ;
			AddTiffTag(pifd,TGNEWSUBFILETYPE,TIFFLONG,1,0) ;
			AddTiffTag(pifd,TGIMAGEWIDTH,TIFFLONG,1,width) ;
			AddTiffTag(pifd,TGIMAGELENGTH,TIFFLONG,1,height) ;
			value = 8*sizeof(float) ; value << 16 ; value +=  8*sizeof(float) ;
			AddTiffTag(pifd,TGBITSPERSAMPLE,TIFFSHORT,1,value) ;
			AddTiffTag(pifd,TGCOMPRESSION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGPHOTOMETRICINTERPRETATION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGSAMPLEFORMAT,TIFFSHORT,1,3) ;
			AddTiffTag(pifd,TGSAMPLESPERPIXEL,TIFFSHORT,1,2) ;
			AddTiffTag(pifd,TGSTRIPOFFSETS,TIFFLONG,1,tgstripoffset) ;
			AddTiffTag(pifd,TGROWSPERSTRIP,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGSTRIPBYTECOUNTS,TIFFLONG,1,2*width*height*sizeof(float)) ;
			AddTiffTag(pifd,TGXRESOLUTION,TIFFRATIONAL,1,tgxoffset) ;
			AddTiffTag(pifd,TGYRESOLUTION,TIFFRATIONAL,1,tgyoffset) ;
			AddTiffTag(pifd,TGRESOLUTIONUNIT,TIFFSHORT,1,2) ;
			WriteTiffIFD(fp,pifd) ;
			rational[0] = 72 ;
			rational[1] = 1 ;
			rational[2] = 72 ;
			rational[3] = 1 ;
			numBytes = 4*sizeof(int) ;
			fwrite((char *)rational,numBytes,1,fp) ;
			delete[] pifd->entry ; // free(pifd->entry) ;
			delete pifd ; //free(pifd) ;
			break ;
	}
}
void WriteTiffPalette( FILE *fp, PaletteColor *colorTable, int numEntries) {
	short shortZero;
	int numBytes,i ;
	
	numBytes = sizeof(short) ;
	shortZero = 0 ;
	for(i=0;i<256;i++) {
		if (i < numEntries)
			fwrite(&colorTable[i].red,numBytes,1,fp) ;
		else 
			fwrite(&shortZero,numBytes,1,fp) ;
	}
	for(i=0;i<256;i++) {
		if (i < numEntries)
			fwrite(&colorTable[i].green,numBytes,1,fp) ;
		else 
			fwrite(&shortZero,numBytes,1,fp) ;
	}
	for(i=0;i<256;i++) {
		if (i < numEntries)
			fwrite(&colorTable[i].blue,numBytes,1,fp) ;
		else 
			fwrite(&shortZero,numBytes,1,fp) ;
	}
}
void WriteTiffStrip( FILE *fp, Ptr data,int nBytes) {
	int numBytes ;
	
	/* Write the data	*/
	numBytes = nBytes ;
	fwrite(data,numBytes,1,fp) ;
}
void WriteTiffHeader(FILE *fp) {
	TIFFHDR header ;
	int numBytes ;
	
	header.thByteOrder = BYTEORDER; 	
	header.thVersion = 42 ;
	header.thIfdOffset = 8 ;
	numBytes = sizeof(TIFFHDR) ;
	fwrite(&header,numBytes,1,fp) ;
}
void WriteTiffTag( short TagID, short TagType, int numValues,int  offset) {
	DIRENTRY ifd ;
	
	ifd.deTag = TagID ;
	ifd.deType = TagType ;
	ifd.deLength = numValues ;
	ifd.deVal = offset ;
}

void CreateAppColorMap(PaletteColor *colorTable) {
	int         i,ndx ;
	int 	numEntries = 104 ;
	//extern unsigned long gLut[] ;


	for(i=0;i<256;i++) {
		(colorTable+i)->red = 0 ;
		(colorTable+i)->green = 0 ;
		(colorTable+i)->blue = 0 ;
	}
    	if (gMonitorDisplayType == PSEUDOCOLOR) {
		for(i=0;i < numEntries;i++) {
     			ndx = gColors[i].pixel ;// This only works on a Pseudocolor
									// display since otherwise the index goes 
									// out of range
       			(colorTable+ndx)->red  			= gColors[i].red ;
        		(colorTable+ndx)->blue  		= gColors[i].green ;
        		(colorTable+ndx)->green  		= gColors[i].blue ;
		}
	}
	else {
		for(i=0;i < numEntries;i++) {
       			(colorTable+i)->red  		= gColors[i].red ;
        		(colorTable+i)->blue  		= gColors[i].green ;
        		(colorTable+i)->green  		= gColors[i].blue ;
		}
	}
}
void CreateInverseColorTable() {
    int         i,ndx ;
    int		numEntries=104 ;

    for(i=0;i<256;i++) {
		gInverseColorTable[i] = 0 ; 
    }
    if (gMonitorDisplayType == PSEUDOCOLOR) {
    	for(i=0;i < numEntries;i++) {
    		ndx = gColors[i].pixel ;
			gInverseColorTable[ndx]	= i ; 
    	}
    }
    else {
    	for(i=0;i < numEntries;i++) {
			gInverseColorTable[i]	= i ; 
    	}
    }
}

Boolean	CreateImageFromTiffFile(char *filename,TImage **imageH) {
	TIFFHDR thdr ;
	FILE *fp ;
	char	text[256] ;
	long	size,i,length ;
	float	*p1 ;
	
	fp = fopen(filename,"r") ;
	if (fp == NULL) {
		sprintf(text,"Unable to open the file %s !",filename) ;
		PutMessage(text) ; 
		return false ;
	}
	if (ReadTiffHeader(fp,&thdr)) {
		if (ReadTiffDirectory(fp,&thdr)) {
			fclose(fp) ;
			*imageH = (TImage *)malloc(sizeof(TImage)) ;
			if (*imageH != nil) {
				(*imageH)->mWidth = gTiff.width ;
				(*imageH)->mHeight = gTiff.height ;
				size = gTiff.width*gTiff.height*sizeof(float) ;
				(*imageH)->mData = (float *)malloc(size) ;
				if ((*imageH)->mData == nil) {
					free((*imageH)) ;
					return nil ;
				}
				(*imageH)->mName = nil ;
				(*imageH)->mMeanIsZero = false ;
				(*imageH)->mIsVisible = false ;
				(*imageH)->mIsDirty = false ;	// It has to be forced to true 
				(*imageH)->mWindow = nil ;
				(*imageH)->mOffscreen = nil ;
				
				// Zero the image
				p1 = (*imageH)->mData ;
				length = gTiff.width*gTiff.height ;
				for( i=0; i < length ; i++) {
					*p1++ = 0. ;
				}
				// Transfer the data from gTiff.data -> image->data
				SetImageDataFromTiffData(*imageH,&gTiff) ;
				delete[] gTiff.data ;
				return true ;
			}
		}
		else {
			fclose(fp) ;
			PutMessage("Error Reading Tiff Directory") ; 
		}
	}
	else {
		fclose(fp) ;
		PutMessage("Error Reading Tiff Header") ; 
	}	
		
	return false ;
}

void	SetImageDataFromTiffData(TImage *image,TiffRecord *tiff) {
	// Map the data in the tiff record in whatever format it is to
	// a floating point format stored in the image
	MapDataToFloat(tiff,image->mData) ;
}

Boolean ReadTiffFile(char *filename,unsigned char **imagedata) {
	TIFFHDR thdr ;
	FILE *fp ;
	
	fp = fopen(filename,"r") ;
	if (fp == NULL) {
		printf("\nUnable to open the file %s !\n",filename) ; fflush(stdout) ;
		return false ;
	}
	if (ReadTiffHeader(fp,&thdr)) {
		if (ReadTiffDirectory(fp,&thdr)) {
			*imagedata = gTiff.data ;	// We actually don't destroy the data
										// in the tiff reader at all
			MapTiffDataToDisplayData(imagedata) ;	
			fclose(fp) ;
			return true ;
		}
		else {
			printf("\nError Reading Tiff Directory\n") ; fflush(stdout) ;
			fclose(fp) ;
		}
	}
	else {
		printf("\nError Reading Tiff Header\n") ; fflush(stdout) ;
		fclose(fp) ;
	}
	return false ;
}

void MapTiffDataToDisplayData(unsigned char **data) {
	int i,j ;
	float	maxDisplayValue = gNumGrays-1 ;
	int bytesPerSample = gTiff.bitsPerSample/8 ;

	if (gTiff.photoMetricInterpretation < 2) {
		// *data already points to the gTiff.data which we preserve
		/* We must change the data so that it can map properly to our LUT	*/
		MapToAppLut(gTiff.data,gTiff.width,gTiff.height,
			gTiff.sampleFormat,gTiff.bitsPerSample,gTiff.samplesPerPixel) ;
			
		// Data has been assigned the pixel value to  used in looking up the RGB value
		// We want to 
 		unsigned char *ucp = (unsigned char *)gTiff.data ;
		unsigned char *usp  ;
		if (gMonitorDisplayType == PSEUDOCOLOR) {
			// We should actually map the data to our LUT
			// Since we are not modifying our look up table
			// I will treat all the data as if it is gray scale
			// and map the data to the NUM_GRAYS grayscales by the red component
			usp = (unsigned char *)gTiff.data ;
			int	ndx,val ;
			for (i=0;i<gTiff.height;i++) {
				for(j=0;j<gTiff.width;j++) {
					val = *usp ;
					ndx = val*maxDisplayValue/255 ;
					*ucp++ = ndx ;
					usp += 3 ;
				}
			}
		}
		else {
			switch(gImageDepth) {
				case 15:
				case 16:
					*data = new unsigned char[gTiff.width*gTiff.height*2] ;
					break ;
				case 24:
				case 32:
					*data = new unsigned char[gTiff.width*gTiff.height*4] ;
					break ;
			}
			usp = (unsigned char *)gTiff.data ;
			ucp = *data ;
			unsigned short	*uusp = (unsigned short *)ucp ;
			unsigned char	rb,gb,bb ;

			switch(gCPUByteOrder) {
				case kMMByteOrder:
					switch(gImageDepth) {
						case 15:
			    		case 16:
				 			for (i=0;i<gTiff.height;i++) {
								for(j=0;j<gTiff.width;j++) {
									rb = gColors[*usp].red ;
									if (bytesPerSample  == 1) {
										rb = gMapFrom24To16BitColor[rb] ;	
										gb = rb ;
										bb = rb ;
									}
									else {	
										gb = gColors[*usp].green ;
										bb = gColors[*usp].blue ;
										rb = gMapFrom24To16BitColor[rb] ;
										gb = gMapFrom24To16BitColor[gb] ;
										bb = gMapFrom24To16BitColor[bb] ;
									}
									*uusp++ = rb << 10 | gb << 5 | bb ;
									usp++ ;
								}
							}
							break ;
						case 24:
			    		case 32:
				 			for (i=0;i<gTiff.height;i++) {
								for(j=0;j<gTiff.width;j++) {
									*ucp++ = 0 ;
									*ucp++ = gColors[*usp].red ;
									*ucp++ = gColors[*usp].green ;
									*ucp++ = gColors[*usp].blue ;
									usp++ ;
								}
							}
							break ;
					}	
					break ;
				case kIntelByteOrder:
					switch(gImageDepth) {
						case 15:
			    		case 16:
				 			for (i=0;i<gTiff.height;i++) {
								for(j=0;j<gTiff.width;j++) {
									rb = gColors[*usp].red ;
									if (bytesPerSample  == 1) {
										rb = gMapFrom24To16BitColor[rb] ;	
										gb = rb ;
										bb = rb ;
									}
									else {	
										gb = gColors[*usp].green ;
										bb = gColors[*usp].blue ;
										rb = gMapFrom24To16BitColor[rb] ;
										gb = gMapFrom24To16BitColor[gb] ;
										bb = gMapFrom24To16BitColor[bb] ;
									}
									*uusp++ = bb & 0x1F | (gb & 0x1F) << 6 | (rb & 0x1F) << 11 ;
									usp++ ;
								}
							}
							break ;
						case 24:
			    		case 32:
				 			for (i=0;i<gTiff.height;i++) {
								for(j=0;j<gTiff.width;j++) {
									*(ucp+3) = 0 ;
									*(ucp+2) = gColors[*usp].red ;
									*(ucp+1) = gColors[*usp].green ;
									*ucp = gColors[*usp].blue ;
									ucp += 4 ;
									usp++ ;
								}
							}
							break ;
					}	
					break ;
			}
		}					
	}
	else if (gTiff.photoMetricInterpretation == 2) {
 		unsigned char *ucp = (unsigned char *)gTiff.data ;
		unsigned char *usp  ;
		if (gMonitorDisplayType == PSEUDOCOLOR) {
			// We should actually map the data to our LUT
			// Since we are not modifying our look up table
			// I will treat all the data as if it is gray scale
			// and map the data to the NUM_GRAYS grayscales by the red component
			usp = (unsigned char *)gTiff.data ;
			int	ndx,val ;
			for (i=0;i<gTiff.height;i++) {
				for(j=0;j<gTiff.width;j++) {
					val = *usp ;
					ndx = val*maxDisplayValue/255 ;
					*ucp++ = ndx ;
					usp += 3 ;
				}
			}
		}
		else {
			// This is RGB data
			// We only know how to treat our own color TIFF files where the
			// planar configuration is 1
			// The data is stored as RRGGBBRRGGBB 	and we must convert it to
			// 		- 00RRGGBB00RRGGBB	BIGENDIAN architecture
			// or 	- BGR00	Intel architecture
			*data = new unsigned char[gTiff.width*gTiff.height*4] ;
			usp = (unsigned char *)gTiff.data ;
			ucp = *data ;
			unsigned short	*uusp = (unsigned short *)ucp ;
			unsigned char	rb,gb,bb ;

			switch(gCPUByteOrder) {
				case kMMByteOrder:
					switch(gImageDepth) {
						case 15:
			    		case 16:
				 			for (i=0;i<gTiff.height;i++) {
								for(j=0;j<gTiff.width;j++) {
									rb = *usp++ ;
									gb = *usp++ ;
									bb = *usp++ ;
									rb = gMapFrom24To16BitColor[rb] ;
									gb = gMapFrom24To16BitColor[gb] ;
									bb = gMapFrom24To16BitColor[bb] ;
									*uusp++ = rb << 10 | gb << 5 | bb ;
								}
							}
							break ;
						case 24:
			    		case 32:
				 			for (i=0;i<gTiff.height;i++) {
								for(j=0;j<gTiff.width;j++) {
									*ucp++ = 0 ;
									*ucp++ = *usp++ ;
									*ucp++ = *usp++ ;
									*ucp++ = *usp++ ;
								}
							}
							break ;
					}	
					break ;
				case kIntelByteOrder:
					switch(gImageDepth) {
						case 15:
			    		case 16:
				 			for (i=0;i<gTiff.height;i++) {
								for(j=0;j<gTiff.width;j++) {
									rb = *usp++ ;
									gb = *usp++ ;
									bb = *usp++ ;
									rb = gMapFrom24To16BitColor[rb] ;
									gb = gMapFrom24To16BitColor[gb] ;
									bb = gMapFrom24To16BitColor[bb] ;
									*uusp++ = bb & 0x1F | (gb & 0x1F) << 6 | (rb & 0x1F) << 11 ;
								}
							}
							break ;
						case 24:
			    		case 32:
				 			for (i=0;i<gTiff.height;i++) {
								for(j=0;j<gTiff.width;j++) {
									*(ucp+3) = 0 ;
									*(ucp+2) = *usp++ ;
									*(ucp+1) = *usp++ ;
									*ucp = *usp++ ;
									ucp += 4 ;
								}
							}
							break ;
					}	
					break ;
			}
		}	
	}
}

Boolean ReadTiffDirectory(FILE *fp,TIFFHDR *thdr) {
	PDIRENTRY dirEntry,nextEntry ;
	long	nextIFD ;
	short	numEntries,i ;
	int		currFilePosition ;
	unsigned long	numBytes,numDataBytes ;
	unsigned char *address ;
	
	/* Move to the position of the IFD	*/
	currFilePosition = ftell(fp) ;
	fseek(fp,thdr->thIfdOffset,0) ;
	/* Read the number of entries	*/
	if (fread(&numEntries,sizeof(short),1,fp) != 1) {
		printf("\nError reading number of entries\n") ; fflush(stdout) ;
		return false;
	}

	if (thdr->thByteOrder != BYTEORDER) {
		numEntries = ByteSwapShort(numEntries) ;
	}
    
	dirEntry = new DIRENTRY[numEntries] ; // (PDIRENTRY)malloc(numEntries*sizeof(DIRENTRY)) ;
	if (dirEntry == 0) {
		printf("\nError allocating %d number of entries\n",numEntries) ; 
		fflush(stdout) ;
		return false ;
	}
	
	gTiff.samplesPerPixel = 1 ; /* In case it does not get set later */
	gTiff.sampleFormat = 1 ;

	
	/* Read the directory entries 	*/
	if (fread(dirEntry,numEntries*sizeof(DIRENTRY),1,fp) != 1) {
		delete[] dirEntry ; // free(dirEntry) ;
		printf("\nError reading entries\n") ; fflush(stdout) ;
		return false ;
	}	
	/*	Read the pointer to the next directory	*/	
	if (fread((char *)&nextIFD,sizeof(int),1,fp) != 1) {
		printf("\nError reading next IFD\n") ; fflush(stdout) ;
		delete[] dirEntry ; // free(dirEntry) ;
		return false ;
	}
	/* Process the tags	*/
	nextEntry = dirEntry ;
	for(i=0;i<numEntries;i++) {
		ProcessTiffDirEntry(fp,thdr,nextEntry) ;
		nextEntry++ ;
	} 
	
	/* Check to see if this is a kind of TIFF file that we support	*/
	switch(gTiff.photoMetricInterpretation) {
		case 0:	/* WhiteIsZero	*/
		case 1: /* WhiteIsBlack	*/
			break ;	
		case 2:	/* RGB	*/
			break ;
		case 3: /* RGB Palette */
			break ;
		case 4:	/* Transparency Mask */
			/* Unsupported	*/
			printf("\nApplication does not support Transparency Mask Tiff Files\n") ; fflush(stdout) ;
			delete[] dirEntry ; // free(dirEntry) ;
			return false ;
			break ;
		case 5:	/* CMYK */
			/* Unsupported	*/
			printf("\nApplication does not support CMYK Tiff Files\n") ; fflush(stdout) ;
			delete[] dirEntry ; // free(dirEntry) ;
			return false ;
			break ;
		case 6:	/* YCbCr */
			/* Unsupported	*/
			printf("\nApplication does not support YCbCr Tiff Files\n") ; fflush(stdout) ;
			delete[] dirEntry ; // free(dirEntry) ;
			return false ;
			break ;
		case 8:	/* CIELab */
			/* Unsupported	*/
			printf("\nApplication does not support CIELab Tiff Files\n") ; fflush(stdout) ;
			delete[] dirEntry ; // free(dirEntry) ;
			return false ;
			break ;
	}
	switch(gTiff.compression) {
		case 1:		// Uncompressed
			break ;
		default:	// Compressed file
			// Unsupported
			printf("\nApplication does not support compressed Tiff Files\n") ; 
			fflush(stdout) ;
			delete[] dirEntry ; // free(dirEntry) ;
			return false ;
			break ;
	}
			
	/* Read the colormap if one exists	*/
	if (gTiff.photoMetricInterpretation == 3 && gTiff.numPaletteColors != 0) {	
		// Palette Image
		currFilePosition = ftell(fp) ;
		fseek(fp,gTiff.paletteOffset,0) ;
		if (gTiff.palette != nil) {
			delete[] gTiff.palette ; // free(gTiff.palette) ;
		}
		numBytes = gTiff.numPaletteColors*sizeof(PaletteColor) ; 
		gTiff.palette = new PaletteColor[gTiff.numPaletteColors] ; // (PaletteColor *)malloc(numBytes) ;
		if (gTiff.palette != nil) {
			fread(gTiff.palette,sizeof(PaletteColor),gTiff.numPaletteColors,fp) ;
			// Arrange the palette into its regular order. 
			// It was read in as RRRR..GGGG...BBB
			// while it should be organized into RGBRGBRGB
			// Also it may have to be byte-swapped
			delete[] gTiff.palette ; // free(gTiff.palette) ;
			gTiff.palette = NULL ;
		}
		fseek(fp,currFilePosition,0) ;
	}
	
	/* Read the Image data	*/
	/* Move to the correct position	*/
	currFilePosition = ftell(fp) ;
	
	if (gTiff.numberOfStrips > 1) {
		/* Get the appropriate offsets and bytecounts	*/
		fseek(fp,*gTiff.stripOffset,0) ;
		for(i=0;i<gTiff.numberOfStrips;i++) {
			fread(gTiff.stripOffset+i,sizeof(int),1,fp) ;
		    if (thdr->thByteOrder != BYTEORDER) {
		    	*(gTiff.stripOffset+i) = ByteSwapLong(*(gTiff.stripOffset+i)) ;
		    }
			if (gTiff.stripOffsetType == TIFFSHORT && thdr->thByteOrder == kMM) {
				*(gTiff.stripOffset+i) >> 16 ;
			}
		}
		fseek(fp,*gTiff.stripByteCount,0) ;
		numDataBytes = 0 ;
		for(i=0;i<gTiff.numberOfStrips;i++) {
			fread(gTiff.stripByteCount+i,sizeof(int),1,fp) ;
		    if (thdr->thByteOrder != BYTEORDER) {
		    	*(gTiff.stripByteCount+i) = ByteSwapLong(*(gTiff.stripByteCount+i)) ;
		    }
			if (gTiff.stripByteCountType == TIFFSHORT && thdr->thByteOrder == kMM) {
				*(gTiff.stripByteCount+i) >> 16 ;
			}
			numDataBytes += *(gTiff.stripByteCount+i) ;
		}
	}
	else {
		if (gTiff.stripByteCount == nil) {
			gTiff.stripByteCount = new int[32] ;
			numDataBytes = gTiff.width*gTiff.height*gTiff.bitsPerSample/8 ;
			gTiff.stripByteCount[0] = numDataBytes ;
		}
		else {
			numDataBytes = gTiff.stripByteCount[0] ;
		}
	}				
			
	gTiff.data = new unsigned char[numDataBytes] ; 
	if (gTiff.data == NULL) {
		fprintf(stderr,"\nMemory allocation error") ; fflush(stderr) ;
		return false ;
	}
		
	address = gTiff.data ;
	for(i=0;i<gTiff.numberOfStrips;i++) {
		fseek(fp,gTiff.stripOffset[i],0) ;
		/* Read the values	*/
		fread(address,sizeof(char),gTiff.stripByteCount[i],fp) ;
		address += gTiff.stripByteCount[i] ;
	}
	
	delete[] gTiff.stripOffset ; // free(gTiff.stripOffset) ;
	delete[] gTiff.stripByteCount ; // free(gTiff.stripByteCount) ;
	
	if (thdr->thByteOrder != BYTEORDER) {
		SwapDataBytes(gTiff.data,gTiff.width,gTiff.height,
    			gTiff.sampleFormat,gTiff.bitsPerSample,gTiff.samplesPerPixel) ;
	}
	
	fseek(fp,currFilePosition,0) ;
	delete[] dirEntry ; // free(dirEntry) ;
	return true ;
}

void SwapDataBytes(unsigned char *data,int width,int height,
		short sampleFormat,short bitsPerSample,short samplesPerPixel) {
	int bytesPerSample,i,numValues ;
	unsigned short	*sp ;
	unsigned int	*lp ;
	float	*fp ;

	bytesPerSample = bitsPerSample/8 ;
	numValues = samplesPerPixel*width*height ;
	
	switch(sampleFormat) {
		case 1:	/* Unsigned integer data */
		case 2:	/* Twos's complement signed integer data	*/
		case 4:	/* Undefined, treat as unsigned integer data	*/
			switch(bytesPerSample) {
				case 2:
					sp = (unsigned short *)data ;
					for(i=0;i<numValues;i++) {
						*sp = ByteSwapShort(*sp) ;
						sp++ ;
					}
					break ;
				case 4:
					lp = (unsigned int *)gTiff.data ;
					for(i=0;i<numValues;i++) {
						*lp = ByteSwapLong(*lp) ;
						lp++ ;
					}
					break ;
				default:
					break ;
			}
			break ;
		case 3:	/* IEEE floating point data [IEEE]	*/
			lp = (unsigned int *)data ;
			fp = (float *)data ;
			for(i=0;i<numValues;i++) {
				*fp = ByteSwapFloat(*lp) ;
				lp++ ;
				fp++ ;
			}
			break ;
	}

}
void MapToAppLut(unsigned char *data,int width,int height,short sampleFormat,
		short bitsPerSample,short samplesPerPixel) {
	float *p1,value,minValue,maxValue,range ;
	int	iminValue,imaxValue ;
	unsigned char *cp ;
	char	*ccp ;
	unsigned short *usp ;
	short	*sp ;
	unsigned int  *uip ;
	int *ip ;
	int i,ivalue ;
	short bytesPerSample ;
	long	numValues ;
	
	bytesPerSample = bitsPerSample/8 ;
	numValues = samplesPerPixel*width*height ;
		
	switch(sampleFormat) {
		case 1:	/* Unsigned integer data */
		case 4:	/* Undefined, treat as unsigned integer data	*/
			switch (bytesPerSample) {
				case 1:	/* Grayscale	*/
					cp = data ;
					iminValue = *cp ;
					imaxValue = *cp ;
					for(i=0;i<numValues;i++) {
						iminValue = (*cp < iminValue ? *cp : iminValue) ;
						imaxValue = (*cp > imaxValue ? *cp : imaxValue) ;
						cp++ ;
					}
					range = imaxValue- iminValue ;
					for(i=0,cp=data;i<numValues;i++) {
						ivalue = *cp ;
						value = (gNumGrays-1)*(ivalue-iminValue)/range ;
						ivalue = (int)value ;
						if (gTiff.photoMetricInterpretation == 0)
							ivalue = (gNumGrays-1) - ivalue ;
						*cp++ = gColors[ivalue].pixel ;
					}
					break ;
				case 2:
					cp = data ;
					usp=(unsigned short *)data ;
					iminValue = *usp ;
					imaxValue = *usp ;
					for(i=0;i<numValues;i++) {
						iminValue = (*usp < iminValue ? *usp : iminValue) ;
						imaxValue = (*usp > imaxValue ? *usp : imaxValue) ;
						usp++ ;
					}
					range = imaxValue- iminValue ;
					for(i=0,cp=data,usp=(unsigned short *)data;i<numValues;i++) {
						ivalue = *usp ;
						value = (gNumGrays-1)*(ivalue-iminValue)/range ;
						ivalue = (int)value ;
						if (gTiff.photoMetricInterpretation == 0)
							ivalue = (gNumGrays-1) - ivalue ;
						*cp++ = gColors[ivalue].pixel ;
						usp++ ;
					}
					break ;
				case 4:
					uip=(unsigned int *)data ;
					iminValue = *uip ;
					imaxValue = *uip ;
					for(i=0;i<numValues;i++) {
						iminValue = (*uip < iminValue ? *uip : iminValue) ;
						imaxValue = (*uip > imaxValue ? *uip : imaxValue) ;
						uip++ ;
					}
					range = imaxValue- iminValue ;
					for(i=0,cp=data,uip=(unsigned int *)data;i<numValues;i++) {
						ivalue = *uip ;
						value = (gNumGrays-1)*(ivalue-iminValue)/range ;
						ivalue = (int)value ;
						if (gTiff.photoMetricInterpretation == 0)
							ivalue = (gNumGrays-1) - ivalue ;
						*cp++ = gColors[ivalue].pixel ;
						uip++ ;
					}
					break ;
			}
			break ;
		case 2:	/* Twos's complement signed integer data	*/
			switch (bytesPerSample) {
				case 1:	/* Grayscale	*/
					ccp=(char *)data ;
					iminValue = *ccp ;
					imaxValue = *ccp ;
					for(i=0;i<numValues;i++) {
						iminValue = (*ccp < iminValue ? *ccp : iminValue) ;
						imaxValue = (*ccp > imaxValue ? *ccp : imaxValue) ;
						ccp++ ;
					}
					range = imaxValue- iminValue ;
					for(i=0,ccp=(char *)data,cp=(unsigned char *)data;i<numValues;i++) {
						ivalue = *ccp ;
						value = (gNumGrays-1)*(ivalue-iminValue)/range ;
						ivalue = (int)value ;
						if (gTiff.photoMetricInterpretation == 0)
							ivalue = (gNumGrays-1) - ivalue ;
						*cp++ = gColors[ivalue].pixel ;
						ccp++ ;
					}
					break ;
				case 2:
					sp=( short *)data ;
					iminValue = *sp ;
					imaxValue = *sp ;
					for(i=0;i<numValues;i++) {
						iminValue = (*sp < iminValue ? *sp : iminValue) ;
						imaxValue = (*sp > imaxValue ? *sp : imaxValue) ;
						sp++ ;
					}
					range = imaxValue- iminValue ;
					for(i=0,cp=data,sp=(short *)data;i<numValues;i++) {
						ivalue = *sp ;
						value = (gNumGrays-1)*(ivalue-iminValue)/range ;
						ivalue = (int)value ;
						if (gTiff.photoMetricInterpretation == 0)
							ivalue = (gNumGrays-1) - ivalue ;
						*cp++ = gColors[ivalue].pixel ;
						sp++ ;
					}
					break ;
				case 4:
					ip=( int *)data ;
					iminValue = *ip ;
					imaxValue = *ip ;
					for(i=0;i<numValues;i++) {
						iminValue = (*ip < iminValue ? *ip : iminValue) ;
						imaxValue = (*ip > imaxValue ? *ip : imaxValue) ;
						ip++ ;
					}
					range = imaxValue- iminValue ;
					for(i=0,cp=data,ip=(int *)data;i<numValues;i++) {
						ivalue = *ip ;
						value = (gNumGrays-1)*(ivalue-iminValue)/range ;
						ivalue = (int)value ;
						if (gTiff.photoMetricInterpretation == 0)
							ivalue = (gNumGrays-1) - ivalue ;
						*cp++ = gColors[ivalue].pixel ;
						ip++ ;
					}
					break ;
			}
			break ;
		case 3:	/* IEEE floating point data	*/
			switch(samplesPerPixel) {
				case 1: /* Floating point	*/
					p1=(float *)data ;
					minValue = *p1 ;
					maxValue = *p1 ;
					for(i=0;i<numValues;i++) {
						minValue = (*p1 < minValue ? *p1 : minValue) ;
						maxValue = (*p1 > maxValue ? *p1 : maxValue) ;
						p1++ ;
					}
					range = maxValue- minValue ;
					for(i=0,p1=(float *)data,cp=data;i<numValues;i++) {
						value = (gNumGrays-1)*(*p1-minValue)/range ;
						ivalue = (int)value ;
						*cp = gColors[ivalue].pixel ;
						cp++ ;
						p1++ ;
					}
					break ;
				case 2:	/* ComplexFloatingPoint */
					/* let us display the real part */
					p1=(float *)data ;
					minValue = *p1 ;
					maxValue = *p1 ;
					for(i=0;i<width*height;i++) {
						minValue = (*p1 < minValue ? *p1 : minValue) ;
						maxValue = (*p1 > maxValue ? *p1 : maxValue) ;
						p1 += 2 ;
					}
					range = maxValue- minValue ;
					for(i=0,p1=(float *)data,cp=data;i<width*height;i++) {
						value = (gNumGrays-1)*(*p1-minValue)/range ;
						ivalue = (int)value ;
						*cp = gColors[ivalue].pixel ;
						cp++ ;
						p1 += 2 ;
					}
					break ;
			}
			break ;
	}
}
Boolean ProcessTiffDirEntry(FILE *fp,TIFFHDR *thdr,PDIRENTRY dirEntry){
	short	TagID,TagType ;
	int		numValues,value ;

    if (thdr->thByteOrder != BYTEORDER) {
    	TagID = ByteSwapShort(dirEntry->deTag) ;
    	TagType = ByteSwapShort(dirEntry->deType) ;
    	numValues = ByteSwapLong(dirEntry->deLength) ;
    	value = ByteSwapLong(dirEntry->deVal) ;
    	if (TagType == TIFFSHORT && thdr->thByteOrder == kMM && numValues == 1 ) 
			value >>= 16 ;
    }
	else {
		TagID = dirEntry->deTag ;
		TagType = dirEntry->deType ;
		numValues = dirEntry->deLength ;
		value = dirEntry->deVal ;
    	if (TagType == TIFFSHORT && thdr->thByteOrder == kMM && numValues == 1) 
			value >>= 16 ;
	}
		
	switch(TagID) {
		case TGNEWSUBFILETYPE:
		case TGOLDSUBFILETYPE:
			break ;
		case TGIMAGEWIDTH:
			gTiff.width = value ;
			break ;
		case TGIMAGELENGTH:
			gTiff.height = value ;
			break ;
		case TGBITSPERSAMPLE:
			gTiff.bitsPerSample = value ;
			break ;
		case TGCOMPRESSION:
			gTiff.compression = value ;
			break ;
		case TGPHOTOMETRICINTERPRETATION:
			gTiff.photoMetricInterpretation = value ;
			break ;
		case TGSTRIPOFFSETS:
			gTiff.numberOfStrips = numValues ;
			gTiff.stripOffset = new int[numValues] ; // (int *)malloc(numValues*sizeof(int)) ;
			*gTiff.stripOffset = value ;
			break ;
		case TGSAMPLESPERPIXEL:
			gTiff.samplesPerPixel = value ;
			break ;
		case TGROWSPERSTRIP:
			gTiff.rowPerStrip = value ;
			break ;
		case TGSTRIPBYTECOUNTS:
			gTiff.numberOfStrips = numValues ;
			gTiff.stripByteCount = new int[numValues] ; // (int *)malloc(numValues*sizeof(int)) ;
			*gTiff.stripByteCount = value ;
			break ;
		case TGXRESOLUTION:
			gTiff.tgxOffset = value ;
			break ;
		case TGYRESOLUTION:
			gTiff.tgyOffset = value ;
			break ;
		case TGRESOLUTIONUNIT:
			gTiff.resolutionUnit = value ;
			break ;
		case TGCOLORMAP:
			gTiff.numPaletteColors = numValues/3 ;	/* RGB */
			gTiff.paletteOffset = value ;
			break ;
		case TGSAMPLEFORMAT:
			gTiff.sampleFormat = value ;
			break ;
		case TGORIENTATION:
			/* We don't process	*/
			break ;
		default:
			break ;
	}
	return true ;
}

Boolean ReadTiffHeader(FILE *fp,TIFFHDR *header) {
	int numBytes ;
	
	numBytes = sizeof(TIFFHDR) ;
	if (fread(header,numBytes,1,fp) != 1) {
        	return false;
    	}
    	/* If we need to byteswap, do so ! */
   	if (header->thByteOrder != BYTEORDER) {
    		header->thVersion = ByteSwapShort(header->thVersion) ;
    		header->thIfdOffset = ByteSwapLong(header->thIfdOffset) ;
    	}
    	return true ;
}

unsigned short ByteSwapShort(unsigned short value) {
	unsigned short leftByte,rightByte ;
	unsigned short retValue ;
	
	leftByte = (value & 0xFF00) >> 8 ;
	rightByte = (value & 0x00FF) << 8 ;
	
	retValue = leftByte | rightByte ;	
	return retValue ;
}


unsigned long ByteSwapLong(unsigned long value) {
	unsigned long leftByte,rightByte,leftMiddleByte,rightMiddleByte ;
	unsigned long retValue ;
	
	leftByte = (value & 0xFF000000) >> 24 ;
	leftMiddleByte = (value & 0x00FF0000) >> 8 ;
	rightByte = (value & 0x000000FF) << 24 ;
	rightMiddleByte = (value & 0x0000FF00) << 8 ;
	
	retValue = leftByte | rightByte | leftMiddleByte | rightMiddleByte ;	
	return retValue ;
}

float ByteSwapFloat(unsigned long value) {
	unsigned long leftByte,rightByte,leftMiddleByte,rightMiddleByte,ivalue ;
	float retValue,*fp;
		
	leftByte = (value & 0xFF000000) >> 24 ;
	leftMiddleByte = (value & 0x00FF0000) >> 8 ;
	rightByte = (value & 0x000000FF) << 24 ;
	rightMiddleByte = (value & 0x0000FF00) << 8 ;
		
	ivalue = leftByte | rightByte | leftMiddleByte | rightMiddleByte ;	
	
	fp = (float *)&ivalue ;
	retValue = *fp ;
	return retValue ;
}

void MapLutToGrayValues(unsigned char *data,short width,short height){
	unsigned char *cp ;
	int i ;

	CreateInverseColorTable() ;
	for (i=0,cp=data; i < width*height;i++,cp++) {
		*cp = gInverseColorTable[*cp] ;
	}
} 

void	MapDataToFloat(TiffRecord *tiff,float *imagedata) {
	long width,height ;
	short sampleFormat,bitsPerSample,samplesPerPixel ;
	float *p1,*p2 ;
	unsigned char *cp,*data ;
	char	*ccp ;
	unsigned short *usp ;
	short	*sp ;
	unsigned int  *uip ;
	int *ip ;
	int i ;
	short bytesPerSample ;
	long	numValues ;
		
	width = tiff->width ;
	height = tiff->height ;
	sampleFormat = tiff->sampleFormat ;
	bitsPerSample = tiff->bitsPerSample ;
	
	bytesPerSample = bitsPerSample/8 ;
	samplesPerPixel = tiff->samplesPerPixel ;
	numValues = samplesPerPixel*width*height ;
	data = tiff->data ;

	if (tiff->photoMetricInterpretation == 2) {
		// Our data is RGB
		// We will convert the real values according to the
		// highest value of the red/green/blue component
		float red,green,blue ;
		cp = data ;
		p1 = imagedata ;
		for(i=0;i<width*height;i++) {
			red 	= *cp++ ;
			green 	= *cp++ ;
			blue 	= *cp++ ;			
			*p1++ = max(max(red,green),blue) ;
		}
		return ;
	}
	
	switch(sampleFormat) {
		case 1:	/* Unsigned integer data */
		case 4:	/* Undefined, treat as unsigned integer data	*/
			switch (bytesPerSample) {
				case 1:	/* Grayscale	*/
					cp = data ;
					p1 = imagedata ;
					for(i=0;i<numValues;i++) {
						*p1++ = *cp++ ;
					}
					break ;
				case 2:
					usp=(unsigned short *)data ;
					p1 = imagedata ;
					for(i=0;i<numValues;i++) {
						*p1++ = *usp++ ;
					}
					break ;
				case 4:
					uip=(unsigned int *)data ;
					p1 = imagedata ;
					for(i=0;i<numValues;i++) {
						*p1++ = *uip++ ;
					}
					break ;
			}
			break ;
		case 2:	/* Twos's complement signed integer data	*/
			switch (bytesPerSample) {
				case 1:	/* Grayscale	*/
					ccp=(char *)data ;
					p1 = imagedata ;
					for(i=0;i<numValues;i++) {
						*p1++ = *ccp++ ;
					}
					break ;
				case 2:
					sp=( short *)data ;
					p1 = imagedata ;
					for(i=0;i<numValues;i++) {
						*p1++ = *sp++ ;
					}
					break ;
				case 4:
					ip=( int *)data ;
					p1 = imagedata ;
					for(i=0;i<numValues;i++) {
						*p1++ = *ip++ ;
					}
					break ;
			}
			break ;
		case 3:	/* IEEE floating point data	*/
			switch(samplesPerPixel) {
				case 1: /* Floating point	*/
					p1=(float *)data ;
					p2 = imagedata ;
					for(i=0;i<numValues;i++) {
						*p2++ = *p1++ ;
					}
					break ;
				case 2:	/* ComplexFloatingPoint */
					/* let us display the real part */
					p1=(float *)data ;
					p2 = imagedata ;
					for(i=0;i<numValues;i++) {
						*p2++ = *p1++ ;
						p1++ ;
					}
					break ;
			}
			break ;
	}

}
